# Project Structure

We list here the contents of the artefact, and tie the foremost
concepts in this submission to the files which implement them. More
detailed explanations are provided as comments on each file.

## Overview
### `core/` directory

Contains various preliminary developments.

| File name   | Description                                                         |
| :---------- | :------------------------------------------------------------------ |
| axioms.v    | Axioms: propositional and functional extensionality.                |
| prelude.v   | Preliminaries.                                                      |
| pred.v      | Support for set-based notion; i.e., writing x \In X.                |
| ordtype.v   | Implementation of ordered types.                                    |
| interval.v  | Theory of sorted list surgery based on intervals.                   |
| seqperm.v   | A theory of permutations over non-equality types.                   |
| finmap.v    | Theory of finite maps.                                              |
| rtc.v       | Reflexive and transitive closures of a relation with finite domain. |

### `pcm/` directory

Contains the theory of PCMs and their morphisms. Implements the
different PCMs used throughout the paper, e.g. the `natmaps` used to
implement histories, or `mutex` PCM used for exclusive ownership.

| File name    | Description                                                                                                                                            |
| :----------- | :----------------------------------------------------------------------------------------------------------------------------------------------------- |
| pcm.v        | Implementation of partial commutative monoids (PCMs).                                                                                                  |
| morphism.v   | PCM morphisms and their separating relations.                                                                                                          |
| invertible.v | Invertible PCM morphisms and their separating relations (Section 4).                                                                                   |
| lift.v       | Lifting (total) commutative monoids to PCMs.                                                                                                           |
| unionmap.v   | Reference implementation of PCM of finite maps under disjoint union.                                                                                   |
| automap.v    | Some automation in the form of overloaded lemmas, for reasoning about finite maps under commutativity and associativity of disjoint union.             |
| heap.v       | Refined implementation of a heaps.                                                                                                                     |
| natmap.v     | Finite maps with nat domain. Includes reasoning about last key in the domain, fresh key unused in the domain, etc. The basis of the type of histories. |
| mutex.v      | A generalized ownership PCM and its instances.                                                                                                         |
| notation.v   | Notation for accessing the ghosts of morphed resources, and associated rewriting lemmas.                                                               |

### `fcsl/` directory

Contains the implementation of FCSL as a shallow embedding in Coq.

| File name  | Description                                                                                                                                                                       |
| :--------- | :-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------- |
| domain.v   | Basic domain theory: partially-ordered sets, complete lattices, CPOs, monotonicity, continuity, Knaster-Tarski and Kleene theorems. Provides the foundation for the soundness of the `fix` rule. |
| state.v    | Definition of subjective states, pairing, projection, and transposition.                                                                                                          |
| world.v    | Definition of resources, transitions, and some constructions over them: e.g., state product, transition coupling, resource restriction.                                           |
| tensor.v   | Definition of the tensor combination of resources.                                                                                                                                |
| schedule.v | Data type of paths (schedules). Paths select a position in a term where to reduce.                                                                                                |
| process.v  | Definition of action trees.                                                                                                                                                       |
| always.v   | Definition of the predicate transformers `always`, and `after` for adherence to a resource.                                                                                       |
| model.v    | Implementation of the denotational semantics of programs as sets of trees, the `vrf` predicate transformer, the inference rules and the lemmas for reasoning about Hoare triples. |

### `examples/` directory

The `examples` folders list the case studies.

| File name    | Description                                                                                                               |
| :----------- | :-------------------------------------------------------------------------------------------------------------------------|
| ticketlock.v | The ticketed lock example presented in this submission.                                                                   |
| CAP.v        | Comparison of morphisms and separating relations to Concurrent Abstract Predicates (Section 5).                           |
| priv.v       | Implementation of a resources of: private heaps Priv, shared heaps Shar, and a simple concurrent allocator.               |
| span.v       | A concurrent version of a DFS-style in-place computation of a graph spanning tree (full functional correctness verified). |
| treiber.v    | Implementation of Treiber's non-blocking stack. Implements quiescent clients.                                             |
| spinlock.v   | An implementation of a CSL-like spinlock resource, using PCM and resource morphisms.                                      |
| rwlock.v     | An implementation of a Reader-Writer lock resource, using PCM and resource morphisms.                                     |
| flatcomb.v   | Implementation of a flat combiner algorithm.                                                                              |
| stackcomb.v  | Instantiating flatcomb with stacks; also coupling the flat combiner with another resource, to simultaneously execute the transitions of that resource with the transitions for helping of the flat combiner.  |

### Project root directory (`./`)

Contains build system files, auxiliary scripts and docs

| File name            | Description                                                                                                                                |
| :------------------- | :----------------------------------------------------------------------------------------------------------------------------------------- |
| Makefile             | Makefile file for building the project using `coq_makefile` utility                                                                        |
| \_CoqProject         | List of the project's Coq files (`coq_makefile` generates Makefiles from it, also some IDE-like and auxiliary tools recognize \_CoqProject |
| dependency\_graph.sh | Script to generate a graphical representation of the library dependencies (see the script to check its dependencies)                       |
| options.v            | Global project's Coq options                                                                                                               |

## Code Navigation and Highlights

The names in the source code *mostly* correspond to the names in the
paper (modulo LaTeX-style representation of Greek letters and
similar changes).

### Some notation/name correspondence

| Notation on paper | Notation in source code | Name on paper | Name in source code | Note |
| :---------------: | :---------------------: | :------------ | :------------------ | :--- |
| • | `\+` | join | `join` | |
| ◁ | `\<` | self-framing | `zig` | |
| ▷ | `\>` | other-framing | `zag` | |
| ⇒ | `\->` | singleton finite map (from nats) | `nm_pts` | |
|                       | `(idx).` | | | This means we keep the projection after the dot (`.`) intact |
| ⊥ | | separating relation | `orth` | |
|                       | | partial functional | `funcP` | |
|                       | | other-fixity | `otherP` | |
|                       | | locality | `zagzigP` | |
|                       | | invariance of Sigma | `coherentP` | |
| [Γ]. {P} A {Q}@V | `{Γ}, ST [V] (P, Q)` | Hoare type | The type of programs inhabiting a resource V, returning values of type A, with pre-condition P and post-condition Q. Γ binds logical variables in P and Q. |

### Mapping Formal Concepts

We map some of the important concepts and formal structures to their
implementation constructors and notations.

| Name or Convention | Meaning | Note |
| :---------------- | :------- | :--- |
| `CohPx2` | Constructor of a *resource invariant* | a definition which uses this constructor ends in `coh` |
| `TransPx2` | Constructor of a *transition* | a definition which uses this constructor ends in `_tr` |
| `ResPx` | Constructor of a *resource* | |
| `State` | *state* constructor | |
| `[morphism of ...]` | PCM morphism constructor | |
| `[pcm of ...]` | PCM constructor | e.g. `[pcm of A * B]` is a product PCM constructor |

## Examples

### Ticket Lock

The file `examples/ticketlock.v` contains an implementation of ticket lock.
This example is the leading example in this submission and is
described in more detail in [DETAIL.md](DETAIL.md), and also in Appendix D in
the [submission supplement](../paper25-ext.pdf).

### Concurrent Abstract Predicates

The files `examples/CAP.v` contains a comparison of concurrent abstract
predicates to morphisms and separating relations.

### Other Examples

The artefact implements further case studies, which are not discussed
in the submission. We present them briefly.

#### CSL Spinlock

The file `examples/spinlock.v` presents the implementation of a CSL
spinlock, using PCM and resource morphisms to incrementally and
compositionally build the CSL resource. The file is structured as
follows.

- `CSLX`. The Module `CSLX` is the implementation of a resource for
  CSL-like exclusive heap-ownership transfer.

- `Spin`. The Module `Spin` is the implementation of *spinlocks*, and
its verification against a generic spec based on permissions to call
lock/unlock.

- `CSL`, The Module `SCSL` resource combines the modules above,
coupling `Spin.genlock/CSLX.open` and `Spin.genunlock/CSLX.close` to
implement the desired CSL-like spinlock for exclusive heap-ownership
transfer.

#### Reader-Writer lock

The file `examples/rwlock.v` presents a Reader-Writer lock
implementation, using PCM and resource morphisms. The file is
structured as follows.

- `Spin`. The Module `Spin` is the implementation of *spinlocks* and
verification against a generic specification in terms of histories.

- `RWX`. The Module `RWX` is the implementation of a resource for heap
ownership transfer and reader counting, specific to a *reader-writer*
locking discipline.

- `RW`. The Module `RW` is the final combination of `Spin` (writer) +
`RWX` + `Spin` (reader) into the module `RW` for reader-writer
locking, with the lifting and combination of `genlock` and `genunlock`
methods from `Spin` (writer) and `Spin` (reader) to a number of
methods for reader-writer locking, e.g. `wr_lock`, `wr_unlock`,
`prologue`, `epilogue`, and `read`.

#### Treiber Stack

- `examples/treiber.v` presents the re-implementation using morphisms
of the previous FCSL resource and specification for a Treiber
stack. Here, resource morphisms are used to combine the internal
Treiber resource with the concurrent allocator `Priv`, and also to
implement _quiescence_.

#### Flat Combiner

- `examples/flatcomb.v` presents the re-implementation of previous
FCSL developments which implement _helping_. The generic flat combiner
resource, FCI(X, M) takes two parameters X and M, where X is the ghost
state PCM—e.g. it can be instantiated with histories or natural
numbers—, and M specifies the helping interface, that is the methods
that helper threads execute. The exported API consists of the methods
fc_request and fc_collect, used for respectively requesting and
collecting help, and the fc_help method by which helper threads
perform their tasks, as specified by M. PCM and resource morphisms are
used to obtain a stateful flat combiner FC(X, M) as a combination of
FCI(X, M) and Priv.

- `examples/stackcomb.v` presents some clients of the flat
combiner. In the first one, we instantiate X in FC(X, M) with the PCM
of histories, and M with a complete stack API (i.e. `push` and `pop`
methods) to implement a flat combining Stack. The API of the latter,
consisting of the `fc_push` and `fc_pop` methods, provides a stack
whose push and pop methods are implemented via helping. In the second
client, we instantiate X in FC(X, M) with the PCM of natural numbers,
and again M with a stack API. The resulting resource, StackN, combines
the stack instance of the flat combiner with a simple counter method
where the collect_tr transition of the flat combiner is coupled with
the increment transition of the counter. As a result, the final client
(obtained through _morphing_) counts calls to `fc_push`. As the
`collect_tr` transition is the linearization point of the flat
combiner helping functionality, the coupling achieves synchronicity:
whenever the flat combiner helps to push or pop an element off the
stack, the counter, which is in a separate object, increments.

#### Concurrent Spanning Tree Algorithm

- `example/span.v`. This module presents the re-implementation of a
concurrent and recursive spanning-tree algorithm already verified in
FCSL. The Span resource keeps a pointer-based graph in its joint heap
ghost `σ_j` , and the self and other ghosts, `μs` and `μo`, are finite
sets of marked pointers which keep track of the nodes in the graph
that have already been visited by span. PCM and resource morphisms are
used for projecting the components of the state, as well as to
implement a quiescent client, which hides the helping sub-threads of
the spanning-tree graph from the parent thread’s environment.

