From Coq Require Import ssreflect ssrbool ssrfun.
From mathcomp Require Import ssrnat eqtype seq fintype tuple finfun.
From fcsl Require Import pred prelude pcm mutex unionmap automap.
From fcsl Require Import heap notation morphism state world model priv.
From fcsl Require Import options.
Set Warnings "-projection-no-head-constant".

(******************************************************)
(* Implementation of generic Flat combiner            *)
(* We can instantiate it with the PCM of histories,   *)
(* to get a stack implementation                      *)
(* Or with natural numbers PCM, to get an incrementor *)
(******************************************************)

(* flat combiner requires methods that work over private heaps only *)
(* so they should be specified by specs over heaps only *)

Structure flat_combineT := FlatCombineT {
  fc_orig : ptr; (* origin of the side-chanel array *)
  fc_lk : ptr; (* the lock for grabing the resource *)
  fc_n : nat; (* size of the array; i.e., number of admitted threads *)
  fc_pcm : pcm; (* resource pcm *)
  fc_inv : fc_pcm -> Pred heap; (* resource invariant *)
  fc_precise : forall g, precise (fc_inv g);
  fc_code : Type;  (* code for each method *)
  fc_inT : fc_code -> Type; (* input type of a coded method *)
  fc_outT : fc_code -> Type; (* output type of a coded method *)
  (* method c specification over result value and input/output ghosts *)
  fc_R : forall c, fc_inT c -> fc_outT c -> fc_pcm -> fc_pcm -> Prop;
  (* program of the coded method *)
  fc_interp : forall c (v : fc_inT c),
    {g}, STsep [priv]
      (fun i : st priv => fc_inv g (self i),
       fun r m => exists g',
         [/\ valid (g \+ g'), fc_inv (g \+ g') (self m) &
             fc_R v r g g']);
  (* the ghost to be returned by help is uniquely determined *)
  (* i.e. fc_R is a partial function *)
  fc_uniq : forall m (v : fc_inT m) (w : fc_outT m) g g1' g2',
    valid (g \+ g1') -> valid (g \+ g2') ->
    fc_R v w g g1' -> fc_R v w g g2' -> g1' = g2'}.

(* a structure for threads to make requests *)
Inductive req (T : flat_combineT) :=
  NoReq                             (* thread is not making a request *)
| Req (c : fc_code T) of fc_inT c   (* thread wants to run method c, with input iT c *)
| Ans (c : fc_code T) of fc_outT c. (* thread got back result of c *)

Section BigJoin.
Variables (X : pcm) (n : nat).

Definition sumg (g : {ffun 'I_n -> X}) := foldr (fun r t => r \+ t) Unit (fgraph g).

Lemma sumE (r1 r2 : {ffun 'I_n -> X}) : r1 =1 r2 -> sumg r1 = sumg r2.
Proof. by move=>E; congr sumg; apply/ffunP. Qed.

Lemma sum0R (r : {ffun 'I_n -> X}) x :
        sumg r = sumg [ffun z => if z == x then Unit else r z] \+ r x.
Proof.
rewrite /sumg !fgraph_codom /= !codomE /=.
have: x \in enum 'I_n by rewrite mem_enum.
have: uniq (enum 'I_n) by apply: enum_uniq.
elim: (enum 'I_n)=>[|y ys IH] //=.
rewrite (ffunE _ y) eq_sym inE.
case/andP; case: eqP=>[<- N _ _|_ N Un]; last first.
- by move/(IH Un)=>->; rewrite joinA.
rewrite joinAC unitL; congr (_ \+ foldr _ _ _).
apply/eq_in_map=>z; rewrite ffunE.
by case: eqP=>// ->; rewrite (negbTE N).
Qed.

Lemma sum0L (r : {ffun 'I_n -> X}) x :
        sumg r = r x \+ sumg [ffun z => if z == x then Unit else r z].
Proof. by rewrite joinC; apply: sum0R. Qed.

Lemma sumIR (r : {ffun 'I_n -> X}) x u :
        sumg r \+ u = sumg [ffun z => if z == x then r x \+ u else r z].
Proof.
rewrite (sum0R _ x) -joinA [in X in _ = X](sum0R _ x) (ffunE _ x) eq_refl.
by congr (_ \+ _); apply: sumE=>z; rewrite !ffunE; case: eqP.
Qed.

End BigJoin.

Section BigOps.
Variable T : flat_combineT.
Notation U := (fc_pcm T).
Notation n := (fc_n T).

Import bigop Monoid.

Canonical U_monoid := Law (@joinA U) (@unitL U) (@unitR U).
Canonical U_comoid := ComLaw (@joinC U).

Lemma sum0R' (r : {ffun 'I_n -> U}) (x : 'I_n) :
        \big [PCM.join/Unit]_(i < n) r i =
        \big [PCM.join/Unit]_(i < n | i != x) r i \+ r x.
Proof. by rewrite (bigD1 x) //= joinC. Qed.

Lemma sumIR' (r : {ffun 'I_n -> U}) (x : 'I_n) (u : U) :
        \big [PCM.join/Unit]_(i < n) r i \+ u =
        \big [PCM.join/Unit]_(i < n) (if i == x then r i \+ u else r i).
Proof.
rewrite (bigD1 x) //= [X in _ = X](bigD1 x) //= eq_refl joinAC.
by congr (_ \+ _); apply: eq_bigr=>i /negbTE=>->.
Qed.

End BigOps.

Module FC.
Section FlatCombine.
Variable F : flat_combineT.
Notation a := (fc_orig F).
Notation lk := (fc_lk F).
Notation n := (fc_n F).
Notation X := (fc_pcm F).
Notation I := (@fc_inv F).

Definition sync (r : {ffun 'I_n -> req F}) (g : {ffun 'I_n -> X}) :=
  forall t : 'I_n, if r t is (NoReq | Req _ _) then g t = Unit else True.

Definition U := [pcm of fset nat_ordType * mtx2 * X].
Structure T :=  layout {hpj : heap; ghosts : {ffun 'I_n -> X};
                        resps : {ffun 'I_n -> req F}}.
Implicit Type s : state U T.

(* self thread ids *)
Notation "\tid_s s" := ((self s).1.1) (at level 24).
(* other thread ids *)
Notation "\tid_o s" := ((other s).1.1) (at level 24).
(* total thread ids *)
Notation "\tid s" := (\tid_s s \+ \tid_o s) (at level 24).
(* self mutex *)
Notation "\mu_s s" := ((self s).1.2) (at level 24).
(* other mutex *)
Notation "\mu_o s" := ((other s).1.2) (at level 24).
(* total mutex *)
Notation "\mu s" := (\mu_s s \+ \mu_o s) (at level 24).
(* self user ghost *)
Notation "\gh_s s" := ((self s).2) (at level 24).
(* other user ghost *)
Notation "\gh_o s" := ((other s).2) (at level 24).
(* uncollected ghosts *)
Notation "\ghs s" := (ghosts (joint s)) (at level 24).
(* total user ghost *)
Notation "\gh s" := (\gh_s s \+ \gh_o s \+ sumg (\ghs s)) (at level 24).
(* uncollected responses *)
Notation "\rsp s" := (resps (joint s)) (at level 24).
(* protected heap *)
Notation "\hpj s" := (hpj (joint s)) (at level 24).
Notation "\hp s" := (lk :-> (\mu s == own) \+ \hpj s \+
                     updi a (fgraph (\rsp s))) (at level 0).

(* resource invariant *)

Program Definition fcoh :=
  CohPx2 (fun s =>
    [/\ sync (\rsp s) (\ghs s), forall x, x \in dom (\tid s) -> x < n,
        if \mu s == nown then I (\gh s) (\hpj s) else \hpj s = Unit &
        [/\ valid (\tid s), valid (\mu s), valid (\gh s) & valid (\hp s)]])
        (fun s => \hp s) _.
Next Obligation.
split=>[|p] s; last by rewrite !zgE !joinA.
by case=>??? [V1 V2 /validL V3 ->]; rewrite allvalidE /= V1 V2 V3.
Qed.

(*
Internal transitions for flat combiner:
- transition for requesting help
- transition for helping (only the locking thread can do it)
- transition for collecting the help results
*)

(* m : fc_code F - the method that tid request help with *)
(* tid : 'I_n    - the thread id *)
(* v : fc_inT m  - the arguments to the method m *)

Program Definition reqhelp_tr m tid (v : fc_inT m) :=
  TransPx (fun s x => [/\ valid (\tid s), nat_of_ord tid \in dom (\tid_s s) &
                          (\rsp s) tid = NoReq F] /\
   x = State (self s) (layout (\hpj s) (\ghs s)
                       [ffun z => if z == tid then Req v else (\rsp s) z])
             (other s)) fcoh _.
Next Obligation.
move=>x y; split=>[s [_ ->][_ ->]|[_ ->]|p|]; rewrite -?lT //=.
- rewrite !zgE joinA=>O [[V D ->]]; rewrite !zg_unlock; case=>->->.
  by rewrite domUn inE !(validLE3 V) D O.
case=>[[H1 H2 H3]] -> /= [H4 H5 H6][V1 V2 V3 V4]; split=>//.
- by move=>z; rewrite ffunE; case: ifP (H4 z)=>// /eqP ->; rewrite H3.
by rewrite (domeqVUn _ (domeqUP _ (size_fgraph _ _)) V4).
Qed.

(* m : fc_code F - the method that tid requests help with *)
(* tid : 'I_n    - the thread id to whom we're helping *)
(* w : fc_outT m - the result to install *)

Program Definition dohelp_tr m tid (w : fc_outT m) :=
  TransPx (fun s x => exists v g, [/\ \mu_s s = own, valid (\mu s),
    (\rsp s) tid = Req v, (\ghs s) tid = Unit, valid (\gh s \+ g) &
    fc_R v w (\gh s) g] /\
    x = State (self s) (layout (\hpj s)
                          [ffun z => if z == tid then g else (\ghs s) z]
                          [ffun z => if z == tid then Ans w else (\rsp s) z])
              (other s)) fcoh _.
Next Obligation.
move=>m tid w x y; split=>[s [v][g]|[v][g][_ ->]|p /lbO O
[v][g]|[v][g][[H V E ??? ->]]]; rewrite -?lT //=.
- by case=>[[_ _->->V H]]->[_][?][[_ _]][/inj_pair2<-] _ W /(fc_uniq V W H)->->.
- rewrite !zgE /=; case=>[[->]] /mxOx /mxE0 [->->] T G W H.
  by rewrite !zg_unlock; case=>->->; exists v, g; rewrite !joinA O in W H *.
rewrite H in V *; rewrite (mxOx V) /=.
case=>H1 H2 -> [-> _ V1 V2]; do !split=>//.
- by move=>t; rewrite !ffunE; case: ifP (H1 t).
- rewrite (_ : g = (\ghs x) tid \+ g) -?sumIR ?joinA //.
  by move: (H1 tid); rewrite E=>->; rewrite unitL.
by rewrite (domeqVUn _ (domeqUP _ (size_fgraph _ _)) V2).
Qed.

(* m : fc_code F - the method for which tid obtained help *)
(* tid : 'I_n    - the thread id which is collecting the help *)
(* w : fc_outT m - the result of the help *)

Program Definition collect_tr m tid (w : fc_outT m) g :=
  TransPx (fun s x => [/\ g = (\ghs s) tid, valid (\tid s),
    nat_of_ord tid \in dom (\tid_s s) & (\rsp s) tid = Ans w] /\
    x = State (\tid_s s, \mu_s s, \gh_s s \+ (\ghs s) tid)
              (layout (\hpj s)
                      [ffun z => if z == tid then Unit else (\ghs s) z]
                      [ffun z => if z == tid then NoReq F else (\rsp s) z])
              (other s)) fcoh _.
Next Obligation.
move=>m tid w g x y; split=>[s [_ ->][_ ->]|[_ ->]|[[p q] r]|]; rewrite -?lT //=.
- rewrite !zgE joinA=>O [[E H1 H2 H3]]; rewrite !zg_unlock=>[[->-> _ _ _ _]] _.
  by rewrite domUn inE !(validLE3 H1) H2 H3 -O joinAC.
case=>[[E H1 H2 H3]] -> /= [X1 X2 X3][V1 V2 V3 V4].
rewrite -joinA joinAC -!joinA - (sum0R (\ghs x) tid) !joinA.
split=>//; first by move=>t; move: (X1 t); rewrite !ffunE; case: ifP.
by rewrite (domeqVUn _ (domeqUP _ (size_fgraph _ _)) V4).
Qed.

(* External transitions for the flat combiner *)

Program Definition take_tr h :=
  TransPx (fun s x => [/\ \mu_s s = own, \hpj s = Unit, valid (\mu s),
                          valid (h \+ \hp s) & I (\gh s) h]  /\
    x = State (\tid_s s, nown, \gh_s s)
              (layout h (\ghs s) (\rsp s)) (other s)) fcoh _.
Next Obligation.
move=>h x y; split=>[s [_ ->][_ ->]|[_ ->]|[[p q] r]|]; rewrite -?lT //=.
- rewrite !zgE -!joinA /= =>O [[->]] -> /mxOx/mxE0 [->-> V H].
  by rewrite !zg_unlock; case=>->->; rewrite -O.
case=>[[->]] -> /mxOx E V H ->; rewrite {}E /= in V H *.
by case=>H1 H2 _ [-> _ -> _]; rewrite (validX V).
Qed.

Program Definition give_tr h :=
  TransPx (fun s x => [/\ \mu s = nown, valid (\mu s) & h = \hpj s] /\
    x = State (\tid_s s, own, \gh_s s)
              (layout Unit (\ghs s) (\rsp s)) (other s)) fcoh _.
Next Obligation.
move=>x y; split=>[s [_ ->][_ ->]|[_ ->]|[[p q] r]|]; rewrite -?lT //=.
- rewrite !zgE /==>O [[/mxE0]][->] /mxE0 [->->] _ ->.
  by rewrite !zg_unlock; case=>->->; rewrite O.
by case=>[[/mxE0]][-> E _] <- -> /= [???][??? V]; rewrite E (validX V).
Qed.

(***********************)
(* Resource definition *)
(***********************)

Inductive fcTr :=
  reqhelp_t (m : fc_code F) of 'I_n & fc_inT m |
  dohelp_t (m : fc_code F) of 'I_n & fc_outT m |
  collect_t (m : fc_code F) of 'I_n & fc_outT m & X |
  take_t of heap | give_t of heap | id_t P of monoP fcoh P.

Definition finterp tr :=
  match tr with
    reqhelp_t m t v => reqhelp_tr t v
  | dohelp_t m t w => dohelp_tr t w
  | collect_t m t w g => collect_tr t w g
  | take_t h => take_tr h
  | give_t h => give_tr h
  | id_t P pf => pre_tr (id_tr fcoh) pf
  end.

Definition fintern tr := if tr is (take_t _ | give_t _)  then false else true.

Program Canonical fc := Eval hnf in ResPx fcoh finterp fintern _.
Next Obligation.
case=>[m t v|m t w|m t w g|||P pf] // _ s x _;
[case=>_ ->|case=>v [g][_ ->]|case=>_ ->|case=>_ ->//];
by apply: domeqUn (domeqUP _ (size_fgraph _ _)).
Qed.

(*************)
(* Stability *)
(*************)

Definition requested m tid (v : fc_inT m) g s :=
  (\rsp s) tid = Req v /\ [pcm g <= \gh s] \/
  exists w g1, [/\ (\rsp s) tid = Ans w, [pcm g <= g1],
    [pcm g1 <= \gh s] & fc_R v w g1 ((\ghs s) tid)].

Lemma ogh g s1 s2 : osteps fc s1 s2 -> [pcm g <= \gh s1] -> [pcm g <= \gh s2].
Proof.
move: s1 s2; apply: iter_ind=>// s1 s2 [][m t v|m t w|m t w x|h|h|P pf];
last by [case=>_ /tr_inj ->]; rewrite !tr_unlock; try by [case=>_ [->->->]].
- case=>v [g2][[_ _ _ E _ _]][->->->] /pleq_trans H /=.
  by rewrite -[g2]unitL -E -sumIR joinA; apply: H.
by case=>_ [->->->]; rewrite -!joinA -sum0L.
Qed.

(* only the thread itself can make a request on its own behalf *)
Lemma onreq tid s1 s2 :
        osteps fc s1 s2 -> nat_of_ord tid \in dom (\tid_s s1) ->
        (\rsp s1) tid = NoReq F -> (\rsp s2) tid = NoReq F.
Proof.
move: s1 s2; apply: iterf_ind=>// s1 s2 [][m t v|m t w|m t w x|h|h|P pf];
last by [case=>_ /tr_inj ->]; rewrite !tr_unlock; try by [case=>_ [_ ->->]].
- case=>[[V D _]][_ ->->] D2; rewrite ffunE; case: ifP V D2=>// /eqP ->.
  by case: validUn=>// _ _ /(_ _ D) /negbTE ->.
- case=>v [g][[_ _ E _ _ _]][_ ->->] D2; rewrite ffunE.
  by case: ifP E D2=>// /eqP ->->.
by case=>[_][_ ->->] D2; rewrite ffunE; case: ifP.
Qed.

Lemma oreq m tid (v : fc_inT m) g s1 s2 :
        osteps fc s1 s2 -> nat_of_ord tid \in dom (\tid_s s1) ->
        requested tid v g s1 -> requested tid v g s2.
Proof.
move: s1 s2; apply: iterf_ind=>// s1 s2 [][n t x|n t x|n t x y|h|h|P pf];
rewrite /requested; last by [case=>_ /tr_inj ->]; rewrite !tr_unlock;
try by [case=>_ [->->->]].
- case=>[[_ _ E]][->->->] D; rewrite ffunE; case: ifP E D=>// /eqP ->->.
  by split=>//; case=>[|[?][?]][].
- case=>w [r][[/= _ _ H1 H2 _]]; rewrite (joinC (\gh_o _)) /= => H3.
  case=>->->-> D; rewrite !ffunE -{1 2}[r]unitL -H2 -sumIR joinA /=.
  split=>//; case: ifP D=>[/eqP ->|_] D.
  - rewrite H1; case=>[[[X]]|[?][?][//]]; subst n=>/inj_pair2 <-.
    by right; exists x, (\gh s1).
  case=>[[->] H|[wx][g1][-> ? H ?]]; [left | right; exists wx, g1];
  by split=>//; apply: pleq_trans H _.
case=>[[E H D _]][->->->] D2; split=>//=; rewrite !ffunE -!joinA -sum0L !joinA.
by case: ifP H D D2=>//= /eqP ->; case: validUn=>// _ _ H _ /H /negbTE ->.
Qed.

(* non-helpers cannot help, i.e., cannot serve requests *)
Lemma ohelp m tid (v : fc_inT m) s1 s2 :
        osteps fc s1 s2 -> \mu_s s1 = own ->
        (\rsp s1) tid = Req v -> (\rsp s2) tid = Req v.
Proof.
move: s1 s2; apply: iterf_ind=>// s1 s2 [][n t x|n t x|n t x y|h|h|P pf];
last by [case=>_ /tr_inj ->]; rewrite !tr_unlock; try by [case=>_ [_ ->->]];
[case=>[[/= _ _ E]][_ ->->]| by case=>w [g][[->]] /mxOx ->//|
 case=>[[/= ? _ _ E]][_ ->->]];
by rewrite ffunE; case: ifP E=>// /eqP ->->.
Qed.

(* non-helpers can shuffle the ghosts, but can't change the sum total *)
Lemma osum s1 s2 : osteps fc s1 s2 -> \mu_s s1 = own -> \gh s1 = \gh s2.
Proof.
move: s1 s2; apply: iterf_ind=>// s1 s2 [][n t x|n t x|n t x y|h|h|P pf];
last by [case=>_ /tr_inj ->]; rewrite !tr_unlock; try by case=>_ [->->->].
- by case=>v [g][[/= ->]] /mxOx ->.
by case=>_ [->->->]; rewrite -!joinA -sum0L.
Qed.

(***********)
(* Actions *)
(***********)

Program Definition req_help m tid (v : fc_inT m) : {x g}, STsep [fc]
  (fun s => [/\ self s = x, nat_of_ord tid \in dom (\tid_s s),
                (\rsp s) tid = NoReq F & [pcm g <= \gh s]],
   fun _ s => self s = x /\ requested tid v g s) :=
  Do (actSx (fun _ : unit => reqhelp_t tid v)).
Next Obligation. by split=>// ??? [][]. Qed.
Next Obligation.
move=>m tid v i [x][g][[H1 H2 H3 H4]] C; apply: vrf_atm=>// j O {C}.
move/coh_valid; rewrite allvalidE=>/andP [V _].
move/(onreq O H2): H3 H4=>H3 /(ogh O) H4; rewrite {i O}(oself O) in H1 H2.
case/andP: V; eexists tt, _. rewrite effE; split=>//= k O _; rewrite -(oself O).
by split=>//; apply: oreq O H2 _; left; rewrite ffunE eq_refl.
Qed.

(* dohelp needs to specify that the helping doesn't *)
(* remove the request condition of the helper thread *)
(* for other threads, that follows by stability *)
(* but for self thread, it has to be said directly *)
(* hence, we propagate requested m' t' v' g' *)

(* tid : 'I_n    -- the thread id of the thread *to be* helped *)
(* w : fc_outT m -- the result that the helper should write into the helping array *)
(* t' : 'I_n     -- can be any thread id, but in use we always instantiate it *)
(*                  with the helper thread's id *)

Program Definition do_help m tid (w : fc_outT m) : {x g1 g t' m' v' g'}, STsep [fc]
  (fun s => [/\ self s = x, \mu_s s = own, \gh s = g1, exists v,
                [/\ (\rsp s) tid = Req v, valid (g1 \+ g) & fc_R v w g1 g],
                nat_of_ord t' \in dom (\tid_s s) & @requested m' t' v' g' s],
   fun _ s => [/\ self s = x, g1 \+ g = \gh s & requested t' v' g' s]) :=
  Do (actSx (fun _ : unit => dohelp_t tid w)).
Next Obligation. by split=>// ??? [][]. Qed.
Next Obligation.
move=>m tid w i [x][_][g][t'][m'][v'][g'][[/= H1 H2 <- [v][H3 H4 H5] H6 H7]] C.
apply: vrf_atm=>// j O {}C; move/(ohelp O H2): H3 H7=>H3 /(oreq O H6) H7.
rewrite (osum O H2) in H4 H5 *; rewrite {i O}(oself O) in H1 H2 H6.
have [E V] : (\ghs j) tid = Unit /\ valid (\mu j).
- by case: C=>/(_ tid); rewrite H3 => -> _ _ [].
eexists tt, _. rewrite effE; split=>[|k O _]; first by exists v, g.
rewrite /= -{1}(oself O) -(osum O H2) /= -{2}[g]unitL -E -sumIR joinA; split=>//=.
apply: oreq O H6 _; rewrite /requested !ffunE -[g]unitL -E -sumIR joinA.
case: ifP H7=>[/eqP ->|_]; rewrite /requested.
- rewrite H3 E unitL=>[[|[?][?]]][//][X]; subst m'=>/inj_pair2 <-.
  by right; exists w, (\gh j).
case=>[[? X]|[q][r][?? X ?]]; [left | right; exists q, r];
by split=>//; apply: pleq_trans X _.
Qed.

Lemma ansM td : monoP fcoh (fun s => if (\rsp s) td is Ans _ _ then false else true).
Proof. by move=>??; rewrite !zgE. Qed.

(* try_collect tries to take the help from the tid element of the array *)
(* if the help has been placed there *)
(* we have it also return the ghost *)

Program Definition try_collect m tid : {x v g}, STsep [fc]
  (fun s => [/\ self s = x, nat_of_ord tid \in dom (\tid_s s) &
                requested tid v g s],
   fun (r : option (fc_outT m * X)) s => if r is Some (w, g') then
     exists g1, [/\ self s = (x.1, x.2 \+ g'), (\rsp s) tid = NoReq F,
       (\ghs s) tid = Unit, [pcm g <= g1], [pcm g1 <= \gh s] & fc_R v w g1 g']
     else self s = x /\ requested tid v g s) :=
  Do (actSx (fun r : option (fc_outT m * X) =>
        if r is Some (w, g') then collect_t tid w g' else id_t (@ansM tid))).
Next Obligation.
split=>[|[[]|]] // s s1 s2 [[r1 g1]|][[r2 g2]|] // _ [H1 _][H2 _];
first by [case: H1 H2=>-> _ _ -> [-> _ _][/inj_pair2 ->]];
by [case: H1 H2=>_ _ _ -> | case: H2 H1=>_ _ _ ->].
Qed.
Next Obligation.
move=>m tid i [x][v][g][[/= H1 H2 H3]] C; apply: vrf_atm=>// j O {C}.
move/coh_valid; rewrite allvalidE=>/andP [V _].
move/(oreq O H2): H3=>H3; rewrite {i O}(oself O) in H1 H2.
move: H3; rewrite /requested; case E: ((\rsp j) tid)=>[|mx vx|mx wx].
- by case=>[|[w][g1]][//].
- case=>[[[X]]|[w][g1][//]]; subst mx=>/inj_pair2 <- X.
  eexists None, _. rewrite effE; split; first by rewrite /= E.
  move=>k O _; rewrite -{1}(oself O); split=>//.
  by apply: oreq O H2 _; left; rewrite E.
case=>[[//]|[w][g1][[X]]]; subst mx=>/inj_pair2 <- ? X ?.
case/andP: V; eexists (Some (wx, _)), _. rewrite effE.
split=>// k O /(ocoh O) [/(_ tid)] S _ _ _.
exists g1; have R : (\rsp k) tid = NoReq F.
- by apply: onreq O H2 _; rewrite ffunE eq_refl.
rewrite -{1}(oself O) H1 /= -prod_eta; split=>//; first by rewrite R in S.
by apply: ogh O _; rewrite -joinA joinCA -!joinA -sum0L joinCA joinA.
Qed.

Lemma reqM tid x : monoP fcoh (fun s => (\rsp s) tid = x).
Proof. by move=>??; rewrite !zgE. Qed.

(* tid : 'I_n   - the index in the array from which to read the request *)
(* t : 'I_n - our thread id *)

Program Definition read_request tid : {x g1 m t v g}, STsep [fc]
  (fun s => [/\ self s = x, \mu_s s = own, \gh s = g1,
     nat_of_ord t \in dom (\tid_s s) & @requested m t v g s],
   fun r s => [/\ self s = x, \gh s = g1,
     (\rsp s) tid = if r is Req _ _ then r else (\rsp s) tid &
     requested t v g s]) :=
  Do (actSx (fun x => id_t (@reqM tid x))).
Next Obligation. by split=>// ????? _ [<-<-][<-<-]. Qed.
Next Obligation.
move=>tid i [x][_][m][t][v][g][[/= H1 H2 <- H3 H4]] C; apply: vrf_atm=>// j O {}C.
move/(oreq O H3): H4=>H4; rewrite (osum O H2); rewrite {i O}(oself O) in H1 H2 H3.
eexists ((\rsp j) tid), _. rewrite effE.
split=>// k O _; rewrite -{1}(oself O) (osum O H2).
by split=>//; [case E : ((\rsp j) tid)=>//; apply: ohelp E | apply: oreq H4].
Qed.

End FlatCombine.
End FC.

(********************************************)
(* flat combiner coupled with private heaps *)
(********************************************)

(* utilizes morphisms *)

Module FP.
Section FlatPriv.
Variable F : flat_combineT.
Notation a := (fc_orig F).
Notation lk := (fc_lk F).
Notation n := (fc_n F).
Notation X := (fc_pcm F).
Notation I := (@fc_inv F).
Notation fc := (FC.fc F).

Definition U := [pcm of Priv.U * FC.U F].
Definition T : Type := Priv.T * FC.T F.

(* switching notation *)

(* self thread ids *)
Notation "\tid_s s" := ((self s\2).1.1) (at level 24).
(* other thread ids *)
Notation "\tid_o s" := ((other s\2).1.1) (at level 24).
(* total thread ids *)
Notation "\tid s" := (\tid_s s \+ \tid_o s) (at level 24).
(* self mutex *)
Notation "\mu_s s" := ((self s\2).1.2) (at level 24).
(* other mutex *)
Notation "\mu_o s" := ((other s\2).1.2) (at level 24).
(* total mutex *)
Notation "\mu s" := (\mu_s s \+ \mu_o s) (at level 24).
(* self user ghost *)
Notation "\gh_s s" := ((self s\2).2) (at level 24).
(* other user ghost *)
Notation "\gh_o s" := ((other s\2).2) (at level 24).
(* uncollected ghosts *)
Notation "\ghs s" := (FC.ghosts (joint s\2)) (at level 24).
(* total user ghost *)
Notation "\gh s" := (\gh_s s \+ \gh_o s \+ sumg (\ghs s)) (at level 24).
(* uncollected responses *)
Notation "\rsp s" := (FC.resps (joint s\2)) (at level 24).
(* protected heap *)
Notation "\hpj s" := (FC.hpj (joint s\2)) (at level 24).

Notation "\hp s" := (lk :-> (\mu s == own) \+ \hpj s\2 \+
                        updi a (fgraph (\rsp s\2))) (at level 0).
(* private heap *)
Notation "\hp_s s" := (self s\1) (at level 24).
(* other heap *)
Notation "\hp_o s" := (other s\1) (at level 24).
(* allocator "stuff" *)
Notation "\hp_j s" := (joint s\1) (at level 24).

Inductive fpTr :=  in_flat of Tr fc | in_priv of Tr priv |
                   lock_t | unlock_t | id_t P of monoP [* coh priv, coh fc] P.

Program Definition lock_tr :=
  ex_tr [* coh priv, coh fc] (fun h =>[| interp priv (Priv.acq_t h),
                                         interp fc (FC.give_t F h)]) _.
Next Obligation. by move=>????? _ [_][[_ _ ->]] _ _ [_][[_ _ ->]]. Qed.

Program Definition unlock_tr :=
  ex_tr [* coh priv, coh fc] (fun h=> [| interp priv (Priv.rel_t h),
                                         interp fc (FC.take_t F h)]) _.
Next Obligation.
move=>????? _ [[h1]][E1 /validL V _][[_ _ _ _ J1]] _ _ [[h2]][E _ _][[_ _ _ _ J2]].
by move=>_ _; rewrite {}E1 in V E; apply: fc_precise J1 J2 E.
Qed.

Definition finterp tr :=
  match tr with
    in_flat t => [| id_tr (coh priv), interp fc t]
  | in_priv tp => [| interp priv tp, id_tr (coh fc)]
  | lock_t => lock_tr
  | unlock_t => unlock_tr
  | id_t P pf => pre_tr (id_tr [* coh priv, coh fc]) pf
  end.

Definition fintern t :=
  if t is in_priv t then internal priv t else
   if t is in_flat t then internal fc t else true.

Program Canonical fp := Eval hnf in ResPx [* coh priv, coh fc] finterp fintern _.
Next Obligation.
case=>[t|t|||P pf] //; try by [rewrite /tcoh /= (tcohE t)];
move=>I s s' [Cp Cf _] H; rewrite /flat /=; last by [case: H=>_ ->].
- by case: H=>-> H _; apply: domeqUn (iint _ _ H).
- by case: H=>H -> _; apply: domeqUn (iint _ _ H) _.
- case: H=>h [H1 H2 _]; apply: domeqfUn (Priv.acqF Cp H1) _.
  by case: H2=>[[_ _ ->]] ->; apply: domeqX.
case: H=>h [H1 H2 _]; apply: domeqUnf (Priv.relF Cp H1) _.
by rewrite /flat /=; case: H2=>[[_ -> _ _ _]] ->; apply: domeqX.
Qed.

(* injecting priv into fp *)

Lemma px_local : morph_localP (fun (_: unit) s sp => sp = s\1).
Proof. by exists (fun x : Ux _ => x.1); case=>????? ->; rewrite !prodE; eauto. Qed.

Lemma px_other : morph_otherP (fun (_:unit) s sp => sp = s\1).
Proof. by case; case=>???? _ _ _ _ ->-> /prod_inj []; rewrite !prodE. Qed.

Lemma px_func : morph_funcP (fun (_:unit) s sp => sp = s\1).
Proof. by case=>??? _ _ _ ->->. Qed.

Lemma px_back : morph_backwardP (fun (_:unit) s sp => sp = s\1).
Proof. by case; case. Qed.

Lemma px_int : morph_internalP (fun (_:unit) t _ t' => t' = in_priv t).
Proof. by move=>t _ _ _ I _ ->. Qed.

Lemma px_tfunc : morph_tfuncP (fun (_:unit) t _ t' => t' = in_priv t).
Proof. by case=>t _ _ _ _ ->->. Qed.

Lemma px_tloc : morph_tlocalP (fun (_:unit) t _ t' => t' = in_priv t).
Proof. by []. Qed.

Lemma px_coh : morph_cohP' (fun (_:unit) s sp => sp = s\1).
Proof. by case=>?? [C _ _] ->. Qed.

Definition pxMixin := MorphMixin px_local px_other px_func px_back
                                 px_int px_tfunc px_tloc px_coh.

Canonical px := Eval hnf in Morph pxMixin.

Lemma px_ostep : morph_ostepP px (fun => xPredT).
Proof.
case=>s1 s _ _ _ -> _ []; case=>[t|t|[h]|[h]|P pf][];
last by [move=>_ /tr_inj <-; eexists _]; rewrite !prodE=>H; exists s\1; split=>//;
try by [move/tr_inj: H <-].
- by apply: iosteps H.
- by apply: iosteps (Priv.acq_t _) H.
by apply: iosteps (Priv.rel_t _) H.
Qed.

Lemma px_forw : morph_forwardP px (fun => xPredT).
Proof.
move=>t [] ic xx jr I /coh_flat_valid V Cp E _ H; exists (in_priv t), tt, [@ jr, ic\2].
rewrite /Morph.phi /= in E *; subst xx; rewrite /= /prod_rel !sbeta.
by rewrite -(domeqVUnE (iint I Cp H) (domeq_refl _)).
Qed.

Definition pxsim : simulation px := Simulation2 px_ostep px_forw.

(* injecting fc into fp *)

Lemma fx_local : morph_localP (fun (_: unit) s sf => sf = s\2).
Proof. by exists (fun x : Ux _ => x.2); case=>????? ->; rewrite !prodE; eauto. Qed.

Lemma fx_other : morph_otherP (fun (_:unit) s sf => sf = s\2).
Proof. by case; case=>???? _ _ _ _ ->-> /prod_inj []; rewrite !prodE. Qed.

Lemma fx_func : morph_funcP (fun (_:unit) s sf => sf = s\2).
Proof. by case=>??? _ _ _ ->->. Qed.

Lemma fx_back : morph_backwardP (fun (_:unit) s sf => sf = s\2).
Proof. by case; case. Qed.

Lemma fx_int : morph_internalP (fun (_:unit) t _ t' => t' = in_flat t).
Proof. by move=>t _ _ _ I _ ->. Qed.

Lemma fx_tfunc : morph_tfuncP (fun (_:unit) t _ t' => t' = in_flat t).
Proof. by case=>t _ _ _ _ ->->. Qed.

Lemma fx_tloc : morph_tlocalP (fun (_:unit) t _ t' => t' = in_flat t).
Proof. by []. Qed.

Lemma fx_coh : morph_cohP' (fun (_:unit) s sp => sp = s\2).
Proof. by case=>?? [_ C _] ->. Qed.

Definition fxMixin := MorphMixin fx_local fx_other fx_func fx_back
                                 fx_int fx_tfunc fx_tloc fx_coh.

Canonical fx := Eval hnf in Morph fxMixin.

Lemma fx_ostep : morph_ostepP fx (fun => xPredT).
Proof.
case=>s1 s _ _ _ -> _ []; case=>[t|t|[h]|[h]|P pf][];
last by [move=>_ /tr_inj <-; eexists _]; rewrite !prodE=>? H;
exists s\2; split=>//; try by [move/tr_inj: H <-];
by apply: iosteps H.
Qed.

Lemma fx_forw : morph_forwardP fx (fun => xPredT).
Proof.
move=>t [] ic xx jr I /coh_flat_valid V Cf E _ H; exists (in_flat t), tt, [@ ic\1, jr].
rewrite /Morph.phi /= in E *; subst xx; rewrite /= /prod_rel !sbeta.
by rewrite -(domeqVUnE (domeq_refl _) (iint I Cf H)).
Qed.

Definition fxsim : simulation fx := Simulation2 fx_ostep fx_forw.

(* f-stability *)

Lemma ofpc s1 s2 : osteps fp s1 s2 -> coh fp s1 -> osteps fc s1\2 s2\2.
Proof.
by move=>O [???]; case/(@morph_osteps _ _ _ fx fxsim tt _ _ s1\2): O=>// x [->].
Qed.

Lemma mshp s1 s2 : msteps (fx tt) s1 s2 -> \hp_s s1 = \hp_s s2.
Proof.
move=>H; move: s1 s2 H (erefl tt); apply: iterf_ind=>// s1 s2.
by case=>[a _ I -> [->] //|a /iosteps /oselfR ->].
Qed.

Lemma mssf s1 s2 : msteps (px tt) s1 s2 -> coh fp s1 -> osteps fc s1\2 s2\2.
Proof.
move: s1 s2; apply: iterf_ind=>//; first by move=>???; apply: iter_trans.
move=>s1 s2 H C; split; first by apply: mcoh (iter1 H) C.
by case: H C=>[a _ I -> [_ ->] //|a /iosteps]; apply: ofpc.
Qed.

(* additional actions in the large resource *)

Lemma lockedM : monoP (coh fp) (fun s : st fp => \mu s = own).
Proof. by move=>p s; rewrite !prodE !zgE joinA. Qed.

Program Definition trylock : {h x m tid (v : fc_inT m) g}, STsep [fp]
  (fun s => [/\ \hp_s s = h, self s\2 = x, nat_of_ord tid \in dom (\tid_s s) &
                FC.requested tid v g s\2],
   fun b s => if b then exists hr,
      [/\ \hp_s s = hr \+ h, self s\2 = (x.1.1, own, x.2),
          I (\gh s) hr & FC.requested tid v g s\2]
     else [/\ \hp_s s = h, self s\2 = x & FC.requested tid v g s\2]) :=
  Do (actSx (fun b=>if b then lock_t else id_t lockedM)).
Next Obligation.
by split=>[|[]] // ??? [[|_ X [Y]]|[_ [Y _] X|]] //; case: X Y=>h [_][[->]].
Qed.
Next Obligation.
move=>i [_][_][m][tid][v][g][[/= <-<- H1 H2]] Ci; apply: vrf_atm=>// j O Cj.
move/(FC.oreq (ofpc O Ci) H1): H2=>H2; rewrite !(oselfR O) /= in H1 *.
have V : valid (\mu j) by case: Cj=>_ [_ _ _][].
case E: (\mu j) (V)=>[||[[]|]] // _; last first.
- eexists false, _. rewrite effE; split=>// k {}O _; rewrite -!(oselfR O) /=.
  by split=>//; apply: FC.oreq (ofpc O Cj) H1 H2.
eexists true, _; split=>[|k {}O Ck].
- rewrite effE; exists (\hpj j); apply: act2_safe=>//.
  - by do ?split=>//; apply: (validX (coh_flat_valid Cj)).
  (* hack until everything is ported to new notation *)
  move: (coh_flat_valid Cj); rewrite /flat /= /flat /= -!lprodX.
  by rewrite !unlock_priv /= => /validX ->.
exists (\hpj j); rewrite -{1}(oselfR O) -{1}(oselfR O) /= !prodE /=.
split=>//; move: (ofpc O Ck); rewrite sbeta; last first.
- by move/FC.oreq; apply.
- by move/FC.osum <-; case: Cj=>_ []; rewrite E.
(* hack until ported to new notation *)
by move=>_; rewrite -lprodX !unlock_priv -!prodEQ0.
Qed.

Program Definition unlock : {h x m tid (v : fc_inT m) g}, STsep [fp]
  (fun s => exists hr, [/\ \hp_s s = hr \+ h, x = self s\2, \mu_s s = own,
     I (\gh s) hr, nat_of_ord tid \in dom (\tid_s s) & FC.requested tid v g s\2],
   fun _ s => [/\ \hp_s s = h, self s\2 = (x.1.1, nown, x.2) &
                  FC.requested tid v g s\2]) :=
  Do (actSx (fun _ : unit => unlock_t)).
Next Obligation. by split=>// ??? [][]. Qed.
Next Obligation.
move=>i [h][x][m][tid][v][g][[/= hr]][H1 -> H2 H3 H4 H5] Ci.
apply: vrf_atm=>// j O Cj; rewrite (FC.osum (ofpc O Ci) H2) in H3.
move/(FC.oreq (ofpc O Ci) H4): H5=>H5; rewrite !(oselfR O) /= in H1 H2 H4 *.
have [V E] : valid (\mu j) /\ \hpj j = Unit.
- by case: Cj=>_ [_ _ H][_ V]; rewrite H2 in H V *; rewrite (mxOx V) in H *.
move: (coh_flat_valid Cj).
(* hack *)
rewrite /flat /= /flat /= -!lprodX !unlock_priv /= !prodE H1 => W.
eexists tt, _; split=>[|k {}O Ck].
- rewrite effE; exists hr; apply: act2_safe=>//.
  - by exists h; rewrite -!lprodX !unlock_priv /= !prodE H1 (validX W).
  - by do !split=>//; apply: validX W.
  by rewrite /flat /= /flat /= !unlock_priv (validX W).
rewrite -!(oselfR O) /= !prodE; split=>//.
by move: (ofpc O Ck); rewrite sbeta=>/FC.oreq; apply.
Qed.

(****************)
(* main program *)
(****************)

Variables (m : fc_code F) (tid : 'I_n) (v : fc_inT m).

(* first we frame the provided methods *)
Program Definition ffc_interp m (v : fc_inT m) : {h g}, STsep [priv]
   (fun i : st priv => exists hr, self i = hr \+ h /\ I g hr,
    fun r j => exists hr g', [/\ self j = hr \+ h, valid (g \+ g'),
                I (g \+ g') hr & fc_R v r g g']) :=
  Do (@fc_interp F m v).
Next Obligation.
move=>m' v' i [h][g][[hr]][H1 H2] C /=; apply: (frame3 priv) (H1) _.
move=>hr2 k Ck {i H1 C} _ H1 _.
rewrite -{}H1 in H2; apply: [gE g]=>//= w i [g'] H _ hr' j _ E -> _ {hr2}.
by exists hr', g'; rewrite E /= in H; case: H.
Qed.

Definition help_loop_tp := {h x g}, STsep [fp]
  (fun s => exists hr, [/\ \hp_s s = hr \+ h, self s\2 = x, \mu_s s = own,
     I (\gh s) hr, nat_of_ord tid \in dom (\tid_s s) & FC.requested tid v g s\2],
   fun (_ : unit) s => exists hr, [/\ \hp_s s = hr \+ h, self s\2 = x,
     FC.requested tid v g s\2 & I (\gh s) hr]).

Program Definition help_loop :=
  ffix (fun (help_loop : nat -> help_loop_tp) t =>
    Do (if @decP _ (t < n) idP is left pf return _ then (
          x <-- morph_app fx tt (FC.read_request (Ordinal pf));
          if x is Req m' v' then (
            w' <-- morph_app px tt (ffc_interp v');
            morph_app fx tt (FC.do_help (Ordinal pf) w');;
            help_loop t.+1)
          else help_loop t.+1)
        else ret tt)) 0.
Next Obligation.
move=>loop t i [h][_][g][[hr]][H1 <- H2 H3 H4 H5] C /=.
case: decP=>[pf|_]; last first.
- apply: vrf_ret=>// k O _; exists hr; rewrite -{1}(oselfR O) -{1}(oselfR O) /=.
  by split=>//; [apply: (FC.oreq (ofpc O C)) | rewrite -(FC.osum (ofpc O C))].
apply/step/(vrf_morph fxsim)=>[||_] //.
apply: [gE self i\2, \gh i, m, tid, v, g]=>// v' xx [X1 X2 R {}H5] _.
case=>{C} j Exx _ C /mxms M; rewrite {xx}Exx {M}(mshp M) in X1 X2 R H1 H5.
rewrite -{}X1 in H2 H4 *; rewrite -{i}X2 in H3.
case: v' R=>[|m' v'|??] R; try by [apply: [gE h, self j\2, g]=>//; exists hr].
apply/step/(vrf_morph pxsim)=>[||_] //.
apply: [gE h, \gh j]=>//; first by exists hr.
move=>{H1 H3} w k [hr2][g'][H1 V H3 G] _ [i] E _ Ci; rewrite {k}E in H1.
move/mxms/mssf/(_ C)=>{C} O; rewrite (FC.osum O H2) in V H3 G.
move/(FC.ohelp O H2): R=>R; move/(FC.oreq O H4): H5=>H5.
rewrite {j O}(oself O) in H2 H4 *; apply/step/(vrf_morph fxsim)=>[||_] //.
apply: [gE self i\2, \gh i, g', tid, m, v, g]=>//; first by split=>//; exists v'.
case=>{G R pf v' m' w H5 Ci} k [X1 X2 H5] _ [j] E _ C /mxms M.
rewrite {k}E (mshp M) in X1 X2 H1 H5; rewrite {}X2 in V H3.
by rewrite -{i M}X1 in H2 H4 *; apply: [gE h, self j\2, g]=>//=; exists hr2.
Qed.

Program Definition lock_help : {h x g}, STsep [fp]
  (fun s => [/\ \hp_s s = h, self s\2 = x, \mu_s s = nown,
                 nat_of_ord tid \in dom (\tid_s s) & FC.requested tid v g s\2],
   fun _ s => [/\ \hp_s s = h, self s\2 = x & FC.requested tid v g s\2]) :=
  Do (x <-- trylock; if x then help_loop;; unlock else ret tt).
Next Obligation.
move=>i [h][_][g][[/= H1 <- H2 H3 H4]] C; apply: [stepE h, _, _, tid, v, g]=>//=.
case=>j; last first.
- case=>{H2 H4 C}{}H1 E H4 C; rewrite -{i}E in H3 *; apply: vrf_ret=>// k O.
  by rewrite -!(oselfR O); split=>//; apply: FC.oreq (ofpc O C) H3 H4.
rewrite [_ i\2]prod_eta [(_ i\2).1]prod_eta /= {}H2 {H1 H4 C}.
case=>hr [H1 H2 H5 H4] C; case/prod_inj: H2 H3=>/prod_inj [<- H2 <-] H3 {i}.
apply: [stepE h, _, g]=>//=; first by exists hr.
case=>{H1 H4 H5 C} k [hr'][H1 E H4 H5] C; rewrite -{j}E in H2 H3 *.
by apply: [gE h, _, _, tid, v, g]=>//=; [exists hr' | case=>j [->->]].
Qed.

Definition lock_loop_tp := {h x g}, STsep [fp]
  (fun s => [/\ \hp_s s = h, self s\2 = x, \mu_s s = nown,
     nat_of_ord tid \in dom (\tid_s s) & FC.requested tid v g s\2],
   fun w s => exists g' g1, [/\ \hp_s s = h,
     self s\2 = (x.1.1, nown, x.2 \+ g'), [pcm g1 <= \gh s],
     [pcm g <= g1], fc_R v w g1 g' & (\rsp s) tid = NoReq F]).

Program Definition lock_loop :=
  ffix (fun (lock_loop : unit -> lock_loop_tp) xx =>
    Do (lock_help;;
        ow <-- morph_app fx tt (FC.try_collect m tid);
        if ow is Some (w, _) return _ then ret w else lock_loop tt)) tt.
Next Obligation.
move=>loop [i][h][_][g][[/= H1 <- H2 H3 H4]] _; apply: [stepE h, _, g]=>//=.
case=>{H1 H4} j [H1 E H4] _; rewrite -{i}E in H2 H3 *.
apply/step/(vrf_morph fxsim)=>[||_] //.
apply: [gE _, v, g]=>//; case=>[[w g'] k [g1]|k] {H4} /=; last first.
- case=>E H4 _ [i] X _ _ /mxms M; rewrite {k}X in E H4.
  by rewrite {M}(mshp M) in H1; rewrite -{j}E in H2 H3 *; apply: [gE h, _, g].
case=>X1 X2 _ X4 X5 X6 _ [i] X _ C /mxms M; rewrite {k}X in X1 X2 X5.
rewrite (mshp M) in H1; case/prod_inj: X1 H2 H3=>/prod_inj [<-<- X1] H2 H3.
apply: vrf_ret=>// k O _; exists g', g1; rewrite -X1 -H2 -2!prod_eta.
rewrite -(oselfR O) -{1}(oselfR O); split=>//;
by [apply: FC.ogh (ofpc O C) X5 | apply: FC.onreq (ofpc O C) H3 X2].
Qed.

Program Definition flat_combine : {g}, STsep [fp]
  (fun s => [/\ \hp_s s = Unit, self s\2 = (#(tid : nat), nown, Unit),
                [pcm g <= \gh s] & (\rsp s) tid = NoReq F],
   fun w s => exists g' g1, [/\ \hp_s s = Unit,
     self s\2 = (#(tid:nat), nown, g'), [pcm g1 <= \gh s],
     [pcm g <= g1], fc_R v w g1 g' & (\rsp s) tid = NoReq F]) :=
   Do (morph_app fx tt (FC.req_help tid v);;
       lock_loop).
Next Obligation.
move=>i [g][[H1 H2 H3 H4]] _; apply/step/(vrf_morph fxsim)=>// _.
apply: [gE _, g]=>//=; first by rewrite {2}H2 domPt inE /=.
case=>{H3 H4} j [E H3] _ [k] X _ _ /mxms M; rewrite {j}X in E H3.
rewrite {M}(mshp M) in H1; rewrite -{i}E in H2.
apply: [gE _, _, g]=>//=; first by rewrite H2 domPt inE /=.
by rewrite H1=>w i [g'][g1]; rewrite unitL; exists g', g1.
Qed.

End FlatPriv.
End FP.
