From Coq Require Import ssreflect ssrbool ssrfun.
From mathcomp Require Import ssrnat seq.
From fcsl Require Import axioms pred prelude pcm unionmap automap heap.
From fcsl Require Import morphism notation state world model.
From fcsl Require Import options.
Set Warnings "-projection-no-head-constant".

(******************************)
(******************************)
(* Resource for private state *)
(******************************)
(******************************)

(**************************************************************************)
(* Private state resource consists of self and other heap, and a joint    *)
(* state keeping the internal data of the allocator.                      *)
(*                                                                        *)
(* The joint state consists of two pointers: lck and f. Lck is the lock   *)
(* for taking the exlusive onwership of the allocator, and f is the       *)
(* pointer to a list of free pointers, stored as a seq. These are         *)
(* hardwired into the definition, i.e., lck = 1 and f = 2, where pointers *)
(* are nats.                                                              *)
(*                                                                        *)
(* In applications, the joint state should be kept abstract, as the       *)
(* exported methods do not specify how it is manipulated. The exported    *)
(* methods include: read (lookup), write (mutate), allocate and           *)
(* deallocate functions.                                                  *)
(*                                                                        *)
(* The pcm U = heap. The joint state is T = bool * seq ptr. The boolean b *)
(* stores the lock status of the allocator, and the sequence xs is the    *)
(* list of free pointers.                                                 *)
(*                                                                        *)
(* Notice that the joint state does not store the real joint heap of the  *)
(* allocator, as this heap can be computed out of T as                    *)
(*                                                                        *)
(*   flat s := lck :-> b \+ f :-> xs.                                     *)
(*                                                                        *)
(* In other words, we only keep the abstract state of the resource, and   *)
(* compute the real state by flattening.                                  *)
(**************************************************************************)

(* First, we go with a resource without allocation *)
(* This is ocassionally useful as a component of other resources *)

Module PrivCore.
Definition U := [pcm of heap].
Definition T := unit.
Implicit Type s : state U T.

Definition hp : U -> _ := idfun.
Canonical hp_morph := Eval hnf in [morphism of hp].

(**********************)
(* Resource invariant *)
(**********************)

Program Definition pc_coh :=
  CohPx2 (fun s => valid (hp.t s)) (fun s => hp.t s) _.
Next Obligation. by split=>*; rewrite ?allvE ?lO // !lTE'. Qed.

(* read-modify-write transition *)
(* we go with weak updates; i.e., no changes to the type of x here *)
Section RMWTrans.
Variable (A : Type) (x : ptr) (P : A -> Prop) (f : A -> A).

Program Definition rmw_tr :=
  TransPx2 (fun s s' =>
    exists v h, [/\ hp.s s = x :-> v \+ h, valid (hp.t s), P v &
    s' = State (x :-> f v \+ h) (idfun.j s) (idfun.o s)]) pc_coh _.
Next Obligation.
split=>[s|[v][h][???->]|p ?? O [v][h]|[v][h][E ?? ->]]; rewrite /= ?lE //=.
- case=>v [h][E1 /validL V _ ->][v2][h2][E _ _ ->] _.
  by rewrite {}E1 in V E; case/(hcancelV V): E=>-> _ ->.
- case=>-> V H [E1 E2 _ _ _]; exists v, (h \+ hp p).
  by rewrite !lE {}E1 {}E2 (validX V) -!joinA -O.
by rewrite E=>/validX ->.
Qed.

End RMWTrans.


(* Acquire/Release transitions: explicitly pass the heap to allocate *)
Section AcqRelTrans.
Variable h : heap.

Program Definition acq_tr :=
  TransPx2 (fun s s' => valid (h \+ hp.t s) /\
    s' = State (h \+ hp.s s) (idfun.j s) (idfun.o s)) pc_coh _.
Next Obligation.
split=>[s [_ ->][_ ->]|[_ ->]|p ?? O []|] //=; first by rewrite !lE.
- by rewrite !lE //= =>V [->->]; rewrite (validX V) // -joinA -O.
by case=>V -> _; rewrite !lE (validX V).
Qed.

Lemma acqF : footacqP pc_coh acq_tr h.
Proof. by move=>?? _ [_ ->]; rewrite /flat /= !lE !joinA. Qed.

Program Definition rel_tr :=
  TransPx2 (fun s s' => exists h', [/\ hp.s s = h \+ h', valid (hp.t s) &
    s' = State h' (idfun.j s) (idfun.o s)]) pc_coh _.
Next Obligation.
split=>[s|[?][_ _ ->]|p ?? O [k][]|]; rewrite ?lE //=.
- case=>h1 [E1 /validL V ->][h2][E _ ->].
  by rewrite {}E1 in V E; rewrite (joinxK V E).
- move=> -> V [E1 E2 _ _ _]; exists (k \+ p).
  by rewrite (validX V) !lE -joinA -O E1 E2.
by case=>k [-> _ ->]; rewrite !lE => /validX ->.
Qed.

Lemma relF : footrelP pc_coh rel_tr h.
Proof. by move=>s _ _ [k][E _ ->]; rewrite /flat /= !lE E !joinA. Qed.

End AcqRelTrans.


(********************************)
(* PrivCore resource definition *)
(********************************)

Inductive pcTr :=
  rmw_t A of ptr & A -> Prop & A -> A | acq_t of heap | rel_t of heap.

Definition pc_interp t :=
  match t with
    rmw_t A x P f => rmw_tr x P f
  | acq_t h => acq_tr h
  | rel_t h => rel_tr h
  end.

Definition pc_intern t := if t is rmw_t _ _ _ _ then true else false.

Program Canonical pc := Eval hnf in ResPx pc_coh pc_interp pc_intern _.
Next Obligation.
case=>[A x P f||] // _ s s' _; rewrite /flat /=.
by case=>v [h][-> V _ ->]; rewrite !lE domeqX.
Qed.

(***********)
(* Actions *)
(***********)

(* These are such commonly used functions that we give large footprint *)
(* specs, to avoid constantly invoking the framing rule. *)
Program Definition rmw A B x f (g : A -> B) : {(v : A) h}, STsep [pc]
  (fun s => hp.s s = x :-> v \+ h,
   fun r s => r = g v /\ hp.s s = x :-> f v \+ h) :=
  Do (actSx (fun r => rmw_t x (fun w => r = g w) f)).
Next Obligation.
split=>// ????? C H [v][h][E _ -> _].
by case: H E=>w [k][-> /validL V -> _] /(hcancelV V) [->].
Qed.
Next Obligation.
move=>A B x f g i [v][h][/= E1 _]; apply: vrf_atm=>// j O C.
rewrite (osf O) in E1; eexists (g v), _; split=>//.
- by rewrite effE; exists v, h.
by move=>m {} O _; rewrite -(osf O) !lE.
Qed.

Definition read A x := @rmw A A x id id.

(* write inherits equation tt = tt in the post; we remove it here *)
Program Definition write A x (v : A) : {h}, STsep [pc]
  (fun s => exists w : A, hp.s s = x :-> w \+ h,
   fun _ s => hp.s s = x :-> v \+ h) :=
  Do (@rmw A unit x (fun => v) (fun => tt)).
Next Obligation.
by move=>A x v i [h][][w E _]; apply: [gE w, h]=>// [[m]][].
Qed.

Module Exports.
Notation pc := pc.
Notation pc_coh := pc_coh.
Canonical pc_coh.
Canonical pc.
End Exports.

End PrivCore.

Export PrivCore.Exports.

(*************************************************************************)
(* We also want allocation. It is possible to implement allocator        *)
(* separately, and then just take a product with PrivCore, this is       *)
(* done in privalt.v. However, there are cases where it is not optimal,  *)
(* in that the PCM type of the combination is not equal to the PCM type  *)
(* of PrivCore.                                                          *)
(*                                                                       *)
(* As I expect to often interchange the two in experiments, it is        *)
(* convenient to implement Priv with an Allocator monolithically, and    *)
(* have the same PCM type in PrivCore and Priv.                          *)
(*                                                                       *)
(*************************************************************************)

Module Priv.
Definition U := heapPCM.
Structure T := ghosts {lk : bool; sq : seq ptr}.
Implicit Type s : state U T.

Definition hp : U -> _ := idfun.
Canonical hp_morph := Eval hnf in [morphism of hp].

Local Definition l := ptr_nat 1.  (* the lock of the allocator *)
Local Definition f := ptr_nat 2.  (* the pointer to the free list *)

Fixpoint flist' xs := if xs is x :: xs' then x :-> tt \+ flist' xs' else Unit.

Definition flist xs := f :-> xs \+ flist' xs.

Definition layout b xs := if b then l :-> true else l :-> false \+ flist xs.

(**********************)
(* Resource invariant *)
(**********************)

Program Definition priv_coh := Eval hnf in
  CohPx2 (fun s => valid (hp.t s \+ layout (lk.j s) (sq.j s)))
         (fun s => hp.t s \+ layout (lk.j s) (sq.j s)) _.
Next Obligation. by split=>[s V|*]; rewrite ?(lE,joinA) // allvE (validX V). Qed.

Section CorePrivTrans.
Variable (A : Type) (x : ptr) (P : A -> Prop) (f : A -> A).

(* read-modify-write transition *)
(* we go with weak updates; i.e., no changes to the type of x here *)

Program Definition rmw_tr :=
  TransPx2 (fun s s' =>
    exists v h, [/\ hp.s s = x :-> v \+ h, valid (hp.t s), P v &
    s' = State (x :-> f v \+ h) (idfun.j s) (idfun.o s)]) priv_coh _.
Next Obligation.
split=>[s|[v][h][???->]|p ?? O [v][h]|[v][h][E ?? ->]]; rewrite /= ?lE //=.
- case=>v [h][E1 /validL V _ ->][v2][h2][E _ _ ->] _.
  by rewrite {}E1 in V E; case/(hcancelV V): E=>-> _ ->.
- case=>-> V H [E1 E2 _ _ _]; exists v, (h \+ hp p).
  by rewrite !lE /= {}E1 {}E2 (validX V) -!joinA -O.
by rewrite E=>/validX ->.
Qed.

(* strong writes are needed to implement alloc *)
Program Definition stwr_tr (v : A) :=
  TransPx2 (fun s s' =>
    exists B (w : B) h, [/\ hp.s s = x :-> w \+ h, valid (hp.t s) &
    s' = State (x :-> v \+ h) (idfun.j s) (idfun.o s)]) priv_coh _.
Next Obligation.
split=>[s|[B][w][h][_ _ ->]|p ?? O [B][w][h][]|[B][w][h]]; rewrite ?lE //=.
- case=>B [w][h][E1 /validL V ->][B'][w'][h'][E _ ->].
  by rewrite {}E1 in V E; case/(cancel V): E=>_ _ ->.
- move=> -> V [E1 E2 _ _ _]; exists B, w, (h \+ p).
  by rewrite lE E1 E2 -O !joinA (validX V).
by case=>E _ ->; rewrite !lE E => /validX ->.
Qed.

End CorePrivTrans.


(* Acquire/Release transitions: explicitly pass the heap to allocate *)
Section AcqRelTrans.
Variable h : heap.

Program Definition acq_tr :=
  TransPx2 (fun s s' => valid (h \+ hp.t s \+ layout (lk.j s) (sq.j s)) /\
    s' = State (h \+ hp.s s) (idfun.j s) (idfun.o s)) priv_coh _.
Next Obligation.
split=>[s [_ ->][_ ->]|[_ ->]|p ?? O []|]; rewrite ?lE //=.
- by move=>V [->->]; rewrite !(validX V) // -!joinA -O.
by case=>V -> _; rewrite /= !lE (validX V).
Qed.

Lemma acqF : footacqP priv_coh acq_tr h.
Proof. by move=>?? _ [_ ->]; rewrite /flat /= !lE !joinA. Qed.

Program Definition rel_tr :=
  TransPx2 (fun s s' => exists h', [/\ hp.s s = h \+ h', valid (hp.t s) &
    s' = State h' (idfun.j s) (idfun.o s)]) priv_coh _.
Next Obligation.
split=>[s|[?][_ _ ->]|p ?? O [k][]|]; rewrite ?lE //=.
- case=>h1 [E1 /validL V ->][h2][E _ ->].
  by rewrite {}E1 in V E; rewrite (joinxK V E).
- move => -> V [E1 E2 _ _ _]; exists (k \+ p).
  by rewrite (validX V) !lE -joinA -O E1 E2.
by case=>k [-> _ ->]; rewrite !lE => /validX ->.
Qed.

Lemma relF : footrelP priv_coh rel_tr h.
Proof. by move=>s _ _ [k][E _ ->]; rewrite /flat /= !lE E !joinA. Qed.

End AcqRelTrans.


(* Transitions for Locking/Unlocking the allocator *)
Program Definition lock_tr :=
  TransPx2 (fun s s' => ~~ lk.j s /\
    s' = State (flist (sq.j s) \+ hp.s s)
               (ghosts true (sq.j s)) (idfun.o s)) priv_coh _.
Next Obligation.
split=>[s [_ ->][_ ->]|[_ ->]|p ?? O []|]; rewrite ?lE //=.
- by move => -> [->->]; rewrite -O -!joinA.
by case=>N ->; rewrite !lE (negbTE N) => /validX ->.
Qed.

Program Definition unlock_tr :=
  TransPx2 (fun s s' => exists xs h,
    [/\ hp.s s = flist xs \+ h, valid (hp.t s \+ layout (lk.j s) (sq.j s)) &
    s' = State h (ghosts false xs) (idfun.o s)]) priv_coh _.
Next Obligation.
split=>[s|[xs][h][_ _ ->]|p ?? O [xs][h][]|[xs][h]]; rewrite ?lE //=.
- case=>xs1 [h1][E /validL/validL V ->][xs2][h2][E' _ ->].
  move: E'; rewrite E /flist -!joinA in V *.
  by case/(hcancelV V)=><- {}V /(joinxK V) ->.
- move => -> V [E1 E2 _ _ _]; exists xs, (h \+ p).
  by rewrite !lE // E1 E2 -!joinA -O (validX V).
by case=>E _ ->; rewrite !lE /= E /layout; case: ifP=>_; apply: validX.
Qed.

(****************************)
(* Priv resource definition *)
(****************************)

Inductive privTr := rmw_t A of ptr & A -> Prop & A -> A |
                    stwr_t A of ptr & A | lock_t | unlock_t |
                    acq_t of heap | rel_t of heap | id_t P of monoP priv_coh P.

Definition priv_interp t :=
 match t with
    rmw_t A x P f => rmw_tr x P f
  | stwr_t A x v => stwr_tr x v
  | lock_t => lock_tr
  | unlock_t => unlock_tr
  | acq_t h => acq_tr h
  | rel_t h => rel_tr h
  | id_t _ pf => pre_tr (id_tr priv_coh) pf
  end.

Definition priv_intern t := if t is (acq_t _ | rel_t _) then false else true.

Program Canonical priv := Eval hnf in ResPx priv_coh priv_interp priv_intern _.
Next Obligation.
case=>[A x P f|A p v|||||P pf] // _ s s' _; rewrite /flat /=; last by case=>_ ->.
- by case=>v [h][-> _ _ ->]; rewrite !lE domeqX.
- by case=>B [w][h][-> _ ->]; rewrite !lE domeqX.
- by case=>N ->; rewrite !lE /layout (negbTE N) domeqX.
case=>xs [h][-> V ->]; rewrite !lE /layout in V *.
by case: ifP V=>[_|] _; [apply: domeqX | rewrite invalidX].
Qed.


(***********)
(* Actions *)
(***********)

Program Definition rmw A B x f (g : A -> B) : {(v : A) h}, STsep [priv]
  (fun s => hp.s s = x :-> v \+ h,
   fun r s => r = g v /\ hp.s s = x :-> f v \+ h) :=
  Do (actSx (fun r => rmw_t x (fun w => r = g w) f)).
Next Obligation.
split=>// ????? C H [v][h][E _ -> _].
by case: H E=>w [k][-> /validL V -> _] /(hcancelV V) [->].
Qed.
Next Obligation.
move=>A B x f g =>i [v][h][/= E1 _].
apply: vrf_atm=>// j O C; rewrite (osf O) in E1.
eexists (g v), _; rewrite effE; split=>[|m {}O].
- by exists v, h; rewrite (validX C).
by rewrite -(osf O) !lE.
Qed.

Definition read A x := @rmw A A x id id.

(* write inherits the equation tt = tt in the post; we simplify here *)
Program Definition write A x (v : A) : {h}, STsep [priv]
  (fun s => exists w : A, hp.s s = x :-> w \+ h,
   fun _ s => hp.s s = x :-> v \+ h) :=
  Do (@rmw A unit x (fun => v) (fun => tt)).
Next Obligation.
by move=>A x v i [h][[/= w E] C]; apply: [gE w, h]=>//= [[m]][].
Qed.

(* strong writes *)
Program Definition stwr A x (v : A) : {h}, STsep [priv]
  (fun s => exists B (w : B), hp.s s = x :-> w \+ h,
   fun (_ : unit) s => hp.s s = x :-> v \+ h) :=
  Do (actSx (fun => @stwr_t A x v)).
Next Obligation. by split=>// ??? [[]]. Qed.
Next Obligation.
move=>A x v i [h][[/= B]][w E1 _]; apply: vrf_atm=>// j O C.
rewrite (osf O) in E1; eexists tt, _; rewrite effE; split=>[|m {}O] /=.
- by exists B, w, h; rewrite (validX C).
by rewrite -(osf O) !lE.
Qed.

(* trylock *)

Lemma failM : monoP priv_coh (fun s => lk.j s).
Proof. by move=>??; rewrite !lE. Qed.

Program Definition trylock : {h}, STsep [priv]
  (fun s => hp.s s = h,
   fun r s => if r then exists xs, hp.s s = flist xs \+ h else hp.s s = h) :=
  Do (actSx (fun b => if b then lock_t else id_t failM)).
Next Obligation.
by split=>[|[] //] ??? [][] // _ [N _][]; rewrite ?N ?(negbTE N).
Qed.
Next Obligation.
move=>i [h][/= E _]; apply: vrf_atm=>// j O C.
rewrite (osf O) in E; case X : (lk.j j).
- by eexists false, _; rewrite effE; split=>// m /osf <-.
move/negbT: X; eexists true, _; rewrite effE; split=>// m /osf <-.
by rewrite !lE E; eexists _.
Qed.

Program Definition unlock : {xs h}, STsep [priv]
  (fun s => hp.s s = flist xs \+ h, fun (_ : unit) s => hp.s s = h) :=
  Do (actSx (fun => unlock_t)).
Next Obligation. by split=>// ??? [][]. Qed.
Next Obligation.
move=>i [xs][h][/= E _]; apply: vrf_atm=>// j O C.
rewrite (osf O) in E; eexists tt, _; rewrite effE; split; first by exists xs, h.
by move=>m /osf <-; rewrite !lE.
Qed.

(*************************************)
(* alloc and dealloc are implemented *)
(*************************************)

Definition alloc_tp A (v : A) := forall _ : unit,
  {h}, STsep [priv] (fun s => hp.s s = h, fun x s => hp.s s = x :-> v \+ h).

Program Definition alloc A (v : A) :=
  ffix (fun (loop : alloc_tp v) xx =>
    Do (b <-- trylock;
        if b then
          xs <-- read (seq ptr) f;
          if xs is x :: xs' then (
            write f xs';;
            unlock;;
            stwr x v;;
            ret x)
          else (
            unlock;;
            loop tt)
        else loop tt)) tt.
Next Obligation.
move=>A v loop [i][h][/= E C]; apply: [stepE h]=>//.
case=>{i E C} m; last by move=>E C; apply: [gE h].
case=>xs Ehm C; apply: [stepE _, _]=>//=; first by rewrite Ehm -joinA.
move=> _ {}m [->] {}Ehm {}C; case: xs Ehm=>[|x xs] Ehm.
- apply: [stepE [::], _]=>//=; first by rewrite Ehm /= -joinA.
  by case=>i Ehi Ci; apply: [gE h].
apply: [stepE _]=>//=; first by eauto.
case{m Ehm C}=>m Ehm C; apply: [stepE xs, _]=>//=.
- by rewrite Ehm -!joinA (joinCA (x:->tt)); eauto.
case{m C Ehm}=> m /= Ehm C; apply: [stepE _]=>//=; first by eauto.
by case{m C Ehm}=>m Ehm C; apply: vrf_ret=>// k O; rewrite -(osf O).
Qed.

Definition dealloc_tp x := forall _ : unit,
  {h}, STsep [priv] (fun s => exists A (v : A), hp.s s = x :-> v \+ h,
                     fun (_ : unit) s => hp.s s = h).

Program Definition dealloc x :=
  ffix (fun (loop : dealloc_tp x) xx =>
    Do (b <-- trylock;
        if b then (
          stwr x tt;;
          xs <-- read (seq ptr) f;
          write f (x :: xs);;
          unlock)
        else loop tt)) tt.
Next Obligation.
move=>x loop [i][h][][A][v Ehi C]; apply: [stepE x :-> v \+ h]=>//=.
case; last by move=>{Ehi C} i Ehi C; apply: [gE h]=>//=; eauto.
move=>{Ehi C} i [xs] Ehi C; apply: [stepE _]=>//=.
- by rewrite Ehi /= joinCA; eauto.
case{i Ehi C}=>i Ehi C; apply: [stepE _, _]=>//=.
- by rewrite /= Ehi -joinA joinCA.
move{i Ehi C}=> _ i [->] Ehi C; apply: [stepE _]=>//=; first by eauto.
case{i Ehi C}=>i Ehi C; apply: [gE x::xs, _]=>//=.
- by rewrite Ehi /= -!joinA.
by case.
Qed.

Module Exports.
Notation priv := priv.
Notation priv_coh := priv_coh.
Canonical priv_coh.
Canonical priv.
End Exports.

End Priv.

Export Priv.Exports.


(* We also need the Shar resource *)

Module Shar.
Section Shar.
Variables (X : pcm) (inv : X -> heap -> Prop).

Definition U := X.                              (* user ghost *)
Structure T :=  layout {hp : heap; here: bool}. (* shared heap, here flag *)
Implicit Type s : state U T.

Definition gh : U -> U := idfun.
Canonical gh_morph := Eval hnf in [morphism of gh].

Program Definition scoh :=
  CohPx2 (fun s => [/\ valid (gh.t s), valid (hp.j s) &
           if here.j s then inv (gh.t s) (hp.j s) else hp.j s = Unit])
         (fun s => hp.j s) _.
Next Obligation.
by split=>[s [V1 V2]|p s ?]; rewrite ?(allvE,V1,V2) ?lE //= -!joinA.
Qed.

(* open transition; give the shared heap away *)
Program Definition open_tr h :=
  TransPx2 (fun s x => [/\ h = hp.j s & here.j s] /\
   x = State (gh.s s) (layout Unit false) (idfun.o s)) scoh _.
Next Obligation.
move=>x y; split=>[s [_ ->][_ ->]|[_ ->]|p ?? O [[]]|[_ ->][]] //=;
by rewrite !lE //= => ->-> [->->]; rewrite -O.
Qed.

(* close transition; get the shared heap back and fix the self auxiliary *)

(*************************************************************************************)
(* Using the existential over r here makes it necessary to use precision             *)
(* later on. It's never a good idea to use existentials in the definition of actions *)
(* Instead, one should make r a parameter to the transition                          *)
(*************************************************************************************)

Program Definition close_tr f (pf : local f) h :=
  TransPx2 (fun s x => exists r, [/\ ~~ here.j s, hp.j s = Unit, valid h,
    valid (gh.t s), f (gh.s s) (gh.o s) = Some r & inv (r \+ gh.o s) h] /\
    x = State r (layout h true) (idfun.o s)) scoh _.
Next Obligation.
move=>f pf h x y; split=>[s|[r][_ ->]|p ?? O [r][[]]|]; rewrite ?lE //=.
- by case=>r1 [[_ _ _ _ -> _] ->][r2][[_ _ _ _ [->] _] ->].
- rewrite !lE //= => -> E V1 V2 F J [E1 E2 _ _ _]; exists (r \+ p).
  by case/(pf _ _ _ _ V2): F=>_ ->; rewrite !lE E1 E2 -!joinA -O.
by case=>r [[_ _ V1 V2 F J] ->]; rewrite !lE; move/localV: F=>->.
Qed.

(* assembled resource *)

Inductive Tr :=
  open_t of heap | close_t f of @local X f & heap | id_t P of monoP scoh P.

Definition sinterp t :=
  match t with
    open_t h => open_tr h
  | close_t f pf h => close_tr pf h
  | id_t P pf => pre_tr (id_tr scoh) pf
  end.

Definition sintern t := if t is id_t _ _ then true else false.

End Shar.

Module Exports.
Program Canonical shar X inv :=
  Eval hnf in ResPx (@scoh X inv) (@sinterp X inv) (@sintern X inv) _.
Next Obligation. by move=>X inv [||??] // _ s x _ [_ ->]. Qed.

End Exports.
End Shar.

Export Shar.Exports.
