From Coq Require Import ssreflect ssrbool ssrfun.
From mathcomp Require Import ssrnat eqtype seq.
From fcsl Require Import axioms pred prelude pcm automap heap natmap.
From fcsl Require Import morphism notation lift mutex.
From fcsl Require Import state world schedule process always model.
From fcsl Require Import tensor.
From fcsl Require Import options.

(**************************************************************************)
(* This file contains the following development. Please see the           *)
(* associated paper for a more detailed description.                      *)
(*                                                                        *)
(* 1. Module Spin: implementation of Spin locks and verification against  *)
(*    a generic spec in terms of histories.                               *)
(*                                                                        *)
(* 2. Module CSLX: implementation of a resource for heap ownership        *)
(*    transfer specific to exclusive locking.                             *)
(*                                                                        *)
(* 3. Module CSL: combination of Spin + CSLX into the module CSL for      *)
(*    exclusive locking, with the lifting of the genlock and genunlock    *)
(*    methods from Spin to exlock and exunlock in CSL.                    *)
(*                                                                        *)
(* 4. Module RWX: implementation of a resource for heap ownership         *)
(*    transfer and reader counting, specific to reader-writer locking.    *)
(*                                                                        *)
(* 5. Module RW: combination of Spin(writer) + RWX + Spin(reader) into    *)
(*    the module RW for reader-writer locking, with the lifting and       *)
(*    combination of genlock and genunlock methods from Spin(writer) and  *)
(*    Spin(reader) to a number of methods for reader-writer locking, e.g. *)
(*    wr_lock, wr_unlock, prologue, epilogue, and read.                   *)
(**************************************************************************)


(*************************************************************************)
(* Module Spin for spinlocks                                             *)
(*************************************************************************)
(*************************************************************************)
(*************************************************************************)
(* 1. State space of Spin contains the following subjective fields:      *)
(*                                                                       *)
(*    pi - permission to unlock                                          *)
(*                                                                       *)
(*    ht - history of lock/unlock events                                 *)
(*                                                                       *)
(* 2. Spin contains internal transitions for locking and unlocking, and  *)
(*    external transitions for acquiring and releasing the permission to *)
(*    unlock.                                                            *)
(*************************************************************************)

Module Spin.
Section Spin.
Variable l : ptr.
Definition U := [pcm of mtx2 * natmap (bool * bool)].
Definition T : Type := unit.
Implicit Type s : state U T.

(* permission to unlock *)
Definition pi (x : U) := x.1.
Canonical pi_morph := Eval hnf in [morphism of pi].

(* history *)
Definition ht (x : U) := x.2.
Canonical ht_morph := Eval hnf in [morphism of ht].

(* eta *)
Lemma idE : idfun = [** pi & ht].
Proof. by apply: fext=>x; apply: prod_eta. Qed.

Definition lockP (v : bool * bool) := v.2 = ~~ v.1.

(* the lock is taken if the total history ends with a lock event *)
Notation "\lkj s" := (last_val false (ht.t s)) (at level 24).

(* state space of Spin: resource invariant and a flattening function *)
Program Definition scoh :=
  CohPx2 (fun s => [/\ valid (pi.t s) && valid (ht.t s),
           l != null, (pi.t s == own) ==> \lkj s,
           um_all (fun => lockP) (ht.t s) & complete false (ht.t s) (\lkj s)])
         (fun s => l :-> (\lkj s)) _.
Next Obligation. by split=>[s[]|*]; rewrite ?(validPt,allvE,idE) ?lO // !lTE'. Qed.

(* lock transition *)
Program Definition lock_tr :=
  TransPx2 (fun s x => ~~ \lkj s /\
    x = State (own, fresh (ht.t s) \-> (false, true) \+ ht.s s)
              (idfun.j s) (idfun.o s)) scoh _.
Next Obligation.
move=>x y; split=>[s [_ ->][_ ->]|[_ ->] /=|[p q] ?? O|]; rewrite /= ?lE //=.
- case=>/negbTE -> [->-> _ _][/andP [X _] _].
  by rewrite -O pcmPJ -joinA=>/(mxON X)/mxE0 [->].
case=>/negbTE ->-> [/andP [V ?]] -> /(mxON V) /mxE0 [_ E] ??.
rewrite !lE E -!joinA valid_fresh !lastval_fresh //=.
by split=>//; [apply: umallPtUn | rewrite cm_fresh].
Qed.

(* unlock transition *)
Program Definition unlock_tr :=
  TransPx2 (fun s x => pi.s s = own /\
    x = State (nown, fresh (ht.t s) \-> (true, false) \+ ht.s s)
              (idfun.j s) (idfun.o s)) scoh _.
Next Obligation.
move=>x y; split=>[s [_ ->][_ ->]|[_ ->]|[p q] ?? O|]; rewrite /= ?lE //=.
- by case=>-> [->-> _ _][/andP [/mxMx/mxE0[->->]]]; rewrite -O !joinA.
case=>->-> []; rewrite !lE=>/andP [/mxMx ->/= V] ->-> *.
rewrite -!joinA valid_fresh lastval_fresh //.
by split=>//; [apply: umallPtUn | rewrite cm_fresh].
Qed.

(* external transition to release permission to unlock *)
Program Definition relperm_tr :=
  TransPx2 (fun s x => pi.s s = own /\
    x = State (nown, ht.s s) (idfun.j s) (idfun.o s)) scoh _.
Next Obligation.
move=>x y; split=>[s [_ ->][_ ->]|[_ ->]|[p q] ?? O|]; rewrite /= ?lE //=.
- by case=>->[->-> _ _][/andP[/mxMx/mxE0[->->]]]; rewrite -O.
by case=>->-> []; rewrite !lE=>/andP [/mxOx ->].
Qed.

(* external transition to acquire permission to unlock *)
Program Definition acqperm_tr :=
  TransPx2 (fun s x => [/\ \lkj s & pi.t s = nown] /\
    x = State (own, ht.s s) (idfun.j s) (idfun.o s)) scoh _.
Next Obligation.
move=>x y; split=>[s [_ ->][_ ->]|[_ ->]|[p q] ?? O|]; rewrite /= ?lE //=.
- by case; case=> -> /mxE0 [->->][->->]; rewrite -O.
by case; case=>E1 /mxE0 [-> E2] -> []; rewrite !lE E1 E2.
Qed.

Inductive Tr := lock_t | unlock_t | relperm_t | acqperm_t | id_t P of monoP scoh P.

Definition sinterp t :=
  match t with
    lock_t => lock_tr
  | unlock_t => unlock_tr
  | relperm_t => relperm_tr
  | acqperm_t => acqperm_tr
  | id_t P pf => pre_tr (id_tr scoh) pf
  end.

Definition sintern t := if t is (relperm_t | acqperm_t) then false else true.

Program Canonical spin := Eval hnf in ResPx scoh sinterp sintern _.
Next Obligation. by case=>[||||P pf] // _ ?? _ _; apply: domeqPt. Qed.

(* stability lemma *)
Lemma st x s1 s2 : osteps spin s1 s2 -> coh spin s1 ->
        x <= last_key (ht.t s1) -> x <= last_key (ht.t s2).
Proof.
move: s1 s2; apply: iters_ind=>// s1 s2 O C; first by apply: ocoh C.
case: O (ocoh (iter1 O) C); case=>[||||P pf][_]; last by [move=>/tr_inj ->];
move/trS=>-> [/andP [_]]; rewrite ?idE ?lE // joinCA => V *;
by apply: leq_trans (lastkeyUnf V).
Qed.

(* trylock method: erases to CAS-ing true into the pointer l *)
(* the erasure property is obvious, though we currently don't *)
(* issue proof obligations to prove it *)

(* a helper lemma to build the action of CAS-ing into l *)
Lemma takenM : monoP scoh (fun s => \lkj s).
Proof. by move=>?? /coh_valid A; rewrite !lE //= joinA. Qed.

Program Definition trylock : {h k}, STsep [spin]
  (fun s => ht.s s = h /\ k <= last_key (ht.t s),
   fun b s => if b then exists t,
     [/\ pi.s s = own, ht.s s = t \-> (false, true) \+ h & k < t]
   else [/\ ht.s s = h & k <= last_key (ht.t s)]) :=
  Do (actSx (fun b => if b then lock_t else id_t takenM)).
Next Obligation.
by split=>[|[]] // ??? [[|_ X Y]|[_ Y X|]] //; case: X; case: Y=>->.
Qed.
Next Obligation.
move=>i [h][k][[/= H1 H2]] C; apply: vrf_atm=>// s O [].
move/(st O C): H2=>H2; rewrite (osf O) {O} in H1; case: (boolP (\lkj s))=>E.
- exists false, s; rewrite effE.
 by split=>// m O; split; [rewrite -(osf O)|apply: (st O)].
case/andP=>/mxON X _ _ {}/X /= H _ _; eexists true, _; rewrite effE.
by split=>//= m O _; exists (fresh (ht.t s)); rewrite -!(osf O) !lE -H1.
Qed.

(* then genlock loops over trylock *)

(* the loop invariant *)
Definition lock_tp := {h k}, STsep [spin]
  (fun s => ht.s s = h /\ k <= last_key (ht.t s),
   fun (_ : unit) s => exists t,
     [/\ pi.s s = own, ht.s s = t \-> (false, true) \+ h & k < t]).

Program Definition genlock :=
  ffix (fun (loop : unit -> lock_tp) xx =>
    Do (b <-- trylock;
        if b then ret tt else loop tt)) tt.
Next Obligation.
move=>loop [i][h][k][/= H] C; apply: [stepE h, k]=>//= {i H C}.
case=>i; last by move=>*; apply: [gE h, k].
by case=>t [???] C; apply: vrf_ret=>// m O; exists t; rewrite -!(osf O).
Qed.

(* genunlock is equally simple; it erases to writing false into l *)
(* but, again, we don't issue proof obligations to prove erasure *)
Program Definition genunlock : {h k}, STsep [spin]
  (fun s => [/\ pi.s s = own, ht.s s = h & k <= last_key (ht.t s)],
   fun _ s => exists t, [/\ pi.s s = nown,
     ht.s s = t \-> (true, false) \+ h & k < t]) :=
  Do (actSx (fun _ : unit => unlock_t)).
Next Obligation. by split=>// ??? [][]. Qed.
Next Obligation.
move=>i [h][k][[/= H1 H2 H3]] C; apply: vrf_atm=>// s O.
move/(st O C): H3=>H3; rewrite !(osf O) {O} in H1 H2=>[[/andP []]] *.
eexists tt, _; rewrite effE; split=>// m O; exists (fresh (ht.t s)).
by rewrite -!(osf O) !lE -H2.
Qed.

End Spin.

(* we export that field names are morphisms *)
Module Exports.
Canonical pi_morph.
Canonical ht_morph.
Arguments Spin.relperm_t {l}.
Arguments Spin.acqperm_t {l}.
Arguments Spin.lock_t {l}.
Arguments Spin.unlock_t {l}.
Arguments Spin.id_t {l}.
End Exports.
End Spin.

Export Spin.Exports.


(*************************************************************************)
(* Module CSLX for heap ownership transfer for exclusive locking         *)
(*************************************************************************)
(*************************************************************************)
(*************************************************************************)
(* 1. State space of CSLX contains the following fields                  *)
(*                                                                       *)
(* - auth (subjective): authorization to move the heap (alt. reservation *)
(*   of the heap).                                                       *)
(*                                                                       *)
(*   if auth = nown, we can't move the heap.                             *)
(*                                                                       *)
(*   if auth = mx auth, we can move the heap, but haven't yet done so.   *)
(*                                                                       *)
(*   if auth = own, we have already moved the heap.                      *)
(*                                                                       *)
(* - hp (subjective): the private heap of the thread                     *)
(*                                                                       *)
(* - hpj (joint) : the shared heap to be moved                           *)
(*                                                                       *)
(* 2. CSLX has the internal transitions open_tr and close_tr for         *)
(*    acquiring and releasing the heap, and external transitions for     *)
(*    acquiring and releasing an authorization.                          *)
(*************************************************************************)

Module CSLX.
Section CSLX.
Variables (I : heap -> Prop) (prec_pf : precise I). (* resource invariant *)

Definition U := [pcm of mtx3 * heap].
Definition T : Type := heap.

(* authorization to move the heap (reservation of the heap) *)
Definition alpha : U -> _ := fst.
Canonical alpha_morph := Eval hnf in [morphism of alpha].

(* the private heap *)
Definition hp : U -> _ := snd.
Canonical hp_morph := Eval hnf in [morphism of hp].

(* the shared heap *)
Notation hpj := (@id T).

Lemma idE : (idfun = [** alpha & hp]) * (idfun = hpj).
Proof. by split; apply: fext=>x //; apply: prod_eta. Qed.

(* state space: the heap hp.j is shared if mu.t = nown *)
(* otherwise hp.j is empty, as the heap is given away *)
Program Definition xcoh :=
  CohPx2 (fun s : state U T =>
    [/\ valid (alpha.t s), valid (hpj.j s \+ hp.t s) &
        if alpha.t s == own then hpj.j s = Unit else I (hpj.j s)])
    (fun s => hpj.j s \+ hp.t s) _.
Next Obligation.
by split=>[s [M V ?]|*]; rewrite ?(lE,joinA) // allvE idE lTE' /= M !(validX V).
Qed.

(* open transition to acquire the heap if authorized *)
Program Definition open_tr :=
  TransPx2 (fun s x => alpha.s s = auth1 /\
    x = State (own, hpj.j s \+ hp.s s) Unit (idfun.o s)) xcoh _.
Next Obligation.
split=>[s [_ ->][_ ->]|[_ ->]|[p q] ?? O|]; rewrite /= ?lE //=.
- case=>-> [->->??][/mxMx/mxE0 [->->]].
  by rewrite -O !joinA pcmPE !unitR.
by case=>->-> [/mxMx E] V H; rewrite idE !lE E (validX V).
Qed.

(* close transition to release the heap if owning it *)
(* and the heap satisfies the resource invariant *)
Program Definition close_tr :=
  TransPx2 (fun s x => [/\ alpha.s s = own,
    hp.s s = hpj.j x \+ hp.s x & I (hpj.j x)] /\
    x = State (auth1, hp.s x) (@hpj.j x) (idfun.o s)) xcoh _.
Next Obligation.
split=>[s|[_ ->]|[p q] ?? O|]; rewrite /= ?lE //=; last 1 first.
- by case; case=>->-> X -> []; rewrite idE !lE=>/mxOx -> /validX ->.
- case; case=>_ -> I1 -> [[_ E I2 ->]][_ /validR V _].
  by rewrite !lE /= in E V; rewrite !(cancPLX _ V I1 I2 E).
case; case => ->-> X [-> _ _][/mxMx/mxE0 [->->]].
by rewrite O !joinA pcmPE !unitR.
Qed.

(* external transition for acquiring authorization *)
Program Definition acqauth_tr :=
  TransPx2 (fun s x => alpha.t s = Unit /\
    x = State (auth1, hp.s s) (idfun.j s) (idfun.o s)) xcoh _.
Next Obligation.
split=>[s [_ ->][_ ->]|[_ ->]|[p q] ?? O|]; rewrite /= ?lE //=.
- by case=>/mxE0 [->->][->-> _ _]; rewrite -O pcmPE.
by case; case/mxE0=>-> E -> []; rewrite !(idE,lE) E.
Qed.

(* external transition for releasing authorization *)
Program Definition relauth_tr :=
  TransPx2 (fun s x => alpha.s s = auth1 /\
    x = State (Unit, hp.s s) (idfun.j s) (idfun.o s)) xcoh _.
Next Obligation.
split=>[s [_ ->][_ ->]|[_ ->]|[p q] ?? O|]; rewrite /= ?lE //=.
- by rewrite -O; case=>-> [->->][/mxMx/mxE0 [->->]]; rewrite pcmPE.
by case=>->-> [/mxMx E]; rewrite !(idE,lE) E.
Qed.

Inductive xTr := open_t | close_t | acqauth_t | relauth_t | id_t P of monoP xcoh P.

Definition xinterp t :=
  match t with
    open_t => open_tr
  | close_t => close_tr
  | acqauth_t => acqauth_tr
  | relauth_t => relauth_tr
  | id_t P pf => pre_tr (id_tr xcoh) pf
  end.

Definition xintern t := if t is (acqauth_t | relauth_t) then false else true.

Program Canonical cslx := Eval hnf in ResPx xcoh xinterp xintern _.
Next Obligation.
case=>[||||P pf] // _ s x C []; last by move=>_ ->.
- by move=>_ ->; rewrite /flat /= !(idE,lE) domeqX.
case: C=>V _ X [Y E _ ->]; rewrite /flat /= !(idE,lE) E.
by case: mxPJ V Y X=>//= t _ _ _ ->->; rewrite domeqX.
Qed.

(* stable specs for open *)
Program Definition open : STsep [cslx]
  (fun s => [/\ alpha.s s = auth1 & hp.s s = Unit],
   fun _ s => [/\ alpha.s s = own & I (hp.s s)]) :=
  Do (actSx (fun _ : unit => open_t)).
Next Obligation. by split=>// ??? [[]]. Qed.
Next Obligation.
move=>s [[/= H1 H2]] C; apply: vrf_atm=>// x O [V _ H3].
rewrite !(osf O) {O} in H1 H2; eexists tt, _; rewrite effE; split=>//= m O _.
rewrite -!(osf O) !lE //=.
by rewrite H1 H2 mxE /= in V H3 *; rewrite (mxMx V) in H3.
Qed.

(* stable specs for close *)
Program Definition close : STsep [cslx]
  (fun s => [/\ alpha.s s = own & I (hp.s s)],
   fun _ s => [/\ alpha.s s = auth1 & hp.s s = Unit]) :=
  Do (actSx (fun _ : unit => close_t)).
Next Obligation. by split=>// ??? [[]]. Qed.
Next Obligation.
move=>s [[/= H1 H2]] C; apply: vrf_atm=>// j O {C} _.
rewrite !(osf O) {O} in H1 H2; eexists tt, (State (_, Unit) (hp.s j) _).
by rewrite effE /= !lE unitR; split=>// m O; rewrite -!(osf O) !lE.
Qed.

End CSLX.

Module Exports.
Canonical alpha_morph.
Canonical hp_morph.
Arguments CSLX.acqauth_t {I}.
Arguments CSLX.relauth_t {I}.
Arguments CSLX.open_t {I}.
Arguments CSLX.close_t {I}.
Arguments CSLX.id_t {I}.
End Exports.
End CSLX.

Export CSLX.Exports.


(**************************************************************************)
(* Combining Spin and CSLX into a CSL module.                             *)
(*                                                                        *)
(* The combination is carried out in three stages.                        *)
(*                                                                        *)
(* 1. We first use the tensor construction to construct CSL' module.      *)
(* This combines the state spaces of Spin and CSLX, and couples the       *)
(* Spin.relperm_t transition with CSLX.acqauth_t, and Spin.acqperm_t with *)
(* CSLX.relauth_t.                                                        *)
(*                                                                        *)
(* 2. Then we restrict CSL' to CSL by a sub-resource construction. This   *)
(* restricts the histories to those for exclusive locking, and restricts  *)
(* the other components appropriately (as described in the acompanying    *)
(* paper).                                                                *)
(*                                                                        *)
(* 3. Finally, exlock and exunlock are the obtained by morphing of        *)
(* Spin.genlock and Spin.genunlock along the canonical morphisms from     *)
(* Spin to CSL, with some additional glue ghost code (also explained in   *)
(* the paper).                                                            *)
(**************************************************************************)

Module CSL.
Section CSL.
Variables (lk : ptr) (I : heap -> Prop).
Variable prec_pf : precise I.
Notation spin := (Spin.spin lk).
Notation cslx := (CSLX.cslx prec_pf).

Definition U := [pcm of Spin.U * CSLX.U].
Definition T : Type := Spin.T * CSLX.T.

(* Stage 1: csl' resource obtained as tensor of spin and cslx *)
Definition couple (t1 : Tr spin) (t2 : Tr cslx) :=
  match t1, t2 with
  | Spin.relperm_t, CSLX.acqauth_t => true
  | Spin.acqperm_t, CSLX.relauth_t => true
  | Spin.lock_t , CSLX.id_t _ _ => true
  | Spin.unlock_t , CSLX.id_t _ _ => true
  | Spin.id_t _ _, CSLX.open_t => true
  | Spin.id_t _ _, CSLX.close_t => true
  | Spin.id_t _ _, CSLX.id_t _ _ => true
  | _, _  => false
  end.

Definition new_ints (t1 : Tr spin) (t2 : Tr cslx) : bool :=
  match t1, t2 with
  | Spin.relperm_t, CSLX.acqauth_t => true
  | Spin.acqperm_t, CSLX.relauth_t => true
  | _, _  => false
  end.

Lemma pfint : footpresP new_ints.
Proof.
case=>[||||??][] //= _ s s' _ []; rewrite /flat /= /flat /=;
by case=>_ -> [_ ->]; rewrite !lE.
Qed.

(* Definition of the cs' as a tensor *)
Definition csl' := tensor couple pfint.

(* Morphisms from spin and xfer to csl' *)
Definition couple_l (t : Tr spin) : Tr cslx :=
  match t with
  | Spin.relperm_t => CSLX.acqauth_t
  | Spin.acqperm_t => CSLX.relauth_t
  | _ => CSLX.id_t _ idM
  end.

Program Definition inl' := InlMorphPx couple pfint couple_l _.
Next Obligation. by split; case. Qed.

Program Definition inl_sim' : simulation inl' :=
  inl_sim (couple:=couple) pfint (couple_l:=couple_l) _ _ _.
Next Obligation. by move=>t ?? s2; case: t; exists s2. Qed.

Definition couple_r (t : Tr cslx) : Tr spin :=
  match t with
  | CSLX.acqauth_t => Spin.relperm_t
  | CSLX.relauth_t => Spin.acqperm_t
  | _ => Spin.id_t _ idM
  end.

Program Definition inr' := InrMorphPx couple pfint couple_r _.
Next Obligation. by split; case. Qed.

Program Definition inr_sim' : simulation inr' :=
  inr_sim (couple:=couple) pfint (couple_r:=couple_r) _ _ _.
Next Obligation. by move=>t ?? s1; case: t; exists s1. Qed.

(* abbreviated names for new internal transitions obtained by coupling *)
Definition relpermauth_t : Tr csl' :=
  couple_tx couple Spin.relperm_t CSLX.acqauth_t.
Definition acqpermauth_t : Tr csl' :=
  couple_tx couple Spin.acqperm_t CSLX.relauth_t.

(* Stage 2: restricting csl' to csl *)

(* give the components inheritted from spin and xfer local names *)
Definition pi : U -> _ := Spin.pi \o fst.
Canonical pi_morph := Eval hnf in [morphism of pi].
Definition alpha : U -> _ := CSLX.alpha \o snd.
Canonical alpha_morph := Eval hnf in [morphism of alpha].
Definition hp : U -> _ := CSLX.hp \o snd.
Canonical hp_morph := Eval hnf in [morphism of hp].

(* we don't need a name for the spin history, as for exlcusive *)
(* locking, we just need one bit of information; is the lock taken or not *)
(* we define morphism tx to check the last entry in Spin history *)
(* returning own if the history says we have the lock and nown otherwise *)
(* please see definition of morphism omg for details *)
Definition tx : U -> _ := omg \o Spin.ht \o fst.
Canonical tx_morph := Eval hnf in [morphism of tx].

(* Now we can define the restriction that we want to impose on csl'
(details are in the paper). The restriction consists of three parts.

1. 'sep tx is the compatibility relation of morphism tx.

   if 'sep tx x y, then x and y are compatible histories for exclusive
   locking

2. pi x \+ mxown (alpha x) == tx x

  says that we hold the lock (tx x = own) iff we have permission to
  unlock or have exchanged the permission for a reservation of the
  heap. similarly for

3. pi y \+ mxown (alpha y) == tx y is subjective dual of the previous
  conjunct. x refers to self state, y to other state, hence this
  conjunct imposes the same requirement on other threads. *)

Definition cinv (x y : Ux csl') :=
  [&& 'sep tx x y, pi x \+ mxown (alpha x) == tx x &
                   pi y \+ mxown (alpha y) == tx y].

(* cinv is an equalizer, as claimed in the paper *)
(* the following lemma is not really necessary for the development *)
(* but we prove it here, to confirm the claim in the paper *)
Lemma cinv_eqlz : cinv =s 'eqlz ('join pi (mxown \o alpha)) tx.
Proof.
move=>x y V; apply/idP/idP; rewrite !sepE /= !sepE /=.
- case/and3P=>S H1 H2; apply/and4P.
  by rewrite (eqP H1) (eqP H2) !eq_refl (pfV2 _ V).
by case/and4P=>/= S H1 H2 H3; apply/and3P.
Qed.

(* cinv is a compatibility relation *)
Lemma cslinv_sep_ax : seprel_axiom cinv.
Proof. by rewrite (sepXEQ cinv_eqlz); apply: sepax. Qed.
Canonical cinv_seprel := Eval hnf in seprel cinv cslinv_sep_ax.
Definition newpcm : sub_pcm cinv := xsepSubPCM cinv_seprel.

(* cinv is an invariant for csl' *)
(* i.e., it is preserved by the transitions *)
Lemma cslpf : sub_persP cinv PredT.
Proof.
move=>t s x C; rewrite prodVE=>/andP [V _] _ H; rewrite /cinv !lA !lprodE /=.
have V2 : valid (Spin.ht.t s\1) by case: C; case=>/andP [].
case: t H=>[[||||P pf][||||Q qf] _|?? [_ ->]] //= [[H1 ->]][H2 ->] _;
rewrite ?lE ?H2 //=.
- case/and3P=>S E ->; rewrite !sepE /omg_s /= /omg !lA !lastval_freshUn //=.
  rewrite lastkey_freshUn // ltnS lastkeyUnf //.
  by move: H1 E; rewrite !omgP //=; do 2![case: mxPJ].
- rewrite H1; case/and3P=>/(stVf (omg \o Spin.ht) V) S E ->.
  rewrite !sepE /omg_s /= /omg !lA !lastval_freshUn // !lastkey_freshUn //=.
  by move: S; rewrite omgP -(eqP E)=>/mxOxy /mxE0 [->->].
- by rewrite H1 (mxE0 H2) (lCSX omg) !sepE !lA.
- by case: H1=>_ /mxE0 [-> _]; rewrite (lCSX omg) !sepE !lA.
by case: H2=>->.
Qed.

(* CSL is now obtained by sub-resource construction. Sub_res combines
into one construction, the two constructions of subjective
sub-resource and global sub-resource described in the paper. Here we
invoke sub_res with the trivial global predicate globalT, because we
only need subjective sub-resource *)
Definition csl := @sub_res csl' cinv_seprel _ newpcm globalT cslpf.

(* morphisms from csl'/spin/xfer to csl *)
Definition sub_morf : morph unit csl' csl :=
  @sub_morph csl' cinv_seprel _ _ globalT cslpf.
Definition sub_sim' : simulation sub_morf :=
  @sub_sim csl' cinv_seprel _ _ globalT cslpf.
Definition inl_morf : morph unit spin csl := comp_morph inl' sub_morf.
Definition inr_morf : morph unit cslx csl := comp_morph inr' sub_morf.
Definition inls : simulation inl_morf := comp_sim inl_sim' sub_sim'.
Definition inrs : simulation inr_morf := comp_sim inr_sim' sub_sim'.

(* The following stability lemmas relate f-stability of inl_morf and
inr_morf morphisms to stability wrt stepping of other threads.

Specifically in this case, if csl steps by inr_morf (i.e., the cslx
component of csl steps), then that corresponds to other steps in
spin. This is so, because internal transitions of cslx are paired with
identity in csl. Dually for inl_morf and spin.  *)
Lemma mxc s1 s2 :
        msteps (inr_morf tt) s1 s2 -> osteps spin (spval s1)\1 (spval s2)\1.
Proof. by apply: msteps_subr=>[??|[]] //; case. Qed.

Lemma msc s1 s2 :
        msteps (inl_morf tt) s1 s2 -> osteps cslx (spval s1)\2 (spval s2)\2.
Proof. by apply: msteps_subl=>[??|[]] //; case. Qed.

(* We next introduce the morphism mu, defined in terms of the existing
components, to signal if we hold the lock. This morphism will be used
in the final specs for exlock and exunlock below (details in the
paper). *)
Definition mu : Ux csl -> _ := 'res_('ker pi) (mxundef \o alpha) \o pval.
Canonical mu_morph := Eval hnf in [morphism of mu].

(* It turns out we need another helper morphism, a different name for the *)
(* heap component. This is a simple coercion of hp from csl' to csl *)
(* but must be made explicit in the type theory *)
Definition hx : Ux csl -> _ := hp \o pval.
Canonical hx_morph := Eval hnf in [morphism of hx].

(* Exlock defined by composing genlock; relpermauth; open.

We could put mu.s s = nown in the precondition, but we elide it. The
condition is irrelevant, as the program works correctly (though it
loops infinitely) when mu.s s = own. The condition can easily be
added by rule consequence, so we don't bother with it. *)
Program Definition exlock : STsep [csl]
  (fun s => hx.s s = Unit,
   fun _ s => mu.s s = own /\ I (hx.s s)) :=
  Do (morph_app inl_morf tt (Spin.genlock lk);;
      morph_app sub_morf tt (actSx (fun _ : unit => relpermauth_t));;
      morph_app inr_morf tt (CSLX.open prec_pf)).
Next Obligation. by split=>[??? [][]|[]] //; right. Qed.
Next Obligation.
move=>i [/= H1] _; apply: step; apply: (vrf_morph inls) =>[||_]//; first by eexists.
apply: [gE Spin.ht.s (spval i)\1, last_key (Spin.ht.t (spval i)\1)]=>//=.
case=>j [_][H2 _ _] _ [x][w][-> Y _] C M; rewrite {j}Y in H1 H2 *.
rewrite lsubX lprodE {i M}(osf (msc (mxms inl_morf M))) -!lprodX in H1 H2.
apply: step; apply: (vrf_morph sub_sim')=>// _; apply: vrf_atm=>// j O C'.
rewrite !(osf O) !lprodE in H1 H2.
case: (ospval_subX (pf:=globalT) (qf:=cslpf) O C)=>k [?] {} O.
subst j; move/(ocoh O): C=>C.
have P : alpha.t (spval k) = Unit.
- case/coh_sep: C; case/and3P=>/(stVf tx (coh_valid C')).
  rewrite !lA !lprodE /= =>S X Y _; move: S; rewrite -(eqP X) -(eqP Y) H2.
  by case/mxOxy/mxE0=>/mxownP -> /mxE0 [_ /mxownP ->].
rewrite {O} lprodE in P; eexists tt, _; split; rewrite ?effE.
- by apply: act2_safe=>//; case: C'=>?? W; rewrite /flat /= !lE (validX W).
move=>{P H2 C' C} m O [_ X _][j] E _ C _; rewrite {m}E in O.
apply: (vrf_morph inrs)=>[||_] //; first by eexists.
apply: gE=>[||{x H1 X C}[k'][H1 H2] _] //=.
- by rewrite -!lprodX -!(osf O) !lprodE !lE.
case=>z [_][->] E _ [/= H _ _] M; rewrite {k'}E in H1 H2.
have H3 : Spin.pi.s (spval z)\1 = Unit.
- by rewrite -(osf (mxc (mxms inr_morf M))) -lprodX -(osf O) !lprodE !lE.
by rewrite !lsubX lSXid !sepE /= !lA !lprodE H1 H3.
Qed.

(* Exunlock is dually defined by composing close; acqpermauth; genunlock *)
Program Definition exunlock : STsep [csl]
  (fun s => mu.s s = own /\ I (hx.s s),
   fun _ s => mu.s s = nown /\ hx.s s = Unit) :=
  Do (morph_app inr_morf tt (CSLX.close prec_pf);;
      morph_app sub_morf tt (actSx (fun _ : unit => acqpermauth_t));;
      morph_app inl_morf tt (Spin.genunlock lk)).
Next Obligation. by split=>[??? [][]|[]] //; right. Qed.
Next Obligation.
move=>i [[/= H1 H2]] _; apply: step; apply: (vrf_morph inrs)=>//; first by eexists.
move: H1 H2; rewrite !lsubX lSXid !sepE /= !lA !lprodE.
rewrite pfunit andbT; case: unitbP=>// _; case: mxundefP=>// H1 _ H2 _.
apply: gE=>//; case=>{H1 H2} m [H1 H2] _ [xx][_][-> H _] C _; rewrite {m}H in H1 H2 *.
apply: step; apply: (vrf_morph sub_sim')=>[||[???]] //.
apply: vrf_atm=>// j O C'; rewrite -!lprodX !(osf O) !lprodE in H1 H2.
case: (ospval_subX (pf:=globalT) (qf:=cslpf) O C)=>x [?] {} O; subst j.
move/(ocoh O): C=>Ck.
have [P1 P2] : last_val false (Spin.ht.t (spval x)\1) /\ pi.t (spval x) = nown.
- case/coh_sep: Ck=>/and3P []; rewrite !lA /= => S.
  move: (stVf tx (coh_valid C') S); rewrite !lprodE in S * => V X Y _ [/coh_valid A] _ _.
  rewrite omgP //; rewrite -(eqP X) -(eqP Y) H1 in V *.
  by case/mxyOx/mxE0: V=>-> /mxE0 [->->].
rewrite !lprodE in P2; eexists tt, _; rewrite effE; split.
- by apply: act2_safe=>//; case: C'=>_ _ V; rewrite /flat /= !lE (validX V).
move=>m {} O _ [j] E _ C _; rewrite {m}E in O.
apply: (vrf_morph inls)=>//; first by eexists.
move=>_; apply: [gE Spin.ht.s (spval j)\1, Unit]=>//=.
- by rewrite -!lprodX -!(osf O) !lprodE !lE.
case=>z [_][{}H1 _ _] _ [y][_][->] E _ [/= H _ _] /(mxms inl_morf) M.
rewrite {z}E in H1 *; rewrite !lsubX lSXid !sepE /= !lA !lprodE.
by rewrite H1 -!(osf (msc M)) -!lprodX /= -!(osf O) !lprodE !lE.
Qed.

End CSL.
End CSL.


(**************************************************************************)
(* Module RWX for heap ownership transfer and reader counting             *)
(* discipline of reader-writer locking; but without the physical reader   *)
(* and writer locks themselves. The physical locks will be provided by    *)
(* composing with two instances of Spin (one for writers, one for         *)
(* readers).                                                              *)
(**************************************************************************)
(**************************************************************************)
(**************************************************************************)
(*                                                                        *)
(* 1. The state space of RWX contains the following components            *)
(*                                                                        *)
(*    - alpha (subjective) : writer authorization                         *)
(*                                                                        *)
(*       - if nown, then we don't have permission to take the heap.       *)
(*                                                                        *)
(*       - if auth, the we have permission to take the heap and become a  *)
(*         writer, of we can exchange the permission to become the first  *)
(*         reader.                                                        *)
(*                                                                        *)
(*       - if own, we are a writer and have already taken the heap.       *)
(*                                                                        *)
(*     - rho (subjective) : reader authorization                          *)
(*                                                                        *)
(*       - if nown, we don't hold the reader lock                         *)
(*                                                                        *)
(*       - if own, we do hold the reader lock                             *)
(*                                                                        *)
(*     - kappa (subjective) : count of registered readers                 *)
(*                                                                        *)
(*     - hp (subjective) : private heap                                   *)
(*                                                                        *)
(*     - hpj (joint) : the shared heap to be transfered to writers or     *)
(*       read by readers.                                                 *)
(*                                                                        *)
(*     - ct (joint) : the contents of the counter of registered readers.  *)
(*                                                                        *)
(* 2. Transitions of RWX are numerous, and we comment on them as they are *)
(* implemented.                                                           *)
(**************************************************************************)
Module RWX.

(* We define U as an explicit 4-tuple *)
(* to avoid somewhat annoying iteration of PCM products *)
Definition U := [pcm of Prod4 mtx3 mtx2 lnat heap].
Definition T : Type := nat * heap.

(* authorization to take the heap *)
Definition alpha : U -> mtx3 := pcm41.
Canonical alpha_morph := Eval hnf in [morphism of alpha].

(* do we hold the reader lock *)
Definition rho : U -> mtx2 := pcm42.
Canonical rho_morph := Eval hnf in [morphism of rho].

(* how many readers are in the system *)
Definition kappal : U -> lnat := pcm43.
Canonical kappal_morph := Eval hnf in [morphism of kappal].
(* kappal uses topped nats, i.e. nats with an extra undefined elements *)
(* which is necessary for subPCM construction. But everywhere else *)
(* we use the version which removes the undefined element *)
Definition kappa : U -> nat := nat_unlift \o kappal.
Canonical kappa_morph := Eval hnf in [morphism of kappa].

(* the heap *)
Definition hp : U -> heap := pcm44.
Canonical hp_morph := Eval hnf in [morphism of hp].

Lemma idE : idfun = fun x => mk4 (alpha x) (rho x) (kappal x) (hp x).
Proof. by apply: fext; case. Qed.

(* some helper lemmas to manipulate the tuples of all components *)
Lemma lSU (s : U * U) :
        (idfun.s s = mk4 (alpha.s s) (rho.s s) (kappal.s s) (hp.s s)) *
        (idfun.o s = mk4 (alpha.o s) (rho.o s) (kappal.o s) (hp.o s)).
Proof. by rewrite !Private_Notation.so_appE !Private_Notation.os_appE idE. Qed.

Lemma lTU (s : U * U) :
        idfun.t s = mk4 (alpha.t s) (rho.t s) (kappal.t s) (hp.t s).
Proof. by rewrite !lSU. Qed.


Section RWX.
Variables (count : ptr) (I : heap -> Prop) (prec_pf : precise I).
(* count is the pointer for keeping track of registered readers *)
(* I is a resource invariant, which is required to be precise. *)

(* the contents of pointer count *)
Definition ct := @fst nat heap.

(* the shared heap *)
Definition hpj := @snd nat heap.

(* the erasure function declares ct as contents of count *)
Let rwflat :=
  [fun s : state U T => count :-> ct.j s \+ hpj.j s \+ hp.t s].

(* Definition of state space of RWX and declaration of rwflat *)
(* as the erasure function *)
Program Definition xcoh :=
  CohPx2 (fun s =>
    [/\ valid (alpha.t s), valid (rho.t s), valid (kappal.t s), valid (rwflat s),
        (alpha.t s == nown) || (kappa.t s == 0),
        (rho.t s == nown) ==> (kappa.t s == ct.j s) &
        if alpha.t s == own then hpj.j s = Unit else I (hpj.j s)]) rwflat _.
Next Obligation.
split=>[s [V1 V2 V3 V4 *]|*]; last by rewrite !lE //= !lE.
by rewrite allvE lTU /= V1 V2 V3 !(validX V4).
Qed.

(* internal transition for giving the heap to the writer *)
Program Definition open_tr :=
  TransPx2 (fun s x => alpha.s s = auth1 /\
    x = State (mk4 own (rho.s s) (up (kappa.s s)) (hpj.j s \+ hp.s s))
              (ct.j s, Unit) (idfun.o s)) xcoh _.
Next Obligation.
split=>[s [_ ->][_ ->]|[_ ->]|[p q r x] ?? O|]; rewrite /= ?lE //=.
- case=> -> [->->????][].
  by case: liftPJ=>// ?? -> _ /mxMx/mxE0 ->; rewrite !joinA -O.
by case=>->-> []; rewrite !lE; case: liftPJ=>// ?? _ -> /mxMx ->-> _ /validX ->.
Qed.

(* internal transition for writer giving the heap back *)
Program Definition close_tr :=
  TransPx2 (fun s x => [/\ alpha.s s = own,
    hp.s s = hpj.j x \+ hp.s x & I (hpj.j x)] /\
    x = State (mk4 auth1 (rho.s s) (up (kappa.s s)) (hp.s x))
              (ct.j s, hpj.j x) (idfun.o s)) xcoh _.
Next Obligation.
split=>[s|[_ ->]|[p q r x] ?? O|]; rewrite /= ?lE //=.
- case; case=>_ -> I1 -> [[_ E I2 ->]][_ _ _ /validR V _ _].
  by rewrite !lE /= in E V *; rewrite !(cancPLX _ V I1 I2 E).
- case; case=>->-> X [->-> ????][].
  by case: liftPJ=>// ?? -> _ /mxMx/mxE0 ->; rewrite !joinA -O.
case; case=>->-> X -> []; rewrite !lE; case: liftPJ=>// ?? _ ->.
by move/mxOx=>->-> _ /validX ->.
Qed.

(* internal transition for first reader to register *)
(* the reader has to be in critical section (i.e., hold reader lock) *)
(* and the writer lock has to be taken *)
(* the writer lock permission is exchanged for increasing the reader count *)
Program Definition getrd1_tr :=
  TransPx2 (fun s x => [/\ rho.s s = own & alpha.s s = auth1] /\
    x = State (mk4 nown own (up 1) (hp.s s))
              (idfun.j s) (idfun.o s)) xcoh _.
Next Obligation.
split=>[s [_ ->][_ ->]|[_ ->]|[p q r x] ?? O|]; rewrite /= ?lE //=.
- rewrite /kappa /=; case; case=>->-> [->-> ????][].
  case: liftPJ=>// ?? -> _ /mxMx/mxE0 [->->] /mxMx/mxE0 -> ??.
  by rewrite addn_eq0 /= -O=>/andP [/= /eqP ->].
case; case=>->->-> []; rewrite !lE; case: liftPJ=>// ?? _ ->.
by move/mxMx=>-> /mxMx -> _ /validX ->.
Qed.

(* internal transition for last reader to deregister *)
(* the reader has to be in critical section *)
(* produces wx = mx auth *)
Program Definition relrd1_tr :=
  TransPx2 (fun s x => [/\ rho.s s = own, kappa.s s = 1 & kappa.o s = 0] /\
    x = State (mk4 auth1 own (up 0) (hp.s s))
              (idfun.j s) (idfun.o s)) xcoh _.
Next Obligation.
split=>[s [_ ->][_ ->]|[_ ->]|[p q r x] ?? O|]; rewrite /= ?lE //=; case; case.
- rewrite /kappa /= =>->-> E [->-> ????][].
  case: liftPJ E=>// ?? -> _ /= /eqP; rewrite addn_eq0=>/andP [/eqP -> /eqP ->].
  by case: mxPJ=>//= ??? /mxMx/mxE0 ->; rewrite -O.
move=>->-> E -> []; rewrite !lE E; case: liftPJ=>//= ?? _ ->.
by case: mxPJ=>//= ??? /mxxM -> _ /validX ->.
Qed.

(* internal transition to increase reader count if not first reader *)
Program Definition incrd_tr :=
  TransPx2 (fun s x => [/\ rho.s s = own & kappa.t s > 0] /\
    x = State (mk4 (alpha.s s) own (up (kappa.s s).+1) (hp.s s))
              (idfun.j s) (idfun.o s)) xcoh _.
Next Obligation.
split=>[s [_ ->][_ ->]|[_ ->]|[p q r x] ?? O|]; rewrite /= ?lE //=; case; case.
- rewrite /kappa /= =>-> H [->-> ????][].
  by case: liftPJ H=>// ?? -> _ H _ /mxMx/mxE0 ->; rewrite -O.
case: (kappa.t _)=>// n -> _ -> []; rewrite !lE.
by case: liftPJ=>// ?? _ ->-> /mxMx -> _ /validX ->.
Qed.

(* internal transition to decrease reader count if not last reader *)
Program Definition decrd_tr :=
  TransPx2 (fun s x => [/\ rho.s s = own, kappa.s s > 0 & kappa.t s > 1] /\
    x = State (mk4 (alpha.s s) own (up (kappa.s s).-1) (hp.s s))
              (idfun.j s) (idfun.o s)) xcoh _.
Next Obligation.
split=>[s [_ ->][_ ->]|[_ ->]|[p q r x] ?? O|]; rewrite /= ?lE //=; case; case.
- case H1 : (kappa.s s1)=>[//|n]; rewrite /kappa /= in H1 *.
  rewrite joinCA=>-> _ H2 [->-> ????][]; case: liftPJ H2=>// ?? -> _ /= H2 _.
  by move/mxMx/mxE0 ->; rewrite -O.
case: (kappa.s _)=>// n -> _ _ -> []; rewrite !lE; case: liftPJ=>//= ?? _ ->->.
by move/mxxM=>-> _ /validX ->; rewrite orbF=>->.
Qed.

(* internal transition for writing c into the physical field ct for reader count *)
Program Definition writect_tr (c : nat) :=
  TransPx2 (fun s x => rho.s s = own /\
    x = State (idfun.s s) (c, hpj.j s) (idfun.o s)) xcoh _.
Next Obligation.
split=>[s [_ ->][_ ->]|[_ ->]|[p q r x] ?? O|]; rewrite /= ?lE //=; case.
- by move=>-> [->-> ????][_ /mxMx/mxE0 ->]; rewrite -O.
by move=>E -> []; rewrite !lE E => -> /mxMx ->-> /validX ->.
Qed.

(* external transition to take writer authorization *)
Program Definition acqauth_tr :=
  TransPx2 (fun s x => (alpha.t s = nown /\ kappa.t s = 0) /\
    x = State (mk4 auth1 (rho.s s) (up (kappa.s s)) (hp.s s))
              (idfun.j s) (idfun.o s)) xcoh _.
Next Obligation.
split=>[s [_ ->][_ ->]|[_ ->]|[p q r x] ?? O|]; rewrite /= ?lE //=; case; case.
- rewrite /kappa /= =>/mxE0 [->->] /eqP.
  rewrite addn_eq0=>/andP [/eqP E1 /eqP E2][->-> ????][].
  by case: liftPJ E1=>// ?? -> /= _ ->; rewrite -O.
by case/mxE0=>-> E1 E2 -> []; rewrite !lE E1 E2; case: liftPJ=>// ?? _ ->.
Qed.

(* external transition to release writer authorization *)
Program Definition relauth_tr :=
  TransPx2 (fun s x => alpha.s s = auth1 /\
    x = State (mk4 nown (rho.s s) (up (kappa.s s)) (hp.s s))
              (idfun.j s) (idfun.o s)) xcoh _.
Next Obligation.
split=>[s [_ ->][_ ->]|[_ ->]|[p q r x] ?? O|]; rewrite /= ?lE //=.
- case=>-> [->->????][].
  by case: liftPJ=>// ?? -> /= _ /mxMx/mxE0 ->; rewrite -O.
case=>->-> []; rewrite !lE; case: liftPJ=>// ?? _ -> /mxMx -> ???.
by case: eqP=>//= _ /andX ->.
Qed.

(* external transition to get readers authorization *)
Program Definition acqro_tr :=
  TransPx2 (fun s x => rho.t s = nown /\
    x = State (mk4 (alpha.s s) own (up (kappa.s s)) (hp.s s))
              (idfun.j s) (idfun.o s)) xcoh _.
Next Obligation.
split=>[s [_ ->][_ ->]|[_ ->]|[p q r x] ?? O|]; rewrite /= ?lE //=.
- case=>/mxE0 [->->][->-> ????][].
  by case: liftPJ=>// ?? ->; rewrite -O.
by case; case/mxE0=>-> E -> []; rewrite !lE E; case: liftPJ=>// ?? _ ->.
Qed.

(* external transition to release reader lock *)
(* we check that the invariant is restored before returning *)
Program Definition relro_tr :=
  TransPx2 (fun s x => [/\ rho.s s = own & kappa.t s = ct.j s] /\
    x = State (mk4 (alpha.s s) nown (up (kappa.s s)) (hp.s s))
              (idfun.j s) (idfun.o s)) xcoh _.
Next Obligation.
split=>[s [_ ->][_ ->]|[_ ->]|[p q r x] ?? O|[]]; rewrite /= ?lE //=.
- case; case=>->-> [->->????][].
  by case: liftPJ=>// ?? /= -> _ _ /mxMx/mxE0 ->; rewrite -O.
by case=>-> E -> []; rewrite !lE E; case: liftPJ=>// ?? _ ->-> /mxMx -> /=.
Qed.

Inductive xTr :=
  open_t | close_t | getrd1_t | relrd1_t | incrd_t | decrd_t |
  acqauth_t | relauth_t | acqro_t | relro_t |
  writect_t of nat | id_t P of monoP xcoh P.

Definition xinterp t :=
  match t with
    open_t => open_tr | close_t => close_tr | getrd1_t => getrd1_tr
  | relrd1_t => relrd1_tr | incrd_t => incrd_tr | decrd_t => decrd_tr
  | acqauth_t => acqauth_tr | relauth_t => relauth_tr | acqro_t => acqro_tr
  | relro_t => relro_tr | writect_t c => writect_tr c
  | id_t P pf => pre_tr (id_tr xcoh) pf
  end.

Definition xintern t :=
  if t is (acqauth_t | relauth_t | acqro_t | relro_t) then false else true.

Program Canonical rwx := Eval hnf in ResPx xcoh xinterp xintern _.
Next Obligation.
case=>[||||||||||n|P pf] // _ s x C []; last by [move=>_ ->];
try by [move=>_ ->; rewrite /flat /= !lE domeqX].
case: C=>V ????? X [Y E _ ->]; rewrite /flat /= !lE E.
by case: mxPJ V Y X=>//= t _ _ _ ->->; rewrite domeqX.
Qed.

(* stability properties *)

(* When we hold the reader lock, others can't change rd.o and ctj *)
Lemma orx s1 s2 :
       osteps rwx s1 s2 -> coh rwx s1 -> rho.s s1 = own ->
       (kappa.o s1 = kappa.o s2) * (ct.j s1 = ct.j s2).
Proof.
move: s1 s2; apply: (pres_iters_ind (@ocoh1 rwx))=>// s1 s2.
- by move=>? [->->][->->].
- by move=>O C; rewrite (osf O).
case; case=>[||||||||||c|P pf][]; rewrite ?lE /=; last by [move=>_ /tr_inj ->].
- by move=>_ ->; rewrite !lE.
- by move=>_ ->; rewrite !lE.
- by case=>-> _ _ [_ /mxMx ->].
- by case=>-> _ _ _ [_ /mxMx ->].
- by case=>-> _ _ [_ /mxMx ->].
- by case=>-> _ _ _ [_ /mxMx ->].
- by move=>_ ->; rewrite !lE.
- by move=>_ ->; rewrite !lE.
- by case/mxE0=>->.
- by case=>-> _ -> [_ /mxMx ->].
by move=>-> _ [_ /mxMx ->].
Qed.

(* if there are readers in the system, the shared heap doesn't change *)
(* will be used to verify the method for reading from the shared heap *)
(* when registered as a reader *)
Lemma ohp s1 s2 :
        osteps rwx s1 s2 -> coh rwx s1 -> kappa.s s1 > 0 ->
        hpj.j s1 = hpj.j s2.
Proof.
move: s1 s2; apply: (pres_iters_ind (@ocoh1 rwx))=>//= s1 s2.
- by move=>O C; rewrite (osf O).
case; case=>[||||||||||c|P pf][]; rewrite ?lE /=; last by [move=>_ /tr_inj ->];
try by [move=>_ ->; rewrite !lE].
- by move=>->-> []; rewrite !lE=>/mxxM->???; rewrite addn_eq0=>/andP [/eqP ->].
- by case=>-> _ _ -> [/mxxM -> ???]; rewrite addn_eq0=>/andP [/eqP ->].
Qed.

(* stable specs for the transitions *)
(* a little bit of a boring development, but stability is of essence *)
(* and once established, simplifies proofs later on *)

(* stable spec for open *)
Program Definition open : STsep [rwx]
  (fun s => [/\ alpha.s s = auth1 & hp.s s = Unit],
   fun _ s => [/\ alpha.s s = own & I (hp.s s)]) :=
  Do (actSx (fun _ : unit => open_t)).
Next Obligation. by split=>// ??? [[]]. Qed.
Next Obligation.
move=>s [[/= H1 H2]] C; apply: vrf_atm=>// x O [V _ _ _ _ _ H].
rewrite !(osf O) {O} in H1 H2; eexists tt, _.
rewrite effE; split=>//= m O _; rewrite -!(osf O) !lE //=.
by rewrite H1 H2 mxE /= in V H *; rewrite (mxMx V) /= in H.
Qed.

(* stable spec for close *)
Program Definition close : STsep [rwx]
  (fun s => [/\ alpha.s s = own & I (hp.s s)],
   fun _ s => [/\ alpha.s s = auth1 & hp.s s = Unit]) :=
  Do (actSx (fun _ : unit => close_t)).
Next Obligation. by split=>// ??? [[]]. Qed.
Next Obligation.
move=>s [[/= H1 H2]] C; apply: vrf_atm=>// j O {C} _.
rewrite !(osf O) in H1 H2; eexists tt, (State (mk4 _ _ _ Unit) (_, hp.s j) _).
by rewrite {O} effE /= !lE unitR; split=>// m O; rewrite -!(osf O) !lE.
Qed.

(* stable spec for getrd1 *)
Program Definition getrd1 : STsep [rwx]
  (fun s => [/\ rho.s s = own, alpha.s s = auth1,
     hp.s s = Unit & ct.j s = 0],
   fun _ s => [/\ rho.s s = own, alpha.s s = nown, hp.s s = Unit,
     kappa.s s = 1, kappa.o s = 0 & ct.j s = 0]) :=
  Do (actSx (fun _ : unit => getrd1_t)).
Next Obligation. by split=>// ??? [[]]. Qed.
Next Obligation.
move=>s [[/= H1 H2 H3 H4]] C; apply: vrf_atm=>// j O _.
have H5 : kappa.o s = 0.
- by move: H2 C=>-> [/mxMx -> _ _ _]; rewrite addn_eq0=>/andP [_ /eqP].
rewrite !(orx O C H1) in H4 H5; rewrite !(osf O) in H1 H2 H3.
eexists tt, _; rewrite effE.
by split=>// x {}O {}C; rewrite -!(orx O C) -?(osf O) ?lE.
Qed.

(* stable spec for relrd1 *)
Program Definition relrd1 : {h k}, STsep [rwx]
  (fun s => [/\ rho.s s = own, hp.s s = h,
     kappa.s s = 1, kappa.o s = 0 & ct.j s = k],
   fun _ s => [/\ rho.s s = own, alpha.s s = auth1, hp.s s = h,
     kappa.s s = 0, kappa.o s = 0 & ct.j s = k]) :=
  Do (actSx (fun _ : unit => relrd1_t)).
Next Obligation. by split=>// ??? [[]]. Qed.
Next Obligation.
move=>s [h][k][[/= H1 H2 H3 H4 H5]]  C; apply: vrf_atm=>// j O _.
rewrite !(orx O C H1) in H4 H5; rewrite !(osf O) {O C} in H1 H2 H3.
eexists tt, _; rewrite effE.
by split=>[|x O C]; rewrite ?H2 // -!(orx O C) -?(osf O) ?lE.
Qed.

(* stable spec for incrd *)
Program Definition incrd : {m n k}, STsep [rwx]
  (fun s => [/\ rho.s s = own, hp.s s = Unit,
      kappa.s s = m, kappa.o s = n, ct.j s = k & m + n > 0],
   fun _ s => [/\ rho.s s = own, hp.s s = Unit,
      kappa.s s = m.+1, kappa.o s = n & ct.j s = k]) :=
  Do (actSx (fun _ : unit => incrd_t)).
Next Obligation. by split=>// ??? [[]]. Qed.
Next Obligation.
move=>s [m][n][k][[H1 H2 H3 H4 H5 H6]] C; apply: vrf_atm=>// j O _.
rewrite !(orx O C H1) in H4 H5; rewrite !(osf O) {O C} in H1 H2 H3.
eexists tt, _; rewrite effE; split=>[|x O C] //=; first by rewrite H3 H4.
by rewrite -!(orx O C) -?(osf O) ?lE.
Qed.

(* stable spec for decrd *)
Program Definition decrd : {h m n k}, STsep [rwx]
  (fun s => [/\ rho.s s = own, hp.s s = h,
     kappa.s s = m, kappa.o s = n, ct.j s = k, m > 0 & m + n > 1],
   fun _ s => [/\ rho.s s = own, hp.s s = h,
     kappa.s s = m.-1, kappa.o s = n & ct.j s = k]) :=
  Do (actSx (fun _ : unit => decrd_t)).
Next Obligation. by split=>// ??? [[]]. Qed.
Next Obligation.
move=>s [h][m][n][k][[H1 H2 H3 H4 H5 H6 H7]] C; apply: vrf_atm=>// j O _.
rewrite !(orx O C H1) in H4 H5; rewrite !(osf O) {O C} in H1 H2 H3.
eexists tt, _; rewrite effE; split=>[|x O C] /=; first by rewrite H3 H4.
by rewrite -!(orx O C) -?(osf O) ?lE.
Qed.

(* stable spec for reading the contents ct of reader counter *)
Lemma ctM r : monoP xcoh (fun s => ct.j s = r).
Proof. by move=>???; rewrite !lE. Qed.

Program Definition read_ct : {h m n c}, STsep [rwx]
  (fun s => [/\ rho.s s = own, hp.s s = h,
     kappa.s s = m, kappa.o s = n & ct.j s = c],
   fun (r : nat) s => [/\ r = c,
     rho.s s = own, hp.s s = h, kappa.s s = m,
     kappa.o s = n & ct.j s = c]) :=
  Do (actSx (fun r => id_t (@ctM r))).
Next Obligation. by split=>// ????? _ [<-<-][<-<-]. Qed.
Next Obligation.
move=>s [_][_][_][_][[H <-<-<-<-]] C; apply:vrf_atm=>// j O _.
rewrite !(orx O C H); rewrite !{O C}(osf O) in H *.
eexists (ct.j j), _; rewrite effE.
by split=>//= x O C; rewrite -!(orx O C) -?(osf O).
Qed.

(* stable spec for setting the contents ct of reader counter to c *)
Program Definition write_ct (c : nat) : {h m n k}, STsep [rwx]
  (fun s => [/\ rho.s s = own, hp.s s = h,
     kappa.s s = m, kappa.o s = n & alpha.s s = k],
   fun _ s => [/\ rho.s s = own, hp.s s = h,
     kappa.s s = m, kappa.o s = n, ct.j s = c & alpha.s s = k]) :=
  Do (actSx (fun _ : unit => writect_t c)).
Next Obligation. by split=>// ??? [[]]. Qed.
Next Obligation.
move=>c s [h][m][n][k][[/= H1 H2 H3 H4 H5]] C; apply: vrf_atm=>// j O _.
rewrite !(orx O C H1) in H4; rewrite !(osf O) {O C} in H1 H2 H3 H5.
eexists tt, _; rewrite effE.
by split=>// x O C; rewrite -!(orx O C) -?(osf O) ?lE.
Qed.

(* stable spec for reading the shared heap if registered as a reader *)
Lemma rdhM A (x : ptr) (r : A) : monoP xcoh (fun s => [pcm x :-> r <= hpj.j s]).
Proof. by move=>???; rewrite !lE. Qed.

Program Definition read A x : {(v : A) m h k}, STsep [rwx]
  (fun s => [/\ kappa.s s = m.+1, hpj.j s = x :-> v \+ k & hp.s s = h],
   fun r s => [/\ kappa.s s = m.+1, hpj.j s = x :-> v \+ k,
     hp.s s = h & r = v]) :=
  Do (actSx (fun r : A => id_t (@rdhM A x r))).
Next Obligation.
split=>// ????? [_ _ _ /validL/validR V _ _ _][/=][h1] E1 _ [[h2 E2]] _.
by rewrite E1 in V E2; case/(hcancelV V): E2.
Qed.
Next Obligation.
move=>A x i [v][m][_][k][[/= H1 H2 <-]] C; apply: vrf_atm=>// j O _.
rewrite !(ohp O C) ?H1 // in H2 *; rewrite !(osf O) in H1 *.
exists v, j; rewrite effE; split=>[|s {}O {}C]; first by split=>//; exists k.
by rewrite -!(osf O) -(ohp O C) ?H1.
Qed.

End RWX.
End RWX.

Lemma kappaE a b c d : RWX.kappa (mk4 a b (up c) d) = c. Proof. by []. Qed.

Canonical RWX.alpha_morph.
Canonical RWX.kappa_morph.
Canonical RWX.hp_morph.
Canonical RWX.rho_morph.
Arguments RWX.open_t {count I}.
Arguments RWX.close_t {count I}.
Arguments RWX.getrd1_t {count I}.
Arguments RWX.relrd1_t {count I}.
Arguments RWX.incrd_t {count I}.
Arguments RWX.decrd_t {count I}.
Arguments RWX.acqauth_t {count I}.
Arguments RWX.relauth_t {count I}.
Arguments RWX.acqro_t {count I}.
Arguments RWX.relro_t {count I}.
Arguments RWX.writect_t {count I}.
Arguments RWX.id_t {count I}.


(**************************************************************************)
(* Combining RWX with two instances of Spin                               *)
(* (one for writers one for readers)                                      *)
(**************************************************************************)
(**************************************************************************)
(**************************************************************************)
(* The combination follows the three stage process described for CSL, but *)
(* with somewhat different invariants. The details are given in the       *)
(* accompanying paper.                                                    *)
(**************************************************************************)

Module RW.
Section RW.
Variables (wlk rlk count : ptr) (I : heap -> Prop).
Variable prec_pf : precise I.
Notation wspin := (Spin.spin wlk).
Notation rspin := (Spin.spin rlk).
Notation rwx := (RWX.rwx count prec_pf).

Definition U := [pcm of Spin.U * RWX.U * Spin.U].
Definition T : Type := Spin.T * RWX.T * Spin.T.

(* We have to make the tensor of 3 resources *)
(* we proceed in two steps.                  *)

(* First, combine wspin with rwx *)
(* we couple all internals with id of the other resource *)

Definition wrwcouple (t1 : Tr wspin) (t2 : Tr rwx) :=
  match t1, t2 with
  | Spin.lock_t, RWX.id_t _ _ => true | Spin.relperm_t, RWX.acqauth_t => true
  | Spin.id_t _ _, RWX.open_t => true | Spin.id_t _ _, RWX.close_t => true
  | Spin.acqperm_t, RWX.relauth_t => true | Spin.unlock_t, RWX.id_t _ _ => true
  | Spin.id_t _ _, RWX.acqro_t => true | Spin.id_t _ _, RWX.relro_t => true
  | Spin.id_t _ _, RWX.getrd1_t => true | Spin.id_t _ _, RWX.relrd1_t => true
  | Spin.id_t _ _, RWX.incrd_t => true | Spin.id_t _ _, RWX.decrd_t => true
  | Spin.id_t _ _, RWX.writect_t n => true | Spin.id_t _ _, RWX.id_t _ _ => true
  | _, _  => false
  end.

Definition wrwints (t1 : Tr wspin) (t2 : Tr rwx) : bool :=
  match t1, t2 with
  | Spin.relperm_t, RWX.acqauth_t => true | Spin.acqperm_t, RWX.relauth_t => true
  | _, _  => false
  end.

Lemma wrwpint : footpresP wrwints.
Proof.
case=>[||||??][] //= _ s s' _ []; rewrite /flat /= /flat /=;
by case=>_ -> [_ ->]; rewrite !lE.
Qed.

(* The auxiliary tensor product *)
Definition wrw := tensor wrwcouple wrwpint.

(* Morphism from wspin and rwx to wrw *)
Definition wrwcouple_l (t : Tr wspin) : Tr rwx :=
  match t with
  | Spin.relperm_t => RWX.acqauth_t | Spin.acqperm_t => RWX.relauth_t
  | _ => RWX.id_t _ idM
  end.
Program Definition inl1 := InlMorphPx wrwcouple wrwpint wrwcouple_l _.
Next Obligation. by split; case. Qed.

Program Definition inl1_sim : simulation inl1 :=
  inl_sim (couple:=wrwcouple) wrwpint (couple_l:=wrwcouple_l) _ _ _.
Next Obligation. by move=>t ?? s2; case: t; exists s2. Qed.

Definition wrwcouple_r (t : Tr rwx) : Tr wspin :=
  match t with
  | RWX.acqauth_t => Spin.relperm_t | RWX.relauth_t => Spin.acqperm_t
  | _ => Spin.id_t _ idM
  end.
Program Definition inr1 := InrMorphPx wrwcouple wrwpint wrwcouple_r _.
Next Obligation. by split; case. Qed.

Program Definition inr1_sim : simulation inr1 :=
  inr_sim (couple:=wrwcouple) wrwpint (couple_r:=wrwcouple_r) _ _ _.
Next Obligation. by move=>t ?? s1; case: t; exists s1. Qed.

(* Coupling wrw with rspin *)

Definition couple (t1 : Tr wrw) (t2 : Tr rspin) :=
  match t1, t2 with
  | couple_t Spin.relperm_t RWX.acqauth_t _, Spin.id_t _ _ => true
  | couple_t Spin.lock_t (RWX.id_t _ _) _, Spin.id_t _ _ => true
  | couple_t (Spin.id_t _ _) RWX.open_t _, Spin.id_t _ _ => true
  | couple_t (Spin.id_t _ _) RWX.close_t _, Spin.id_t _ _ => true
  | couple_t Spin.acqperm_t RWX.relauth_t _, Spin.id_t _ _ => true
  | couple_t Spin.unlock_t (RWX.id_t _ _) _, Spin.id_t _ _ => true
  | Tensor.id_t _ _, Spin.lock_t => true
  | couple_t (Spin.id_t _ _) (RWX.acqro_t) _, Spin.relperm_t => true
  | couple_t (Spin.id_t _ _) (RWX.relro_t) _, Spin.acqperm_t => true
  | Tensor.id_t _ _ , Spin.unlock_t => true
  | couple_t (Spin.id_t _ _) (RWX.getrd1_t) _, Spin.id_t _ _ => true
  | couple_t (Spin.id_t _ _) (RWX.relrd1_t) _, Spin.id_t _ _ => true
  | couple_t (Spin.id_t _ _) (RWX.incrd_t) _, Spin.id_t _ _ => true
  | couple_t (Spin.id_t _ _) (RWX.decrd_t) _, Spin.id_t _ _ => true
  | couple_t (Spin.id_t _ _) (RWX.writect_t n) _, Spin.id_t _ _ => true
  | couple_t (Spin.id_t _ _) (RWX.id_t _ _) _, Spin.id_t _ _ => true
  | Tensor.id_t _ _ , Spin.id_t _ _ => true
  | _, _  => false
  end.

(* description of new internal transitions *)
Definition ints (t1 : Tr wrw) (t2 : Tr rspin) : bool :=
  match t1, t2 with
  | couple_t (Spin.id_t _ _) RWX.acqro_t _, Spin.relperm_t => true
  | couple_t (Spin.id_t _ _) (RWX.relro_t) _, Spin.acqperm_t => true
  | _, _  => false
  end.

Lemma pint : footpresP ints.
Proof.
case=>[[]|] // P pf [] // f; case=>// _ s s' _;
case; case=>H1 H2 _ _ _; do ![rewrite /flat /=];
by case: H1 H2=>_ -> [_ ->]; rewrite !lE domeqX.
Qed.

(* Tensor product between wrw and rspin *)
(* equivalently, a 3-way product between wspin, rwx, rspin *)
Definition rw' : resource := tensor couple pint.

(* names for the four new internal transitions *)
Definition relpermauth_t : Tr rw' :=
  couple_tx couple (couple_tx wrwcouple
    Spin.relperm_t RWX.acqauth_t) (Spin.id_t _ idM).
Definition acqpermauth_t : Tr rw' :=
  couple_tx couple (couple_tx wrwcouple Spin.acqperm_t RWX.relauth_t)
                   (Spin.id_t _ idM).
Definition acqroperm_t : Tr rw' :=
  couple_tx couple (couple_tx wrwcouple (Spin.id_t _ idM) RWX.acqro_t)
                   Spin.relperm_t.
Definition relroperm_t : Tr rw' :=
  couple_tx couple (couple_tx wrwcouple (Spin.id_t _ idM) RWX.relro_t)
                   Spin.acqperm_t.

(* morphism from from wrw to rw' *)
Definition couple_l (t : Tr wrw) : Tr rspin :=
  match t with
   | couple_t (Spin.id_t _ _) RWX.acqro_t _ => Spin.relperm_t
   | couple_t (Spin.id_t _ _) RWX.relro_t _ => Spin.acqperm_t
   | _ => Spin.id_t _ idM
  end.

Program Definition inl2 := InlMorphPx couple pint couple_l _.
Next Obligation.
split; first by case=>// x y; case: x; case: y.
by case=>//; case=>// P pf [] // H; case=>// [[]].
Qed.

Program Definition inl2_sim : simulation inl2 :=
  inl_sim (couple:=couple) pint (couple_l := couple_l) _ _ _.
Next Obligation.
by move=>t ?? s2 H K; exists s2; case: t H K=>// [[]??] // [] // ? [] // [].
Qed.

(* morphism from rspin to rw' *)
Definition couple_r (t : Tr rspin) : Tr wrw :=
  match t with
   | Spin.relperm_t => couple_tx wrwcouple (Spin.id_t _ idM) RWX.acqro_t
   | Spin.acqperm_t => couple_tx wrwcouple (Spin.id_t _ idM) RWX.relro_t
   | _ => Tensor.id_t _ idM
  end.

Program Definition inr' := InrMorphPx couple pint couple_r _.
Next Obligation. by split; case. Qed.

Program Definition inr'_sim : simulation inr' :=
  inr_sim (couple:=couple) pint (couple_r := couple_r) _ _ _.
Next Obligation. by move=>t ?? s1; case: t; exists s1. Qed.

(* morphism from wspin, rwx to rw' *)
Definition inw' : morph unit wspin rw' := comp_morph inl1 inl2.
Definition inx' : morph unit rwx rw' := comp_morph inr1 inl2.
Definition inw'_sim : simulation inw' := comp_sim inl1_sim inl2_sim.
Definition inx'_sim : simulation inx' := comp_sim inr1_sim inl2_sim.

(* we can reduce stability in rw' to stability of wspin, xfer and rspin *)
Lemma ost_wspin s1 s2 : osteps rw' s1 s2 -> osteps wspin s1\1\1 s2\1\1.
Proof. by move/ost_fst/ost_fst. Qed.

Lemma ost_rspin s1 s2 : osteps rw' s1 s2 -> osteps rspin s1\2 s2\2.
Proof. by move/ost_snd. Qed.

Lemma ost_xfer s1 s2 : osteps rw' s1 s2 -> osteps rwx s1\1\2 s2\1\2.
Proof. by move/ost_fst/ost_snd. Qed.

(* restricting rw' to rw *)

(* fields inheritted from wspin *)
Definition wpi : U -> _ := Spin.pi \o fst \o fst.
Canonical wpi_morph := Eval hnf in [morphism of wpi].
Definition ht : U -> _ := Spin.ht \o fst \o fst.
Canonical ht_morph := Eval hnf in [morphism of ht].
(* fields inheritted from rwx *)
Definition alpha : U -> _ := RWX.alpha \o snd \o fst.
Canonical alpha_morph := Eval hnf in [morphism of alpha].
Definition hp : U -> heap := RWX.hp \o snd \o fst.
Canonical hp_morph := Eval hnf in [morphism of hp].
Definition rho : U -> _ := RWX.rho \o snd \o fst.
Canonical rho_morph := Eval hnf in [morphism of rho].
Definition kappa : U -> _ := RWX.kappa \o snd \o fst.
Canonical kappa_morph := Eval hnf in [morphism of kappa].
Definition ct : T -> _ := RWX.ct \o snd \o fst.
Definition hpj : T -> _ := RWX.hpj \o snd \o fst.
(* fields inheritted from rspin *)
Definition rpi : U -> _ := Spin.pi \o snd.
Canonical rpi_morph := Eval hnf in [morphism of rpi].

(* additional morphisms specific to the combined resource *)
(* defined in terms of the above primitive fields *)

(* mu = own iff we are the writer, i.e., if we passed all *)
(* the stages of taking the writer lock: *)
(*  (1) we locked the writer lock and set wpi.s = own, *)
(*  (2) we transfered wpi to rwx, *)
(*  (3) we took the heap *)

Definition mu : U -> _ := 'res_('ker wpi) (mxundef \o alpha).
Canonical mu_morph := Eval hnf in [morphism of mu].

(* tx = own iff we hold the reader lock *)
(* strictly speaking, this morphism is not needed in the exported procedures *)
(* so it is not used in the paper, but we define it here for internal reasons *)

Definition tx : U -> _ := omg \o Spin.ht \o snd.
Canonical tx_morph := Eval hnf in [morphism of tx].

(* global invariant *)
Definition rwginv (s : state U T) :=
  valid (nxown (kappa.t s) \+ wpi.t s \+ mxown (alpha.t s)) /\
  last_val false (ht.t s) =
  (nxown (kappa.t s) \+ wpi.t s \+ mxown (alpha.t s) == own).

Lemma rwginvG : cglobalP (coh rw') rwginv.
Proof. by move=>p s /coh_valid A; rewrite /rwginv !lE. Qed.

(* subjective invariant *)
Definition rwsinv (x y : U) :=
  [&& 'sep tx x y, rpi x \+ rho x == tx x & rpi y \+ rho y == tx y].

(* the following lemma says that the subjective invariant is an equalizer *)
(* as claimed in the paper. the lemma is not needed, but we prove it *)
(* to confirm the claim from the paper *)

Lemma rwsinv_eqlz : rwsinv =s 'eqlz ('join rpi rho) tx.
Proof.
move=>x y V; apply/idP/idP; rewrite !sepE /= !sepE /=.
- case/and3P=>S H1 H2; apply/and4P.
  by rewrite (eqP H1) (eqP H2) !eq_refl (pfV2 _ V).
by case/and4P=>/= S H1 H2 H3; apply/and3P.
Qed.

Lemma rwsinv_sep_ax : seprel_axiom rwsinv.
Proof. by rewrite (sepXEQ rwsinv_eqlz); apply: sepax. Qed.
Canonical rwsinv_seprel := Eval hnf in seprel rwsinv rwsinv_sep_ax.
Definition newpcm : sub_pcm rwsinv := xsepSubPCM rwsinv_seprel.

(* The two invariants are indeed invariants, i.e., they are *)
(* preserved by the transitions of rw' *)

Lemma rwsginv_pers : @sub_persP rw' rwsinv rwginv.
Proof.
move=>t s x C; rewrite prodVE=>/andP [_ V] _ H.
rewrite /rwsinv /rwginv !lA /= !lprodE.
have [V1 V2] : valid (RWX.alpha.t s\1\2) /\ valid (Spin.ht.t s\1\1).
- by case: C; case=>C1 C2; case: C1=>/andP [_ ->]; case: C2=>->.
have V3 : valid (Spin.ht.t s\2) by case: C=>_ [/andP][].
case: t H=>[[t1 y X1 t2 X2|P pf t2 X2 [[_ ->]]]|??[_ ->] //]; last first.
- case: t2 X2=>[||||?? _ [_ ->]] //= _ [H3 ->] _; rewrite ?lE //=.
  (* id + id + lock *)
  - case/and3P=>S E -> [->->]; rewrite !sepE /omg_s /omg !lA !lastval_freshUn //=.
    rewrite lastkey_freshUn // ltnS lastkeyUnf //.
    by move: H3 E; rewrite !omgP //; do 2![case: mxPJ].
  (* id + id + unlock *)
  rewrite H3; case/and3P=>/(stVf (omg \o Spin.ht) V) S E ->.
  rewrite !sepE /omg_s /omg !lA !lastval_freshUn // !lastkey_freshUn //=.
  by move: S; rewrite omgP -(eqP E)=>/mxMxy /mxE0 [->->].
case: t1 y t2 X1 X2=>[||||P pf]; case=>[||||||||||n|Q qf] //;
case=>[||||R rf] //= _ _ [[[H1 ->]]][H2 -> _][H3 -> _]; rewrite ?lE //= ?kappaE.
- (* lock + id + id *)
  rewrite (negbTE H1) -!joinA lastval_fresh //= => -> [].
  by case: mxPJ=>//=; [move=>_ /mxE0 ->|case|case].
- (* unlock+ id + id *)
  by rewrite H1 -!joinA lastval_fresh // =>->; case=>/mxxOy/mxE0 [->->].
- (* relperm + acqauth + id *)
  by case: H2=>/mxE0 [->->] _; rewrite H1 !unitR=>-> [/mxxOy/mxE0 [->->]].
- (* acqperm + relauth + id *)
  by case: H1 H2 V1=>_ /mxE0 [->->] -> /mxMx ->; rewrite !unitR.
- (* id + open + id *)
  by move: H2 V1=>-> /mxMx ->; case: (RWX.kappa.t _).
- (* id + close + id *)
  by case: H2 V1=>-> _ _ /mxMx ->.
- (* id + getrd1 + id *)
  by case: H2 V1=>->-> /mxMx ->-> [] /mxxO/mxE0 [->->].
- (* id + reldrd1 + id *)
  by case: H2=>->->->-> [/mxMxy/mxE0][->]; case: mxPJ.
- (* id + incrd + id *)
  by case: (RWX.kappa.t _) H2=>[|n][->].
- (* id + decrd + id *)
  case: (RWX.kappa.s _) H2=>[|n][-> _] //=.
  by rewrite !(@pcm_joinE natPCM) /= addSnnS addnS ltnS; case: (_ + _).
- (* id + acqauth + relperm *)
  by rewrite (mxE0 H2) H3 !sepE /omg_s /omg !lA omgP=>/and3P [-> /eqP <-].
(* id + relauth + acqperm *)
case: H2=>-> _; case: H3=>_ /mxE0 [-> _].
by rewrite !sepE /omg_s /omg !lA omgP=>/and3P [-> /eqP <-].
Qed.

(* now rw is a subresource of rw' *)
Definition rw := @sub_res rw' rwsinv_seprel _ newpcm rwginvG rwsginv_pers.

(* morphisms from rw'/wspin/rwx/rspin to rw *)
Definition sub_morf : morph unit rw' rw :=
  @sub_morph rw' rwsinv_seprel _ _ rwginvG rwsginv_pers.
Definition sub_sim : simulation sub_morf :=
  @sub_sim rw' rwsinv_seprel _ _ rwginvG rwsginv_pers.
Definition inw : morph unit wspin rw := comp_morph inw' sub_morf.
Definition inx : morph unit rwx rw := comp_morph inx' sub_morf.
Definition inr : morph unit rspin rw := comp_morph inr' sub_morf.
Definition inw_sim : simulation inw := comp_sim inw'_sim sub_sim.
Definition inx_sim : simulation inx := comp_sim inx'_sim sub_sim.
Definition inr_sim : simulation inr := comp_sim inr'_sim sub_sim.

(* stability of subcomponents transfers *)

Lemma mwx s1 s2 :
        msteps (inw tt) s1 s2 -> osteps rwx (spval s1)\1\2 (spval s2)\1\2.
Proof.
by move/msub_spval_comp/msteps_inl; apply: ost_mst_inl_id=>[??|[]//]; case.
Qed.

Lemma mwr s1 s2 :
        msteps (inw tt) s1 s2 -> osteps rspin (spval s1)\2 (spval s2)\2.
Proof.
move/msub_spval_comp. apply: ost_mst_inl; last by move=>t _ _ J ->; case: t J.
by move=>?? [].
Qed.

Lemma mxw s1 s2 :
        msteps (inx tt) s1 s2 -> osteps wspin (spval s1)\1\1 (spval s2)\1\1.
Proof.
by move/msub_spval_comp/msteps_inl; apply: ost_mst_inr_id=>[??|[]//]; case.
Qed.

Lemma mxr s1 s2 :
        msteps (inx tt) s1 s2 -> osteps rspin (spval s1)\2 (spval s2)\2.
Proof.
move/msub_spval_comp; apply: ost_mst_inl; last by move=>t _ _ J ->; case: t J.
by move=>?? [].
Qed.

Lemma mrw s1 s2 :
        msteps (inr tt) s1 s2 -> osteps wspin (spval s1)\1\1 (spval s2)\1\1.
Proof. by move/msteps_subr=>M; apply: ost_fst; apply: M=>[??|[]//]; case. Qed.

Lemma mrx s1 s2 :
        msteps (inr tt) s1 s2 -> osteps rwx (spval s1)\1\2 (spval s2)\1\2.
Proof. by move/msteps_subr=>M; apply: ost_snd; apply: M=>[??|[]//]; case. Qed.


(* additional helper lemmas *)
Lemma rwV (s : st rw) :
       coh rw s -> valid (rpi.t (spval s) \+ rho.t (spval s)).
Proof.
move=>C; case: (C); rewrite spVE=>/andP [A] /and3P [/(stVf tx A) V].
rewrite !lA => /eqP X /eqP Y _ _; rewrite -X -Y /= in V.
by rewrite joinAC !joinA in V; rewrite joinA joinAC.
Qed.

Lemma txP s : coh rw s -> rho.s (spval s) = own -> tx.s (spval s) = own.
Proof.
move=>C E; case: C; rewrite spVE=>/andP [A] /and3P [/(stVf tx A) V].
by rewrite !lA E=>X; rewrite -(eqP X) /= in V *; case/mxyOx/mxE0: V=>->.
Qed.

Lemma muP s : coh rw s -> (alpha.s (spval s) = own <-> mu.s (spval s) = own).
Proof.
rewrite (lSXid mu) /mu !sepE /= !lA.
move=>C; case: C=>_ [V1 _][[_]][V2]; split=>/= E; last first.
- by case: ifP E=>// _; case: (alpha.s _)=>//; case.
move: V2 V1; rewrite -!lprodX /= E=>/mxMx ->.
by do 2![case: mxPJ=>//]=>_ /mxE0 ->.
Qed.

(* additional stability lemmas derived out of RWX.orx by lifing *)
Lemma orx s1 s2 :
        osteps rwx s1\1\2 s2\1\2 -> coh rw' s1 -> rho.s s1 = own ->
        (kappa.o s1 = kappa.o s2) * (ct.j s1 = ct.j s2).
Proof. by move=>O [[_ C ???]]; rewrite !lprodE=>H; rewrite !(RWX.orx O). Qed.

Lemma orx1 s1 s2 c :
        osteps rwx s1\1\2 s2\1\2 -> coh rw' s1 ->
        (rho.s s1 == own) ==> (ct.j s1 == c) ->
        (rho.s s2 == own) ==> (ct.j s2 == c).
Proof.
move=>O C /implyP H; apply/implyP=>/eqP X.
by rewrite -(orx O C) ?H // !lprodE (osf O) -!lprodX X.
Qed.

Lemma orx2 s1 s2 c :
        osteps rwx s1\1\2 s2\1\2 -> coh rw' s1 ->
        (rho.s s1 == own) ==> (kappa.o s1 == 0) && (ct.j s1 == c) ->
        (rho.s s2 == own) ==> (kappa.o s2 == 0) && (ct.j s2 == c).
Proof.
move=>O C /implyP H; apply/implyP=>/eqP X.
by rewrite -!(orx O C) ?H // !lprodE (osf O) -!lprodX X.
Qed.


(* Now, we implement the functions for locking and unlocking the
reader and writer locks, together with transfering permissions. These
will be used in the exported methods. *)

(* acquire the writer lock, whether reader or writer *)
(* locks writer in Spin, then transfers permission to Xfer *)
Program Definition wrlock : {r t c}, STsep [rw]
  (fun s : st rw =>
    [/\ hp.s (spval s) = Unit, rho.s (spval s) = r,
        tx.s (spval s) = t & (r == own) ==> (ct.j (spval s) == c)],
   fun _ s =>
     [/\ alpha.s (spval s) = auth1,
         hp.s (spval s) = Unit, rho.s (spval s) = r,
         tx.s (spval s) = t & (r == own) ==> (ct.j (spval s) == c)]) :=
  Do (morph_app inw tt (Spin.genlock wlk);;
      morph_app sub_morf tt (actSx (fun _ : unit  => relpermauth_t))).
Next Obligation. by split; [move=>??? [][]|case; left; split=>//; right]. Qed.
Next Obligation.
move=>m [_][_][c][[/= H <-<- E]] C; apply: step.
apply: (vrf_morph inw_sim)=>[|//|_]; first by eexists; split=>//; eexists.
apply: [gE _, last_key _]=>//.
case: C=>_ _ C' [j][t][Wpu _ _] _ [x][_][->][_][->] X _ C /(mxms inw) M.
rewrite {j}X in Wpu *; rewrite !lprodE !(osf (mwx M)) !(osf (mwr M)) in H *.
move/(orx1 (mwx M) C'): E=>E; case: (C)=>_ _ {}C'.
apply: (vrf_morph sub_sim)=>// _; apply: vrf_atm=>// k O /coh_flat_valid V.
case: (ospval_subX O C)=>y [?] O' _; subst k.
rewrite -!lprodX in H Wpu *; rewrite !(osf O) in H Wpu *.
move/(orx1 (ost_xfer O) C'): E=>E {O}.
have {O' C'} [Wx Rt]: (alpha.t (spval y) = nown /\ kappa.t (spval y) = 0).
- case: (ocoh O' C)=>_ []; rewrite !Wpu -!joinA.
  by move/mxxOy; do 3!case: mxPJ=>//; case: nxP.
rewrite !lprodE in H Wpu Wx Rt; eexists tt, _; rewrite effE; split.
- apply: act2_safe=>//.
  - by apply: act2_safe=>//; rewrite /flat /= !lE (validX V).
  by rewrite flatPE /flat /= !lE (validX V).
move=>k O {}C [j] X _ _ /mxms M2; rewrite {k}X in O.
have Rx : rho.s (spval y) = rho.s (spval j) by rewrite -!(osf O) !lprodE !lE.
split=>//; rewrite -?(osf O) ?lprodE ?lE // -!lprodX Rx.
by rewrite (orx1 (ost_xfer O) C) // !lprodE !lE -!lprodX.
Qed.

(* release the writer lock, whether reader or writer *)
(* transfers permission from Xfer to writer Spin, then unlocks *)
Program Definition wrunlock : {r t}, STsep [rw]
  (fun s: st rw =>
    [/\ alpha.s (spval s) = auth1, hp.s (spval s) = Unit,
        rho.s (spval s) = r, tx.s (spval s) = t &
        (r == own) ==> (kappa.o (spval s) == 0) && (ct.j (spval s) == 1)],
   fun _ s =>
     [/\ mu.s (spval s) = nown, hp.s (spval s) = Unit,
         kappa.s (spval s) = 0, rho.s (spval s) = r, tx.s (spval s) = t &
         (r == own) ==> (kappa.o (spval s) == 0) && (ct.j (spval s) == 1)]) :=
  Do (morph_app sub_morf tt (actSx (fun _ : unit  => acqpermauth_t));;
      morph_app inw tt (Spin.genunlock wlk)).
Next Obligation. by split; [move=>??? [][]|case; left; split=>//; right]. Qed.
Next Obligation.
move=>m [_][_][[Wx H <-<- E]] C; case: (C)=>_ _ C'; apply: step.
have Rs : kappa.s (spval m) = 0.
- by case: C=>_ []; rewrite Wx joinC; do 3!case: mxPJ=>//; case: nxPJ.
apply: (vrf_morph sub_sim)=>// _.
apply: vrf_atm=>//= j O /coh_flat_valid V.
rewrite !(osfw rw' O) in Wx Rs H *; move/(orx2 (ost_xfer O) C'): E =>E.
case: (ospval_subX (s1:=m) O C)=>x [?] O' _; subst j.
have [Ht Wpu] : [/\ last_val false (ht.t (spval x)) & wpi.t (spval x) = nown].
- case: (ocoh O' C)=>_ []; rewrite Wx joinC.
  by do 2![case: mxPJ]=>// t1 _ _ t2 <- /mxE0 [].
rewrite !lprodE in Wx Rs H Ht Wpu; eexists tt, _; rewrite effE; split.
- apply: act2_safe=>//.
  - by apply: act2_safe=>//; rewrite /flat /= !lE (validX V).
  by rewrite flatPE /flat /= !lE (validX V).
move{O' C}=>k O' C [y] X _ [_ _ Cy'] /(mxms sub_morf)=>M; rewrite {k}X in O'.
apply: (vrf_morph inw_sim)=>[||_] //; first by eexists; split=>//; eexists.
have {}Wpu : Spin.pi.s (spval y)\1\1 = own.
-by rewrite -!lprodX -(@osf rw' _ _ _ _ O') !lprodE !lE.
apply: [gE _, last_key _]=>//=.
case=>k [t][{} Wpu _ _] _ [z][_][->][_][->] X _ _ /(mxms inw) M2.
rewrite {k}X in Wpu; have Rx : rho.s (spval x) = rho.s (spval z).
- rewrite !lprodE -!(osf (mwx M2)) -!lprodX -(@osf rw' _ _ _ _ O').
  by rewrite !lprodE !lE.
rewrite lSXid /mu !sepE /= !lA /= !lprodE -(osf (mwr M2)).
rewrite -!(osf (mwx M2)) -!lprodX -!(osfw rw' O') !lprodE !lE Wpu.
split=>//; rewrite -!lprodX Rx (orx2 (mwx M2) Cy') //.
by rewrite (orx2 (ost_xfer O') C) // !lprodE !lE -!lprodX.
Qed.

(* acquire reader lock *)
(* locks reader in Spin then transfers permission to Xfer *)
Program Definition rdlock : {r}, STsep [rw]
  (fun s: st rw => [/\ kappa.s (spval s) = r & hp.s (spval s) = Unit],
   fun _ s =>
    [/\ kappa.s (spval s) = r, hp.s (spval s) = Unit,
        rho.s (spval s) = own & ct.j (spval s) = kappa.t (spval s)]) :=
  Do (morph_app inr tt (Spin.genlock rlk);;
      morph_app sub_morf tt (actSx (fun _ : unit  => acqroperm_t))).
Next Obligation. by split=>[??? [][]|] //; right. Qed.
Next Obligation.
move=>m [_][[/= <- H]] _; apply: step.
apply: (vrf_morph inr_sim)=>[|//|_]; first by eexists.
apply: [gE _, last_key _]=>//; case=>j [t][Rpu _ _] _ [x][_][->] X _ C /(mxms inr) M.
rewrite {j}X in Rpu *; rewrite !lprodE !(osf (mrx M)) {m M} in H *.
apply: (vrf_morph sub_sim)=>// _.
apply: vrf_atm=>//= j O C'; move: (coh_flat_valid C')=>V.
case: (ospval_subX (W:=rw') O C)=>z [?] _ {} C; subst j.
rewrite -!lprodX !(osfw rw' O) !lprodE {O} in H Rpu *.
have [Rx E] : RWX.rho.t (spval z)\1\2 = nown /\
              ct.j (spval z) = kappa.t (spval z).
  case: (C)=>_ _ [[_ Cx _ _ _]]; case: Cx; move/rwV: C.
  by rewrite !lprodE Rpu=>/mxOxy/mxE0 [_ ->] ????? /eqP ->.
eexists tt, _; rewrite effE; split=>[|m {C C'} O C [y] X _ _ _].
- apply: act2_safe=>//.
  - by apply: act2_safe=>//; rewrite /flat /= !lE (validX V).
  by rewrite flatPE /flat /= !lE (validX V).
rewrite {m}X in O; split; rewrite -!(osfw rw' O) !lprodE !lE /= ?rdE //.
by rewrite -!lprodX -!(orx (ost_xfer O) C) !lprodE !lE // -!lprodX.
Qed.

(* release reader lock, after the number of readers has been adjusted *)
(* transfers permission from Xfer to reader Spin, then unlocks *)
Program Definition rdunlock : {r}, STsep [rw]
  (fun s: st rw => [/\ kappa.s (spval s) = r, hp.s (spval s) = Unit,
                mu.s (spval s) = nown,
                rho.s (spval s) = own & ct.j (spval s) = kappa.t (spval s)],
   fun (_ : unit) s =>
    [/\ kappa.s (spval s) = r, hp.s (spval s) = Unit, tx.s (spval s) = nown &
        mu.s (spval s) = nown]) :=
  Do (morph_app sub_morf tt (actSx (fun _ : unit  => relroperm_t));;
      morph_app inr tt (Spin.genunlock rlk)).
Next Obligation. by split=>[???[][]|] //; right. Qed.
Next Obligation.
move=>m [_][[<- H Wtx Rx E]] C; apply: step; case: (C)=>_ _ C'.
apply: (vrf_morph sub_sim)=>// _; apply: vrf_atm2=>// j O /coh_flat_valid =>V.
rewrite !(orx (ost_xfer O) C') // !(osf O) in E.
rewrite !(osf O) in H Wtx Rx *; case: (ospval_subX O C)=>x [?] O' _; subst j.
move: (ocoh O' C)=>{} C; case: (C)=>_ _ {} C'.
have [X Rpu] : last_val false (Spin.ht.t (spval x)\2) /\
               Spin.pi.t (spval x)\2 = nown.
- case/coh_sep: (C)=>/and3P []; rewrite !lA /= => S.
  move: (stVf tx (coh_valid C') S); rewrite !lprodE in S *.
  move=>W X Y _ [_ /coh_valid A] _; rewrite omgP // -(eqP X) -(eqP Y) in W *.
  by rewrite -!lprodX Rx in W *; case/mxyMx/mxE0: W=>-> /mxE0 [->->].
rewrite !lprodE in H Rx Wtx E; eexists tt, _; rewrite effE; split.
- apply: act2_safe=>//.
  - by apply: act2_safe=>//=; rewrite /flat /= !lE (validX V).
  by rewrite flatPE /flat /= !lE (validX V).
move=>k {}O _ J [y] Y _ _ _; rewrite {k}Y in O.
case/(ospval_subX1 relroperm_t O C): J=>i [{C'} _ {}O' /(ocoh O'){O'} {}C].
apply: (vrf_morph inr_sim)=>[|//|_]; first by eexists.
have {} Rpu : Spin.pi.s (spval y)\2 = own.
- by rewrite -!lprodX -(osf O) !lprodE !lE.
apply: [gE _, last_key _]=>[//|/=[n][t][{}Rpu Ht]].
have {}V : valid (Spin.ht.t (spval y)\2) by case: C=>_ _ [_][/andP][].
rewrite lastkeyUn {}V gtn_max=>Lk {C} _ [q][_][->] Y _ C /(mxms inr) M /=.
rewrite {n}Y in Rpu Ht; move: Wtx; have N : t != 0 by case: t Lk Ht.
have V : valid (Spin.ht.s (spval q)\2) by case: C=>_ _ [_][/andP][_ /validL].
rewrite !(lSXid mu) /mu !sepE /= !lA !lprodE -!(osf (mrx M)) -!(osf (mrw M)).
rewrite -!lprodX -!(osf O) !lprodE !lE /= !kappaE; split=>//.
rewrite lSXid /= /omg -lSXid Ht lastvalUn -Ht V lastkeyPt //.
by case: ltngtP Lk=>//; rewrite lastvalPt.
Qed.

(* The remaining code in this module is simple lifting of stable specs
from RWX. *)

Program Definition read_ct : {m n k}, STsep [rw]
  (fun s : st rw => [/\ rho.s (spval s) = own, hp.s (spval s) = Unit,
    kappa.s (spval s) = m, kappa.o (spval s) = n & ct.j (spval s) = k],
   fun (r : nat) s => [/\ r = k,
     rho.s (spval s) = own, hp.s (spval s) = Unit,
     kappa.s (spval s) = m, kappa.o (spval s) = n & ct.j (spval s) = k]) :=
  Do (morph_app inx tt (RWX.read_ct count prec_pf)).
Next Obligation.
move=>s [m][n][k][/= H] _; apply: (vrf_morph inx_sim)=>[|//|_].
- by eexists; split=>//; eexists.
rewrite !lprodE in H; apply: [gE Unit, m, n, k]=>//= r z {}H _.
by case=>y [_][->][_][/= ->] E _ _ _; rewrite {z}E in H; rewrite !lprodE.
Qed.

Program Definition write_ct (k : nat) : {m n}, STsep [rw]
  (fun s : st rw => [/\ rho.s (spval s) = own, hp.s (spval s) = Unit,
     kappa.s (spval s) = m, kappa.o (spval s) = n & mu.s (spval s) = nown],
   fun _ s => [/\ rho.s (spval s) = own, hp.s (spval s) = Unit,
     kappa.s (spval s) = m, kappa.o (spval s) = n, ct.j (spval s) = k &
     mu.s (spval s) = nown]) :=
  Do (morph_app inx tt (RWX.write_ct count prec_pf k)).
Next Obligation.
move=>k s [m][n][/= H] _; apply: (vrf_morph inx_sim)=>[|//|_].
- by eexists; split=>//; eexists.
rewrite !lprodE in H; apply: [gE Unit, m, n, _]=>//=; first by case: H.
case=>z H' _ [y][_][->][_][->] E _ _ /(mxms inx)/mxw O; rewrite {z}E in H'.
rewrite !lprodE; case: H'=>->->->->-> H'; split=>//; case: H=>_ _ _ _.
by rewrite !(lSXid mu) /mu /wpi !sepE /= !lA !lprodE -!(osf O) H'.
Qed.

Program Definition getrd1 : STsep [rw]
  (fun s: st rw => [/\ rho.s (spval s) = own, alpha.s (spval s) = auth1,
     hp.s (spval s) = Unit & ct.j (spval s) = 0],
   fun _ s => [/\ rho.s (spval s) = own, alpha.s (spval s) = nown,
      hp.s (spval s) = Unit, kappa.s (spval s) = 1,
      kappa.o (spval s) = 0 & ct.j (spval s) = 0]) :=
  Do (morph_app inx tt (RWX.getrd1 count prec_pf)).
Next Obligation.
move=>s [/= H] _; apply: (vrf_morph inx_sim)=>[|//|_].
- by eexists; split=>//; eexists.
rewrite !lprodE in H; apply: gE=>//=; case=>m {} H _.
by case=>z [_][->][_][->] E _ _ _; rewrite {m}E in H; rewrite !lprodE.
Qed.

Program Definition relrd1 : STsep [rw]
  (fun s: st rw => [/\ rho.s (spval s) = own, hp.s (spval s) = Unit,
     kappa.s (spval s) = 1, kappa.o (spval s) = 0 & ct.j (spval s) = 1],
   fun _ s => [/\ rho.s (spval s) = own,  alpha.s (spval s) = auth1,
     hp.s (spval s) = Unit, kappa.s (spval s) = 0,
     kappa.o (spval s) = 0 & ct.j (spval s) = 1]) :=
  Do (morph_app inx tt (RWX.relrd1 count prec_pf)).
Next Obligation.
move=>s [/= H] _; apply: (vrf_morph inx_sim)=>[|//|_].
- by eexists _;split=>//; eexists.
rewrite !lprodE in H; apply: [gE Unit, 1]=>//=; case=>m {} H _.
by case=>z [_][->][_][->] E _ _ _; rewrite {m}E in H; rewrite !lprodE.
Qed.

Program Definition incrd : {n}, STsep [rw]
  (fun s: st rw => [/\ rho.s (spval s) = own, hp.s (spval s) = Unit,
      kappa.s (spval s) = 0, kappa.o (spval s) = n.+1 & ct.j (spval s) = n.+1],
   fun _ s => [/\ rho.s (spval s) = own, hp.s (spval s) = Unit,
      kappa.s (spval s) = 1, kappa.o (spval s) = n.+1 &
      ct.j (spval s) = n.+1]) :=
  Do (morph_app inx tt (RWX.incrd count prec_pf)).
Next Obligation.
move=>s [n][/= H] _; apply: (vrf_morph inx_sim)=>[|//|_].
- by eexists; split=>//; eexists.
rewrite !lprodE in H; apply: [gE 0, n.+1, n.+1]=>//=; first by case: H.
by case=>z X Y [y][_][->][_][->] E _ _ _; rewrite {z}E in X Y; rewrite !lprodE.
Qed.

Program Definition decrd : {n}, STsep [rw]
  (fun s: st rw => [/\ rho.s (spval s) = own, hp.s (spval s) = Unit,
     kappa.s (spval s) = 1, kappa.o (spval s) = n.+1 & ct.j (spval s) = n.+2],
   fun _ s => [/\ rho.s (spval s) = own, hp.s (spval s) = Unit,
     kappa.s (spval s) = 0, kappa.o (spval s) = n.+1 & ct.j (spval s) = n.+2]) :=
  Do (morph_app inx tt (RWX.decrd count prec_pf)).
Next Obligation.
move=>s [n][/= H] _; apply: (vrf_morph inx_sim)=>[|//|_].
- by eexists; split=>//; eexists.
rewrite !lprodE in H; apply: [gE Unit, 1, n.+1, n.+2]=>//=; first by case: H.
by case=>z X Y [y][_][->][_][->] E _ _ _; rewrite {z}E in X Y; rewrite !lprodE.
Qed.

(************************)
(* Procedures to export *)
(************************)

(* we express the specs in terms of 4 morphisms *)
Definition hhp : Ux rw -> heap := hp \o pval. (* private heap *)
Canonical hhp_morph := Eval hnf in [morphism of hhp].
Definition kkappa : Ux rw -> _ := kappa \o pval. (* number of readers *)
Canonical kkappa_morph := Eval hnf in [morphism of kkappa].
Definition mmu : Ux rw -> _ := mu \o pval. (* are we a writer? *)
Canonical mmu_morph := Eval hnf in [morphism of mmu].
Definition ttx : Ux rw -> _ := tx \o pval. (* do we have the reader lock? *)
Canonical ttx_morph := Eval hnf in [morphism of ttx].
(* We expose ttx in the specs below, though we elide it in the paper *)
(* the info of ttx seems pointless for the clients, since *)
(* clients shouldn't access reader lock directly, but only via *)
(* prologue and epilogue. Nevertheless, we leave ttx in for now *)
(* as it is easy to remove *)

(* Take the writer lock, and become a writer, i.e., take the heap. *)
(* Doesn't change the staus of reader lock, so we set ttx = nown *)
(* and by framing, ttx can be own as well, and thus removed *)
Program Definition wr_lock : STsep [rw]
  (fun s => [/\ hhp.s s = Unit & ttx.s s = nown],
   fun _ s => [/\ mmu.s s = own, I (hhp.s s) & ttx.s s = nown]) :=
  Do (wrlock;;
      morph_app inx tt (RWX.open count prec_pf)).
Next Obligation.
move=>m [[H Rtx]] C; rewrite !lsubX in H Rtx; have Rx : rho.s (spval m) = Unit.
- by case/coh_sep: C=>/and3P [_]; rewrite !lA Rtx; case: mxPJ=>// _ /mxownP.
apply: [stepE Unit, Unit, 0]=>//; case=>k[{H Rx Rtx} Wx H Rx Rtx] _ {C m} _.
rewrite !lprodE in Wx H Rx Rtx.
apply: (vrf_morph inx_sim)=>[|//|_]; first by eexists; split=>//; eexists.
apply: gE=>//=; case{Wx H}=>m [Wx H] _ [j][_][->][_][-> E] _ C /(mxms inx) M.
rewrite {m}E in Wx H; rewrite !lsubX !lprodE -!(osf (mxr M)).
by split=>//; apply/muP=>//; rewrite !lprodE.
Qed.

(* If writer, release the writer lock. Also set ttx = nown. *)
Program Definition wr_unlock : STsep [rw]
  (fun s => [/\ mmu.s s = own, I (hhp.s s) & ttx.s s = nown],
   fun _ s => [/\ mmu.s s = nown, hhp.s s = Unit & ttx.s s = nown]) :=
  Do (morph_app inx tt (RWX.close count prec_pf);;
      wrunlock).
Next Obligation.
move=>m [[Wtx H Rtx]] C; rewrite !lsubX in Wtx H Rtx; move/(muP C): Wtx=>Wx.
apply: step; apply: (vrf_morph inx_sim)=>[|//|_].
- by eexists; split=>//; eexists.
rewrite !lprodE in Wx H ; apply: gE=>//=.
case=>{Wx H}j [Wx H] _ [k][_][->][_][->] E _ _ /(mxms inx) M.
rewrite {j}E in Wx H; rewrite !lprodE (osf (mxr M)) -!lprodX in Wx H Rtx.
move: (mcoh M C)=>{m M}{}C; have Rx : rho.s (spval k) = Unit.
- by case/coh_sep: C=>/and3P [_]; rewrite !lA Rtx; case: mxPJ=>// _ /mxownP.
by apply: [gE Unit, Unit]=>//=; case=>m; case; rewrite !lsubX.
Qed.

(* Post says: we don't have reader lock in the end *)
(* and the self heap is unchanged. We don't explicitly say *)
(* that we're not a writer, because it follows from having the *)
(* reader count > 0 and the invariant *)

Program Definition prologue : STsep [rw]
  (fun s => [/\ kkappa.s s = 0 & hhp.s s = Unit],
   fun _ s => [/\ kkappa.s s = 1, hhp.s s = Unit & ttx.s s = nown]) :=
  Do (rdlock;;
      x <-- read_ct;
      (if x == 0 then wrlock;; getrd1 else incrd);;
      write_ct x.+1;;
      rdunlock).
Next Obligation.
move=>m [/= H] C; rewrite !lsubX in H.
apply: [stepE 0]=>//=; case=>k [Rs {}H Rx E]{m}C.
apply: [stepE 0, RW.kappa.o _, ct.j _]=>//=.
rewrite {}E {}Rs pcmE /= add0n =>{C k Rx H} _ m [-> Rx H Rs <- E] C.
have Rtx : tx.s (spval m) = own by apply: txP.
apply: step; case Ro: (kappa.o _) E=>[|n] E /=;
[apply: [stepE own, own, 0]=>//=; first by [rewrite E];
  case=>{H Rx Rtx Rs Ro E}k [Wx H Rx Rtx /eqP E] {m}C;
  apply: gE=>//=; case=>{Rx Wx H E Rtx}x [ Rx _ H Rs Ro E] {k} C |
 apply: [gE n]=>//=; case=>{H Rx Rtx Rs Ro E}x [ Rx H Rs Ro E] {m} C];
have Wtx : RW.mu.s (spval x) = nown
  by [case: C=>_ []; rewrite Rs (lSXid RW.mu) /RW.mu !sepE /= !lA;
      do !case: mxPJ=>//];
apply: [stepE 1, kappa.o _]=>//=;
rewrite {}Ro; case=>{Wtx Rs Rx H E}y [ Rx H Rs Ro E Wtx]{}C;
by apply: [gE 1]=>//=; rewrite ?Rs ?Ro //; case=>m []; rewrite !lsubX=>->->->.
Qed.

(* Post says: we deregistered one reader, the self heap is unchanged *)
(* we don't have reader lock, and we are not a writer *)
(* Notice that we don't say anything about last_val (ht.s s) = false *)
(* i.e., about the writer lock not being taken *by us*. *)
(* We can't guarantee that, because as a thread we could have locked *)
(* the writer lock, then forked a couple of readers; *)
(* the epilogue will deregister one of them, but we as a thread may still *)
(* hold the lock on behalf of remaining readers *)
(* Thus, we only say that we are not a writer (wtx.s = nown) *)
(* though we may actually be the thread that last locked the writer lock *)

Program Definition epilogue : STsep [rw]
  (fun s => [/\ kkappa.s s = 1 & hhp.s s = Unit],
   fun _ s => [/\ kkappa.s s = 0, hhp.s s = Unit,
                  ttx.s s = nown & mmu.s s = nown]) :=
  Do (rdlock;;
      x <-- read_ct;
      (if x == 1 then relrd1;; wrunlock else decrd);;
      write_ct x.-1;;
      rdunlock).
Next Obligation.
move=>m [/= H] C; rewrite !lsubX in H.
apply: [stepE 1]=>//=; case=>k [Rs {}H Rx E]{}C{m}.
apply: [stepE 1, RW.kappa.o _, ct.j _]=>//=.
rewrite {}E {}Rs pcmE /= add1n=>{Rx H k C}_ m [ -> Rx H Rs <- E] C.
apply: step; case Ro: (kappa.o _) E=>[|n] E /=;
[apply: stepE=>//=; case=>{Rx Rs Ro E H C m}k [ Rx Wx H Rs Ro E] C;
  have Rtx : RW.tx.s (spval k) = own by [apply: txP];
  apply: [gE own, own]=>//=; first by [rewrite Ro E];
  case=>{Wx H Rs Rx Rtx Ro E C k}m [ Wtx H Rs Rx Rtx /andP [/eqP Ro /eqP E]] C|
 apply: [gE n]=>//=; case=>{Rx Rs Ro E H C m}k [ Rx H Rs Ro E] C;
 have Wtx: mu.s (spval k) = nown
   by case: C=>_ []; rewrite Ro (lSXid RW.mu) /RW.mu !sepE /= !lA;
      case: nxPJ; do !case: mxPJ=>//];
apply: [stepE 0, kappa.o _]=>//=;
rewrite {}Ro; case=>{Rx H Rs E Wtx C}z [Rx H Rs Ro E Wtx] C;
by apply: [gE 0]=>//=; rewrite ?Rs ?Ro //; case=>? []; rewrite !lsubX.
Qed.

(* read from the pointer x, if x we can prove that x is in the shared heap *)
(* and we are a writer *)

Program Definition read A x : {(v : A) k}, STsep [rw]
  (fun s => [/\ kkappa.s s = 1, hpj.j s = x :-> v \+ k & hhp.s s = Unit],
   fun r s => [/\ kkappa.s s = 1, hpj.j s = x :-> v \+ k,
     hhp.s s = Unit & r = v]) :=
  Do (morph_app inx tt (RWX.read count prec_pf A x)).
Next Obligation.
move=>A x s [v][k][/= H] _; apply: (vrf_morph inx_sim)=>[|//|_].
- by eexists; split=>//; eexists.
rewrite !lsubE !lprodE in H; apply: [gE v, 0, Unit, k]=>//=.
move=>_ m [{H} H1 H2 H3] -> C [y][_][->][_][->] E _ _ _.
by rewrite {m}E in H1 H2 H3 C; rewrite !lsubE !lprodE.
Qed.

End RW.
End RW.
