From Coq Require Import ssreflect ssrbool ssrfun.
From mathcomp Require Import ssrnat eqtype seq path.
From fcsl Require Import axioms prelude pred pcm automap unionmap heap.
From fcsl Require Import morphism notation state world model priv.
From fcsl Require Import options.

(********************************************************)
(********************************************************)
(* Verification of a concurrent spanning-tree algorithm *)
(********************************************************)
(********************************************************)

(************************************************)
(* Graphs in adjacency-matrix representation    *)
(* (edge relation is boolean matrix of nodes)   *)
(************************************************)

Section GraphTheory.
Variables (node : ordType) (edge : node -> node -> bool).
Implicit Type t : union_map node unit.

(* front of t is included in t' *)
Definition front t t' := {subset dom t <= dom t'} /\
                         forall x y, x \in dom t -> edge x y -> y \in dom t'.
Definition maximal t := front t t.

(* tree rooted at x is a subgraph that has unique paths from x *)
Definition tree (x : node) t :=
  [/\ x \in dom t &
      forall y, y \in dom t ->
        exists !p, [/\ y = last x p, path edge x p & {subset p <= dom t}]].

(* set of nodes t is connected from x if it has a path from x to every other node *)
Definition connected x t :=
  forall y, y \in dom t -> exists p, path edge x p /\ last x p = y.

(**********************)
(* Lemmas about trees *)
(**********************)

(* a tree has no cycles *)
Lemma tree_cycle x y t p :
        tree x t -> {subset p <= dom t} -> y \in dom t -> ~~ cycle edge (y::p).
Proof.
case=>_ T Sp Dy; apply/negP => /= Py.
case: (T y Dy) => p' [[Exy Px Sp'] Uq].
suff : p' = p' ++ rcons p y.
- by elim: {Exy Px Sp' Uq Sp Py} p'=>[|x' p' IH]; [case: p | case].
apply: Uq; rewrite last_cat last_rcons cat_path Px -Exy; split=>//.
move=>z; rewrite mem_cat mem_rcons inE.
by case/or3P; [apply: Sp' | move/eqP=>-> | apply: Sp].
Qed.

(* we can never reach out of a maximal t *)
Lemma max_path t x p :
        maximal t -> path edge x p -> x \in dom t -> {subset x :: p <= dom t}.
Proof.
case=>_ M Px Dx; elim: p x Dx Px=>[|y p IH] x Dx //=.
- by move=>_ z; rewrite inE; move/eqP => ->.
case/andP => /(M _ _ Dx) Dy Py z; rewrite inE.
by case/orP; [move/eqP=>-> | apply: IH Dy Py z].
Qed.

Lemma tree_edge x y t : edge x y -> tree y t -> x \notin dom t.
Proof.
move=>H T; case: (T)=>D Tx; apply/negP => /Tx [p][Ex _]; case: Ex=>Ex P S.
by move: (tree_cycle T S D); rewrite /= rcons_path P -Ex H.
Qed.

(* three basic front lemmas *)

Lemma frontPt x t : {subset (edge x) <= dom t} -> x \in dom t -> front (#x) t.
Proof. by move=>E Dx; split=>[u|u v]; rewrite domPt inE=>/eqP <- // /E. Qed.

Lemma frontUn t1 t2 t : front t1 t -> front t2 t -> front (t1 \+ t2) t.
Proof.
case=>S1 F1 [S2 F2]; split=>[z|x y]; rewrite domUn; move/andP/proj2.
- by case/orP; [apply: S1 | apply: S2].
by case/orP; [apply: F1 | apply: F2].
Qed.

Lemma front_leq t t1 t2 : {subset dom t1 <= dom t2} -> front t t1 -> front t t2.
Proof.
move=>S [S1 F1]; split; first by move=>z /S1 /S.
by move=>x y Dx Ex; apply: S; apply: F1 Dx Ex.
Qed.

Definition if_edge x y :=
  [fun ty => if edge x y then tree y ty /\ maximal ty else ty = Unit].

Lemma max_join x y1 y2 ty1 ty2 :
        {subset edge x <= pred2 y1 y2} ->
        if_edge x y1 ty1 -> if_edge x y2 ty2 -> maximal (#x \+ (ty1 \+ ty2)).
Proof.
move=>B /= G1 G2; split=>// u v; rewrite domPtUn inE.
case/andP=>Vx; move: (validR Vx)=>V.
have [Dx1 Dx2] : x \notin dom ty1 /\ x \notin dom ty2.
- by rewrite validPtUn domUn inE V /= negb_or in Vx; case/andP: Vx.
rewrite domPtUn inE Vx /= !domUn inE V /= !inE; case/orP.
- by move/eqP=><-{u} E; move: {E} (B _ E) (E); case/orP=>/eqP -> E;
     [move: E G1 | move: E G2]=>->[[->]]; rewrite !orbT.
case/orP=>Du E; [move: G1 Du | move: G2 Du];
by case: ifP=>[_ [_ [_ M]] /M /(_ E) ->|_ ->]; rewrite ?dom0 ?orbT.
Qed.

Lemma tree_join x y1 y2 ty1 ty2 :
        {subset edge x <= pred2 y1 y2} -> valid (ty1 \+ ty2) ->
        if_edge x y1 ty1 -> if_edge x y2 ty2 -> tree x (#x \+ (ty1 \+ ty2)).
Proof.
move=>B V /= G1 G2; have L z t : if_edge x z t -> x \notin dom t.
- by simpl; case: ifP=>[E [/(tree_edge E) //]|_ ->]; rewrite dom0.
have [Dx1 Dx2] := (L _ _ G1, L _ _ G2).
have Vx : valid (#x \+ (ty1 \+ ty2)).
- by rewrite validPtUn V /= domUn inE V negb_or Dx1 Dx2.
split; first by rewrite domPtUn inE Vx eq_refl.
move=>u; rewrite domPtUn inE Vx /=; case/orP.
- move/eqP=><-{u}; exists [::]; split=>//.
  case=>//= y p [Exy /andP [H Px] _]; move/B: H (H) Exy Px G1 G2.
  case/orP=>/eqP ->{y} -> Exy Px; [case=>[[T _] M _]|move=>_ [[T _] M]].
  - by rewrite Exy (max_path _ _ _ (mem_last _ p)) // in Dx1.
  by rewrite Exy (max_path _ _ _ (mem_last _ p)) in Dx2.
rewrite domUn inE V /=.
wlog Du : y1 y2 ty1 ty2 B Dx1 Dx2 V Vx G1 G2 / u \in dom ty1.
- move=>G; case/orP=>H; [|rewrite (joinC ty1) in Vx V *].
  - by apply: (G y1 y2)=>//; rewrite H.
  by apply: (G y2 y1)=>//; [move=>z /B; rewrite !inE orbC | rewrite H].
move=>_; case: ifP G1 Du=>[Exy1 [[Dy1 Ty1] My1] Du|_ ->]; last by rewrite dom0.
case: (Ty1 u Du)=>p [][Ey1 Py1 Sy1] Uq.
have Du2 : u \notin dom ty2 by case: validUn V Du=>// _ _ Y _ /Y.
exists (y1 :: p); split=>/=.
- rewrite Exy1; split=>// z; rewrite inE => X.
  have Y : z \in dom ty1 by case: eqP X=>[->//|_]; apply: Sy1.
  by rewrite domPtUn inE Vx domUn inE V /= Y orbT.
case=>[|xx p'][/= Exy]; first by rewrite -Exy Du in Dx1.
case/andP=>Exy2; move: (B _ Exy2)=>H Px _.
case/orP: H Exy2 Exy Px=>/eqP ->{xx} Exy2 Exy Px; last first.
- move: G2; rewrite Exy2; case; case=>Dy2 _ My2.
  by rewrite Exy (max_path My2 Px Dy2 (mem_last _ p')) in Du2.
rewrite (Uq p') //; split=>// z X.
by apply: (max_path My1 Px Dy1); rewrite inE X orbT.
Qed.

End GraphTheory.

(**************************************************)
(* Binary graphs in adjacency-list representation *)
(*                                                *)
(* nodes keep left/right pointers to children     *)
(* if either is null, that child does not exist   *)
(**************************************************)

(* pcm is finite sets of pointers *)
Definition U := [pcm of union_map [ordType of ptr] unit].

Structure raw_graph := RawGr {nodes : U; left : ptr -> ptr; right : ptr -> ptr}.

Inductive choice := l | r.

Notation ch g x := (left g x, right g x).
Notation rd c g x := (if c is l then left g x else right g x).

Definition goodP g :=
  valid (nodes g) /\ forall x : ptr,
    if x \in dom (nodes g) then
      x != null /\ forall c, rd c g x \in [predU pred1 null & dom (nodes g)]
    else forall c, rd c g x = null.

Structure graph := Graph {rg :> raw_graph; _ : goodP rg}.

Lemma graphN x (g : graph) : x \in dom (nodes g) -> x != null.
Proof. by case: g=>r [V H] D; move: (H x); rewrite D; case. Qed.

Lemma graphV (g : graph) : valid (nodes g). Proof. by case: g=>r []. Qed.

Lemma graphD x (g : graph) c :
        x \in dom (nodes g) -> rd c g x \in [predU pred1 null & dom (nodes g)].
Proof. by case: g=>r [V H] D; move: (H x); rewrite D; case. Qed.

(* definition of prunning the c edge of x in graph g *)
Definition pr' c x g :=
  RawGr (nodes g) [eta left g with x |-> if c is l then null else left g x]
                  [eta right g with x |-> if c is r then null else right g x].

Lemma pr_good c x (g : graph) : goodP (pr' c x g).
Proof.
case: g=>r [V G]; split=>// z; move: (G z); case: ifP=>[D [Nz H]|Nz H d] /=.
- by split=>//= d; case: ifP (H d)=>// /eqP <-; case: c; case: d.
by case: ifP (H d)=>// /eqP <-; case: c; case: d.
Qed.

Definition pr c x (g : graph) := Graph (@pr_good c x g).

(*******************************************)
(* Specification components and properties *)
(*******************************************)

Definition subgraph (m : U) (g1 g2 : graph) :=
  nodes g1 = nodes g2 /\
  forall x c, if x \notin dom m then rd c g1 x = rd c g2 x
              else rd c g1 x \in pred2 (rd c g2 x) null.

Lemma subgr_mark x m g1 g2 :
        valid (#x \+ m) -> subgraph m g1 g2 -> subgraph (#x \+ m) g1 g2.
Proof.
move=>V [N H]; split=>// y c; case: ifP (H y c)=>D E.
- by rewrite E inE eq_refl; case: ifP.
by rewrite domPtUn V inE /= (negbFE D) orbT.
Qed.

Lemma subgr_prune c x g (m : U) : x \in dom m -> subgraph m (pr c x g) g.
Proof.
move=>Dx; split=>//= y d; case: eqP=>[->|_].
- by rewrite Dx; case: c; case: d; rewrite !inE !eq_refl // !orbT.
by rewrite inE !eq_refl; case: ifP.
Qed.

Lemma subgr_refl m g : subgraph m g g.
Proof. by split=>// x c; rewrite inE !eq_refl; case: ifP. Qed.

Hint Resolve subgr_refl : core.

Lemma subgr_trans m g1 g2 g3 :
        subgraph m g1 g2 -> subgraph m g2 g3 -> subgraph m g1 g3.
Proof.
case=>N1 H1 [N2 H2]; split; first by rewrite N1 N2.
move=>x d; case: ifP (H1 x d) (H2 x d) => _; first by move=>->.
suff H a b c : a \in pred2 b null -> b \in pred2 c null -> a \in pred2 c null.
- by case: d; apply: H.
by case/orP=>/eqP -> // _; rewrite !inE eq_refl orbT.
Qed.

(* subspan *)

Definition edges g x := [predD pred2 (left g x) (right g x) & pred1 null].

Definition subspan (x : ptr) t (m : U) (g1 g2 : graph) :=
  [/\ x != null, tree (edges g1) x t,
      front (edges g1) t t & front (edges g2) t m].

Lemma subsp_mark x y t m g1 g2 :
        valid (#y \+ m) -> subspan x t m g1 g2 -> subspan x t (#y \+ m) g1 g2.
Proof.
move=>V [/= N T M F]; split=>//; case: F=>Dx F; split.
- by move=>z /Dx; rewrite domPtUn V inE /= orbC=>->.
by move=>a b A /(F _ _ A); rewrite domPtUn V inE /= orbC=>->.
Qed.

Lemma subsp_prune x y t c (m : U) g1 g2 :
        y \notin dom t -> subspan x t m g1 g2 -> subspan x t m (pr c y g1) g2.
Proof.
have path_mono p (t' : U) g :
  {in dom t', forall y, ch g1 y = ch g y} -> x \in dom t' ->
  {subset p <= dom t'} -> path (edges g1) x p = path (edges g) x p.
- move=>H; elim: p x=>[|z zs IH] //= x Dx S; rewrite IH.
  - by case: (H _ Dx)=>->->.
  - by apply: S; rewrite inE eq_refl.
  by move=>w Hw; apply: S; rewrite inE Hw orbT.
have tmono (t' : U) g : {in dom t', forall y, ch g1 y = ch g y} ->
                        tree (edges g1) x t' -> tree (edges g) x t'.
- move=>H [Dx Ht]; split=>// w Dw.
  case/(_ _ Dw): Ht=>p [[Ew Pw Sw] Up]; exists p.
  split; first by rewrite -(path_mono _ _ _ H Dx Sw).
  move=>p' [Ey' Py' Sy']; apply: Up; split=>//.
  by rewrite (path_mono _ _ _ H Dx Sy').
have fmono (t' : U) u g : {in dom t', forall y, ch g1 y = ch g y} ->
                    front (edges g1) t' u -> front (edges g) t' u.
- move=>H [S Ht]; split=>// q w Dq Dw; apply: Ht (Dq) _.
  by rewrite /edges in Dw *; case: (H _ Dq)=>->->.
move=>Dy [/= N T M F]; split=>//; [apply: tmono T | apply: fmono M];
by move=>z Dz; case: c Dz Dy=>/=; case: eqP=>// ->->.
Qed.

(* helper predicate from the post of span *)

Definition sp_post r x t mu gr (g : graph) :=
  if r then subspan x t mu gr g
  else t = Unit /\ x \in [predU pred1 null & dom mu].

Definition if_post c x m t (g1 g : graph) :=
  exists r, sp_post r (rd c g x) t m g1 g /\
            rd c g1 x = if r then rd c g x else null.

Lemma subspan_join x tl tr (m : U) (g1 g : graph) :
        x != null -> valid (#x \+ (tl \+ tr)) -> x \in dom m ->
        if_post l x m tl g1 g -> if_post r x m tr g1 g ->
        subspan x (#x \+ (tl \+ tr)) m g1 g.
Proof.
rewrite /if_post/sp_post=>Dn V Dx [r1][H1 E1][r2][H2 E2].
have Sb: {subset edges g1 x <= pred2 (left g1 x) (right g1 x)}.
- by move=>z; rewrite inE; case/andP=>_; apply.
have [L1 L2] : if_edge (edges g1) x (left g1 x) tl /\
               if_edge (edges g1) x (right g1 x) tr.
- split; rewrite /= !inE /=; [case: r1 E1 H1 | case: r2 E2 H2];
  by move=>-> [->] //; rewrite eq_refl ?orbT.
have [F1 F2] : front (edges g) tl m /\ front (edges g) tr m.
- split; [case: r1 E1 H1 | case: r2 E2 H2]; move=>_;
  by (case=>_ _ _ // || case=>->; split=>?; rewrite dom0 //).
have [D1 D2] : left g x \in [predU pred1 null & dom m] /\
               right g x \in [predU pred1 null & dom m].
- split; [case: r1 E1 H1 | case: r2 E2 H2]; try by [move=>_ []];
  try by move=><- [_ T _ M]; rewrite !inE (proj1 M _ (proj1 T)) orbT.
split=>//; first by apply: tree_join Sb (validR V) L1 L2.
- by apply: max_join Sb L1 L2.
apply: frontUn (frontUn F1 F2); apply: frontPt Dx.
by move=>z /andP /= [Nz] H; case/orP: H Nz=>/eqP ->{z} Nz;
   [move: Nz D1 | move: Nz D2]; rewrite !inE => /negbTE ->.
Qed.

(******************************)
(* Graphs layed out in a heap *)
(******************************)

Section Layout.
Implicit Types (g : graph) (m : U).

Definition f := (fun g m h x => x :-> (x \in dom m, ch g x) \+ h).

Definition layout g m := foldl (f g m) Unit (dom (nodes g)).

Lemma foldf g m h xs : foldl (f g m) h xs = h \+ foldl (f g m) Unit xs.
Proof.
elim: xs h=>[|x xs /= IH] h; first by rewrite unitR.
by rewrite (IH (f g m h x)) (IH (f g m Unit x)) joinCA joinA /f unitR.
Qed.

Lemma layoutE g m : valid (layout g m) * (dom (layout g m) = dom (nodes g)).
Proof.
rewrite /layout; set zs := dom (nodes g).
have [] : [/\ uniq zs, null \notin zs & sorted ord zs].
- by rewrite /zs (contraL (@graphN null g)).
elim: zs=>[//|z zs IH]; rewrite inE negb_or; case/andP=>H U /andP [Z N] S.
case: (IH U N (path_sorted S))=>{IH U N} V D; rewrite /= foldf {1 3}/f unitR.
split; first by rewrite validPtUn //= V D eq_sym Z.
apply: eq_sorted_irr S _=>//=.
by move=>s; rewrite domPtUn !inE validPtUn /= V D !(eq_sym z) Z H.
Qed.

End Layout.

(*********************************************)
(* Resource invariant, transitions and atoms *)
(*********************************************)

(* joint state is a heap and a graph *)
Structure T := ghost {gr : graph}.
Implicit Type s : state U T. (* speeds up typechecking *)

Definition mu : U -> _ := idfun.
Canonical mu_morph := Eval hnf in [morphism of mu].

Lemma idJ s : idfun.j s = ghost (gr.j s).
Proof. by rewrite -(lJA gr); case: (idfun.j s). Qed.

(* resource invariant *)
Program Definition span_coh :=
  CohPx2 (fun s => valid (mu.t s) /\ {subset dom (mu.t s) <= dom (nodes (gr.j s))})
         (fun s => layout (gr.j s) (mu.t s)) _.
Next Obligation. by split=>[?[*]|*]; rewrite ?allvE ?lO ?lTE' ?lE // layoutE. Qed.

(* trymark transition *)
Program Definition trymark_tr x b :=
  TransPx2 (fun s r => [/\ b = (x \notin dom (mu.t s)),
    x \in dom (nodes (gr.j s)), valid (mu.t s) &
    r = if b then State (#x \+ mu.s s) (ghost (gr.j s)) (idfun.o s) else s])
  span_coh _.
Next Obligation.
split=>[s [-> ?? ->][??? ->]|[-> ?? ->]|p ?? O|]; rewrite /= ?lE //=.
- by case: ifP; rewrite ?lE.
- by case: ifP; rewrite !lE // =>_ [<- ->->] [->->]; rewrite -?joinA -O.
case=>-> D _ -> [V S]; case: ifP=>H //; rewrite !lE /= -joinA.
have V' : valid (#x \+ mu.t s1) by rewrite validPtUn V H.
by split=>// z; rewrite domPtUn inE V'; case: eqP=>[<-|_] //; apply: S.
Qed.

(* prune transition *)
Program Definition prune_tr x c :=
  TransPx2 (fun s r => [/\ x \in dom (mu.s s), valid (mu.t s) &
    r = State (mu.s s) (ghost (pr c x (gr.j s))) (idfun.o s)]) span_coh _.
Next Obligation.
split=>[s [_ _ ->][_ _ ->]|[_ _ ->]|p ?? O []|[_ _ ->]] //=;
by rewrite !lE // => D V [->->]; rewrite -O domUn [x \in _]inE D !(validX V).
Qed.

(* read transition *)
Lemma readM x c r : monoP span_coh (fun s => x \in dom (nodes (gr.j s)) /\
                                             r = rd c (gr.j s) x).
Proof. by move=>??; rewrite !lE. Qed.

(* the assembled resource *)
Inductive spanTr := trymark_t of ptr & bool | prune_t of ptr & choice |
                    read_t of ptr & choice & ptr.

Definition span_interp tr :=
  match tr with
    trymark_t x b => trymark_tr x b
  | prune_t x c => prune_tr x c
  | read_t x c y => pre_tr (id_tr span_coh) (@readM x c y)
  end.

Program Canonical sp := Eval hnf in ResPx span_coh span_interp (fun => true) _.
Next Obligation.
case=>[p b|p c|p c q] _ s x _ H; apply/domeqP; rewrite !layoutE;
by case: H=>/= *; subst x=>//; do ?[case: ifP]; rewrite ?lE.
Qed.

(*************)
(* Stability *)
(*************)

Lemma osubgr g s1 s2 :
        osteps sp s1 s2 ->
        subgraph (mu.t s1) (gr.j s1) g -> subgraph (mu.t s2) (gr.j s2) g.
Proof.
move: s1 s2; apply: iter_ind=>s1 s2 []; case=>//=; rewrite !lE.
- move=>x [][/= H _ V]; rewrite lE=>-> Sg //; rewrite !lE /=.
  by rewrite joinCA; apply: subgr_mark Sg; rewrite validPtUn joinC V -H.
- move=>c x [Dx] V /trS ->; rewrite !lE.
  by apply/subgr_trans/subgr_prune; rewrite joinC domUn inE Dx V.
by move=>c x _ [_] /tr_inj ->.
Qed.

Lemma osubsp x t g s1 s2 :
        osteps sp s1 s2 -> {subset dom t <= dom (mu.s s1)} ->
        subspan x t (mu.t s1) (gr.j s1) g -> subspan x t (mu.t s2) (gr.j s2) g.
Proof.
move: s1 s2; apply: iterf_ind=>// s1 s2 []; case=>//=; rewrite !lE.
- move=>z [][/= H _ V]; rewrite lE=>-> Sd; split=>//; rewrite !lE //.
  by rewrite joinCA; apply: subsp_mark; rewrite validPtUn joinC V -H.
- move=>y c /= [Dy] V /trS -> Sd; split=>//; rewrite !lE //; apply: subsp_prune.
  by apply: contra (Sd _) _; case: validUn V=>// _ _ /(_ _ Dy).
by move=>y c p [_] /tr_inj ->.
Qed.

Lemma och x s1 s2 :
        osteps sp s1 s2 -> x \in dom (mu.s s1) -> ch (gr.j s1) x = ch (gr.j s2) x.
Proof.
move: s1 s2; apply: iterf_ind=>// s1 s2 []; case=>//=; rewrite !lE.
- by move=>z [][_ _ _]; rewrite lE =>-> //; rewrite !lE.
- move=>y c [Dy] V /trS -> Sd; split; rewrite !lE //=; case: eqP Dy Sd=>// <-.
  by case: validUn V=>// _ _ L _ /L /negbTE ->.
by move=>y c p [_] /tr_inj ->.
Qed.

(* a common use form; we will package program specs to match this lemma *)
Lemma och' x xx s1 s2 :
        osteps sp s1 s2 -> mu.s s1 = #x /\ ch (gr.j s1) x = xx ->
                           mu.s s2 = #x /\ ch (gr.j s2) x = xx.
Proof. by move=>O [E ?]; rewrite -(osf O) -(och O) // E domPt inE /=. Qed.

Lemma omu x s1 s2 : osteps sp s1 s2 -> x \in dom (mu.t s1) -> x \in dom (mu.t s2).
Proof.
move: s1 s2; apply: iter_ind=>// s1 s2 []; case=>//=; rewrite !lE; last first.
- by move=>y c p [_] /tr_inj ->.
- by move=>y c [Dz] V /trS ->; rewrite !lE.
move=>z [][/= H Dz V]; rewrite lE=>-> Dx //.
by rewrite !lE joinCA domPtUn validPtUn inE Dx joinC V -H orbT.
Qed.

Lemma osp_post r x t (g : graph) s1 s2 :
        osteps sp s1 s2 -> {subset dom t <= dom (mu.s s1)} ->
        sp_post r x t (mu.t s1) (gr.j s1) g -> sp_post r x t (mu.t s2) (gr.j s2) g.
Proof.
move=>O Dt; case: r; first by apply: osubsp O Dt.
by case=>Et H; split=>//; move: H; rewrite !inE; case: eqP=>//= _; apply: omu.
Qed.

(***********)
(* Actions *)
(***********)

Program Definition trymark x : {g}, STsep [sp]
  (fun s1 => [/\ subgraph (mu.t s1) (gr.j s1) g, mu.s s1 = Unit &
                 x \in dom (nodes g)],
   fun r s2 => subgraph (mu.t s2) (gr.j s2) g /\
               if r then mu.s s2 = #x /\ ch (gr.j s2) x = ch g x
               else mu.s s2 = Unit /\ x \in dom (mu.t s2)) :=
  Do (actSx (trymark_t x)).
Next Obligation. by split=>// ????? _ [-> _ _ _][->]. Qed.
Next Obligation.
move=>x i [g][][Sg Em Dg] C; apply: vrf_atm=>// j O {}C.
move/(osubgr O): Sg=>Sg; rewrite {O i}(osf O) in Em.
eexists (x \notin dom (mu.t j)), _; split; rewrite ?effE.
- by split=>//; rewrite ?(proj1 Sg, proj1 C).
rewrite if_neg=>m; case: ifPn=>Dx O /=.
- by split; [apply: osubgr O Sg | rewrite (omu O Dx) -(osf O)].
move: (proj2 Sg x); rewrite Dx /= => H; split; last first.
- by apply: (och' O); rewrite !lE /= Em unitR (H l) (H r).
apply: osubgr O _; rewrite !lE -joinA; apply: subgr_mark Sg.
by rewrite validPtUn (proj1 C).
Qed.

Program Definition prune x c : {g xl xr}, STsep [sp]
  (fun s1 => [/\ subgraph (mu.t s1) (gr.j s1) g &
                 mu.s s1 = #x /\ ch (gr.j s1) x = (xl, xr)],
   fun _ s2 => [/\ subgraph (mu.t s2) (gr.j s2) g & mu.s s2 = #x /\
                   ch (gr.j s2) x = (if c is l then null else xl,
                                     if c is r then null else xr)]) :=
  Do (actSx (fun _ : unit => prune_t x c)).
Next Obligation. by split=>// ??? [][]. Qed.
Next Obligation.
move=>x c i [g][xl][xr][/= R C]; apply: vrf_atm=>// j O {}C.
case: R=>/(osubgr O) /= Sg /(och' O) [Em Ech]; eexists tt, _.
rewrite effE /= {1}Em domPt inE /= (proj1 C).
split=>//= m {} O []; rewrite !lE /==>Vx Sx.
split; last by apply: (och' O); rewrite !lE /= eq_refl; case: Ech {O} c=>->-> [].
apply: osubgr O _; rewrite !lE /=; apply: subgr_trans (subgr_prune _ _ _) Sg.
by rewrite domUn inE Vx Em domPt inE eq_refl.
Qed.

Program Definition read x c : {g xl xr}, STsep [sp]
  (fun s1 => [/\ subgraph (mu.t s1) (gr.j s1) g &
                 mu.s s1 = #x /\ ch (gr.j s1) x = (xl, xr)],
   fun r s2 => [/\ r = if c is l then xl else xr, subgraph (mu.t s2) (gr.j s2) g &
                   mu.s s2 = #x /\ ch (gr.j s2) x = (xl, xr)]) :=
  Do (actSx (read_t x c)).
Next Obligation. by split=>// ????? _ [[_ -> _]][[_ ->]]. Qed.
Next Obligation.
move=>x c i [/= g][xl][xr][/= R C]; apply: vrf_atm=>// j O {}C; case: R.
move=>/(osubgr O) Sg /(och' O) [Em Ech]; eexists _, _.
rewrite effE; split=>[|m {}O].
- do !split=>//; apply: (proj2 C).
  by rewrite domUn inE (proj1 C) Em domPt inE eq_refl.
by split; [case: c Ech; case | apply: osubgr Sg | apply: och' O _].
Qed.

(****************)
(* Main program *)
(****************)

(* helper procedure *)
Program Definition if_prune b x c : {g tl tr rl rr xl xr}, STsep [sp]
  (fun s1 => [/\ subgraph (mu.t s1) (gr.j s1) g, mu.s s1 = #x \+ (tl \+ tr) /\
                 ch (gr.j s1) x = (xl, xr), if c is l then rl = b else rr = b &
                 sp_post rl (left g x) tl (mu.t s1) (gr.j s1) g /\
                 sp_post rr (right g x) tr (mu.t s1) (gr.j s1) g],
  fun _ s2 => [/\ subgraph (mu.t s2) (gr.j s2) g, mu.s s2 = #x \+ (tl \+ tr) /\
                  ch (gr.j s2) x = (if c is l then if rl then xl else null else xl,
                                   if c is r then if rr then xr else null else xr) &
                  sp_post rl (left g x) tl (mu.t s2) (gr.j s2) g /\
                  sp_post rr (right g x) tr (mu.t s2) (gr.j s2) g]) :=
  Do (if b then ret tt else prune x c).
Next Obligation.
move=>b x c i [g][tl][tr][rl][rr][xl][xr][] /= [Sg [Eis Ech] Er [H1 H2]] C.
have Dx : x \in dom (mu.s i).
- by rewrite Eis domPtUn inE eq_refl -Eis (validE2 (proj1 C)).
have V : valid (#x \+ (tl \+ tr)) by rewrite -Eis (validE2 (proj1 C)).
have [Sl Sr] : {subset dom tl <= dom (tl \+ tr)} /\
               {subset dom tr <= dom (tl \+ tr)}.
- by split=>z D; rewrite domUn inE D /= ?orbT andbT (validRE3 V).
case: b Er=>Er.
- apply: vrf_ret=>// k O; split; first by apply: osubgr Sg.
  by rewrite -(osf O); case: c Er=>->; rewrite -(och O).
- split; apply: osp_post O _ _=>// z D;
  by rewrite Eis domPtUn inE V ?(Sl _ D) ?(Sr _ D) orbT.
apply: (frame4 sp (#x) (tl \+ tr))=>//= {Eis C} y j C X Ejs Ejo.
rewrite {i}X !lE /= -joinA -Ejs -Ejo in Sg Ech H1 H2 Dx.
apply: [gE g, xl, xr]=>//= [{Sg Ech C}][k][Sg' [Eks Ech']] _ z i _ X Eis O.
rewrite {k}X !lE /= joinA -Eis in Sg' Eks Ech'; rewrite -Eks in O *.
split=>//; first by case: c Ech' Er {H1 H2}; [case: rl | case: rr].
set x1 := State _ _ _ in O; set x2 := State _ _ _ in O.
rewrite (_ : gr.j i = gr.j x2); last by rewrite lE.
rewrite (_ : mu.t i = mu.t x2); last by rewrite !lE joinCA joinA -Eis.
by split; apply: osp_post O _ _; rewrite !lE //= joinCA Eks -Ejs -Ejo.
Qed.

(* main procedure *)
Notation post x g :=
  (fun r s => subgraph (mu.t s) (gr.j s) g /\
              if r then subspan x (mu.s s) (mu.t s) (gr.j s) g
              else mu.s s = Unit /\ x \in [predU pred1 null & dom (mu.t s)]).

Program Definition span : forall x, {g}, STsep [sp]
  (fun s1 => [/\ subgraph (mu.t s1) (gr.j s1) g, mu.s s1 = Unit &
                 x \in [predU pred1 null & dom (nodes g)]], post x g) :=
  ffix (fun loop x =>
    Do (if x == null then ret false else (
        b <-- trymark x;
        if b then
          xl <-- read x l;
          xr <-- read x r;
          rs <-- par (loop xl) (loop xr);
          if_prune rs.1 x l;;
          if_prune rs.2 x r;;
          ret true
        else ret false))).
Next Obligation.
move=>loop x i [g][[/= Sg Em Dx]] C;  case: ifP=>[|/negbT] N.
- by apply: vrf_ret=>// m O; move/(osubgr O): Sg; rewrite -(osf O) !inE N.
rewrite inE (negbTE N) in Dx; apply: [stepE g]=>//=.
move/graphD: Dx=>X; have {X i C Sg Em} [Ln Rn] := (X l, X r).
case=>m R C; last first.
- apply: vrf_ret=>// k O; case: R=>/(osubgr O) Sg [Em Dx].
  by rewrite -{2}(osf O) inE (negbTE N) /= (omu O Dx).
apply: [stepE g, left g x, right g x]=>//= {m R C} _ k /= [-> Sg Ekch] C.
apply: [stepE g, left g x, right g x]=>//= {k C Sg Ekch} _ m /= [-> Sg [Em Ech]] C.
apply: step; apply: (frame4 sp Unit (#x))=>//=; first by rewrite unitL.
move=> y j {}C X Ejs Ejo; rewrite {m Em}X !lE /= unitL -Ejo in Sg Ech.
apply: (par_do3 sp Unit Unit (post (left g x) g) (post (right g x) g))=>/=;
try by [move=>{}C; apply: [gE g]=>//=; rewrite !lE !unitL];
try by [rewrite ?lE unitL -Ejs].
case=>{Sg C} rl rr tl tr k _ Eks [/= Sg H1][_ H2] z m C X Ems /och O.
rewrite !lE (joinCA tr) joinA {k}X !lE joinCA joinA /= in Eks Sg H1 H2 O.
rewrite joinC {z}Eks in Ems O *; rewrite -Ems in Sg H1 H2.
rewrite O /= ?inE ?domPt ?inE //= {j Ejs Ejo O} in Ech.
have V : valid (#x \+ (tl \+ tr)) by rewrite -Ems (validE2 (proj1 C)).
apply: [stepE g, tl, tr, rl, rr, left g x, right g x]=>//=.
case=>{Sg Ems Ech H1 H2 C m} k [Sg [Ek Ech]] H C.
apply: [stepE g, tl, tr, rl, rr, if rl then left g x else null, right g x]=>//=.
case=>{Sg Ek Ech H C k} m [Sg [Em Ech][H1 H2]] C.
apply: vrf_ret=>// k O; split; first by apply: osubgr Sg.
have {Sg} Dx : x \in dom (mu.t m).
- by rewrite Em domUn inE -{1}Em (proj1 C) domPtUn inE V eq_refl.
rewrite -{1}(osf O); apply: osubsp O _ _=>//; rewrite {1}Em {Em}.
by apply: subspan_join=>//; [exists rl | exists rr]; case: Ech.
Qed.

(*******************)
(* Hiding and span *)
(*******************)

(* Lemmas for reasoning about layouts, i.e., heap erasures *)
Lemma layF x m (g : graph) :
        find x (layout g m) =
        if x \in dom (nodes g) then Some (idyn (x \in dom m, ch g x)) else None.
Proof.
move: (fst (layoutE g m)); rewrite /layout; elim: (dom (nodes g))=>//= z zs IH.
rewrite foldf {1 3}/f unitR=>V; rewrite findPtUn2 // inE.
by case: eqP=>[<-| _]//; apply: IH (validR V).
Qed.

Lemma layE m1 m2 (g1 g2 : graph) :
        layout g1 m1 = layout g2 m2 ->
        g1 = g2 /\ {in dom (nodes g1), dom m1 =i dom m2}.
Proof.
move=>E; have E1 : dom_eq (nodes g1) (nodes g2).
- apply/domeqP; rewrite -(layoutE g1 m1).2 E layoutE.
  by split=>//; case: g1 g2 {E}=>r1 [-> _][r2 [-> _]].
have E2 x : x \in dom (nodes g1) ->
  (x \in dom m1) = (x \in dom m2) /\ ch g1 x = ch g2 x.
- move=>D; move: (layF x m1 g1); rewrite D E layF.
  by case/domeqP: E1=>_ <-; rewrite D; case/Some_inj/dyn_inj=>->->->.
case: g1 g2 E1 E2 E; case=>n1 l1 r1 G1 [[n2 l2 r2] G2] E1 E2 E /=.
move/domeq_eta: E1=>/= ?; subst n2; split=>[|x /E2 [] //].
suff L x : (l1 x = l2 x) * (r1 x = r2 x).
- have [??]: l1 = l2 /\ r1 = r2 by split; apply: fext=>x; rewrite L.
  by subst l2 r2; congr Graph; apply: pf_irr.
case Dx : (x \in dom n1); first by case: (E2 x Dx)=>_ [->->].
case: G1 {E E2}=>_ /(_ x); case: G2=>_ /= /(_ x).
by rewrite Dx=>X1 X2; rewrite (X1 l) (X1 r) (X2 l) (X2 r).
Qed.

Lemma lay_readE x m (g : graph) :
        x \in dom (nodes g) ->
        layout g m = x :-> (x \in dom m, ch g x) \+ free (layout g m) x.
Proof.
move=>D.
have V : valid (x :-> (x \in dom m, ch g x) \+ free (layout g m) x).
- by rewrite validPtUn /= (graphN D) validF layoutE domF inE eq_refl.
apply/find_eta=>//; first by rewrite layoutE.
by move=>k; rewrite findPtUn2 // findF !layF; case: eqP=>// ->; rewrite D.
Qed.

Lemma lay_markE x m (g : graph) :
        valid (#x \+ m) -> free (layout g (#x \+ m)) x = free (layout g m) x.
Proof.
move=>V; apply/find_eta; try by rewrite validF layoutE.
by move=>k; rewrite !findF !layF domPtUn inE V eq_sym; case: eqP.
Qed.

Lemma lay_pruneE x b m (g : graph) :
        x \in dom (nodes g) ->
        free (layout (pr b x g) m) x = free (layout g m) x.
Proof.
move=>V; apply/find_eta; try by rewrite validF layoutE.
by move=>k; rewrite !findF !layF /=; case: eqP.
Qed.

Arguments lay_readE [x m g] _.
Arguments lay_markE [x m g] _.
Arguments lay_pruneE [x b m g] _.

Section Quiescence.
Variable x : ptr.
Implicit Type v : bool * (ptr * ptr).

Definition spriv_phi (d : U) (s1 : st priv) (s2 : st sp) :=
  [/\ coh sp s2, d = nodes (gr.j s2), mu.o s2 = Unit &
      [pcm layout (gr.j s2) (mu.t s2) <= idfun.s s1]].

Definition spriv_tau' (d : U) (tr : Tr sp) : Tr priv :=
  match tr with
    trymark_t x b => Priv.rmw_t x (fun v => v.1 = ~~ b)
      (fun v => if v.1 then v else (true, v.2))
  | prune_t x c => Priv.rmw_t x (fun => True)
      (fun v => (v.1, if c is l then (null, v.2.2) else (v.2.1, null)))
  | read_t x c y => Priv.rmw_t x
      (fun v => y = if c is l then v.2.1 else v.2.2) id
  end.

Definition spriv_tau (d : U) (tr : Tr sp) (_ : st priv) (t' : Tr priv) :=
  t' = spriv_tau' d tr.

Lemma spriv_local : morph_localP spriv_phi.
Proof.
exists (fun => Unit)=>d p s s'' Cp Cs [_ -> E H]; rewrite lE in H.
exists s''; split; first by rewrite zgE0.
rewrite zgE0; split=>//=; apply: pleq_trans H _.
by rewrite lE // (coh_valid Cp).
Qed.

Lemma spriv_other : morph_otherP spriv_phi.
Proof. by move=>?????????? [?? Ev _][?? Ew _]; rewrite !lT Ev Ew. Qed.

Lemma spriv_func : morph_funcP spriv_phi.
Proof.
move=>d s s1 s2 /cohso V [V1 S1][V2 S2][_ -> E1 [h1 H1]][_ E E2 [h2 H]].
apply/stE; rewrite H1 in H; move/dom_prec: H; rewrite !layoutE E -H1 !idJ.
case/(_ (validL V) _)/layE=>// ->; rewrite E E1 E2 !unitR in V1 S1 V2 S2 *=>D.
split=>//; apply/domeq_eta/domeqP; rewrite V1 V2; split=>// k.
case Dk : (k \in dom (nodes (gr.j s2))); first by apply: D.
by rewrite (contraFF (S1 k)) // (contraFF (S2 k)).
Qed.

Lemma spriv_backw : morph_backwardP spriv_phi.
Proof. by move=>????? _ _ _ [_ -> _ _][_ ->]. Qed.

Lemma spriv_int : morph_internalP spriv_tau.
Proof. by move=>t _ p s I C ->; case: t I. Qed.

Lemma spriv_tfunc : morph_tfuncP spriv_tau.
Proof. by move=>p t sp t1 t2 C ->->. Qed.

Lemma spriv_tlocal : morph_tlocalP spriv_tau.
Proof. by []. Qed.

Lemma spriv_coh : morph_cohP' spriv_phi.
Proof. by move=>??? _ []. Qed.

Definition sprivMorphMixin :=
  MorphMixin spriv_local spriv_other spriv_func spriv_backw
             spriv_int spriv_tfunc spriv_tlocal spriv_coh.
Canonical spriv := Morph sprivMorphMixin.

Lemma spriv_ostep : morph_ostepP spriv (fun _ _ => True).
Proof.
move=>d s1 s2 s1' Cp Cs [_ E Em H] _ /iter1 O; exists s1'.
by do !split=>//; rewrite -(osf O).
Qed.

Lemma spriv_forw : morph_forwardP spriv (fun _ _ => True).
Proof.
move=>t d s1 s1' s2' _ C1 C1' [_ Ed Em [h H]] _ J; exists (spriv_tau' d t), d.
exists (State (flat (coh sp) s2' \+ h) (idfun.j s1) (idfun.o s1)); rewrite /flat /=.
move: (icoh J C1'); case: t J=>[p b|p c|p c y] J; last 1 first.
- case: J; case=>D ->{y} ->{s2'} _; do !split=>//=; first by rewrite !lE /=.
  by rewrite /= (cohso C1) H (lay_readE D) -joinA -!lT; eexists _, _.
- case: J=>-> D V -> /= C2; do !split=>//;
    try by [case: ifP; rewrite ?lE /=].
  rewrite (cohso C1) H (lay_readE D) negbK -joinA.
  eexists _, _; split=>//; case: ifP=>/= Dm; last first.
  - by rewrite {1}(lay_readE D) (negbFE Dm) joinA.
  have V' : valid (#p \+ mu.t s1') by rewrite validPtUn V.
  rewrite !lE /= (negbTE Dm) -joinA (lay_readE D) /=.
  by rewrite domPtUn inE V' eq_refl /= lay_markE // joinA.
case: J=>Dm V -> /= C2'; do !split=>//=; rewrite ?lE //=.
have D : p \in dom (nodes (gr.j s1')) by apply: (proj2 C1'); rewrite Em unitR.
have D' : p \in dom (nodes (pr c p (gr.j s1'))) by [].
rewrite (cohso C1) H (lay_readE D) -joinA.
rewrite (lay_readE D') Em unitR Dm /= eq_refl lay_pruneE //.
by eexists _, _; split=>//; rewrite joinA; case: c {C2' D'}.
Qed.

Definition spriv_sim := Simulation2 spriv_ostep spriv_forw.

Program Definition span_top (d : U) : {g : graph}, STsep [priv]
  (fun s1 : st priv => exists s1', [/\ d = nodes g, [morph spriv] d s1 s1',
     subgraph (mu.t s1') (gr.j s1') g, mu.s s1' = Unit,
     x \in dom d & connected (edges g) x d],
   fun _ (s2 : st priv) => exists s2', [/\ [morph spriv] d s2 s2',
     subgraph (mu.t s2') (gr.j s2') g & tree (edges (gr.j s2')) x d]) :=
  Do (morph_app spriv d (span x)).
Next Obligation.
move=>d i [g][[i']][-> M S] E Dx Cx _; apply: (vrf_morph spriv_sim) M I _=>C.
apply: [gE g]=>//; first by rewrite /= inE Dx orbT.
move=>{E S C} b m' [S] H /cohso/validL V d' m M {i} _ C _; exists m'.
have [? E] : d' = nodes g /\ mu.o m' = Unit by case: M S=>_ ->-> _ [].
(* b can't be false *)
case: b H; last by case=>->; rewrite inE (negbTE (graphN Dx)) E unitL dom0.
subst d'=>[[_ T _ F]]; split=>//; suff : nodes g = mu.s m' by move=>->.
rewrite E unitR in F; have {T} Dx : x \in dom (mu.s m') by case: T.
apply: domeq_eta; apply/domeqP; rewrite V graphV; split=>//.
move=>y; apply/idP/idP; last first.
- by case: M; case; rewrite E unitR=>_ X -> _ _; apply: X.
by case/Cx=>p [P <-]; apply: max_path F P Dx _ _; apply: mem_last.
Qed.

End Quiescence.
