From Coq Require Import ssreflect ssrbool ssrfun.
From mathcomp Require Import ssrnat eqtype seq.
From fcsl Require Import axioms pred prelude ordtype pcm unionmap automap heap.
From fcsl Require Import natmap morphism notation lift mutex.
From fcsl Require Import state world schedule process always model.
From fcsl Require Import options.

(********************************************************)
(* CSLX resource for heap transfer of exclusive locking *)
(********************************************************)

Module CSLX.
Section CSLX.
Variables (R : heap -> Prop) (prec_pf : precise R). (* resource invariant *)

Definition U := [pcm of mtx3 * heap].
Definition T : Type := heap.

(* lock taken/free *)
Definition mu : U -> _ := fst.
Canonical mu_morph := Eval hnf in [morphism of mu].

(* the private heap *)
Definition hp : U -> _ := snd.
Canonical hp_morph := Eval hnf in [morphism of hp].

(* the shared heap *)
Notation hpj := (@id T).

Lemma idE : (idfun = [** mu & hp]) * (idfun = hpj).
Proof. by split; apply: fext=>// x; rewrite [x]prod_eta. Qed.

(* state space: the heap \hp.j is shared if mu.t = nown *)
(* otherwise \hp.j is empty, as the heap is given away *)
Program Definition xcoh :=
  CohPx2 (fun s : state U T =>
    [/\ valid (mu.t s), valid (hpj.j s \+ hp.t s) &
        if (mu.t s == own) || (mu.t s == auth1) then hpj.j s = Unit else R (hpj.j s)])
    (fun s => hpj.j s \+ hp.t s) _.
Next Obligation.
split=>[s [M V ?]|*]; rewrite ?lO // ?(allvE, idE) !lTE' ?lE //.
by rewrite pcmE /= M !(validX V).
Qed.

(* open transition to acquire the heap *)
Program Definition open_tr :=
  TransPx2 (fun s x => mu.t s = nown /\
    x = State (own, hpj.j s \+ hp.s s) Unit (idfun.o s)) xcoh _.
Next Obligation.
split=>[s [_ ->][_ ->]|[_ ->]|[p q] ?? O|]; rewrite /= ?lE //=.
- by case=>/mxE0 [->->][->->]; rewrite -O !joinA !pcmPE unitR.
by case; case/mxE0=>->E -> [_ V]; rewrite !lE E (validX V).
Qed.

(* close transition to release the heap if owning it *)
(* and the heap satisfies the resource invariant *)
Program Definition close_tr :=
  TransPx2 (fun s x => [/\ mu.s s = auth1,
    hp.s s = hpj.j x \+ hp.s x & R (hpj.j x)] /\
    x = State (nown, hp.s x) (@hpj.j x) (idfun.o s)) xcoh _.
Next Obligation.
split=>[s|[_ ->]|[p q] ?? O|]; rewrite /= ?lE //=; case; last 1 first.
- by case=>->-> X -> []; rewrite idE !lE=>/mxMx -> /validX ->.
- case=>_ -> I1 -> [[_ E I2 ->]][_ /validR V _].
  by rewrite !lE /= in E V; rewrite !(cancPL _ (validL V) I1 I2 E).
by case=> ->-> X [-> _ _][/mxMx/mxE0 [->->]]; rewrite O !joinA pcmPE !unitR.
Qed.

(* i should have some transitions that modify the heap *)
Program Definition write_tr A (x : ptr) (v : A) :=
  TransPx2 (fun s s' => exists B (w : B) h, hp.s s = x :-> w \+ h /\
    s' = State (mu.s s, x :-> v \+ h) (@hpj.j s) (idfun.o s)) xcoh _.
Next Obligation.
split=>[s||[p q] ?? O|]; rewrite /= ?lE //=.
- case=>B1 [w1][h1][->->][B2][w2][h2][E ->][_ /validR/validL V _].
  by move: (hcancelT V E)=>?; subst B2; case/(hcancelV V): E=>_ _ ->.
- by case=>B [w][h][_ ->]; rewrite !lE.
- case=>B [w][h][]; rewrite !lE //= => E1 [E2 E3 _ _] _.
  by exists B, w, (h \+ q); rewrite !lE E1 E2 E3 -O !joinA.
by case=>B [w][h][->->][]; rewrite !lE=>-> /validX ->.
Qed.

Program Definition give_tr :=
  TransPx2 (fun s s' => mu.s s = own /\
    s' = State (auth1, hp.s s) (hpj.j s) (idfun.o s)) xcoh _.
Next Obligation.
split=>[s [_ ->][_ ->]|[_ ->]|[p q] ?? O|]; rewrite /= ?lE //=.
- by case=>-> [->-> _ _][/mxMx/mxE0 [->->]]; rewrite -O.
by case=>->->; rewrite !lE; case=>/mxMx ->.
Qed.

Inductive xTr := open_t | close_t | write_t A of ptr & A |
                 give_t | id_t.

Definition xinterp t :=
  match t with
    open_t => open_tr
  | close_t => close_tr
  | write_t A x v => write_tr x v
  | give_t => give_tr
  | id_t => id_tr xcoh
  end.

Definition xintern (t : xTr) := true.

Program Canonical cslx := ResPx xcoh xinterp xintern _.
Next Obligation.
case=>[||A p v||] // _ s x C; last by move=>->.
- by case=>_ ->; rewrite /flat /= !(idE,lE) domeqX.
- case: C=>V _ X [[Y E _ ->]]; rewrite /flat /= !(idE,lE) E.
  by case: mxPJ V Y X=>//= t _ _ _ ->->; rewrite domeqX.
- by case=>B [w][h][E ->]; rewrite /flat /= !lE E domeqX.
by case=>_ ->; rewrite /flat /= !lE.
Qed.

End CSLX.

Module Exports.
Canonical mu_morph.
Canonical hp_morph.
Arguments CSLX.write_t [A].
End Exports.
End CSLX.

Export CSLX.Exports.

(********************************************)
(********************************************)
(* Spinlocks with nat permissions to unlock *)
(********************************************)
(********************************************)

(* and an external transition to grant permission *)
(* but lock transition doesn't grant permission to unlock *)

Module Spin.
Section Spin.
Variable l : ptr.
(* we take lifted nats, to be able to build subpcms *)
Definition U := [pcm of lnat * natmap (bool * bool)].
Definition T : Type := unit.
Implicit Type s : state U T.

(* permission field is a nat *)
Definition pi_raw (x : U) := x.1.
Canonical pi_raw_morph := Eval hnf in [morphism of pi_raw].
Definition pi : U -> nat := nat_unlift \o pi_raw.
Canonical pi_morph := Eval hnf in [morphism of pi].

(* history field *)
Definition ht (x : U) := x.2.
Canonical ht_morph := Eval hnf in [morphism of ht].

(* eta *)
Lemma idE : idfun = [** pi_raw & ht].
Proof. by apply: fext=>x; apply: prod_eta. Qed.

Definition lockP (v : bool * bool) := v.2 = ~~ v.1.

(* the lock is taken if the total history ends with a lock event *)
Notation "\lkj s" := (last_val false (ht.t s)) (at level 24).

(* state space; we don't require that \pi.t > 0 implies \lkj s *)
(* it is possible to have a permission, but the lock is free *)
(* unlocking programs will deal with failure to unlock separately *)
Program Definition scoh :=
  CohPx2 (fun s => [/\ valid (pi_raw.t s), valid (ht.t s), l != null,
           um_all (fun => lockP) (ht.t s) & complete false (ht.t s) (\lkj s)])
         (fun s => l :-> (\lkj s)) _.
Next Obligation.
split=>[s []|*]; last by rewrite /= !lE.
by rewrite validPt allvE idE !lTE' /= =>->->->.
Qed.

(* lock transition doesn't grant permission to unlock. *)
(* the permission is not exclusive and doesn't prevent *)
(* other threads from unlocking, assuming they also got a permission *)
Program Definition lock_tr :=
  TransPx2 (fun s x => ~~ \lkj s /\
    x = State (up (pi.s s), fresh (ht.t s) \-> (false, true) \+ ht.s s)
              (idfun.j s) (idfun.o s)) scoh _.
Next Obligation.
move=>x y; split=>[s [_ ->][_ ->]|[_ ->] /=|[p q] ?? O|]; rewrite /= ?lE //=.
- case => /negbTE -> [->-> _ _][]; rewrite -O pcmPJ -!joinA.
  by case: liftPJ=>// ?? ->.
case=>/negbTE ->-> [W V -> H C]; rewrite !lE -!joinA valid_fresh lastval_fresh //=.
case: liftPJ W=>// ?? _ -> _; split=>//; first by apply: umallPtUn.
by rewrite cm_fresh.
Qed.

(* unlock transition *)

(* we can only unlock if we have permission, *and* find the lock taken *)
(* in the process, we consume the permission *)
Program Definition unlock_tr :=
  TransPx2 (fun s x => [/\ pi.s s > 0 & \lkj s] /\
    x = State (up (pi.s s).-1, fresh (ht.t s) \-> (true, false) \+ ht.s s)
              (idfun.j s) (idfun.o s)) scoh _.
Next Obligation.
move=>x y; split=>[s [_ ->][_ ->]|[_ ->]|[p q] ?? O|]; rewrite /= ?lE //=.
- case; case=>H -> [->-> _ _][]; rewrite -O pcmPJ -joinA.
  rewrite (leq_trans H (leq_addr _ _)).
  by case: liftPJ=>// ?? ->; case: (pi.s x) H.
case; case=>H1 -> -> [W V -> H2 H3]; rewrite !lE -!joinA valid_fresh lastval_fresh //=.
case: liftPJ W=>// ?? _ -> _; split=>//; first by apply: umallPtUn.
by rewrite cm_fresh.
Qed.

Program Definition take_tr :=
  TransPx2 (fun s x =>
    x = State (up (pi.s s).+1, ht.s s) (idfun.j s) (idfun.o s)) scoh _.
Next Obligation.
move=>x y; split=>[s ->->|->|[p q] ?? O|]; rewrite /= ?lE //=.
- by case=>->-> _ _ []; rewrite -O; case: liftPJ=>// ?? ->.
by move=>-> []; rewrite !lE; case: liftPJ=>// ?? _ ->.
Qed.

Inductive Tr := lock_t | unlock_t | take_t | id_t.

Definition sinterp t :=
  match t with
    lock_t => lock_tr
  | unlock_t => unlock_tr
  | take_t => take_tr
  | id_t => id_tr scoh
  end.

Definition sintern (t : Tr) := if t is take_t then false else true.

Program Canonical spin := ResPx scoh sinterp sintern _.
Next Obligation. by case=>[|||] // _ ?? _ _; apply: domeqPt. Qed.

(* stability *)
Lemma sth x s1 s2 :
        osteps spin s1 s2 -> coh spin s1 ->
        [pcm x <= ht.t s1] -> [pcm x <= ht.t s2].
Proof.
move: s1 s2; apply: (pres_iterf_ind (@ocoh1 spin))=>// s1 s2.
case; case=>[|||]; last by [move/tr_inj=>->];
last by [move/trS ->; rewrite !lE];
by case=>H /trS -> _; rewrite !lE joinCA=>/pleq_trans X; split; apply: X.
Qed.

Lemma stk x s1 s2 : osteps spin s1 s2 -> coh spin s1 ->
        x <= last_key (ht.t s1) -> x <= last_key (ht.t s2).
Proof.
move=>O C H; apply: leq_trans H _; case: (ocoh O C)=>_ V _ _ _.
by apply: umpleq_lastkey V (sth O C pleq_refl).
Qed.

Lemma takenM b : monoP scoh (fun s => \lkj s = b).
Proof. by move=>?? /coh_valid A; rewrite !lE //= joinA. Qed.

(* trylock erases to a CAS *)
Program Definition trylock : {p h k}, STsep [spin]
  (fun s => [/\ pi.s s = p, ht.s s = h & k <= last_key (ht.t s)],
   fun b s => if b then exists t,
     [/\ pi.s s = p, ht.s s = t \-> (false, true) \+ h & k < t]
   else [/\ pi.s s = p, ht.s s = h & k <= last_key (ht.t s)]) :=
  Do (actPx (fun s t b =>
       (t, b) = if \lkj s then (id_t, false) else (lock_t, true))).
Next Obligation.
split=>[?????|???|????] /coh_valid V; last by [rewrite !lE];
by case: ifP=>H [->] // -> [->->].
Qed.
Next Obligation.
move=>i [p][h][k][[/= P H K]] C; apply: vrf_atm=>// s O [].
move/(stk O C): K=>K; rewrite !(osf O) {i C O} in P H; case E: (\lkj s).
- exists false, s; split=>[|m O C]; first by rewrite /eff /= E; exists id_t.
  by rewrite (stk O) //= -?(osf O) // E.
eexists true, _; split=>[|m O C]; first by exists lock_t; rewrite /eff /= E.
by exists (fresh (ht.t s)); rewrite -?(osf O) !lE /= -H P.
Qed.

(* then lock loops over trylock *)
Definition lock_tp := {p h k}, STsep [spin]
  (fun s => [/\ pi.s s = p, ht.s s = h & k <= last_key (ht.t s)],
   fun (_ : unit) s => exists t,
     [/\ pi.s s = p, ht.s s = t \-> (false, true) \+ h & k < t]).

Program Definition lock :=
  ffix (fun (loop : unit -> lock_tp) xx =>
    Do (b <-- trylock;
        if b then ret tt else loop tt)) tt.
Next Obligation.
move=>loop [i][p][h][k][[/= P H K]] C; apply: [stepE p, h, k]=>//= {i P H K C}.
case=>i; last by move=>*; apply: [gE p, h, k].
by case=>t [<-??] C; apply: vrf_ret=>// m O; exists t; rewrite -!(osf O).
Qed.

(* unlock action adds unlock entry to history if it finds *)
(* the lock taken; otherwise it doesn't add anything *)
Program Definition unlock : {p h k}, STsep [spin]
  (fun s => [/\ pi.s s = p.+1, ht.s s = h & k <= last_key (ht.t s)],
   fun _ s => exists t, k <= t <= last_key (ht.t s) /\
     ((pi.s s = p /\ ht.s s = t \-> (true, false) \+ h) \/
      [/\ pi.s s = p.+1, ht.s s = h & atval false t (ht.t s) = false])) :=
  Do (actPx (fun s t (_ : unit) =>
        t = if \lkj s then unlock_t else id_t)).
Next Obligation.
split=>[??[]?[] _ ->->|?? _ ? ->|??? _ /coh_valid V ->] //;
by [case: ifP | rewrite !lE].
Qed.
Next Obligation.
move=>i [p][h][k][[/= P H K]] C; apply: vrf_atm=>// s O.
rewrite !(osf O) in P H; move/(stk O C): K=>{i C O}K C; case E: (\lkj s);
rewrite /eff /= E; eexists tt, _.
- split=>[|m O {}C]; first by eexists _; split=>//=; rewrite E P.
  case: (ocoh O C)=>_ V _ _ _; exists (fresh (ht.t s)).
  rewrite -!(osf O) !lE -joinA /= -H in V *.
  rewrite (leq_trans K (ltnW _)) ?(lastkey_max,domPtUn,V,inE,eq_refl) //.
  by split=>//; left.
split=>[|m O {}C]; first by eexists _.
case: (ocoh O C)=>_ V _ _ _; exists (last_key (ht.t s)).
have G: gapless (ht.t s) by case: C=>???? [].
rewrite (leq_trans K) //= (stk O C) //; split=>//; right.
by rewrite -(osf O) -{1}(osf O) (umpleq_atval _ G V (sth O C pleq_refl)).
Qed.

End Spin.

(* we export that field names are morphisms *)
Module Exports.
Canonical pi_morph.
Canonical ht_morph.
End Exports.
End Spin.

Export Spin.Exports.


(**********************************************)
(* CSL resource combines a spinlock with CSLX *)
(* by coupling Spin.genlock/CSLX.open and     *)
(* Spin.genunlock/CSLX.close.                 *)
(**********************************************)

(* Unlike in the first example, We use the new spinlocks from above,
rather than ones form lock.v.

This requires more interesting proofs, as we actually have to reason
out of the histories in the specs of Spin.unlock, in order to
determine if the unlocking succeeded (turned lk from true to false) or
failed (found lk already set to false). *)

Module CSL.
Section CSL.
Variables (lk : ptr) (inv : heap -> Prop).
Variable prec_pf : precise inv.
Notation spin := (Spin.spin lk).
Notation cslx := (CSLX.cslx prec_pf).

Definition U := [pcm of Spin.U * CSLX.U].
Definition T : Type := Spin.T * CSLX.T.

(* this is a self-contained resource, so everything is internal *)
Inductive cslTr := lock_t | unlock_t | write_t A of ptr & A |
                   grant_t |
                   id_t.

Definition csl_interp t :=
  match t with
    lock_t => [| interp spin Spin.lock_t, interp cslx CSLX.open_t]
  | unlock_t => [| interp spin Spin.unlock_t, interp cslx CSLX.close_t]
  | write_t A p v =>
      [| interp spin Spin.id_t, interp cslx (CSLX.write_t p v)]
  | grant_t => [| interp spin Spin.take_t, interp cslx CSLX.give_t]
  | id_t => id_tr [* coh spin, coh cslx]
  end.

Definition csl_intern (t : cslTr) := true.

(* first an intermedate resource csl' *)
(* that takes a product of Spin and Xfer *)
(* We will eventually produce csl by restricting it *)
Program Canonical csl' := ResPx [* coh spin, coh cslx] csl_interp csl_intern _.
Next Obligation.
case=>[||A p v||] _ s x [_ C _]; last by [move=>->];
case=>_ H _; (apply: domeqUn=>//; first by apply: domeqX); rewrite /flat /=.
- by case: H=>_ ->; rewrite !(CSLX.idE, lE) domeqX.
- by case C; case: H; case=>-> E _ -> /mxMx -> _ ->; rewrite !lE E domeqX.
- by case: H=>B [w][h][E ->]; rewrite !lE E domeqX.
by case: H=>_ ->; rewrite !lE.
Qed.

Definition pi : U -> _ := Spin.pi \o fst.
Canonical pi_morph := Eval hnf in [morphism of pi].
Definition ht : U -> _ := Spin.ht \o fst.
Canonical ht_morph := Eval hnf in [morphism of ht].
Definition mu : U -> mtx3 := CSLX.mu \o snd.
Canonical mu_morph := Eval hnf in [morphism of mu].
Definition hp : U -> _ := CSLX.hp \o snd.
Canonical hp_morph := Eval hnf in [morphism of hp].
Definition tx : U -> _ := omg \o Spin.ht \o fst.
Canonical tx_morph := Eval hnf in [morphism of tx].

Lemma idE : @idfun U = [** fst & snd].
Proof. by apply/fext=>x; rewrite /= -prod_eta. Qed.

(* we will eventually restrict csl' into csl *)
Definition cinv (x y : Ux csl') :=
  [&& 'sep tx x y,
      pi x == if mu x == auth1 then 1 else 0,
      pi y == if mu y == auth1 then 1 else 0,
      tx x == if mu x == nown then nown else own &
      tx y == if mu y == nown then nown else own].

Arguments cinv x y /.

Lemma cslinv_sep_ax : seprel_axiom cinv.
Proof.
split=>[|x y V|x y V|x y z V] /=; first by rewrite sep00 !pfunit.
- by rewrite (sepC _ V); do !bool_congr.
- by case/and5P=>S -> _ -> _; rewrite (sepx0 V S) !pfunit.
case/and5P=>Sxy Mx My Tx Ty /and5P [Sz _ Mz _ Tz].
rewrite !pfjoin ?(validL V) ?(validAR V) ?(sepAxx V Sxy Sz) //=.
rewrite (eqP Tx) (eqP Ty) (eqP Tz)(eqP Mx) (eqP My) (eqP Mz) !eq_refl /=.
have: valid (mu y \+ mu z) by case/andP: V=>_ /andP []; rewrite -joinA=>/validR.
by case: mxPJ=>//; case.
Qed.

Canonical cinv_seprel := Eval hnf in seprel cinv cslinv_sep_ax.

(* cinv is an invariant for csl' *)
Lemma cslpf : sub_persP cinv PredT.
Proof.
move=>t s x C V _ H; rewrite /= !lA /=.
case: t H=>[||A p a||]; last by [move=>->].
- rewrite !lprodE; case; case=>H1 -> [H2 ->] _.
  rewrite !lE -sepSXst -!lprodX /= in H1 H2 *.
  case/and5P=>S /eqP Ms /eqP Mo /eqP Ts /eqP To _; rewrite omgP //= in H1.
  move/eqP/mxE0: H1=>H1; rewrite !H1 in Ts To *.
  case: eqP Ts Ms=>// -> _ ->; case: eqP To Mo=>// -> _ ->.
  by rewrite !sepE /= -comp1f /= (lCSX Spin.ht) !omg_fresh // H1.
- rewrite !lprodE; case; case=>[[N _]] -> [_ ->] _;
  rewrite !lE -sepSXst -!lprodX /= in N *.
  case: C=>_ [W _ _ _]; rewrite -lprodX in W.
  case/and5P=>S /eqP Ms /eqP Mo /eqP Ts /eqP To _.
  case E: (pi.s s) N Ms W Mo Ts To=>[//|n].
  case: eqP E=>// ->; case: n=>// E _ _ /mxMx -> /= -> _ To.
  by rewrite To !sepE /= -comp1f (lCSX Spin.ht) !omg_fresh.
- by rewrite !lprodE; case=>-> [B][w][h][_ ->] _; rewrite !lE.
rewrite !lprodE; case=>[->][E ->] _; rewrite !lE /=.
case: C=>_ [W _ _ _]; rewrite {}E in W *; move/mxMx: W=>-> /=.
by rewrite !sepE !lA (lCSX omg); case/and5P=>-> /eqP ->->->->.
Qed.

(* morphism from spin to csl' *)

Section MorphSpinCSL'.

Definition sphi (_ : unit) (sw : st csl') :=
  [fun sv : st spin => sv = sw\1].

(* the transition part of the morphism, the following *)
Definition stau' (_ : unit) (t : Tr spin) : Tr csl' :=
  match t with
    Spin.lock_t => lock_t
  | Spin.unlock_t => unlock_t
  | Spin.take_t => unlock_t (* doesn't matter *)
  | Spin.id_t => id_t
  end.

Definition stau x t (s : st csl') t' := t' = stau' x t.

(* the simulation condition *)
Definition ssim (_ : unit) :=
  [fun sw : st csl' =>
     cinv (idfun.s sw) (idfun.o sw) /\
     if (mu.s sw == own) || (mu.s sw == auth1) then inv (hp.s sw) else hp.s sw = Unit].

Lemma slocal : morph_localP sphi.
Proof.
exists (fun x : Ux csl' => x.1); case=>p sw sv ?? ->.
by exists sw\1; rewrite /= !prodE.
Qed.

Lemma sother : morph_otherP sphi.
Proof. by case; case=>???? _ _ _ _ ->-> O; rewrite -!prodEQ0 O. Qed.

Lemma sfunc : morph_funcP sphi. Proof. by case=>??? _ _ _ ->->. Qed.
Lemma sback : morph_backwardP sphi. Proof. by case; case. Qed.
Lemma sint : morph_internalP stau. Proof. by case. Qed.
Lemma stfunc : morph_tfuncP stau. Proof. by case=>t ???? ->->. Qed.
Lemma stloc : morph_tlocalP stau. Proof. by []. Qed.
Lemma scoh : morph_cohP' sphi. Proof. by case=>?? [C _ _] ->. Qed.

Definition sm := Morph (MorphMixin slocal sother sfunc sback
                                   sint stfunc stloc scoh).

Lemma sostep : morph_ostepP sm ssim.
Proof.
case=>sw sw' sv [Cs Cx V _] /= E.
rewrite /Morph.phi /= in E *; subst sv; case=>S M O; exists sw'\1.
have {Cs} Aw : allvalid sw by rewrite prodVE (coh_valid Cs) (coh_valid Cx).
have St : 'sep tx (idfun.s sw) (idfun.o sw) by case/and5P: S.
rewrite !lA /= in S *.
case: O; case=>[||A p v||] H; last by [move: H=>/tr_inj ->];
(split=>//; last by [case: H; rewrite !prodE=>/iosteps]);
case: H; rewrite /= !lprodE !prodE ?lTtr !lE /=.
- case=>H1 -> [H2 ->] _; rewrite !lE !sepE !lA -!lprodX omgP //= in H1 H2 *.
  move/eqP/mxE0: H1 H2=>H1 /mxE0 H2; rewrite !H1 !H2 /= in S M *.
  by rewrite !omg_fresh ?H1 //=; case/and4P: S=>_ /eqP -> /eqP ->.
- case=>H1 -> [H2 ->] _; rewrite !lE !sepE !lA -!lprodX omgP //= in H1 H2 *.
  case: H1=>N _; case/and5P: S M=>S /eqP Ms /eqP Mo /eqP Ts /eqP To.
  have W : valid (mu.s sw \+ mu.o sw) by rewrite lprodX; case: Cx.
  case Es : (mu.s sw) W Ms Mo N Ts To=>//=; last by move/mxMx=>-> _ ->.
  case: eqP=>//= Eo; last by move=>_ _ ->.
  by rewrite Eo=>_ ->-> _ /= Ts To ->; rewrite !omg_fresh //= Ts.
- by move=>-> [B][w][h][_ /trS ->]; rewrite !lE -sepSXst -!lprodX.
move=>-> [Eo ->] _; rewrite !lE; case: Cx; rewrite Eo=>/mxMx Es _ _.
rewrite !lprodE Es Eo /= (lCSX omg) !sepE !lA /= in M S *.
by case/and5P: S=>-> /eqP -> /eqP ->->->.
Qed.

Lemma sforw : morph_forwardP sm ssim.
Proof.
move=>t [] sw sv sv1 J [Cs Cx V] _ /= E [S M] H; exists (stau' tt t), tt.
rewrite /Morph.phi /= in E *; subst sv.
have {Cs} Aw : allvalid sw by rewrite prodVE (coh_valid Cs) (coh_valid Cx).
have St : 'sep tx (idfun.s sw) (idfun.o sw) by case/and5P: S.
rewrite /flat /= -!lprodX !lA in V.
case: t J H=>[|||] //= J H; last by move: H=>->; exists sw.
- exists [@ sv1, State (own, snd.j sw) Unit (snd.o sw)]; case: Cx=>_ _.
  rewrite /prod_rel !lA !lprodE !prodE !lE /= -!lprodX !sepE !omgP //= in S H *.
  case: H=>/eqP H ->; rewrite /flat /= !lA !lE /= (lCSX CSLX.mu);
  rewrite (lCSX Spin.pi) (lCSX Spin.ht) (lCSX omg) (lCSX CSLX.hp) (validX V).
  move/mxE0: H S=>H /and5P [S Ms Mo /eqP Ts /eqP To].
  rewrite !H in Ts To *. case: eqP Ts=>// Ts _; case: eqP To=>// To _.
  rewrite {}Ts {}To /= in Ms Mo M *.
  by rewrite (eqP Ms) (eqP Mo) M !unitR !omg_fresh ?H.
exists [@ sv1, State (nown, Unit) (hp.s sw) (snd.o sw)].
rewrite /prod_rel !lA !lprodE !prodE !lE /= -!lprodX !sepE !omgP //= in H S *.
case: H; case=>N _ ->; rewrite /flat /= !lA !lE /= (lCSX CSLX.mu);
rewrite (lCSX Spin.pi) (lCSX Spin.ht) (lCSX omg) (lCSX CSLX.hp) (validX V).
case/and5P: S M=>S /eqP Ms /eqP Mo /eqP Ts /eqP To.
have W : valid (mu.t sw) by rewrite lprodX; case: Cx.
case Es: (mu.s sw) W Ms Mo N Ts To=>[||t] //=; first by move=>_ ->.
move/mxMx=>-> /=; case: t Es=>[[]|] /=; last by move=>_ ->.
by move=>Ms ->-> _ Ts To Eh; rewrite !omg_fresh // Ts To //= unitR.
Qed.

Definition smsim : simulation sm := Simulation2 sostep sforw.

End MorphSpinCSL'.

Program Definition grant : STsep [csl']
  (fun s => [/\ mu.s s = own, smsim tt s & inv (hp.s s)],
   fun _ s => [/\ mu.s s = auth1, smsim tt s & inv (hp.s s)]) :=
  Do (actSx (fun _ : unit => grant_t)).
Next Obligation. by split=>// ??? [][]. Qed.
Next Obligation.
move=>i [[/= M S J]] C; apply: vrf_atm=>// j O [_ _ V].
move: (morph_osteps_sim smsim tt C (erefl _) S O)=>{C}{}S.
rewrite !(osf O) !lprodE /= in M; eexists tt, _; split.
- by rewrite effE; apply: act2_safe=>//; rewrite /flat /= !lE (validX V).
move=>m {}O C; set k := [@ _, _] in O C.
split; last 1 first.
- by case: S=>_; rewrite -(osf O) !lprodE !lE M.
- by rewrite -(osf O) !lprodE !lE /=.
apply: (morph_osteps_sim smsim tt C)=>//=.
rewrite !sepE !lA !lprodX  /k !sbeta !lA !lE /= (lCSX omg) -!lprodX /=.
rewrite /= !sepE !lA !lprodE !lA M -!lprodX /= in S.
by case: S=>/and5P [-> /eqP ->->->->].
Qed.

(* we shouldn't need equations on hp.s as they follow from smsim and mu *)
(* but they don't hurt *)

Program Definition lock' : STsep [csl']
  (fun s => [/\ mu.s s = nown, hp.s s = Unit & smsim tt s],
   fun _ s => [/\ mu.s s = own, ssim tt s & inv (hp.s s)]) :=
  Do (morph_app sm tt (Spin.lock lk)).
Next Obligation.
move=>i [[/= Em Eh S]] _; apply: (vrf_morph smsim)=>// C.
have {C Eh} Whi : valid (ht.t i) by case: C; rewrite !lprodE.
have {S} Ep : pi.s i == 0 by case: S=>/and5P []; rewrite !lA Em.
apply: [gE Spin.pi.s i\1, Spin.ht.s i\1, last_key (Spin.ht.t i\1)]=>//=.
case=>m [t][H1 H2 H3] _ [j] /= X [S M] C _; rewrite {m}X in H1 H2.
rewrite S; rewrite -!lprodX !lA /= in H1 H2 H3 S M C *; rewrite -{}H1 in Ep.
have Whj : valid (ht.s j) by case: C=>[[_]] /validL; rewrite lprodE.
have Wmj : valid (mu.t j) by case: C=>_ []; rewrite lprodE.
rewrite !sepE !lA /= !(lCSX Spin.ht) in S.
have {Whi H2 H3} E : last_val false (ht.s j).
have N : t != 0 by move: Whj; rewrite H2=>/validPtUn_cond.
- rewrite lastkeyUn Whi gtn_max in H3; case/andP: H3=>H3 _.
  rewrite H2 lastvalUn -H2 Whj lastkeyPt // lastvalPt //=.
  by rewrite ltnNge leq_eqVlt H3 orbT.
rewrite omgP in E; rewrite (eqP E) (eqP Ep) in S.
by case/and5P: S M; case: (mu.s j) Wmj=>[||[]].
Qed.

Program Definition unlock' : STsep [csl']
  (fun s => [/\ mu.s s = own, smsim tt s & inv (hp.s s)],
   fun _ s => [/\ mu.s s = nown, ssim tt s & hp.s s = Unit]) :=
  Do (grant;;
      morph_app sm tt (Spin.unlock lk)).
Next Obligation.
move=>j [/= H] _; apply: stepE=>//= {H} [[i]][Em S J] C.
apply: (vrf_morph smsim)=>// _.
have {S Em} [/eqP P Ts] : pi.s i == 1 /\ last_val false (ht.s i).
- by case: S=>/and5P []; rewrite !lA Em omgP.
have {C} W1 : valid (ht.t i) by case: C=>[[]]; rewrite lprodE.
apply: [gE 0, Spin.ht.s i\1, last_key (Spin.ht.t i\1)]; rewrite /= -!lprodX //.
case=>{P} m [t] H _ [xx] X [S Em] C _; rewrite {m}X !lA -!lprodX /= in S H.
rewrite !lA S; rewrite !sepE /= !lA !(lCSX Spin.ht) /= in S.
have {C} W2 : valid (ht.t xx) by case: C=>[[]]; rewrite lprodE.
case: H=>/andP [H1 H2]; case/and5P: S=>S /= /eqP Msx _ /eqP Tsx _.
case; last first.
- case=>_ Eh {Em Msx Tsx J}.
  rewrite -{}Eh in Ts W1 H1; rewrite !lastkeyUn {}W1 W2 in H1 H2.
  rewrite geq_max in H1; case/andP: H1=>H1 _.
  rewrite /omg_s Ts /= in S; case/andP: S=>S _.
  rewrite maxnC /maxn S leq_eqVlt ltnNge {}H1 orbF in H2.
  have {S} N : t > 0 by rewrite (eqP H2); apply: leq_trans S.
  rewrite {H2}(eqP H2) in N *.
  rewrite /atval findUnL // {N} (dom_lastkeyE N).
  by rewrite /last_val /atval in Ts; rewrite Ts.
case=>Ps Eh; have N : t != 0.
- by rewrite Eh -joinA in W2; move/validPtUn_cond: W2.
rewrite lastkeyUn W1 geq_max in H1; case/andP: H1=>H1 _.
have: last_val false (ht.s xx) = false.
- rewrite Eh lastvalUn -Eh (validL W2).
  by rewrite lastkeyPt // leqNgt ltnS H1 lastvalPt.
rewrite omgP; case Ex2 : (mu.s xx == _) Tsx=>-> // _.
by rewrite (eqP Ex2) /= in Em *.
Qed.

(* now we package things up *)

(* first we promote cinv into the coherence predicate *)
(* i.e., all coherent states will satisfy the requirements of cinv *)
Definition newpcm : sub_pcm cinv := xsepSubPCM cinv_seprel.
Definition csl := @sub_res csl' cinv_seprel _ newpcm globalT cslpf.
Definition sub_morf : morph unit csl' csl :=
  @sub_morph csl' cinv_seprel _ _ globalT cslpf.
Definition sub_sim : simulation sub_morf :=
  @sub_sim csl' cinv_seprel _ _ globalT cslpf.

(* now we create abstract morphisms to extract state components *)
(* importantly, mx is undefined on mu's that are auth1 *)
Definition hx : Ux csl -> _ := hp \o pval.
Canonical hx_morph := Eval hnf in [morphism of hx].
Definition mx : Ux csl -> _ := mxundef \o mu \o pval.
Canonical mx_morph := Eval hnf in [morphism of mx].
Definition px : Ux csl -> _ := pi \o pval.
Canonical px_morph := Eval hnf in [morphism of px].

Program Definition lock : STsep [csl]
  (fun s => [/\ mx.s s = nown & hx.s s = Unit],
   fun _ s => [/\ mx.s s = own & inv (hx.s s)]) :=
  Do (morph_app sub_morf tt lock').
Next Obligation.
move=>i [[/= M H]] /coh_sep [] ///= S _ _; apply: (vrf_morph sub_sim)=>// _.
rewrite !lsubX -lCSX in M H; case: mxundefP M=>// M _.
apply: gE=>//=; first by rewrite S M.
case=>m {M S H} [M S H] _ [j] X _ _ _; rewrite {m}X in S M H.
by rewrite !lsubX -lCSX M.
Qed.

Program Definition unlock : STsep [csl]
  (fun s => [/\ mx.s s = own & inv (hx.s s)],
   fun _ s => [/\ mx.s s = nown & hx.s s = Unit]) :=
  Do (morph_app sub_morf tt unlock').
Next Obligation.
move=>i [[/= M H]] /coh_sep [] ///= S _ _; apply: (vrf_morph sub_sim)=>// _.
rewrite !lsubX -lCSX in M H; case: mxundefP M=>// M _.
apply: gE=>//=; first by rewrite S M.
case=>m {M S H} [M S H] _ [j] X _ _ _; rewrite {m}X in S M H.
by rewrite !lsubX -lCSX M.
Qed.

Definition modal (a : mtx2) (s : st csl) := mx.s s = a /\ valid a.

Lemma mx_law1 (s : st csl) a b :
        coh csl s -> star (modal a) (modal b) s -> modal (a \+ b) s.
Proof.
move=>C [s1][s2][S Ma Mb].
have W : valid (mu.s (spval s)).
- by case/coh_sep: C=>_ _ [_][/validL]; rewrite lprodE.
have {S} Em : mu.s (spval s) = mu.s (spval s1) \+ mu.s (spval s2).
- case: S; rewrite !lT -!lsubX=>H _ _ _ _.
  rewrite (_ : mu \o pval = (mu \o pval) \o idfun) //.
  rewrite -!(lCSX _ idfun) H pfjoin // -H.
  by move/cohso/validL: C.
case: Ma=>Ma Va; case: Mb=>Mb Vb; rewrite /modal.
rewrite !lsubX -!(lCSX mxundef) /mxundef in Ma Mb *.
have {}Ma : mu.s (spval s1) = own /\ a = own \/
            mu.s (spval s1) = nown /\ a = nown.
- by case: (mu.s _) Ma Va=>[||[[]|]] <- //; [right | left].
have {}Mb : mu.s (spval s2) = own /\ b = own \/
            mu.s (spval s2) = nown /\ b = nown.
- by case: (mu.s _) Mb Vb=>[||[[]|]] <- //; [right | left].
by rewrite Em in W *; case: Ma Mb W; case=>->-> []; case=>->->.
Qed.

Lemma mx_law2 (s : st csl) a b :
        coh csl s -> modal (a \+ b) s -> star (modal a) (modal b) s.
Proof.
move=>C; suff {a b} L : forall a, modal a s -> star (modal a) (modal Unit) s.
- case; case: b=>[||[]] //=; first by rewrite join_undef.
  - by rewrite unitR=>M V; apply: L.
  by move=>M /mxMx ?; subst a; rewrite starC; apply: L.
by move=>a [M V]; exists s, (outs s); do !split=>//; rewrite lO pfunit.
Qed.

End CSL.
End CSL.
