From Coq Require Import ssreflect ssrbool ssrfun.
From mathcomp Require Import ssrnat eqtype seq fintype finfun.
From fcsl Require Import pred prelude pcm automap heap natmap.
From fcsl Require Import pcm morphism notation lift mutex state world model priv.
From fcsl Require Import treiber flatcomb.
From fcsl Require Import options.

(**************************************************************************)
(* Some client examples for the flat combiner.                            *)
(*                                                                        *)
(* 1. First example instantiates the flat combiner with stacks, specified *)
(* by means of histories, as we did for Treiber, or in Section 5 of the   *)
(* paper.                                                                 *)
(*                                                                        *)
(* 2. Second example couples the stack instance of the flat combiner with *)
(* a simple counter method. The collect_tr transition of the flat         *)
(* combiner is coupled with the increment transition of the counter.      *)
(*                                                                        *)
(* As the collect_tr transition is the "linearization point" of the flat  *)
(* combiner helping functionality, the coupling achieves synchronicity:   *)
(* whenever the flat combiner helps to push or pop an element off the     *)
(* stack, the counter, which is in a separate object, increments.  In the *)
(* related work, such functionality is always achieved by means of        *)
(* higher-order ghost code, but here we do it with morphism.              *)
(*                                                                        *)
(* This demonstrates that morphisms works with helping.                   *)
(**************************************************************************)

(*******************************************)
(* Instantiating flat combiner with stacks *)
(*******************************************)

Module SC.
Section StackCombine.
Variables (a lk snt : ptr) (n : nat) (A : Type).
Notation hist := (natmap (seq A * seq A)).

Definition inv (ht : hist) h :=
  exists p h' al,
    [/\ h = snt :-> p \+ h', llist p al h',
        um_all (fun => Treiber.stackP) ht & complete [::] ht al].

Lemma prec_inv ht : precise (inv ht).
Proof.
move=>h1 h2 t1 t2 V [p1][h1'][ls1][E1 L1 K1 C1][p2][h2'][ls2][E2 L2 K2 C2].
rewrite E1 E2 -!joinA in V *; case/(hcancelV V)=>? {}V; subst p2.
by case/(llist_prec V L1 L2)=>???; subst ls2 h2' t2.
Qed.

Inductive code := pushc | popc.

Definition inT c : Type := if c is pushc then A else unit.
Definition outT c : Type := if c is pushc then unit else option A.

Definition specR c :=
  if c is pushc as c' return inT c' -> outT c' -> hist -> hist -> _
  then fun e _ ht g =>
    let: vs := last_val [::] ht in g = fresh ht \-> (vs, e :: vs)
  else fun _ (ow : option A) ht g =>
    if ow is Some v then
      exists vs, last_val [::] ht = v :: vs /\ g = fresh ht \-> (v :: vs, vs)
    else last_val [::] ht = [::] /\ g = Unit.

Lemma uniqR m (v : inT m) (w : outT m) g g1' g2' :
        valid (g \+ g1') -> valid (g \+ g2') ->
        specR v w g g1' -> specR v w g g2' -> g1' = g2'.
Proof.
by move=>_ _; case: m v w=>/=[v _ ->->|_ [w [ls1][->->][ls2][[->]->]|[_ ->][]]].
Qed.

Variable seq_push : forall v, {g}, STsep [priv]
  (fun s : st priv => inv g (self s),
   fun r s => exists g', [/\ valid (g \+ g'),
     inv (g \+ g') (self s) & @specR pushc v r g g']).

Variable seq_pop : forall (_ : unit), {g}, STsep [priv]
  (fun s : st priv => inv g (self s),
   fun r s => exists g', [/\ valid (g \+ g'),
     inv (g \+ g') (self s) & @specR popc tt r g g']).

Definition s_interp c :=
  if c is pushc as c' return forall v : inT c', {g}, STsep [priv]
   (fun s : st priv => inv g (self s),
    fun r m => exists g', [/\ valid (g \+ g'),
      inv (g \+ g') (self m) & @specR c' v r g g'])
  then seq_push else seq_pop.

Definition F := FlatCombineT a lk n prec_inv s_interp uniqR.
Notation fp := (FP.fp F).

Implicit Type s : st fp.

(* We recover the notation for the flat combiner's ghosts *)

(* self thread ids *)
Notation "\tid_s s" := ((self s\2).1.1) (at level 24).
(* other thread ids *)
Notation "\tid_o s" := ((other s\2).1.1) (at level 24).
(* total thread ids *)
Notation "\tid s" := (\tid_s s \+ \tid_o s) (at level 24).
(* self mutex *)
Notation "\mu_s s" := ((self s\2).1.2) (at level 24).
(* other mutex *)
Notation "\mu_o s" := ((other s\2).1.2) (at level 24).
(* total mutex *)
Notation "\mu s" := (\mu_s s \+ \mu_o s) (at level 24).
(* self user ghost *)
Notation "\gh_s s" := ((self s\2).2) (at level 24).
(* other user ghost *)
Notation "\gh_o s" := ((other s\2).2) (at level 24).
(* uncollected ghosts *)
Notation "\ghs s" := (FC.ghosts (joint s\2)) (at level 24).
(* total user ghost *)
Notation "\gh s" := (\gh_s s \+ \gh_o s \+ sumg (\ghs s)) (at level 24).
(* uncollected responses *)
Notation "\rsp s" := (FC.resps (joint s\2)) (at level 24).
(* protected heap *)
Notation "\hpj s" := (FC.hpj (joint s\2)) (at level 24).
Notation "\hp s" := (lk :-> (\mu s == own) \+ \hpj s\2 \+
                        updi a (fgraph (\rsp s\2))) (at level 0).
(* private heap *)
Notation "\hp_s s" := (self s\1) (at level 24).
(* other heap *)
Notation "\hp_o s" := (other s\1) (at level 24).
(* allocator's joint state *)
Notation "\hp_j s" := (joint s\1) (at level 24).

Program Definition fc_push tid v : {ht}, STsep [fp]
  (fun s => [/\ \hp_s s = Unit, self s\2 = (#nat_of_ord tid, nown, Unit),
                [pcm ht <= \gh s] & (\rsp s) tid = NoReq F],
   fun _ s => exists t vs, [/\ \hp_s s = Unit,
     self s\2 = (#nat_of_ord tid, nown, t \-> (vs, v :: vs)),
     [pcm ht <= \gh s], last_key ht < t & (\rsp s) tid = NoReq F]) :=
  Do (@FP.flat_combine F pushc tid v).
Next Obligation.
move=>tid v i [ht][[/= H1 H2 H3 H4]] C; apply: [gE ht]=>// {H1 H2 H4 C} [[k]][g'][g1].
case=>H1 H2 H5 H6 H7 H4 C; subst g'; exists (fresh g1), (last_val [::] g1).
have V : valid g1 by case: C=>_ [_ _ _][_ _ V _ _]; case: H5 V=>x -> /validL.
by split=>//; [apply: pleq_trans H6 H5 | apply: umpleq_lastkey H6].
Qed.

Program Definition fc_pop tid : {ht}, STsep [fp]
  (fun s => [/\ \hp_s s = Unit, self s\2 = (#nat_of_ord tid, nown, Unit),
                [pcm ht <= \gh s] & (\rsp s) tid = NoReq F],
   fun v s => [pcm ht <= \gh s] /\ if v is Some e then
       exists t vs, last_key ht < t /\
                    self s\2 = (#nat_of_ord tid, nown, t \-> (e :: vs, vs))
     else exists t,
       [/\ last_key ht <= t, self s\2 = (#nat_of_ord tid, nown, Unit) &
           atval [::] t (\gh s) = [::]]) :=
  Do (@FP.flat_combine F popc tid tt).
Next Obligation.
move=>tid i [ht][[/= H1 H2 H3 H4]] C; apply: [gE ht]=>// {H1 H2 H4 C} v k [g'][g1].
case=>H1 H2 H5 H6 H7 H4 C; split; first by apply: pleq_trans H6 H5.
have [V1 V] : valid g1 /\ valid (\gh k).
- by case: C=>_ [_ _ _][_ _ W]; case: H5 W (W)=>x -> /validL.
case: v H7=>[w [vs]|][X ?]; subst g'.
- by exists (fresh g1), vs; split=>//; apply: umpleq_lastkey H6.
exists (last_key g1); split=>//; first by apply: umpleq_lastkey H6.
rewrite -{2}X /last_val /atval; case: H5 V=>x -> V.
case E : (unitb g1); last by rewrite findUnL // dom_lastkey // E.
by move/unitbE: E=>->; rewrite !lastkey0 !find_cond.
Qed.

End StackCombine.
End SC.

(**************************************************************************)
(* Let's try to entangle with a counting object. The idea is to create a  *)
(* resource with the functionality of counting how many times we have     *)
(* invoked a stack operation. This will excercise the coupling of a stack *)
(* transitions with a non-idle incrementing transition in the counter     *)
(*                                                                        *)
(* Eventually, one can discern between the operations. E.g., we can count *)
(* pushes only, or pops only, or pushes of even numbers, etc., but we     *)
(* don't bother with that for now.                                        *)
(**************************************************************************)

Module Ctr.
Section Counter.
Local Definition U := [pcm of nat].
Local Definition T := unit.
Implicit Type s : state U T.

Program Definition ccoh := CohPx2 (fun s => True) (fun s => Unit) _.
Next Obligation.  by []. Qed.

Program Definition incr_tr n :=
  TransPx (fun s x => x = State (self s + n) (joint s) (other s)) ccoh _.
Next Obligation.
split=>[s ->->|->|p /lbO O|]; rewrite -?lT // !zg_unlock /= O addnAC.
by case=> ->->.
Qed.

Inductive cTr := incr_t of nat | id_t P of monoP ccoh P.

Definition cinterp tr :=
  match tr with
    incr_t n => incr_tr n
  | id_t _ pf => pre_tr (id_tr ccoh) pf
  end.

Lemma cres_int : res_internalP cinterp xpredT. Proof. by []. Qed.

Definition ctrMixin := ResourceMixin cres_int.
Canonical ctr := Eval hnf in Resource ctrMixin.
End Counter.
End Ctr.

(* Now let's entangle *)

Section Entangling.
Variables (a lk snt : ptr) (n : nat) (A : Type).
Notation hist := (natmap (seq A * seq A)).
Notation inv := (@SC.inv snt A).
Variable seq_push : forall v, {g}, STsep [priv]
  (fun s : st priv => inv g (self s),
   fun r s => exists g', [/\ valid (g \+ g'),
     inv (g \+ g') (self s) & @SC.specR A SC.pushc v r g g']).
Variable seq_pop : forall (_ : unit), {g}, STsep [priv]
  (fun s : st priv => inv g (self s),
   fun r s => exists g', [/\ valid (g \+ g'),
     inv (g \+ g') (self s) & @SC.specR A SC.popc tt r g g']).
Definition F := (@SC.F a lk snt n _ seq_push seq_pop).

Notation ctr := Ctr.ctr.
Notation fp := (FP.fp F).

Definition U := [pcm of FP.U F * Ctr.U].
Definition T : Type := FP.T F * Ctr.T.

(* We pair up the flat combiner instantiated with stacks *)
(* with the counter *)

Inductive combTr :=
  reqhelp_t (m : fc_code F) of 'I_n & fc_inT m |
  dohelp_t (m : fc_code F) of 'I_n & fc_outT m |
  collect_t (m : fc_code F) of 'I_n & fc_outT m |
  take_t of heap | give_t of heap |
  in_priv of Tr priv | lock_t | unlock_t | id_t P of monoP [* coh fp,coh ctr] P.

(* the collect transition of the flat combiner is coupled *)
(* with the increment transition of counter *)
(* everything else is coupled with identity *)

Program Definition collect_tr m tid (w : fc_outT m) :=
  ex_tr [* coh fp, coh ctr]
    (fun g => [| interp fp (FP.in_flat (@FC.collect_t F _ tid w g)),
                 interp ctr (Ctr.incr_t (size (dom g)))]) _.
Next Obligation. by move=>???????? _ [[_ [[-> _ _ _ _ _]]]] _ _ [[_ [[->]]]]. Qed.

Definition comb_interp tr :=
  match tr with
    reqhelp_t m tid v => [| interp fp (FP.in_flat (@FC.reqhelp_t F _ tid v)),
                            id_tr (coh ctr)]
  | dohelp_t m tid w => [| interp fp (FP.in_flat (@FC.dohelp_t F _ tid w)),
                            id_tr (coh ctr)]
  | collect_t m tid w => collect_tr tid w
  | take_t h => [| interp fp (FP.in_flat (@FC.take_t F h)), id_tr (coh ctr)]
  | give_t h => [| interp fp (FP.in_flat (@FC.give_t F h)), id_tr (coh ctr)]
  | in_priv t => [| interp fp (FP.in_priv F t), id_tr (coh ctr)]
  | lock_t => [| interp fp (FP.lock_t F), id_tr (coh ctr)]
  | unlock_t => [| interp fp (FP.unlock_t F), id_tr (coh ctr)]
  | id_t _ pf => pre_tr (id_tr [* coh fp, coh ctr]) pf
  end.

Definition comb_intern t : Prop :=
  if t is (take_t _ | give_t _) then false else
   if t is in_priv t' then internal priv t' else true.

Program Canonical comb := Eval hnf in ResPx [* coh fp, coh ctr] comb_interp comb_intern _.
Next Obligation.
case=>[m t v|m t w|m t w|h|h|t|||P pf] //;
try by [rewrite /tcoh /= (@tcohE fp)];
move=>I s x [Cf Cc _] H; rewrite /flat /=;
try by [case: H=>J -> _; apply: domeqUn (iint _ _ J) _].
- by case: H=>g [J1 J2 _]; apply: domeqUn=>//; apply: iint J1.
by case: H=>_ ->.
Qed.

(* we need a morphism from fp to comb *)

Definition cphi (_ : unit) (s : st comb) (x : st fp) := x = s\1.
Definition ctau' (_ : unit) (t : Tr fp) :=
  match t with
    FP.in_flat (FC.reqhelp_t m t v) => reqhelp_t t v
  | FP.in_flat (FC.dohelp_t m t w) => dohelp_t t w
  | FP.in_flat (FC.collect_t m t w g) => collect_t t w
  | FP.in_flat (FC.take_t h) => take_t h
  | FP.in_flat (FC.give_t h) => give_t h
  | FP.in_flat (FC.id_t P pf) => id_t (prodM (prodM idM pf) idM)
  | FP.in_priv t => in_priv t
  | FP.lock_t => lock_t
  | FP.unlock_t => unlock_t
  | FP.id_t _ pf => id_t (prodM pf idM)
  end.

Definition ctau x t (s : st comb) t' := t' = ctau' x t.

Lemma cp_local : morph_localP cphi.
Proof.
exists (fun x => x.1)=>[[p s _]] _ _ ->.
by rewrite /cphi !prodE; exists s\1.
Qed.

Lemma cp_other : morph_otherP cphi.
Proof. by case; case=>???? _ _ _ _ ->-> /prod_inj []; rewrite !prodE. Qed.

Lemma cp_func : morph_funcP cphi.
Proof. by case=>??? _ _ _ ->->. Qed.

Lemma cp_back : morph_backwardP cphi. Proof. by case; case. Qed.

Lemma cp_int : morph_internalP ctau.
Proof. by move=>t _ [] _ I _ ->; case: t I=>//; case. Qed.

Lemma cp_tfunc : morph_tfuncP ctau. Proof. by case=>t ???? ->->. Qed.

Lemma cp_tloc : morph_tlocalP ctau. Proof. by []. Qed.

Lemma cp_coh : morph_cohP' cphi. Proof. by case=>?? [C _ _] ->. Qed.

Definition cp :=
  Morph (MorphMixin cp_local cp_other cp_func cp_back
                    cp_int cp_tfunc cp_tloc cp_coh).

Lemma cp_ostep : morph_ostepP cp (fun => xPredT).
Proof.
case=>s1 s _ _ _ -> _ []; case=>[m t v|m t w|m t w|h|h|t|||P pf] H;
exists s\1; split=>//; last by [case: H=>_ /tr_inj ->];
try by case: H; rewrite !prodE=>H _ _; apply: iosteps H.
by case: H=>g []; rewrite !prodE=>H _ _; apply: iosteps H.
Qed.

Lemma cp_forw : morph_forwardP cp (fun => xPredT).
Proof.
move=>t [] sc _ sf J C _ -> _ H; exists (ctau' tt t), tt; case: t J H=>[||||];
last by [move=>p pf _ [H ->]; exists sc];
[ |move=>t J[H1 H2 V]|move=>J [h][H1 H2 V]|move=>J [h][H1 H2 V]];
try by [exists [@ sf, sc\2];
        rewrite /Morph.phi /= /cphi /prod_rel !prodE !(validX V) //;
        do !split=>//; exists h].
case=>[m t v|m t w|m t w g|h|h|P pf] // J [];
last by [move=>H1 H2 _; exists [@ sf, sc\2]; rewrite /Morph.phi /=
         /cphi !prodE [sf]seta; case: H2=>H2 ->; rewrite H1 -!seta];
try by [move=>H1 H2 V; exists [@sf, sc\2];
        rewrite /Morph.phi /= /cphi /prod_rel !prodE !(validX V)].
move=>H1 H2 V.
exists [@ sf, State (self sc\2 + size (dom g)) (joint sc\2) (other sc\2)].
rewrite /Morph.phi /= /cphi /= /prod_rel !prodE; split=>//; exists g.
by rewrite !prodE !(validX V) //.
Qed.

Definition cpsim : simulation cp := Simulation2 cp_ostep cp_forw.

(* Now, we can already lift the flat combiner programs to the *)
(* composed resource. However, the lifting alone will not let us *)
(* infer that the change in the stack entailed a change in the counter. *)
(* We need to impose that as an invariant, by means of resource restriction. *)
(* Then we prove that the invariant is inductive, as in other, simpler *)
(* examples. Then, we can easily lift the programs, and exploit the *)
(* information that the change in the stack also changed the counter *)

(* We first need to state the invariant that we want. *)
(* Which is that the domain of the stack history equals the counter *)
(* That way, we maintain the property that we actually count the *)
(* operations. We could have more involved relations too, as already *)
(* mentioned. But we don't bother for now *)

(* The invariant is in the subjective components, not in the joint *)
(* So, we need to use a self/other form of restriction. This *)
(* has not been explained in the paper, but it's similar to the *)
(* ordinary restriction construction *)

(* here's the invariant *)
(* it says that the size of the self history equals the counter *)
(* if the history is undefined, then we have 0 by default *)
(* but this case will never arise in programs *)
Definition hist_count (x : Ux comb) :=
  size (dom x.1.2.2) = if valid (x.1.2.2) then x.2 else 0.

(* this particular restriction construction requires *)
(* proving some properties of the hist_count invariant *)

Lemma hc_pcm : pcm_closed hist_count.
Proof.
split=>[|u v]; rewrite /hist_count; first by rewrite dom0 valid_unit.
move=>H1 H2; case: ifP=>[V|/negbT/invalidE ->]; last by rewrite dom_undef.
by rewrite (size_domUn V) H1 H2 !(validE2 V).
Qed.

(* Next we prove that the invariant hist_count is inductive *)
(* I'm reusing here the all_cohP definition, which states inductivity *)
(* of a helper predicate that there exists s1 satisfying the invariant *)
(* to which s erases. In the future, I should define a separate *)
(* version of all_cohP, which states this directly. It will be easier to *)
(* follow the proofs. *)

Lemma hc_ind : all_cohP comb (fun s => exists s1, s = ustval hc_pcm xpredT s1).
Proof.
case=>[m t v|m t w|m t w|h|h|t|||P pf] s s' // _ _; last by [case=>_ ->];
move=>H [s1 ?]; subst s; move: (ustps s1) (ustpo s1)=>Xs Xo;
suff [H1 H2 H3] : [/\ hist_count (self s'), true & hist_count (other s')]
  by [exists (ustpair (pc:=hc_pcm) _ H1 H2 H3); rewrite ustE];
rewrite /hist_count !prodEQ0.
- by case: H; case=>_ [_ ->] _ -> _; rewrite -!prodEQ0.
- by case: H; case=>_ [g][g'][_ ->] _ -> _;rewrite -!prodEQ0.
- case: H=>g [[->]][[-> _ _ _] ->] _ -> _;rewrite -!prodEQ0 /=.
  case: ifP=>[V|/negbT/invalidE ->]; last by rewrite dom_undef Xo.
  by rewrite size_domUn // Xs Xo (validE2 V).
- by case: H=>[[_ -> _]] -> _; rewrite -!prodEQ0.
- by case: H=>[[h]][_][_ ->] _ -> _; rewrite -!prodEQ0.
by case: H=>[[h]][_][_ ->] _ -> _; rewrite -!prodEQ0.
Qed.

Definition ucomb := ust_res xpredT hc_pcm.
Definition um := ust_morph xpredT hc_pcm.
Definition umsim : simulation um := ust_sim hc_ind.
Definition sm := comp_morph cp um.
Definition smsim := comp_sim cpsim umsim.

(* We can now morph a program. The push will now not only do a push, *)
(* but also change the counter *)

(* we first adjust indexes *)

(* self user ghost *)
Notation "\gh_s s" := ((self (ustval _ _ s)\1\2).2) (at level 24).
(* other user ghost *)
Notation "\gh_o s" := ((other (ustval _ _ s)\1\2).2) (at level 24).
(* uncollected ghosts *)
Notation "\ghs s" := (FC.ghosts (joint (ustval _ _ s)\1\2)) (at level 24).
(* total user ghost *)
Notation "\gh s" := (\gh_s s \+ \gh_o s \+ sumg (\ghs s)) (at level 24).
(* uncollected responses *)
Notation "\rsp s" := (FC.resps (joint (ustval _ _ s)\1\2)) (at level 24).
(* private heap *)
Notation "\hp_s s" := (self (ustval _ _ s)\1\1) (at level 24).
(* the counter *)
Notation "\ct_s s" := (self (ustval _ _ s)\2) (at level 24).

(* and can now lift the program, so that it induces the change to the counter *)

Program Definition counting_push tid v : {ht n}, STsep [ucomb]
  (fun s : st ucomb => [/\ \hp_s s = Unit,
    self (ustval _ _ s)\1\2 = (#nat_of_ord tid, nown, Unit),
    [pcm ht <= \gh s], (\rsp s) tid = NoReq F & \ct_s s = n],
   fun _ s => exists t vs, [/\ \hp_s s = Unit,
     self (ustval _ _ s)\1\2 = (#nat_of_ord tid, nown, t \-> (vs, v :: vs)),
     [pcm ht <= \gh s], last_key ht < t, (\rsp s) tid = NoReq F & \ct_s s = n.+1]) :=
  Do (morph_app sm tt (SC.fc_push tid v)).
Next Obligation.
move=>tid v i [ht][m][[/= H1 H2 H3 H4 H5]] _; apply: (vrf_morph smsim)=>[|//|_].
- by exists (ustval _ _ i).
apply: [gE ht]=>//{H1 H3 H4}[[k]][/= t][vs][H1 X2 H3 X3 H4] C [j].
have {C} V : valid (self k\2).2 by case: C=>_ [_ _ _ [_ _]] /validL/validL.
case=>_ [->] E _ _ _; rewrite {k}E in H1 X2 H3 X3 H4 V.
exists t, vs; split=>// {H1 H3 H4 X3}; move: V (ustps i) (ustps j).
by rewrite /hist_count !prodEQ0 H2 X2 dom0 domPtK valid_unit validPt -H5=>-><-<-.
Qed.

End Entangling.
