From Coq Require Import ssreflect ssrbool ssrfun.
From mathcomp Require Import ssrnat eqtype seq.
From fcsl Require Import axioms pred prelude pcm automap heap natmap lift mutex.
From fcsl Require Import morphism notation state world model priv invertible.
From fcsl Require Import options.

(*****************************************************************************)
(*  * Implementation and Verification of Ticketed Lock: a tutorial           *)
(*                                                                           *)
(*  A ticket lock is a locking algorithm, a type of spinlock that uses       *)
(*  tickets to control which thread of execution is allowed to enter         *)
(*  a critical section.                                                      *)
(*                                                                           *)
(*  The basic concept of a ticket lock is similar to the ticket queue        *)
(*  management system. This is the method that many bakeries and delis use   *)
(*  to serve customers in the order that they arrive, without making them    *)
(*  stand in a line. There is a dispenser from which customers pull          *)
(*  sequentially numbered tickets upon arrival and there is also a display   *)
(*  with the ticket number that is now being served.                         *)
(*                                                                           *)
(*  When the next ticket number (customer) is ready to be served, the        *)
(*  "Now Serving" sign is incremented to call the customer with the          *)
(*  consecutive number.                                                      *)
(*                                                                           *)
(*****************************************************************************)




(* ** Ticket Labels                                                          *)
(*                                                                           *)
(* We define the type of labels, define ordered prdicate, and show its       *)
(* properties                                                                *)
(*                                                                           *)

Module Labels.

(* label type *)
Inductive label := waitT | serveT | usedT.

Definition eq_lbl (x y : label) : bool :=
match x, y with
  waitT, waitT => true
| serveT, serveT => true
| usedT, usedT => true
| _, _ => false
end.

Lemma eq_lblP : Equality.axiom eq_lbl.
Proof. by case=>[||][||] //=; constructor. Qed.

Definition lbl_eqMixin := EqMixin eq_lblP.
Canonical lbl_eqType := Eval hnf in EqType label lbl_eqMixin.

(* defining ordered predicates on natmaps to labels *)

Definition altnb (m : natmap label) (a b : label) :=
  all (fun t => all (ltn t)
    (dom (um_filter [pts _ v|v == b] m)))
    (dom (um_filter [pts _ v|v == a] m)).

Definition altn (m : natmap label) (a b : label) :=
 forall t k, (t, a) \In m -> (k, b) \In m -> t < k.

Lemma altnP m a b : reflect (altn m a b) (altnb m a b).
Proof.
case V : (altnb m a b); constructor; last first.
- move=>H; apply/(elimF idP V)/allP=>x /dom_umfilt [v][/eqP ->] /H K.
  by apply/allP=>k /dom_umfilt [w][/eqP ->]; apply: K.
move/allP: V=>H t k T K.
have /H/allP : t \in dom (um_filter [pts _ v | v == a] m).
- by apply/dom_umfilt; exists a.
by apply; apply/dom_umfilt; exists b.
Qed.

Definition ordered m :=
  [&& altnb m serveT waitT, altnb m usedT serveT & altnb m usedT waitT].

(* we proove a weak variant, the lemma can be strenghtened to equality since   *)
(* ordered is downward closed                                                  *)
Lemma ord_fresh m : ordered m -> ordered (fresh m \-> waitT \+ m).
Proof.
case V : (valid m); last first.
- by move/invalidE: (negbT V)=>->; rewrite /ordered !join_undef.
have W : valid (fresh m \-> waitT \+ m) by rewrite valid_fresh.
case/and3P=>/altnP H1 /altnP H2 /altnP H3.
apply/and3P; split; apply/altnP=>t k; rewrite !InPtUnE //.
- by case=>// T [[->]|]; [move/In_dom/dom_ordfresh: T|apply: H1].
- by case=>// T [] //; apply: H2.
by case=>// T [[->]|]; [move/In_dom/dom_ordfresh: T|apply: H3].
Qed.

Lemma ordered0 : ordered Unit.
Proof. by apply/and3P;split; apply/altnP;rewrite /altn=>t k /In0. Qed.

(* downward closure of ordered *)

Lemma orderedDL x y : valid (x \+ y) -> ordered (x \+ y) -> ordered x.
Proof.
move=>V; case/and3P=>/altnP H1 /altnP H2 /altnP H3; apply/and3P.
by split; apply/altnP=>t k T K; [apply: H1|apply: H2|apply: H3]; apply/InL.
Qed.

Lemma orderedDR x y : valid (x \+ y) -> ordered (x \+ y) -> ordered y.
Proof. by rewrite joinC; apply: orderedDL. Qed.

(* ordered can be lifted to a relation *)

Definition omega_rel x y := ordered (x \+ y).

(* moreover, the relation is separating *)

Lemma omega_seprel : seprel_axiom omega_rel.
Proof.
rewrite /omega_rel; split=>[|x y z|x y z|x y z].
- by rewrite unitL; apply/and3P; split; apply/altnP=>k t /In0.
- by rewrite joinC.
- by rewrite unitR; apply: orderedDL.
by rewrite -joinA=>V _ H; rewrite H (orderedDR V H).
Qed.

End Labels.

Export Labels.

(* ** Example of non-separating relation                                    *)
(*                                                                          *)
(* We show here that gapless, one of the predicates we use to constrain the *)
(* state space of TL resource is not a separating relation (and thus cannot *)
(* be imposed on the underlying PCM                                         *)

Module NoGaps.
Parameter A : Type.
Implicit Type h : natmap A.

(* lifting gapless to a relation *)

(** printing upsilon_rel %\ensuremath{\bot_\upsilon}% #&bot;<sub>&upsilon;</sub># *)
Definition  upsilon_rel h1 h2 := gaplessb (h1 \+ h2).

Lemma upsilon_non_seprel (a : A) : ~ seprel_axiom upsilon_rel.
Proof.
rewrite /upsilon_rel;move=>[H1 H2 H3 H4].
have V1: valid ((2 \-> a) \+ (1 \-> a) \+ (3 \-> a)).
rewrite -[_ \+ _ \+ _]unitR -!joinA !validPtUn  //= -!andbA;apply/and4P;split.
- done.
- by rewrite dom0.
- by rewrite unitR domPtK.
- by rewrite unitR domUn inE;apply/negP=>/andP[]?/orP[];rewrite domPtK.
have G1: gaplessb ((2 \-> a) \+ (1 \-> a)).
apply/gaplessP=>t.
rewrite lastkeyUn (validL V1) !lastkeyPt //= /maxn //= -mem_iota //= !inE //=.
move=>/orP[/eqP->|/eqP->].
- by rewrite domUnPtE //= (validL V1).
- by rewrite domPtUnE //= (validL V1).
have G2: gaplessb ((2 \-> a) \+ (1 \-> a) \+ (3 \-> a)).
apply/gaplessP=>t.
rewrite lastkeyUn V1 !lastkeyPt //= /maxn //= lastkeyUn (validL V1) !lastkeyPt /maxn //=.
rewrite -mem_iota //= !inE //= =>/or3P[]/eqP->.
- by rewrite -joinA domUn joinA  V1 //= inE;apply/orP;right;
    rewrite domPtUnE //= (validAR V1) //=.
- by rewrite -joinA domPtUnE //= joinA V1 //=.
- by rewrite -joinA domUn joinA V1 inE //=;apply/orP;right;
    rewrite domUnPtE //= (validAR V1) //=.
move:(H4 _ _ _ V1 G1 G2)=>/andP[]/gaplessP H;move: (H 2)=>H'.
have: (2 \in dom (1 \-> a \+ 3 \-> a))
  by apply:(H' _);rewrite lastkeyUn (validAR V1) //= !lastkeyPt //= /maxn.
rewrite !domPtUn //= inE (validAR V1) //=.
rewrite -[3 \-> a]unitR domPtUn unitR (validR V1) inE dom0 //=.
Qed.

End NoGaps.


(* ** Ticket lock resource                                                      *)

(* Ticketed lock is a structure containing two pointers : *)
Structure ticketT := TicketT {
  tdr : ptr;   (* ticket dispenser pointer *)
  dsp : ptr;   (* display pointer *)
}.

Module Ticket.
Section Ticket.
Variable R : ticketT.
Notation pTdr := (tdr R).
Notation pDsp := (dsp R).

Definition U := [tpcm of natmap label].
Definition T := unit.

(** printing sg %\ensuremath{\sigma}% #&sigma;# *)
(** printing sg.l %\ensuremath{\sigma_l}% #&sigma;<sub>l</sub># *)
(** printing sg.r %\ensuremath{\sigma_r}% #&sigma;<sub>r</sub># *)
(** printing sg.x %\ensuremath{\sigma_x}% #&sigma;<sub>x</sub># *)
Definition sg : U -> _ := idfun.
Canonical sg_morph := Eval hnf in [morphism of sg].
Arguments sg /.


Definition tick m := last_key (um_filter [pts _ v | v == usedT] m).

Lemma psi_gt0 x : (tick x).+1 > 0.
Proof. by []. Qed.

(** printing psi %\ensuremath{\psi}% #&psi;# *)
(** printing psi.l %\ensuremath{\psi_l}% #&psi;<sub>l</sub># *)
(** printing psi.r %\ensuremath{\psi_r}% #&psi;<sub>r</sub># *)
(** printing psi.x %\ensuremath{\psi_x}% #&psi;<sub>x</sub># *)
Definition psi x : posNat := Sub _ (psi_gt0 x).

Lemma psi_morph_ax : morph_axiom (@sepT U) psi.
Proof.
rewrite /psi; split=>[|x y V _].
- by apply/eqP; rewrite -val_eqE /= /tick umfilt0 lastkey0.
split=>//; apply/eqP; rewrite -val_eqE /= /tick.
by rewrite umfiltUn // lastkeyUn -umfiltUn //= valid_umfilt V /= maxnSS.
Qed.

Canonical psi_morph := Eval hnf in Morphism' _ psi_morph_ax.
Arguments psi /.

Notation disp m := (val (psi m)).


(* ** helper lemmas for ordered *)

(* display is an upper bound of used tickets *)
Lemma disp_upb m t a :
        ordered m -> (t, a) \In sg m -> (disp m <= t) = (a != usedT).
Proof.
case/and3P=>_ /altnP O1 /altnP O2; case: eqP=>[->|Kn] K /=.
- by apply: negbTE; rewrite -leqNgt lastkey_max //; apply/dom_umfilt; exists usedT.
case: (posnP (tick m))=>[->|]; first by move/In_dom/dom_cond: K; rewrite lt0n.
case/dom_lastkeyE/dom_umfilt=>v [/eqP ->] X.
by case: a K Kn=>// K _; [apply: O2 | apply: O1].
Qed.

(* updating display to serveT and to usedT preservers ordered *)
(* we don't require that disp m \-> waitT *)

Lemma ord_lock m : ordered m -> ordered (upd (disp m) serveT m).
Proof.
move=>O; case/and3P: (O)=>/altnP H1 /altnP H2 /altnP H3; apply/and3P.
split; apply/altnP=>t k; rewrite !InU /=; case=>_ T [_];
case: eqP T=>/eqP Tn T; case: eqP=>/eqP Kn K //;
try by [apply: H1 | apply: H2 | apply: H3]; last first.
- by rewrite (eqP Kn) ltnS leqNgt (disp_upb O T).
by rewrite (eqP Tn) ltn_neqAle eq_sym Kn (disp_upb O K).
Qed.

(* ditto, not required that disp m \-> serveT *)
Lemma ord_unlock m : ordered m -> ordered (upd (disp m) usedT m).
Proof.
move=>O; case/and3P: (O)=>/altnP H1 /altnP H2 /altnP H3; apply/and3P.
split; apply/altnP=>t k; rewrite !InU /=; case=>_ T [_];
case: eqP T=>/eqP Tn T; case: eqP=>/eqP Kn K //;
try by [apply: H1 | apply: H2 | apply: H3];
by rewrite (eqP Tn) ltn_neqAle eq_sym Kn (disp_upb O K).
Qed.

(* ** Resource TL *)

(* we erase to the heap in parameter R *)

Let xflat s : heap := pTdr :-> last_key (sg.x s) \+ pDsp :-> val (psi.x s).

(* resource invariant *)

Program Definition xcoh :=
  CohPx2 (fun s : state U T => [/\ valid (sg.x s), ordered (sg.x s),
    gapless (sg.x s) & valid (xflat s)]) xflat _.
Next Obligation.
split=>[s [V ???]|p s A]; first by rewrite allvalidE lT V.
by rewrite /xflat !lE // !joinA.
Qed.

(* transitions *)

(* take_tr transition for taking a fresh ticket *)
Program Definition taketx_tr :=
  TransPx2 (fun s s' =>
    s' = State (fresh (sg.x s) \-> waitT \+ sg.l s) tt (sg.r s)) xcoh _.
Next Obligation.
split=>[s ->->|->|p ?? O|-> [V O G Vf]]; rewrite //= ?lE //=.
- by case=>->-> *; rewrite !joinA O.
by rewrite -joinA valid_fresh (ord_fresh O) gp_fresh (validX Vf).
Qed.

(* lock_tr transition for grabbing the lock *)
Program Definition lock_tr :=
  TransPx2 (fun s s' => (val (psi.x s), waitT) \In sg.l s /\
    s' = State (upd (val (psi.x s)) serveT (sg.l s)) tt (sg.r s)) xcoh _.
Next Obligation.
split=>[s [?->][?->]|[?->]|p A ? O|[H ->][V O G Vf]]; rewrite ?lE //=.
- rewrite -maxnA O; case=>H [->->] _ _; split; last first.
  - by rewrite updL // (In_dom H).
  by apply/InL=>//; rewrite -allvL in A; rewrite (validL A).
rewrite !lE /= -updL ?(In_dom H) // maxnSS validU V (validX Vf); split=>//.
- by move/ord_lock: O; rewrite pfjoin // -maxnSS; apply; apply/InL.
move=>k /andP [Nk Gh]; rewrite domU inE /=; case: eqP=>// N.
rewrite lastkeyU in Gh; first by apply: G; rewrite Nk Gh.
by rewrite domUn inE V -maxnSS (In_dom H).
Qed.

(* unlock_tr transition for releasing the lock *)
Program Definition unlock_tr :=
  TransPx2 (fun s s' => (val (psi.x s), serveT) \In sg.l s /\
    s' = State (upd (val (psi.x s)) usedT (sg.l s)) tt (sg.r s)) xcoh _.
Next Obligation.
split=>[s [?->[?->]]|[?->]|p A ? O|[H ->][V O G Vf]]; rewrite ?lE //=.
- rewrite -maxnA O; case=>H [->->] _ _; split; last first.
  - by rewrite updL // (In_dom H).
  by apply/InL=>//; rewrite -allvL in A; rewrite (validL A).
rewrite !lE /= -updL ?(In_dom H) // maxnSS validU V (validX Vf); split=>//.
- by move/ord_unlock: O; rewrite pfjoin // -maxnSS; apply; apply/InL.
move=>k /andP [Nk Gh]; rewrite domU inE /=; case: eqP=>// N.
rewrite lastkeyU in Gh; first by apply: G; rewrite Nk Gh.
by rewrite domUn inE V -maxnSS (In_dom H).
Qed.

(* we put the resource together *)

Inductive xTr := taketx_t | lock_t | unlock_t | id_t.

Definition xinterp t :=
  match t with
    lock_t => lock_tr
  | unlock_t => unlock_tr
  | taketx_t => taketx_tr
  | id_t => id_tr xcoh
  end.

Definition xintern (t : xTr) := true.

Program Canonical TL := ResPx xcoh xinterp xintern _.
Next Obligation.
case=>_ ?? _ /=; try by [move=>->; apply: domeqX];
by case=>_ ->; apply: domeqX.
Qed.

(* ** Stability properties *)

(* others can't increase display past out waitT ticket *)

Lemma lock_stab s1 s2 t :
        osteps TL s1 s2 -> coh TL s1 ->
        (t, waitT) \In sg.l s1 ->
        val (psi.x s1) <= t -> val (psi.x s2) <= t.
Proof.
move: s1 s2; apply: pres_iterf_ind=>// s1 s2; first by apply: ocoh1.
case; case; last by move/tr_inj=>->.
- move/trS=>-> C H; rewrite !lE [in fresh _]joinC; split=>//=.
  apply: leq_trans; rewrite {2}/tick umfiltPtUn /=.
  by rewrite valid_freshUn ?(validX (cohso C)).
- case=>H1 /trS -> C H2; rewrite !lE [in upd _]joinC; split=>//=.
  apply: leq_trans; rewrite {2}/tick umfiltU !maxnSS /=.
  rewrite ltnS geq_max leq_maxl leq_max orbC lastkey_mono //.
  by move=>x; rewrite domF inE; case: ifP.
case=>H1 /trS -> C H2; rewrite !lE // [in upd _]joinC; split=>//=.
rewrite !lE in H1; case: (C)=>V /and3P [/altnP O _ _] _ _ _.
rewrite {2}/tick umfiltU /= !maxnSS /= gtn_max.
have T : val (psi.x s1) < t by apply: O; [apply/InR | apply/InL].
apply/andP; split; apply: leq_ltn_trans T; rewrite /= maxnSS in H1 *.
- by rewrite ltnW // ltnS leq_maxl.
rewrite lastkeyUmax validU valid_umfilt (validR V) /=.
case: ifP=>[|_]; first by case/dom_umfilt=>v [/eqP ->] /(In_fun H1).
by rewrite geq_max ltnS leqnn /= ltnW // ltnS leq_maxr.
Qed.

(* strong monotonicity of display *)
(* this holds also for msteps but we don't need the stronger result *)
Lemma psi_mono s1 s2 :
        osteps TL s1 s2 -> coh TL s1 ->
        val (psi.r s1) <= val (psi.r s2).
Proof.
move: s1 s2; apply: iters_ind=>// s1 s2; first by move=>s3; apply: leq_trans.
- by apply: ocoh.
case; case; last by move/tr_inj=>->.
- move/trS=>-> C; rewrite !lE /= {2}/tick umfiltPtUn.
  by rewrite valid_freshUn ?(validX (cohso C)).
- case=>H /trS -> C; rewrite !lE /= !maxnSS in H *.
  apply: lastkey_mono=>z /dom_umfilt [v][/eqP -> Z].
  rewrite umfiltU domF inE; case: eqP Z=>[<- /(In_fun H)|Nz Z] //.
  by apply/dom_umfilt; exists usedT.
case=>H /trS -> C; rewrite !lE /= !maxnSS in H *.
apply: lastkey_mono=>z /dom_umfilt [v][/eqP -> Z].
rewrite umfiltU /= domU inE /= valid_umfilt (validX (cohso C)).
by case: eqP=>Nz //; apply/dom_umfilt; exists usedT.
Qed.

(* weak monotonicity of display *)
Lemma psi_mono_weak s1 s2 :
        osteps TL s1 s2 -> coh TL s1 ->
        val (psi.x s1) <= val (psi.x s2).
Proof.
by move=>O C; rewrite -(osf O) geq_max !leq_max leqnn (psi_mono O C) orbT.
Qed.

(* strengtening of stability *)
Lemma lock_stab_str s1 s2 :
        osteps TL s1 s2 -> coh TL s1 ->
        (val (psi.x s1), waitT) \In sg.l s1 ->
        val (psi.x s2) = val (psi.x s1).
Proof.
by move=>O C H; apply/eqP; rewrite eqn_leq (lock_stab O C) ?(psi_mono_weak O C).
Qed.

(* only we can unlock *)

Lemma unlock_stab s1 s2 :
        osteps TL s1 s2 -> coh TL s1 ->
        (val (psi.x s1), serveT) \In sg.l s1 ->
        val (psi.x s2) = val (psi.x s1).
Proof.
move: s1 s2; apply: pres_iters_ind=>// s1 s2;
  first by [apply: ocoh1]; first by [move=>s3 ->];
  [move: s1 s2; apply: iters_ind=>// s1 s2; first by apply: ocoh|];
(case; case; last by [move/tr_inj=>->];
first by [move/trS=>-> C K; rewrite !lE /= {2}/tick umfiltPtUn
  valid_freshUn ?(validX (cohso C))];
case=>/In_dom H /trS -> C /In_dom; rewrite !lE /= in H *;
by rewrite (negbTE (dom_inNR (cohso C) H))).
Qed.


(* ** Actions *)


(* Take ticket action  *)
(* I'm making it large foot right away, to remove the framing step later on *)
Program Definition taketx : {h},STsep [TL]
  (fun s => sg.l s = h,
   fun t s => sg.l s = t \-> waitT \+ h) :=
  Do (actPx (fun s tr t => (tr, t) = (taketx_t, (fresh (sg.x s))))).
Next Obligation.
by split=>[?????? [->->][->->]||???? /coh_valid A] //; rewrite !lE //= joinA.
Qed.
Next Obligation.
move=>s [h][/= P] C; apply vrf_atm=>// j O C'.
eexists (fresh (sg.x j)), _; split=>[|m Om _]; first by exists taketx_t.
by rewrite -(osf Om) !lE -(osf O) P.
Qed.

(* Fetch display action - !dsp *)
Program Definition fetch_dsp : {h}, STsep [TL]
  (fun s => sg.l s = h,
   fun t' s => sg.l s = h /\ t' <= val (psi.x s)) :=
   Do (actPx (fun s tr t' => (tr, t') = (id_t, val (psi.x s)))).
Next Obligation.
by split=>[?????? [->->][->->]||???? /coh_valid A] //; rewrite !lE //= -maxnA.
Qed.
Next Obligation.
move=>s [h][/= P] _; apply vrf_atm=>// j O C'.
eexists (val (psi.x j)), _; split=>[|m Om _]; first by exists id_t.
by rewrite -{1}(osf Om) -{1}(osf O) psi_mono_weak.
Qed.

(* Lock action *)
Program Definition locktx : {h}, STsep [TL]
  (fun s => sg.l s = val (psi.x s) \-> waitT \+ h,
   fun _ s => sg.l s = val (psi.x s) \-> serveT \+ h) :=
  Do (actPx (fun s tr u => (tr, u) = (lock_t, tt))).
Next Obligation. by split=>[??????[->->][]||]. Qed.
Next Obligation.
move=>s [h][S] C; apply: vrf_atm=>// j O Cj.
have S' : (val (psi.x s), waitT) \In sg.l s.
- by rewrite S; apply/InPtUnL; rewrite -S (validX (cohso C)).
have {O C S' s} S : sg.l j = val (psi.x j) \-> waitT \+ h.
- by rewrite (lock_stab_str O C) // -(osf O).
have S' : (val (psi.x j), waitT) \In sg.l j.
- by rewrite S; apply/InPtUnL; rewrite -S (validX (cohso Cj)).
eexists tt, _; split=>[|m Om Cm] /=; first by exists lock_t.
set k := State _ _ _ in Om Cm.
have /= X : val (psi.x j) = val (psi.x k).
- rewrite !lE; congr (maxn _.+1 _); rewrite {2}/tick umfiltU /= {1}maxnSS.
by rewrite freeND //; apply/negP=>/dom_umfilt [v][/eqP ->] /(In_fun S').
move: (unlock_stab Om Cm)=>/= ->; rewrite -X.
- by rewrite -(osf Om) !lE {1}S /= updPtUn.
by rewrite !lE InU /= eq_refl validU maxnSS (validX (cohso Cj)).
Qed.

(* We define loop invariant of lock separately *)
Definition lock_tp t := {h},STsep [TL]
  (fun (s : st TL) => sg.l s = t \-> waitT \+ h,
   fun (_ : unit) (s : st TL) => sg.l s = t \-> waitT \+ h /\ val (psi.x s) = t).

Program Definition lock : {h}, STsep [TL]
  (fun s => sg.l s = h,
   fun _ s => sg.l s = val (psi.x s) \-> serveT \+ h) :=
  Do (t <-- taketx;
      ffix (fun (loop : unit -> lock_tp t) _ => Do (
        t' <-- fetch_dsp;
        if t == t' then ret tt else loop tt)) tt;;
      locktx).
Next Obligation.
move=>t loop [s][h][/= Sg] C; apply: [stepM _]=>[|t' m [P1 P2] Cm M] //.
case: ifP=>[/eqP E|_]; last by apply: [gM h]=>//=; rewrite P1.
rewrite -{}E in P2; apply: vrf_ret=>j Oj Cj; case: (Cm)=>Vm Or _ _.
split; first by rewrite -(osf Oj) -Sg.
have W : (t, waitT) \In sg.l m.
- by rewrite P1 Sg InPtUnE -?Sg -?P1 ?(validL Vm) //; left.
have T : val (psi.x m) = t.
- by apply/eqP; rewrite eqn_leq P2 andbT -pfjoin // (disp_upb Or (InL _ W)).
by rewrite -T in W; move: (lock_stab_str Oj Cm)=>/= ->.
Qed.
Next Obligation.
move=>s [h][/= P] _; apply: [stepM h]=>//= t' m {P} P' C M.
apply: vrf_bnd; apply: [gM h]=>//; case=>j [{P' C M} P1 P2] C M.
by apply: [gM h]=>//=; rewrite {1}P1 -P2.
Qed.

(** Unlock action *)

Program Definition unlock : {h t}, STsep [TL]
  (fun s => sg.l s = (t \-> serveT \+ h) /\ val (psi.x s) = t,
   fun (_ : unit) s => sg.l s = (t \-> usedT \+ h)) :=
  Do (actSx (fun _ => unlock_t)).
Next Obligation. by split=>[???[[]]|]. Qed.
Next Obligation.
move=>s [h][t][[S E]] C; apply: vrf_atm=>j O Cj; rewrite -{}E in S *.
have S' : (val (psi.x s), serveT) \In sg.l s.
- by rewrite S; apply/InPtUnL; rewrite -S (validX (cohso C)).
rewrite -(unlock_stab O C S').
have {O C S' s} S : sg.l j = val (psi.x j) \-> serveT \+ h.
- by rewrite (unlock_stab O C S') -(osf O).
have S' : (val (psi.x j), serveT) \In sg.l j.
- by rewrite S; apply/InPtUnL; rewrite -S (validX (cohso Cj)).
eexists tt, _; split=>[|m Om Cm] /=; first by rewrite effE.
by rewrite -(osf Om) !lE S updPtUn.
Qed.

(** Alpha morphism *)

(** printing serve %\ensuremath{\#_\textt{serve}}% #&#35;<sub>serve</sub># *)
Notation serve x := (um_count [pts _ v | v == serveT] x).

(** printing alpha_rel %\ensuremath{\bot_\alpha}% #&bot;<sub>&alpha;</sub># *)
Definition alpha_rel x y := serve (sg x) + serve (sg y) <= 1.

Lemma alpha_sep_ax : seprel_axiom alpha_rel.
Proof.
rewrite /alpha_rel; split=>[|x y z|x y z|x y z V].
- by rewrite umcnt0.
- by rewrite addnC.
- by rewrite umcnt0 addn0; apply: leq_trans (leq_addr _ _).
rewrite !umcntUn ?(validX V) //= -addnA andbC =>_ H.
by rewrite H; apply: leq_trans (leq_addl _ _) H.
Qed.

Canonical alpha_seprel := Eval hnf in seprel alpha_rel alpha_sep_ax.


(** printing alpha %\ensuremath{\alpha}% #&alpha;# *)
(** printing alpha.l %\ensuremath{\alpha_l}% #&alpha;<sub>l</sub># *)
(** printing alpha.r %\ensuremath{\alpha_r}% #&alpha;<sub>r</sub># *)
(** printing alpha.x %\ensuremath{\alpha_x}% #&alpha;<sub>x</sub># *)
Definition alpha x : mtx2 := if serve (sg x) == 0 then nown else own.

Lemma alpha_ax : morph_axiom alpha_rel alpha.
Proof.
rewrite /alpha/alpha_rel; split=>[|x y z]; first by rewrite umcnt0.
rewrite umcntUn //= addn_eq0; case: (um_count _ _)=>[|n1];
by case: (um_count _ _)=>[|n2] //=; rewrite addSn addnS.
Qed.

Canonical alpha_morph := Morphism' alpha alpha_ax.

(* if the called ticket is waiting, there's no other tickets served *)
(* follows from the ordering property *)
Lemma cnt_wait s :
        coh TL s -> (val (psi.x s), waitT) \In sg.x s -> serve (sg.x s) = 0.
Proof.
case=>V O _ _ K; apply/eqP/contraT; rewrite -lt0n -has_predT.
case/hasP=>x /dom_umfilt [v][/eqP ->] K' _.
have : val (psi.x s) <= x by move: (disp_upb O K'); rewrite pfjoin.
by case/and3P: O=>/altnP /(_ _ _ K' K); case: ltngtP.
Qed.

(* if we have more than one serve ticket then the smallest is called *)
Lemma cnt_serve s :
        coh TL s -> serve (sg.x s) > 0 -> (val (psi.x s), serveT) \In sg.x s.
Proof.
move=>C S; case: (C)=>V O G _; have [x X] : exists x, (x, serveT) \In sg.x s.
- by move: S; rewrite -has_predT=>/hasP [x]/dom_umfilt[v][/eqP ->]; exists x.
have /In_domX [w W] : val (psi.x s) \in dom (sg.x s).
- apply: G; rewrite /= {1}maxnSS; apply: leq_trans (lastkey_max (In_dom X)).
  by move: (disp_upb O X); rewrite pfjoin.
case: w W S=>//; first by move/(cnt_wait C)=>->.
by move/(disp_upb O); rewrite pfjoin //= leqnn.
Qed.

(* stability of rel_alpha *)
Lemma alpha_rel_stab s1 s2 :
        msteps (id_morph TL tt) s1 s2 ->
        coh TL s1 ->
        alpha_rel (sg.l s1) (sg.r s1) -> alpha_rel (sg.l s2) (sg.r s2).
Proof.
move: s1 s2; apply: omind=>s1 s2 M C; move: (cohso C)=>V; case: M=>[a _ _ ->|a].
- case: a; rewrite /alpha_rel; last by move=>->.
  - by move=>->; rewrite !lE umcntPtUn ?valid_freshUn ?(validX V).
  - case=>H ->; rewrite !lE (umcntU _ _ H) /= addSn.
    by rewrite -umcntUn // (cnt_wait C) //; apply/InL.
  case=>H ->; rewrite !lE; apply: leq_trans=>/=.
  by rewrite leq_add2r (umcntU _ _ H) leq_pred.
case: a; rewrite /alpha_rel; last by move/tr_inj=>->.
- by move/trS=>->; rewrite !lE umcntPtUn ?valid_freshUn ?(validX V).
- case=>H /trS ->; rewrite !lE in H *.
  by rewrite (umcntU _ _ H) addnS -umcntUn // (cnt_wait C) //; apply/InR.
case=>H /trS ->; rewrite !lE in H *; apply: leq_trans.
by rewrite leq_add2l (umcntU _ _ H) leq_pred.
Qed.

(* Deriving abstract specs *)

Program Definition lock_alpha : STsep [TL]
  (fun s => alpha.l s = nown /\ alpha_rel (idfun.l s) (idfun.r s),
   fun _ s => alpha.l s = own /\ alpha_rel (idfun.l s) (idfun.r s)) :=
  Do lock.
Next Obligation.
move=>s [[H1 H2]] C; apply: [gM _]=>//; case=>m P Cm M; split.
- by rewrite /alpha P umcntPtUn //= -P (validX (cohso Cm)).
by apply: alpha_rel_stab M C H2.
Qed.

Program Definition unlock_alpha : STsep [TL]
  (fun s => alpha.l s = own /\ alpha_rel (idfun.l s) (idfun.r s),
   fun _ s => alpha.l s = nown /\ alpha_rel (idfun.l s) (idfun.r s)) :=
  Do unlock.
Next Obligation.
move=>s [[Hs HD]] C; rewrite /alpha /alpha_rel /= in Hs HD; move: (cohso C)=>V.
have [{}Hs Ho] : serve (sg.l s) = 1 /\ serve (sg.r s) = 0.
- case: posnP Hs HD=>//; case: (serve _)=>// n1 _ _; rewrite addSn.
  by case: n1=>//; rewrite add0n; case: (serve _).
have [h /= Hsg] : exists h, sg.l s = val (psi.x s) \-> serveT \+ h.
- exists (free (sg.l s) (val (psi.x s))); apply: In_eta.
  have : serve (sg.x s) > 0 by rewrite umcntUn ?(validX V) // Hs Ho.
  by case/(cnt_serve C)/InUn=>//; move/umcnt_mem0: Ho=>Ho /Ho.
have Sh : serve h = 0.
- by move: HD; rewrite Hsg umcntPtUn /= -?Hsg ?(validX V) //; case: (serve h).
apply: [gM h, val (psi.x s)]=>//=; case=>m {} Hsg [Vm _ _ _] M.
split; last by apply: alpha_rel_stab M C HD.
by rewrite /alpha Hsg umcntPtUn //= -?Hsg ?(validX Vm) // Sh.
Qed.

(* setting up for the subPCM construction *)

Lemma sublock : sub_persP alpha_rel PredT.
Proof.
move=>??? C ?? K H _; split=>//; apply: alpha_rel_stab C H.
by apply: iter1; apply: SelfMStep K.
Qed.

Definition U' : sub_pcm alpha_rel := xsepSubPCM alpha_seprel.
Definition TL' : resource := sub_res U' globalT sublock.
Definition TL_morph : morph unit TL TL' := sub_morph U' globalT sublock.
Definition TL_sim : simulation TL_morph := sub_sim U' globalT sublock.

(** printing alpha' %\ensuremath{\alpha'}% #&alpha;'# *)
(** printing alpha'.l %\ensuremath{\alpha'_l}% #&alpha;'<sub>l</sub># *)
(** printing alpha'.r %\ensuremath{\alpha'_r}% #&alpha;'<sub>r</sub># *)
(** printing alpha'.x %\ensuremath{\alpha'_x}% #&alpha;'<sub>x</sub># *)
Definition alpha' : U' -> _ := alpha \o pval.
Canonical alpha'_morph := Eval hnf in [morphism of alpha'].

Program Definition lock' : STsep [TL']
  (fun s => alpha'.l s = nown, fun _ s => alpha'.l s = own) :=
  Do (morph_app TL_morph tt lock_alpha).
Next Obligation.
move=>s [/= H] C; apply: (vrf_morph TL_sim)=>// _; apply: gM=>//=.
- by case/coh_sep: C=>// -> _ _; rewrite lV -lsubX -lV.
case=>m [{} H _] _ _ [sw'] X _ /coh_sep [] // _ _ _ _.
by move: X H=>->; rewrite lV -lsubX -lV.
Qed.

Program Definition unlock' : STsep [TL']
  (fun s => alpha'.l s = own, fun _ s => alpha'.l s = nown) :=
  Do (morph_app TL_morph tt unlock_alpha).
Next Obligation.
move=>s [/= H] C; apply: (vrf_morph TL_sim)=>// _; apply: gM=>//=.
- by case/coh_sep: C=>// -> _ _; rewrite lV -lsubX -lV.
case=>m [{} H _] _ _ [sw'] X _ /coh_sep [] // _ _ _ _.
by move: X H=>->; rewrite lV -lsubX -lV.
Qed.

Section Invertibility.

Lemma inv_alpha_seprel : inv_rel alpha_seprel.
Proof.
move=>a1 a2 a' V _; rewrite /= /alpha_rel !pfjoin ?(validLE3 V) //=.
rewrite !umcntUn ?(pfV2,validLE3 V) //= addnCA addnA => H.
by rewrite H andbT; apply: leq_trans H; apply: leq_addr.
Qed.

Lemma inv_alpha : inv_morph [morphism of alpha].
Proof.
split=>[|a b1 b2 Va /=]; first by apply: inv_alpha_seprel.
rewrite /alpha/alpha_rel; case: posnP=>[H _ /esym/mxE0 [->->]|H X].
- by exists a, Unit; rewrite H !pfunit !unitR umcnt0 /=.
case: mxPJ=>//; case=>// _ _ _; [exists a, Unit|exists Unit, a; rewrite addnC];
by rewrite X pfunit umcnt0 -if_neg -lt0n H /= ?unitL ?unitR.
Qed.

Corollary inv_alpha' : inv_morph [morphism of alpha'].
Proof. by apply: (inv_comp_sub U' inv_alpha). Qed.

End Invertibility.
End Ticket.
End Ticket.
