From Coq Require Import ssreflect ssrbool ssrfun.
From mathcomp Require Import ssrnat eqtype seq.
From fcsl Require Import axioms pred prelude pcm automap heap natmap.
From fcsl Require Import morphism notation state world model priv.
From fcsl Require Import options.
Set Warnings "-projection-no-head-constant".

(***********************************)
(* Implementation and verification *)
(* of Treiber's non-blocking stack *)
(***********************************)

(* Main info for the Treiber stack *)
Structure treiberT := TreiberT {
  tb_set : Type;   (* type of elements *)
  tb_snt : ptr}.   (* sentinel pointer *)

Module Treiber.
Section Treiber.
Variable R : treiberT.
Notation A := (tb_set R).
Notation snt := (tb_snt R).

Definition U := [pcm of natmap (seq A * seq A)].
Definition ht : U -> _ := idfun. (* stack history *)
Canonical ht_morph := Eval hnf in [morphism of ht].
Arguments ht /.

(* pi      - sentinel contents *)
(* al(pha) - abstract stack value *)
(* s(tac)k - stack heap *)
(* g(ar)b  - garbage heap *)
(* i(ni)t  - initial stack value *)
Structure T := layout {pi:  ptr; sk: heap; gb: heap; al: seq A; it: seq A}.
Notation "\stgb s" := (sk.j s \+ gb.j s) (at level 24).
Implicit Type s : state U T. (* speeds up typechecking *)

Lemma etaL (s : T) : s = layout (pi s) (sk s) (gb s) (al s) (it s).
Proof. by case: s. Qed.

Definition stackP {A} (v : seq A * seq A) :=
  exists e : A, v.1 = e :: v.2 \/ v.2 = e :: v.1.

(**********************)
(* Resource invariant *)
(**********************)

Program Definition xtreib_coh :=
  CohPx2 (fun s => [/\ valid (snt :-> pi.j s \+ \stgb s),
    llist (pi.j s) (al.j s) (sk.j s), um_all (fun => stackP) (ht.t s) &
    complete (it.j s) (ht.t s) (al.j s)])
         (fun s => snt :-> pi.j s \+ \stgb s) _.
Next Obligation.
by split=>[s []|*]; rewrite ?lO ?lTE' ?lE // allvE=>-> ?? C; rewrite (cm_valid C).
Qed.

(**********************************************)
(* Push transition; p = old, q = new sentinel *)
(**********************************************)

Program Definition push_tr p q v :=
  TransPx2 (fun s x =>
    [/\ pi.j s = p, valid (q :-> (v, p) \+ snt :-> pi.j s \+ \stgb s) &
        x = State (fresh (ht.t s) \-> (al.j s, v :: al.j s) \+ ht.s s)
                  (layout q (q :-> (v, p) \+ sk.j s) (gb.j s) (v :: al.j s) (it.j s))
                  (idfun.o s)]) xtreib_coh _.
Next Obligation.
split=>[s [??->][??->]|[??->]|x ?? O []|]; rewrite ?lE //=.
- by move => <- V [->->]; rewrite -O !joinA in V *.
case=><- V -> []; rewrite !lE //= !(validX V) -joinA; split; try by [eauto];
by [apply: umallPtUn=>//; exists v; right | apply/cm_fresh].
Qed.

(*********************************************)
(* Pop transition; p = old, q = new sentinel *)
(*********************************************)

Program Definition pop_tr p q :=
  TransPx2 (fun s x => exists h v vs,
     [/\ pi.j s = p, sk.j s = p :-> (v, q) \+ h, al.j s = v :: vs, valid (sk.j s) &
         x = State (fresh (ht.t s) \-> (v :: vs, vs) \+ ht.s s)
                   (layout q h (p :-> (v, q) \+ gb.j s) vs (it.j s))
                   (idfun.o s)]) xtreib_coh _.
Next Obligation.
split=>[s||x ?? O|][h][v][vs]; rewrite ?lE //=.
- by case=>->->-> V -> [h2][v2][vs2][_] /(hcancelV V) [[<- _]]<-[<-].
- by case=>???? ->; rewrite !lE.
- case=>??? V [E1 E2 _]; exists h, v, vs.
  by rewrite !lE E1 E2 !joinA -O.
move=>H []; case: H=>->->-> W -> V L ?? /=.
rewrite !lE //= (validX V) -joinA; split=>//; last by apply/cm_fresh.
- by case: L=>r [k][/(hcancelV W)][[->] _ ->].
by apply: umallPtUn=>//; exists v; left.
Qed.

End Treiber.

Module Exports.
Notation xtreib_coh := xtreib_coh.
Canonical ht_morph.
Notation push_tr := push_tr.
Notation pop_tr := pop_tr.
Arguments ht /.
End Exports.

End Treiber.

Export Treiber.Exports.


(* Switching to notation with both priv and treib *)

Module TreiberP.
Section TreiberPriv.
Variable R : treiberT.
Notation A := (tb_set R).
Notation snt := (tb_snt R).
Definition U := [pcm of Priv.U * Treiber.U R].
Definition T : Type := Priv.T * Treiber.T R.
Definition treib_coh := [* priv_coh, xtreib_coh R].
Implicit Type x : state U T.

Definition hp : U -> _ := Priv.hp \o fst.
Canonical hp_morph := Eval hnf in [morphism of hp].
Definition ht : U -> _ := (@Treiber.ht R) \o snd.
Canonical ht_morph := Eval hnf in [morphism of ht].
Definition pi : T -> _ := (@Treiber.pi R) \o snd.
Definition al : T -> _ := (@Treiber.al R) \o snd.
Definition sk : T -> _ := (@Treiber.sk R) \o snd.
Definition gb : T -> _ := (@Treiber.gb R) \o snd.
Definition it : T -> _ := (@Treiber.it R) \o snd.
Notation "\stgb s" := (sk.j s \+ gb.j s) (at level 24).

(***********************)
(* Resource definition *)
(***********************)

Inductive treibTr :=
  push_t of ptr & ptr | pop_t of ptr & ptr |
  id_t P of monoP treib_coh P | in_priv of Tr priv.

Lemma push_ex p q : funcrP treib_coh (fun v => [| Priv.rel_tr (q :-> (v,p)),
                                                  Treiber.push_tr p q v]).
Proof.
move=>????? _ [[?]][E /validL V _] _ _ [[?]][].
by rewrite E in V *; case/(hcancelV V); case.
Qed.

Definition tr_interp tr :=
  match tr with
    push_t p q => ex_tr _ _ (@push_ex p q)
  | pop_t p q => [| id_tr priv_coh, pop_tr R p q]
  | id_t P pf => pre_tr (id_tr treib_coh) pf
  | in_priv tr => [| interp priv tr, id_tr (xtreib_coh R)]
  end.

Definition tr_intern t : Prop := if t is in_priv t then internal priv t else true.

Program Canonical treib := Eval hnf in ResPx treib_coh tr_interp tr_intern _.
Next Obligation.
case=>[p q|p q|P pf|t] /= I s x [Cp _ _]; rewrite /flat /= /flat /=.
- case=>v [X][_ _ ->] _; rewrite !lE.
  by apply: domeqUnf (Priv.relF Cp X) (domeqX _).
- by case=>-> [h][v][vs][_ -> _ _ ->]; rewrite !lE domeqX.
- by case=>_ ->.
by case=>K -> _; apply: domeqUn (iint I Cp K) _.
Qed.

(*****************************)
(* Injecting priv into treib *)
(*****************************)

Implicit Type s : st treib.

Lemma px_local : morph_localP (fun (_:unit) s sp => sp = s\1).
Proof. by exists (fun x:Ux _ => x.1); case=>????? ->; rewrite !prodE; eauto. Qed.

Lemma px_other : morph_otherP (fun (_:unit) s sp => sp = s\1).
Proof. by case; case=>???? _ _ _ _ ->-> /prod_inj []; rewrite !prodE. Qed.

Lemma px_func : morph_funcP (fun (_:unit) s sp => sp = s\1).
Proof. by case=>??? _ _ _ ->->. Qed.

Lemma px_back : morph_backwardP (fun (_:unit) s sp => sp = s\1).
Proof. by case; case. Qed.

Lemma px_int : morph_internalP (fun (_:unit) t _ t' => t' = in_priv t).
Proof. by move=>t _ _ s I C ->. Qed.

Lemma px_tfunc : morph_tfuncP (fun (_:unit) t _ t' => t' = in_priv t).
Proof. by case=>t s t1 t2 C ->->. Qed.

Lemma px_tlocal : morph_tlocalP (fun (_:unit) t _ t' => t' = in_priv t).
Proof. by []. Qed.

Lemma px_coh : morph_cohP' (fun (_ : unit) s sp => sp = s\1).
Proof. by case=>s _ [C _ _] ->. Qed.

Definition pxMixin := MorphMixin px_local px_other px_func px_back
                                 px_int px_tfunc px_tlocal px_coh.
Canonical px := Eval hnf in Morph pxMixin.

Lemma px_ostep : morph_ostepP px (fun => xPredT).
Proof.
case=>s1 s _ _ _ -> _ []; case=>[p q [v]|p q|P pf|t][]; try
  by [move=>_ /tr_inj <-; eexists _]; rewrite !prodE=>H; exists s\1; split=>//.
- by apply: iosteps (Priv.rel_t _) H.
- by move/tr_inj: H <-.
by apply: iosteps H.
Qed.

Lemma px_forw : morph_forwardP px (fun => xPredT).
Proof.
move=>t [] ic _ jr I [Cp _ V] _ -> _ H; exists (in_priv t), tt, [@ jr, ic\2].
rewrite /Morph.phi /= /= /prod_rel !sbeta.
by rewrite -(domeqVUnE (iint I Cp H) (domeq_refl _)).
Qed.

Canonical px_sim := Simulation2 px_ostep px_forw.

(*************)
(* Stability *)
(*************)

Lemma oht h s1 s2 :
        msteps (px tt) s1 s2 -> [pcm h <= ht.t s1] -> [pcm h <= ht.t s2].
Proof.
rewrite !lprodE; move: s1 s2; apply: iter_ind=>// s1 s2 [a _ _ -> [_ ->] //|].
case=>[p q [v]|p q|P pf|t][_]; rewrite ?prodE; try by [move/tr_inj=>->].
- by case=>_ _ /trS -> _ /pleq_trans; rewrite !lE joinCA; eauto.
by case=>k [v][vs][????] /trS -> _ /pleq_trans; rewrite !lE joinCA; eauto.
Qed.

Lemma ostgb h s1 s2 :
        msteps (px tt) s1 s2 -> [pcm h <= \stgb s1] -> [pcm h <= \stgb s2].
Proof.
rewrite !lprodE; move: s1 s2; apply: iter_ind=>// s1 s2 [a _ _ -> [_ ->] //|].
case=>[p q [v]|p q|P pf|t][_]; rewrite ?prodE; try by [move/tr_inj=>->].
- by case=>_ _ /trS -> _ /pleq_trans; rewrite !lE -joinA; eauto.
case=>k [v][vs][_]; rewrite lE=>-> _ _ /trS -> _.
by rewrite !lE joinCA -joinA; eauto.
Qed.

Lemma oit s1 s2 : msteps (px tt) s1 s2 -> it.j s1 = it.j s2.
Proof.
rewrite !lprodE=>O; move: s1 s2 O I; apply: iterf_ind=>// s1 s2 [a _ _ -> [_ ->] //|].
case=>[p q [v]|p q|P pf|t][_]; rewrite ?prodE; try by [move/tr_inj=>->];
by (case=>_ _ /trS -> _ ||case=>k [v][vs][????] /trS ->); rewrite !lE.
Qed.

Lemma ohts s1 s2 : msteps (px tt) s1 s2 -> ht.s s1 = ht.s s2.
Proof.
rewrite !lprodE=>O; move: s1 s2 O I; apply: iterf_ind=>// s1 s2 [a _ _ -> [_ ->] //|].
by move=>a /(iosteps a) O; rewrite -!lprodX (osf O).
Qed.

(**********************************)
(* Actions and their stable specs *)
(**********************************)

Lemma failM p : monoP treib_coh (fun x => pi.j x <> p).
Proof. by move=>???; rewrite !lE. Qed.

Program Definition trypush p q : {v h}, STsep [treib]
  (fun s => [/\ hp.s s = q :-> (v, p), ht.s s = Unit & [pcm h <= ht.t s]],
   fun r s => if r then exists t vs, [/\ hp.s s = Unit,
     ht.s s = t \-> (vs, v::vs), [pcm h <= ht.t s] & last_key h < t]
   else [/\ hp.s s = q :-> (v, p), ht.s s = Unit & [pcm h <= ht.t s]]) :=
  Do (actSx (fun b=>if b then push_t p q else id_t (@failM p))).
Next Obligation.
split=>[|[]] // ??? [[|_ X Y]|[_ Y X|]] //; case: X Y=>v [_][E???][];
by rewrite !lprodE E.
Qed.
Next Obligation.
move=>p q i [v][h][[/= Eh Et Ot]] _; apply: vrf_atm=>// j O.
move/(oht (oms _ O)): Ot=>Ot []; rewrite !(osf O) /= in Eh Et *.
rewrite /flat /= -!lprodX=>{i O} /validL Vh [_ _ _ /cm_valid Vt] Vf.
rewrite Eh in Vf; case: (pi.j j =P p)=>E; last first.
- exists false, j; rewrite effE; split=>// m O; rewrite -(osf O) -{1}(osf O).
  by move/oht: (oms (px tt) O)=>/(_ _ Ot).
eexists true, _. rewrite effE; split.
- exists v; apply: act2_safe=>//; first by exists Unit; rewrite -!lprodX unitR.
  by split=>//; rewrite -!lprodX // (validX Vf).
  by rewrite -!lprodX /flat /= !lE (lCJX Priv.lk) (lCJX Priv.sq) (validX Vf).
move=>m O _; move/oht: (oms (px tt) O); rewrite !lprodE !lE -!lprodX -joinA=>Otx.
rewrite -(osf O) -{1}(osf O) !lprodE !lE -!lprodX /= {2}Et unitR; eexists _, _.
by split=>//; [apply: pleq_trans Ot (Otx _ _) | apply: umpleq_lastkey Ot].
Qed.

Program Definition trypop (p q : ptr) : {v h}, STsep [treib]
  (fun s => [/\ hp.s s = Unit, ht.s s = Unit, [pcm h <= ht.t s] &
                [pcm p :-> (v, q) <= \stgb s]],
   fun r s => hp.s s = Unit /\ if r then exists t vs,
     [/\ ht.s s = t \-> (v::vs, vs), [pcm h <= ht.t s] & last_key h < t]
   else ht.s s = Unit /\ [pcm h <= ht.t s]) :=
  Do (actSx (fun b => if b then pop_t p q else id_t (@failM p))).
Next Obligation.
by split=>[|[]] // ??? [[|_ X Y]|[_ Y X|]] //; case: Y; case: X=>_ [?][?][?][E];
rewrite !lprodE E.
Qed.
Next Obligation.
move=>p q i [v][h][[/= Eh Et Ot Oh C]]; apply: vrf_atm=>// j O {C}.
move/(oht (oms _ O)): Ot=>Ot; move/(ostgb (oms _ O)): Oh=>Oh.
case=>_ [/validR Vsg L _][Vt _ _ _] Vf.
rewrite /flat /= -!lprodX in Vsg L Vt Vf.
move: (validL Vsg)=>Vst; rewrite !{i O}(osf O) /= in Eh Et *.
case: (pi.j j =P p) Oh=>[<- Oh|]; last first.
- exists false, j; rewrite effE; split=>// m O; rewrite -(osf O) -{1}(osf O) /=.
  by move/oht: (oms (px tt) O)=>/(_ _ Ot).
case/first_exists: L; first by move: (pleqV Oh Vsg); rewrite validPt.
move=>w [r][k][El Est] _; case: Oh=>x; rewrite Est -!joinA in Vsg *.
case/(hcancelV Vsg); case=><-<- _ _; eexists true, _; split.
- rewrite effE; apply: act2_safe=>//.
  - by exists k, w, (behead (al.j j)); rewrite -!lprodX.
  by rewrite /flat /= Est in Vf *; rewrite -!lprodX !lE (validX Vf).
move=>m O _; move/oht: (oms (px tt) O); rewrite !lprodE !lE -!lprodX -joinA=>Otx.
rewrite -(osf O) -{1}(osf O) /= !lprodE !lE -!lprodX /= {2}Et unitR.
split=>//; eexists _, _; split=>//;
by [apply: pleq_trans Ot (Otx _ _) | apply: umpleq_lastkey Vt Ot].
Qed.

Lemma readsentM p : monoP treib_coh (fun x => pi.j x = p).
Proof. by move=>???; rewrite !lE. Qed.

Program Definition readsent : {hs h}, STsep [treib]
  (fun s => [/\ hp.s s = hs, ht.s s = Unit & [pcm h <= ht.t s]],
   fun (r : ptr) s => [/\ hp.s s = hs, ht.s s = Unit, [pcm h <= ht.t s] &
     if r == null then exists t, last_key h <= t <= last_key (ht.t s) /\
                                 atval (it.j s) t (ht.t s) = [::]
     else exists (v : A) (q : ptr), [pcm r :-> (v, q) <= \stgb s]]) :=
  Do (actSx (fun p => id_t (@readsentM p))).
Next Obligation. by split=>// ????? _ [<- _][<-]. Qed.
Next Obligation.
move=>i [hs][h][[/= Eh Et Ot]] C; apply: vrf_atm=>// j O {}C.
move/(oht (oms _ O)): Ot=>Ot; rewrite !{i O}(osf O) /= in Eh Et *.
eexists _, _. rewrite effE.
split=>//= m O _; case: C (ocoh O C)=>_ [/validR/validL V].
move=>L _ M _ [_][_ _ _] /cm_valid Vtm _; rewrite !(osf O) /= in Eh Et L.
rewrite -!lprodX in V M Vtm L; split=>//; first by apply: oht (oms _ O) _.
case: ifP=>E; last first.
- case/(first_exists (negbT E)): L=>v [q][k][_ Est _].
  by exists v, q; apply: ostgb (oms _ O) _; rewrite Est -joinA.
rewrite (eqP E) in L; case/(llist_null V): L M=>-> _ [Vj G _ Ek].
exists (last_key (ht.t j)); rewrite -(oit (oms _ O)).
by rewrite !umpleq_lastkey // ?(umpleq_atval _ G) //; apply: (oht (oms _ O)).
Qed.

Definition rn_x p (r : A * ptr) x :=
  [pcm p :-> r <= \stgb x] /\ valid (\stgb x).
Lemma readnodeM p (r : A * ptr) : monoP treib_coh (rn_x p r).
Proof. by move=>??; rewrite /rn_x /= !lE. Qed.

Program Definition readnode p : {h (v : A) (q : ptr)}, STsep [treib]
  (fun s => [/\ hp.s s = Unit, ht.s s = Unit, [pcm h <= ht.t s] &
                [pcm p :-> (v, q) <= \stgb s]],
   fun r s => [/\ r = (v, q), hp.s s = Unit, ht.s s = Unit,
                  [pcm h <= ht.t s] & [pcm p :-> (v, q) <= \stgb s]]) :=
  Do (actSx (fun r => id_t (@readnodeM p r))).
Next Obligation.
by split=>// ????? _ [H _][[[x]]]; case: H=>[[h -> V]] /(hcancelV V) [].
Qed.
Next Obligation.
move=>p i [h][v][q][[/= Eh Et Ot Oh]] C; apply: vrf_atm=>// j O {C}.
move/(oht (oms _ O)): Ot=>Ot; move/(ostgb (oms _ O)): Oh=>Oh.
rewrite !(osf O) in Eh Et; case=>{i O} _ [/validR Vst _ _ _ _].
rewrite -!lprodX in Vst; eexists (v, q), _. rewrite effE; split=>// m O _.
by rewrite !(osf O) in Eh Et; move/(oht (oms _ O)): Ot; move/(ostgb (oms _ O)): Oh.
Qed.

(****************)
(* Main methods *)
(****************)

Definition pop_tp := {h}, STsep [treib]
  (fun s => [/\ hp.s s = Unit, ht.s s = Unit & [pcm h <= ht.t s]],
   fun r s => hp.s s = Unit /\ exists t, if r is Some v then exists vs,
     [/\ ht.s s = t \-> (v::vs, vs), [pcm h <= ht.t s] & last_key h < t]
   else [/\ ht.s s = Unit, [pcm h <= ht.t s], atval (it.j s) t (ht.t s) = [::] &
            last_key h <= t <= last_key (ht.t s)]).

Program Definition pop :=
  ffix (fun (loop : unit -> pop_tp) xx =>
    Do (p <-- readsent;
        if p == null then ret None
        else (
          vq <-- readnode p;
          ok <-- trypop p vq.2;
          if ok then ret (Some vq.1)
          else loop tt))) tt.
Next Obligation.
move=>loop [i][h][/= X] C; apply: [stepE Unit, h]=>// {i X C}.
move=>p i [Eh Et Ot S] C; case: ifP S=>Ep; last first.
- case=>v [q Hp]; apply: [stepE h, v, q]=>//= {i Eh Et Ot Hp C}.
  move=>_ i [->] Eh Et Ot Hp C; apply: [stepE v, h]=>//= {i Eh Et Ot Hp C}.
  case=>[i [Eh][t][vs][Et Ot K]|i [Eh][Et Ot]] C; last by apply: [gE h].
  apply: vrf_ret=>// m O; rewrite !(osf O) /= in Eh Et; split=>//.
  by exists t, vs; split=>//; apply: pleq_trans Ot (oht (oms _ O) _).
case=>t [/andP [N1 N2] A]; apply: vrf_ret=>// m O Cm.
move/(oht (oms _ O)): Ot=>Ot; rewrite !(osf O) /= in Eh Et.
move: (oht (oms _ O) pleq_refl)=>Om; case: C=>_ [_ _ _ [_] G _ _ _].
have Vm : valid (ht.t m) by case/cohso/andP: Cm; rewrite /= !(lCSX snd).
split=>//; exists t; rewrite N1; split=>//; last first.
- by apply: leq_trans N2  (umpleq_lastkey Vm _).
by rewrite -A (oit (oms _ O)); apply: umpleq_atval=>//; rewrite lprodE.
Qed.

Definition push_post v h :=
  fun (_ : unit) s => exists t vs, [/\ hp.s s = Unit,
    ht.s s = t \-> (vs, v::vs), [pcm h <= ht.t s] & last_key h < t].

Definition push_tp q v := {h}, STsep [treib]
  (fun s => exists p : ptr, [/\ hp.s s = q :-> (v, p),
    ht.s s = Unit & [pcm h <= ht.t s]], push_post v h).

(* The implementation of this fixed point constructor takes *)

Program Definition push v : {h}, STsep [treib]
  (fun s => [/\ hp.s s = Unit, ht.s s = Unit & [pcm h <= ht.t s]],
   push_post v h) :=
  Do (q <-- morph_app px tt (Priv.alloc (v, null));
      ffix (fun (loop : unit -> push_tp q v) xx =>
        Do (p <-- readsent;
            morph_app px tt (Priv.write q (v, p));;
            ok <-- trypush p q;
            if ok then ret tt else loop tt)) tt).
Next Obligation.
move=>v q loop [i][h][[/= p']] X _; apply: [stepE q :-> (v, p'), h]=>//= {i X}.
move=>p i [Eh Et Ot] _ _; apply/step/(vrf_morph px_sim)=>// _.
apply: [gE Unit]=>//=; first by exists (v, p'); rewrite -lprodX unitR.
case=>{Eh} j /= Eh _ [m E _] _ /mxms O; move/(oht O): Ot=>Ot.
rewrite (ohts O) in Et; rewrite {i j O}E unitR /= -lprodX in Eh.
apply: [stepE v, h]=>// {m Et Eh Ot}; case=>[i [t][vs][Eh Et Ot K]|i X] _;
  last by apply: [gE h]=>//; exists p.
apply: vrf_ret=>// m O; rewrite !(osf O) /= in Eh Et.
by exists t, vs; split=>//; apply: oht (oms _ O) Ot.
Qed.
Next Obligation.
move=>v i [h][[/= Eh Et Ot]] _; apply/step/(vrf_morph px_sim)=>// _.
apply: [gE Unit]=>//=; first by rewrite -lprodX.
move=>{Eh} q j Eh _ [m] E ?? /mxms O; rewrite (ohts O) in Et.
move/(oht O): Ot=>Ot; rewrite {j O}E unitR in Eh.
by apply: [gE h]=>//; exists null; rewrite lprodE.
Qed.

Lemma idE : idfun = [** hp & ht].
Proof. by apply: fext=>x; apply: prod_eta. Qed.

End TreiberPriv.

Module Exports.
Canonical hp_morph.
Canonical ht_morph.
Arguments hp /.
Arguments ht /.
Arguments pi !_/.
Arguments al !_/.
Arguments sk !_/.
Arguments gb !_/.
Arguments it !_/.
End Exports.
End TreiberP.

Export TreiberP.Exports.


(**************************************************************************)
(* General pattern of the action spec proofs is as follows.               *)
(*                                                                        *)
(* It consists of two parts: (1) the proof of safety (i.e., that the      *)
(* given precondition implies the safety of the action, and (2) of        *)
(* correctness, (i.e., the poscondition of the action implies the given   *)
(* postcondition).                                                        *)
(*                                                                        *)
(* We always prove safety in the forward manner, i.e., by modifying the   *)
(* hypotheses from the context, typically using the oself and oself2      *)
(* lemmas over a given osteps hypothesis.                                 *)
(*                                                                        *)
(* We (almost) always prove correctness in the backward manner, i.e., by  *)
(* first modifying the goal by using oself and oself2, and *only* then    *)
(* unfolding the action definition. It is (almost) impossible here to     *)
(* work in the other order, because unfolding the action will put the     *)
(* osteps hypothesis in an unworkable format.                             *)
(*                                                                        *)
(* The one except to this (the almost part), is if the action does not    *)
(* change the heap, in which case, unfolding the action does not change   *)
(* the osteps hypothesis.                                                 *)
(**************************************************************************)

(*************************)
(* Hiding and quiescence *)
(*************************)

(* morphism from treib to priv which erases histories *)
(* and a simple client example *)

Section Quiescence.
Variable snt : ptr.
Definition R := TreiberT nat snt.
Notation treib := (TreiberP.treib R).

Local Definition hist := [pcm of natmap (seq nat * seq nat)].

Definition hp := (@TreiberP.hp R).
Canonical hp_morph := Eval hnf in [morphism of hp].
Definition ht := (@TreiberP.ht R).
Canonical ht_morph := Eval hnf in [morphism of ht].
Definition pi := (@TreiberP.pi R).
Definition al := (@TreiberP.al R).
Definition sk := (@TreiberP.sk R).
Definition gb := (@TreiberP.gb R).
Definition it := (@TreiberP.it R).
Notation "\stgb s" := (sk.j s \+ gb.j s) (at level 24).
Arguments hp /.
Arguments ht /.
Arguments pi !_/.
Arguments al !_/.
Arguments sk !_/.
Arguments gb !_/.
Arguments it !_/.

(* definition of the state component of the morphism *)
Definition tpriv_phi x (p : st priv) (t : st treib) :=
  [/\ p = State (hp.s t \+ snt :-> pi.j t \+ \stgb t) (fst.j t) (hp.o t),
      [/\ ht.s t = x.1, gb.j t = x.2, it.j t = [::] & ht.o t = Unit] &
      coh treib t].

(* the next two definitions are needed to morph id_tr's *)
Definition tp P := fun p => exists x t, [/\ tpriv_phi x p t & P t].
Definition fr gb P := fun (p : st priv) =>
  exists q (xs : seq nat) h1 h2, [/\ idfun.s p = h1 \+ snt :-> q \+ h2 \+ gb,
    llist q xs h2 & P (State h1 (idfun.j p) (idfun.o p))].

Lemma tp_mono P : monoP (coh _) P -> monoP (coh _) (tp P).
Proof.
move=>pf px p _; rewrite /tp /tpriv_phi; case=>x [t][H X].
rewrite !lE in H; case: H; case=>E1 E2 E3 H C.
set s := State (hp.s t \+ px, ht.s t) (idfun.j t) (idfun.o p, ht.o t).
have C2 : coh treib s.
- case: C; rewrite /= /prod_raw /flat /= -!lprodX !lE /= =>_ D.
  by rewrite -E3 => V; split=>//; rewrite (validX V).
have S1 : s = State (idfun.s t) (idfun.j t) (idfun.o p, ht.o t) \< (px, Unit).
- by rewrite /s zg_unlock (TreiberP.idE R) lSE' pcmPE unitR.
have S2 : t = State (idfun.s t) (idfun.j t) (idfun.o p, ht.o t) \> (px, Unit).
- rewrite /s zg_unlock /= pcmPE unitL E3.
  by rewrite {1}[t]stEX {2}(TreiberP.idE R) lSE'.
have D2 : P s by rewrite S1; apply: pf; rewrite -?S1 -?S2.
by exists x, s; rewrite !lE /= /Priv.hp /= E1 E2 !(pull px) !joinA.
Qed.

Lemma fr_mono gb P : monoP (coh priv) P -> monoP (coh priv) (fr gb P).
Proof.
move=>pf px p C [q][xs][h1][h2][]; move: (coh_valid C)=>A.
rewrite !lE //= => H1 H2 H3; rewrite /= !lE //= H1 in C.
exists q, xs, (h1 \+ px), h2; rewrite !lE //= H1 !(pull px) !joinA; split=>//.
by rewrite joinC zgEX; apply: pf; rewrite -zgEX //= !lE /= (validX C).
Qed.

(* definition of transition component of the morphism *)
Definition tpriv_tau' (x : hist * heap) (tr : Tr treib) : Tr priv :=
  match tr with
    TreiberP.push_t p q => Priv.rmw_t snt (fun v => v = p) (fun => q)
  (* this is CAS(snt,p,q) success *)
  | TreiberP.pop_t p q => Priv.rmw_t snt (fun v => v = p) (fun => q)
  | TreiberP.id_t P pf => Priv.id_t (tp_mono pf)
  | TreiberP.in_priv tr =>
      if tr is Priv.id_t P pf then Priv.id_t (fr_mono (gb:=x.2) pf) else tr
  end.

Definition tpriv_tau x t (s : st priv) t' := t' = tpriv_tau' x t.

(* morphism properties *)
Lemma tpriv_local : morph_localP tpriv_phi.
Proof.
exists (fun p => (p, Unit))=>x px p t Cp _ H; move: (coh_valid Cp)=>A /=.
case: H Cp; rewrite !lE //; case=>E1 E2 O H C Cp.
exists (State (idfun.s t) (idfun.j t) (idfun.o p, ht.o t)).
rewrite {1}[t]stEX -!zgEX pcmE /= O unitL prod_feta !lE; split=>//.
rewrite /tpriv_phi !lE /= E1 E2; split.
- by rewrite /Priv.hp /= !(pull px) !joinA.
- by rewrite unitR.
rewrite /= !lE //= E1 in Cp; case: C=>_ /=; rewrite /= /flat /= -!lprodX=>C V.
rewrite -O -!(lCJX _ fst) -E2 !lJA in C V.
rewrite /prod_raw /flat /= !prodEX !lE /=!lJA unitR !(lCJX _ snd) -E2 !lJA.
by rewrite (validX Cp); split=>//; rewrite (validX V).
Qed.

Lemma tpriv_other : morph_otherP tpriv_phi.
Proof.
move=>??????????; rewrite !prodX !lT -!lprodX.
by case=>-> [???->] _ [->][???->] _; rewrite !lE=>/= ->.
Qed.

Lemma tpriv_func : morph_funcP tpriv_phi.
Proof.
move=>x p t1 t2 /cohso/validL V [_ [_ L1 _ _ _]][_ [_ L2 _ _ _]] H1 H2.
apply/stE; rewrite -!lT !prodX !lT -!lprodX.
case: H1 H2 V L1 L2=>-> [-> G1 I1 -> _][[E <-<-][-> G2 I2 -> _]] /= V.
rewrite !lE -!lprodX !lA /= in V *.
rewrite -!joinA G1 G2 !(pull x.2) !(pull (snt :-> _)) !(joinC (hp.s _)) in E V.
case/(hcancelV V): E=>Ep {}V /(joinxK V) E L1 L2; rewrite -Ep in L2.
case/(llist_prec (validR V) L1 L2): E=>{L1 L2} Ea Es ->; split=>//; congr (_, _).
rewrite [snd.j t1]Treiber.etaL [snd.j t2]Treiber.etaL !(lCJX _ snd).
by rewrite G1 I1 G2 I2 Ep Ea Es.
Qed.

Lemma tpriv_backw : morph_backwardP tpriv_phi.
Proof. by case=>??[??]??? _ _ _ [_ [/= <-<- _ _ _]][_ [/= <-<-]]. Qed.

Lemma tpriv_int : morph_internalP tpriv_tau.
Proof. by move=>t _ x _ I _ ->; case: t I=>//; case. Qed.

Lemma tpriv_tfunc : morph_tfuncP tpriv_tau.
Proof. by move=>x t _ _ _ _ ->->. Qed.

Lemma tpriv_tlocal : morph_tlocalP tpriv_tau.
Proof. by []. Qed.

Lemma tpriv_coh : morph_cohP' tpriv_phi.
Proof. by move=>??? _ []. Qed.

Definition tprivMorphMixin :=
  MorphMixin tpriv_local tpriv_other tpriv_func tpriv_backw
             tpriv_int tpriv_tfunc tpriv_tlocal tpriv_coh.
Canonical tpriv := Morph tprivMorphMixin.

Lemma tpriv_ostep : morph_ostepP tpriv (fun => xPredT).
Proof.
move=>x xx p2 t1 Cp _ [E H Cs] _ O; subst xx.
set t2 := State (idfun.s t1) (idfun.j p2, snd.j t1) (idfun.o p2, ht.o t1); exists t2.
have Ct : coh treib t2.
- move: (ocoh (iter1 O) Cp); rewrite /= /prod_raw /= /t2 /flat /=.
  rewrite -!lprodX !lE !lA /= !lJA !(lCJX _ snd) -!(osf (iter1 O)) !lE => V.
  by rewrite !(validX V); case: Cs=>_ []; rewrite -!lprodX.
have T : (tr t2)\2 = (tr t1)\2.
- by rewrite !prodE [t1]stEX /t2 !lE !prodEX !lA /= lJA.
have V : valid (flat priv_coh (tr t2)\1 \+ flat (xtreib_coh R) (tr t2)\2).
- move: (ocoh (iter1 O) Cp); rewrite /= /flat /t2 /= -!lprodX !lE /=.
  by rewrite !lA !lJA !(lCJX _ snd) -(osf (iter1 O)) !lE=>/validX ->.
split=>//=.
- rewrite /Morph.phi /= /tpriv_phi /t2 !lE [p2]stEX !lE -(osf (iter1 O)) !lE.
  rewrite {2}/sk/TreiberP.sk {2 3}/gb/TreiberP.gb/it/TreiberP.it.
  by rewrite {2}/pi/TreiberP.pi /=; split=>//; rewrite !(lCJX _ snd).
case: (O)=>a; move: {1 3}a (erefl a)=>[A p P f|A p f|||h|h|P pf] E /=;
rewrite !lE /= => X;
  last by [rewrite /t2 /= {1}[t1]stEX (proj2 X) !lE 2!prod_feta !lE];
apply: (@iosteps treib _ _ (TreiberP.in_priv _ a)); split=>//;
rewrite !prodE -{}E /t2 /= !prodEX !lE ?trEX -!lprodX !lA /=.
- by case: X=>v [h][? W ?] /trS ->; exists v, h; rewrite !lE (validX W).
- by case: X=>B [w][h][? W] /trS ->; exists B, w, h; rewrite !lE (validX W).
- by case: X=>? ->; rewrite !lE -!(lCJX _ fst).
- case: X=>xs [h][/= ? W] /trS ->; exists xs, h.
  by rewrite !lE -!(lCJX _ fst) (validX W).
- by case: X=>W ->; rewrite !lE -!(lCJX _ fst) (validX W).
by case: X=>h1 [/= ? W] /trS ->; exists h1; rewrite !lE (validX W).
Qed.

Lemma tpriv_forw : morph_forwardP tpriv (fun => xPredT).
Proof.
move=>t x _ t1 t2 I _ Ct1 [-> [? X1 X2 X3 _]] _ H; move: (icoh H Ct1)=>Ct2.
exists (tpriv_tau' x t); case: t I H=>[p q|p q|P pf|t] // I.
- case=>z [[/= h1]][H1 _ H2][Hp _ H3]; rewrite /flat H2 H3 X1 /= !lE /= => V.
  eexists (_ \-> (al.j t1, _) \+ _, _), _; rewrite -!lprodX in H1 Hp V.
  rewrite /Morph.phi/=/tpriv_phi/tpriv_tau /= !lprodE {}H2 {}H3 -!lprodX !lE /=.
  rewrite !joinA X1 X3; split=>//; exists p, (hp.s t1 \+ sk.j t1 \+ x.2).
  rewrite H1 -!joinA !(pull (snt :-> _)) !(pull (q :-> _)).
  by rewrite Hp X1 in V *; split=>//; rewrite (validX V).
- case=>/= H1 [h][v][vs][Hp H2 _ _ H3]; rewrite /flat H1 H3 X1 /= !lE /= => V.
  eexists (_ \+ ht.s t1, _), _; rewrite -!lprodX in Hp H2 V.
  rewrite /Morph.phi/=/tpriv_phi !lprodE {}H1 {}H3 -!lprodX !lE /=.
  rewrite !joinA X1 X3; split=>//; exists p, (hp.s t1 \+ sk.j t1 \+ x.2).
  rewrite H2 -!joinA !(pull (snt :-> _)) !(pull h).
  by rewrite Hp X1 in V *; split=>//; rewrite (validX V).
- by case=>H ->; eexists x, _; do !split=>//; exists x, t1.
case=>/= H1 H2; rewrite /flat H2 X1 /= => V; eexists x, _.
rewrite /Morph.phi/=/tpriv_phi !lprodE H2; move: (trans_other H1); rewrite !lT -!lprodX=>->.
split=>//; rewrite [t1\1]stEX [t2\1]stEX /= X1 !lE -!lprodX !lA in V H1 *.
case: t I H1 V=>[A z P f|A z v|||||P pf] // _; last 1 first.
- case=>H [->->->]; split=>//; eexists _, (al.j t1), _, (sk.j t1).
  by rewrite !lE /= !joinA; case: Ct1=>_ [_]; rewrite -!lprodX.
- case=>w [h][/=]; rewrite !lE /= =>-> _ H [->->->]; rewrite X1 => V.
  by eexists w, _; rewrite -!joinA (validX V).
- case=>B [w][h][/=]; rewrite !lE /= => -> _ [->->->]; rewrite X1 => V.
  by eexists B, w, _; rewrite -!joinA (validX V).
- by rewrite /= !lE; case=>-> [->->->]; rewrite -!joinA.
case=>xs [h][/=]; rewrite !lE /= -!(lCJX _ fst)=>-> W [->->->].
rewrite X1=>V; eexists xs, _; rewrite -!joinA; split=>//.
move: W; rewrite /Priv.layout; case: (fst.j t1); case=>s;
by rewrite ?(validX V) // invalidX.
Qed.

Definition tpriv_sim := Simulation2 tpriv_ostep tpriv_forw.

Program Definition push_top :  STsep [priv]
  (fun s : st priv => exists q h, llist (A:=nat) q [::] h /\ idfun.s s = snt :-> q \+ h,
   fun _ s => exists q h g, self s = snt :-> q \+ h \+ g /\
                            (llist q [:: 1; 2] h \/ llist q [:: 2; 1] h)) :=
  Do (morph_app tpriv Unit (par (@TreiberP.push R 1) (@TreiberP.push R 2))).
Next Obligation.
move=>i [[_][_][[->->] E]] C; set t : st treib := State Unit (idfun.j i,
  Treiber.layout (R:=R) null Unit Unit [::] [::]) (idfun.o i, Unit : hist).
apply: (vrf_morph tpriv_sim t)=>[|//|Ct].
- rewrite /Morph.phi /= /tpriv_phi /= /prod_raw /= !prodEX /flat;
  rewrite !lE /= !lE !unitL -E -stEM !lJA !(validX C).
  by do 3!split=>//; [apply: umall0 | apply: cm0].
pose Q x := @TreiberP.push_post R x Unit.
have Et : idfun.s t = Unit by rewrite !lE.
apply: (par_do3 _ Unit Unit (Q 1) (Q 2))=>//;
try by move=>?; apply: [gE Unit]=>//=; rewrite !lE.
- by rewrite unitL.
move=>v ?? m [_ [_ L _ Cm _]] Sm [t1][vs1][_ H1 _ _][t2][vs2][_ H2 _ _].
rewrite !lE /= in H1 H2; case=>y g p [-> /= [_ _ H3 H4]] _ _ _ _.
rewrite lA in H4; rewrite prod_feta !lE in Sm. case: Sm=>Sm1 Sm2.
rewrite -!lprodX -!(lCSX _ snd) Sm2 H4 unitR {}H1 {}H2 {}H3 in Cm.
exists (pi.j m), (sk.j m), (hp.s m \+ gb.j m); split.
- by rewrite -!joinA !(pull (snt :-> _)) (joinCA (sk.j m)).
have [Nt1 Nt2] : t1 != 0 /\ t2 != 0.
- by case: Cm; rewrite validPtUn validPt; case/and3P.
case: (ltngtP t1 t2) Cm=>N Cm; [right; rewrite joinC in Cm|left|];
try by [case: (cmCons Cm _) L; rewrite ?lastkeyPt // -!lprodX=>_ ->/cmPt[_][-><-]].
by move: N Cm=>-> []; rewrite invalidX.
Qed.

End Quiescence.
