From Coq Require Import ssreflect ssrbool ssrfun.
From mathcomp Require Import ssrnat seq.
From fcsl Require Import axioms pred prelude pcm state world process.
From fcsl Require Import options.

Section Always.
Variable W : resource.
Notation U := (Ux W).
Notation T := (Tp W).
Notation coh := (coh W).
Notation tree A := (tree W A).

(* Definition of the operator for adherence to a prototol *)
Fixpoint always_path A s t scs (Q : state U T -> tree A -> Prop) :=
  if scs is sc :: scs' then
    forall s2, osteps W s s2 ->
      [/\ tsafe t sc s2, Q s2 t &
          forall s3 q, tstep t sc q s2 s3 -> always_path s3 q scs' Q]
  else forall s2, osteps W s s2 -> Q s2 t.

(* unfolding of the fixed point in always_path *)
(* to pull out the common parts of the branches *)

Lemma alwpE A s (t : tree A) scs Q :
        always_path s t scs Q <->
        forall s2, osteps W s s2 -> Q s2 t /\
          if scs is sc :: scs' then
             tsafe t sc s2 /\
             forall s3 q, tstep t sc q s2 s3 -> always_path s3 q scs' Q
          else True.
Proof. by split; case: scs=>[|sc scs] /= H s2 /H // [H'] // []. Qed.

Definition always A s (p : tree A) Q := forall scs, always_path s p scs Q.

(* some structural properties of always_path and always *)

Lemma alwp_safe A s (t : tree A) sc scs Q :
        always_path s t (sc :: scs) Q -> tsafe t sc s.
Proof. by case/alwpE/(_ _ iter_refl)=>? []. Qed.

Lemma alw_safe A s (t : tree A) Q sc :
        always s t Q -> tsafe t sc s.
Proof. by move=>H; apply: alwp_safe (H [:: sc]). Qed.

Lemma alwp_refl A s (t : tree A) sc Q : always_path s t sc Q -> Q s t.
Proof. by case/alwpE/(_ _ iter_refl). Qed.

Lemma alw_refl A s (t : tree A) Q : always s t Q -> Q s t.
Proof. by move/(_ [::])/alwp_refl. Qed.

Lemma alwp_envs A s1 (t : tree A) scs s2 Q :
        always_path s1 t scs Q -> osteps W s1 s2 -> always_path s2 t scs Q.
Proof. by move/alwpE=>T E; apply/alwpE=>s3 /(iter_trans E)/T. Qed.

Lemma alw_envs A s1 (t : tree A) s2 Q :
        always s1 t Q -> osteps W s1 s2 -> always s2 t Q.
Proof. by move=>S E scs; apply: alwp_envs (S scs) E. Qed.

Lemma alwp_envsq A s1 (t : tree A) scs (Q : _ -> _ -> Prop) :
        always_path s1 t scs Q ->
        always_path s1 t scs (fun s2 p2 => forall s3, osteps W s2 s3 -> Q s3 p2).
Proof.
elim: scs s1 t=>[|sc scs IH] /= s1 t H s2 M.
- by move=>s3 /(iter_trans M); apply: H.
split; first by case: (H _ M).
- by move=>s3 /(iter_trans M) /H; case.
by move=>s3 q T; apply: IH; case: (H _ M)=>_ _; apply.
Qed.

Lemma alw_envsq A s1 (t : tree A) (Q : _ -> _ -> Prop) :
        always s1 t Q ->
        always s1 t (fun s2 p2 => forall s3, osteps W s2 s3 -> Q s3 p2).
Proof. by move=>H scs; apply: alwp_envsq (H scs). Qed.

Lemma alw_tstep A s1 (t : tree A) sc s2 q Q :
        always s1 t Q -> tstep t sc q s1 s2 -> always s2 q Q.
Proof. by move=>H T scs; case:(H (sc::scs) _ iter_refl)=>_ _; apply. Qed.

Lemma alwp_unfin A s1 scs (Q : state U T -> tree A -> Prop) :
        (forall s2, osteps W s1 s2 -> Q s2 Unfinished) ->
        always_path s1 Unfinished scs Q.
Proof.
case: scs=>[|sc scs] H // s2 E; split;
by [move=>G; case: sc G | apply: H E | move=>s3 q /tstUnfin].
Qed.

Lemma alw_unfin A s1 (Q : state U T -> tree A -> Prop) :
        (forall s2, osteps W s1 s2 -> Q s2 Unfinished) ->
        always s1 Unfinished Q.
Proof. by move=>H scs; apply: alwp_unfin. Qed.

Lemma alwp_ret A s1 (v : A) scs (Q : state U T -> tree A -> Prop) :
        (forall s2, osteps W s1 s2 -> Q s2 (Ret v)) ->
        always_path s1 (Ret v) scs Q.
Proof.
case: scs=>[|sc scs] H // s2 E; split;
by [move=>G; case: sc G | apply: H E | move=>s3 q /tstUnfin].
Qed.

Lemma alw_ret A s1 (v : A) (Q : state U T -> tree A -> Prop) :
        (forall s2, osteps W s1 s2 -> Q s2 (Ret v)) ->
        always s1 (Ret v) Q.
Proof. by move=>H ps; apply: alwp_ret. Qed.

(* We don't compare to W here. We will use alw_ext for that *)
Lemma alw_act A s1 (a : action W A) Q :
        (forall s2, osteps W s1 s2 ->
          [/\ act_safe a s2, Q s2 (Act a) &
              forall s3 v s4, eff a v s2 s3 ->
                              osteps W s3 s4 -> Q s4 (Ret v)]) ->
        always s1 (Act a) Q.
Proof.
move=>H [|sc scs] s2 /H [//] H1 H2 H3; split=>//; last first.
- move=>s3 q /tstAct [w][_ -> S]; apply: alwp_ret=>s4; apply: H3.
  by case: S=>sw [sw'][tv][tw][Ha][->->]; exists tv.
move=>G; case: sc G=>// _; case: H1=>s3 [v Ha].
exists s3, (Ret v), v; split=>//.
by case: Ha=>tv [Ha Hb]; exists s2, s3, tv, tv.
Qed.

(* always is a natural operator *)
(* this is essentially our rule for strengthening posts *)

Lemma alwp_imp A s (t : tree A) scs (Q1 Q2 : state U T -> tree A -> Prop) :
         (forall s p, Q1 s p -> Q2 s p) ->
         always_path s t scs Q1 -> always_path s t scs Q2.
Proof.
elim: scs s t=>[|sc scs IH] s t /= I L s2 E; first by apply: I (L _ E).
case/L: (E)=>S H T; split=>//; first by apply: I H.
by move=>s3 q; move/T; apply: IH I.
Qed.

Lemma alw_imp A s (t : tree A) (Q1 Q2 : state U T -> tree A -> Prop) :
        (forall s t, Q1 s t -> Q2 s t) ->
        always s t Q1 -> always s t Q2.
Proof. by move=>I H scs; apply: alwp_imp I _. Qed.

(* rule of conjunction *)

Lemma alwp_and A s (t : tree A) scs (Q1 Q2 : state U T -> tree A -> Prop) :
        always_path s t scs Q1 -> always_path s t scs Q2 ->
        always_path s t scs (fun s t => Q1 s t /\ Q2 s t).
Proof.
elim: scs s t=>[|sc scs IH] s t /= H1 H2 s2 E.
- by split; [apply: H1 | apply: H2].
case: (H1 _ E)=>S {}H1 Ls1; case: (H2 _ E)=>_{}H2 Ls2.
by split=>// s3 q T; apply: IH; [apply: Ls1 | apply: Ls2].
Qed.

Lemma alw_and A s (t : tree A) (Q1 Q2 : state U T -> tree A -> Prop) :
        always s t Q1 -> always s t Q2 ->
        always s t (fun s t => Q1 s t /\ Q2 s t).
Proof. by move=>Ls1 Ls2 scs; apply: alwp_and; [apply: Ls1 | apply: Ls2]. Qed.

(* always commutes with universals; or, infinitary conjunction rule *)
(* we could also have a version where we require that A is inhabitted *)
(* but the version below is more useful in separation logic, as it lets *)
(* is commute logical variables, when we know that the program is safe *)
(* (which we always do, irrespective of A being inhabitted) *)

Lemma alwpA A B s (t : tree A) scs (Q : B -> state U T -> tree A -> Prop) :
        always_path s t scs (fun _ _ => True) ->
        (always_path s t scs (fun s' t' => forall x, Q x s' t') <->
         forall x, always_path s t scs (fun s' t' => Q x s' t')).
Proof.
move=>Ls; split=>[{Ls}|].
- elim: scs s t=>[|sc scs IH] s t /= Et x s2 /Et //.
  by case=>S Ha L; split=>// s3 q; move/L/IH.
elim: scs s t Ls=>[|sc scs IH] s t /= Et Ha s2 E.
- by move=>x; apply: Ha.
case/Et: (E)=>/= S _ L; split=>//.
- by move=>x; case/(Ha x): E.
move=>s3 q T; apply: IH; first by apply: L T.
by move=>x; case/(Ha x): E=>_ _; apply.
Qed.

Lemma alwA A B s (t : tree A) (Q : B -> state U T -> tree A -> Prop) :
        always s t (fun _ _ => True) ->
        (always s t (fun s' t' => forall x, Q x s' t') <->
         forall x, always s t (fun s' t' => Q x s' t')).
Proof.
move=>Ls; split.
- by move=>H x scs; move: x; apply/(alwpA _ (Ls scs)).
by move=>H scs; apply/(alwpA _ (Ls scs))=>x; apply: H.
Qed.

Arguments alwA [A B s t Q] _.

(* always commutes with implication, which is easy consequence *)
(* of commuting with universals *)

Lemma alwI A s (t : tree A) (P : Prop) (Q : state U T -> tree A -> Prop) :
        always s t (fun _ _ => True) ->
        always s t (fun s' t' => P -> Q s' t') <->
        (P -> always s t (fun s' t' => Q s' t')).
Proof. by apply: alwA. Qed.

Arguments alwI [A s t P Q] _.

Lemma alwp_msteps A s (t : tree A) scs :
        coh s ->
        always_path s t scs (fun s _ => True) ->
        always_path s t scs (fun s' t' => msteps (id_morph W tt) s s').
Proof.
elim: scs s t=>[|sc scs IH] s t /= C H1 s2 E; first by apply: oms.
case/H1: (E)=>S _ {}H1; split=>//{S}; first by apply: oms.
move: (ocoh E C)=>C2=>s3 q T; move: (tst_coh T C2)=>C3.
apply: alwp_imp (IH _ _ C3 (H1 _ _ T))=>s' _ M.
by apply: iter_trans M; apply: iter_trans (tst_msteps C2 T); apply: oms E.
Qed.

Lemma alw_msteps A s (t : tree A) :
        coh s ->
        always s t (fun _ _ => True) ->
        always s t (fun s' t' => msteps (id_morph W tt) s s').
Proof. by move=>C Ls scs; apply: alwp_msteps. Qed.

(*******************)
(* After predicate *)
(*******************)

(* specialization of always to the termination case *)
(* we also demand that states preserve the invariant throughout *)
(* the execution, as one would require in CSL *)

Definition after A s (t : tree A) (Q : A -> state U T -> Prop) :=
  always s t (fun s2 t2 => coh s2 /\ forall v, t2 = Ret v -> Q v s2).

Lemma aft_coh A s (t : tree A) (Q : A -> state U T -> Prop) :
       after s t Q -> coh s.
Proof. by case/alw_refl. Qed.

Lemma aft_and A s (t : tree A) (Q1 Q2 : A -> state U T -> Prop) :
        after s t Q1 -> after s t Q2 -> after s t (fun v s => Q1 v s /\ Q2 v s).
Proof.
move=>H1 H2; apply: alw_imp (alw_and H1 H2)=>x q [[C1 X1]][C2 X2].
by split=>// v E; split; [apply: X1 | apply: X2].
Qed.

(* the following is our rule of consequence; we add coh into it *)
(* because resource invariant should be usable whenever we need it *)
Lemma aft_imp A s (t : tree A) (Q1 Q2 : A -> state U T -> Prop) :
        (forall v s, coh s -> Q1 v s -> Q2 v s) ->
        after s t Q1 -> after s t Q2.
Proof. by move=>I; apply: alw_imp=>s1 t1 [C H]; split=>// v /H; apply: I. Qed.

Lemma aftA A B s (t : tree A) (Q : B -> A -> state U T -> Prop) :
        after s t (fun _ _ => True) ->
        (after s t (fun v s' => forall x, Q x v s') <->
         forall x, after s t (fun v s' => Q x v s')).
Proof.
move=>Ls; rewrite -alwA; last by apply: alw_imp Ls.
split; first by apply: alw_imp=>s' q [C I] x; split=>// v /I.
move/(alw_and Ls); apply: alw_imp=>s' q [[C _]] I.
by split=>// v E x; apply: (proj2 (I x)).
Qed.

Arguments aftA [A B s t Q] _.

Lemma aftI A s (t : tree A) (P : Prop) (Q : A -> state U T -> Prop) :
        after s t (fun _ _ => True) ->
        after s t (fun v s' => P -> Q v s') <->
        (P -> after s t (fun v s' => Q v s')).
Proof.
move=>Ls; rewrite -alwI; last by apply: alw_imp Ls.
split; first by apply: alw_imp=>s' q [C I] v; split=>// x /I; apply.
move/(alw_and Ls); apply: alw_imp=>s' q [[C _]] I.
by split=>// v E /I [_]; apply.
Qed.

Arguments aftI [A s t P Q] _.

Lemma aft_msteps A s (t : tree A) :
        after s t (fun _ _ => True) ->
        after s t (fun _ s' => msteps (id_morph W tt) s s').
Proof.
move=>H1; move: (aft_coh H1)=>C.
have/(alw_msteps C) H2 : always s t (fun _ => xPredT) by apply: alw_imp.
by apply: {H1 H2 t} alw_imp (alw_and H1 H2)=>s' t [[]].
Qed.

(* we can combine aft_inv and aft_msteps *)
Lemma aft_invst A s (t : tree A) (Q1 Q2 : A -> state U T -> Prop) :
        (forall v s', coh s -> msteps (id_morph W tt) s s' ->
                      Q1 v s' -> Q2 v s') ->
        after s t Q1 -> after s t Q2.
Proof.
move=>R H.
have {}H : after s t (fun v s' => Q1 v s' /\ msteps (id_morph W tt) s s').
- by apply: aft_and=>//; apply: aft_msteps; apply: aft_imp H.
by apply: aft_imp (H)=>v x C [X M]; apply: R=>//; apply: aft_coh H.
Qed.

Lemma aft_unfin A s (t : tree A) Q :
        after s t Q -> after s Unfinished Q.
Proof.
case/alw_refl=>C _; apply/alw_unfin=>s2 O.
by split=>//; apply: ocoh O C.
Qed.

(************************************)
(* after and sequential composition *)
(************************************)

Lemma aft_bnd A B (t1 : tree A) (t12 : tree B) tt2 s1 Q :
        t12 \In tcat t1 tt2 ->
        after s1 t1 (fun v s =>
          forall p, p \In tt2 v -> after s p Q) ->
        after s1 t12 Q.
Proof.
move=>Tc Ls scs.
elim: scs s1 t1 t12 Tc Ls=>[|sc scs IH] s1 t1 t12 Tc Ls /=.
- move=>s2 O; split; last by case: Tc=>k [->].
  by case/alw_refl: Ls; move/(ocoh O).
move=>s2; case: Tc IH=>k [->{t12}] H IH O.
split; last first.
- move=>s3 q T; case/tstSeq: T Ls.
  - case=>v [_ ->-> _ ->]; case/alw_refl=>_ /(_ _ (erefl _)).
    by move/(_ _ (H v))/alw_envs; apply.
  case=>sc' [p'][_ ->{q}] T /alw_envs/(_ O)/alw_tstep.
  by move/(_ _ _ _ T)/IH; apply; exists k.
- by split=>//; case/alw_refl: Ls; move/(ocoh O).
move=>G; case: sc G=>//=; first by case: t1 {Ls}=>// a; exists s2, (k a).
move=>sc G; move/(alw_envs Ls): O=>{Ls} /alw_safe/(_ G).
by case=>s' [t2]; exists s', (Seq t2 k), t2.
Qed.

(*******************)
(* after and frame *)
(*******************)

Section AfterFrame.
Variable A : Type.

(* We write the helper lemmas using zigging and zagging, but then
export the main lemma in the self/other notation *)

(* Helper lemma describing the initial interference to running a
thread t. Thread t starts in a state where it owns self part x = self
s1, p is the frame that t doesn't touch, and the environment runs in
the rest. Technically, t runs in s1 \> p, and environment runs in
transposition of (s1 \< p). One should think of the setup as the
environment going first. The lemma says that if the environment
produces s', then p and x are the same as before. Moreover, if we knew
that t has postcondition Q when run in s1 \> p, then it has the same
postcondition when run in s2 \> p, because "after" is defined to be
stable (among other things) under initial interference to t. *)

Lemma aft_frame_other p s1 s' (t : tree A) Q :
        after (s1 \> p) t Q ->
        osteps W (s1 \< p) s' ->
        exists s2, [/\ s' = s2 \< p, after (s2 \> p) t Q &
                       osteps W (State p (joint s1) (self s1 \+ other s1))
                                (State p (joint s2) (self s2 \+ other s2))].
Proof.
move=>Ls E; have C : coh (s1 \< p) by apply/cohZ; case: (alw_refl Ls).
case/(ozig C): E (E)=>s2 [->{s'}] E E'.
exists s2; split=>//; first by apply: alw_envs Ls E.
have X : self s1 = self s2 by move: (oself E); rewrite !zgE.
rewrite (zgEX (self s1)) (zgEX (self s2)) X.
by apply/ozigzag=>//=; rewrite -!zgEX !(joinC p) !zgEX -{1}X -!stEX'.
Qed.

(* The frame now uses the above lemma as the first step, to account
for the base case when the environment interferes before t starts. For
the later interferences of the environment, the proof uses induction
hypthesis, but we need a separate lemma for the first
interference. The conjunct about osteps is provided to support
reasoning with stability. Postcondition Q says what t does, but, by
giving p to the frame, we may have fixed some other property of the
joint state that Q does not say anything about. For example, p may
stand for some permission, without which, some quantity of joint state
remains fixed. Now, t knows it cannot change the quantify, but maybe
the environment can, so Q does not say anything about it. But since p
is frame, we know that the environment cannot change the quantity
either, and thus we chould be able to deduce the stability of that
property in the overall conclusion. The technical way to say that is
that the steps of t, and the environment, both look like environment
steps to the state p. Our actions are interefence to others, just like
other's actions are interference to us. *)

(* We give forward version of the rule, because that's the most
used. I think that frame can be stated in the backward way too, unlike
parallel composition rule, which I couldn't figure out how to state in
a backward form, which is significantly different from the forward
form (i.e., which doesn't existentially quantify over
postconditions. *)

Lemma aft_frame' p s1 (t : tree A) Q :
        after (s1 \> p) t Q ->
        after (s1 \< p) t (fun v' s' => exists s2,
          [/\ s' = s2 \< p, Q v' (s2 \> p) &
              osteps W (State p (joint s1) (self s1 \+ other s1))
                       (State p (joint s2) (self s2 \+ other s2))]).
Proof.
move=>Ls scs.
elim: scs s1 p t Ls =>
       [|sc scs IH] s1 p t /= Ls _ /(aft_frame_other Ls) [s2][->] {}Ls E.
- by case/alw_refl: Ls=>C L; move/cohZ: (C)=>C'; split=>// v /L; exists s2.
have C2 : coh (s2 \< p) by apply/cohZ; case: (alw_refl Ls).
split; first by apply: tst_safe_mono C2 (alw_safe Ls).
- by case/alw_refl: Ls=>C L; move/cohZ: (C)=>C'; split=>// v /L; exists s2.
move=>x q T; case/(tst_frame C2 (alw_safe Ls)): T (T)=>s3 [->{x}] T.
move/cohZ/(tst_coh T)/cohZ: (C2)=>C.
move/tst_other; rewrite !zgE=>O; move/IH: {IH Ls} (alw_tstep Ls T).
apply: alwp_imp=>s4 t' {C} [C] Ls; split=>// v {C scs t'}/Ls.
case=>s5 [->{s4}] H E'; exists s5; split=>//.
apply: iter_trans E'; apply: {v H Q} iter_trans E _.
move/cohZ: (C2)=>C2'.
rewrite (zgEX (other s2)) (zgEX (other s3)) -O.
rewrite [s2]stEX' -zgEX joinC (zgEX (self s2)) in C2; move/cohZ: C2=>C2.
rewrite -zgEX joinC (zgEX (other s2)) in C2; move/cohZ: C2=>C2.
apply/ozigzag=>//; rewrite {2}O. move/(tst_osteps C2'): T.
by rewrite !tr_unlock !zg_unlock.
Qed.

(* The above are helper lemmas stated using zig-zagging. In practice,
I expect that explicitly splitting the self of the initial state into
disjoint portions will be a more understandable idiom. Thus, I
reformulate aft_frame' into the following. *)

Lemma aft_frame p x1 s1 (t : tree A) Q :
        self s1 = x1 \+ p ->
        after (State x1 (joint s1) (p \+ other s1)) t Q ->
        after s1 t (fun v s2 => exists x2,
          [/\ self s2 = x2 \+ p,
              Q v (State x2 (joint s2) (p \+ other s2)) &
              osteps W (State p (joint s1) (x1 \+ other s1))
                       (State p (joint s2) (x2 \+ other s2))]).
Proof.
rewrite (zgEX p)=>S /aft_frame'; rewrite /= -zgEX -S -stEX'.
by apply: aft_imp=>v _ _ [s2][->]; rewrite !zg_unlock; exists (self s2).
Qed.

(* a somewhat shorter way of stating the lemma *)
(* though not clear that proving it directly in this form would have been better *)

Lemma aft_frame2 s s1 s2 (t : tree A) Q :
        splits s s1 s2 -> after s1 t Q ->
        after s t (fun v s' => exists s1' s2',
          [/\ splits s' s1' s2', Q v s1' & osteps W s2 s2']).
Proof.
case=>X X1 X2 J1 J2 H; rewrite [s1]stEX' -J1 X1 in H.
apply: aft_imp (aft_frame X H)=>v s' C' [x'][X'].
set s1' := State _ _ _; rewrite J2 -X2 -stEX'; set s2' := State _ _ _.
by exists s1', s2'.
Qed.

End AfterFrame.

(**********************************)
(* after and parallel composition *)
(**********************************)

Section AfterPar.
Variables A1 A2 : Type.

(* We write the helper lemmas in expanded state notation, similar to
those in the case of aft_frame, but not quite equal as here we don't
have a fixed p, but we now have a thread that changes it. Thus, we now
split the self ot the initial state into explicit x1 and x2, and use j
and o for the other components. The main lemma will export joint and
other notation instead of j and o. *)

Lemma aft_par_safe x1 x2 j o (t1 : tree A1) (t2 : tree A2) Q1 Q2 sc :
        after (State x1 j (x2 \+ o)) t1 Q1 ->
        after (State x2 j (x1 \+ o)) t2 Q2 ->
        tsafe (Par t1 t2 Ret) sc (State (x1 \+ x2) j o).
Proof.
rewrite (zgEX x2 x1) (zgEX x1 x2)=>/aft_frame' Ls1 /aft_frame' /=.
rewrite joinC => Ls2 G; case: sc G=>//=.
- by case: {Ls1 Ls2} t1=>// v1; case: t2=>// v2; eexists _, _.
- move=>sc G; case: (Ls1 [:: sc] _ iter_refl); case=>// s [q].
  by rewrite /tstep !zg_unlock /=; eauto.
move=>sc G; case: (Ls2 [:: sc] _ iter_refl); case=>// s [q].
by rewrite /tstep /= zgEX /= joinC !zg_unlock; eauto.
Qed.

Lemma aft_par_other x1 x2 j o s (t1 : tree A1) (t2 : tree A2) Q1 Q2 :
        after (State x1 j (x2 \+ o)) t1 Q1 ->
        after (State x2 j (x1 \+ o)) t2 Q2 ->
        osteps W (State (x1 \+ x2) j o) s ->
        exists j' o',
          [/\ s = State (x1 \+ x2) j' o', coh s,
              after (State x1 j' (x2 \+ o')) t1 Q1 &
              after (State x2 j' (x1 \+ o')) t2 Q2].
Proof.
move=>Ls1 Ls2; have C : coh (State x2 j o \< x1).
- by rewrite cohZ -zgEX; case/alw_refl: Ls2.
 rewrite (zgEX x1)=>E; case/(ozig C): E (E).
case=>s' j' o' [->{s}] /oself; rewrite !zgE /= => <-{s'}.
rewrite joinC => E; exists j', o'; rewrite zgEX; split=>//.
- by apply: ocoh E _; rewrite -zgEX (zgEX x2) cohZ -zgEX; case/alw_refl: Ls1.
- apply: alw_envs Ls1 _; rewrite (zgEX x2) [X in _ _ X](zgEX x2).
  by apply/ozigzag=>//; rewrite -!zgEX joinC !zgEX.
apply: alw_envs Ls2 _; rewrite (zgEX x1) [X in _ _ X](zgEX x1).
by apply/ozigzag=>//; rewrite joinC.
Qed.

(* Again forward rules; it seems parallel composition can't be given *)
(* as a trully backward rule. One can fake a backward rule, by using *)
(* existentials to request that the user gives posts for t1 and t2 *)
(* but that's not an improvement over the forward rule *)

(* the parallel composition lemma to export *)

Lemma aft_par x1 x2 s (t1 : tree A1) (t2 : tree A2) Q1 Q2 :
        self s = x1 \+ x2 ->
        after (State x1 (joint s) (x2 \+ other s)) t1 Q1 ->
        after (State x2 (joint s) (x1 \+ other s)) t2 Q2 ->
        after s (Par t1 t2 Ret)
          (fun v' s' => exists x1' x2',
            [/\ self s' = x1' \+ x2',
                Q1 v'.1 (State x1' (joint s') (x2' \+ other s')) &
                Q2 v'.2 (State x2' (joint s') (x1' \+ other s'))]).
Proof.
case: s=>_ j o /= ->; move=>Ls1 Ls2 scs.
elim: scs x1 x2 j o t1 t2 Ls1 Ls2 => [|sc scs IH]
          x1 x2 j o t1 t2 Ls1 Ls2 _ /(aft_par_other Ls1 Ls2)
          [j1][o1][->] C {}Ls1 {}Ls2 //.
split=>//; first by apply: aft_par_safe Ls1 Ls2.
move=>s3 q T; case/tstPar: T C Ls1 Ls2.
- case=>_ [v1][v2][->->-> _ ->{t1 t2 q s3}] C Ls1 Ls2; apply: alw_ret.
  move=>_ /(aft_par_other Ls1 Ls2) {C Ls1 Ls2 j o} [j'][o'][->] C.
  case/alw_refl=>_ /(_ _ (erefl _)) Ls1 /alw_refl [_ /(_ _ (erefl _)) Ls2].
  by split=>// _ [<-]; exists x1, x2.
case=>t1' [t2'][sc'][->]; case; case=>_ -> T C Ls1 Ls2.
- rewrite (zgEX x2) in T C; move/(tst_frame C): T (T); rewrite -zgEX.
  case/(_ (alw_safe Ls1)); case=>x1' j' o' [->{s3}] T /tst_other.
  rewrite !zgE /= => E; rewrite -{o'}E /= -!zgEX in T *.
  apply: IH (alw_tstep Ls1 T) (alw_envs Ls2 _).
  rewrite (zgEX o1) [in X in _ _ X](zgEX o1); apply: ozigzag=>//.
  - by rewrite -coh_tr trE tr_unlock /= -zgEX (zgEX x2) -cohZ.
  rewrite -!zgEX; move/tst_osteps: T; rewrite !tr_unlock; apply.
  by rewrite zgEX -cohZ.
rewrite (zgEX x1) in T C; move/(tst_frame C): T (T); rewrite -zgEX.
case/(_ (alw_safe Ls2)); case=>x2' j' o' [->{s3}] T /tst_other.
rewrite !zgE /= => E; rewrite -{o'}E /= -!zgEX in T *.
rewrite /zig joinC; apply: IH (alw_envs Ls1 _) (alw_tstep Ls2 T).
rewrite (zgEX o1) [in X in _ _ X](zgEX o1); apply: ozigzag=>//.
- by rewrite -coh_tr trE tr_unlock /= -zgEX (zgEX x1) -cohZ.
rewrite -!zgEX; move/tst_osteps: T; rewrite !tr_unlock; apply.
by rewrite zgEX -cohZ.
Qed.

(* a more compact form *)

Definition star (P1 P2 : st W -> Prop) s :=
  exists s1 s2, [/\ splits s s1 s2, P1 s1 & P2 s2].

Lemma starC P1 P2 s : star P1 P2 s <-> star P2 P1 s.
Proof. by split; case=>s1 [s2][/spC S *]; exists s2, s1. Qed.


Lemma aft_par2 s (t1 : tree A1) (t2 : tree A2) Q1 Q2 :
        star (fun s1 => after s1 t1 Q1) (fun s2 => after s2 t2 Q2) s ->
        after s (Par t1 t2 Ret) (fun v => star (Q1 v.1) (Q2 v.2)).
Proof.
case=>s1 [s2][[X X1 X2 J1 J2]] H1 H2.
rewrite [s1]stEX' [s2]stEX' -J1 -J2 X1 X2 in H1 H2.
apply: aft_imp (aft_par X H1 H2); case=>v1 v2 s' C' [x1'][x2'][X'].
by set s1' := State _ _ _; set s2' := State _ _ _; exists s1', s2'.
Qed.

End AfterPar.

End Always.


(********************************)
(* After and resource morphisms *)
(********************************)

Section AfterMorph.
Variables (X : Type) (W' W : resource).
Notation after W s t Q := (@after W _ s t Q).

Lemma aft_morph A (tau : morph X W' W) (sim : simulation tau) (t : tree W' A) Q x1 s1 s1' :
        coh W s1 -> [morph tau] x1 s1 s1' -> sim x1 s1 -> after W' s1' t Q ->
        after W s1 (Morf x1 tau t) (fun v s2 =>
          exists x2 s2', [/\ [morph tau] x2 s2 s2', sim x2 s2, Q v s2' &
                             mxsteps tau x1 s1 x2 s2]).
Proof.
move=>C X1 S1 Ls scs.
elim: scs A t Q x1 s1 s1' C X1 S1 Ls=>
      [|sc scs IH] A t Q x1 s1 s1' C X1 S1 Ls s O;
 case: (morph_osteps C X1 S1 O)=>s' [Xs S1'] /(alw_envs Ls) {} Ls;
 move: (omxs tau x1 O) (ocoh O C)=>M {O} C'; split=>//.
- move=>G; case: sc G=>//; first by case: t Ls=>// v; exists s, (Ret v).
  move=>sc /(alw_safe Ls) [s2'][q] /(tst_forw C' Xs S1') [x2][s2][].
  by exists s2, (Morf x2 tau q), x2, q.
move=>sx q; rewrite /tstep tstMorf; case=>H.
- case: H Ls=>v [_ ->-> _ ->{sx t q}] Ls.
  apply: alwp_ret=>s2 O; split; first by apply: ocoh C'.
  case: (morph_osteps C' Xs S1' O)=>s2' [X2 S2'] /(alw_envs Ls)/alw_refl.
  case=>C2' /(_ _ (erefl _)) H _ [<-]; exists x1, s2'; split=>//.
  by apply: iter_trans M (omxs _ _ O).
case: H=>sc' [xx][t2][_ ->{sc q} T]; case/(tst_backw C' Xs S1'): (T)=>sx' [Xx Sx].
move/(alw_tstep Ls)/(IH _ _ _ _ _ _ (tst_coh T C') Xx Sx).
apply: alwp_imp=>s2 q [C2 H].
split=>// v /H [x2][s2'][X2 {C2 H} S H N]; exists x2, s2'; split=>//.
apply: iter_trans N; apply: iter_trans M _.
case: (tst_wcovered C' T); first by case=>->->.
case=>a [I P]; apply: iter1; apply: SelfMxStep I _.
by case: P=>sw [sw'][tw][sw1][sw1'][tw'][[->->-> Hw]][]; exists tw, sw, sw'.
Qed.

End AfterMorph.
