From Coq Require Import ssreflect ssrbool ssrfun.
From mathcomp Require Import ssrnat eqtype.
From fcsl Require Import axioms pred prelude.
From fcsl Require Import options.

(**********)
(* Posets *)
(**********)

Module Poset.

Section RawMixin.

Record mixin_of (T : Type) := Mixin {
  mx_leq : T -> T -> Prop;
  _ : forall x, mx_leq x x;
  _ : forall x y, mx_leq x y -> mx_leq y x -> x = y;
  _ : forall x y z, mx_leq x y -> mx_leq y z -> mx_leq x z}.

End RawMixin.

Section ClassDef.

Record class_of T := Class {mixin : mixin_of T}.

Structure type : Type := Pack {sort : Type; _ : class_of sort}.
Local Coercion sort : type >-> Sortclass.

Variables (T : Type) (cT : type).
Definition class := let: Pack _ c as cT' := cT return class_of cT' in c.
Definition clone c of phant_id class c := @Pack T c.

(* produce a poset type out of the mixin *)
(* equalize m0 and m by means of a phantom *)
Definition pack (m0 : mixin_of T) :=
  fun m & phant_id m0 m => Pack (@Class T m).

Definition leq := mx_leq (mixin class).

End ClassDef.

Module Exports.
Coercion sort : type >-> Sortclass.
Notation poset := Poset.type.
Notation PosetMixin := Poset.Mixin.
Notation Poset T m := (@pack T _ m id).

Notation "[ 'poset' 'of' T 'for' cT ]" := (@clone T cT _ idfun)
  (at level 0, format "[ 'poset'  'of'  T  'for'  cT ]") : form_scope.
Notation "[ 'poset' 'of' T ]" := (@clone T _ _ id)
  (at level 0, format "[ 'poset'  'of'  T ]") : form_scope.

Notation "x <== y" := (Poset.leq x y) (at level 70).

(* re-state lattice properties using the exported notation *)
Section Laws.
Variable T : poset.

Lemma poset_refl (x : T) : x <== x.
Proof. by case: T x=>S [[leq B R]]. Qed.

Lemma poset_asym (x y : T) : x <== y -> y <== x -> x = y.
Proof. by case: T x y=>S [[l R A Tr]] *; apply: (A). Qed.

Lemma poset_trans (y x z : T) : x <== y -> y <== z -> x <== z.
Proof. by case: T y x z=>S [[l R A Tr]] ? x y z; apply: (Tr). Qed.

End Laws.

Hint Resolve poset_refl : core.

Add Parametric Relation (T : poset) : T (@Poset.leq T)
  reflexivity proved by (@poset_refl _)
  transitivity proved by (fun x y => @poset_trans _ y x) as poset_rel.

End Exports.

End Poset.

Export Poset.Exports.

(**********************)
(* monotone functions *)
(**********************)

Definition monotone (T1 T2 : poset) (f : T1 -> T2) :=
  forall x y, x <== y -> f x <== f y.

Section MonotoneFunStructure.
Variables T1 T2 : poset.

Structure monotone_type := Monotone {
  mfun :> T1 -> T2;
  _ : monotone mfun}.

Definition monotone_for of phant (T1 -> T2) := monotone_type.

Definition clone_monotone f :=
  let: @Monotone _ pf := f
    return {type of @Monotone for f} ->
      monotone_for (Phant (T1 -> T2))
  in fun k => k pf.

End MonotoneFunStructure.

Arguments Monotone [T1 T2] mfun _.

Notation "{ 'monotone' fT }" := (monotone_for (Phant fT))
  (at level 0, format "{ 'monotone'  fT }") : form_scope.
Notation "[ 'monotone' 'of' f ]" :=
  (clone_monotone (@Monotone _ _ f))
  (at level 0, format "[ 'monotone'  'of'  f ]") : form_scope.

Lemma monotoneP (T1 T2 : poset) (f : {monotone T1 -> T2}) x y :
        x <== y -> f x <== f y.
Proof. by case: f=>f; apply. Qed.

(**************************)
(* some basic definitions *)
(**************************)

Section IdealDef.
Variable T : poset.

Structure ideal (P : T) := Ideal {id_val : T; id_pf : id_val <== P}.

(* Changing the type of the ideal *)

Lemma relaxP (P1 P2 : T) : P1 <== P2 -> forall p, p <== P1 -> p <== P2.
Proof. by move=>H1 p H2; apply: poset_trans H1. Qed.

Definition relax (P1 P2 : T) (x : ideal P1) (pf : P1 <== P2) :=
  Ideal (relaxP pf (id_pf x)).

End IdealDef.

(***********************)
(* poset constructions *)
(***********************)

Section SubPoset.
Variables (T : poset) (s : Pred T).

Local Notation tp := {x : T | x \In s}.

Definition sub_leq (p1 p2 : tp) := sval p1 <== sval p2.

Lemma sub_refl x : sub_leq x x.
Proof. by rewrite /sub_leq. Qed.

Lemma sub_asym x y : sub_leq x y -> sub_leq y x -> x = y.
Proof.
move: x y=>[x Hx][y Hy]; rewrite /sub_leq /= => H1 H2.
move: (poset_asym H1 H2) Hy=> <- Hy; congr exist.
by apply: pf_irr.
Qed.

Lemma sub_trans x y z : sub_leq x y -> sub_leq y z -> sub_leq x z.
Proof.
move: x y z=>[x Hx][y Hy][z Hz]; rewrite /sub_leq /=.
by apply: poset_trans.
Qed.

(* no need to put canonical here, because the system won't be *)
(* able to get the proof C from the {x : T | x \In s} syntax *)
(* actually, it will *)
Definition subPosetMixin := PosetMixin sub_refl sub_asym sub_trans.
Canonical subPoset := Eval hnf in Poset tp subPosetMixin.
End SubPoset.

(* pairing of posets *)

Section PairPoset.
Variable (A B : poset).
Local Notation tp := (A * B)%type.

Definition pair_leq (p1 p2 : tp) := p1.1 <== p2.1 /\ p1.2 <== p2.2.

Lemma pair_refl x : pair_leq x x.
Proof. by []. Qed.

Lemma pair_asym x y : pair_leq x y -> pair_leq y x -> x = y.
Proof.
move: x y=>[x1 x2][y1 y2][/= H1 H2][/= H3 H4].
by congr (_, _); apply: poset_asym.
Qed.

Lemma pair_trans x y z : pair_leq x y -> pair_leq y z -> pair_leq x z.
Proof.
move: x y z=>[x1 x2][y1 y2][z1 z2][/= H1 H2][/= H3 H4]; split=>/=.
- by apply: poset_trans H3.
by apply: poset_trans H4.
Qed.

Definition pairPosetMixin :=
  PosetMixin pair_refl pair_asym pair_trans.
Canonical pairPoset := Eval hnf in Poset tp pairPosetMixin.

End PairPoset.

(* functions into a poset form a poset *)

Section FunPoset.
Variable (A : Type) (B : poset).
Local Notation tp := (A -> B).

Definition fun_leq (p1 p2 : tp) := forall x, p1 x <== p2 x.

Lemma fun_refl x : fun_leq x x.
Proof. by []. Qed.

Lemma fun_asym x y : fun_leq x y -> fun_leq y x -> x = y.
Proof.
move=>H1 H2; apply: fext=>z;
by apply: poset_asym; [apply: H1 | apply: H2].
Qed.

Lemma fun_trans x y z : fun_leq x y -> fun_leq y z -> fun_leq x z.
Proof. by move=>H1 H2 t; apply: poset_trans (H2 t). Qed.

Definition funPosetMixin := PosetMixin fun_refl fun_asym fun_trans.
Canonical funPoset := Eval hnf in Poset tp funPosetMixin.

End FunPoset.

(* dependent functions into a poset form a poset *)

Section DFunPoset.
Variables (A : Type) (B : A -> poset).
Local Notation tp := (forall x, B x).

Definition dfun_leq (p1 p2 : tp) := forall x, p1 x <== p2 x.

Lemma dfun_refl x : dfun_leq x x.
Proof. by []. Qed.

Lemma dfun_asym x y : dfun_leq x y -> dfun_leq y x -> x = y.
Proof.
move=>H1 H2; apply: fext=>z;
by apply: poset_asym; [apply: H1 | apply: H2].
Qed.

Lemma dfun_trans x y z : dfun_leq x y -> dfun_leq y z -> dfun_leq x z.
Proof. by move=>H1 H2 t; apply: poset_trans (H2 t). Qed.

(* no point in declaring this canonical, since it's keyed on forall *)
(* and forall is not a symbol *)
Definition dfunPosetMixin :=
  PosetMixin dfun_refl dfun_asym dfun_trans.
Definition dfunPoset := Eval hnf in Poset tp dfunPosetMixin.

End DFunPoset.

(* ideal of a poset is a poset *)

Section IdealPoset.
Variable (T : poset) (P : T).

Definition ideal_leq (p1 p2 : ideal P) := id_val p1 <== id_val p2.

Lemma ideal_refl x : ideal_leq x x.
Proof. by case: x=>x H; rewrite /ideal_leq. Qed.

Lemma ideal_asym x y : ideal_leq x y -> ideal_leq y x -> x = y.
Proof.
move: x y=>[x1 H1][x2 H2]; rewrite /ideal_leq /= => H3 H4; move: H1 H2.
rewrite (poset_asym H3 H4)=>H1 H2.
congr Ideal; apply: pf_irr.
Qed.

Lemma ideal_trans x y z : ideal_leq x y -> ideal_leq y z -> ideal_leq x z.
Proof. by apply: poset_trans. Qed.

Definition idealPosetMixin :=
  PosetMixin ideal_refl ideal_asym ideal_trans.
Canonical idealPoset := Eval hnf in Poset (ideal P) idealPosetMixin.

End IdealPoset.

(* Prop is a poset *)

Section PropPoset.

Definition prop_leq (p1 p2 : Prop) := p1 -> p2.

Lemma prop_refl x : prop_leq x x.
Proof. by []. Qed.

Lemma prop_asym x y : prop_leq x y -> prop_leq y x -> x = y.
Proof. by move=>H1 H2; apply: pext. Qed.

Lemma prop_trans x y z : prop_leq x y -> prop_leq y z -> prop_leq x z.
Proof. by move=>H1 H2; move/H1. Qed.

Definition propPosetMixin :=
  PosetMixin prop_refl prop_asym prop_trans.
Canonical propPoset := Eval hnf in Poset Prop propPosetMixin.

End PropPoset.


(*********************)
(* Complete lattices *)
(*********************)

Module Lattice.

Section RawMixin.

Variable T : poset.

Record mixin_of := Mixin {
  mx_sup : Pred T -> T;
  _ : forall (s : Pred T) x, x \In s -> x <== mx_sup s;
  _ : forall (s : Pred T) x,
        (forall y, y \In s -> y <== x) -> mx_sup s <== x}.

End RawMixin.

Section ClassDef.

Record class_of (T : Type) := Class {
  base : Poset.class_of T;
  mixin : mixin_of (Poset.Pack base)}.

Local Coercion base : class_of >-> Poset.class_of.

Structure type : Type := Pack {sort : Type; _ : class_of sort}.
Local Coercion sort : type >-> Sortclass.

Variables (T : Type) (cT : type).
Definition class := let: Pack _ c as cT' := cT return class_of cT' in c.
Definition clone c of phant_id class c := @Pack T c.

(* produce a lattice type out of the mixin *)
(* equalize m0 and m by means of a phantom *)
Definition pack b0 (m0 : mixin_of (Poset.Pack b0)) :=
  fun m & phant_id m0 m => Pack (@Class T b0 m).

Definition sup (s : Pred cT) : cT := mx_sup (mixin class) s.

Definition poset := Poset.Pack class.

End ClassDef.

Module Exports.
Coercion base : class_of >-> Poset.class_of.
Coercion sort : type >-> Sortclass.
Coercion poset : type >-> Poset.type.
Canonical Structure poset.

Notation lattice := Lattice.type.
Notation LatticeMixin := Lattice.Mixin.
Notation Lattice T m := (@pack T _ _ m id).

Notation "[ 'lattice' 'of' T 'for' cT ]" := (@clone T cT _ idfun)
  (at level 0, format "[ 'lattice'  'of'  T  'for' cT ]") : form_scope.
Notation "[ 'lattice' 'of' T ]" := (@clone T _ _ id)
  (at level 0, format "[ 'lattice'  'of'  T ]") : form_scope.

Arguments Lattice.sup [cT] s.
Prenex Implicits Lattice.sup.
Notation sup := Lattice.sup.

(* re-state lattice properties using the exported notation *)
Section Laws.
Variable T : lattice.

Lemma supP (s : Pred T) x : x \In s -> x <== sup s.
Proof. by case: T s x=>S [[p]][/= s H1 *]; apply: H1. Qed.

Lemma supM (s : Pred T) x : (forall y, y \In s -> y <== x) -> sup s <== x.
Proof. by case: T s x=>S [[p]][/= s H1 H2 *]; apply: (H2). Qed.

End Laws.

End Exports.

End Lattice.

Export Lattice.Exports.

(* we have greatest lower bounds too *)
Section Infimum.
Variable T : lattice.

Definition inf (s : Pred T) :=
  sup [Pred x : T | forall y, y \In s -> x <== y].

Lemma infP s x : x \In s -> inf s <== x.
Proof. by move=>H; apply: supM=>y; apply. Qed.

Lemma infM s x : (forall y, y \In s -> x <== y) -> x <== inf s.
Proof. by apply: supP. Qed.

End Infimum.

(* we can compute least and greatest fixed points *)

Section Lat.
Variable T : lattice.

Definition lbot : T := sup Pred0.

Definition tarski_lfp (f : T -> T) := inf [Pred x : T | f x <== x].
Definition tarski_gfp (f : T -> T) := sup [Pred x : T | x <== f x].

Definition sup_closed (T : lattice) :=
  [Pred s : Pred T | forall d, d <=p s -> sup d \In s].

Definition sup_closure (T : lattice) (s : Pred T) :=
  [Pred p : T | forall t : Pred T, s <=p t -> t \In sup_closed T -> p \In t].

End Lat.

Arguments lbot {T}.
Arguments sup_closed {T}.
Arguments sup_closure [T] s.
Prenex Implicits sup_closed sup_closure.


Section BasicProperties.
Variable T : lattice.

Lemma sup_mono (s1 s2 : Pred T) : s1 <=p s2 -> sup s1 <== sup s2.
Proof. by move=>H; apply: supM=>y /H/supP. Qed.

Lemma supE (s1 s2 : Pred T) : s1 =p s2 -> sup s1 = sup s2.
Proof. by move=>H; apply: poset_asym; apply: supM=>y /H/supP. Qed.

(* Knaster-Tarski *)
Lemma tarski_lfp_fixed' (f : T -> T) :
        monotone f -> f (tarski_lfp f) = tarski_lfp f.
Proof.
move=>M.
suff L: f (tarski_lfp f) <== tarski_lfp f.
- by apply: poset_asym=>//; apply/infP/M.
by apply: infM=>x L; apply: poset_trans (L); apply/M/infP.
Qed.

Lemma tarski_lfp_fixed (f : {monotone T -> T}) :
        f (tarski_lfp f) = tarski_lfp f.
Proof. by apply: tarski_lfp_fixed'; apply: monotoneP. Qed.

Lemma tarski_lfp_least f (x : T) : f x = x -> tarski_lfp f <== x.
Proof. by move=>H; apply: infP; rewrite InE /= H. Qed.

Lemma tarski_gfp_fixed' (f : T -> T) :
        monotone f -> f (tarski_gfp f) = tarski_gfp f.
Proof.
move=>M.
suff L: tarski_gfp f <== f (tarski_gfp f).
- by apply: poset_asym=>//; apply/supP/M.
by apply: supM=>x L; apply: poset_trans (L) _; apply/M/supP.
Qed.

Lemma tarski_gfp_fixed (f : {monotone T -> T}) :
        f (tarski_gfp f) = tarski_gfp f.
Proof. by apply: tarski_gfp_fixed'; apply: monotoneP. Qed.

Lemma tarski_gfp_greatest f (x : T) : f x = x -> x <== tarski_gfp f.
Proof. by move=>H; apply: supP; rewrite InE /= H. Qed.

Lemma tarski_lfp_mono' (f1 f2 : T -> T) :
        monotone f2 -> f1 <== f2 -> tarski_lfp f1 <== tarski_lfp f2.
Proof.
move=>M H; apply: infP; apply: poset_trans (H (tarski_lfp f2)) _.
by rewrite tarski_lfp_fixed'.
Qed.

Lemma tarski_lfp_mono (f1 : T -> T) (f2 : {monotone T -> T}) :
        f1 <== f2 -> tarski_lfp f1 <== tarski_lfp f2.
Proof. by apply: tarski_lfp_mono'; apply: monotoneP. Qed.

Lemma tarski_gfp_mono' (f1 f2 : T -> T) :
        monotone f1 -> f1 <== f2 -> tarski_gfp f1 <== tarski_gfp f2.
Proof.
move=>M H; apply/supP/poset_trans/(H (tarski_gfp f1)).
by rewrite tarski_gfp_fixed'.
Qed.

Lemma tarski_gfp_mono (f1 : {monotone T -> T}) (f2 : T -> T) :
        mfun f1 <== f2 -> tarski_gfp f1 <== tarski_gfp f2.
Proof. by apply: tarski_gfp_mono'; apply: monotoneP. Qed.

(* closure contains s *)
Lemma sup_clos_sub (s : Pred T) : s <=p sup_closure s.
Proof. by move=>p H1 t H2 H3; apply: H2 H1. Qed.

(* closure is smallest *)
Lemma sup_clos_min (s : Pred T) :
        forall t, s <=p t -> sup_closed t -> sup_closure s <=p t.
Proof. by move=>t H1 H2 p; move/(_ _ H1 H2). Qed.

(* closure is closed *)
Lemma sup_closP (s : Pred T) : sup_closed (sup_closure s).
Proof.
move=>d H1 t H3 H4; move: (sup_clos_min H3 H4)=>{}H3.
by apply: H4=>x; move/H1; move/H3.
Qed.

Lemma sup_clos_idemp (s : Pred T) : sup_closed s -> sup_closure s =p s.
Proof. by move=>p; split; [apply: sup_clos_min | apply: sup_clos_sub]. Qed.

Lemma sup_clos_mono (s1 s2 : Pred T) :
        s1 <=p s2 -> sup_closure s1 <=p sup_closure s2.
Proof.
move=>H1; apply: sup_clos_min (@sup_closP s2).
by move=>t /H1; apply: sup_clos_sub.
Qed.

End BasicProperties.

(* lattice constructions *)

Section SubLattice.
Variables (T : lattice) (s : Pred T) (C : sup_closed s).
Local Notation tp := (subPoset s).

Definition sub_sup' (u : Pred tp) : T :=
  sup [Pred x : T | exists y, y \In u /\ x = sval y].

Lemma sub_supX (u : Pred tp) : sub_sup' u \In s.
Proof. by apply: C=>x [][y H][_] ->. Qed.

Definition sub_sup (u : Pred tp) : tp :=
  exist _ (sub_sup' u) (sub_supX u).

Lemma sub_supP (u : Pred tp) (x : tp) : x \In u -> x <== sub_sup u.
Proof. by move=>H; apply: supP; exists x. Qed.

Lemma sub_supM (u : Pred tp) (x : tp) :
        (forall y, y \In u -> y <== x) -> sub_sup u <== x.
Proof. by move=>H; apply: supM=>y [z][H1] ->; apply: H H1. Qed.

Definition subLatticeMixin := LatticeMixin sub_supP sub_supM.
Definition subLattice := Eval hnf in Lattice {x : T | x \In s} subLatticeMixin.

End SubLattice.

(* pairing *)

Section PairLattice.
Variables (A B : lattice).
Local Notation tp := (A * B)%type.

Definition pair_sup (s : Pred tp) : tp :=
            (sup [Pred p | exists f, p = f.1 /\ f \In s],
             sup [Pred p | exists f, p = f.2 /\ f \In s]).

Lemma pair_supP (s : Pred tp) (p : tp) : p \In s -> p <== pair_sup s.
Proof. by move=>H; split; apply: supP; exists p. Qed.

Lemma pair_supM (s : Pred tp) (p : tp) :
        (forall q, q \In s -> q <== p) -> pair_sup s <== p.
Proof. by move=>H; split; apply: supM=>y [f][->]; case/H. Qed.

Definition pairLatticeMixin := LatticeMixin pair_supP pair_supM.
Canonical pairLattice := Eval hnf in Lattice tp pairLatticeMixin.

End PairLattice.

(* functions into a latice form a lattice *)

Section FunLattice.
Variables (A : Type) (B : lattice).
Local Notation tp := (A -> B).

Definition fun_sup (s : Pred tp) : tp :=
  fun x => sup [Pred p | exists f, f \In s /\ p = f x].

Lemma fun_supP (s : Pred tp) (p : tp) : p \In s -> p <== fun_sup s.
Proof. by move=>H x; apply: supP; exists p. Qed.

Lemma fun_supM (s : Pred tp) (p : tp) :
        (forall q, q \In s -> q <== p) -> fun_sup s <== p.
Proof. by move=>H t; apply: supM=>x [f][H1] ->; apply: H. Qed.

Definition funLatticeMixin := LatticeMixin fun_supP fun_supM.
Canonical funLattice := Eval hnf in Lattice tp funLatticeMixin.

End FunLattice.

(* dependent functions into a lattice form a lattice *)

Section DFunLattice.
Variables (A : Type) (B : A -> lattice).
Local Notation tp := (dfunPoset B).

Definition dfun_sup (s : Pred tp) : tp :=
  fun x => sup [Pred p | exists f, f \In s /\ p = f x].

Lemma dfun_supP (s : Pred tp) (p : tp) :
        p \In s -> p <== dfun_sup s.
Proof. by move=>H x; apply: supP; exists p. Qed.

Lemma dfun_supM (s : Pred tp) (p : tp) :
       (forall q, q \In s -> q <== p) -> dfun_sup s <== p.
Proof. by move=>H t; apply: supM=>x [f][H1] ->; apply: H. Qed.

Definition dfunLatticeMixin := LatticeMixin dfun_supP dfun_supM.
Definition dfunLattice := Eval hnf in Lattice (forall x, B x) dfunLatticeMixin.

End DFunLattice.

(* applied sup equals the sup of applications *)

Lemma sup_appE A (B : lattice) (s : Pred (A -> B)) (x : A) :
        sup s x = sup [Pred y : B | exists f, f \In s /\ y = f x].
Proof. by []. Qed.

Lemma sup_dappE A (B : A -> lattice) (s : Pred (dfunLattice B)) (x : A) :
        sup s x = sup [Pred y : B x | exists f, f \In s /\ y = f x].
Proof. by []. Qed.

(* ideal of a lattice forms a lattice *)

Section IdealLattice.
Variables (T : lattice) (P : T).

Definition ideal_sup' (s : Pred (ideal P)) :=
  sup [Pred x | exists p, p \In s /\ id_val p = x].

Lemma ideal_supP' (s : Pred (ideal P)) : ideal_sup' s <== P.
Proof. by apply: supM=>y [[x]] H /= [_] <-. Qed.

Definition ideal_sup (s : Pred (ideal P)) := Ideal (ideal_supP' s).

Lemma ideal_supP (s : Pred (ideal P)) p :
        p \In s -> p <== ideal_sup s.
Proof. by move=>H; apply: supP; exists p. Qed.

Lemma ideal_supM (s : Pred (ideal P)) p :
        (forall q, q \In s -> q <== p) -> ideal_sup s <== p.
Proof. by move=>H; apply: supM=>y [q][H1] <-; apply: H. Qed.

Definition idealLatticeMixin := LatticeMixin ideal_supP ideal_supM.
Canonical idealLattice := Eval hnf in Lattice (ideal P) idealLatticeMixin.

End IdealLattice.

(* Prop is a lattice *)

Section PropLattice.

Definition prop_sup (s : Pred Prop) : Prop := exists p, p \In s /\ p.

Lemma prop_supP (s : Pred Prop) p : p \In s -> p <== prop_sup s.
Proof. by exists p. Qed.

Lemma prop_supM (s : Pred Prop) p :
        (forall q, q \In s -> q <== p) -> prop_sup s <== p.
Proof. by move=>H [r][]; move/H. Qed.

Definition propLatticeMixin := LatticeMixin prop_supP prop_supM.
Canonical propLattice := Eval hnf in Lattice Prop propLatticeMixin.

End PropLattice.


(* monotonicity *)

Section MonoCanonicals.
Variables T T1 T2 T3 S1 S2 : poset.

Lemma id_mono_ax : monotone (@idfun T). Proof. by []. Qed.
Canonical id_mono := Eval hnf in Monotone idfun id_mono_ax.

Definition const_fun (y : T2) (_ : T1) := y.

Lemma const_mono_ax (y : T2) : monotone (const_fun y). Proof. by []. Qed.
Canonical const_mono y := Eval hnf in Monotone (const_fun y) (const_mono_ax y).

Lemma proj1_mono_ax : monotone (@fst T1 T2).
Proof. by case=>x1 x2 [y1 y2][]. Qed.
Canonical proj1_mono := Eval hnf in Monotone fst proj1_mono_ax.

Lemma proj2_mono_ax : monotone (@snd T1 T2).
Proof. by case=>x1 x2 [y1 y2][]. Qed.
Canonical proj2_mono := Eval hnf in Monotone snd proj2_mono_ax.

Definition diag (x : T) := (x, x).

Lemma diag_mono_ax : monotone diag. Proof. by []. Qed.
Canonical diag_mono := Eval hnf in Monotone diag diag_mono_ax.

Lemma sval_mono_ax (P : Pred T) : monotone (@sval T P).
Proof. by []. Qed.
Canonical sval_mono P := Eval hnf in Monotone sval (@sval_mono_ax P).

Definition app A B (x : A) := fun f : A -> B => f x.
Lemma app_mono_ax A (x : A) : monotone (@app A T x).
Proof. by move=>f1 f2; apply. Qed.
Canonical app_mono A x := Eval hnf in Monotone (@app A T x) (app_mono_ax x).

Definition dapp A (B : A -> poset) (x : A) := fun f : dfunPoset B => f x.
Lemma dapp_mono_ax A (B : A -> poset) (x : A) : monotone (@dapp A B x).
Proof. by move=>f1 f2; apply. Qed.
Canonical dapp_mono A B x := Eval hnf in Monotone (@dapp A B x) (dapp_mono_ax x).

Section Comp.
Variables (f1 : {monotone T2 -> T1}) (f2 : {monotone T3 -> T2}).

Lemma comp_mono_ax : monotone (f1 \o f2).
Proof. by move=>x y H; apply/monotoneP/monotoneP. Qed.
Canonical comp_mono := Eval hnf in Monotone (f1 \o f2) comp_mono_ax.

End Comp.

Section Prod.
Variables (f1 : {monotone S1 -> T1}) (f2 : {monotone S2 -> T2}).

Lemma prod_mono_ax : monotone (f1 \* f2).
Proof. by case=>x1 x2 [y1 y2][/= H1 H2]; split=>/=; apply: monotoneP. Qed.
Canonical prod_mono := Eval hnf in Monotone (f1 \* f2) prod_mono_ax.

End Prod.
End MonoCanonicals.

Prenex Implicits diag app.

(**********)
(* Chains *)
(**********)

Section Chains.
Variable T : poset.

Definition chain_axiom (s : Pred T) :=
  (exists d, d \In s) /\
  (forall x y, x \In s -> y \In s -> x <== y \/ y <== x).

Structure chain := Chain {
  pred_of :> Pred T;
  _ : chain_axiom pred_of}.

Definition chain_for of phant T := chain.

Canonical chainPredType := Eval hnf in @mkPredType T chain pred_of.

Definition clone_chain (s : chain) :=
  let: Chain _ pf := s
    return {type of @Chain for s} -> chain_for (Phant T)
  in fun k => k pf.

End Chains.

Arguments Chain [T] _.

Notation "{ 'chain' T }" := (chain_for (Phant T))
 (at level 0, format "{ 'chain'  T }") : form_scope.
Notation "[ 'chain' 'of' s ]" :=
  (clone_chain (@Chain _ s))
  (at level 0, format "[ 'chain'  'of'  s ]") : form_scope.

Lemma chainE (T : poset) (s1 s2 : chain T) :
        s1 = s2 <-> pred_of s1 =p pred_of s2.
Proof.
split=>[->//|]; move: s1 s2=>[s1 H1][s2 H2] /= E; move: H1 H2.
suff: s1 = s2 by move=>-> H1 H2; congr Chain; apply: pf_irr.
by apply: fext=>x; apply: pext; split=> /E.
Qed.

Section ImageChain.
Variables (T1 T2 : poset) (f : {monotone T1 -> T2}) (s : chain T1).

Lemma image_chainP : chain_axiom (Image f s).
Proof.
case: s=>p [[d H1] H2]; split=>[|x y]; first by exists (f d); exists d.
case=>x1 -> H3 [y1 -> H4]; rewrite -!toPredE /= in H3 H4.
by case: (H2 x1 y1 H3 H4)=>L; [left | right]; apply: monotoneP L.
Qed.

Canonical image_chain := Eval hnf in Chain (Image f s) image_chainP.

End ImageChain.

(* some chain lemmas and constructions *)
(* most of these lemmas hold for Pred, not just for chain *)

Lemma id_chainE T (s : chain T) f : f =1 idfun -> Image f s =p s.
Proof.
move=>H x; split; first by case=>y ->; rewrite H.
by exists x=>//; rewrite H.
Qed.

Section ChainConst.
Variables (T1 T2 : poset) (y : T2).

Definition const_pred : Pred _ := Pred1 y.
Lemma const_chainP : chain_axiom const_pred.
Proof. by split; [exists y | move=>x1 x2 <-<-; left]. Qed.

Canonical const_chain := Eval hnf in Chain const_pred const_chainP.

Lemma const_chainE (s : chain T1) : Image (const_fun y) s =p const_chain.
Proof.
move=>z1; split; first by case=>z2 ->.
by case: s=>s [[d] H1] H2 <-; exists d.
Qed.

End ChainConst.


Section ChainCompose.
Variables T1 T2 T3 : poset.
Variables (f1 : {monotone T2 -> T1}) (f2 : {monotone T3 -> T2}).
Variable s : chain T3.

Lemma comp_chainE : Image f1 (Image f2 s) =p Image (f1 \o f2) s.
Proof. by move=>x; apply: ImageIm. Qed.

End ChainCompose.


(*********)
(* CPO's *)
(*********)

Module CPO.

Section RawMixin.

Record mixin_of (T : poset) := Mixin {
  mx_bot : T;
  mx_lim : chain T -> T;
  _ : forall x, mx_bot <== x;
  _ : forall (s : chain T) x, x \In s -> x <== mx_lim s;
  _ : forall (s : chain T) x,
        (forall y, y \In s -> y <== x) -> mx_lim s <== x}.

End RawMixin.

Section ClassDef.

Record class_of (T : Type) := Class {
  base : Poset.class_of T;
  mixin : mixin_of (Poset.Pack base)}.

Local Coercion base : class_of >-> Poset.class_of.

Structure type : Type := Pack {sort; _ : class_of sort}.
Local Coercion sort : type >-> Sortclass.

Variables (T : Type) (cT : type).
Definition class := let: Pack _ c as cT' := cT return class_of cT' in c.
Definition clone c of phant_id class c := @Pack T c.

Definition pack b0 (m0 : mixin_of (Poset.Pack b0)) :=
  fun m & phant_id m0 m => Pack (@Class T b0 m).

Definition poset := Poset.Pack class.
Definition lim (s : chain poset) : cT := mx_lim (mixin class) s.
Definition bot : cT := mx_bot (mixin class).

End ClassDef.

Module Import Exports.
Coercion base : class_of >-> Poset.class_of.
Coercion sort : type >-> Sortclass.
Coercion poset : type >-> Poset.type.
Canonical Structure poset.

Notation cpo := type.
Notation CPOMixin := Mixin.
Notation CPO T m := (@pack T _ _ m id).

Notation "[ 'cpo' 'of' T 'for' cT ]" := (@clone T cT _ idfun)
  (at level 0, format "[ 'cpo'  'of'  T  'for'  cT ]") : form_scope.
Notation "[ 'cpo' 'of' T ]" := (@clone T _ _ id)
  (at level 0, format "[ 'cpo'  'of'  T ]") : form_scope.

Arguments CPO.bot {cT}.
Arguments CPO.lim [cT] s.
Prenex Implicits CPO.lim.
Prenex Implicits CPO.bot.
Notation lim := CPO.lim.
Notation bot := CPO.bot.

Section Laws.
Variable D : cpo.

Lemma botP (x : D) : bot <== x.
Proof. by case: D x=>S [[p]][/= b l H1 *]; apply: H1. Qed.

Lemma limP (s : chain D) x : x \In s -> x <== lim s.
Proof. by case: D s x=>S [[p]][/= b l H1 H2 *]; apply: (H2). Qed.

Lemma limM (s : chain D) x : (forall y, y \In s -> y <== x) -> lim s <== x.
Proof. by case: D s x=>S [[p]][/= b l H1 H2 H3 *]; apply: (H3). Qed.

End Laws.

End Exports.

End CPO.

Export CPO.Exports.

Hint Resolve botP : core.

Definition lim_phant (D : cpo) (s : chain D) of phantom (Pred _) s := lim s.

Notation "''lim' s" := (lim_phant (Phantom (Pred _) s))
  (at level 10, s at level 8, format "''lim'  s") : form_scope.

(* this necessary to avoid dependent rewriting *)
Lemma limE (D : cpo) (s1 s2 : chain D) :
        s1 =p s2 -> 'lim s1 = 'lim s2.
Proof. by move/chainE=>->. Qed.

Lemma limX (D1 D2 : cpo) (f : {monotone D1 -> D2}) (s : chain D1) :
        'lim (Image f s) = lim [chain of Image f s].
Proof. by []. Qed.

(* common chain constructions *)

(* adding bot to a chain *)

Definition lift (D : cpo) (s : Pred D) : Pred D :=
  fun x => x = bot \/ x \In s.

Section LiftChain.
Variable (D : cpo) (s : chain D).

Lemma lift_chainP : chain_axiom (lift s).
Proof.
case: s=>p [[d H1] H2] /=; split=>[|x y]; first by exists d; right.
by case=>[->|H3][->|H4]; auto.
Qed.

Canonical lift_chain := Eval hnf in Chain (lift s) lift_chainP.

End LiftChain.

(****************************)
(* common cpo constructions *)
(****************************)

(* pairs *)

Section PairCPO.
Variables (A B : cpo).

Let pair_bot := (@bot A, @bot B).

Let pair_lim (s : {chain A * B}) :=
  ('lim (Image fst s), 'lim (Image snd s)).

Lemma pair_botP x : pair_leq pair_bot x.
Proof. by split; apply: botP. Qed.

Lemma pair_limP (s : {chain A * B}) x : x \In s -> x <== pair_lim s.
Proof. by split; apply: limP; exists x. Qed.

Lemma pair_limM (s : {chain A * B}) x :
        (forall y, y \In s -> y <== x) -> pair_lim s <== x.
Proof. by split; apply: limM=>y [z ->]; case/H. Qed.

Definition pairCPOMixin := CPOMixin pair_botP pair_limP pair_limM.
Canonical pairCPO := Eval hnf in CPO (A * B) pairCPOMixin.

End PairCPO.

(* functions *)

Section FunCPO.
Variable (A : Type) (B : cpo).

Let fun_bot (x : A) := @bot B.

Let fun_lim (s : {chain A -> B}) x := 'lim (Image (app x) s).

Lemma fun_botP f : fun_leq fun_bot f.
Proof. by move=>y; apply: botP. Qed.

Lemma fun_limP (s : {chain A -> B}) f : f \In s -> f <== fun_lim s.
Proof. by move=>H t; apply: limP; exists f. Qed.

Lemma fun_limM (s : {chain A -> B}) f :
        (forall y, y \In s -> y <== f) -> fun_lim s <== f.
Proof. by move=>H1 t; apply: limM=>y [g ->] H2; apply: H1. Qed.

Definition funCPOMixin := CPOMixin fun_botP fun_limP fun_limM.
Canonical funCPO := Eval hnf in CPO (A -> B) funCPOMixin.

End FunCPO.
(* dependent functions *)

Section DFunCPO.
Variable (A : Type) (B : A -> cpo).

Let dfun_bot : dfunPoset B := fun x => bot.

Let dfun_lim (s : chain (dfunPoset B)) : dfunPoset B :=
  fun x => 'lim (Image (dapp x) s).

Lemma dfun_botP x : dfun_bot <== x.
Proof. by move=>y; apply: botP. Qed.

Lemma dfun_limP (s : chain _) x : x \In s -> x <== dfun_lim s.
Proof. by move=>H t; apply: limP; exists x. Qed.

Lemma dfun_limM (s : chain _) x :
        (forall y, y \In s -> y <== x) -> dfun_lim s <== x.
Proof. by move=>H1 t; apply: limM=>y [f ->] H2; apply: H1. Qed.

Definition dfunCPOMixin := CPOMixin dfun_botP dfun_limP dfun_limM.
Definition dfunCPO := Eval hnf in CPO (forall x, B x) dfunCPOMixin.

End DFunCPO.

(* Prop *)

Section PropCPO.
Local Notation tp := [poset of Prop].

Let prop_bot : tp := False.

Let prop_lim (s : chain tp) : tp := exists p, p \In s /\ p.

Lemma prop_botP x : prop_bot <== x.
Proof. by []. Qed.

Lemma prop_limP (s : chain tp) p : p \In s -> p <== prop_lim s.
Proof. by exists p. Qed.

Lemma prop_limM (s : chain tp) p :
        (forall q, q \In s -> q <== p) -> prop_lim s <== p.
Proof. by move=>H [r][]; move/H. Qed.

Definition propCPOMixin := CPOMixin prop_botP prop_limP prop_limM.
Canonical propCPO := Eval hnf in CPO Prop propCPOMixin.

End PropCPO.

(* Pred *)

Section PredCPO.
Variable A : Type.

Definition predCPOMixin := funCPOMixin A propCPO.
Canonical predCPO := Eval hnf in CPO (Pred A) predCPOMixin.

End PredCPO.

(* every complete lattice is a cpo *)

Section LatticeCPO.
Variable A : lattice.
Local Notation tp := (Lattice.poset A).

Let lat_bot : tp := lbot.

Let lat_lim (s : Pred tp) : tp := sup s.

Lemma lat_botP x : lat_bot <== x.
Proof. by apply: supM. Qed.

Lemma lat_limP (s : chain tp) x : x \In s -> x <== lat_lim s.
Proof. by apply: supP. Qed.

Lemma lat_limM (s : chain tp) x :
        (forall y, y \In s -> y <== x) -> lat_lim s <== x.
Proof. by apply: supM. Qed.

Definition latCPOMixin := CPOMixin lat_botP lat_limP lat_limM.
Definition latCPO := Eval hnf in CPO tp latCPOMixin.

End LatticeCPO.

(* sub-CPO's *)

(* every chain-closed subset of a cpo is a cpo *)

Section AdmissibleClosure.
Variable T : cpo.

Definition chain_closed :=
  [Pred s : Pred T |
     bot \In s /\ forall d : chain T, d <=p s -> 'lim d \In s].

(* admissible closure of s is the smallest closed set containing s *)
(* basically extends s to include the sups of chains *)
Definition chain_closure (s : Pred T) :=
  [Pred p : T | forall t : Pred T, s <=p t -> chain_closed t -> p \In t].

(* admissible closure contains s *)
Lemma chain_clos_sub (s : Pred T) : s <=p chain_closure s.
Proof. by move=>p H1 t H2 H3; apply: H2 H1. Qed.

(* admissible closure is smallest *)
Lemma chain_clos_min (s : Pred T) t :
        s <=p t -> chain_closed t -> chain_closure s <=p t.
Proof. by move=>H1 H2 p; move/(_ _ H1 H2). Qed.

(* admissible closure is closed *)
Lemma chain_closP (s : Pred T) : chain_closed (chain_closure s).
Proof.
split; first by move=>t _ [].
move=>d H1 t H3 H4; move: (chain_clos_min H3 H4)=>{}H3.
by case: H4=>_; apply=>// x; move/H1; move/H3.
Qed.

Lemma chain_clos_idemp (s : Pred T) :
        chain_closed s -> chain_closure s =p s.
Proof.
move=>p; split; last by apply: chain_clos_sub.
by apply: chain_clos_min=>// /chain_closP.
Qed.

Lemma chain_clos_mono (s1 s2 : Pred T) :
        s1 <=p s2 -> chain_closure s1 <=p chain_closure s2.
Proof.
move=>H1; apply: chain_clos_min (chain_closP s2)=>p H2.
by apply: chain_clos_sub; apply: H1.
Qed.

(* intersection of admissible sets is admissible *)
Lemma chain_closI (s1 s2 : Pred T) :
       chain_closed s1 -> chain_closed s2 -> chain_closed (PredI s1 s2).
Proof.
move=>[H1 S1][H2 S2]; split=>// d H.
by split; [apply: S1 | apply: S2]=>// x; case/H.
Qed.

End AdmissibleClosure.

Arguments chain_closed {T}.
Prenex Implicits chain_closed.

(* diagonal of an admissible set of pairs is admissible *)
Lemma chain_clos_diag (T : cpo) (s : Pred (T * T)) :
        chain_closed s -> chain_closed [Pred t : T | (t, t) \In s].
Proof.
case=>B H1; split=>// d H2; move: (H1 [chain of Image diag d]).
rewrite /= limX /= /lim /=.
rewrite !(limE (comp_chainE _ _ _)) !(limE (id_chainE _ _)) //.
by apply; case=>x1 x2 [z ->]; apply: H2.
Qed.

Section SubCPO.
Variables (D : cpo) (s : Pred D) (C : chain_closed s).
Local Notation tp := (subPoset s).

Let sub_bot := exist _ (@bot D) (proj1 C).

Lemma sub_botP (x : tp) : sval sub_bot <== sval x.
Proof. by apply: botP. Qed.

Lemma sub_limX (u : chain tp) : 'lim (Image sval u) \In s.
Proof. by case: C u=>P H u; apply: (H)=>t [[y H1 ->]]. Qed.

Definition sub_lim (u : chain tp) : tp :=
  exist _ ('lim (Image sval u)) (sub_limX u).

Lemma sub_limP (u : chain tp) x : x \In u -> x <== sub_lim u.
Proof. by move=>H; apply: limP; exists x. Qed.

Lemma sub_limM (u : chain tp) x :
        (forall y, y \In u -> y <== x) -> sub_lim u <== x.
Proof. by move=>H; apply: limM=>y [z ->]; apply: H. Qed.

Definition subCPOMixin := CPOMixin sub_botP sub_limP sub_limM.
Definition subCPO := Eval hnf in CPO {x : D | x \In s} subCPOMixin.

End SubCPO.

(***********************************************)
(* Continuity and Kleene's fixed point theorem *)
(***********************************************)

Lemma lim_mono (D : cpo) (s1 s2 : chain D) :
        s1 <=p s2 -> lim s1 <== lim s2.
Proof. by move=>H; apply: limM=>y /H/limP. Qed.

Lemma lim_liftE (D : cpo) (s : chain D) :
        'lim s = 'lim (lift_chain s).
Proof.
apply: poset_asym; apply: limM=>y H; first by apply: limP; right.
by case: H; [move=>-> | apply: limP].
Qed.

(* applied lim equals the lim of applications *)
(* ie., part of continuity of application *)
(* but is so often used, I give it a name *)

Lemma lim_appE A (D : cpo) (s : {chain A -> D}) (x : A) :
        'lim s x = 'lim (Image (app x) s).
Proof. by []. Qed.

Lemma lim_dappE A (D : A -> cpo) (s : chain (dfunCPO D)) (x : A) :
        'lim s x = 'lim (Image (dapp x) s).
Proof. by []. Qed.

(************************)
(* continuous functions *)
(************************)

Definition continuous (D1 D2 : cpo) (f : {monotone D1 -> D2}) :=
  forall s : chain D1, f ('lim s) = 'lim (Image f s).

Module ContinuousFun.
Section ClassDef.
Variables D1 D2 : cpo.

Structure mixin_of (f : {monotone D1 -> D2}) := Mixin {
  _ : continuous f}.

Record class_of (cfun : D1 -> D2) := Class {
  base : monotone cfun;
  mixin : mixin_of (Monotone cfun base)}.

Local Coercion base : class_of >-> monotone.

Structure continuous_type : Type := Pack {cfun; _ : class_of cfun}.
Local Coercion cfun : continuous_type >-> Funclass.

Variables (cfun : D1 -> D2) (cT : continuous_type).
Definition class := let: Pack _ c  as cT' := cT return class_of cT' in c.
Let xT := let: Pack T _ := cT in T.
Notation xclass := (class : class_of xT).
Definition clone c of phant_id class c := @Pack cfun c.
Definition pack (b0 : monotone cfun) (m0 : mixin_of (Monotone cfun b0)) :=
  fun m & phant_id m0 m => Pack (@Class cfun b0 m).

Definition monotone_fun := Monotone cT xclass.

Definition continuous_for of phant (D1 -> D2) := continuous_type.

End ClassDef.

Module Exports.
Coercion base : class_of >-> monotone.
Coercion mixin : class_of >-> mixin_of.
Coercion cfun : continuous_type >-> Funclass.
Coercion monotone_fun : continuous_type >-> monotone_type.
Canonical monotone_fun.
Notation continuous_type := continuous_type.
Notation contMixin := (@Mixin _ _).
Notation Continuous F m := (@pack _ _ F _ _ (Mixin m) id).

Notation "{ 'continuous' fT }" := (continuous_for (Phant fT))
  (at level 0, format "{ 'continuous'  fT }") : form_scope.
Notation "[ 'continuous' 'of' f ]" :=
  (@clone _ _ f _ _ _ id)
  (at level 0, format "[ 'continuous'  'of'  f ]") : form_scope.

Section Lemmas.
Variables (D1 D2 : cpo) (f : {continuous D1 -> D2}).

Lemma contE (s : chain D1) : f ('lim s) = 'lim (Image f s).
Proof. by case: f=>g [b][]; apply. Qed.


End Lemmas.
End Exports.
End ContinuousFun.

Export ContinuousFun.Exports.

Module Kleene.
(* just for this proof, we want that nat is a poset under <= *)
(* in other scopes, we'll use the equality as ordering *)
Section NatPoset.

Lemma nat_refl x : x <= x. Proof. by []. Qed.

Lemma nat_asym x y : x <= y -> y <= x -> x = y.
Proof. by move=>H1 H2; apply: anti_leq; rewrite H1 H2. Qed.

Lemma nat_trans x y z : x <= y -> y <= z -> x <= z.
Proof. by apply: leq_trans. Qed.

Definition natPosetMixin := PosetMixin nat_refl nat_asym nat_trans.
Canonical natPoset := Eval hnf in Poset nat natPosetMixin.

End NatPoset.

Section NatChain.

Definition nats : Pred nat := PredT.

Lemma nat_chain_axiom : chain_axiom nats.
Proof.
split=>[|x y _ _]; first by exists 0.
rewrite /Poset.leq /= [y <= x]leq_eqVlt.
by case: leqP; [left | rewrite orbT; right].
Qed.

Canonical nats_chain := Eval hnf in Chain nats nat_chain_axiom.

End NatChain.

Section Kleene.
Variables (D : cpo) (f : {continuous D -> D}).

Fixpoint pow m := if m is n.+1 then f (pow n) else bot.

Lemma pow_mono_ax : monotone pow.
Proof.
move=>m n; elim: n m=>[|n IH] m /=; first by case: m.
rewrite {1}/Poset.leq /= leq_eqVlt ltnS.
case/orP; first by move/eqP=>->.
move/IH=>{IH} H; apply: poset_trans H _.
by elim: n=>[|n IH] //=; apply: monotoneP IH.
Qed.

Canonical pow_mono := Eval hnf in Monotone pow pow_mono_ax.

Lemma reindex : Image pow nats =p lift (Image f (Image pow nats)).
Proof.
move=>x; split.
- case; case=>[|n] -> _; first by left.
  by right; exists (pow n)=>//; exists n.
case=>/=; first by move=>->; exists 0.
by case=>y -> [n ->]; exists n.+1.
Qed.

Definition kleene_lfp := 'lim (Image pow nats).

Lemma kleene_lfp_fixed : f kleene_lfp = kleene_lfp.
Proof. by rewrite /kleene_lfp contE /= (limE reindex) /= lim_liftE. Qed.

Lemma kleene_lfp_least x : f x = x -> kleene_lfp <== x.
Proof.
move=>H; apply: limM=>y [n ->] _.
by elim: n=>[|n IH] //=; rewrite -H; apply: monotoneP IH.
Qed.

End Kleene.

Lemma kleene_lfp_mono (D : cpo) (f1 f2 : {continuous D -> D}) :
        (f1 : D -> D) <== f2 -> kleene_lfp f1 <== kleene_lfp f2.
Proof.
move=>H; apply: limM=>x [n ->] X.
have N : forall n, pow f1 n <== pow f2 n.
- by elim=>//= m /(monotoneP f2); apply: poset_trans (H _).
by apply: poset_trans (N n) _; apply: limP; exists n.
Qed.

Module Exports.
Notation kleene_lfp := kleene_lfp.
Notation kleene_lfp_fixed := kleene_lfp_fixed.
Notation kleene_lfp_least := kleene_lfp_least.
Notation kleene_lfp_mono := kleene_lfp_mono.
End Exports.

End Kleene.

Export Kleene.Exports.

(**********************************)
(* Continuity of common functions *)
(**********************************)

Lemma id_cont_ax (D : cpo) : continuous [monotone of @idfun D].
Proof. by move=>d; rewrite (limE (id_chainE _ _)). Qed.
Canonical id_cont D := Eval hnf in Continuous idfun (@id_cont_ax D).

Lemma const_cont_ax (D1 D2 : cpo) (y : D2) :
        continuous [monotone of (@const_fun D1 D2 y)].
Proof.
move=>s; apply: poset_asym.
- by apply: limP; case: s=>[p][[d H1] H2]; exists d.
by apply: limM=>_ [x ->].
Qed.

Canonical const_cont (D1 D2 : cpo) y :=
  Eval hnf in Continuous (const_fun y) (@const_cont_ax D1 D2 y).

Section ContCompose.
Variables D1 D2 D3 : cpo.
Variable (f1 : {continuous D2 -> D1}) (f2 : {continuous D3 -> D2}).

Lemma comp_cont_ax : continuous [monotone of f1 \o f2].
Proof. by move=>s; rewrite /= !contE (limE (comp_chainE _ _ _)). Qed.

Canonical comp_cont :=  Eval hnf in Continuous (f1 \o f2) comp_cont_ax.
End ContCompose.

Lemma fst_cont_ax (D1 D2 : cpo) : continuous [monotone of @fst D1 D2].
Proof. by []. Qed.
Canonical fst_cont D1 D2 := Eval hnf in Continuous fst (@fst_cont_ax D1 D2).

Lemma snd_cont_ax (D1 D2 : cpo) : continuous [monotone of @snd D1 D2].
Proof. by []. Qed.
Canonical snd_cont D1 D2 := Eval hnf in Continuous snd (@snd_cont_ax D1 D2).

Lemma diag_cont_ax (D : cpo) : continuous [monotone of @diag D].
Proof.
move=>d; apply: poset_asym; split=>/=;
by rewrite (limE (comp_chainE _ _ _)) (limE (id_chainE _ _)).
Qed.
Canonical diag_cont D := Eval hnf in Continuous diag (@diag_cont_ax D).

Lemma app_cont_ax A (D : cpo) x : continuous [monotone of @app A D x].
Proof. by []. Qed.

Canonical app_cont A (D : cpo) x :=
  Eval hnf in Continuous (@app A D x) (@app_cont_ax A D x).

Definition dapp2 A (B : A -> cpo) (x : A) := fun f : dfunCPO B => f x.

Lemma dapp2_mono_ax A (D : A -> cpo) x : monotone (@dapp2 A D x).
Proof. by move=>f1 f2; apply. Qed.

Canonical dapp2_mono A D x :=
  Eval hnf in Monotone (@dapp2 A D x) (@dapp2_mono_ax A D x).

Lemma dapp2_cont_ax A (D : A -> cpo) x :
        continuous [monotone of @dapp2 A D x].
Proof. by move=>s; apply: limE. Qed.

Canonical dapp2_cont A D x :=
  Eval hnf in Continuous (@dapp2 A D x) (@dapp2_cont_ax A D x).

Section ProdCont.
Variables S1 S2 T1 T2 : cpo.
Variables (f1 : {continuous S1 -> T1}) (f2 : {continuous S2 -> T2}).

Lemma prod_cont_ax : continuous [monotone of f1 \* f2].
Proof.
move=>d; apply: poset_asym; split=>/=;
by rewrite contE !(limE (comp_chainE _ _ _)); apply: lim_mono.
Qed.

Canonical prod_cont := Eval hnf in Continuous (f1 \* f2) prod_cont_ax.
End ProdCont.
