From Coq Require Import ssreflect ssrbool ssrfun.
From mathcomp Require Import ssrnat.
From fcsl Require Import axioms pred prelude domain pcm state world.
From fcsl Require Import morphism notation schedule process always.
From fcsl Require Import options.

(**********************)
(* Helper definitions *)
(**********************)

(** pre-condition is a predicate on the states of a resource [V] *)
Definition pre V := st V -> Prop.
(** post-condition is a relation
    between the return result of type [A] and the states of a resource [V] *)
Definition post V A := A -> st V -> Prop.
(** specification is a conjunction of a precondition and a postcondition
    possibly parameterized by a logical variable of type [G] *)
Definition spec V G A := G -> prod (pre V) (post V A).

Definition stab V (R : st V -> Prop) s :=
  forall s', osteps V s s' -> R s'.
Definition mstab V W X (tau : morph X V W) (Q : X -> st W -> Prop) x sw :=
  forall x' sw', mxsteps tau x sw x' sw' -> Q x' sw'.
Definition hat V W X (f : morph X V W) P x sw :=
  exists sv, [/\ [morph f] x sw sv & P sv].
Definition idstab V (Q : st V -> Prop) s :=
  forall s', msteps (@world.id_morph V tt) s s' -> Q s'.

(*************************************************************)
(* List of inference rules, with vrf predicate kept abstract *)
(*************************************************************)

Module Type VrfSig.

Parameter ST : resource -> Type -> Type.
Parameter ret : forall V A, A -> ST V A.
Parameter bnd : forall V A1 A2, ST V A1 -> (A1 -> ST V A2) -> ST V A2.
Parameter par : forall V A1 A2, ST V A1 -> ST V A2 -> ST V (A1 * A2).
Parameter atomic : forall V A, action V A -> ST V A.
Parameter morph_app : forall V W X A, morph X V W -> X -> ST V A -> ST W A.
Arguments ret [V][A] x.

Parameter vrf : forall V A, ST V A -> post V A -> st V -> Prop.

Parameter vrf_post_inv : forall V A (c : ST V A) (Q1 Q2 : post V A) I s,
  coh V s -> idstab I s -> (forall x s, idstab I s -> Q1 x s -> Q2 x s) ->
  vrf c Q1 s -> vrf c Q2 s.
Parameter vrf_ret : forall V A (v : A) (Q : A -> st V -> Prop) s,
  coh V s -> stab (Q v) s -> vrf (ret v) Q s.
Parameter vrf_bnd : forall V A1 A2 (c1 : ST V A1) (c2 : A1 -> ST V A2) Q s,
  coh V s -> vrf c1 (fun x => vrf (c2 x) Q) s -> vrf (bnd c1 c2) Q s.
Parameter vrf_par : forall V A1 A2 (c1 : ST V A1) (c2 : ST V A2) Q1 Q2 s,
  coh V s -> star (vrf c1 Q1) (vrf c2 Q2) s ->
  vrf (par c1 c2) (fun v : A1 * A2 => star (Q1 v.1) (Q2 v.2)) s.
Parameter vrf_frame : forall V A (c : ST V A) Q1 Q2 s,
  coh V s -> star (vrf c Q1) (stab Q2) s -> vrf c (fun v => star (Q1 v) Q2) s.
Parameter vrf_atm : forall V A (a : action V A) Q s,
  coh V s -> stab (fun s => exists v s',
  eff a v s s' /\ stab (Q v) s') s -> vrf (atomic a) Q s.
Parameter vrf_morph : forall V W X A (f : morph X V W) (sim : simulation f)
                        (c : ST V A) x sw Q,
  coh W sw -> hat f (vrf c Q) x sw -> sim x sw ->
  vrf (morph_app f x c) (fun r sw' =>
    exists x', hat f (Q r) x' sw' /\ sim x' sw') sw.

(* for ease of proof, it turns out we actually want to work *)
(* with a somewhat changed definition of vrf, which is preconditioned *)
(* by coh. That version, however, if exported directly, needs one more *)
(* axiom. So, to keep axioms at the minimum, we introduce the following *)
(* and then prove lemmas about it from the axioms *)

Definition vrf2 V A (c : ST V A) Q s := coh V s -> vrf c Q s.

Definition has_spec V G A (pq : spec V G A) (c : ST V A) :=
  forall s g, (pq g).1 s -> coh V s -> vrf2 c (pq g).2 s.

Structure STbin V G A (pq : spec V G A) := do' {
  prog_of :> ST V A;
  proof_of : has_spec pq prog_of}.

Arguments STbin V G [A] pq.

(* some notation *)

(*
Notation "x '<--' c1 ';' c2" := (bnd c1 (fun x => c2))
  (at level 81, right associativity).
Notation "' x '<--' c1 ';' c2" := (bnd c1 (fun x => c2))
  (at level 81, right associativity, x strict pattern).
Notation "c1 ';;' c2" := (bnd c1 (fun _ => c2))
  (at level 81, right associativity).
*)

Notation "x '<--' c1 ';' c2" := (bnd c1 (fun x => c2))
  (at level 201, right associativity, format
  "'[v' x  '<--'  c1 ';' '//' c2 ']'").
Notation "' x '<--' c1 ';' c2" := (bnd c1 (fun x => c2))
  (at level 201, right associativity, x strict pattern, format
  "'[v' ' x  '<--'  c1 ';' '//' c2 ']'").
Notation "c1 ';;' c2" := (bnd c1 (fun _ => c2))
  (at level 201, right associativity).

(* The implementation of this fixed point constructor takes *)
(* monotone closure of the function *)
(* this is a hack, to avoid proving monotonicity all the time *)
(* the later is needed because in the shallow embedding, we can *)
(* escape the abstraction, and implement functions which are not monotone *)
(* We do prove below that all constructors are monotone, so if we stick *)
(* with the provided constructions only, monotonicity is not needed *)
Parameter ffix : forall V G A (B : A -> Type) (pq : forall x, spec V G (B x)),
  ((forall x : A, STbin V G (pq x)) -> forall x : A, STbin V G (pq x)) ->
    forall x : A, STbin V G (pq x).

(*
(* the next fixed point constructor requires explicit proof of monotonicity *)
(* we don't use it in code, but I put it here for experimentation *)
(* it also requires exposing that STbin is a poset *)
(* we also expose lattice structure *)

Parameter STbinPosetMix : forall V G A (pq : spec V G A),
  Poset.mixin_of (STbin V G pq).
Canonical STbinPoset V G A pq :=
  Eval hnf in Poset (STbin V G pq) (@STbinPosetMix V G A pq).

Parameter STbinLatticeMix : forall V G A (pq : spec V G A),
  Lattice.mixin_of (STbinPoset pq).
Canonical STbinLattice V G A pq :=
  Eval hnf in Lattice (STbin V G pq) (@STbinLatticeMix V G A pq).

Parameter ffix2 : forall V G A (B : A -> Type) (pq : forall x, spec V G (B x))
  (f : (dfunPoset (fun x => STbinPoset (pq x))) ->
        dfunPoset (fun x => STbinPoset (pq x))),
   monotone f -> forall x : A, STbin V G (pq x).
*)
End VrfSig.


(*****************************)
(* Inference rules are sound *)
(*****************************)

Module Vrf : VrfSig.

(* Semantics of programs as sets of trees. We require that each *)
(* program's denotation contains at least one element -- the unfinished *)
(* tree. This way, an infinite loop is not modeled by an empty set, but *)
(* by a set containing approximating finite trees. *)

Structure ST' V A := Prog' {
  set_of :> tree V A -> Prop;
  _ : set_of Unfinished}.

(* assignments to make Coq's typechecking of modules happy *)
Definition ST := ST'.
Definition Prog V A t pf : ST V A := @Prog' V A t pf.

Arguments Prog [V A] _ _.

(* admit \In notation *)
Canonical prog_PredType W A := Eval hnf in mkPredType (@set_of W A).

Lemma stE V A (T1 T2 : ST V A) :
        T1 = T2 <-> set_of T1 = set_of T2.
Proof.
split=>[->//|]; case: T1 T2=>m1 H1 [m2 H2] /= E.
by rewrite -E in H2 *; rewrite (pf_irr H1 H2).
Qed.

(*********************)
(* prog constructors *)
(*********************)

Definition ret_set V A (x : A) t := t = Unfinished \/ t = @Ret V A x.
Definition ret V A (x : A) := Prog (@ret_set V A x) (or_introl (erefl _)).

Arguments ret [V][A] x.

Definition bnd_set V A B (c1 : ST V A) (c2 : A -> ST V B) t :=
  t = Unfinished \/ exists t', t \In tcat t' c2 /\ t' \In c1.
Definition bnd V A B c1 c2 :=
  Prog (@bnd_set V A B c1 c2) (or_introl (erefl _)).

Definition par_set V A B (c1 : ST V A) (c2 : ST V B) t :=
  t = Unfinished \/ exists t1 t2, [/\ t = Par t1 t2 Ret, t1 \In c1 & t2 \In c2].
Definition par V A B c1 c2 :=
  Prog (@par_set V A B c1 c2) (or_introl (erefl _)).

Definition atomic_set V A (a : action V A) t := t = Unfinished \/ t = Act a.
Definition atomic V A a := Prog (@atomic_set V A a) (or_introl (erefl _)).

Definition morphapp_set V W X A (tau : morph X V W) (x : X) (c : ST V A) t :=
  t = Unfinished \/ exists t', t' \In c /\ t = Morf x tau t'.
Definition morph_app V W X A tau x c :=
  Prog (@morphapp_set V W X A tau x c) (or_introl (erefl _)).

(* fix will be defined later, once we introduce specs *)

(* vrf computes weakest coherent preconditions *)
Definition vrf V A (c : ST V A) Q s := forall t, c t -> after s t Q.

(* weakening post *)
(* we also allow weakening post with an invariant property I for V *)
(* even if I is not made part of V's coh *)
Lemma vrf_post_inv V A (c : ST V A) (Q1 Q2 : post V A) I s :
        coh V s -> idstab I s -> (forall x s, idstab I s -> Q1 x s -> Q2 x s) ->
        vrf c Q1 s -> vrf c Q2 s.
Proof.
move=>_ H1 H2 H t /H; apply: aft_invst=>v s' C' M.
by apply: H2=>x M'; apply: H1; apply: iter_trans M M'.
Qed.

(* coherence is one of the invariant properties *)
(* don't export; reprove outside from vrf_post_inv *)
Lemma vrf_post V A (c : ST V A) (Q1 Q2 : post V A) s :
        coh V s ->
        (forall x s, coh V s -> Q1 x s -> Q2 x s) ->
        vrf c Q1 s -> vrf c Q2 s.
Proof.
move=>C H X; have I : idstab (coh V) s by move=>s' /mcoh; apply.
by apply: vrf_post_inv C I _ X=>x s' I; apply: H (I _ _).
Qed.

(* pre's and post's are stable *)

(* don't export for now *)
Lemma vrf_pre_stable V A (c : ST V A) Q s :
        vrf c Q s -> coh V s -> stab (vrf c Q) s.
Proof. by move=>H C s' O t /H X; apply: aft_imp (alw_envs X O). Qed.

(* don't export for now *)
Lemma vrf_post_stable V A c (Q : post V A) s :
        coh V s -> vrf c Q s -> vrf c (fun x => stab (Q x)) s.
Proof.
move=>C H t /H /alw_envsq; apply: alw_imp=>s' p Hs.
by split; [case: (Hs _ iter_refl) | move=>v E m /Hs [_] /(_ _ E)].
Qed.

(* rule of conjunction; also don't export for now *)
Lemma vrf_conj V A c (Q1 Q2 : post V A) s :
        coh V s ->
        vrf c Q1 s -> vrf c Q2 s -> vrf c (fun r s => Q1 r s /\ Q2 r s) s.
Proof.
move=>C H1 H2 t X; apply: alw_imp (alw_and (H1 t X) (H2 t X)).
by move=>{H1 H2 C t X} s t [[C H1]][_ H2]; split=>// v E; intuition.
Qed.

(* ret *)
Lemma vrf_ret V A (v : A) (Q : A -> st V -> Prop) s :
        coh V s -> stab (Q v) s -> vrf (ret v) Q s.
Proof.
move=>C S t [] ->.
- by apply: alw_unfin=>m O; split=>//; apply: ocoh O C.
apply: alw_ret=>m O; split=>[|w [<-]]; first by apply: ocoh O C.
by apply: S O.
Qed.

(* Hoare-style lemma for sequential composition; not exported *)
Lemma vrf_hoare V A1 A2 (c1 : ST V A1) (c2 : A1 -> ST V A2) Q1 Q2 s :
        coh V s ->
        vrf c1 Q1 s -> (forall y j, coh V j -> Q1 y j -> vrf (c2 y) Q2 j) ->
        vrf (bnd c1 c2) Q2 s.
Proof.
move=>C H1 H2 t [->|]; first by apply: alw_unfin=>m /ocoh/(_ C).
case=>t' [H3 /H1 H4]; apply: aft_bnd H3 (aft_imp _ H4).
by move=>v s' C' /H2; apply.
Qed.

(* Dijkstra-style lemma for sequential composition *)
(* this is the one exported *)
Lemma vrf_bnd V A1 A2 (c1 : ST V A1) (c2 : A1 -> ST V A2) Q s :
        coh V s ->
        vrf c1 (fun x => vrf (c2 x) Q) s -> vrf (bnd c1 c2) Q s.
Proof. by move=>C H; apply: (vrf_hoare (Q1:=(fun x=>vrf (c2 x) Q))). Qed.

(* parallel composition *)
Lemma vrf_par V A1 A2 (c1 : ST V A1) (c2 : ST V A2) Q1 Q2 s :
        coh V s ->
        star (vrf c1 Q1) (vrf c2 Q2) s ->
        vrf (par c1 c2) (fun v => star (Q1 v.1) (Q2 v.2)) s.
Proof.
move=>C [s1][s2][S H1 H2] t.
have C1 : coh V s1 by apply/(coh_gbL _ S).
have C2 : coh V s2 by apply/(coh_gbR _ S).
case=>[->|]; first by apply: alw_unfin=>m /ocoh /(_ C).
case=>t1 [t2][->{t}] /H1 {}H1 /H2 {}H2.
by apply: aft_par2; exists s1, s2.
Qed.

Arguments vrf_par [V][A1][A2][c1][c2] Q1 Q2 [s] _.

(* frame *)
Lemma vrf_frame V A (c : ST V A) Q1 Q2 s :
        coh V s ->
        star (vrf c Q1) (stab Q2) s ->
        vrf c (fun v => star (Q1 v) Q2) s.
Proof.
move=>C [s1][s2][S H1 H2] t; have C1 : coh V s1 by apply/(coh_gbL _ S).
move/H1 {H1}; move/(aft_frame2 S).
apply: aft_imp=>v s' C' [s1'][s2'][S' H' O].
by exists s1', s2'; split=>//; apply: H2 O.
Qed.

(* atomic *)
Lemma vrf_atm V A (a : action V A) Q s :
        coh V s ->
        stab (fun s => exists v s', eff a v s s' /\ stab (Q v) s') s ->
        vrf (atomic a) Q s.
Proof.
move=>C H t [] ->; first by apply: alw_unfin=>s' /ocoh /(_ C).
apply: alw_act=>s' Os; move: (ocoh Os C)=>C'; case: (H _ Os)=>v [x][I X].
have Cx : coh V x by case: I=>tv [_ H']; apply: icoh H' C'.
split=>//; first by exists x, v.
move=>??? /(eff_func C' I) [<-<-] Ox; move: (ocoh Ox Cx).
by split=>// _ [<-]; apply: X.
Qed.

(* lifting by morphism *)
Lemma vrf_morph V W X A (f : morph X V W) (sim : simulation f) (c : ST V A) x sw Q :
        coh W sw -> hat f (vrf c Q) x sw -> sim x sw ->
        vrf (morph_app f x c)
            (fun r sw' => exists x', hat f (Q r) x' sw' /\ sim x' sw') sw.
Proof.
move=>Cw [sv][T H] M t; move: (morph_coh Cw T)=>Cv.
case=>[->|[t']]; first by apply: alw_unfin=>? /ocoh/(_ Cw).
case=>/H {}H ->; apply: aft_imp (aft_morph Cw T M H).
move=>r s' Cw' [x'][sv'][T' S' H'] M'.
by exists x'; split=>//; exists  sv'.
Qed.

(* now the part with specs and fixed points *)

Definition vrf2 V A (c : ST V A) Q s := coh V s -> vrf c Q s.

Definition has_spec V G A (pq : spec V G A) (c : ST V A) :=
  forall s g, (pq g).1 s -> coh V s -> vrf2 c (pq g).2 s.

Structure STbin V G A (pq : spec V G A) := do' {
  prog_of :> ST V A;
  proof_of : has_spec pq prog_of}.

(* lattice structure on prog *)

Section STLattice.
Variables (V : resource) (A : Type).

Definition leq (c1 c2 : ST V A) := set_of c1 <== set_of c2.

Lemma leq_refl c : leq c c.
Proof. by []. Qed.

Lemma leq_trans c1 c2 c3 : leq c1 c2 -> leq c2 c3 -> leq c1 c3.
Proof. by move=>H1 H2 p x; apply/H2/H1. Qed.

Lemma leq_asym c1 c2 : leq c1 c2 -> leq c2 c1 -> c1 = c2.
Proof.
move: c1 c2=>[m1 N1][m2 N2]; rewrite /leq /= => H1 H2.
by apply/stE; apply: poset_asym H1 H2.
Qed.

Definition sup_set (cs : Pred (ST V A)) t :=
  t = Unfinished \/ exists c : ST V A, t \In c /\ c \In cs.
Definition sup_prg cs := Prog (sup_set cs) (or_introl (erefl _)).

(* sup is an upper bound *)
Lemma sup_supB cs c : c \In cs -> leq c (sup_prg cs).
Proof. by move=>H1 x H2; right; exists c. Qed.

(* sup is the least upper bound *)
Lemma sup_supL cs c : (forall x, x \In cs -> leq x c) -> leq (sup_prg cs) c.
Proof. by case: c=>m N H /= t [->//|[T][H2]] /H; apply. Qed.

Definition stPosetMix := PosetMixin leq_refl leq_asym leq_trans.
Canonical stPoset := Eval hnf in Poset (ST V A) stPosetMix.

Definition stLatticeMix := LatticeMixin sup_supB sup_supL.
Canonical stLattice := Eval hnf in Lattice (ST V A) stLatticeMix.

End STLattice.

(* lattice structure on STbin *)

Section STbinLattice.
Variables (V : resource) (G A : Type) (pq : spec V G A).
Notation STbin s := (@STbin V G A s).

Definition bleq (c1 c2 : STbin pq) := prog_of c1 <== prog_of c2.

Lemma bleq_refl c : bleq c c.
Proof. by []. Qed.

Lemma bleq_trans c1 c2 c3 : bleq c1 c2 -> bleq c2 c3 -> bleq c1 c3.
Proof. by apply: poset_trans. Qed.

Lemma bleq_asym c1 c2 : bleq c1 c2 -> bleq c2 c1 -> c1 = c2.
Proof.
case: c1 c2=>c1 pf1 [c2 pf2] H1 H2; move: (poset_asym H1 H2)=>/= E.
by rewrite -E in pf2 H1 H2 *; rewrite (pf_irr pf1 pf2).
Qed.

Definition bsup_set (cs : Pred (STbin pq)) t :=
  t = Unfinished \/ exists c : STbin pq, t \In prog_of c /\ c \In cs.
Definition bsup_prog cs := Prog (bsup_set cs) (or_introl (erefl _)).

Lemma bsup_spec cs : has_spec pq (bsup_prog cs).
Proof.
move=>g s H C _ t [->|[c [Ht _]]].
- by apply: alw_unfin=>s' /ocoh/(_ C).
by case: c Ht=>/= p /(_ _ _ H C); apply.
Qed.

Definition bsup cs := do' (@bsup_spec cs).

(* sup is an upper bound *)
Lemma bsup_supB cs c : c \In cs -> bleq c (bsup cs).
Proof. by move=>H1 x H2; right; exists c. Qed.

(* sup is the least upper bound *)
Lemma bsup_supL cs c : (forall x, x \In cs -> bleq x c) -> bleq (bsup cs) c.
Proof. by case: c; case=>m N H /= H1 t [->//|][T][H2]; move/H1; apply. Qed.

Definition STbinPosetMix := PosetMixin bleq_refl bleq_asym bleq_trans.
Canonical STbinPoset := Eval hnf in Poset (STbin pq) STbinPosetMix.

Definition STbinLatticeMix := LatticeMixin bsup_supB bsup_supL.
Canonical STbinLattice := Eval hnf in Lattice (STbin pq) STbinLatticeMix.

End STbinLattice.

Section Fix.
Variables (V : resource) (G A : Type) (B : A -> Type).
Variable pq : forall x, spec V G (B x).
Notation tp := (forall x, STbin (pq x)).
Notation lat := (dfunLattice (fun x => [lattice of STbin (pq x)])).
Variable (f : tp -> tp).

(* fixed point constructor over monotone closure *)
Definition f' (c : lat) := sup [Pred t : lat | exists c', c' <== c /\ t = f c'].
Definition ffix := tarski_lfp f'.

(* fixed point constructor which requires explicit proof of monotonicity *)
Definition monotone' := monotone (f : lat -> lat).
Definition ffix2 (pf : monotone') : tp := tarski_lfp (f : lat -> lat).
End Fix.

Arguments ffix2 [V G A B pq] f pf x.

Notation "'Do' e" := (@do' _ _ _ _ e _) (at level 80).
Notation "x '<--' c1 ';' c2" := (bnd c1 (fun x => c2))
  (at level 81, right associativity).
Notation "c1 ';;' c2" := (bnd c1 (fun _ => c2))
  (at level 81, right associativity).

(* Monotonicity of the constructors *)

(* ret is monotone only on types A that are small, i.e., A can't have an ST. *)
(* Thus, the lattice on A is wrt. equality, so we just take v1 = v2 *)
(* The model won't allow working with types A which contain ST *)
(* on which the lattice would be more complicated (i.e., the subset lattice). *)
(* Trying to work with such will raise universe inconsistency. *)

(* Now, that said, if we wanted ret to be monotone in that general case *)
(* the denotational semantics of ret should change. We should not just denote *)
(* it by an Unfinished and the trivial action returning the value v, but we should *)
(* have a bunch of trees; ret x for each element x <= v in the latice A *)
(* To make this all work smoothly, one should require that the denotation of v *)
(* is saturated in the following sense: if x \in [[v]] and y <= x then y \in [[v]] *)
(* i think these are called downward-closed sets. It seems pretty basic. *)

(* But for us here, since we can't nest ST types, and only return values from simple *)
(* types where the lattice is the identity (that's not even a lattice, since it has *)
(* no bottom element), the ownward closed sets include only the given element *)
(* In a very precise sense, taking the monotone closure of the function before doing *)
(* the lfp computation, precisely corresponds to adding the elements in a downward-closed sets *)
Lemma ret_mono V A (v1 v2 : A) : v1 = v2 -> @ret V A v1 <== @ret V A v2.
Proof. by move=>->. Qed.

Lemma do_mono V G A (e1 e2 : ST V A) (s : spec V G A)
        (pf1 : has_spec s e1) (pf2 : has_spec s e2) :
        e1 <== e2 -> @do' _ _ _ _ e1 pf1 <== @do' _ _ _ _ e2 pf2.
Proof.  by []. Qed.

Lemma bnd_mono V A B (e1 e2 : ST V A) (k1 k2 : A -> ST V B) :
        e1 <== e2 -> k1 <== k2 -> bnd e1 k1 <== bnd e2 k2.
Proof.
move=>pf1 pf2 t [->|[t1][X2 X1]]; [left | right]=>//.
exists t1; split; last by apply: pf1 X1.
by case: X2=>t2 [Y1 Y2]; exists t2; split=>// x; apply: pf2 (Y2 _).
Qed.

Lemma par_mono V A1 A2 (e1 k1 : ST V A1) (e2 k2 : ST V A2) :
        e1 <== k1 -> e2 <== k2 -> par e1 e2 <== par k1 k2.
Proof.
move=>pf1 pf2 t; case; first by left.
case=>t1 [t2][X X1 X2]; right; exists t1, t2.
by split=>//; [apply: pf1 | apply: pf2].
Qed.

(* On actions, the ordering is equality *)
(* hence, we can't have higher-order effectful into actions *)
(* but the size of the model won't allow this anyway, just like with ret *)
Lemma atm_mono V A (a1 a2 : action V A) :
        a1 = a2 -> @atomic V A a1 <== @atomic V A a2.
Proof. by move=>->. Qed.

Lemma morph_mono V W X A (tau : morph X V W) (e1 e2 : ST V A) (x : X) :
        e1 <== e2 -> morph_app tau x e1 <== morph_app tau x e2.
Proof.
move=>pf t; case=>[|[t1][X1 X2]]; [left | right]=>//.
by exists t1; split=>//; apply: pf.
Qed.

Section FixMono.
Variables (V : resource) (G A : Type) (B : A -> Type).
Variable s : forall x, spec V G (B x).
Notation tp := (forall x, STbin V G (s x)).
Notation lat := (dfunLattice (fun x => [lattice of STbin (s x)])).
Variables (f1 f2 : lat -> lat).

Lemma fix_mono : f1 <== f2 -> ffix f1 <== ffix f2.
Proof.
move=>Hf; apply: tarski_lfp_mono'.
- move=>x1 x2 Hx; apply: supM=>z [x][H ->]; apply: supP; exists x.
  by split=>//; apply: poset_trans H Hx.
move=>y; apply: supM=>_ [x][H1 ->].
by apply: poset_trans (Hf x) _; apply: supP; exists x.
Qed.
End FixMono.

(* an example over an empty resource; (i.e. a purely functional program) *)

Fixpoint fib (x : nat) := if x is x'.+1 then x * fib x' else 1.

Definition fibtp := forall x : nat, @STbin emp_res unit nat
  (fun => (fun i => True, fun v m => v = fib x)).

(* for the example, we need a quick lemma for function calls *)
(* this need not be exported, as it follows from the signature *)
(* will reprove it outside of the module *)
Lemma gE V G A (pq : spec V G A) (e : STbin pq) g (Q : A -> st V -> Prop) i :
        coh V i -> (pq g).1 i ->
        (forall v m, (pq g).2 v m -> coh V m -> Q v m) ->
        vrf e Q i.
Proof.
case: e=>e H C /H X Y.
apply: (vrf_post (Q1:=(pq g).2))=>//; last by apply: X.
by move=>*; apply: Y.
Qed.

(* now we can do the example with monotonicity explicitly *)
Program Definition fibx :=
  ffix2 (fun (loop : fibtp) (x : nat) =>
    Do (if x is x'.+1 then
          t <-- loop x';
          ret (x * t)
        else ret 1)) _.
Next Obligation.
move=>loop x i [_] C _; case: x; first by apply: vrf_ret.
by move=>n; apply: vrf_bnd=>//; apply: gE=>// x m -> _; apply: vrf_ret.
Qed.
Next Obligation.
move=>x1 x2 H; case=>// n; apply: do_mono.
by apply: bnd_mono=>//; move: (H n).
Qed.

End Vrf.

Export Vrf.

Definition do2 V G A (pq : spec V G A) (c : ST V A) (pf : has_spec pq c) := locked (do' pf).

Notation "'Do' e" := (@do2 _ _ _ _ e _) (at level 80).


(*****************************************************)
(* Derived definitions and lemmas for easier proving *)
(*****************************************************)

(* It turns out it's a bit more convenient to redefine vrf so that *)
(* it keeps the premise that the state is coherent; just makes it *)
(* easier to thread these premises through the proofs *)

Notation vrf := vrf2.

(* to prove vrf, we can always assume coherence *)
Lemma vrf_coh V A c (Q : post V A) s : (coh V s -> vrf c Q s) -> vrf c Q s.
Proof. by move=>H C; apply: H. Qed.

(* we can drop the coherence assumption now, as it's baked into the *)
(* definition of vrf *)

Lemma vrf_post_inv V A (c : ST V A) (Q1 Q2 : post V A) I s :
        idstab I s -> (forall x s, idstab I s -> Q1 x s -> Q2 x s) ->
        vrf c Q1 s -> vrf c Q2 s.
Proof. by move=>H1 H2 H C; apply: vrf_post_inv C H1 H2 (H C). Qed.


(* special case of vrf_post_inv when I = coh V s /\ msteps s *)
Lemma vrf_post_msteps V A (c : ST V A) s (Q1 Q2 : post V A) :
        (forall x s', coh V s -> msteps (world.id_morph V tt) s s' ->
                      Q1 x s' -> Q2 x s') ->
        vrf c Q1 s -> vrf c Q2 s.
Proof.
move=>R H; apply: vrf_coh=>C.
have I : idstab (fun s' => coh V s' /\ msteps (world.id_morph V tt) s s') s.
- by move=>s' M; split=>//; apply: mcoh M C.
apply: vrf_post_inv I _ (H)=>x s' X.
by apply: R=>//; case: (X _ iter_refl).
Qed.

(* special case of vrf_post_inv when I = coh V *)
(* this is the more customary Hoare rule for post strenghtening *)
Lemma vrf_post V A c (Q1 Q2 : post V A) s :
        (forall x s, coh V s -> Q1 x s -> Q2 x s) ->
        vrf c Q1 s -> vrf c Q2 s.
Proof.
move=>R; apply: vrf_post_msteps=>// x s' C /mcoh M.
by apply: R; apply: M C.
Qed.

(* vrf_ret with coherence hypothesis inserted for convenience *)
Lemma vrf_ret V A (v : A) (Q : post V A) s :
       (forall m, osteps V s m -> coh V m -> Q v m) ->
       vrf (ret v) Q s.
Proof. by move=>H C; apply: vrf_ret=>// m O; apply: H (ocoh O C). Qed.


Lemma vrf_bnd V A1 A2 (c1 : ST V A1) (c2 : A1 -> ST V A2) Q s :
        vrf c1 (fun x => vrf (c2 x) Q) s -> vrf (bnd c1 c2) Q s.
Proof.
move=>H C; apply: vrf_bnd=>//;
by apply: vrf_post H C=>x s' C; apply.
Qed.

(* shorter name for vrf_bnd *)
Definition step := vrf_bnd.

Lemma vrf_par V A1 A2 (c1 : ST V A1) (c2 : ST V A2) Q1 Q2 s :
        star (vrf c1 Q1) (vrf c2 Q2) s ->
        vrf (par c1 c2) (fun v : A1 * A2 => star (Q1 v.1) (Q2 v.2)) s.
Proof.
case=>s1 [s2][S V1 V2] C; apply: vrf_par (C) _.
have C1 : coh V s1 by apply/(coh_gbL _ S).
have C2 : coh V s2 by apply/(coh_gbR _ S).
by exists s1, s2; split=>//; [apply: V1 | apply: V2].
Qed.

(* forward form of vrf_par, with explicit shuffle of self/other *)
(* and coherence added as hypotheses to subgoals *)
Lemma par_do2 V A1 A2 (c1 : ST V A1) (c2 : ST V A2) x1 x2 Q1 Q2 (Q : post _ _) i :
        (* one of the two conditions is superfluous, if we take vrf *)
        (* to be prefixed by coh *)
        self i = x1 \+ x2 ->
        (coh V (State x1 (joint i) (x2 \+ other i)) ->
         vrf c1 Q1 (State x1 (joint i) (x2 \+ other i))) ->
        (coh V (State x2 (joint i) (x1 \+ other i)) ->
         vrf c2 Q2 (State x2 (joint i) (x1 \+ other i))) ->
        (forall v y1 y2 m,
           coh V m ->
           self m = y1 \+ y2 ->
           Q1 v.1 (State y1 (joint m) (y2 \+ other m)) ->
           Q2 v.2 (State y2 (joint m) (y1 \+ other m)) ->
           Q v m) ->
        vrf (par c1 c2) Q i.
Proof.
move=>Si; set s1 := State _ _ _; set s2 := State _ _ _ => Ls1 Ls2 H.
apply: vrf_coh=>C; suff/vrf_par : star (vrf c1 Q1) (vrf c2 Q2) i.
- apply: vrf_post=>// v m Cs [m1][m2][M]; rewrite [m1]stEX' [m2]stEX'.
  by case: M=>X ->-> <-<-; apply: (H v (self m1) (self m2) m).
exists s1, s2; split=>//.
- by apply: Ls1; rewrite /s1 zgEX -cohZ zg_unlock /= -Si -stEX'.
by apply: Ls2; rewrite /s2 zgEX -cohZ zg_unlock /= joinC -Si -stEX'.
Qed.

(* using morphism notation *)
Lemma par_do3 V A1 A2 (c1 : ST V A1) (c2 : ST V A2) x1 x2 Q1 Q2 (Q : post _ _) (i : st V) :
        idfun.s i = x1 \+ x2 ->
        (coh V (State x1 (idfun.j i) (x2 \+ idfun.o i)) ->
         vrf c1 Q1 (State x1 (idfun.j i) (x2 \+ idfun.o i))) ->
        (coh V (State x2 (idfun.j i) (x1 \+ idfun.o i)) ->
         vrf c2 Q2 (State x2 (idfun.j i) (x1 \+ idfun.o i))) ->
        (forall v y1 y2 m,
           coh V m ->
           idfun.s m = y1 \+ y2 ->
           Q1 v.1 (State y1 (idfun.j m) (y2 \+ idfun.o m)) ->
           Q2 v.2 (State y2 (idfun.j m) (y1 \+ idfun.o m)) ->
           Q v m) ->
        vrf (par c1 c2) Q i.
Proof.
rewrite -!lT=>Si; set s1 := State _ _ _; set s2 := State _ _ _ => Ls1 Ls2 H.
apply: vrf_coh=>C; have S : splits i s1 s2 by [].
have C1 : coh V s1 by apply/(coh_gbL _ S).
have C2 : coh V s2 by apply/(coh_gbR _ S).
have /vrf_par : star (vrf c1 Q1) (vrf c2 Q2) i.
- by exists s1, s2; do !split=>//; [apply: Ls1 (C1) | apply: Ls2 (C2)].
apply: vrf_post=>// v m Cs [m1][m2][M]; rewrite [m1]stEX' [m2]stEX'.
case: M=>Y ->-><-<-; rewrite !lT; apply: (H v (idfun.s m1) (idfun.s m2) m)=>//.
by rewrite -!lT.
Qed.

Arguments par_do2 V [A1][A2][c1][c2] x1 x2 Q1 Q2 [Q][i] _ _ _ _.
Arguments par_do3 V [A1][A2][c1][c2] x1 x2 Q1 Q2 [Q][i] _ _ _ _.

Lemma vrf_frame V A (c : ST V A) Q1 Q2 s :
        star (vrf c Q1) (stab Q2) s -> vrf c (fun v => star (Q1 v) Q2) s.
Proof.
case=>s1 [s2][S V1 V2] C; apply: vrf_frame (C) _.
have C1 : coh V s1 by apply/(coh_gbL _ S).
by exists s1, s2; split=>//; apply: V1 C1.
Qed.

(* I'm experimenting with what's the best form of frame lemma in practice *)
(* Here's a bunch of them; eventually to be wittled down *)

Section FrameLemmas.
Variables (V : resource) (A : Type) (c : ST V A) (Q : post V A).

(* with explicit proof of stability of R *)
Lemma vrf_frames (R : st V -> Prop) s :
        star (vrf c Q) R s ->
        (forall s s', R s -> osteps V s s' -> R s') ->
        vrf c (fun v => star (Q v) R) s.
Proof.
case=>s1 [s2][S H1 H2] H; apply: vrf_frame=>//.
by exists s1, s2; split=>// s2'; apply: H H2.
Qed.

(* backward form *)
Lemma vrf_frame2 s s1 s2 :
        splits s s1 s2 ->
        vrf c (fun r s1' => stab (fun s2' =>
                forall s', splits s' s1' s2' -> Q r s') s2) s1 ->
        vrf c Q s.
Proof.
move=>S H; set Q1 := fun _ _ => _ in H.
apply: vrf_coh=>C.
have/vrf_frame : star (vrf c Q1) (stab (fun x => osteps V s2 x)) s.
- by exists s1, s2; do !split=>//.
by apply vrf_post=>// v s' C' [s1'][s2'][S' H'] /H'; apply.
Qed.

(* explicit backward form *)
Lemma frame1 x p (R : _ -> Prop) i :
        self i = x \+ p ->
        (forall m, osteps V (State p (joint i) (x \+ other i)) m -> R m) ->
        vrf c (fun v m =>
               forall y, other m = p \+ y -> R (State p (joint m) (self m \+ y)) ->
                Q v (State (self m \+ p) (joint m) y))
             (State x (joint i) (p \+ other i)) ->
        vrf c Q i.
Proof.
move=>Si Hs H; apply: vrf_coh=>C; set s2 := State _ _ _ in Hs.
set s1 := State x (joint i) (p \+ other i); have S : splits i s1 s2 by [].
apply: (vrf_frame2 S); apply: vrf_post H=>w s1' _ H s2' O s H2.
case: s H2 H=>??? [/= ->-> H2 <- ->].
rewrite -(oself O); apply=>//; rewrite -H2.
by rewrite (_ : p = self s2') -?stEX' -?(oself O) //; apply: Hs.
Qed.

(* more optimized form, where R is taken to be osteps itsefl *)
Lemma frame2 x p i :
        self i = x \+ p ->
        vrf c (fun v m => forall y, other m = p \+ y ->
                osteps V (State p (joint i) (x \+ other i))
                         (State p (joint m) (self m \+ y))  ->
                Q v (State (self m \+ p) (joint m) y))
             (State x (joint i) (p \+ other i)) ->
        vrf c Q i.
Proof.
pose R m := osteps V (State p (joint i) (x \+ other i)) m.
by move=>S H; apply: (frame1 (R:=R)) S _ H.
Qed.

(* this is how it really wants to look like *)
(* talk in terms of self, not other *)
Lemma frame3 x p i :
        self i = x \+ p ->
        (forall y j, coh V j ->
                     i = State (x \+ p) (joint j) y ->
                     self j = x ->
                     other j = p \+ y ->
                     vrf c (fun v m =>
          forall z k,
            coh V k ->
            m = State z (joint k) (p \+ other k) ->
            self k = z \+ p ->
            osteps V (State p (joint j) (x \+ y))
                     (State p (joint k) (z \+ other k)) ->
            Q v k) j) ->
        vrf c Q i.
Proof.
move=>Si H; apply: vrf_coh=>C.
set s1 := State x (joint i) (p \+ other i).
set s2 := State p (joint i) (x \+ other i).
have S : splits i s1 s2 by [].
have C1 : coh V s1 by apply/(coh_gbL _ S).
have C2 : coh V s2 by apply/(coh_gbR _ S).
move/H: (C1)=>{H} /(_ (other i)) /=; rewrite -Si -stEX' => H.
apply: (frame2 Si); apply: vrf_post (H _ _ _)=>// v s Cv {}H y E O.
apply: H=>//=; last by rewrite -E -stEX'.
by rewrite 1?zgEX cohZ zg_unlock /= -?E -?stEX'.
Qed.

(* change to use morphism notation *)
Lemma frame4 x p (i : st V) :
        idfun.s i = x \+ p ->
        (forall y j, coh V j ->
                     i = State (x \+ p) (idfun.j j) y ->
                     idfun.s j = x ->
                     idfun.o j = p \+ y ->
                     vrf c (fun v m =>
          forall z k,
            coh V k ->
            m = State z (idfun.j k) (p \+ idfun.o k) ->
            idfun.s k = z \+ p ->
            osteps V (State p (idfun.j j) (x \+ y))
                     (State p (idfun.j k) (z \+ idfun.o k)) ->
            Q v k) j) ->
        vrf c Q i.
Proof.
rewrite -!lT=>Si H; apply: frame3 Si _=>y j C.
rewrite !lT=>H1 H2 /(H _ _ C H1 H2).
by apply: vrf_post=>v s Cs Hs z k /Hs; rewrite !lT; apply.
Qed.

(* In fact, I rarely use the osteps part *)
(* Because it appears in the post, it always gets pushed really *)
(* far down from the place of framing. So why not just drop it? *)
Lemma frame x p i :
        self i = x \+ p ->
        vrf c (fun v m => forall y, other m = p \+ y ->
                Q v (State (self m \+ p) (joint m) y))
             (State x (joint i) (p \+ other i)) ->
        vrf c Q i.
Proof.
pose R m := osteps V (State p (joint i) (x \+ other i)) m.
by move=>S H; apply: (frame2 S)=>//; apply: vrf_post H=>// w m C H z /H.
Qed.

End FrameLemmas.

Arguments frame4 V [A][c][Q] x p [i] _ _.
Arguments frame3 V [A][c][Q] x p [i] _ _.
Arguments frame2 V [A][c][Q] x p [i] _ _.
Arguments frame V [A][c][Q] x p [i] _ _.

Lemma vrf_atm_help V A (a : action V A) Q s :
        stab (fun s => exists v s',
        eff a v s s' /\ stab (Q v) s') s -> vrf (atomic a) Q s.
Proof. by move=>S C; apply: vrf_atm. Qed.

(* add coherence hypotheses to vrf_atm *)
Lemma vrf_atm V A (a : action V A) (Q : post V A) s :
        (forall j, osteps V s j -> coh V j ->
           exists v k, eff a v j k /\
             forall m, osteps V k m -> coh V k -> Q v m) ->
        vrf (atomic a) Q s.
Proof.
move=>H; apply: vrf_coh=>C; apply: vrf_atm_help=>// s' O; move: (ocoh O C)=>C'.
case: (H _ O C')=>v [k][H1 H2]; exists v, k; split=>// x /H2.
by apply; apply: (eff_coh H1 C').
Qed.

(* repeat the interp hypothesis when proving the postcondition *)
Lemma vrf_atm2 V A (a : action V A) (Q : post V A) s :
        (forall j, osteps V s j -> coh V j ->
           exists r k, eff a r j k /\
             forall m, osteps V k m -> coh V k ->
                       eff a r j k -> Q r m) ->
        vrf (atomic a) Q s.
Proof.
move=>H; apply: vrf_coh=>C; apply: vrf_atm=>// s' O; move: (ocoh O C)=>C'.
case: (H _ O C')=>v [k][H1 H2] C2; exists v, k.
by split=>// x /H2; move/(_ (eff_coh H1 C') H1).
Qed.

Definition actPx' V A g (pf : actionP g) := atomic (@actF V A g pf).
Definition actSx' V A g (pf : smart_actP g) := atomic (@actS V A g pf).

Arguments actPx' [V A] g _.
Arguments actSx' [V A] g _.
Notation "'actPx' g" := (actPx' g _) (at level 80).
Notation "'actSx' g" := (actSx' g _) (at level 80).

(* Also experimenting with morph lemmas *)

Lemma vrf_morph0 V W X A (f : morph X V W) (sim : simulation f)
                 (c : ST V A) x sw Q :
        hat f (vrf c Q) x sw -> sim x sw ->
        vrf (morph_app f x c) (fun r sw' =>
        exists x', hat f (Q r) x' sw' /\ sim x' sw') sw.
Proof.
case=>sv [M H1] S Cw; apply: vrf_morph=>//.
by exists sv; split=>//; apply: H1 (morph_coh Cw M).
Qed.


Section MorphLemmas.
Variables (V W : resource) (X A : Type) (tau : morph X V W) (c : ST V A).
Variables (sim : simulation tau) (sv : st V).

(* forward rule with the addition of explicit stability *)
(* wrt. the morphism; we build a new simulation on the fly *)
(* and apply vrf_morph0 rule *)
Lemma vrf_morph1 x sw Q :
        [morph tau] x sw sv -> sim x sw ->
        vrf c Q sv ->
        vrf (morph_app tau x c) (fun r sw' =>
          exists x' sv', [/\ [morph tau] x' sw' sv', sim x' sw',
                             Q r sv' & mxsteps tau x sw x' sw']) sw.
Proof.
move=>X1 S H; have J : hat tau (vrf c Q) x sw by exists sv.
pose I' := fun x' sw' => mxsteps tau x sw x' sw'.
pose morph_sim_raw x' sw' := sim x' sw' /\ I' x' sw'.
have O : morph_ostepP tau morph_sim_raw.
- move=>x1 sw1 sw1' sv1 Cw Cv M [S1 H1] /iter1 O.
  case: (morph_osteps Cw M S1 O)=>sv' [M' S' O'].
  by exists sv'; do !split=>//; apply: iter_trans H1 (omxs _ _ O).
have F : morph_forwardP tau morph_sim_raw.
- move=>t x1 sw1 sv1 sv1' J1 Cw Cv M [S1 H1] K.
  case: (morph_forw J1 Cw M S1 K)=>t' [x'][sw'][Y1 Y2 Y3 Y4].
  exists t', x', sw'; do !split=>//; apply: iter_trans H1 _.
  by apply: iter1; apply: SelfMxStep J1 _; exists t', sv1, sv1'.
pose sim' := Simulation2 O F.
have S' : sim' x sw by split=>//; apply: iter_refl.
apply: vrf_post (vrf_morph0 J S')=>x0 s Cs [x'][[sv'][M1 Q1][S1' M1']].
by exists x', sv'.
Qed.

(* backward version with coherence hypotheses inserted for convenience *)
(* this is the one I use, so I rename it vrf_morph *)
Lemma vrf_morph x sw (Q : post W A) :
        [morph tau] x sw sv -> sim x sw ->
        (coh V sv -> vrf c (fun r sv' => forall x' sw',
           [morph tau] x' sw' sv' -> sim x' sw' ->
           coh W sw' -> mxsteps tau x sw x' sw' -> Q r sw') sv) ->
        vrf (morph_app tau x c) Q sw.
Proof.
move=>T S H; apply: vrf_coh=>Cw; move: (H (morph_coh Cw T)).
move=>/(vrf_morph1 T S); apply: vrf_post=>//.
by move=>r sw' Cw' [x'][sv'][T' S']; apply.
Qed.

End MorphLemmas.

Arguments vrf_morph [V W X A tau c] sim sv.

(* version of vrf_morph where sim is unfolded *)
Lemma vrf_morphX V W X A (f : morph X V W) (c : ST V A)
                 (sim : X -> st W -> Prop) sv x sw (Q : post W A) :
        morph_simP f sim -> [morph f] x sw sv -> sim x sw ->
        (coh V sv -> vrf c (fun r sv' => forall x' sw',
           [morph f] x' sw' sv' -> sim x' sw' ->
           coh W sw' -> mxsteps f x sw x' sw' -> Q r sw') sv) ->
        vrf (morph_app f x c) Q sw.
Proof. by move=>M; apply: (vrf_morph (Simulation M)). Qed.

Arguments vrf_morphX [V W X A f c] sim sv.

(******************************************)
(* Notation for logical variable postexts *)
(******************************************)

Definition logbase V A (p : pre V) (q : post V A) : spec V unit A :=
  fun => (p, q).

Definition logvar V B A (G : A -> Type) (s : forall x : A, spec V (G x) B) :
                  spec V {x : A & G x} B :=
  fun t => let (x, g) := t in s x g.

Notation "'STsep' [ V ] ( p , q ) " :=
  (STbin V unit (fun => (p, q))) (at level 0,
   format "'[hv ' STsep  [ V ]  ( '[' p , '/' q ']' ) ']'").

Notation "{ x .. y }, 'STsep' [ V ] ( p , q )" :=
  (STbin V _ (logvar (fun x => .. (logvar (fun y => logbase p q)) .. )))
   (at level 0, x binder, y binder, right associativity,
    format "'[hv ' { x .. y }, '/ ' STsep  [ V ]  ( '[' p , '/' q ']' ) ']'").

Notation "[> h1 , .. , hn <]" :=
  (existT _ h1 .. (existT _ hn tt) .. ) (at level 0).

(************************************************************)
(* Lemmas for pulling out and instantiating ghost variables *)
(************************************************************)

(* the lemma for instantiating ghosts of a function call *)
Lemma gE V G A pq (c : STbin V G pq) g (Q : A -> st V -> Prop) s :
        (pq g).1 s ->
        (forall v m, (pq g).2 v m -> coh V m -> Q v m) ->
        vrf c Q s.
Proof.
case: c=>c H /H X Y; apply: vrf_coh=>C.
apply: (vrf_post (Q1:=(pq g).2)) (X C)=>// *.
by apply: Y.
Qed.

Arguments gE [V G A pq c] g [Q] _ _ _.

Notation "'[gE' x1 , .. , xn ']'" :=
  (gE (existT _ x1 .. (existT _ xn tt) ..)) (at level 0).

(* this lemma always follows step, so let's combine them *)
Lemma stepE V G A B (s : spec V G A) (e : STbin V G s) e2 i (Q : post V B) (g : G) :
        (s g).1 i ->
        (forall v m, (s g).2 v m -> coh V m ->
            vrf (e2 v) Q m) ->
        vrf (bnd e e2) Q i.
Proof. by move=>H1 H2; apply: step=>//; apply: gE H1 H2. Qed.

Arguments stepE [V G A B s e e2 i Q] g _ _.

Notation "'[stepE' x1 , .. , xn ']'" :=
  (stepE (existT _ x1 .. (existT _ xn tt) ..)) (at level 0).

(* a version with id invariants (aka., msteps) *)
Lemma gM V G A pq (c : STbin V G pq) g (Q : A -> st V -> Prop) s :
        (pq g).1 s ->
        (forall v m, (pq g).2 v m -> coh V m ->
          msteps (world.id_morph V tt) s m -> Q v m) ->
        vrf c Q s.
Proof.
case: c=>c H /H X Y; apply: vrf_coh=>C.
pose I s' := msteps (world.id_morph V tt) s s'.
have S : idstab I s by [].
apply: (vrf_post_inv (Q1:=(pq g).2)) S _ (X C)=>v s' I' H'.
by apply: Y H' _ _; [apply: mcoh (I' _ iter_refl) C | apply: I'].
Qed.

Arguments gM [V G A pq c] g [Q] _ _ _.

Notation "'[gM' x1 , .. , xn ']'" :=
  (gM (existT _ x1 .. (existT _ xn tt) ..)) (at level 0).

(* the stepE lemma with invariants (aka. msteps) *)
Lemma stepM V G A B (s : spec V G A) (e : STbin V G s) e2 i (Q : post V B) (g : G) :
        (s g).1 i ->
        (forall v m, (s g).2 v m -> coh V m ->
            msteps (world.id_morph V tt) i m -> vrf (e2 v) Q m) ->
        vrf (bnd e e2) Q i.
Proof. by move=>H1 H2; apply: step=>//; apply: gM H1 H2. Qed.

Arguments stepM [V G A B s e e2 i Q] g _ _.

Notation "'[stepM' x1 , .. , xn ']'" :=
  (stepM (existT _ x1 .. (existT _ xn tt) ..)) (at level 0).
