From Coq Require Import ssreflect ssrbool ssrfun.
From mathcomp Require Import ssrnat eqtype seq fintype path tuple finfun.
From fcsl Require Import domain pred prelude ordtype pcm.
From fcsl Require Import world schedule process always.
From fcsl Require Import options.

(* A tight model (fully abstract?) *)
(* It makes sure that all trees in a denotation set capture one and the same *)
(* behavior. The behavior may be approximated better in some trees, *)
(* but it will never be the case that the two trees in the same set *)
(* are totally different; i.e., that they have different actions in the same *)
(* syntactic position. Two trees may differ in a given position *)
(* only in that one of them may have Unfinished there. *)
(* But they cannot have two different defined elements in the same position *)

(* The setup will allow us to prove adequacy, progress and preservation lemmas *)

(* The model uses axioms of choice and of exluded middle, which is why *)
(* I don't make it the main model file. *)

(* for bnd_ret and similarity, we need classical logic *)
(* because of these axioms, I don't make this file be the main model *)
(* but just keep it as an alternative *)

Require Import Coq.Logic.ChoiceFacts.

Axiom cid : forall A, ConstructiveIndefiniteDescription_on A.

Axiom func_choice : forall A B (R:A->B->Prop),
  (forall x, exists y, R x y) ->
  (exists f, forall x, R x (f x)).

Axiom classicT : forall (P : Prop), {P} + {~ P}.

(**********************)
(* Helper definitions *)
(**********************)

Definition pre V := st V -> Prop.
Definition post V A := A -> st V -> Prop.
Definition spec V G A := G -> prod (pre V) (post V A).
Definition post V A := A -> st V -> Prop.
Definition stab V (R : st V -> Prop) s :=
  forall s', osteps V s s' -> R s'.
Definition mstab V W X (tau : morph X V W) (Q : X -> st W -> Prop) x sw :=
  forall x' sw', mxsteps tau x sw x' sw' -> Q x' sw'.
Definition img V W X (tau : morph X V W) Q1 Q2 x sw :=
  exists sv, [/\ \\tau x sw sv, Q1 sv & Q2 x sw].

(*************************************************************)
(* List of inference rules, with vrf predicate kept abstract *)
(*************************************************************)

Module Type VrfSig.

Parameter ST : resource -> Type -> Type.
Parameter ret : forall V A, A -> ST V A.
Parameter bnd : forall V A1 A2, ST V A1 -> (A1 -> ST V A2) -> ST V A2.
Parameter par : forall V A1 A2, ST V A1 -> ST V A2 -> ST V (A1 * A2).
Parameter atomic : forall V A, action V A -> ST V A.
Parameter morph_app : forall V W X A, morph X V W -> X -> ST V A -> ST W A.
Arguments ret [V][A] x.

Parameter vrf : forall V A, ST V A -> post V A -> st V -> Prop.

Parameter vrf_coh : forall V A (c : ST V A) Q s, vrf c Q s -> coh V s.
Parameter vrf_post : forall V A (c : ST V A) (Q1 Q2 : post V A) s,
  (forall x s, coh V s -> Q1 x s -> Q2 x s) -> vrf c Q1 s -> vrf c Q2 s.
Parameter vrf_ret : forall V A (v : A) Q s,
  coh V s -> stab (Q v) s -> vrf (ret v) Q s.
Parameter vrf_bnd : forall V A1 A2 (c1 : ST V A1) (c2 : A1 -> ST V A2) Q s,
  vrf c1 (fun x => vrf (c2 x) Q) s -> vrf (bnd c1 c2) Q s.
Parameter vrf_par : forall V A1 A2 (c1 : ST V A1) (c2 : ST V A2) Q1 Q2 s,
  star (vrf c1 Q1) (vrf c2 Q2) s ->
  vrf (par c1 c2) (fun v : A1 * A2 => star (Q1 v.1) (Q2 v.2)) s.
Parameter vrf_frame : forall V A (c : ST V A) Q1 Q2 s,
  star (vrf c Q1) (stab Q2) s -> vrf c (fun v => star (Q1 v) Q2) s.
Parameter vrf_atm : forall V A (a : action V A) Q s,
   coh V s -> stab (fun s => exists v s',
     interp V (a v) s s' /\ stab (Q v) s') s -> vrf (atomic a) Q s.
Parameter vrf_morph : forall V W X A (tau : morph X V W) (c : ST V A) x sw Q I,
  img tau (vrf c Q) (mstab tau I) x sw ->
  vrf (morph_app tau x c) (fun r sw' => exists x', img tau (Q r) I x' sw') sw.

Definition has_spec V G A (pq : spec V G A) (c : ST V A) :=
  forall s g, (pq g).1 s -> coh V s -> vrf c (pq g).2 s.

Structure STbin V G A (pq : spec V G A) := do' {
  prog_of :> ST V A;
  proof_of : has_spec pq prog_of}.

Arguments STbin V G [A] pq.

(* some notation *)

Notation "x '<--' c1 ';' c2" := (bnd c1 (fun x => c2))
  (at level 81, right associativity).
Notation "c1 ';;' c2" := (bnd c1 (fun _ => c2))
  (at level 81, right associativity).

(* The implementation of this fixed point constructor takes *)
(* monotone closure of the function *)
(* this is a hack, to avoid proving monotonicity all the time *)
(* the later is needed because in the shallow embedding, we can *)
(* escape the abstraction, and implement functions which are not monotone *)
(* We do prove below that all constructors are monotone, so if we stick *)
(* with the provided constructions only, monotonicity is not needed *)
(* This can be avoided by internalizing parametricity, but that's fancy *)
Parameter ffix : forall V G A (B : A -> Type) (pq : forall x, spec V G (B x)),
  ((forall x : A, STbin V G (pq x)) -> forall x : A, STbin V G (pq x)) ->
    forall x : A, STbin V G (pq x).

End VrfSig.

(*****************************)
(* Inference rules are sound *)
(*****************************)

Module Vrf : VrfSig.

(* Semantics of programs as sets of trees. We require that each *)
(* program's denotation contains at least one element -- the unfinished *)
(* tree. This way, an infinite loop is not modeled by an empty set, but *)
(* by a set containing approximating finite trees. *)

(* in the denotations, we want to allow only similar trees *)
(* these basically look exactly the same *)
(* except the one on the left may have more places undefined *)

Fixpoint approx V A (t1 t2 : tree V A) : Prop :=
  match (t1, t2) with
  | (Ret v1, Ret v2) => v1 = v2
  | (Act a1, Act a2) => a1 = a2
  | (Seq A1 x1 k1, Seq A2 x2 k2) =>
      exists pf : A1 = A2,
        approx x1 (cast (@tree V) pf x2) /\
        forall x : A2, approx (k1 (icast pf x)) (k2 x)
  | (Par B1 C1 x1 y1 k1, Par B2 C2 x2 y2 k2) =>
      exists (pf : (B1, C1) = (B2, C2)),
        [/\ approx x1 (cast (fun BC => @tree V BC.1) pf x2),
            approx y1 (cast (fun BC => @tree V BC.2) pf y2) &
        forall x : B2 * C2,
          approx (k1 (cast (fun BC => BC.1 * BC.2 : Type) pf x)) (k2 x)]
  | (Morf U1 X1 x1 f1 q1, Morf U2 X2 x2 f2 q2) =>
       exists (pf : (U1, X1) = (U2, X2)),
         [/\ x1 = cast (fun UX=>UX.2) pf x2,
             f1 = cast (fun UX=>morph UX.2 UX.1 V) pf f2 &
             approx q1 (cast (fun UX=>tree UX.1 A) pf q2)]
  | (_, _) => t1 = Unfinished
  end.


Definition similar V A (t1 t2 : tree V A) :=
  exists t, approx t1 t /\ approx t2 t.

Definition saturated V A (X : tree V A -> Prop) :=
  forall t1 t2, X t2 -> approx t1 t2 -> X t1.

Definition directed V A (X : tree V A -> Prop) :=
  forall t1 t2, X t1 -> X t2 ->
    exists t, [/\ approx t1 t, approx t2 t & X t].

Structure ST' V A := Prog' {
  set_of :> tree V A -> Prop;
  _ : set_of Unfinished;
  _ : saturated set_of;
  _ : directed set_of}.

(* assignments to make Coq's typechecking of modules happy *)
Definition ST := ST'.
Definition Prog V A t pf1 pf2 pf3 : ST V A := @Prog' V A t pf1 pf2 pf3.

Arguments Prog [V A] _ _ _ _.

(* admit \In notation *)
Canonical prog_PredType W A := Eval hnf in mkPredType (@set_of W A).

Lemma stE V A (T1 T2 : ST V A) :
        T1 = T2 <-> set_of T1 = set_of T2.
Proof.
split=>[->//|]; case: T1 T2=>m1 H1 S1 U1 [m2 H2 S2 U2] /= E.
move: H2 S2 U2; rewrite -E => H2 S2 U2.
by rewrite (pf_irr H1) (pf_irr S1) (pf_irr U1).
Qed.

Lemma prog_unfin V A (e : ST V A) : Unfinished \In e.
Proof. by case: e. Qed.

Lemma prog_satur V A (e : ST V A) : saturated e.
Proof. by case: e. Qed.

Lemma prog_direct V A (e : ST V A) : directed e.
Proof. by case: e. Qed.

Hint Resolve prog_unfin prog_satur prog_direct : core.

(* the verify lemma, and the decorated Hoare type *)

Definition vrf V A (c : ST V A) Q s := coh V s /\ forall t, c t -> after s t Q.

Definition has_spec V G A (pq : spec V G A) (c : ST V A) :=
  forall s g, (pq g).1 s -> coh V s -> vrf c (pq g).2 s.

Structure STbin V G A (pq : spec V G A) := do' {
  prog_of :> ST V A;
  proof_of : has_spec pq prog_of}.

Arguments STbin V G [A] pq.

(*******************************)
(* Approximation is a preorder *)
(*******************************)

Lemma apx_refl V A (t : tree V A) : approx t t.
Proof.
by elim: {V A} t=>[V A|V A v|V A a|V A B t1 IH1 t2 IH2|
 V A B1 B2 t1 IH1 t2 IH2 k IH|V A W B x f t IH] //; exists (erefl _).
Qed.

Hint Resolve apx_refl : core.

Lemma apx_trans V A (t1 t2 t3 : tree V A) :
        approx t1 t2 -> approx t2 t3 -> approx t1 t3.
Proof.
elim: {V A} t1 t2 t3=>[V A|V A v|V A a|V A B t IH k IHk|
V A B1 B2 t1 IH1 t2 IH2 k IHk|V A W B y f t IH] tx ty //.
- by case: tx=>//= _ <-; case: ty.
- by case: tx=>//= _ <-; case: ty.
- case: tx=>//= B1' t1' k1' [pf1][]; subst B1'.
  case: ty=>//= B2' t2' k2' S1 S2 [pf2][S3 S4]; subst B2'.
  exists (erefl _); rewrite !eqc in S1 S3 *.
  by split; [apply: IH S1 S3 | move=>x; apply: IHk (S2 x) (S4 x)].
- case: tx=>//= B1' B2' t1' t2' k' [pf][]; case: pf (pf)=>??.
  subst B1' B2'=>pf S1 S2 S.
  case: ty=>//= B1' B2' tx1 tx2 kx [px][]; case: px (px)=>??.
  subst B1' B2'=>px H1 H2 H.
  rewrite !eqc in S1 S2 H1 H2; exists (erefl _); rewrite !eqc.
  split; first by [apply: IH1 S1 H1]; first by [apply: IH2 S2 H2].
  by move=>x; apply: IHk (H x); move: (S x); rewrite !eqc.
case: tx=>//= W' X' x' f' t' [pf][]; case: pf (pf)=>??.
subst W' X'=>pf ->->; rewrite !eqc=>S.
case: ty=>//= W' X' xx fx tx [px][]; case: px (px)=>??.
subst W' X'=>px ->->; rewrite !eqc=>H.
by exists (erefl _); split=>//; apply: IH S H.
Qed.

(**************************************)
(* Inversion lemmas for approximation *)
(**************************************)

Lemma inv_unf V A (t : tree V A) : approx t Unfinished -> t = Unfinished.
Proof. by case: t. Qed.

Lemma inv_ret V A (t : tree V A) v :
        approx t (Ret v) -> t = Unfinished \/ t = Ret v.
Proof. by case: t=>//=; [left | move=>a <-; right]. Qed.

Lemma inv_act V A (a : action V A) t :
        approx t (Act a) -> t = Unfinished \/ t = Act a.
Proof. by case: t=>//=; [left | move=>x <-; right]. Qed.

Lemma inv_seq V A B (t1 : tree V A) (t2 : A -> tree V B) t :
        approx t (Seq t1 t2) ->
        t = Unfinished \/
        exists t1' t2', [/\ t = Seq t1' t2', approx t1' t1 &
                            forall x, approx (t2' x) (t2 x)].
Proof.
case: t t1 t2=>//=; first by left.
by move=>B' t' t1 t2 tx [pf][]; subst B'=>S H; right; exists t', t1.
Qed.

Lemma inv_par V A A1 A2 (t1 : tree V A1) (t2 : tree V A2) k t :
        approx t (Par t1 t2 k) ->
        t = Unfinished \/
        exists t1' t2' k', [/\ t = Par t1' t2' k',
          approx t1' t1, approx t2' t2 &
          forall x, @approx V A (k' x) (k x)].
Proof.
case: t=>//; first by left.
move=>B1 B2 t t1' t2' /= [pf]; case: pf (pf)=>??; subst B1 B2=>pf [].
rewrite !eqc=>S1 S2 S; right; exists t, t1', t2'; split=>//.
by move=>x; move: (S x); rewrite eqc.
Qed.

Lemma inv_morph V W X A (f : morph X V W) (t1 : tree V A) t x :
        approx t (Morf x f t1) ->
        t = Unfinished \/ exists t1', t = Morf x f t1' /\ approx t1' t1.
Proof.
case: t=>//; first by left.
move=>W' B y f' t /= [pf]; case: pf (pf)=>??; subst W' B=>pf [].
by rewrite !eqc=>->-> S; right; exists t.
Qed.

(***********************************************)
(* Approximation with stepping, safety, always *)
(***********************************************)

Lemma apx_good V A (t1 t2 : tree V A) sc :
        approx t1 t2 -> good t1 sc -> good t2 sc.
Proof.
elim: t2 t1 sc=>{V A} V A.
- by move=>t1 sc /inv_unf ->.
- by move=>a t1 sc /(@inv_ret _ _ _ a) [] ->.
- by move=>a t1 sc /inv_act [] ->; case: sc.
- move=>B t1 IH1 t2 IH2 t sc /inv_seq [->|]; first by case: {IH1 IH2} sc.
  case=>t1' [t2'][-> S H]; case: sc=>//; last by move=>sc;  apply: IH1 S.
  by case: t1' S=>// v; case: t1 {IH1}.
- move=>B1 B2 t1 IH1 t2 IH2 t IH tx sc /inv_par [->|]; first by case: sc.
  case=>t1' [t2'][k'][->{tx} S1 S2 S]; case: sc=>//.
  - case: t1' S1=>// v1; case: t1 {IH1} =>//= _ _.
    by case: t2' S2=>// v2 S2 _; case: t2 IH2 S2.
  - by move=>sc; apply: IH1 S1.
  by move=>sc; apply: IH2 S2.
move=>W B b f t IH t1 sc /(@inv_morph _ _ _ _ f) [->|]; first by case: sc.
case=>t1' [->{t1}] S; case: sc=>//; last by move=>sc; apply: IH S.
by case: t1' S=>// v; case: t IH=>//.
Qed.

Lemma apx_forw W V A (cx : morphs V W) (t1 t2 : tree V A)
               sc (x x' : mtp cx) s s' t1' :
        approx t1 t2 -> tstep' cx t1 sc t1' x x' s s' ->
        exists t2', tstep' cx t2 sc t2' x x' s s' /\ approx t1' t2'.
Proof.
elim: t2 t1 t1' W cx x x' s s' sc=>{V A} V A.
- by move=>t1 t1' W cx x x' s s' sc /inv_unf ->; case: sc.
- move=>v t1 t1' W cx x x' s s' sc /(@inv_ret _ _ _ v) [] ->;
  by [case: sc | exists t1'].
- move=>a t1 t1' W cx x x' s s' sc /inv_act [] ->;
  by [case: sc | exists t1'].
- move=>B t IH t1 IH1 t2 t3 W cs x x' s s' sc /inv_seq [->|]; first by case: sc.
  case=>t1' [t2'][->{t2} S1 S2]; case: sc=>//.
  - by case: t1' S1=>// v1; case: t IH=>// v2 IH /= <- [->->->]; exists (t1 v1).
  move=>sc /= [t2][->{t3}] /IH-/(_ S1) [tx][T' S'].
  by eexists (Seq tx t1); split; [exists tx | exists (erefl _)].
- move=>B1 B2 t IH t1 IH1 tk IHk t2 t3 W cx x x' s s' sc /inv_par.
  case=>[->|[t1'][t2'][k'][->{t2} S1 S2 S]]; case: sc=>//.
  - case: t1' S1=>// v1; case: t {IH}=>//= _ <-; case: t2' S2=>// v2.
    by case: t1 {IH1}=>// _ <- [->->->]; exists (tk (v1, v2)).
  - move=>sc [r1][->{t3}] /IH-/(_ S1) [tx][T' S'].
    by exists (Par tx t1 tk); split; [exists tx | exists (erefl _)].
  move=>sc [r2][->{t3}] /IH1-/(_ S2) [tx][T' S'].
  by exists (Par t tx tk); split; [exists tx | exists (erefl _)].
move=>W Y y f t IH t1 t1' W' f' x x' s s' sc /(@inv_morph _ _ _ _ f).
case=>[->|[tx][->{t1} S]]; case: sc=>//.
- by case: tx S=>// v; case: t {IH}=>// _ <- [->->->]; exists (Ret v).
move=>sc [x2][t2] [->] /IH-/(_ S) [t2'][T' S'].
by exists (Morf x2 f t2'); split; [exists x2, t2' | exists (erefl _)].
Qed.

(* the other direction holds only for good schedules *)
Lemma apx_backw W V A (cx : morphs V W) (t1 t2 : tree V A)
                sc (x x' : mtp cx) s s' t2' :
        approx t1 t2 -> good t1 sc -> tstep' cx t2 sc t2' x x' s s' ->
        exists t1', tstep' cx t1 sc t1' x x' s s' /\ approx t1' t2'.
Proof.
elim: t2 t1 t2' W cx x x' s s' sc=>{V A} V A.
- by move=>t1 t2' W cx x x' s s' [].
- by move=>v t1 t2' W cx x x' s s' [].
- move=>a t1 t2' W cx x x' s s' [] // /inv_act.
  by case=>-> //= _ [v][-> H]; exists (Ret v); split=>//; exists v.
- move=>B t IH t1 IH1 t2 t3 W cs x x' s s' sc /inv_seq.
  case=>[->|[t1'][t2'][-> S1 S2]]; case: sc=>//.
  - by case: t1' S1=>// v; case: t IH=>// _ _ <- _ [->->->]; exists (t2' v).
  move=>sc /= G [t4][->] /IH-/(_ _ S1 G) [tx][T' S'].
  by exists (Seq tx t2'); split; [exists tx | exists (erefl _)].
- move=>B1 B2 t1 IH1 t2 IH2 k IHk t1' t2' W cx x x' s s' sc /inv_par.
  case=>[->|[tx][tx'][k'][-> S1 S2 S]]; case: sc=>//.
  - case: tx S1=>// v; case: t1 IH1=>// _ _ <-.
    case: tx' S2=>// w; case: t2 IH2=>// _ _ <- [[->->->]].
    by exists (k' (v, w)).
  - move=>sc G [r1][->] /IH1-/(_ _ S1 G) [ty][T' S'].
    by exists (Par ty tx' k'); split; [exists ty | exists (erefl _)].
  move=>sc G [r2][->] /IH2-/(_ _ S2 G) [ty][T' S'].
  by exists (Par tx ty k'); split; [exists ty | exists (erefl _)].
move=>W' B y f t IH t1 t2' W cx x x' s s' sc /(@inv_morph _ _ _ _ f).
case=>[->|[t1'][-> S]]; case: sc=>//.
- by case: t1' S=>// v; case: t {IH} =>// _ <- _ [->->->]; exists (Ret v).
move=>sc /= G [y2][t2][->] /IH-/(_ _ S G) [tx][T' S'].
by exists (Morf y2 f tx); split; [exists y2, tx | exists (erefl _)].
Qed.

Lemma apx_tstep W V A (cx : morphs V W) (t1 t2 : tree V A)
                sc (x x1 x2 : mtp cx) s s1 s2 t1' t2' :
        approx t1 t2 ->
        tstep' cx t1 sc t1' x x1 s s1 -> tstep' cx t2 sc t2' x x2 s s2 ->
        [/\ x1 = x2, s1 = s2 & approx t1' t2'].
Proof. by move=>S /(apx_forw S)[t'][T1 S1] /(tst_det T1)[<-<-<-]. Qed.

Lemma apx_tsafe V A (t1 t2 : tree V A) sc s :
        approx t1 t2 -> tsafe t2 sc s -> tsafe t1 sc s.
Proof.
move=>S T G; case: T=>[|x [q]]; first by apply: apx_good G.
by move/apx_backw=>/(_ _ S G) [t1'][T' _]; exists x, t1'.
Qed.

Lemma apx_aft V A P (t1 t1' : tree V A) s :
        approx t1 t1' -> after s t1' P -> after s t1 P.
Proof.
move=>S H sc; elim: sc t1 t1' s S {H} (H sc)=>[|sc scs IH] t1 t1' s S H /= s2 O.
- case: (H _ O)=>C H'; split=>// v E; rewrite {t1}E in S.
  by case: t1' S {H} H'=>//= _ <-; apply.
case: (H _ O)=>T1 [C T2] T3; split.
- by apply: apx_tsafe S T1.
- split=>// v E; rewrite {t1}E in S *.
  by case: t1' S {IH H T1 T3} T2=>//= _ <-; apply.
by move=>s3 q /apx_forw - /(_ _ S) [t2'][T S2]; apply: IH S2 (T3 _ _ _).
Qed.

(********************************)
(* Approximation and similarity *)
(********************************)

Lemma sim_tstep W V A (cx : morphs V W) (c : ST V A) (t1 t2 : tree V A)
                sc (x x1 x2 : mtp cx) s s1 s2 t1' t2' :
        similar t1 t2 ->
        tstep' cx t1 sc t1' x x1 s s1 ->
        tstep' cx t2 sc t2' x x2 s s2 ->
        [/\ x1 = x2, s1 = s2 & similar t1' t2'].
Proof.
case=>t [S1 S2] /(apx_forw S1) [q1][T1 H1] /(apx_forw S2)[q2][T2 H2].
case: (apx_tstep (apx_refl _) T1 T2)=>->-> S; split=>//.
by exists q2; split=>//; apply: apx_trans H1 S.
Qed.

Definition join V A (c : ST V A) t1 t2 (pf1 : t1 \In c) (pf2 : t2 \In c) :=
  sval (cid (prog_direct pf1 pf2)).

Lemma apx_joinL V A (c : ST V A) t1 t2 pf1 pf2 :
        approx t1 (@join V A c t1 t2 pf1 pf2).
Proof. by rewrite /join; case: (cid _)=>t []. Qed.

Lemma apx_joinR V A (c : ST V A) t1 t2 pf1 pf2 :
        approx t2 (@join V A c t1 t2 pf1 pf2).
Proof. by rewrite /join; case: (cid _)=>t []. Qed.

Lemma apx_join V A (c : ST V A) t1 t2 pf1 pf2 :
        @join V A c t1 t2 pf1 pf2 \In c.
Proof. by rewrite /join; case: (cid _)=>t []. Qed.

Hint Resolve apx_joinL apx_joinR apx_join : core.

(*****************************)
(* CPO structure of programs *)
(*****************************)

Section STCPO.
Variables (V : resource) (A : Type).

Definition leq (c1 c2 : ST V A) := set_of c1 <== set_of c2.

Lemma leq_refl c : leq c c.
Proof. by []. Qed.

Lemma leq_trans c1 c2 c3 : leq c1 c2 -> leq c2 c3 -> leq c1 c3.
Proof. by move=>H1 H2 p x; apply: H2; apply: H1. Qed.

Lemma leq_asym c1 c2 : leq c1 c2 -> leq c2 c1 -> c1 = c2.
Proof.
move: c1 c2=>[m1 N1 R1 U1][m2 N2 R2 U2]; rewrite /leq /= => H1 H2.
by apply/stE; apply: poset_asym H1 H2.
Qed.

Definition stPosetMix := PosetMixin leq_refl leq_asym leq_trans.
Canonical stPoset := Eval hnf in Poset (ST V A) stPosetMix.

Definition bot_set := Pred1 (@Unfinished V A).

Lemma unfbotP : bot_set Unfinished.
Proof. by []. Qed.

Lemma satbotP : saturated bot_set.
Proof. by move=>t1 t2 <- /inv_unf ->. Qed.

Lemma dirbotP : directed bot_set.
Proof. by move=>t1 t2 <-<-; exists Unfinished. Qed.

Definition bot_prg := Prog bot_set unfbotP satbotP dirbotP.

Lemma pbotP x : bot_prg <== x.
Proof. by move=>t <-; apply: prog_unfin. Qed.

Definition lim_set (cs : chain [poset of ST V A]) t :=
  exists c : ST V A, t \In c /\ c \In cs.

Lemma unflimP cs : lim_set cs Unfinished.
Proof. by case: cs=>cs [[c H1 H2]]; exists c. Qed.

Lemma satlimP cs : saturated (lim_set cs).
Proof. by move=>t1 t2 [c][H C] S; exists c; split=>//; apply: prog_satur S. Qed.

Lemma satdirP cs : directed (lim_set cs).
Proof.
move=>t1 t2 [c1][H1 C1][c2][H2 C2].
case: cs C1 C2=>cs [C X] C1 C2; case: (X _ _ C1 C2)=>H.
- case: (prog_direct (H _ H1) H2)=>t [S1 S2 S].
  by exists t; split=>//; exists c2.
case: (prog_direct H1 (H _ H2))=>t [S1 S2 S].
by exists t; split=>//; exists c1.
Qed.

Definition lim_prg cs :=
  Prog (@lim_set cs) (unflimP cs) (@satlimP cs) (@satdirP cs).

Lemma plimP cs x : x \In cs -> x <== lim_prg cs.
Proof. by exists x. Qed.

Lemma plimM cs x : (forall y, y \In cs -> y <== x) -> lim_prg cs <== x.
Proof. by move=>H t [c][H1 H2]; apply: H H1. Qed.

Definition stCPOMix := CPOMixin pbotP plimP plimM.
Canonical stCPO := Eval hnf in CPO (ST V A) stCPOMix.

End STCPO.

(* cpo structure on STbin *)

Section STbinCPO.
Variables (V : resource) (G A : Type) (pq : spec V G A).
Notation STbin s := (@STbin V G A s).

Definition bleq (c1 c2 : STbin pq) := prog_of c1 <== prog_of c2.

Lemma bleq_refl c : bleq c c.
Proof. by []. Qed.

Lemma bleq_trans c1 c2 c3 : bleq c1 c2 -> bleq c2 c3 -> bleq c1 c3.
Proof. by apply: poset_trans. Qed.

Lemma bleq_asym c1 c2 : bleq c1 c2 -> bleq c2 c1 -> c1 = c2.
Proof.
case: c1 c2=>c1 pf1 [c2 pf2] H1 H2; move: (poset_asym H1 H2)=>/= E.
by rewrite -E in pf2 H1 H2 *; rewrite (pf_irr pf1 pf2).
Qed.

Definition STbinPosetMix := PosetMixin bleq_refl bleq_asym bleq_trans.
Canonical STbinPoset := Eval hnf in Poset (STbin pq) STbinPosetMix.

Lemma bbot_spec : has_spec pq bot.
Proof.
move=>s g H C; split=>// t <-; apply: alw_unfin=>s2 O.
by split=>//; apply: ocoh C.
Qed.

Definition bbot := do' bbot_spec.

Lemma bbotP x : bbot <== x.
Proof. by case: x=>x pf; apply: botP. Qed.

Lemma chain_pf (cs : chain [poset of STbin pq]) :
        chain_axiom (fun s : ST V A => exists pf : has_spec pq s, do' pf \In cs).
Proof.
case: cs=>cs [[[d pf] H1] H2]; split; first by exists d, pf.
by move=>x y [qf1] /H2 X [qf2] /X.
Qed.

Lemma blim_spec cs : has_spec pq (lim (Chain (chain_pf cs))).
Proof.
move=>s g H C; split=>// t [c][H1][pf H2].
by apply: (proj2 (pf _ _ H C)) H1.
Qed.

Definition blim cs := do' (blim_spec cs).

Lemma blimP cs x : x \In cs -> x <== blim cs.
Proof. by case: x=>x pf H; apply: limP; exists pf. Qed.

Lemma blimM cs x : (forall y, y \In cs -> y <== x) -> blim cs <== x.
Proof. by case: x=>x pf H; apply: limM=>y [qf]; apply: H. Qed.

Definition STbinCPOMix := CPOMixin bbotP blimP blimM.
Canonical STbinCPO := Eval hnf in CPO (STbin pq) STbinCPOMix.

End STbinCPO.

(***********************)
(* command denotations *)
(***********************)

(* not exported *)

(* ret *)
Definition ret_set V A (x : A) t := t = Unfinished \/ t = @Ret V A x.

Lemma ret_sat V A x : saturated (@ret_set V A x).
Proof.
move=>t1 t2 [-> /inv_unf ->|->]; first by left.
by case: t1=>//; [left | move=>a ->; right].
Qed.

Lemma ret_dir V A x : directed (@ret_set V A x).
Proof. by move=>t1 t2 [] -> [] ->; exists (Ret x); split=>//; right. Qed.

Definition ret V A (x : A) :=
  Prog _ (or_introl (erefl _)) (@ret_sat V A x) (@ret_dir V A x).

Arguments ret [V][A] x.

(* bnd *)

Definition bnd_set V A B (c1 : ST V A) (c2 : A -> ST V B) t :=
  t = Unfinished \/ exists t', t \In tcat t' c2 /\ t' \In c1.

Lemma bnd_sat V A B c1 c2 : saturated (@bnd_set V A B c1 c2).
Proof.
move=>t1 t2 [-> /inv_unf ->|]; first by left.
case=>t1' [[q1]][->{t2}] H1 T1 /inv_seq [->|]; first by left.
case=>tx' [t2'][->{t1} S H]; right; exists tx'; split.
- by exists t2'; split=>// x; apply: prog_satur (H1 x) (H x).
by apply: prog_satur T1 S.
Qed.

Lemma bnd_dir V A B c1 c2 : directed (@bnd_set V A B c1 c2).
Proof.
move=>t1 t2 [->|].
- case=>[->|]; first by exists Unfinished; split=>//; left.
- case=>t' [[q]][-> H1 H2]; exists (Seq t' q); split=>//.
  by right; exists t'; split=>//; exists q.
case=>t' [[q]][-> H1 H2][->|].
- by exists (Seq t' q); split=>//; right; exists t'; split=>//; exists q.
case=>t1' [[q1]][-> H1' H2'].
case: (prog_direct H2 H2')=>t [X1 X2 X].
pose k x := join (H1 x) (H1' x).
exists (Seq t k); split.
- by exists (erefl _); split=>// x; apply: apx_trans (apx_joinL _ _).
- by exists (erefl _); split=>// x; apply: apx_trans (apx_joinR _ _).
by right; exists t; split=>//; exists k; split=>// x; apply: apx_join.
Qed.

Definition bnd V A B c1 c2 :=
  Prog _  (or_introl (erefl _)) (@bnd_sat V A B c1 c2) (@bnd_dir V A B c1 c2).

(* par *)

Definition par_set V A B (c1 : ST V A) (c2 : ST V B) t :=
  t = Unfinished \/ exists t1 t2 k,
  [/\ t = Par t1 t2 k, t1 \In c1, t2 \In c2 & forall x, approx (k x) (Ret x)].

Lemma par_sat V A B c1 c2 : saturated (@par_set V A B c1 c2).
Proof.
move=>t1 t2 [-> /inv_unf ->|]; first by left.
case=>t3 [t4][k][->] T1 T2 H /inv_par [->|]; first by left.
case=>t1' [t2'][k'][-> S H1 H2]; right; exists t1', t2', k'.
split=>//; last by move=>x; apply: apx_trans.
- by apply: prog_satur S.
by apply: prog_satur H1.
Qed.

Lemma par_dir V A B c1 c2 : directed (@par_set V A B c1 c2).
Proof.
move=>t1 t2 [->|].
- case=>[->|]; first by exists Unfinished; split=>//; left.
  case=>t3 [t4][k][-> H1 H2 H]; exists (Par t3 t4 k).
  by split=>//; right; exists t3, t4, k.
case=>t3 [t4][k][-> H1 H2 H][->|].
- by exists (Par t3 t4 k); split=>//; right; exists t3, t4, k.
case=>t5 [t6][q][-> X1 X2 X].
exists (Par (join H1 X1) (join H2 X2) Ret).
split; last by right; eexists _, _, Ret.
- by exists (erefl _); split=>//; apply: apx_joinL.
by exists (erefl _); split=>//; apply: apx_joinR.
Qed.

Definition par V A B c1 c2 :=
  Prog _ (or_introl (erefl _)) (@par_sat V A B c1 c2) (@par_dir V A B c1 c2).

(* atomic *)

Definition atomic_set V A (a : action V A) t := t = Unfinished \/ t = Act a.

Lemma atm_sat V A a : saturated (@atomic_set V A a).
Proof.
move=>t1 t2 [-> /inv_unf|]; first by left.
by move=>-> /inv_act [] ->; [left | right].
Qed.

Lemma atm_dir V A a : directed (@atomic_set V A a).
Proof. by move=>t1 t2 [] -> [] ->; exists (Act a); split=>//; right. Qed.

Definition atomic V A a :=
  Prog _ (or_introl (erefl _)) (@atm_sat V A a) (@atm_dir V A a).

(* morph_app *)

Definition morphapp_set V W X A (tau : morph X V W) (x : X) (c : ST V A) t :=
  t = Unfinished \/ exists t', t' \In c /\ t = Morf x tau t'.

Lemma morph_sat V W X A tau x c : saturated (@morphapp_set V W X A tau x c).
Proof.
move=>t1 t2 [-> /inv_unf|]; first by left.
case=>t' [H1 ->] S; case: {S} (inv_morph S)=>[->|]; first by left.
by case=>t1' [->] S; right; exists t1'; split=>//; apply: prog_satur S.
Qed.

Lemma morph_dir V W X A tau x c : directed (@morphapp_set V W X A tau x c).
Proof.
move=>t1 t2 [->|].
- case=>[->|]; first by exists Unfinished; split=>//; left.
  by case=>t' [H ->]; exists (Morf x tau t'); split=>//; right; exists t'.
case=>t' [H ->][->|].
- by exists (Morf x tau t'); split=>//; right; exists t'.
case=>t1' [H' ->]; exists (Morf x tau (join H H')).
split; last by right; eexists _.
- by exists (erefl _); split=>//; apply: apx_joinL.
by exists (erefl _); split=>//; apply: apx_joinR.
Qed.

Definition morph_app V W X A tau x c :=
  Prog _ (or_introl (erefl _)) (@morph_sat V W X A tau x c)
                               (@morph_dir V W X A tau x c).

(* fix *)

Section Fix.
Variables (V : resource) (G A : Type) (B : A -> Type).
Variable pq : forall x, spec V G (B x).
Notation tp := (forall x, STbin _ _ (pq x)).
Notation cpo := (dfunCPO (fun x => [cpo of STbin _ _ (pq x)])).
Variable (f : cpo -> cpo) (pf : continuous f).
Definition ffix := kleene_lfp pf.
End Fix.

(*************************************************)
(* Lifting operational semantics for trees       *)
(* to semantics of programs; i.e., sets of trees *)
(*************************************************)

(* the following relation is for program stepping *)
(* as opposed to tree stepping that we had before *)

(* we show that thus defined relation is adequate for *)
(* the operational semantics that we have in mind *)
(* In particular, the relation admits the expected *)
(* operational inference rules *)

(* A tree t2 is in e2 if there is an original t1 for it in e1, which *)
(* evaluates to t2. To handle the fact that Undefined also has to be in *)
(* e2, we allow for e1 to evaluate to t2' which t2 approximates. That *)
(* immediately includes Undefined into e2, since Undefined approximates *)
(* every tree. By saturating the sets with respect to the approximation *)
(* relation, we conflate the difference between t2 and t2'. *)

Definition pstep' V W A (cx : morphs V W) (e1 : ST V A) sc (e2 : ST V A)
                 xs1 xs2 (s1 s2 : st W) :=
  [/\ forall t2, t2 \In e2 -> exists t1 t2',
     [/\ t1 \In e1, t2' \In e2, tstep' cx t1 sc t2' xs1 xs2 s1 s2 &
         approx t2 t2'] &
      forall t1 t2 (xs2' : mtp cx) s2', t1 \In e1 ->
        tstep' cx t1 sc t2 xs1 xs2' s1 s2' -> t2 \In e2].

Definition pstep V A e1 sc e2 :=
  @pstep' V V A (mnil (erefl V)) e1 sc e2 tt tt.

Lemma pstep_atm V W A (cx : morphs V W) (a : action V A) (v : A)
        (xs1 xs2 : mtp cx) s1 s2 :
        unwind cx (a v) xs1 xs2 s1 s2 ->
        pstep' (atomic a) ChoiceAct (ret v) xs1 xs2 s1 s2.
Proof.
move=>H; split; last first.
- by move=>t1 t2 ?? [] -> // [w][->] /(unw_det H) [->]; right.
move=>t2 X; exists (Act a), (Ret v); split;
by [right|right|exists v|case: X=>->].
Qed.


Lemma pstep_bnd_ret V W A B (cx : morphs V W) (v : A) (e : A -> ST V B)
                    (xs : mtp cx) (s : st W) :
        pstep' (bnd (ret v) e) SeqRet (e v) xs xs s s.
Proof.
split; last first.
- by move=>t1 t2 ?? [-> //|][_][[q]][-> /(_ v) S][] -> //= [->].
move=>t2 H; pose q := fun x => if classicT (x = v) then t2 else Unfinished.
exists (Seq (Ret v) q), t2; split=>//; last first.
- by split=>//; rewrite /q; case: (classicT _).
right; exists (Ret v); split; last by right.
exists q; split=>// x; rewrite /q; case: (classicT _)=>X; first by subst x.
by apply: prog_unfin.
Qed.

Lemma pstep_bnd V W A B (cx : morphs V W) sc (e1 e1' : ST V A)
        (e2 : A -> ST V B) (xs1 xs2 : mtp cx) (s1 s2 : st W) :
        pstep' e1 sc e1' xs1 xs2 s1 s2 ->
        pstep' (bnd e1 e2) (SeqStep sc) (bnd e1' e2) xs1 xs2 s1 s2.
Proof.
move=>H; split; last first.
- move=>t1 t2 ?? [->//|][t'][[q]][-> S X] /= [t3][->] /(proj2 H _ _ _ _ X).
  by right; exists t3; split=>//; exists q.
move=>t2 [->|].
- case: (proj1 H Unfinished)=>[|t1' [t2'][H1 H2 T]]; first by apply: prog_unfin.
  exists (Seq t1' (fun => Unfinished)), (Seq t2' (fun => Unfinished)).
  split=>//; last by exists t2'.
  - right; exists t1'; split=>//; exists (fun => Unfinished).
    by split=>// x; apply: prog_unfin.
  right; exists t2'; split=>//; exists (fun => Unfinished).
  by split=>// x; apply: prog_unfin.
case=>t' [[q]][->{t2}] X /(proj1 H) [t1][t2'][X1 X2 T] S.
exists (Seq t1 q), (Seq t2' q); split; last by exists (erefl _).
- by right; exists t1; split=>//; exists q.
- by right; exists t2'; split=>//; exists q.
by exists t2'.
Qed.

Lemma pstep_parL V W A1 A2 (cx : morphs V W) sc (e1 e1' : ST V A1)
        (e2 : ST V A2) (xs1 xs2 : mtp cx) s1 s2 :
        pstep' e1 sc e1' xs1 xs2 s1 s2 ->
        pstep' (par e1 e2) (ParL sc) (par e1' e2) xs1 xs2 s1 s2.
Proof.
move=>H; split; last first.
- move=>t1 t2 ?? [->//|][t3][t4][k][-> X1 X2 S] /= [t][->] /(proj2 H _ _ _ _ X1).
  by right; exists t, t4, k.
move=>t2 [->|].
- case: (proj1 H Unfinished)=>[|t1' [t2'][H1 H2 T] _]; first by apply: prog_unfin.
  exists (Par t1' Unfinished Ret), (Par t2' Unfinished Ret).
  split=>//; last by exists t2'.
  - by right; exists t1', Unfinished, Ret; split=>//; apply: prog_unfin.
  by right; exists t2', Unfinished, Ret; split=>//; apply: prog_unfin.
case=>t1 [t3][k][->{t2}] X1 X2 S; case: (proj1 H _ X1)=>t2 [t2'][X3 X4] T S'.
exists (Par t2 t3 Ret), (Par t2' t3 Ret); split=>//; last by exists (erefl _).
- by right; exists t2, t3, Ret.
- by right; exists t2', t3, Ret.
by exists t2'.
Qed.

Lemma pstep_parR V W A1 A2 (cx : morphs V W) sc (e1 : ST V A1)
        (e2 e2' : ST V A2) (xs1 xs2 : mtp cx) s1 s2 :
        pstep' e2 sc e2' xs1 xs2 s1 s2 ->
        pstep' (par e1 e2) (ParR sc) (par e1 e2') xs1 xs2 s1 s2.
Proof.
move=>H; split; last first.
- move=>t1 t2 ?? [->//|][t3][t4][k][-> X1 X2 S] /= [t][->] /(proj2 H _ _ _ _ X2).
  by right; exists t3, t, k.
move=>t2 [->|].
- case: (proj1 H Unfinished)=>[|t1' [t2'][H1 H2 T _]]; first by apply: prog_unfin.
  exists (Par Unfinished t1' Ret), (Par Unfinished t2' Ret).
  split=>//; last by exists t2'.
  - by right; exists Unfinished, t1', Ret; split=>//; apply: prog_unfin.
  by right; exists Unfinished, t2', Ret; split=>//; apply: prog_unfin.
case=>t1 [t3][k][->{t2}] X1 X2 S; case: (proj1 H _ X2)=>t2 [t2'][X3 X4] T S'.
exists (Par t1 t2 Ret), (Par t1 t2' Ret); split; last by exists (erefl _).
- by right; exists t1, t2, Ret.
- by right; exists t1, t2', Ret.
by exists t2'.
Qed.

Lemma pstep_par_ret V W A1 A2 (cx : morphs V W) (v1 : A1) (v2 : A2)
                    (xs : mtp cx) (s : st W) :
        pstep' (par (ret v1) (ret v2)) ParRet (ret (v1, v2)) xs xs s s.
Proof.
split; last first.
- move=>t1 t2 ?? [->//|][t3][t4][k][-> X1 X2 S].
  case: t3 t4 X1 X2=>// w1 [] // w2 [] // [->][] // [->][-> _ _].
  by case/(@inv_ret _ _ _ _): (S (v1, v2))=>->; [left | right].
move=>t2 H; exists (Par (Ret v1) (Ret v2) Ret), (Ret (v1, v2));
split=>//; last by [case: H=>->]; last by right.
by right; exists (Ret v1), (Ret v2), Ret; split=>//; right.
Qed.

Lemma pstep_morph (W W' V : resource) X A (cx : morphs W V) (f : morph X W' W)  sc
                  (e1 e2 : ST W' A) x1 x2 (xs1 xs2 : mtp cx) (s1 s2 : st V) :
        pstep' (cx:=(mrcons cx f))
          e1 sc e2 (trcons f xs1 x1) (trcons f xs2 x2) s1 s2 ->
        pstep' (cx:=cx)
          (morph_app f x1 e1) (MorfStep sc) (morph_app f x2 e2) xs1 xs2 s1 s2.
Proof.
move=>H; split; last first.
- move=>t1 t2 xs2' s2' [->//|][t'][X1 ->] /= [x3][t3][->] T.
  suff [???] : [/\ xs2' = xs2, x3 = x2 & s2' = s2].
  - by subst xs2' x3 s2'; right; exists t3; split=>//; apply: (proj2 H) T.
  case/(proj1 H): (proj2 H _ _ _ _ X1 T)=>t'' [r2][X1' _ T'] _.
  have S : similar t' t''.
  - by case: (prog_direct X1 X1')=>t []; exists t.
  by case: (sim_tstep e1 S T T')=>/trcons_inj [->->]->.
move=>t2 [->|].
- case: (proj1 H Unfinished)=>[|t1' [t2'][H1 H2 T _]]; first by apply: prog_unfin.
  exists (Morf x1 f t1'), (Morf x2 f t2'); split=>//; last by exists x2, t2'.
  - by right; exists t1'.
  by right; exists t2'.
case=>t' [Z ->]; case: (proj1 H _ Z)=>t1' [t2'][H1 H2 T] S.
exists (Morf x1 f t1'), (Morf x2 f t2'); split; last by exists (erefl _).
- by right; exists t1'.
- by right; exists t2'.
by exists x2, t2'.
Qed.

Lemma pstep_morph_ret (W W' V : resource) X A (cx : morphs W V) (f : morph X W' W)
                      x (v : A) (xs : mtp cx) (s : st V) :
        pstep' (cx:=cx) (morph_app f x (ret v)) MorfRet (ret v) xs xs s s.
Proof.
split; last first.
- move=>t1 t2 xs2' s2' [->//|][t'][X1 ->]; case: t' X1=>// w [] // [->].
  by case=>->; right.
exists (Morf x f (Ret v)), (Ret v); split=>//; last by case: H=>->.
- by right; exists (Ret v); split=>//; right.
by right.
Qed.

Section PstepFix.
Variables (V W : resource) (G A : Type) (B : A -> Type) (cx : morphs V W).
Variable pq : forall x, spec V G (B x).
Notation tp := (forall x, STbin _ _ (pq x)).
Notation cpo := (dfunCPO (fun x => [cpo of STbin _ _ (pq x)])).
Variables (f : cpo -> cpo) (pf : continuous f).

Lemma pstep_fix x sc e (xs1 xs2 : mtp cx) (s s'  : st W) :
        pstep' (f (ffix pf) x) sc e xs1 xs2 s s' ->
        pstep' (ffix pf x) sc e xs1 xs2 s s'.
Proof. by rewrite /ffix kleene_lfp_fixed. Qed.

End PstepFix.

(* evaluation is deterministic *)

Lemma pstep_det V W A (cx : morphs V W) (e e1 e2 : ST V A) sc
                (xs xs1 xs2 : mtp cx) (s s1 s2 : st W) :
        pstep' e sc e1 xs xs1 s s1 -> pstep' e sc e2 xs xs2 s s2 ->
        [/\ xs1 = xs2, s1 = s2 & e1 = e2].
Proof.
move=>H1 H2.
case: (proj1 H1 Unfinished (prog_unfin _))=>t1 [x1][T1 X1 Ts1 _].
case: (proj1 H2 Unfinished (prog_unfin _))=>t2 [x2][T2 X2 Ts2 _].
have S : similar t1 t2 by case: (prog_direct T1 T2)=>t []; exists t.
case: (sim_tstep e S Ts1 Ts2)=>->-> _; split=>//.
apply/stE; apply/fext=>t {x1 t1 X1 T1 Ts1 x2 t2 X2 T2 Ts2 S}; apply/pext; split.
- by case/(proj1 H1)=>t1 [x1][X1 T1 /(proj2 H2 _ _ _ _ X1)]; apply: prog_satur.
by case/(proj1 H2)=>t2 [x2][X2 T2 /(proj2 H1 _ _ _ _ X2)]; apply: prog_satur.
Qed.

(* we can also state what would be equivalents of progress and *)
(* preservation lemmas in type system theory *)

(* Progress *)

Definition nonterminal V A (e : ST V A) :=
  exists t sc, t \In e /\ good t sc.

(* first a helper lemma *)
Lemma prog' V G A (pq : spec V G A) (e : STbin V G pq) (s : st V) t sc g :
        (pq g).1 s -> coh V s -> t \In prog_of e -> tsafe t sc s.
Proof.
case: e=>e H H1 C /= T; case: (H s g H1 C)=>_ /(_ _ T).
by move/alw_safe; move/(_ sc).
Qed.

(* and now the main one *)
Lemma pst_progress V G A (pq : spec V G A) (e : STbin V G pq) (s : st V) g :
        (pq g).1 s -> coh V s -> nonterminal e ->
        exists sc s' e', pstep e sc e' s s'.
Proof.
move=>H C [t][sc][H1 H2].
set e' := fun x => exists t q s',
  [/\ t \In prog_of e, tstep t sc q s s' & approx x q].
have Un : e' Unfinished.
- by case: (@prog' _ _ _ _ _ _ _ sc g H C H1)=>// x [q]; exists t, q, x.
have St : saturated e'.
- move=>t1 t2 [r][q][s'][R1 T1 S1] X2.
  by exists r, q, s'; split=>//; apply: apx_trans S1.
have Dr : directed e'.
- move=>t1 t2 [t1'][q1][s1'][X1 T1 S1][t2'][q2][s2'][X2 T2 S2].
  case: (prog_direct X1 X2)=>q [Y1 Y2 Y].
  case: (apx_forw Y1 T1)=>r1 [G1 J1].
  case: (apx_forw Y2 T2)=>r2 [G2 J2].
  case: (tst_det G1 G2) J2 (G2)=><- _ <- J2 {G2} G2.
  exists r1; split; last by [exists q, r1, s1'];
  by [apply: apx_trans J1 | apply: apx_trans J2].
case: (@prog' _ _ _ _ _ _ _ sc g H C H1)=>// s' [q T].
exists sc, s', (Prog e' Un St Dr); split; last first.
- by move=>t1 t2 [] s2' X' T'; exists t1, t2, s2'.
move=>t2 [r1][z1][s1][R1 T1 S1]; exists r1, z1.
have S : similar t r1.
- by exists (join H1 R1); split; [apply: apx_joinL | apply: apx_joinR].
by case: (sim_tstep e S T T1)=> _ -> S'; split=>//; exists r1, z1, s1.
Qed.

(* preservation *)

(* states that if we step in a state satisfying the precondition *)
(* we move to a state from which the residual program still *)
(* satisfies the postcondition *)

Lemma pst_preserve V G A (pq : spec V G A) (e : STbin V G pq) (s : st V)
                   g e' s' sc :
         (pq g).1 s -> coh V s -> pstep e sc e' s s' -> vrf e' (pq g).2 s'.
Proof.
case: e=>e H H1 C /= X; split.
- case: (proj1 X Unfinished)=>[|t1' [t2'][X1 X2 T _]];
  by [apply: prog_unfin | apply: tst_coh T C].
move=>t {X} /(proj1 X) [t1'][t2'][X1 X2] T S.
case: (H _ _ H1 C)=>_ /(_ _ X1) Ls.
by move/(alw_tstep Ls): T=>{Ls}; apply: apx_aft S.
Qed.

(******************************)
(* continuity of constructors *)
(******************************)

(* I couldn't prove continuity for bnd *)
(* The problem is that the construction for tcat *)
(* is infinitary. It constructs a new tree by appending *)
(* the given tree to a whole set; this is preventing me *)
(* from picking the common set in a chain in which both *)
(* the program and the continuation reside *)

Lemma bnd_mono V A B (e1 e2 : ST V A) (k1 k2 : A -> ST V B) :
        e1 <== e2 -> k1 <== k2 -> bnd e1 k1 <== bnd e2 k2.
Proof.
move=>pf1 pf2 t [->|[t1][X2 X1]]; [left | right]=>//.
exists t1; split; last by apply: pf1 X1.
by case: X2=>t2 [Y1 Y2]; exists t2; split=>// x; apply: pf2 (Y2 _).
Qed.

Lemma bnd_mono' V A B : monotone (fun x => @bnd V A B x.1 x.2).
Proof.
case=>e1 k1 [e2 k2][/= H1 H2].
by apply: bnd_mono.
Qed.

Lemma bnd_cont V A B : continuous (fun x => @bnd V A B x.1 x.2).
Proof.
apply/contE2; exists (@bnd_mono' V A B)=>s t /=.
case=>[->|].
- case: s=>s [[c]]/= H1 H2; exists (bnd c.1 c.2).
  by split; [apply: prog_unfin | exists c].
case=>t' [H1 H2]; case: H2=>ts [H2 H2'].
pose ks := lim (proj2_chain s).
exists (bnd ts ks); split.
- by right; exists t'; split=>//; exists ts.
exists (ts, ks); split=>//.
case: H1=>q [?] H1; subst t.
(* and now a problem: we can't prove that ks \In proj2 s *)
(* what we need is a set ks \In proj2_chain s *)
(* which is not the limit of this chain, but some element of the chain *)
(* so that forall x, q x \In ks x *)
(* But I don't see how to do that; for every x, I can choose ks x *)
(* but I can't find their maximum, because chains only let you take a *)
(* maximum of two sets, and here we have a family indexed by x : A *)
(* of which we want to take maximum *)
Abort.

(* So, CPO's don't work with action trees *)
(* If we want to reconcile with CPO's, we may have to drop back *)
(* to Brookes' action traces *)

End Vrf.
