From Coq Require Import ssreflect ssrbool ssrfun.
From fcsl Require Import axioms pred prelude state world schedule.
From fcsl Require Import options.

(*************)
(* Processes *)
(*************)

(* In the semantics, we push morphisms onto a stack *)
(* so that when an action is encountered, it's first *)
(* changed according to all the morphisms in the stack *)

(* build up stacks from the right side *)
Inductive morphs V W := 
  mnil of W = V | mrcons U X of morphs V U & morph X U W.

(* build up stacks from the left side *)
Fixpoint mlcons U V W X (f : morph X U V) (cx : morphs V W) :=
  match cx with 
    mnil pf => mrcons (mnil (erefl _)) (cast _ pf f) 
  | mrcons _ _ cx g => mrcons (mlcons f cx) g
  end.

(* accumulated intermediate types of a morphism stack *) 
Fixpoint mtp V W (cx : morphs V W) : Type := 
  if cx is mrcons V B cx f then mtp cx * B else unit. 

(* adding new seed to the left of existing stack of seeds *)
Fixpoint tlcons U V W X (f : morph X U V) (cx : morphs V W) : 
                X -> mtp cx -> mtp (mlcons f cx) :=
  if cx is mrcons _ _ cx' _ return X -> mtp cx -> mtp (mlcons f cx) 
  then fun x '(ys, y) => (tlcons f x ys, y) else fun x _ => (tt, x).

Arguments tlcons [U V W X] f [cx] _ _.

Lemma tlcons_inj Vt V W X (cx : morphs V W) (f : morph X Vt V) 
                 x1 x2 (xs1 xs2 : mtp cx) : 
        tlcons f x1 xs1 = tlcons f x2 xs2 -> x1 = x2 /\ xs1 = xs2.
Proof.
elim: {W} cx xs1 xs2=>[W pf [[[->]]]|W V' X' cx IH f'] //.
by case=>a1 a2 [b1 b2][/IH][->->] ->. 
Qed.

Prenex Implicits tlcons_inj.

(* context of morphisms forms commuting squares of transitions *)
(* composing on the right *) 
Fixpoint tsquare V W (cx : morphs V W) : mtp cx -> mtp cx -> _ := 
  match cx return mtp cx -> mtp cx -> _ with
    mnil pf => fun x x' sv sv' tv sw sw' tw =>
      [/\ sw = cast (fun W => st W) pf sv, sw' = cast (fun W => st W) pf sv',
          tw = cast Tr pf tv & interp V tv sv sv']
  | mrcons U X cx f => fun '(xs, x) '(xs', x') sv sv' tv sw sw' tw =>  
      exists su su' tu, tsquare xs xs' sv sv' tv su su' tu /\
       [/\ [morph f] x sw su, [morph f] x' sw' su', 
           f x tu sw tw & interp W tw sw sw']
  end.

(* we can compose commuting squares on the left too *)
Lemma tsq_lcons U V W X (f : morph X U V) (cx : morphs V W) 
                x x' (xs xs' : mtp cx) su su' tu sw sw' tw : 
        tsquare (tlcons f x xs) (tlcons f x' xs') su su' tu sw sw' tw <->
        exists sv sv' tv, [/\ [morph f] x sv su, [morph f] x' sv' su', 
                              f x tu sv tv & interp U tu su su'] /\ 
          tsquare xs xs' sv sv' tv sw sw' tw.
Proof.
elim: {W} cx xs xs' sw sw' tw=>[W pf /= _ _|W A Y cx IH g /= [xs y][xs' y']] sw sw' tw.
- subst W; rewrite !eqc; split; last first.
  - by case=>sv [sv'][tv][[M M' Ef Hu]][->->-> Hv]; exists su, su', tu. 
  by case=>_ [_][_][[->->-> Hu]][M M' Ef Hw]; exists sw, sw', tw. 
split. 
- case=>sa [sa'][ta][/IH][sv][sv'][tv][[Mf Mf' Ef Hu S]][Mg Mg' Eg Hw].
  by exists sv, sv', tv; split=>//; exists sa, sa', ta. 
case=>sv [sv'][tv][[Mf Mf' Ef Hu]][sa][sa'][ta][S][Mg Mg' Eg Hw].
by exists sa, sa', ta; split=>//; apply/IH; exists sv, sv', tv. 
Qed.

Lemma tsq_coh V W (cx : morphs V W) (x x' : mtp cx) sv sv' tv sw sw' tw :
        coh W sw -> tsquare x x' sv sv' tv sw sw' tw -> coh V sv.
Proof.
elim: {W} cx x x' sv sv' sw sw' tv tw=>
[W pf /= _ _|W A Y cx IH g /= [xs y][xs' y']] sv sv' sw sw' tv tw Cw.
- by subst W; rewrite !eqc; case=><-.
by case=>su [su'][tu][S][M _ _ _]; apply: IH S; apply: morph_coh Cw M.
Qed.

Lemma tsq_int V W (cx : morphs V W) (x x' : mtp cx) sv sv' tv sw sw' tw :
        internal V tv -> coh W sw ->
        tsquare x x' sv sv' tv sw sw' tw -> internal W tw.
Proof.
elim: {W} cx x x' sv sv' sw sw' tv tw=>
[W pf /= _ _|W A Y cx IH g /= [xs y][xs' y']] sv sv' sw sw' tv tw I Cw.
- by subst W; case=>_ _ ->. 
case=>su [su'][tu][S][M M' Eg H]; move: (morph_coh Cw M)=>Cu.
by move/IH: (S)=>/(_ I Cu)=>J; apply: morph_int J Cw Eg.
Qed.

Lemma tsq_local V W (cx : morphs V W) (x x' : mtp cx) sv sv' sw sw' p tv tw : 
        coh W (sw \< p) -> other sw = other sw' ->
        tsquare x x' sv sv' tv (sw \> p) (sw' \> p) tw ->
        exists q, 
          exists sv1 sv1', [/\ sv = sv1 \> q, sv' = sv1' \> q &
          tsquare x x' (sv1 \< q) (sv1' \< q) tv (sw \< p) (sw' \< p) tw].
Proof.
elim: {W} cx x x' sv sv' sw sw' p tv tw=>
[W pf /= [][]|W U Y cx IH f /= [x a][x' a']] sv sv' sw sw' p tv tw' C' O;
move/cohZ: (C')=>C.
- subst W; rewrite !eqc; case=><-<-<-{sv sv' tv} H; exists p, sw, sw'.
  by rewrite !eqc; do !split=>//; apply: transZ H.
case=> su [su'][tu][S][M M' Ef H]; case: (morph_local f)=>fw L.
case: (L _ _ _ _ C' M)=>su1 [?] M1; subst su. 
move: (icoh H C)=>C2; move/transZ: H; move/(_ C' O)=>H. 
move: (icoh H C')=>C2'; case: (L _ _ _ _ C2' M')=>su1' [?] M1'; subst su'.
case/IH: S; first by apply/cohZ; apply: morph_coh M.
- by move: (morph_other C' C2' M1 M1'); rewrite !zgE; apply. 
move=>fv [sv1][sv1'][?? S']; subst sv sv'; exists fv, sv1, sv1'.
do !split=>//; exists (su1 \< fw p), (su1' \< fw p), tu.
by do !split=>//; apply: morph_tlocal.
Qed.

Lemma tsq_det X V W (cx : morphs V W) (a : action V X)
              (x x1' x2' : mtp cx) sv1 sv1' sv2 sv2' 
              sw sw1' sw2' tv1 tv2 tw1 tw2 v1 v2 :
        coh W sw ->
        a sv1 tv1 v1 -> tsquare x x1' sv1 sv1' tv1 sw sw1' tw1 ->
        a sv2 tv2 v2 -> tsquare x x2' sv2 sv2' tv2 sw sw2' tw2 ->
        [/\ v1 = v2, x1' = x2', sv1 = sv2, tv1 = tv2, 
            sv1' = sv2', sw1' = sw2' & tw1 = tw2].
Proof.
elim: {W} cx x x1' x2' sv1 sv1' sw sw1' sw2' tv1 tv2 tw1 tw2 v1 v2=>
[W pf /= [][][]|W U Y cx IH g /= [x b][x1 b1][x2 b2]] 
sv1 sv1' sw sw1' sw2' tv1 tv2 tw1 tw2 v1 v2 C.
- subst W; rewrite !eqc=>H1 [???] T1 H2 [???] T2.
  subst sv1 sv1' tv1 sv2 sv2' tv2; case: (act_func C H1 H2)=>??.
  by subst tw2 v2; rewrite (trans_func C T1 T2).
move=>H1 [su1][su1'][tu1][S1][M1 M1' Eg1 T1].
move=>H2 [su2][su2'][tu2][S2][M2 M2' Eg2 T2]. 
move: (icoh T1 C) (morph_coh C M1)=>Cw Cu; move: (tsq_coh Cu S1)=>Cv1.
move: (morph_func C M1 M2)=>? {M1 M2}; subst su2.
case: (IH _ _ _ _ _ _ _ _ _ _ _ _ _ _ Cu H1 S1 H2 S2)=>*.
subst v2 x2 sv2 sv2' su2' tv2 tu2.
move: (morph_tfunc C Eg1 Eg2)=>? {Eg1 Eg2}; subst tw2.
move: (trans_func C T1 T2)=>? {T1 T2}; subst sw2'.
by rewrite (morph_backw Cw Cw M1' M2').
Qed.

(* hide a number of components from the commuting squares *)
(* to leave just the ending states *)

Definition unwind_tr V W (cx : morphs V W) (tv : Tr V) (x x' : mtp cx) sw sw' := 
  exists sv sv' tw, tsquare x x' sv sv' tv sw sw' tw.

Arguments unwind_tr [V W] cx tv x x' sw sw'.

Lemma unwtr_coh V W (cx : morphs V W) (t : Tr V) (xs xs' : mtp cx) : 
        preserved (coh W) (unwind_tr cx t xs xs'). 
Proof.
case: cx xs xs'=>[pf /= [][]|V' X' cx g /= [xs a][xs' a']] 
      sw sw' /= [sv][sv'][tw] /=; first by case=>->-> _; subst W; apply: icoh. 
by case=>su [su'][tu][_][_ _ _]; apply: icoh.
Qed.

Lemma unwtr_other V W (cx : morphs V W) (t : Tr V) (xs xs' : mtp cx) : 
        otherP (unwind_tr cx t xs xs').
Proof.
case: cx xs xs'=>[pf /= [][]|U Y cx g /= [xs a][xs' a']]
sw sw' [sv][sv'][tw] /=; first by subst W; case=>->-> _; apply: trans_other. 
by case=>su [su'][tu][_][_ _ _]; apply: trans_other.
Qed.

Lemma unwtr_local V W (cx : morphs V W) (t : Tr V) (xs xs' : mtp cx) : 
        localP (coh W) (unwind_tr cx t xs xs'). 
Proof.
move=>p sw sw' C O [sv][sv'][tw] /(tsq_local C O).
by case=>q [sv1][sv1'][?? S]; exists (sv1 \< q), (sv1' \< q), tw. 
Qed.

Lemma unwtr_forw U V W X (cx : morphs U V) (f : morph X V W)
               (sim : simulation f) tu x (xs xs' : mtp cx) sw sv sv' :
        internal U tu -> 
        coh W sw -> [morph f] x sw sv -> sim x sw -> 
        unwind_tr cx tu xs xs' sv sv' ->
        exists x' sw', [/\ [morph f] x' sw' sv', sim x' sw' &
          unwind_tr (mrcons cx f) tu (xs, x) (xs', x') sw sw'].
Proof.
case: cx xs xs'=>[pf /= [][]|V' X' cx g /= [xs a][xs' a']] I Cw T S /= G. 
- subst V; case: G T=>su [su'][_][->->{sv sv'} _ G]; rewrite eqc=>T.
  case: (morph_forw I Cw T S G)=>tw [x'][sw'][Ef T' S' G'].
  by exists x', sw'; split=>//; exists su, su', tw, su, su', tu. 
case: G T=>su [su'][tu'] /= [sv1][sv1'][tv][Csq][M M' Eg H M2].
move: (morph_coh Cw M2)=>Cv; case: (morph_forw _ Cw M2 S H). 
- by apply: morph_int Eg=>//; apply: tsq_int I _ Csq; apply: morph_coh Cv M.
move=>tw [x'][sw'][Ef M2' S' H']; exists x', sw'; split=>//.
by exists su, su', tw, sv, sv', tu'; split=>//; exists sv1, sv1', tv. 
Qed.

(* we now need to unwind an action as follows *)
(* start with sw, follow the morphisms until we get sv *)
(* then run a in sv (not tv as is the case in tsquare) *)
(* and then go back *)

Definition unwind_act V W A (cx : morphs V W) (a : action V A) (v : A)
                            (x x' : mtp cx) sw sw' := 
  exists sv sv' tv tw, a sv tv v /\ tsquare x x' sv sv' tv sw sw' tw.

Arguments unwind_act [V W A] cx a v x x' sw sw'.

Lemma unwact_coh V W A (cx : morphs V W) (a : action V A) v (xs xs' : mtp cx) : 
        preserved (coh W) (unwind_act cx a v xs xs'). 
Proof.
move=>sw sw' [sv][sv'][tv][tw][H S]; apply: unwtr_coh.
by exists sv, sv', tw; apply: S. 
Qed.

Lemma unwact_det V W X (cx : morphs V W) (a : action V X) 
                       v1 v2 (xs xs1 xs2 : mtp cx) s s1 s2 :
        coh W s ->
        unwind_act cx a v1 xs xs1 s s1 -> 
        unwind_act cx a v2 xs xs2 s s2 ->
        [/\ v1 = v2, xs1 = xs2 & s1 = s2]. 
Proof.
move=>C [sv1][sv1'][tv1][tw1][H1 S1][sv2][sv2'][tv2][tw2][H2 S2].
by case: (tsq_det C H1 S1 H2 S2). 
Qed.

Lemma unwact_other V W A (cx : morphs V W) (a : action V A) v (xs xs' : mtp cx) : 
        otherP (unwind_act cx a v xs xs').
Proof.
move=>sw sw' [sv][sv'][tv][tw][H S]; apply: unwtr_other. 
by exists sv, sv', tw; apply: S.
Qed.

Lemma unwact_local V W A (cx : morphs V W) (a : action V A) v (xs xs' : mtp cx) : 
        localP (coh W) (unwind_act cx a v xs xs'). 
Proof.
move=>p sw sw' C O [sv][sv'][tv][tw][H S]. 
case: (tsq_local C O S)=>q [sv1][sv1'][?? S']; subst sv sv'.
exists (sv1 \< q), (sv1' \< q), tv, tw; split=>//.
by apply: actZ H; apply: tsq_coh S'.
Qed.

Lemma unwact_forw U V W X A (cx : morphs U V) (f : morph X V W)
               (sim : simulation f) (a : action U A) v x (xs xs' : mtp cx) sw sv sv' :
        coh W sw -> [morph f] x sw sv -> sim x sw -> 
        unwind_act cx a v xs xs' sv sv' ->
        exists x' sw', [/\ [morph f] x' sw' sv', sim x' sw' &
          unwind_act (mrcons cx f) a v (xs, x) (xs', x') sw sw'].
Proof.
case: cx xs xs'=>[pf /= [][]|V' X' cx g /= [xs j][xs' j']] Cw T S /= G. 
- subst V; case: G T=>su [su'][tv][tw][H][->->{sv sv'} _ G] T.
  case: (morph_forw (act_int (morph_coh Cw T) H) Cw T S G). 
  move=>tw' [x'][sw'][Ef T' S' G']; exists x', sw'; split=>//. 
  by exists su, su', tv, tw'; split=>//; exists su, su', tv. 
case: G T=>su [su'][tu'] /= [tv][Ha][sv1][sv1'][tv'][Csq][M M' Eg H M2].
move: (morph_coh Cw M2)=>Cv; move: (morph_coh Cv M)=>Cv1.
case: (morph_forw _ Cw M2 S H).
- apply: morph_int Cv Eg; apply: tsq_int (Cv1) (Csq).
  by apply: act_int Ha; apply: tsq_coh Cv1 Csq. 
move=>tw [x'][sw'][Ef M2' S' H']; exists x', sw'; split=>//.
exists su, su', tu', tw; split=>//; exists sv, sv', tv. 
by split=>//; exists sv1, sv1', tv'. 
Qed.

(*********)
(* Trees *)
(*********)

Inductive tree W A := 
    Unfinished 
  | Ret of A 
  | Act of action W A
  | Seq B of tree W B & B -> tree W A 
  | Par B1 B2 of tree W B1 & tree W B2 & B1 * B2 -> tree W A
  | Morf V B of B & morph B V W & tree V A.

Arguments Unfinished {W A}.
Arguments Ret {W A} _.
Arguments Act {W A} a.
Arguments Seq {W A B} _ _.
Arguments Par {W A B1 B2} _ _ _.
Arguments Morf {W A V B} _ _ _.

(*********************************************************)
(* Helper lemmas for getting equality of tree components *)
(*********************************************************)

Lemma seqE W A B1 B2 (t1 : tree W B1) (t2 : tree W B2) k1 k2 : 
        Seq t1 k1 = Seq t2 k2 <-> 
        [/\ B1 = B2, ijmeq t1 t2 &
            jmeq (fun B => B -> tree W A) k1 k2].
Proof.
split; case=>E; rewrite -{B2}E in t2 k2 *.
- by do 2![move/inj_pair2=><-].
by do 2![move/jmE=><-].
Qed.

Lemma seqEX W A B (t1 t2 : tree W B) k1 k2 : 
        Seq t1 k1 = Seq t2 k2 :> tree W A <-> t1 = t2 /\ k1 = k2.
Proof. by rewrite seqE; split=>[[_]|[->->] //]; do 2![move/jmE=>->]. Qed.

Lemma parE W A1 A2 B1 B2 C t1 t2 q1 q2 
             (r1 : A1 * B1 -> tree W C) (r2 : A2 * B2 -> tree W C) :
        Par t1 q1 r1 = Par t2 q2 r2 <-> 
        [/\ A1 = A2, B1 = B2, ijmeq t1 t2, ijmeq q1 q2 & 
            jmeq (fun B => B -> tree W C) r1 r2].
Proof. 
split; case=>E1 E2; rewrite -{A2}E1 -{B2}E2 in t2 q2 r1 r2 *.
- by do 2![move/inj_pair2=>->]; do 2![move/inj_pair2]; move=>->.
by do 3![move/jmE=>->].
Qed.

Lemma parEX W A1 A2 C (t1 t2 : tree W A1) (q1 q2 : tree W A2) r1 r2 :
        Par t1 q1 r1 = Par t2 q2 r2 :> tree W C <-> 
          [/\ t1 = t2, q1 = q2 & r1 = r2].
Proof. 
split; case; last by move=>->->->.
by do 2![move/inj_pair2=>->]; do 2![move/inj_pair2]=>->.
Qed.

Lemma morfE W A W1 W2 B1 B2 (b1 : B1) (b2 : B2)
              (tau1 : morph B1 W1 W) (tau2 : morph B2 W2 W)  
              (t1 : tree W1 A) (t2 : tree W2 A) : 
        Morf b1 tau1 t1 = Morf b2 tau2 t2 :> tree W A <-> 
        [/\ W1 = W2, B1 = B2, ijmeq b1 b2, jmeq (tree^~ A) t1 t2 & 
            @jmeq _ (fun x => morph x.1 x.2 W) (B1, W1) (B2, W2) tau1 tau2].
Proof. 
split; case=>E1 E2; subst W2 B2.  
- by move/inj_pair2=>-> /inj_pair2 /inj_pair2 -> /inj_pair2 ->.
by do 3![move/jmE=>->].
Qed.

Lemma morfEX W A V B (b1 b2 : B) (tau1 tau2 : morph B V W) (t1 t2 : tree V A) : 
        Morf b1 tau1 t1 = Morf b2 tau2 t2 :> tree W A <-> 
        [/\ b1 = b2, tau1 = tau2 & t1 = t2]. 
Proof. 
split; case; last by move=>->->->.
by move/inj_pair2=>-> /inj_pair2 /inj_pair2 -> /inj_pair2.
Qed.

(* catenating a tree with a set *)

Definition tcat W A B (t : tree W A) (k : A -> Pred (tree W B)) :=
  [Pred s | exists q, s = Seq t q /\ forall x, q x \In k x].

(* a schedule is good for a tree if it lands on the intended action *)

Fixpoint good W A (t : tree W A) sc :=
  match sc, t with 
    ChoiceAct, Act a => True
  | SeqRet, Seq _ (Ret _) _ => True
  | SeqStep sc', Seq _ t' _ => good t' sc'
  | ParRet, Par _ _ (Ret _) (Ret _) k => True
  | ParL sc', Par _ _ t' _ _ => good t' sc' 
  | ParR sc', Par _ _ _ t' _ => good t' sc'
  | MorfRet, Morf _ _ _ _ (Ret _) => True
  | MorfStep sc', Morf _ _ _ _ t' => good t' sc'
  | _, _ => False
  end.

(* Stepping through a tree via a schedule. Beta redexes return the same *)
(* initial state and reduced tree. None is returned by trees that get   *)
(* stuck because of a mismatched schedule. This boils down to           *)
(* actions that are safe but get stuck because of a wrong external      *)
(* choice. *)

Fixpoint tstep' W V A (cx : morphs V W) (t : tree V A) sc q xs1 xs2 s1 s2 := 
  match (sc, t) with
    (ChoiceAct, Act a) => 
       exists v, [/\ q = Ret v & unwind_act cx a v xs1 xs2 s1 s2]
  | (SeqRet, Seq _ (Ret v) k) => [/\ q = k v, xs2 = xs1 & s2 = s1]
  | (SeqStep sc', Seq _ t1 k1) => 
      exists t2, q = Seq t2 k1 /\ tstep' cx t1 sc' t2 xs1 xs2 s1 s2
  | (ParRet, Par _ _ (Ret v1) (Ret v2) k) => 
      [/\ q = k (v1, v2), xs2 = xs1 & s2 = s1]
  | (ParL sc', Par _ _ t1 t2 k) => 
      exists r1, q = Par r1 t2 k /\ tstep' cx t1 sc' r1 xs1 xs2 s1 s2
  | (ParR sc', Par _ _ t1 t2 k) => 
      exists r2, q = Par t1 r2 k /\ tstep' cx t2 sc' r2 xs1 xs2 s1 s2
  | (MorfRet, Morf _ _ _ _ (Ret v)) => [/\ q = Ret v, xs2 = xs1 & s2 = s1]
  | (MorfStep sc', Morf U X x1 f t1) => 
       exists x2 t2, q = Morf x2 f t2 /\ 
         tstep' _ t1 sc' t2 (tlcons f x1 xs1) (tlcons f x2 xs2) s1 s2
  | _ => False
  end.

Arguments tstep' [W V A] cx t sc q xs1 xs2 s1 s2.

(* The top level call to tstep' needs its own name *)
Definition tstep W A (t : tree W A) sc q s1 s2 := 
  tstep' (mnil (erefl W)) t sc q tt tt s1 s2.

(* views for tstep', organized to match on the tree *)
(* as we typically know the structure of the tree *)

Section TstepViews.
Variables (W V : resource) (cx : morphs V W).

Lemma tstUnfin A (t : tree V A) sc x1 x2 s1 s2 :
        tstep' cx Unfinished sc t x1 s1 x2 s2 <-> False.
Proof. by split=>//; case: sc. Qed.

Lemma tstRet A (t : tree V A) sc v x1 x2 s1 s2 :
        tstep' cx (Ret v) sc t x1 s1 x2 s2 <-> False.
Proof. by split=>//; case: sc. Qed.

Lemma tstAct A (t : tree V A) sc a x1 x2 s1 s2 :
        tstep' cx (Act a) sc t x1 x2 s1 s2 <->
        exists v, [/\ sc = ChoiceAct, t = Ret v & unwind_act cx a v x1 x2 s1 s2].
Proof.
split; last by case=>v [->-> C]; exists v.
by case: sc=>//; case=>v [->]; exists v.
Qed.

Lemma tstSeq A B (t : tree V B) k sc (q : tree V A) x1 x2 s1 s2 :
        tstep' cx (Seq t k) sc q x1 x2 s1 s2 <->
        (exists v, [/\ sc = SeqRet, t = Ret v, q = k v, x2 = x1 & s2 = s1]) \/
         exists sc' p',
           [/\ sc = SeqStep sc', q = Seq p' k & tstep' cx t sc' p' x1 x2 s1 s2].
Proof.
split; last first.
- case; first by case=>v [->->->->->]. 
  by case=>sc' [t'][->->] H; exists t'.
case: sc=>//. 
- by case: t=>// v [->->->]; left; exists v. 
by move=>sc [t'][-> S']; right; exists sc, t'. 
Qed.

Lemma tstPar A1 A2 B t1 t2 sc k (q : tree V B) x1 x2 s1 s2 :
        tstep' cx (Par t1 t2 k) sc q x1 x2 s1 s2 <-> 
        (sc = ParRet /\ exists (v1 : A1) (v2 : A2),
           [/\ t1 = Ret v1, t2 = Ret v2, q = k (v1, v2), x2 = x1 & s2 = s1]) \/
        exists t1' t2' sc',
          q = Par t1' t2' k /\
          ([/\ sc = ParL sc', t2' = t2 & tstep' cx t1 sc' t1' x1 x2 s1 s2]  \/
           [/\ sc = ParR sc', t1' = t1 & tstep' cx t2 sc' t2' x1 x2 s1 s2]).
Proof.
split; last first.
- case; first by case=>-> [v1][v2][->->->->->]. 
  case=>t1' [t2'][sc'][->]; case; case=>->-> H; 
  by [exists t1' | exists t2']. 
case: sc=>//.
- case: t1=>// v1; case: t2=>// v2 [->->->]; left. 
  by split=>//; exists v1, v2.
- by move=>sc [t1'][->]; right; exists t1', t2, sc; split=>//; left. 
by move=>sc [t2'][->]; right; exists t1, t2', sc; split=>//; right.
Qed.

Lemma tstMorf X U A (tau : morph X U V) (t1 : tree U A) q sc x1 y1 x2 s1 s2 :
        tstep' cx (Morf y1 tau t1) sc q x1 x2 s1 s2 <-> 
        (exists v, [/\ sc = MorfRet, t1 = Ret v, q = Ret v, x2 = x1 & s2 = s1]) \/
        (exists sc' y2 t2, 
          [/\ sc = MorfStep sc', q = Morf y2 tau t2 &
              tstep' _ t1 sc' t2 (tlcons tau y1 x1) (tlcons tau y2 x2) s1 s2]).
Proof.
split; last first.
- case; first by case=>v [->->->->->].
  by case=>sc' [y2][t2][->-> T]; exists y2, t2. 
case: sc=>//.
- by case: t1=>// v [->->->]; left; exists v.
by move=>sc' [y2][t2][-> T]; right; exists sc', y2, t2. 
Qed.

(* we can put these together, in a lemma that serves as a *)
(* analysis over the tree, but with more structure than ordinary casing *)

Lemma tst A (t : tree V A) sc q xs1 xs2 s1 s2 :
  tstep' cx t sc q xs1 xs2 s1 s2 <->
    [\/ exists a v, 
        [/\ sc = ChoiceAct, t = Act a, q = Ret v
            & unwind_act cx a v xs1 xs2 s1 s2],
        exists B (p' : tree V B) k,
          t = Seq p' k /\ 
           ((exists v,
                [/\ sc = SeqRet, p' = Ret v, q = k v, xs2 = xs1 & s2 = s1]) \/
            (exists sc' q',
                [/\ sc = SeqStep sc', q = Seq q' k &
                    tstep' cx p' sc' q' xs1 xs2 s1 s2])),
        exists B1 B2 (t1 : tree V B1) (t2 : tree V B2) k, 
                [/\ t = Par t1 t2 k &
                    (exists v1 v2, 
                        [/\ sc = ParRet, t1 = Ret v1,
                            t2 = Ret v2, xs2 = xs1, s2 = s1 & 
                            q = k (v1, v2)]) \/
                    exists t1' t2' sc',
                      q = Par t1' t2' k /\
                      ([/\ sc = ParL sc', t2' = t2 &
                           tstep' cx t1 sc' t1' xs1 xs2 s1 s2] \/
                       [/\ sc = ParR sc',
                           t1' = t1 & tstep' cx t2 sc' t2' xs1 xs2 s1 s2])] |
        exists X U (tau : morph X U V) x1 t1,
                t = Morf x1 tau t1
                /\  ((exists v,
                         [/\ sc = MorfRet, t1 = Ret v, q = Ret v, 
                             xs2 = xs1 & s2 = s1]) \/
                     exists sc' x2 t2,
                       [/\ sc = MorfStep sc', q = Morf x2 tau t2 & 
                           tstep' _ t1 sc' t2 (tlcons tau x1 xs1)
                                              (tlcons tau x2 xs2) s1 s2])].
Proof.
split; last first.
- case. 
  - by case=>a [v][->->->]; exists v.
  - case=>B [t'][k][->]; case; first by case=>v [->->->->].
    by case=>sc' [q'][->-> H]; exists q'. 
  - case=>B1 [B2][t1][t2][k][->]; case; first by case=>v1 [v2][->->->->->]. 
    by case=>t1' [t2'][sc'][->]; case; case=>->-> H; [exists t1' | exists t2'].
  case=>U [Y][tau'][y1][t1][->]; case; first by case=>v [->->->->->]. 
  by case=>sc' [y2][t2][->-> T]; exists y2, t2. 
case: t=>[|v|a|B t k|B1 B2 t1 t2 k|U X y1 tau' t1]. 
- by rewrite tstUnfin.
- by rewrite tstRet.
- by rewrite tstAct; case=>v [->-> H]; apply: Or41; exists a, v.
- rewrite tstSeq; case.
  - case=>v [->->->->->]; apply: Or42; exists B, (Ret v), k.
    by split=>//; left; exists v. 
  case=>sc' [t'][->-> H]; apply: Or42; exists B, t, k; split=>//.
  by right; exists sc', t'. 
- rewrite tstPar; case.
  - case=>-> [v1][v2][->->->->->]; apply: Or43. 
    exists B1, B2, (Ret v1), (Ret v2), k. 
    by split=>//; left; exists v1, v2. 
  case=>t1' [t2'][sc'][->]; case.
  - case=>->-> H; apply: Or43; exists B1, B2, t1, t2, k; split=>//.
    by right; exists t1', t2, sc'; split=>//; left.
  case=>->-> H; apply: Or43; exists B1, B2, t1, t2, k; split=>//.
  by right; exists t1, t2', sc'; split=>//; right. 
rewrite tstMorf; case.
- case=>v [->->->->->]; apply: Or44. exists X, U, tau', y1, (Ret v). 
  by split=>//; left; exists v.
case=>sc' [y2][t2][->-> H]; apply: Or44. 
by exists X, U, tau', y1, t1; split=>//; right; exists sc', y2, t2. 
Qed.

(* We may be given both the tree and the schedule *)
 
Lemma tstActS A a (t : tree V A) x1 x2 s1 s2 : 
        tstep' cx (Act a) ChoiceAct t x1 x2 s1 s2 <->
        exists v, [/\ t = Ret v & unwind_act cx a v x1 x2 s1 s2].
Proof. 
rewrite tstAct; split; case=>v. 
- by case=>_ ->; exists v.
by case=>->; exists v.
Qed.

Lemma tstSeqS A B (t : tree V B) k sc (q : tree V A) x1 x2 s1 s2 :
        tstep' cx (Seq t k) sc q x1 x2 s1 s2 <->
        match sc with 
          SeqRet => exists v, 
            [/\ t = Ret v, q = k v, x2 = x1 & s2 = s1]
        | SeqStep sc' => exists p', 
            [/\ q = Seq p' k & tstep' cx t sc' p' x1 x2 s1 s2]
        | _ => False
        end.
Proof.
case: sc=>[||sc||sc|sc||sc]; try by [split=>//; case].
split; last by case=>v [->->->]. 
by case: t=>//= v [->->]; exists v.
Qed.

Lemma tstParS A1 A2 B t1 t2 sc k (q : tree V B) x1 x2 s1 s2 :
        tstep' cx (Par t1 t2 k) sc q x1 x2 s1 s2 <-> 
        match sc with 
          ParRet => exists (v1 : A1) (v2 : A2),
            [/\ t1 = Ret v1, t2 = Ret v2, q = k (v1, v2), x2 = x1 & s2 = s1]
        | ParL sc' => exists t1',
            q = Par t1' t2 k /\ tstep' cx t1 sc' t1' x1 x2 s1 s2
        | ParR sc' => exists t2',
            q = Par t1 t2' k /\ tstep' cx t2 sc' t2' x1 x2 s1 s2
        | _ => False
        end.
Proof.
case: sc=>[||sc||sc|sc||sc]; try by split=>//; case.
split; last by case=>v1 [v2][->->->->].
case: t1=>// v1; case: t2=>// v2. 
by case=>->->; exists v1, v2.
Qed.

Lemma tstMorphS U Y A (tau : morph Y U V) t1 (q : tree V A) sc x1 y1 x2 s1 s2 :
        tstep' cx (Morf y1 tau t1) sc q x1 x2 s1 s2 <->
        match sc with 
          MorfRet => exists v, 
            [/\ t1 = Ret v, q = Ret v, x2 = x1 & s2 = s1]
        | MorfStep sc' => exists y2 t2,
            [/\ q = Morf y2 tau t2 &
                tstep' _ t1 sc' t2 (tlcons tau y1 x1) (tlcons tau y2 x2) s1 s2]
        | _ => False
        end.
Proof.
case: sc=>[||sc||sc|sc||sc]; try by split=>//; case.
split; last by case=>v [->->-> S]. 
by case: t1=>// v [->->->]; exists v. 
Qed.

(* this view if we are given a schedule but not a tree *)
(* typically used when we have to induct on schedule *)

Lemma tstS A (t : tree V A) sc q x1 x2 s1 s2 :
        tstep' cx t sc q x1 x2 s1 s2 <->
        match sc with 
          ChoiceAct => exists a v,
            [/\ t = Act a, q = Ret v & unwind_act cx a v x1 x2 s1 s2] 
        | SeqRet => exists B (v : B) k,
            [/\ t = Seq (Ret v) k, q = k v, x2 = x1 & s2 = s1]
        | SeqStep sc' => exists B (t' : tree V B) q' k, 
            [/\ t = Seq t' k, q = Seq q' k & tstep' cx t' sc' q' x1 x2 s1 s2]
        | ParRet => exists B1 B2 (v1 : B1) (v2 : B2) k,
            [/\ t = Par (Ret v1) (Ret v2) k, q = k (v1, v2), x2 = x1 & s2 = s1]
        | ParL sc' => exists B1 B2 (r : tree V B2) (t' q' : tree V B1) k,
            [/\ t = Par t' r k, q = Par q' r k & tstep' cx t' sc' q' x1 x2 s1 s2]
        | ParR sc' => exists B1 B2 (r : tree V B1) (t' q' : tree V B2) k,
            [/\ t = Par r t' k, q = Par r q' k & tstep' cx t' sc' q' x1 x2 s1 s2]
        | MorfRet => exists U Y (tau' : morph Y U V) (y1 : Y) v, 
            [/\ t = Morf y1 tau' (Ret v), q = Ret v, x2 = x1 & s2 = s1]
        | MorfStep sc' => exists U Y (tau : morph Y U V) (y1 y2 : Y) t1 t2,
            [/\ t = Morf y1 tau t1, q = Morf y2 tau t2 &
                tstep' _ t1 sc' t2 (tlcons tau y1 x1) (tlcons tau y2 x2) s1 s2]
        end.
Proof.
case:sc=>[||sc||sc|sc||sc].
- split; last by case=>a [v][->-> H]; exists v.
  by case: t=>// a [v][->]; exists a, v.
- split; last by case=>B [v][k][->]->->.
  by case: t=>// B; case=>// v k [->->]; exists B, v, k.
- split; last by case=>B [t'][q'][k][->->] H; exists q'.
  by case: t=>//= B t' k [q'][-> H]; exists B, t', q', k.
- split; last by case=>B1 [B2][v1][v2][k][->->->]. 
  case: t=>// B1 B2; case=>// v1; case=>//= v2 k [->->]. 
  by exists B1, B2, v1, v2, k.  
- split; last by case=>B1 [B2][r][t'][q'][k][->->] H; exists q'.
  case: t=>// B1 B2 t' r k /= [q'][-> H].
  by exists B1, B2, r, t', q', k.
- split; last by case=>B1 [B2][r][t'][q'][k][->->] H; exists q'.
  case: t=>// B1 B2 r t' k /= [q'][-> H].
  by exists B1, B2, r, t', q', k.
- split; last by case=>U [Y][tau'][y1][v][->->->->]. 
  by case: t=>// U Y y1 tau' [] //= v []; exists U, Y, tau', y1, v. 
split; last first. 
- by case=>U [Y][tau][y1][y2][t1][t2][->-> T]; exists y2, t2. 
case:t=>// U Y y1 tau t1 [y2][t2][-> T]. 
by exists U, Y, tau, y1, y2, t1, t2. 
Qed.

(* some additional lemmas on par *)

Lemma tstParT A B C (t1 t2 : tree V A) sc q1 q2 x1 x2 s1 s2 
              (k1 k2 : A * B -> tree V C) : 
        tstep' cx (Par t1 q1 k1) sc (Par t2 q2 k2) x1 x2 s1 s2 <->
        (k2 = k1 /\ 
           exists sc', 
             [/\ sc = ParL sc', q2 = q1 & tstep' cx t1 sc' t2 x1 x2 s1 s2] \/ 
             [/\ sc = ParR sc', t2 = t1 & tstep' cx q1 sc' q2 x1 x2 s1 s2]) \/
        exists x, exists y, 
           [/\ sc = ParRet, t1 = Ret x, q1 = Ret y, x2 = x1, s2 = s1 &
               k1 (x, y) = Par t2 q2 k2].
Proof.
rewrite tstPar; split.
- case; first by case=>-> [x][y][->-> E ->]; right; exists x, y. 
  by case=>t1' [t2'][sc'][]; case/parEX=>->->-> H; left; split=>//; exists sc'.
case; first by case=>-> [sc' H]; right; exists t2, q2, sc'.  
by case=>x [y][->->->-> E]; left; split=>//; exists x, y. 
Qed.

Lemma tstParL A B C t1 t2 sc q (k : A * B -> tree V C) x1 x2 s1 s2 : 
        tstep' cx t1 sc t2 x1 x2 s1 s2 -> 
        tstep' cx (Par t1 q k) (ParL sc) (Par t2 q k) x1 x2 s1 s2. 
Proof. by exists t2. Qed.

Lemma tstParR A B C t1 t2 sc q (k : A * B -> tree V C) x1 x2 s1 s2 : 
        tstep' cx t1 sc t2 x1 x2 s1 s2 -> 
        tstep' cx (Par q t1 k) (ParR sc) (Par q t2 k) x1 x2 s1 s2. 
Proof. by exists t2. Qed.

End TstepViews.


(*************************************************************)
(* We need the same lemmas for trees that we had for actions *)
(*************************************************************)

Definition tsafe W A (t : tree W A) sc s := 
  good t sc -> exists s' q, tstep t sc q s s'.

Arguments tsafe [W A] t sc s. 

Lemma tst_wcovered W V A (cx : morphs V W) (t : tree V A) sc q x1 x2 s1 s2 : 
        coh W s1 -> tstep' cx t sc q x1 x2 s1 s2 -> 
        (x1, s1) = (x2, s2) \/ 
        exists tv, [/\ internal V tv & unwind_tr cx tv x1 x2 s1 s2].
Proof.
elim: sc W V A cx t q x1 s1 x2 s2=>[||sc IH||sc IH|sc IH||sc IH] 
W V A cx t q x1 s1 x2 s2 C; rewrite tstS.
- case=>a [v][_ _][sv][sv'][tv][tw][H S]; right. 
  exists tv; split; last by exists sv, sv', tw.  
  by apply: act_int (tsq_coh C S) H.
- by case=>B [v][k][_ _ ->->]; left. 
- by case=>B [t1][t2][k][_ _]; apply: IH. 
- by case=>B1 [B2][v1][v2][k][_ _ ->->]; left.
- by case=>B1 [B2][r][t1][t2][k][_ _]; apply: IH.
- by case=>B1 [B2][r][t1][t2][k][_ _]; apply: IH.
- by case=>U [Y][tau][y1][v][_ _ ->->]; left.
case=>U [Y][tau][y1][y2][t1][t2][_ _] /IH. move/(_ C). case; case.
- by case/tlcons_inj=>_ ->->; left. 
move=>a [I H]; right; case: H=>su [su'][tw] /tsq_lcons. 
case=>sv [sv'][tv][[M1 M2 Et H]] Csq; exists tv; split. 
- by apply: morph_int I _ (Et); apply: tsq_coh Csq.
by exists sv, sv', tw. 
Qed.

Lemma tst_ind W V A (cx : morphs V W) (t : tree V A) sc q x1 x2 s1 s2 : 
        tstep' cx t sc q x1 x2 s1 s2 -> good t sc /\ other s1 = other s2.
Proof. 
elim: sc W V A cx t q x1 s1 x2 s2 =>[||sc IH||sc IH|sc IH||sc IH] 
W V X A cx q x1 s1 x2 s2; rewrite tstS.
- by case=>a [v][-> _ H]; rewrite (unwact_other H). 
- by case=>B [v][k][-> _ _ ->]. 
- by case=>B [t1][t2][t][-> _]; apply: IH.
- by case=>B1 [B2][v1][v2][k][-> _ _ ->].
- by case=>B1 [B2][t][t1][t2][k][-> _]; apply: IH.
- by case=>B1 [B2][t][t1][t2][k][-> _]; apply: IH.
- by case=>U [Y][tau'][y][v][-> _ _ ->]. 
by case=>U [Y][tau'][y1][y2][t1][t2][-> _]; apply: IH.
Qed.

Prenex Implicits tst_wcovered tst_ind. 

Lemma tst_coh W V A (cx : morphs V W) (t : tree V A) sc q x x' : 
        preserved (coh W) (tstep' cx t sc q x x'). 
Proof. 
move=>s s' T C; case/(tst_wcovered C): T; first by case=>_ <-.
by case=>a [I] /unwtr_coh; apply.
Qed.

Lemma tst_osteps W A (t : tree W A) sc q s1 s2 : 
        coh W s1 -> tstep t sc q s1 s2 -> osteps W (tr s1) (tr s2). 
Proof.
move=>C /(tst_wcovered C) [[->]//|].
case=>a [I][sw][sw'][tw][->-> _] H; apply: iter1.
by apply: (iostep a); rewrite !trI.  
Qed.

Lemma tst_msteps W A (t : tree W A) sc q s1 s2 : 
        coh W s1 -> tstep t sc q s1 s2 -> msteps (id_morph W tt) s1 s2.
Proof.
move=>C /(tst_wcovered C) [[->]//|]. 
case=>a [I][sw][sw'][tw][->-> _] H; apply: iter1.
by apply: SelfMStep I _ H.
Qed.

Lemma tst_safe W A (t : tree W A) sc q s1 s2 : 
        tstep t sc q s1 s2 -> tsafe t sc s1 /\ good t sc.
Proof. by move=>H; split; [exists s2, q | case/tst_ind: H]. Qed.

Lemma tst_progress W A (t : tree W A) sc s : 
        tsafe t sc s -> good t sc -> exists s' q, tstep t sc q s s'.
Proof. by apply. Qed.

Lemma tst_det W V A (cx : morphs V W) (t : tree V A) sc x s q1 q2 x1 x2 s1 s2 : 
        coh W s ->
        tstep' cx t sc q1 x x1 s s1 -> 
        tstep' cx t sc q2 x x2 s s2 -> [/\ q1 = q2, x1 = x2 & s1 = s2].
Proof. 
elim: sc W V A cx x t s x1 q1 s1 x2 q2 s2=>[||sc IH||sc IH|sc IH||sc IH] 
W V A cx x t s x1 q1 s1 x2 q2 s2 C; rewrite tstS /=.
- by case=>a [v][->-> H1][v'][-> H2]; case:(unwact_det C H1 H2)=>->->->.
- by case=>B[v][k][->->->->][->->->]. 
- case=>B [t1][q][k][->-> H][t2][->]. 
  by case/(IH _ _ _ _ _ _ _ _ _ _ _ _ _ C H)=>->->->.
- by case=>B1 [B2][v1][v2][k][->->->->][->->->].
- case=>B1 [B2][r][t'][q1'][k][->-> H][q2'][->].
  by case/(IH _ _ _ _ _ _ _ _ _ _ _ _ _ C H)=>->->->.
- case=>B1 [B2][r][t'][q1'][k][->-> H][q2'][->].
  by case/(IH _ _ _ _ _ _ _ _ _ _ _ _ _ C H)=>->->->.
- by case=>U [Y][tau'][y1][v][->->->->][->->->].
case=>U [Y][tau'][y1][y2][t1][t2][->-> H][t3][t4][->].
by case/(IH _ _ _ _ _ _ _ _ _ _ _ _ _ C H)=>-> /tlcons_inj [->->].
Qed.

Lemma tst_other W A (t : tree W A) sc q : otherP (tstep t sc q). 
Proof. by move=>s s' /tst_ind []. Qed.

Lemma tst_local W V A (cx : morphs V W) (t : tree V A) sc q x x' : 
        localP (coh W) (tstep' cx t sc q x x'). 
Proof.
elim: sc W V A cx t q x x'=>[||sc IH||sc IH|sc IH||sc IH] 
W V A cx t q x x' p s s' C O; rewrite tstS.
- by case=>a [v][->-> H]; exists v; split=>//; apply: unwact_local H.
- by case=>B [v][k][->->->]; rewrite !zg_unlock O; case=>->->.
- by case=>B [v][t'][q'][->->] T; exists t'; split=>//; apply: IH.
- by case=>B1 [B2][v1][v2][k][->->->]; rewrite !zg_unlock O; case=>->->. 
- by case=>B1 [B2][r][t'][q'][k][->->] T; exists q'; split=>//; apply: IH. 
- by case=>B1 [B2][r][t'][q'][k][->->] T; exists q'; split=>//; apply: IH. 
- by case=>U [Y][tau'][y1][v][->->->]; rewrite !zg_unlock O; case=>->->. 
by case=>U [Y][f'][y1][y2][t1][t2][->->]; exists y2, t2; split=>//; apply: IH.
Qed.

Lemma tst_zag W A (t : tree W A) sc q : zagP (coh W) (tstep t sc q).
Proof. by apply/zagzigE; [apply/tst_other | apply: tst_local]. Qed.

Lemma tst_safe_mono W A (t : tree W A) sc : monoP (coh W) (tsafe t sc). 
Proof. 
by move=>p ? C H /H [?][q] /(tst_zag C) [k][]; exists (k \< p), q. 
Qed.

Lemma tst_frame W A (t : tree W A) sc q : 
        frameP (coh W) (tsafe t sc) (tstep t sc q). 
Proof.
move=>p s s1 C S H; case/tst_safe: (H)=>_{}/S [s'][q'] H'.
by case/(tst_zag C): H' (H')=>s2 [->] /(tst_det C H) [-> _ ->]; exists s2. 
Qed.

Lemma tst_forw Vt V W X A cx (f : morph X V W) (sim : simulation f) 
               (t t' : tree Vt A) sc (xs xs' : mtp cx) x sw sv sv' : 
        coh W sw -> [morph f] x sw sv -> sim x sw -> 
        tstep' cx t sc t' xs xs' sv sv' -> 
        exists x' sw', [/\ [morph f] x' sw' sv', sim x' sw' & 
                           tstep' (mrcons cx f) t sc t' (xs, x) (xs', x') sw sw']. 
Proof.
elim: sc Vt A cx t t' xs xs'=>[||sc IH||sc IH|sc IH||sc IH] Vt A cx t t' 
xs xs' Cw T S; rewrite tstS.
- case=>a [v][->->] /(unwact_forw Cw T S) [x'][sw'][].
  by exists x', sw'; split=>//; exists v. 
- by case=>B [v][k][->->->->]; exists x, sw. 
- case=>X' [t1][q1][k][->->] /(IH _ _ _ _ _ _ _ Cw T S) [x'][sw'][].
  by exists x', sw'; split=>//; exists q1.
- by case=>B1 [B2][v1][v2][k][->->->->]; exists x, sw. 
- case=>B1 [B2][r][t1][q1][k][->->] /(IH _ _ _ _ _ _ _ Cw T S) [x'][sw'][]. 
  by exists x', sw'; split=>//; exists q1.
- case=>B1 [B2][r][t1][q1][k][->->] /(IH _ _ _ _ _ _ _ Cw T S) [x'][sw'][]. 
  by exists x', sw'; split=>//; exists q1.
- by case=>U [Y][tau'][y1][v][->->->->]; exists x, sw. 
case=>U [Y][tau'][y1][y2][t1][t2][->->] /(IH _ _ _ _ _ _ _ Cw T S) [x'][sw'][]. 
by exists x', sw'; split=>//; exists y2, t2. 
Qed.

Lemma tst_backw Vt V W X A cx (f : morph X V W) (sim : simulation f) 
                (t t' : tree Vt A) sc x x' (xs xs' : mtp cx) sw sv sw' : 
        coh W sw -> [morph f] x sw sv -> sim x sw ->
        tstep' (mrcons cx f) t sc t' (xs, x) (xs', x') sw sw' -> 
        exists sv', [/\ [morph f] x' sw' sv', sim x' sw' & 
                        tstep' cx t sc t' xs xs' sv sv'].
Proof.
elim: sc Vt A cx t t' xs xs'=>[||sc IH||sc IH|sc IH||sc IH] Vt A cx t q
xs xs' Cw T S; rewrite tstS.
- case=>a [v][->->] H. 
  case: (H)=>su [su'][tv][tw] /= [Ha][sv1][sv1'][tv'][Sc][M M' Ef Hw].
  move: (morph_func Cw T M)=>?; subst sv1.
  have U : unwind_act cx a v xs xs' sv sv1' by exists su, su', tv, tv'.  
  exists sv1'; split=>//; last by exists v.
  case/(unwact_forw Cw T S): U=>y [sw2][_ S'].
  by case/(unwact_det Cw H)=>_ [->->]. 
- by case=>B [v][k][->->][->->] ->; exists sv. 
- case=>B [t'][q'][k][->->] /(IH _ _ _ _ _ _ _ Cw T S) [z'][].
  by exists z'; split=>//; exists q'.
- by case=>B1 [B2][v1][v2][k][->->][->->] ->; exists sv.
- case=>B1 [B2][r][t'][q'][k][->->] /(IH _ _ _ _ _ _ _ Cw T S) [z'][].
  by exists z'; split=>//; exists q'.
- case=>B1 [B2][r][t'][q'][k][->->] /(IH _ _ _ _ _ _ _ Cw T S) [z'][].
  by exists z'; split=>//; exists q'.
- by case=>U [Y][tau'][y1][v][->->][->->] ->; exists sv. 
case=>U [Y][tau'][y1][y2][t1][t2] /= [->->] /(IH _ _ _ _ _ _ _ Cw T S) [z'][].
by exists z'; split=>//; exists y2, t2.
Qed.
