From Coq Require Import ssreflect ssrbool ssrfun Setoid.
From fcsl Require Import axioms pred prelude pcm morphism notation.
From fcsl Require Import options.

(*************************************************************************)
(* Definition of:                                                        *)
(* - subjective state, self, joint, other, zig, zag, state pairing and   *)
(*   projection, and state transposition.                                *)
(*************************************************************************)

(********************)
(********************)
(* Subjective state *)
(********************)
(********************)

Structure state (U : pcm) T := State {self : U; joint : T; other : U}.

(* stEX and stE are temporary lemmas; will be replaced by morphism-based *)
(* versions and the end of the file. That's why I prime them for now *)

Lemma stEX' U T (s : state U T) : s = State (self s) (joint s) (other s).
Proof. by case: s. Qed.

Lemma stE' U T (s1 s2 : state U T) :
        s1 = s2 <-> [/\ self s1 = self s2,
                        joint s1 = joint s2 &
                        other s1 = other s2].
Proof. by rewrite {1}[s1]stEX' {1}[s2]stEX'; split; case=>->->->. Qed.

Definition outs U T (s : state U T) := State Unit (joint s) (self s \+ other s).

(* helper predicate for splitting *)
(* this should really be a PCM sans unit, but its done using relations *)
(* becase uncomputability *)
Definition splits U T (s s1 s2 : state U T) :=
  [/\ self s = self s1 \+ self s2, other s1 = self s2 \+ other s,
      other s2 = self s1 \+ other s, joint s = joint s1 & joint s = joint s2].

Lemma spC U T (s s1 s2 : state U T) : splits s s1 s2 -> splits s s2 s1.
Proof. by case=>X X1 X2 J1 J2; split=>//; rewrite joinC. Qed.

Lemma spA U T (s s1 s2 s11 s12 : state U T) :
        splits s s1 s2 -> splits s1 s11 s12 ->
        let: s3 := State (self s12 \+ self s2) (joint s) (self s11 \+ other s)
        in splits s s11 s3 /\ splits s3 s12 s2.
Proof.
case=>X X1 X2 J1 J2 [Y Y1 Y2 H1 H2].
do !split=>//=.
- by rewrite X Y joinA.
- by rewrite Y1 X1 joinA.
- by rewrite J1.
- by rewrite Y2 X1 joinCA.
- by rewrite X2 Y joinCA joinA.
by rewrite J1.
Qed.

Lemma spOL U T (s : state U T) : splits s (outs s) s.
Proof. by rewrite /splits !unitL. Qed.

Lemma spOR U T (s : state U T) : splits s s (outs s).
Proof. by apply: spC; apply: spOL. Qed.

Hint Resolve spOL spOR : core.

Definition star U T (P1 P2 : Pred (state U T)) :=
  [Pred s | exists s1 s2, [/\ splits s s1 s2, s1 \In P1 & s2 \In P2]].

Add Parametric Morphism U T : (@star U T) with signature
  @EqPredType _ _ ==> @EqPredType _ _ ==> @EqPredType _ _ as predstar_morph.
Proof.
move=>x1 y1 E1 x2 y2 E2; split; case=>s1 [s2][S1].
- by rewrite E1 E2; exists s1, s2.
by rewrite -E1 -E2; exists s1, s2.
Qed.

Lemma starC U T (P1 P2 : Pred (state U T)) : star P1 P2 <~> star P2 P1.
Proof. by split; case=>s1 [s2][/spC]; exists s2, s1. Qed.

Lemma starCA U T (P1 P2 P3 : Pred (state U T)) :
        star (star P1 P2) P3 <~> star (star P1 P3) P2.
Proof.
split; case=>s1 [s2][S][s11][s12][/spC S1 H1 H2 H3];
case: (spA S S1)=>/spC; set s' := State _ _ _ => S' S1';
by exists s', s12; split=>//; exists s11, s2.
Qed.

Lemma starA U T (P1 P2 P3 : Pred (state U T)) :
        star P1 (star P2 P3) <~> star (star P1 P2) P3.
Proof. by rewrite starC starCA (starC P2). Qed.

Lemma starAC U T (P1 P2 P3 : Pred (state U T)) :
        star P1 (star P2 P3) <~> star P2 (star P1 P3).
Proof. by rewrite starC starA starCA. Qed.

(* Definitions that should be opaque are hidden behind a module *)
(* Could have been done with locking too, but it seems modules perform *)
(* somewhat better *)

Module Type OpaqueDefsSig.

Module Names.
(* framing *)
Implicit Types (U : pcm) (T : Type).
Parameter zig : forall U T (s : state U T), U -> state U T.
Parameter zag : forall U T (s : state U T), U -> state U T.
Parameter gaz : forall U T, U -> state U T -> state U T.
Parameter giz : forall U T, U -> state U T -> state U T.
(* transposition *)
Parameter tr : forall U T (s : state U T), state U T.
(* pairing *)
Parameter spair : forall U1 U2 T1 T2,
  state U1 T1 -> state U2 T2 -> state (prodPCM U1 U2) (T1 * T2).
Parameter sfst : forall U1 U2 T1 T2,
  state (prodPCM U1 U2) (T1 * T2) -> state U1 T1.
Parameter ssnd : forall U1 U2 T1 T2,
  state (prodPCM U1 U2) (T1 * T2) -> state U2 T2.
(* sub-pcm taken as a subset *)
(* obsoleted by the definition below, but I keep it for now *)
Parameter ustpair :
  forall (U : pcm) T (P : U -> Prop) (pc : pcm_closed P) (Q : T -> Prop),
  forall (s : state U T), P (self s) -> Q (joint s) -> P (other s) ->
    state (csub_pcm pc) {x : T | Q x}.
Arguments ustpair [U T P pc Q] s ps pj po.
Parameter ustval :
  forall (U : pcm) T (P : U -> Prop) (pc : pcm_closed P) (Q : T -> Prop),
  state (csub_pcm pc) {x : T | Q x} -> state U T.
(* more general treatment, parametric in the sub-pcms *)
(* this allows not only taking a closed subset, but making the join operation *)
(* more partial than it was. Makes ustpair obsolete *)
Parameter spval : forall V D (U : @sub_pcm V D) T, state U T -> state V T.
Parameter spsub : forall V D (U : @sub_pcm V D) T, state V T -> state U T.
End Names.

Axiom zigE : forall U T (s : state U T) p,
  Names.zig s p = State (self s \+ p) (joint s) (other s).
Axiom zagE : forall U T (s : state U T) p,
  Names.zag s p = State (self s) (joint s) (p \+ other s).
Axiom gazE : forall U T p (s : state U T),
  Names.gaz p s = State p (joint s) (self s \+ other s).
Axiom gizE : forall U T p (s : state U T),
  Names.giz p s = State (self s \+ other s) (joint s) p.
Axiom trE : forall U T (s : state U T),
  Names.tr s = (State (other s) (joint s) (self s)).
Axiom spairE : forall U1 U2 T1 T2 (s1 : state U1 T1) (s2 : state U2 T2),
  Names.spair s1 s2 = State (self s1, self s2)
                            (joint s1, joint s2)
                            (other s1, other s2).
Axiom sfstE : forall U1 U2 T1 T2 (s : state (prodPCM U1 U2) (T1 * T2)),
  Names.sfst s = State (self s).1 (joint s).1 (other s).1.
Axiom ssndE : forall U1 U2 T1 T2 (s : state (prodPCM U1 U2) (T1 * T2)),
 Names.ssnd s = State (self s).2 (joint s).2 (other s).2.
Axiom ustpairE : forall U T P pc Q s pfs pfj pfo,
  @Names.ustpair U T P pc Q s pfs pfj pfo =
  State ((exist _ _ pfs) : (csub_pcm pc)) (exist _ _ pfj) (exist _ _ pfo).
Axiom ustvalE : forall U T P pc Q s,
 @Names.ustval U T P pc Q s =
 State (sval (self s)) (sval (joint s)) (sval (other s)).
Axiom spvalE : forall V D (U : @sub_pcm V D) T (s : state U T),
  @Names.spval V D U T s = State (pval (self s)) (joint s) (pval (other s)).
Axiom spsubE : forall V D (U : @sub_pcm V D) T (s : state V T),
  @Names.spsub V D U T s = State (psub U (self s)) (joint s) (psub U (other s)).
End OpaqueDefsSig.

Module OpaqueDefs : OpaqueDefsSig.

Module Names.
Definition zig U T (s : state U T) p := State (self s \+ p) (joint s) (other s).
Definition zag U T (s : state U T) p := State (self s) (joint s) (p \+ other s).
Definition gaz U T p (s : state U T) := State p (joint s) (self s \+ other s).
Definition giz U T p (s : state U T) := State (self s \+ other s) (joint s) p.
Definition tr U T (s : state U T) := State (other s) (joint s) (self s).
Definition spair U1 U2 T1 T2 (s1 : state U1 T1) (s2 : state U2 T2) :=
  State (self s1, self s2) (joint s1, joint s2) (other s1, other s2).
Definition sfst U1 U2 T1 T2 (s : state (prodPCM U1 U2) (T1 * T2)) :=
  State (self s).1 (joint s).1 (other s).1.
Definition ssnd U1 U2 T1 T2 (s : state (prodPCM U1 U2) (T1 * T2)) :=
  State (self s).2 (joint s).2 (other s).2.
Definition ustpair
  (U : pcm) T (P : U -> Prop) (pc : pcm_closed P) (Q : T -> Prop)
  (s : state U T) (pfs : P (self s)) (pfj : Q (joint s)) (pfo : P (other s)) :=
  State ((exist _ _ pfs) : (csub_pcm pc)) (exist _ _ pfj) (exist _ _ pfo).
Definition ustval
  (U : pcm) T (P : U -> Prop) (pc : pcm_closed P) (Q : T -> Prop)
  (s : state (csub_pcm pc) {x : T | Q x}) :=
  State (sval (self s)) (sval (joint s)) (sval (other s)).
Definition spval V D (U : @sub_pcm V D) T (s : state U T) :=
  State (pval (self s)) (joint s) (pval (other s)).
Definition spsub V D (U : @sub_pcm V D) T (s : state V T) :=
  State (psub U (self s)) (joint s) (psub U (other s)).
End Names.

Definition zigE U T s p := erefl (@Names.zig U T s p).
Definition zagE U T s p := erefl (@Names.zag U T s p).
Definition gazE U T p s := erefl (@Names.gaz U T p s).
Definition gizE U T p s := erefl (@Names.giz U T p s).
Definition trE U T s := erefl (@Names.tr U T s).
Definition spairE U1 U2 T1 T2 s1 s2 := erefl (@Names.spair U1 U2 T1 T2 s1 s2).
Definition sfstE U1 U2 T1 T2 s := erefl (@Names.sfst U1 U2 T1 T2 s).
Definition ssndE U1 U2 T1 T2 s := erefl (@Names.ssnd U1 U2 T1 T2 s).
Definition ustpairE U T P pc Q s pfs pfj pfo :=
  erefl (@Names.ustpair U T P pc Q s pfs pfj pfo).
Definition ustvalE U T P pc Q s := erefl (@Names.ustval U T P pc Q s).
Definition spvalE V D U T s := erefl (@Names.spval V D U T s).
Definition spsubE V D U T s := erefl (@Names.spsub V D U T s).
End OpaqueDefs.

Export OpaqueDefs.Names.


(*******************)
(* Zigging/zagging *)
(*******************)

Notation "s \< p" := (zig s p) (at level 34, left associativity).
Notation "s \> p" := (zag s p) (at level 34, left associativity).
Notation "p \\< s" := (gaz p s) (at level 34, left associativity).
Notation "p \\> s" := (giz p s) (at level 34, left associativity).

Lemma zg_unlock U T (s : state U T) p :
        (s \< p = State (self s \+ p) (joint s) (other s)) *
        (s \> p = State (self s) (joint s) (p \+ other s)) *
        (p \\< s = State p (joint s) (self s \+ other s)) *
        (p \\> s = State (self s \+ other s) (joint s) p).
Proof. by rewrite OpaqueDefs.zigE OpaqueDefs.zagE OpaqueDefs.gazE OpaqueDefs.gizE. Qed.

(* When state is not homogenous, we may want to explicitly fold *)
(* the zig/zag notation *)

Lemma zgEX (U : pcm) T p (s o : U) (j : T) :
        ((State (s \+ p) j o = State s j o \< p) *
         (State s j (p \+ o) = State s j o \> p)) *
        ((State (p \+ s) j o = State s j o \< p) *
         (State s j (o \+ p) = State s j o \> p)) *
        (State p j (s \+ o) = p \\< State s j o).
Proof. by rewrite !zg_unlock !(joinC p). Qed.

Lemma zgE U T (s : state U T) p :
        (((self (s \< p) = self s \+ p) * (self (s \> p) = self s)) *
         ((joint (s \< p) = joint s) * (joint (s \> p) = joint s))) *
        (((other (s \< p) = other s) * (other (s \> p) = p \+ other s)) *
         ((self (p \\< s) = p) * (joint (p \\< s) = joint s) *
          (other (p \\< s) = self s \+ other s))) *
         ((self (p \\> s) = self s \+ other s) * (joint (p \\> s) = joint s) *
          (other (p \\> s) = p)).
Proof. by rewrite !zg_unlock. Qed.


(****************************)
(* Transposition of a state *)
(****************************)

Lemma tr_unlock U T (s : state U T) : tr s = State (other s) (joint s) (self s).
Proof. by rewrite OpaqueDefs.trE. Qed.

Lemma trEX (U : pcm) T (s o : U) (j : T) : tr (State s j o) = State o j s.
Proof. by rewrite tr_unlock. Qed.

Lemma trI U T (s : state U T) : tr (tr s) = s.
Proof. by rewrite !tr_unlock -stEX'. Qed.

Lemma tr_idemp U T : @tr U T \o @tr _ _ = id.
Proof. apply/fext/trI. Qed.

Lemma tr_inj U T (s1 s2 : state U T) : tr s1 = tr s2 <-> s1 = s2.
Proof. by split=>[|->//]; apply/inv_inj/trI. Qed.

(* The next lemma useful only in the stated direction; the other *)
(* direction is by simplification. *)
Lemma trE0 U T (s : state U T) :
        (other (tr s) = self s) *
        (joint (tr s) = joint s) *
        (self (tr s) = other s).
Proof. by rewrite !tr_unlock. Qed.

Lemma trE1 U T p (s : state U T) :
        (tr (s \< p) = tr s \> p) * (tr (s \> p) = tr s \< p).
Proof. by rewrite !tr_unlock !zg_unlock !(joinC p). Qed.

Lemma trE2 U T p (s : state U T) :
        (tr (p \\< s) = p \\> tr s) * (tr (p \\> s) = p \\< tr s).
Proof. by rewrite !tr_unlock !zg_unlock /= (joinC (self s)). Qed.

Definition trE := (trE1, trI).


(* zig/zag-ing PCM unit doesn't change the state *)
Lemma zgE0 (U : pcm) T (s : state U T) :
        (s \< Unit = s) * (s \> Unit = s).
Proof. by rewrite !zg_unlock unitL unitR -stEX'. Qed.

Lemma spZ1 U T (s : state U T) p : splits (s \< p) (s \> p) (p \\< s).
Proof. by rewrite /splits !zgE. Qed.

Lemma spZ2 U T (s : state U T) p : splits (s \< p) (p \\< s) (s \> p).
Proof. by apply/spC; apply: spZ1. Qed.

Lemma spZ3 U T (s : state U T) p : splits (p \\> s) (s \> p) (tr (s \< p)).
Proof. by rewrite /splits !trE !zgE !trE0 !(joinC p). Qed.

Lemma spZ4 U T (s : state U T) p : splits (p \\> s) (tr (s \< p)) (s \> p).
Proof. by apply/spC; apply: spZ3. Qed.

Hint Resolve spZ1 spZ2 spZ3 spZ4 : core.

(***************************************)
(* State pairs and fst/snd projections *)
(***************************************)

Notation "s \1" := (sfst s)
  (at level 2, left associativity, format "s \1").
Notation "s \2" := (ssnd s)
  (at level 2, left associativity, format "s \2").
Notation "[ '@' s1 , s2 ]" := (spair s1 s2)
  (at level 0, format "[ '@'  s1 ,  s2 ]").

Definition prod_unlock :=
  (OpaqueDefs.spairE, OpaqueDefs.sfstE, OpaqueDefs.ssndE).

Section ProdState.
Variables (U1 U2 : pcm) (T1 T2 : Type).

Lemma prodEX : (forall (s1 o1 : U1) (s2 o2 : U2) (j1 : T1) (j2 : T2),
                  [@ State s1 j1 o1, State s2 j2 o2] =
                     State (s1, s2) (j1, j2) (o1, o2)) *
               (forall (s o : U1 * U2) (j : T1 * T2),
                 ((State s j o)\1 = State s.1 j.1 o.1) *
                 ((State s j o)\2 = State s.2 j.2 o.2)).
Proof. by split=>*; rewrite ?prod_unlock. Qed.

(* Some simplifying equations *)

Lemma sbeta (s1 : state U1 T1) (s2 : state U2 T2) :
        ([@ s1, s2]\1 = s1) * ([@ s1, s2]\2 = s2).
Proof. by rewrite !prod_unlock -!stEX'. Qed.

Lemma seta (s : state (prodPCM U1 U2) (T1 * T2)) : s = [@ s\1, s\2].
Proof. by rewrite [s]stEX' !prodEX -!prod_eta. Qed.

(* The next lemma should be used only in the stated direction. We don't *)
(* want .1/.2 projections interjected over self/other/joint, because this *)
(* hampers zig/zag notation, and forces that state is printed in expanded *)
(* form. The lemma helps remove the projections if it appeared by *)
(* accident. *)
Lemma prodEQ0 (s : state (prodPCM U1 U2) (T1 * T2)) :
        (((self s).1 = self s\1) * ((self s).2 = self s\2)) *
        (((joint s).1 = joint s\1) * ((joint s).2 = joint s\2)) *
        (((other s).1 = other s\1) * ((other s).2 = other s\2)).
Proof. by rewrite !prod_unlock. Qed.

Lemma prodEQ1 (s1 : state U1 T1) (s2 : state U2 T2) :
        (self [@ s1, s2] = (self s1, self s2)) *
        (joint [@ s1, s2] = (joint s1, joint s2)) *
        (other [@ s1, s2] = (other s1, other s2)).
Proof. by rewrite !prod_unlock. Qed.

Lemma prodEQ2 p (s : state (prodPCM U1 U2) (T1 * T2)) :
       (((s \< p)\1 = s\1 \< p.1) * ((s \> p)\1 = s\1 \> p.1)) *
       (((s \< p)\2 = s\2 \< p.2) * ((s \> p)\2 = s\2 \> p.2)) *
       (((p \\< s)\1 = p.1 \\< s\1) * ((p \\< s)\2 = p.2 \\< s\2)).
Proof. by rewrite !zg_unlock !prod_unlock. Qed.

Lemma prodEQ3 (s : state (prodPCM U1 U2) (T1 * T2)) :
       ((tr s)\1 = tr s\1) *
       ((tr s)\2 = tr s\2).
Proof. by rewrite !tr_unlock !prod_unlock. Qed.

Lemma tr_beta (s1 : state U1 T1) (s2 : state U2 T2) :
         tr [@ s1, s2] = [@ tr s1, tr s2].
Proof. by rewrite !tr_unlock !prod_unlock. Qed.

Definition prodE := ((sbeta, tr_beta), (prodEQ1, prodEQ2), (prodEQ3, prodEQ0)).

(* A separate rule. Putting this one in prodE may result in unwanted
behavior, because applying prodX makes some of prodE's inapplicable *)

Lemma prodX (s : state (prodPCM U1 U2) (T1 * T2)) :
        ((self s = (self s\1, self s\2)) * (joint s = (joint s\1, joint s\2))) *
        ((other s = (other s\1, other s\2)) * (tr s = [@ tr s\1, tr s\2])).
Proof. by rewrite !tr_unlock !prod_unlock -!prod_eta. Qed.

Definition prodXE := (prodE, prodX).

End ProdState.


(***********************)
(* States and sub-PCMs *)
(***********************)

Section SubUState.
Variables (U : pcm) (T : Type) (P : U -> Prop).
Variables (pc : pcm_closed P) (Q : T -> Prop).

Definition substate := state (csub_pcm pc) {x : T | Q x}.

Notation ustpair := (@ustpair U T P pc Q).
Notation ustval := (@ustval U T P pc Q).

Lemma ust_unlock :
        (forall (s : state U T) ps (pj : Q (joint s)) po,
           ustpair s ps pj po =
           State ((exist _ _ ps):(csub_pcm pc))
                 (exist _ _ pj) (exist _ _ po)) *
        (forall s : state (csub_pcm pc) {x : T | Q x}, ustval s =
           State (sval (self s)) (sval (joint s)) (sval (other s))).
Proof. by do !split=>*; rewrite ?OpaqueDefs.ustpairE? OpaqueDefs.ustvalE. Qed.

Lemma ustEX :
  (forall (s o : U) (j : T) ps (pj : Q j) po,
     ustpair (State s j o) ps pj po =
     State ((exist _ _ ps):(csub_pcm pc)) (exist _ _ pj) (exist _ _ po)) *
  (forall (s o : csub_pcm pc) (j : {x : T | Q x}),
     ustval (State s j o) = State (sval s) (sval j) (sval o)).
Proof. by do !split=>*; rewrite ust_unlock. Qed.

Lemma ust_beta s ps pj po : ustval (ustpair s ps pj po) = s.
Proof. by rewrite !ust_unlock /= -stEX'. Qed.

Lemma ust_pf (s : substate) :
        [/\ P (self (ustval s)), Q (joint (ustval s)) & P (other (ustval s))].
Proof. by case: s; case=>s ps [j pj][o po]; rewrite ust_unlock. Qed.

Lemma ustps (s : substate) : P (self (ustval s)).
Proof. by case: (ust_pf s). Qed.

Lemma ustpj (s : substate) : Q (joint (ustval s)).
Proof. by case: (ust_pf s). Qed.

Lemma ustpo (s : substate) : P (other (ustval s)).
Proof. by case: (ust_pf s). Qed.

Lemma ust_eta s : s = ustpair (ustval s) (ustps s) (ustpj s) (ustpo s).
Proof.
case: s=>s j o; rewrite ust_unlock; apply/stE'=>/=; split.
- case: s=>s pf1; move: (ustps _); rewrite ustEX=>pf2.
  by congr exist; apply: pf_irr.
case: j=>j pf1; move: (ustpj _); rewrite ustEX=>pf2.
- by congr exist; apply: pf_irr.
case: o=>o pf1; move: (ustpo _); rewrite ustEX=>pf2.
by congr exist; apply: pf_irr.
Qed.

Lemma ust_inj (s1 s2 : substate) : ustval s1 = ustval s2 -> s1 = s2.
Proof.
rewrite !ust_unlock; case: s1=>[[s1 ps1]][j1 pj1] [o1 po1].
case: s2=>[[s2 ps2]][j2 pj2][o2 po2] /= [???]; subst s2 j2 o2.
by apply/stE'; split=>/=; congr exist; apply: pf_irr.
Qed.

(* The next lemma should be used only in the stated direction. *)
Lemma ustEQ0 (s : substate) :
        (sval (self s) = self (ustval s)) *
        (sval (joint s) = joint (ustval s)) *
        (sval (other s) = other (ustval s)).
Proof. by rewrite !ust_unlock. Qed.

Lemma ustEQ1 s ps pj po :
        (self (ustpair s ps pj po) = exist P (self s) ps) *
        (joint (ustpair s ps pj po) = exist Q (joint s) pj) *
        (other (ustpair s ps pj po) = exist P (other s) po).
Proof. by rewrite !ust_unlock. Qed.

Lemma ustEQ2 p s : (ustval (s \< p) = ustval s \< sval p) *
                   (ustval (s \> p) = ustval s \> sval p) *
                   (ustval (p \\< s) = (sval p) \\< (ustval s)).
Proof.
by rewrite !zg_unlock !ust_unlock; case: p=>p pf; case: s; case=>??? [].
Qed.

Lemma ustEQ3 s : ustval (tr s) = tr (ustval s).
Proof. by rewrite !tr_unlock !ust_unlock; case: s; case=>s ps j []. Qed.

Lemma usts (s : state U T) (ps : P (self s)) : P (other (tr s)).
Proof. by rewrite trE0. Qed.

Lemma ustj (s : state U T) (pj : Q (joint s)) : Q (joint (tr s)).
Proof. by rewrite trE0. Qed.

Lemma usto (s : state U T) (po : P (other s)) : P (self (tr s)).
Proof. by rewrite trE0. Qed.

Lemma tr_ustbeta s ps pj po :
        tr (ustpair s ps pj po) = ustpair (tr s) (usto po) (ustj pj) (usts ps).
Proof.
rewrite !tr_unlock !ust_unlock; apply/stE'; split=>/=.
- by move: (usto po); rewrite trE0=>pf; rewrite (pf_irr pf po).
- by move: (ustj pj); rewrite trE0=>pf; rewrite (pf_irr pf pj).
by move: (usts ps); rewrite trE0=>pf; rewrite (pf_irr pf ps).
Qed.

Definition ustE := ((ust_beta, tr_ustbeta), (ustEQ1, ustEQ2), (ustEQ3, ustEQ0)).

Lemma ustX s : (self s = exist P (self (ustval s)) (ustps s)) *
               (joint s = exist Q (joint (ustval s)) (ustpj s)) *
               (other s = exist P (other (ustval s)) (ustpo s)).
Proof. by rewrite {1 4 7}[s]ust_eta !ustE. Qed.

Definition ustXE := (ustE, ustX).

End SubUState.

Arguments ustval [U T P] pc Q s.
Arguments ust_inj {U T P pc Q} _ _.

(*************************)
(* States and sub-PCMs 2 *)
(*************************)


Section SubState.
Variables (V : pcm) (D : rel V) (U : sub_pcm D) (T : Type).
Notation spval := (@spval V D U T).
Notation spsub := (@spsub V D U T).

Lemma spunlock :
        (forall s, spval s =
                   State (pval (self s)) (joint s) (pval (other s))) *
        (forall s, spsub s =
                   State (psub U (self s)) (joint s) (psub U (other s))).
Proof. by split=>s; rewrite ?OpaqueDefs.spvalE ?OpaqueDefs.spsubE. Qed.

Lemma spEX : (forall s o j, spval (State s j o) =
                            State (pval s) j (pval o)) *
             (forall s o j, spsub (State s j o) =
                            State (psub U s) j (psub U o)).
 Proof. by split=>*; rewrite spunlock. Qed.

Lemma sp_inj (s1 s2 : state U T) : spval s1 = spval s2 -> s1 = s2.
Proof.
rewrite !spunlock; case: s1=>s1 j1 o1; case: s2=>s2 j2 o2 /=.
by case=>/pval_inj Hs Hj /pval_inj Ho; apply/stE'.
Qed.

(* The next lemma should be used only in the stated direction. *)
(* added note: except if I leave spval unlock, the lemma is unnecessary *)
Lemma spEQ0 (s : state U T) :
        (pval (self s) = self (spval s)) *
        (joint s = joint (spval s)) *
        (pval (other s) = other (spval s)).
Proof. by rewrite spunlock. Qed.

Lemma spEQ2 p s :
        (valid (self s \+ p) ->
           spval (s \< p) = spval s \< pval p) *
        (valid (p \+ other s) ->
           spval (s \> p) = spval s \> pval p) *
        (valid (self s \+ other s) ->
           spval (p \\< s) = pval p \\< spval s).
Proof. by rewrite !zg_unlock; do! split; move=> W; rewrite !spunlock pfjoin. Qed.

Lemma spEQ3 s : spval (tr s) = tr (spval s).
Proof. by rewrite !tr_unlock !spunlock. Qed.

Lemma spEQ4 s : valid (self s \+ other s) -> spval (outs s)  = outs (spval s).
Proof. by move=>W; rewrite /outs !spunlock pfjoin // pfunit. Qed.

Definition spE := ((spEQ2, spEQ3), (spEQ0, spEX)).

End SubState.

Arguments spval [V D U T] s.
Arguments sp_inj [V D U T] _ _.


Section SubState2.

Lemma sp_betax (V : pcm) (D : sep_rel V) (U : sub_pcm D) T (s : state V T) :
        valid (self s \+ other s) ->
        D (self s) (other s) -> spval (spsub U s) = s.
Proof.
by move=>W H; rewrite !spunlock !pval_psub
-?stEX' ?(validL W) ?(validR W) ?(sep0E W H).
Qed.

Lemma sp_eta (V : pcm) (D : rel V) (U : sub_pcm D)  T (s : state U T) :
        s = spsub U (spval s).
Proof. by apply/stE'; rewrite !spunlock !psub_pval. Qed.

Lemma sp_eta' (V : pcm) (D : rel V) (U : sub_pcm D)  T (s : state U T) :
        spsub U (spval s) = s.
Proof. by rewrite -sp_eta. Qed.

End SubState2.

(******************************************)
(* Dealing with notation                  *)
(* of projecting out of a PCM and a state *)
(******************************************)

(* Using morphisms as labels *)

(* we hide the definition of the two applications *)
(* to be able to apply custom rewrite laws later on *)
Module Type Private_NotationSig.
Section Private_Notation.
Variables (U : pcm) (T1 T2 : Type).
Parameter j_app : (T1 -> T2) -> state U T1 -> T2.
Parameter state_to_subj : state U T1 -> U * U.
Axiom j_appE : forall f s, j_app f s = f (joint s).
Axiom state_to_subjE : forall s, state_to_subj s = (self s, other s).
End Private_Notation.
End Private_NotationSig.

Module Private_NotationST : Private_NotationSig.
Section Private_Notation.
Variables (U : pcm) (T1 T2 : Type).
Definition j_app (f : T1 -> T2) (s : state U T1) := f (joint s).
Definition state_to_subj (s : state U T1) := (self s, other s).
Definition j_appE f s := erefl (j_app f s).
Definition state_to_subjE s := erefl (state_to_subj s).
End Private_Notation.
End Private_NotationST.

Notation "f '.j' x" := (Private_NotationST.j_app f x) (at level 10,
  format "f '.j'  x").

Coercion Private_NotationST.state_to_subj : state >-> prod.

Definition unlock_priv := (Private_NotationST.j_appE,
                           Private_NotationST.state_to_subjE,
                           Private_Notation.so_appE,
                           Private_Notation.os_appE, prodE).

(* allvalid lemmas for state *)

Lemma allvalidE U T (s : state U T) : allvalid s = valid (self s \+ other s).
Proof. by rewrite /allvalid -!lock Private_NotationST.state_to_subjE. Qed.

Lemma zgVE U T (s : state U T) p : allvalid (s \< p) = allvalid (s \> p).
Proof. by rewrite /allvalid !zg_unlock !Private_NotationST.state_to_subjE /= joinA. Qed.

Lemma gzVE U T (s : state U T) p : allvalid (p \\< s) = allvalid (s \> p).
Proof. by rewrite /allvalid !zg_unlock !Private_NotationST.state_to_subjE /= joinCA. Qed.

Lemma trVE U T (s : state U T) : allvalid (tr s) = allvalid s.
Proof. by rewrite /allvalid tr_unlock !Private_NotationST.state_to_subjE joinC. Qed.

Lemma prodVE U1 U2 T1 T2 (s : state (prodPCM U1 U2) (T1 * T2)) :
        allvalid s = allvalid s\1 && allvalid s\2.
Proof. by rewrite /allvalid -!lock !prod_unlock !Private_NotationST.state_to_subjE. Qed.

Lemma ustVE (U : pcm) (T : Type) (P : U -> Prop) (pc : pcm_closed P) (Q : T -> Prop) :
        (forall s ps pj po, allvalid (@ustpair U T P pc Q s ps pj po) = allvalid s) *
        (forall s, allvalid (@ustval U T P pc Q s) = allvalid s).
Proof.
split=>[????|[[s ps][j pj][o po]]];
by rewrite /allvalid ust_unlock !Private_NotationST.state_to_subjE.
Qed.

Lemma spVE (V : pcm) (D : rel V) (U : sub_pcm D) (T : Type) (s : state U T) :
        allvalid s = allvalid (spval s) &&
                     D (idfun.s (spval s)) (idfun.o (spval s)).
Proof.
by rewrite spunlock /allvalid -!lock valid_sep !Private_NotationST.state_to_subjE !lE.
Qed.

Lemma spVI (V : pcm) (D : rel V) (U : sub_pcm D) (T : Type) (s : state U T) :
        allvalid s -> allvalid (spval s).
Proof. by rewrite spVE; case/andP. Qed.

Lemma spZE (V : pcm) (D : rel V) (U : sub_pcm D) T (s : state U T) p :
        (allvalid (s \< p) -> spval (s \< p) = spval s \< pval p) *
        (allvalid (s \< p) -> spval (s \> p) = spval s \> pval p).
Proof.
by rewrite allvalidE !zgE; split=>A; rewrite !spEQ2 ?(validL A) ?(validAR A).
Qed.

Lemma sp_beta (V : pcm) (D : sep_rel V) (U : sub_pcm D) T (s : state V T) :
       allvalid s -> D (idfun.s s) (idfun.o s) ->
       spval (spsub U s) = s.
Proof.
rewrite allvalidE !lSX Private_NotationST.state_to_subjE /=.
by apply: sp_betax.
Qed.

Definition spvalE := (sp_beta, spZE, spE).


(* application laws *)

Section LabelSOLaws.
Variables (U : pcm) (V : Type) (f : U -> V).

(* because we locked the coercion, we need special lemmas for  state triples *)
(* NOTE: there is ambiguity between SX/SXst, because they both *)
(* SX applies whenever SXst does (though not vice-versa) *)
(* because a state is coercible into a pair *)
(* Thus, we always first try to rewrite by SXst, then by SX *)
Lemma lSXst T (s : state U T) :
        (f.s s = f (self s)) * (f.o s = f (other s)).
Proof. by rewrite !lSX Private_NotationST.state_to_subjE. Qed.

(* ... or to an explicit state *)
(* NOTE: unlike SX/SXst, there is no ambiguity *)
(* between Sso/Sst (resp. Tso/Tst) *)
Lemma lSst T (s o : U) (j : T) :
        (f.s (State s j o) = Private_Notation.app_lock f s) *
        (f.o (State s j o) = Private_Notation.app_lock f o).
Proof. by rewrite !lSXst !Private_Notation.app_lockE. Qed.

Lemma lStr T (s : state U T) : (f.s (tr s) = f.o s) * (f.o (tr s) = f.s s).
Proof. by rewrite !lSXst !tr_unlock. Qed.

End LabelSOLaws.

Lemma lStrf (U V W : pcm) (T : Type) (f : U -> V) (g : V -> W) (s : state U T) :
        g (f.s (tr s)) \+ g (f.o (tr s)) = g (f.s s) \+ g (f.o s).
Proof. by rewrite !lStr joinC. Qed.


Section LabelTLaws.
Variables (U V : pcm) (f : U -> V).

Lemma lTst T (s o : U) (j : T) :
        f.t (State s j o) = Private_Notation.app_lock f s \+ Private_Notation.app_lock f o.
Proof. by rewrite !lSst. Qed.

Lemma lTtr T (s : state U T) : f.t (tr s) = f.t s.
Proof. by rewrite !lStr joinC. Qed.

End LabelTLaws.

Section LabelZgLaws.
Variables (U V W : pcm) (T : Type).

(* when zigzagging, we expand selfs to the right, others to the left *)
(* this way the relative order of expressions self s, p, other s remains *)
(* the same, modulo parentheses. We can rewrite by associativity to *)
(* make the expressions match, but will not need commutativity as well *)
Lemma lSzg1 (f : U -> V) (s : state U T) p :
        (f.s (s \> p) = f.s s) * (f.o (s \< p) = f.o s).
Proof. by rewrite !lSXst !zgE. Qed.

Lemma lSzg1' (f : U -> V) (s : state U T) p :
        (f.s (p \\< s) = f p) * (f.o (p \\< s) = f (idfun.t s)).
Proof. by rewrite !lSXst !zgE. Qed.

Lemma lSo' (f : U -> V) (s : state U T) :
        (f.s (outs s) = f Unit) * (f.o (outs s) = f (idfun.t s)).
Proof. by rewrite !lSXst. Qed.

(* applying morphism to a zig-zagged state really only makes sense *)
(* if the morphisms definition of separation is validity *)
(* We only do this when defining a resource, and then morphisms *)
(* are just simple projections, i.e., their domain is validity *)
Lemma lSzg2_partial (D : sep_rel U) (f : {morphism D >-> V}) (s : state U T) p :
        (allvalid (s \< p) -> D (idfun.s s) p ->
           f.s (s \< p) = f.s s \+ Private_Notation.app_lock f p) *
        (allvalid (s \< p) -> D (idfun.o s) p ->
           f.o (s \> p) = Private_Notation.app_lock f p \+ f.o s).
Proof.
by do !split; rewrite !lSXst /= /allvalid -!lock !Private_NotationST.state_to_subjE !zgE
                   Private_Notation.app_lockE=>/= X S;
  rewrite pfjoin // 1?sepC // (validLE3 X).
Qed.

Lemma lSzg2_partialf (Df : sep_rel U) (Dg : sep_rel V)
        (f : {morphism Df >-> V}) (g : {morphism Dg >-> W}) (s : state U T) p :
        (allvalid (s \< p) -> Df (idfun.s s) p -> Dg (f.s s) (f p) ->
           g (f.s (s \< p)) = g (f.s s) \+ g (Private_Notation.app_lock f p)) *
        (allvalid (s \< p) -> Df (idfun.o s) p -> Dg (f.o s) (f p) ->
           g (f.o (s \> p)) = g (Private_Notation.app_lock f p) \+ g (f.o s)).
Proof.
rewrite !lSXst /= /allvalid -!lock !Private_NotationST.state_to_subjE /=.
rewrite !zgE Private_Notation.app_lockE; split=>A H1 H2.
- by rewrite pfjoin // ?(validLE3 A) // pfjoin // pfV2 // (validL A).
by rewrite pfjoin 1?sepC ?(validLE3 A) // pfjoin 1?sepC // pfV2 1?sepC ?(validLE3 A).
Qed.

Definition lSzg := (lSzg1, (* lSzg2, *) lSzg2_partialf, lSzg2_partial).

Lemma lTzg1_partial (D : sep_rel U) (f : {morphism D >-> V}) (s : state U T) p :
        (allvalid (s \< p) -> D (idfun.s s) p ->
           f.t (s \< p) = Private_Notation.app_lock f p \+ f.t s) *
        (allvalid (s \< p) -> D (idfun.o s) p ->
           f.t (s \> p) = Private_Notation.app_lock f p \+ f.t s).
Proof. by split=>X S; rewrite !lSzg // joinCA // joinA. Qed.

Lemma lTzg1_partialf (Df : sep_rel U) (Dg : sep_rel V)
        (f : {morphism Df >-> V}) (g : {morphism Dg >-> W}) (s : state U T) p :
        (allvalid (s \< p) -> Df (idfun.s s) (idfun.o s) -> Df (idfun.t s) p ->
                              Dg (f.t s) (f p) ->
           g (f.t (s \< p)) = g (Private_Notation.app_lock f p) \+ g (f.t s)) *
        (allvalid (s \< p) -> Df (idfun.s s) (idfun.o s) -> Df (idfun.t s) p ->
                              Dg (f.t s) (f p) ->
           g (f.t (s \> p)) = g (Private_Notation.app_lock f p) \+ g (f.t s)).
Proof.
rewrite allvalidE !lSXst !zgE /=; split=>A H1 H2 H3;
have X: valid (p \+ (self s \+ other s)) by [rewrite joinCA joinA];
have Y: valid (f p \+ f (self s \+ other s)) by [rewrite pfV2 // sepC];
have Z: valid (f p \+ (f (self s) \+ f (other s)))
  by [rewrite -pfjoin // (validR X)].
- rewrite Private_Notation.app_lockE (pfjoin f).
  - by rewrite -joinA joinCA // pfjoin // sepC.
  - by rewrite (validL A).
  by rewrite (sepAxx _ H1 H2) // joinAC.
rewrite Private_Notation.app_lockE (pfjoin f).
- by rewrite joinCA // pfjoin // sepC.
- by rewrite (validAR A).
by rewrite (sepAxx _ H1 H2) // joinAC.
Qed.

Lemma lTzg3_partialf (Df : sep_rel U) (Dg : sep_rel V)
        (f : {morphism Df >-> V}) (g : {morphism Dg >-> W}) (s : state U T) p :
        (allvalid (s \< p) -> Df (idfun.s s) p -> Dg (f.s s) (f p) ->
           g (f.s (s \< p)) \+ g (f.o (s \< p)) =
           g (f.s s) \+ g (Private_Notation.app_lock f p) \+ g (f.o s)) *
        (allvalid (s \< p) -> Df (idfun.o s) p -> Dg (f.o s) (f p) ->
           g (f.s (s \> p)) \+ g (f.o (s \> p)) =
           g (f.s s) \+ g (Private_Notation.app_lock f p) \+ g (f.o s)).
Proof. by split=>A H1 H2; rewrite !lSzg2_partialf // lSzg1 // joinA. Qed.


Lemma lTo' (D : sep_rel U) (f : {morphism D >-> V}) (s : state U T) :
        allvalid s -> D (idfun.s s) (idfun.o s) -> f.t (outs s) = f.t s.
Proof.
move=>X S; rewrite !lSo' pfunit unitL pfjoin ?lA // !lSXst.
by rewrite allvalidE in X.
Qed.

Definition lTzg := ((* lTzg1, *) lTzg3_partialf, lTzg1_partialf, lTzg1_partial).

(* sometimes a simple switch is what I need *)
Lemma lTZ (D : sep_rel U) (f : {morphism D >-> V}) (s : state U T) p :
        allvalid (s \< p) -> D (id.s s) p -> D (id.o s) p ->
        f.t (s \> p) = f.t (s \< p).
Proof. by move=>X S1 S2; rewrite !lSzg //= joinA. Qed.

End LabelZgLaws.

(* And the lemmas for joint *)

Section LabelJLaws.
Variables (U : pcm) (T1 T2 : Type) (f : T1 -> T2).

(* this one we don't automatically rewrite with *)
(* it basically unfolds the definitions of .s and .o *)
(* so should be used only in libraries, not in clients *)
Lemma lJX (s : state U T1) : f.j s = f (joint s).
Proof. by rewrite Private_NotationST.j_appE. Qed.

(* can't use the locking trick here, because putting the equation in *)
(* for unlocking would be too eager *)
(* So we have a different hack *)
Lemma lJst (s o : U) (j : T1) : f.j (State s j o) = f j.
Proof. by rewrite !lJX. Qed.

Lemma lJstA (s o : U) (x : state U T1) :
        f.j (State s (idfun.j x) o) = f.j x.
Proof. by rewrite !lJX Private_NotationST.j_appE. Qed.

Lemma lJtr (s : state U T1) : f.j (tr s) = f.j s.
Proof. by rewrite !lJX trE0. Qed.

Lemma lJo' (s : state U T1) : f.j (outs s) = f.j s.
Proof. by rewrite !lJX. Qed.

End LabelJLaws.

Lemma lJprod U T T1 T2 (f1 : T -> T1) (f2 : T -> T2) (s : state U T) :
        (f1 \** f2).j s = (f1.j s, f2.j s).
Proof. by rewrite !lJX. Qed.

Lemma lJzg U T1 T2 (f : T1 -> T2) (s : state U T1) p :
        (f.j (s \< p) = f.j s) *
        (f.j (s \> p) = f.j s).
Proof. by rewrite !lJX !zg_unlock. Qed.

Lemma lJA U T T1 (f : T -> T1) (s : state U T) :
        f (idfun.j s) = f.j s.
Proof. by rewrite !lJX. Qed.

Lemma lCJX U T V W (g : V -> W) (f : T -> V) (x : state U T) :
        g (f.j x) = (g \o f).j x.
Proof. by rewrite !lJX. Qed.

(* expand zig, zag in a special case when they are compared to explicit triples *)
(* this is the only case when we need to do that *)
Definition zigX U T (x : state U T) (s o : U) (j : T) p :
             ((x \< p = State s j o) =
                ((idfun.s x \+ p, idfun.j x, idfun.o x) = (s, j, o))) *
             ((x \> p = State s j o) =
                ((idfun.s x, idfun.j x, p \+ idfun.o x) = (s, j, o))).
Proof.
by rewrite !(zg_unlock,lSXst,lJX); split; apply/pext; split; case=><-<-<-.
Qed.

(* actually, we also have one more special case *)
Lemma zigY U T (x y : state U T) p :
       ((x \< p = y \< p) =
          ((idfun.s x \+ p, idfun.j x, idfun.o x) =
           (idfun.s y \+ p, idfun.j y, idfun.o y))) *
       ((x \> p = y \> p) =
          ((idfun.s x, idfun.j x, p \+ idfun.o x) =
           (idfun.s y, idfun.j y, p \+ idfun.o y))).
Proof.
rewrite !(zg_unlock,lSXst,lJX); split;
by apply/pext=> /=; split; case=>->->->.
Qed.

(* Do a similar thing for tr *)
Lemma trX U T (x : state U T) (s o : U) (j : T) :
        (tr x = State s j o) = (x = State o j s).
Proof. by rewrite tr_unlock; apply/pext; split=>[[<-<-<-]|->]; apply/stEX'. Qed.

(* a somewhat more general *)
Lemma trY U T (x y : state U T) :
        (tr x = y) = (x = tr y).
Proof. by case: y=>a b c; rewrite trX trEX. Qed.

(* Transitional properties *)
(* Listed here to be used until we refactor the development *)
(* too use \all.s/o instead of self/other *)
Section Transitional.
Variables (U : pcm) (T : Type) (s s' : state U T).

Lemma lbSA : self s = idfun.s s.
Proof. by rewrite lSXst. Qed.

Lemma lbOA : other s = idfun.o s.
Proof. by rewrite lSXst. Qed.

Lemma lbJA : joint s = idfun.j s.
Proof. by rewrite lJX. Qed.

Lemma lbV : valid (self s \+ other s) = valid (idfun.t s).
Proof. by rewrite !lSXst. Qed.

Lemma lbO : other s = other s' <-> idfun.o s = idfun.o s'.
Proof. by rewrite !lSXst. Qed.

End Transitional.

(* add lbSA/lbOA as hints; to be removed once transitioned *)
Hint Extern 0 (self _ = idfun.s _) => apply: lbSA : core.
Hint Extern 0 (other _ = idfun.o _) => apply: lbOA : core.

Definition lT := (lbV, lbSA, lbOA, lbJA).

(* expansion in morphism notation *)
Lemma stEM (U : pcm) (T : Type) (s : state U T) :
        s = State (idfun.s s) (idfun.j s) (idfun.o s).
Proof. by rewrite -lbSA -lbOA -lbJA -stEX'. Qed.

Lemma stM (U : pcm) (T : Type) (s s' : state U T) :
       s = s' <-> [/\ idfun.s s = idfun.s s',
                      idfun.j s = idfun.j s' &
                      idfun.o s = idfun.o s'].
Proof.
split=>[->//|[H1 H2 H3]]; apply/stE'.
by rewrite !lbSA !lbOA !lbJA H1 H2 H3.
Qed.

(* expansion laws; to be used with caution; never in program proofs *)
Definition lbX := (lSXst, lSX, lJX).

(* rewriting of \f.s and \f.o *)
Definition lSE' := ((lSst, lStrf, lStr), (lSzg, lSE')).

(* rewriting of \f.t *)
Definition lTE' := ((lTzg, lTst), (lTtr, lTE')).

(* rewriting \f.j *)
Definition lJE' := ((lJstA, lJst), (lJtr, lJprod), lJzg).

(* lTE before lSE, because TE sometomes overrides the s/o order *)

Definition lE :=
  ((lTE', lSE', lJE'), ((zigX, zigY, trX), (tr_inj, lTApply)), (lSApply, app_lockE)).

(* separate pfjoin because it can be too agressive *)
Definition lE1 := (lE, pfjoin).

(* a bit slower, so I don't pack in, consider removing zigX, trX too *)
Definition lE2 := (lE1, tr_inj).

(* interaction with all *)
(* lJA rewrites too much *)
Definition lA := (lTA, lSA, (* lJA, *) idfun0E).

Definition lO := (lTo', lSo', lJo').

(* Most times we need trS in implication form *)
Lemma trS U T (x : state U T) (s o : U) (j : T) :
        tr x = State s j o -> x = State o j s.
Proof. by rewrite trX. Qed.

Prenex Implicits trS.

Section LabelProd.
Variables (U1 U2 V : pcm) (T1 T2 T : Type).

(* the Xst lemmas apply to any state indiscriminately *)
(* so be careful in automatic rewriting *)
(* still, they are better than if I had lemmas that apply to pairs *)
Lemma lprodSXst (s : state (prodPCM U1 U2) (T1 * T2)) :
        ((forall f : _ -> V, (f \o fst).s s = f.s s\1) *
         (forall f : _ -> V, (f \o fst).o s = f.o s\1)) *
        ((forall f : _ -> V, (f \o snd).s s = f.s s\2) *
         (forall f : _ -> V, (f \o snd).o s = f.o s\2)).
Proof. by do !split; move=>f; rewrite !lSXst /= !prodEQ0. Qed.

Lemma lprodTXst (s : state (prodPCM U1 U2) (T1 * T2)) :
        (forall f : _ -> V, (f \o fst).t s = f.t s\1) *
        (forall f : _ -> V, (f \o snd).t s = f.t s\2).
Proof. by do !split; move=>f; rewrite !lprodSXst. Qed.

Lemma lprodJXst (s : state (prodPCM U1 U2) (T1 * T2)) :
        (forall f : _ -> T, (f \o fst).j s = f.j s\1) *
        (forall f : _ -> T, (f \o snd).j s = f.j s\2).
Proof. by do !split; move=>f; rewrite !lJX /= !prodEQ0. Qed.

(* special cases when f = id *)
Lemma lprodSXst1 (s : state (prodPCM U1 U2) (T1 * T2)) :
        ((fst.s s = idfun.s s\1) * (fst.o s = idfun.o s\1)) *
        ((snd.s s = idfun.s s\2) * (snd.o s = idfun.o s\2)).
Proof. by do !split; rewrite !lSXst /= !prodEQ0. Qed.

Lemma lprodTXst1 (s : state (prodPCM U1 U2) (T1 * T2)) :
        (fst.t s = idfun.t s\1) * (snd.t s = idfun.t s\2).
Proof. by do !split; rewrite !lSXst /= !prodEQ0. Qed.

Lemma lprodJXst1 (s : state (prodPCM U1 U2) (T1 * T2)) :
        (fst.j s = idfun.j s\1) * (snd.j s = idfun.j s\2).
Proof. by do !split; rewrite !lJX /= !prodEQ0. Qed.

(* when the lemmas are applied to an explicit pair *)
(* it's safe to reduce *)
Lemma lprodSst (s1 : state U1 T1) (s2 : state U2 T2) :
        ((forall f : _ -> V, (f \o fst).s [@ s1, s2] = f.s s1) *
         (forall f : _ -> V, (f \o fst).o [@ s1, s2] = f.o s1)) *
        ((forall f : _ -> V, (f \o snd).s [@ s1, s2] = f.s s2) *
         (forall f : _ -> V, (f \o snd).o [@ s1, s2] = f.o s2)).
Proof. by do !split; move=>f; rewrite !lprodSXst !sbeta. Qed.

Lemma lprodTst (s1 : state U1 T1) (s2 : state U2 T2) :
        (forall f : _ -> V, (f \o fst).t [@ s1, s2] = f.t s1) *
        (forall f : _ -> V, (f \o snd).t [@ s1, s2] = f.t s2).
Proof. by do !split; move=>f; rewrite !lprodSst. Qed.

Lemma lprodJst (s1 : state U1 T1) (s2 : state U2 T2) :
        (forall f : _ -> T, (f \o fst).j [@ s1, s2] = f.j s1) *
        (forall f : _ -> T, (f \o snd).j [@ s1, s2] = f.j s2).
Proof. by do !split; move=>f; rewrite lprodJXst sbeta. Qed.

Lemma lprodSst1 (s1 : state U1 T1) (s2 : state U2 T2) :
        ((fst.s [@ s1, s2] = idfun.s s1) *
         (fst.o [@ s1, s2] = idfun.o s1)) *
        ((snd.s [@ s1, s2] = idfun.s s2) *
         (snd.o [@ s1, s2] = idfun.o s2)).
Proof. by do !split; rewrite !lprodSXst1 !sbeta. Qed.

Lemma lprodTst1 (s1 : state U1 T1) (s2 : state U2 T2) :
        (fst.t [@ s1, s2] = idfun.t s1) *
        (snd.t [@ s1, s2] = idfun.t s2).
Proof. by do !split; rewrite !lprodSst1. Qed.

Lemma lprodJst1 (s1 : state U1 T1) (s2 : state U2 T2) :
        (fst.j [@ s1, s2] = idfun.j s1) *
        (snd.j [@ s1, s2] = idfun.j s2).
Proof. by do !split; rewrite lprodJXst1 sbeta. Qed.

Lemma lprodSpair V1 V2 (s : state V T)
                 (f1 : _ -> V1) (f2 : _ -> V2) :
        ([** f1 & f2].s s = (f1.s s, f2.s s)) *
        ([** f1 & f2].o s = (f1.o s, f2.o s)).
Proof. by rewrite !lSXst. Qed.

Lemma lprodTpair (V1 V2 : pcm) (s : state V T)
                 (f1 : _ -> V1) (f2 : _ -> V2) :
        [** f1 & f2].t s = (f1.t s, f2.t s).
Proof. by rewrite !lprodSpair. Qed.

Lemma lprodJpair (s : state V T)
                 (f1 : _ -> T1) (f2 : _ -> T2) :
        [** f1 & f2].j s = (f1.j s, f2.j s).
Proof. by rewrite !lJX. Qed.

End LabelProd.


(* Reindexing lemmas for sep *)
Section SepProd.
Variables (U1 U2 : pcm) (T1 T2 : Type).

Lemma preimSXst (s : state (prodPCM U1 U2) (T1 * T2)) :
        (forall D : rel U1,
           'preim fst D (idfun.s s) (idfun.o s) = D (idfun.s s\1) (idfun.o s\1)) *
        (forall D : rel U2,
           'preim snd D (idfun.s s) (idfun.o s) = D (idfun.s s\2) (idfun.o s\2)).
Proof. by split=>D; rewrite preimX /= !lSXst !prodEQ0. Qed.

(* state the above lemma in form of sep *)
(* so that I don't have to unfold sep constantly *)
Lemma sepSXst (V : pcm) (s : state (prodPCM U1 U2) (T1 * T2)) :
        (forall D (f : {morphism D >-> V}),
           'sep (f \o fst) (idfun.s s) (idfun.o s) = 'sep f (idfun.s s\1) (idfun.o s\1)) *
        (forall D (f : {morphism D >-> V}),
           'sep (f \o snd) (idfun.s s) (idfun.o s) = D (idfun.s s\2) (idfun.o s\2)).
Proof. by split=>D f; rewrite !sepX preimSXst. Qed.

End SepProd.

Definition lprodX := (lprodTXst, lprodSXst, lprodJXst,
                      lprodTXst1, lprodSXst1, lprodJXst1).
Definition lprodE := (lprodTst, lprodSst, lprodJst,
                      lprodTst1, lprodSst1, lprodJst1,
                      (* I keep lprodX for automation *)
                      (* though I may remove it later *)
                      lprodX, sepSXst).
Definition lprodP := (lprodTpair, lprodSpair, lprodJpair).

(* lemmas for reindexing, i.e., automatically rewriting *)
(* projections out of sub-resource *)
Section LabelSub.
Variables (U V : pcm) (D : rel U) (X : sub_pcm D) (T : Type).

(* the Xst lemmas apply to any state indiscriminately *)
(* so be careful in automatic rewriting *)
(* still, they are better than if I had lemmas that apply to pairs *)
Lemma lvalSXst (s : state X T) :
        (forall f : _ -> V, (f \o pval).s s = f.s (spval s)) *
        (forall f : _ -> V, (f \o pval).o s = f.o (spval s)).
Proof. by split; move=>f; rewrite !lSXst /= !spE. Qed.

Lemma lvalTXst (s : state X T) :
        (forall f : _ -> V, (f \o pval).t s = f.t (spval s)).
Proof. by move=>f; rewrite !lvalSXst. Qed.

Lemma lvalJXst (s : state X T) :
        (forall f : _ -> V, f.j s = f.j (spval s)).
Proof. by move=>f; rewrite !lJX /= !spE. Qed.

(* the other direction should be stated wrt. s which is explicitly an
existential package of a state, and a proof that the subjective
components satisfy the requirements of the restricted PCM. For that,
we first need the form dual to spval (i.e., spval is elimination out
of a package, but we never built the corresponding introduction). It
just wasn't needed so far, but once it appears, its lemmas should be
placed here *)

Lemma lsubSXst (s : state U T) :
        (forall f : _ -> V, (f \o psub X).s s = f.s (spsub X s)) *
        (forall f : _ -> V, (f \o psub X).o s = f.o (spsub _ s)).
Proof. by split=>f; rewrite !lSXst !spunlock. Qed.

Lemma lsubTXst (s : state U T) :
        (forall f : _ -> V, (f \o psub X).t s = f.t (spsub X s)).
Proof. by move=>f; rewrite !lsubSXst. Qed.

Lemma lsubJXst (s : state U T) :
        (forall f : _ -> V, f.j s = f.j (spsub X s)).
Proof. by move=>f; rewrite !lJX /= !spunlock. Qed.

End LabelSub.


Section LabelSep.
Variables (U V : pcm) (D : rel U) (X : sub_pcm D) (T : Type).

Lemma preim_valSXst (s : state X T) :
        'preim pval D (idfun.s s) (idfun.o s) = D (idfun.s (spval s)) (idfun.o (spval s)).
Proof. by rewrite preimX /= -!lvalSXst !lA. Qed.

Lemma sep_valSXst (D' : rel U) (f : {morphism D' >-> V}) (s : state X T) :
        'sep (f \o pval) (idfun.s s) (idfun.o s) =
        'sep f (idfun.s (spval s)) (idfun.o (spval s)).
Proof.
by rewrite !sepX !sepE /= -!lvalSXst !(lCSX pval) -!comp1f -!compf1.
Qed.

Definition lsubX := (lvalTXst, lvalSXst, lvalJXst,
                     lsubTXst, lsubSXst, lsubJXst).
Definition lsubE := (preim_valSXst, sep_valSXst,
                     lvalTXst, lvalSXst, lvalJXst,
                     lsubTXst, lsubSXst).

End LabelSep.

Section SepZagZig.
Variables (U V : pcm) (D : sep_rel U) (T : Type).

Lemma sep_zagzig (s : state U T) (p : U) :
        allvalid (s \< p) -> D p (idfun.o s) ->
        D (idfun.s (s \> p)) (idfun.o (s \> p)) ->
        D (idfun.s (s \< p)) (idfun.o (s \< p)).
Proof.
move=>A X; rewrite !lE //= => H; rewrite sepxA //.
rewrite allvalidE /= !zg_unlock /= in A.
by rewrite !lSXst /= joinA.
Qed.

Lemma sep_zigzag (s : state U T) (p : U) :
        allvalid (s \< p) -> D (idfun.s s) p ->
        D (idfun.s (s \< p)) (idfun.o (s \< p)) ->
        D (idfun.s (s \> p)) (idfun.o (s \> p)).
Proof.
move=>A X; rewrite !lE //= => H; rewrite sepAx //.
by rewrite allvalidE /= !zg_unlock /= in A; rewrite !lSXst.
Qed.

Definition sepZ := (sep_zagzig, sep_zigzag).

End SepZagZig.

(********************************************)
(* a variant of pfV2 lemma adapted to state *)
(********************************************)

Lemma stV2 U V D T (f : @morphism U V D) :
        forall s : state U T, allvalid s ->
          D ((@idfun U).s s) (idfun.o s) -> valid (f.t s).
Proof.
move=>s; rewrite /allvalid -!lock Private_NotationST.state_to_subjE /= =>A.
by rewrite !lSX; apply: pfV2 A.
Qed.

Lemma stV2_phant U V D T (f : @morphism U V D) of phantom (_ -> _) f :
        forall s : state U T, allvalid s ->
          D ((@idfun U).s s) (idfun.o s) -> valid (f.t s).
Proof. by apply: stV2. Qed.

Notation stVf f := (@stV2_phant _ _ _ _ _ (Phantom (_ -> _) f)).


(* to compare to cap/iris, this lemma doesn't hold *)
(* this is why we will be creating subresources *)
(* whereby S is placed into the coh predicate and made global *)
Lemma sep_dup (U : pcm) T (s s1 s2 : state U T) (S : sep_rel U) :
        splits s s1 s2 -> S (idfun.s s) (idfun.o s) <->
                          S (idfun.s s1) (idfun.o s1) /\ S (idfun.s s2) (idfun.o s2).
Proof.
case=>H1 H2 H3 H4 H5.
rewrite !lSXst. rewrite H1 H2 H3 /=.
Abort.

Lemma stEX U T (s : state U T) : s = State (idfun.s s) (idfun.j s) (idfun.o s).
Proof. by rewrite -!lT -stEX'. Qed.

Lemma stE U T (s1 s2 : state U T) :
        s1 = s2 <-> [/\ idfun.s s1 = idfun.s s2,
                        idfun.j s1 = idfun.j s2 &
                        idfun.o s1 = idfun.o s2].
Proof. by rewrite -!lT -stE'. Qed.

Lemma allvL U T (s : state U T) (p : U) :
        valid (idfun.s s \+ p \+ idfun.o s) = allvalid (s \< p).
Proof. by rewrite allvalidE !zgE !lT. Qed.

Lemma allvR U T (s : state U T) (p : U) :
        valid (idfun.s s \+ p \+ idfun.o s) = allvalid (s \> p).
Proof. by rewrite allvalidE !zgE !joinA !lT. Qed.
