From Coq Require Import ssreflect ssrbool ssrfun.
From mathcomp Require Import ssrnat eqtype seq.
From fcsl Require Import pred prelude pcm automap heap natmap.
From fcsl Require Import morphism state world schedule process always model.
From fcsl Require Import options.

(* Definition of the combination operator (aka. tensor) over two resources *)

(* We first define the resource, then the morphisms into it. The latter *)
(* exist only if certain conditions are met. *)

(* Input to the construction are:

- R1 R2 : resources to be combined

- couple : Tr R1 -> Tr R2 -> Prop;

  relation saying which transitions to couple

- new_internals : Tr R1 -> Tr R2 -> bool;

  which of the coupled transitions will be internal.  Right now this
  is somewhat optimized as follows; we allow some omissions from
  it. E.g., we only need to list transition couplings where at least
  one of the transitions is externals. All the internal pairs will be
  automatically assumed internal without the need to list them here.

- pf : forall x y, new_internals x y -> footeqP [* coh R1, coh R2] [|
    interp R1 x, interp R2 y]

  proof that new internals are footprint preserving, so they indeed
  can be made internal transitions of the combined resource. *)

Definition footpresP R1 R2 (ints : Tr R1 -> Tr R2 -> bool) :=
  forall x y, ints x y -> footeqP [* coh R1, coh R2]
                                  [| interp R1 x, interp R2 y].

Module Tensor.
Section Tensor.
Variables R1 R2 : resource.
Variables couple : Tr R1 -> Tr R2 -> bool.
Variables ints : Tr R1 -> Tr R2 -> bool.
Variables pfint : footpresP ints.

(* State space: product PCM and product type *)
Definition U := prodPCM (Ux R1) (Ux R2).
Definition T : Type := Tp R1 * Tp R2.

(* Resorce invariant: the product of each resource invariant *)
Definition tcoh := [* coh R1, coh R2].

(* Names for transitions *)
(* coupling is given by coupling relation; and we just add id *)
Inductive tTr := couple_t x y of couple x y | id_t P of monoP tcoh P.

(* Semantics of the transition *)
Definition tinterp t :=
  match t with
    couple_t x y _ => [| interp R1 x, interp R2 y]
  | id_t P pf => pre_tr (id_tr tcoh) pf
  end.

(* The internal transitions in the tensor resource *)
(* include the old internals coupled with id, plus the *)
(* new internals declared in the new_internals relation *)
Definition tintern t :=
  match t with
    couple_t x y _ =>
      [/\ internal R1 x & internal R2 y] \/ ints x y
  | id_t _ _ => true
  end.

(* Proof that the internal transitions are footprint preserving *)
Program Canonical tensor := Eval hnf in ResPx tcoh tinterp tintern _.
Next Obligation.
case=>[x y qf|P qf _ ?? _ [_ ->]]; last by apply: domeqUn.
case=>[[I1 I2] s s' [C1 C2 _][H1 H2 _]|]; last by apply: pfint.
by apply: domeqUn; [apply: iint H1 | apply: iint H2].
Qed.

End Tensor.

Module Exports.
Notation tensor_coh := tcoh.
Notation tensorTr := tTr.
Notation tensor_interp := tinterp.
Notation tensor_intern := tintern.
Notation tensor := tensor.
Notation couple_t := couple_t.
Notation couple_tx c x y := (@couple_t _ _ c x y (erefl _)).
Canonical tensor.

(* lemmas for transfering stability properties *)
Section Lemmas.
Variables R1 R2 : resource.
Variables couple : Tr R1 -> Tr R2 -> bool.
Variables ints : Tr R1 -> Tr R2 -> bool.
Variables pfint : footpresP ints.

Notation tensor := (tensor couple pfint).

Lemma ost_fst s1 s2 : osteps tensor s1 s2 -> osteps R1 s1\1 s2\1.
Proof.
move=> H; move: s1 s2 H I; apply: iters_ind=>// s1 s2.
- by move=>s3; apply: iter_trans.
case; case=>[x y O []|P qf [_ /tr_inj ->]] //=.
by rewrite !prodE=>/iosteps.
Qed.

Lemma ost_snd s1 s2 : osteps tensor s1 s2 -> osteps R2 s1\2 s2\2.
Proof.
move=> H; move: s1 s2 H I; apply: iters_ind=>// s1 s2.
- by move=>s3; apply: iter_trans.
case; case=>[x y O []|P qf [_ /tr_inj ->]] //=.
by rewrite !prodE=>_ /iosteps.
Qed.

Lemma fst_mono P : monoP (coh R1) P -> monoP (coh tensor) (fun c => P c\1).
Proof. by move=>pf ??; rewrite /=/prod_raw !prodE; case=>???; apply: pf. Qed.

Lemma snd_mono P : monoP (coh R2) P -> monoP (coh tensor) (fun c => P c\2).
Proof. by move=>pf ??; rewrite /=/prod_raw !prodE; case=>???; apply: pf. Qed.

End Lemmas.

End Exports.
End Tensor.

Export Tensor.Exports.


(*************************************************************)
(* The left morphisms, i.e., inl_morph : R1 -> tensor R1 R2. *)
(*************************************************************)

Section LeftMorph.
Variables R1 R2 : resource.
Variables couple ints : Tr R1 -> Tr R2 -> bool.
Variables pfint : footpresP ints.
Notation tensor := (tensor couple pfint).

(* couple_l should "functionalize" the couple relation *)
(* it says how transitions of R1 are mapped to R2 *)
Variables couple_l : Tr R1 -> Tr R2.
Variables couple_lpf : forall t, couple t (couple_l t).
(* so far, our examples have only coupled internals with id *)
(* but the construction allows coupling internals with other internals *)
Variables iint_lpf : forall t, internal R1 t -> internal R2 (couple_l t).
Variables sim_forwardl : forall t s1 s1' s2,
  internal R1 t -> interp R1 t s1 s1' ->
  exists s2', interp R2 (couple_l t) s2 s2'.

(* state and transition component of the morphism *)
Definition inl_s (_ : unit) (s : st tensor) (x : st R1) := x = s\1.
Definition inl_t (_ : unit) (t : Tr R1) (s : st tensor) (t' : Tr tensor) :=
  t' = couple_t (couple_lpf t).

(* so defined, inl satisfies all the morphism properties *)
Lemma inl_local : morph_localP inl_s.
Proof.
exists (fun x => x.1); case=>p s s' _ _ ->.
by exists s\1; rewrite /inl_s !prodE.
Qed.

Lemma inl_other : morph_otherP inl_s.
Proof. by case; case=>???? _ _ _ _ ->-> O; rewrite -!prodEQ0 O. Qed.

Lemma inl_func : morph_funcP inl_s.
Proof. by case=>??? _ _ _ ->->. Qed.

Lemma inl_back : morph_backwardP inl_s.
Proof. by case; case. Qed.

Lemma inl_internal : morph_internalP inl_t.
Proof. by move=>t _ [s I] C ->; left; split=>//; apply: iint_lpf. Qed.

Lemma inl_tfunc : morph_tfuncP inl_t.
Proof. by case=>t sw _ _ C ->->. Qed.

Lemma inl_tlocal : morph_tlocalP inl_t.
Proof. by []. Qed.

Lemma inl_coh : morph_cohP' inl_s.
Proof. by case=>sw sv [C _ _] ->. Qed.

Definition inl_morph :=
  Morph (MorphMixin inl_local inl_other inl_func
         inl_back inl_internal inl_tfunc inl_tlocal inl_coh).

Lemma inl_ostep : morph_ostepP inl_morph (fun => xPredT).
Proof.
case=>s s' sv C C1 -> _ O;  exists s'\1; split=>//.
case: O; case=>[x y _ []|P pf [_ /tr_inj ->]] //.
by rewrite !prodE=>/iosteps.
Qed.

Lemma inl_forw : morph_forwardP inl_morph (fun => xPredT).
Proof.
move=>t [] s s' s1' I [C1 C2 V] _ ->{s'} _ H1.
case: (sim_forwardl s\2 I H1)=>s2' H2.
exists (couple_t (couple_lpf t)), tt, [@ s1', s2'].
rewrite /Morph.phi /= /inl_t /inl_s /prod_rel !sbeta; do !split=>//.
by rewrite (domeqVUn (iint I C1 H1) (iint (iint_lpf I) C2 H2) V).
Qed.

Definition inl_sim : simulation inl_morph := Simulation2 inl_ostep inl_forw.

End LeftMorph.

Arguments inl_s _ _ _ _ _ _ _ _ /.
Arguments inl_t _ _ _ _ _ _ _ _ /.
Arguments inl_morph _ _ _ _ _ _ _ _ /.
Arguments inl_sim _ _ _ _ _ _ _ _ /.

Section LeftMorphPack.
Variables R1 R2 : resource.
Variables couple ints : Tr R1 -> Tr R2 -> bool.
Variables pfint : footpresP ints.
Variables couple_l : Tr R1 -> Tr R2.
Notation tensor := (tensor couple pfint).

Definition inlmixP :=
  [/\ forall t, couple t (couple_l t) &
      forall t, internal R1 t -> internal R2 (couple_l t)].

Definition inlmixP2 :=
  [/\ forall t, couple t (couple_l t) &
      forall t, internal R1 t -> internal R2 (couple_l t)].

Lemma inlmixE : inlmixP2 -> inlmixP.
Proof. by case=>??; split. Qed.

Definition InlMorphPx (pf : inlmixP2) :=
  inl_morph pfint (let: conj a b := inlmixE pf in a)
                  (let: conj a b := inlmixE pf in b).

End LeftMorphPack.

Arguments InlMorphPx [R1 R2] couple [ints] pfint couple_l _.


(*****************************************************)
(* inr_morph : R2 -> tensor R1 R2 is completely dual *)
(*****************************************************)

Section RightMorph.
Variables R1 R2 : resource.
Variables couple : Tr R1 -> Tr R2 -> bool.
Variables new_internals : Tr R1 -> Tr R2 -> bool.
Variables pint: forall x y, new_internals x y  ->
  footeqP [* coh R1, coh R2] [| interp R1 x, interp R2 y].
Notation tensor := (tensor couple pint).

Variables couple_r : Tr R2 -> Tr R1.
Variables couple_rpf : forall t, couple (couple_r t) t.
Variables iint_rpf : forall t, internal R2 t -> internal R1 (couple_r t).
Variables sim_forwardr : forall t s2 s2' s1,
  internal R2 t -> interp R2 t s2 s2' ->
  exists s1', interp R1 (couple_r t) s1 s1'.

(* state and transition component of the morphism *)
Definition inr_s (_ : unit) (s : st tensor) (x : st R2) := x = s\2.
Definition inr_t (_ : unit) (t : Tr R2) (s : st tensor) (t' : Tr tensor) :=
  t' = couple_t (couple_rpf t).

Lemma inr_local : morph_localP inr_s.
Proof.
exists (fun x => x.2); case=>p s s' _ _ ->.
by exists s\2; rewrite /inr_s !prodE.
Qed.

Lemma inr_other : morph_otherP inr_s.
Proof. by case; case=>???? _ _ _ _ ->-> O; rewrite -!prodEQ0 O. Qed.

Lemma inr_func : morph_funcP inr_s.
Proof. by case=>??? _ _ _ ->->. Qed.

Lemma inr_back : morph_backwardP inr_s.
Proof. by case; case. Qed.

Lemma inr_internal : morph_internalP inr_t.
Proof. by move=>t t' [s I] C ->; left; split=>//; apply: iint_rpf. Qed.

Lemma inr_tfunc : morph_tfuncP inr_t.
Proof. by case=>t sw _ _ C ->->. Qed.

Lemma inr_tlocal : morph_tlocalP inr_t.
Proof. by []. Qed.

Lemma inr_coh : morph_cohP' inr_s.
Proof. by case=>s _ [_ C _] ->. Qed.

Definition inr_morph :=
  Morph (MorphMixin inr_local inr_other inr_func
         inr_back inr_internal inr_tfunc inr_tlocal inr_coh).

Lemma inr_ostep : morph_ostepP inr_morph (fun => xPredT).
Proof.
case=>s s' _ [C C1 V] _ -> _ O; exists s'\2; split=>//.
case: O; case=>[x y _ []|P pf [_ /tr_inj ->]] //.
by rewrite !prodE=>_ /iosteps.
Qed.

Lemma inr_forw : morph_forwardP inr_morph (fun => xPredT).
Proof.
move=>t [] s s' s2' I [C1 C2 V] _ -> _ H2.
case: (sim_forwardr s\1 I H2)=>s1' H1.
exists (couple_t (couple_rpf t)), tt, [@ s1', s2'].
rewrite /Morph.phi /= /prod_rel /inr_s !sbeta; do !split=>//.
by rewrite (domeqVUn (iint (iint_rpf I) C1 H1) (iint I C2 H2) V).
Qed.

Definition inr_sim := Simulation2 inr_ostep inr_forw.

End RightMorph.

Arguments inr_s _ _ _ _ _ _ _ _ /.
Arguments inr_t _ _ _ _ _ _ _ _ /.
Arguments inr_morph _ _ _ _ _ _ _ _ /.
Arguments inr_sim _ _ _ _ _ _ _ _ /.

Section RightMorphPack.
Variables R1 R2 : resource.
Variables couple ints : Tr R1 -> Tr R2 -> bool.
Variables pfint : footpresP ints.
Variables couple_r : Tr R2 -> Tr R1.
Notation tensor := (tensor couple pfint).

Definition inrmixP :=
  [/\ forall t, couple (couple_r t) t &
      forall t, internal R2 t -> internal R1 (couple_r t)].

Definition inrmixP2 :=
  [/\ forall t, couple (couple_r t) t &
      forall t, internal R2 t -> internal R1 (couple_r t)].

Lemma inrmixE : inrmixP2 -> inrmixP.
Proof. by case=>??; split. Qed.

Definition InrMorphPx (pf : inrmixP2) :=
  inr_morph pfint (let: conj a b := inrmixE pf in a)
                  (let: conj a b := inrmixE pf in b).

End RightMorphPack.

Arguments InrMorphPx [R1 R2] couple [ints] pfint couple_r _.


(**********************************************************)
(* Useful stability properties of inl_morph and inr_morph *)
(**********************************************************)

Section StabilityProperties.
Variables R1 R2 : resource.
Variables couple : Tr R1 -> Tr R2 -> bool.
Variables new_internals : Tr R1 -> Tr R2 -> bool.
Variables pint : forall x y, new_internals x y  ->
  footeqP [* coh R1, coh R2] [| interp R1 x, interp R2 y].
Notation tensor := (tensor couple pint).

(* we repeat the conditions for inl_morph existence *)
Variables couple_l : Tr R1 -> Tr R2.
Variables couple_lpf : forall t, couple t (couple_l t).
Variables iint_lpf : forall t, internal R1 t -> internal R2 (couple_l t).
Notation inl_morph := (inl_morph pint couple_lpf iint_lpf).

(* we repeat the conditions for inr_morph existence *)
Variables couple_r : Tr R2 -> Tr R1.
Variables couple_rpf : forall t, couple (couple_r t) t.
Variables iint_rpf : forall t, internal R2 t -> internal R1 (couple_r t).
Notation inr_morph := (inr_morph pint couple_rpf iint_rpf).

(* Explicitly passing the identity transitions *)
(* We only actually depend on the identity with the trivial precondition *)
Variable id_R1 : Tr R1.
Variable int_R1 : internal R1 id_R1.
Variables id_R1_eq : forall s s', interp R1 id_R1 s s' -> id_tr (coh R1) s s'.

Variable id_R2 : Tr R2.
Variable int_R2 : internal R2 id_R2.
Variables id_R2_eq : forall s s', interp R2 id_R2 s s' -> id_tr (coh R2) s s'.

(* relating msteps in tensor with osteps in components *)

Lemma ost_mst_inl R s1 s2 (f : morph unit R R1):
        (forall t t' x, internal R t -> f tt t x t' -> couple_l t' = id_R2) ->
        msteps (comp_morph f inl_morph tt) s1 s2 -> osteps R2 s1\2 s2\2.
Proof.
move=>E M; move: s1 s2 M I; apply: iters_ind=>// s1 s2.
- by move=>?; apply: iter_trans.
case=>t; last by move/iostep/iter1/ost_snd.
move=>ft {}/E E [x][t2][_ {}/E /= E] -> [_].
by rewrite E=>/id_R2_eq ->.
Qed.

Lemma ost_mst_inr R s1 s2 (f : morph unit R R2) :
        (forall t t' x, internal R t -> f tt t x t' -> couple_r t' = id_R1) ->
        msteps (comp_morph f inr_morph tt) s1 s2 -> osteps R1 s1\1 s2\1.
Proof.
move=>E M; move: s1 s2 M I; apply: iters_ind=>// s1 s2.
- by move=>?; apply: iter_trans.
case=>t; last by move/iostep/iter1/ost_fst.
move=>ft {}/E E [x][t2][_ {}/E/= E] -> [].
by rewrite E=>/id_R1_eq ->.
Qed.

(* specialize the above lemmas to f = identity *)

Lemma ost_mst_inl_id s1 s2 :
        (forall t, internal R1 t -> couple_l t = id_R2) ->
        msteps (inl_morph tt) s1 s2 -> osteps R2 s1\2 s2\2.
Proof.
move=>E M; move: s1 s2 M I; apply: iters_ind=>// s1 s2.
- by move=>?; apply: iter_trans.
case=>t; last by move/iostep/iter1/ost_snd.
by move=>ft {}/E E -> [_]; rewrite E=>/id_R2_eq ->.
Qed.

Lemma ost_mst_inr_id s1 s2 :
        (forall t, internal R2 t -> couple_r t = id_R1) ->
        msteps (inr_morph tt) s1 s2 -> osteps R1 s1\1 s2\1.
Proof.
move=>E M; move: s1 s2 M I; apply: iters_ind=>// s1 s2.
- by move=>?; apply: iter_trans.
case=>t; last by move/iostep/iter1/ost_fst.
by move=>ft {}/E E -> []; rewrite E=>/id_R1_eq ->.
Qed.

(* relating msteps in tensor with msteps in components *)

Lemma msteps_inl R s1 s2 (f : morph unit R R1) :
      msteps (comp_morph f inl_morph tt) s1 s2 -> msteps (f tt) s1\1 s2\1.
Proof.
move=>H; move: s1 s2 H I; apply: iters_ind=>// s1 s2.
- by move=>?; apply: iter_trans.
case=>[t t' I|].
- by case=>x1 [t1][-> /(SelfMStep I) Ef -> [/Ef/iter1]].
case=>[x y H []|P pf [_ /tr_inj ->]//].
by rewrite !prodE=>/(OtherMStep (f tt))/iter1.
Qed.

Lemma msteps_inr R s1 s2 (f : morph unit R R2) :
      msteps (comp_morph f inr_morph tt) s1 s2 -> msteps (f tt) s1\2 s2\2.
Proof.
move=>H; move: s1 s2 H I; apply: iters_ind=>// s1 s2.
- by move=>?; apply: iter_trans.
case=>[t t' I|].
- by case=>x2 [t2][-> /(SelfMStep I) Ef -> [_ /Ef/iter1]].
case=>[x y H []|P pf [_ /tr_inj ->]//].
by rewrite !prodE=>_ /(OtherMStep (f tt))/iter1.
Qed.

(* a common pattern is that we take a tensored resource *)
(* and then restrict it by a subresource construction *)
(* the following are stability lemmas specialized to that *)
(* scenario, where we take a subresource of a tensor *)
(* NOTE 1: they are **not** instances of the above lemmas *)
(* but require separate proofs *)
(* NOTE 2: we combined subjective and global subresource into *)
(* one construction so we only need one construction here as well *)

Variables (D : sep_rel (Ux tensor)) (P : st tensor -> Prop).
Variables (U : sub_pcm D) (pf : cglobalP (coh tensor) P).
Variables (qf : sub_persP D P).
Notation sub_tensor := (sub_res U pf qf).
Notation subtensor_morph := (sub_morph U pf qf).
Notation inl_submorph := (comp_morph inl_morph subtensor_morph).
Notation inr_submorph := (comp_morph inr_morph subtensor_morph).

Lemma msteps_subl s1 s2 :
        (forall t, internal R1 t -> couple_l t = id_R2) ->
        msteps (inl_submorph tt) s1 s2 -> osteps R2 (spval s1)\2 (spval s2)\2.
Proof. by move=>/ost_mst_inl_id H /msub_spval_comp/H. Qed.

Lemma msteps_subr s1 s2 :
        (forall t, internal R2 t -> couple_r t = id_R1) ->
        msteps (inr_submorph tt) s1 s2 -> osteps R1 (spval s1)\1 (spval s2)\1.
Proof. by move=>/ost_mst_inr_id H /msub_spval_comp/H. Qed.

End StabilityProperties.
