From Coq Require Import ssreflect ssrbool ssrfun.
From fcsl Require Import axioms pred prelude pcm.
From fcsl Require Import unionmap heap morphism notation state.
From fcsl Require Import options.
Set Warnings "-projection-no-head-constant".

(*************************************************************************)
(* Definition of resources and related notions:                          *)
(* - subjective relations, and operations on them: relation              *)
(*   transposition, union, iteration                                     *)
(* - transition = deterministic subjective relations                     *)
(* - atoms = deterministic *families* of transitions                     *)
(*************************************************************************)

Section GlobalPredicates.
Variables (U : pcm) (T : Type).
Implicit Type P C : state U T -> Prop.
Implicit Type flat : state U T -> heap.
Implicit Type s : state U T.

(** The intuition behind the definition of [globalP] is
   if the state is "good" from the point of view of one thread, then
   it is good from the point of view of any other thread as well **)

Definition validP C := forall s, C s -> allvalid s.
Definition flat_validP C flat := forall s, C s -> valid (flat s).

Definition globalP C := forall p s, C (s \< p) <-> C (s \> p).
Definition flat_globalP C flat :=
  forall p s, C (s \< p) -> flat (s \< p) = flat (s \> p).

Definition cglobalP C P := forall p s, C (s \< p) -> P (s \< p) <-> P (s \> p).

End GlobalPredicates.

(*****************************)
(*****************************)
(* Resource invariants       *)
(* aka. coherence predicates *)
(* aka. state spaces         *)
(*****************************)
(*****************************)

Section CohDefs.
Variables (U : pcm) (T : Type).
Variables (coh : state U T -> Prop) (flat : state U T -> heap).
Implicit Type s : state U T.

Open Scope type_scope.

Definition cohP :=
  (forall s, coh s -> allvalid s /\ valid (flat s)) /\
  forall p s, (coh (s \< p) <-> coh (s \> p)) /\ (flat (s \< p) = flat (s \> p)).

Lemma cohmixP :
        cohP -> [/\ validP coh, flat_validP coh flat,
                    globalP coh & flat_globalP coh flat].
Proof.
case=>H1 H2; split; try by [move=>s /H1 []]; by move=>p s; case: (H2 p s).
Qed.

(* we can weaken the second conjunct to give us the validity assumption *)

Definition cohP2 :=
  (forall s, coh s -> allvalid s /\ valid (flat s)) /\
  (forall p s, allvalid (s \< p) -> (coh (s \< p) <-> coh (s \> p)) /\
     (coh (s \< p) -> flat (s \< p) = flat (s \> p))).

Lemma cohmixP2 : cohP2 -> [/\ validP coh, flat_validP coh flat ,
                              globalP coh & flat_globalP coh flat].
Proof.
case=>H1 H2; split; try by move=>s /H1 [].
- move=>p s; split=>C; case: (H1 _ C);
  by rewrite -?zgVE=>V _; apply/(proj1 (H2 p s V)).
by move=>p s C; apply: (proj2 (H2 p s (proj1 (H1 _ C)))).
Qed.

End CohDefs.

Module CohPred.
Section CohPred.
Variables (U : pcm) (T : Type).

Structure mixin_of (coh : state U T -> Prop) := Mixin {
  flat_op : state U T -> heap;
  _ : validP coh;
  _ : flat_validP coh flat_op;
  _ : globalP coh;
  _ : flat_globalP coh flat_op}.

Section ClassDef.

Notation class_of := mixin_of (only parsing).

Structure cohpred : Type := Pack {sort : state U T -> Prop; _ : class_of sort}.
Local Coercion sort : cohpred >-> Funclass.

Variables (coh : state U T -> Prop) (cT : cohpred).

Definition class := let: Pack _ c as cT' := cT return class_of cT' in c.

Definition pack c := @Pack coh c.
Definition clone := fun c & coh = cT & phant_id (pack c) cT => pack c.

Definition flat := flat_op class.
End ClassDef.
End CohPred.

Module Exports.
Coercion sort : cohpred >-> Funclass.
Notation cohpred U T := (cohpred U T).
Notation CohPredMixin := (@Mixin _ _).
Notation CohPred coh m := (@pack _ _ coh m).

Section CombinedConstructor.
Variables (U : pcm) (T : Type).
Variables (coh : state U T -> Prop) (flat : state U T -> heap).

Definition CohMix2 pf :=
  @Mixin U T coh flat (let: And4 a b c d := cohmixP2 pf in a)
                      (let: And4 a b c d := cohmixP2 pf in b)
                      (let: And4 a b c d := cohmixP2 pf in c)
                      (let: And4 a b c d := cohmixP2 pf in d).

Definition CohPx2 pf := @CohPred.pack U T coh (CohMix2 pf).

End CombinedConstructor.

Arguments CohPx2 [U T] coh flat _.

Notation "[ 'cohpred' 'of' coh 'for' C ]" := (@clone _ _ coh C _ (erefl _) idfun)
  (at level 0, format "[ 'cohpred'  'of'  coh  'for'  C ]") : form_scope.
Notation "[ 'cohpred' 'of' coh ]" := (@clone _ _ coh _ _ (erefl _) id)
  (at level 0, format "[ 'cohpred'  'of'  coh ]") : form_scope.

Notation flat := flat.

Section Repack.
Variables (U : pcm) (T : Type) (coh : cohpred U T).

Lemma coh_valid : validP coh.
Proof. by case: coh=>c [? Vs Vf Ts Vz]; apply: Vs. Qed.

Lemma coh_flat_valid : flat_validP coh (flat coh).
Proof. by case: coh=>c [? Vs Vf Ts Vz]; apply: Vf. Qed.

Lemma coh_global : globalP coh.
Proof. by case: coh=>c [? Vs Vf Ts Vz]; apply: Ts. Qed.

Lemma coh_flat_global : flat_globalP coh (flat coh).
Proof. by case: coh=>c [? Vs Vf Ts Vz]; apply: Vz. Qed.

(* some abbreviations *)
Definition cohZ := coh_global.
Definition flatZ := coh_flat_global.

End Repack.

Section Lemmas.
Variables (U : pcm) (T : Type).
Implicit Type s : state U T.

Variables (coh : cohpred U T) (P : state U T -> Prop) (pf : cglobalP coh P).

Lemma coh_validT s : coh s -> valid (self s \+ other s).
Proof. rewrite -allvalidE; exact: coh_valid. Qed.

Lemma cohso s : coh s -> valid (idfun.t s).
Proof. by move/coh_valid; rewrite allvE. Qed.

Lemma coh_out s : coh s -> coh (outs s).
Proof. by move=>C; rewrite /outs zgEX -coh_global -zgEX unitL -stEX'. Qed.

Lemma coh_tr s : coh (tr s) <-> coh s.
Proof.
rewrite [s]stEX' trEX -{1}[other s]unitL zgEX coh_global.
by rewrite -zgEX joinC zgEX -coh_global -zgEX unitL.
Qed.

Lemma coh_flat_tr s : coh s -> flat coh (tr s) = flat coh s.
Proof.
move=>C; rewrite [s]stEX' trEX -{1}[other s]unitL zgEX coh_flat_global -zgEX.
- by rewrite joinC zgEX -coh_flat_global -zgEX unitL -stEX'.
by rewrite unitL -trEX -stEX' coh_tr.
Qed.

Lemma coh_gbL s s1 s2 : splits s s1 s2 -> coh s <-> coh s1.
Proof.
by case=>X X1 X2 J1 J2; rewrite [s]stEX' [s1]stEX' X zgEX coh_global -zgEX X1 J1.
Qed.

Lemma coh_gbR s s1 s2 : splits s s1 s2 -> coh s <-> coh s2.
Proof. by move/spC/coh_gbL. Qed.

Lemma cglob_tr s : coh s -> P (tr s) <-> P s.
Proof.
rewrite [s]stEX' trEX=>C.
rewrite -{1}[other s]unitL zgEX pf; last by rewrite -zgEX unitL -coh_tr trEX.
rewrite -zgEX joinC zgEX -pf; last by rewrite -zgEX unitL.
by rewrite -zgEX unitL.
Qed.

Lemma globalT : cglobalP coh PredT.
Proof. by []. Qed.

End Lemmas.

Arguments globalT [U T coh].
Prenex Implicits globalT.
Hint Resolve globalT : core.

End Exports.
End CohPred.

Export CohPred.Exports.


(*****************************************)
(* Constructors for coherence predicates *)
(*****************************************)

(* empty coherence *)

Module EmpCoh.
Section EmpCoh.
Implicit Type s : state unitPCM unit.

Definition emp_raw s := True.
Definition emp_flat s := Unit : heap.

Lemma emp_valid : validP emp_raw.
Proof. by []. Qed.

Lemma emp_global : globalP emp_raw.
Proof. by []. Qed.

Lemma emp_flat_valid : flat_validP emp_raw emp_flat.
Proof. by []. Qed.

Lemma emp_flat_global : flat_globalP emp_raw emp_flat.
Proof. by []. Qed.

End EmpCoh.

Module Exports.
Notation emp_raw := emp_raw.

Definition empCohMix :=
  CohPredMixin emp_valid emp_flat_valid emp_global emp_flat_global.
Canonical emp_coh := Eval hnf in CohPred emp_raw empCohMix.

Lemma flatEmpE s : flat emp_coh s = Unit.
Proof. by []. Qed.

End Exports.
End EmpCoh.

Export EmpCoh.Exports.


(* coherence for products *)
Module ProdCoh.
Section ProdCoh.
Variables (U1 U2 : pcm) (T1 T2 : Type).
Variables (coh1 : cohpred U1 T1) (coh2 : cohpred U2 T2).

Local Notation prod_flat := (fun s => flat coh1 s\1 \+ flat coh2 s\2).
Definition prod_raw s := [/\ coh1 s\1, coh2 s\2 & valid (prod_flat s)].

Lemma prod_valid : validP prod_raw.
Proof.
move=>s [/coh_valid V1 /coh_valid V2] _; rewrite !allvalidE in V1 V2 *.
by rewrite !prodX /= V1 V2.
Qed.

Lemma prod_flat_valid : flat_validP prod_raw prod_flat.
Proof. by move=>s []. Qed.

Lemma prod_global : globalP prod_raw.
Proof.
move=>p s; rewrite /prod_raw/prod_flat !prodE /=.
split; case=>C1 C2 V.
- by rewrite -!cohZ -!flatZ.
by rewrite !flatZ !cohZ.
Qed.

Lemma prod_flat_global : flat_globalP prod_raw prod_flat.
Proof. by move=>p s; rewrite /prod_raw !prodE=>[[??]]; rewrite !flatZ. Qed.

End ProdCoh.

Module Exports.
Notation prod_raw := prod_raw.

Section Structs.
Variables (U1 U2 : pcm) (T1 T2 : Type).
Variables (coh1 : cohpred U1 T1) (coh2 : cohpred U2 T2).

Definition prodCohMixin :=
  CohPredMixin (@prod_valid _ _ _ _ coh1 coh2)
               (@prod_flat_valid _ _ _ _ coh1 coh2)
               (@prod_global _ _ _ _ coh1 coh2)
               (@prod_flat_global _ _ _ _ coh1 coh2).
Canonical prod_coh :=
  Eval hnf in CohPred (prod_raw coh1 coh2) prodCohMixin.

Lemma flatPE s1 s2 : flat prod_coh [@ s1, s2] = flat coh1 s1 \+ flat coh2 s2.
Proof. by rewrite /flat /= !prodE. Qed.

End Structs.

Notation "[ '*' coh1 , coh2 ]" := (prod_coh coh1 coh2)
  (at level 0, format "[ '*'  coh1 ,  coh2 ]").

End Exports.
End ProdCoh.

Export ProdCoh.Exports.


Module SubUCoh.
Section SubUCoh.
Variables (U : pcm) (P : U -> Prop) (pc : pcm_closed P).
Variables (T : Type) (Q : T -> Prop).
Notation ustval := (@ustval U T P pc Q).
Variable coh : cohpred U T.

Definition sub_coh_raw := coh \o ustval.
Definition sub_flat_raw := flat coh \o ustval.

Lemma sub_coh_valid : validP sub_coh_raw.
Proof.
case; case=>s ps [j pj][o po] /coh_valid.
by rewrite !ustEX !allvalidE.
Qed.

Lemma sub_coh_flat_valid : flat_validP sub_coh_raw sub_flat_raw.
Proof. by case; case=>s ps j [o po] /coh_flat_valid. Qed.

Lemma sub_global : globalP sub_coh_raw.
Proof.
case=>p pf; case; case=>??? [??].
by rewrite /sub_coh_raw /= !zg_unlock !ustEX !zgEX cohZ.
Qed.

Lemma sub_flat_global : flat_globalP sub_coh_raw sub_flat_raw.
Proof.
case=>p pf; case=>[[s ps][j pj][o po]].
rewrite /sub_coh_raw /sub_flat_raw !zg_unlock /= !ustEX !zgEX.
by apply: flatZ.
Qed.

End SubUCoh.

Module Exports.
Section Exports.
Variables (U : pcm) (P : U -> Prop) (pc : pcm_closed P).
Variables (T : Type) (Q : T -> Prop).
Variable coh : cohpred U T.

Definition usubCohMix :=
  CohPredMixin (@sub_coh_valid U P pc T Q coh)
               (@sub_coh_flat_valid U P pc T Q coh)
               (@sub_global U P pc T Q coh)
               (@sub_flat_global U P pc T Q coh).
Definition usub_coh := Eval hnf in CohPred _ usubCohMix.

Lemma flat_ustval (s : substate pc Q) :
        flat coh (ustval pc Q s) = flat usub_coh s.
Proof. by []. Qed.

End Exports.
End Exports.
End SubUCoh.

Export SubUCoh.Exports.


(* 2nd attempt. Restricting the PCM of a coherence predicate *)
(* and adding a conjunct simultaneously *)

(* conditionally global predicate *)

Coercion coh_global : cohpred >-> globalP.

Module SubCoh.
Section SubCoh.
Variables (V : pcm) (D : rel V) (U : sub_pcm D).
Variables (T : Type) (coh : cohpred V T).
Variables (P : state V T -> Prop) (cglobP : cglobalP coh P).
Notation spval := (@spval V D U T).

Definition sub_coh_raw (s : state U T) := [/\ allvalid s, P (spval s) & coh (spval s)].

Definition sub_flat_raw s := flat coh (spval s).

Lemma sub_valid : validP sub_coh_raw.
Proof. by move=>s []. Qed.

Lemma sub_flat_valid : flat_validP sub_coh_raw sub_flat_raw.
Proof. by move=>s [W _ /coh_flat_valid]. Qed.

Lemma sub_global : globalP sub_coh_raw.
Proof.
move=>p s; rewrite /sub_coh_raw !allvalidE !zgE -joinA.
split; case=>W; rewrite !spE ?(validR W) ?(validAL W) // => H C.
- by rewrite -cglobP // -cohZ.
by rewrite cglobP // cohZ.
Qed.

Lemma sub_flat_global : flat_globalP sub_coh_raw sub_flat_raw.
Proof.
move=>p s W; have [V1 V2] : valid (self s \+ p) /\ valid (p \+ other s).
- by case: W; rewrite allvalidE !zgE=>W _ _; rewrite (validL W) (validAR W).
rewrite /sub_flat_raw !spE ?flatZ //.
by case: W; rewrite !spE.
Qed.

Definition sub_cohMix :=
  CohPredMixin sub_valid sub_flat_valid sub_global sub_flat_global.
Canonical sub_coh := Eval hnf in CohPred _ sub_cohMix.

(* massaging of the coh predicate *)
Lemma coh_sep s :
        sub_coh s <->
        [/\ D (idfun.s (spval s)) (idfun.o (spval s)),
            P (spval s) & coh (spval s)].
Proof.
split; first by case; rewrite allvalidE valid_sep !spE !lT; case/andP.
by case=>S H C; split=>//; rewrite allvalidE valid_sep /= !spE (coh_validT C) !lT.
Qed.

End SubCoh.

Module Exports.
Notation sub_coh_raw := sub_coh_raw.
Notation sub_coh := sub_coh.
Canonical sub_coh.
Notation coh_sep := coh_sep.
End Exports.
End SubCoh.

Export SubCoh.Exports.


(************************)
(************************)
(* Subjective Relations *)
(************************)
(************************)

Notation srel U T := (state U T -> state U T -> Prop).

(* Note: Consider making srel a type; will make it easier/possible to *)
(* define richer inheritance patterns using canonical structures if it's *)
(* it's own type then use (x, y) \In R notation, instead of R x y, to *)
(* prevent accidential unfolding. *)

Section SubjectiveRelsDefs.
Variables (U U1 U2 : pcm) (T T1 T2 A : Type).
Implicit Types (P : state U T -> Prop) (g : srel U T).
Variable (coh : state U T -> Prop).

(* All the important concepts for resources and transitions *)
(** _other-preserving_ relation: we are not allowed to change the other part of the state *)
Definition otherP g := forall s s', g s s' -> other s = other s'.
Definition zagP g := forall p s s',
  coh (s \< p) -> g (s \> p) s' ->
  exists s'', s' = s'' \> p /\ g (s \< p) (s'' \< p).

(** _local_ relation: *)
Definition localP (g : srel U T) := forall p s1 s2,
  coh (s1 \< p) -> other s1 = other s2 ->
  g (s1 \> p) (s2 \> p) -> g (s1 \< p) (s2 \< p).

(* zagging for unary predicates is safety monotonicity *)
Definition monoP P :=
  forall p s, coh (s \< p) -> P (s \> p) -> P (s \< p).
(* framing wrt. generic P -- P will usually a custom notion of safety *)
Definition frameP P g := forall p s s',
  coh (s \< p) -> P (s \> p) -> g (s \< p) s' ->
  exists s'', s' = s'' \< p /\ g (s \> p) (s'' \> p).
Definition deterP (a : A -> srel U T) :=
  forall s r1 s1 r2 s2, coh s -> a r1 s s1 -> a r2 s s2 -> r1 = r2 /\ s1 = s2.
Definition funcrP (a : A -> srel U T) :=
  forall s s1 s2 r1 r2, coh s -> a r1 s s1 -> a r2 s s2 -> r1 = r2.
Definition flatP coh (g : srel U T) :=
  forall s s', g s s' -> valid (flat coh s) -> valid (flat coh s').

(* I can pick either zagP or localP for the basis; they are *)
(* equivalent. Proofs of localP tend to be simpler, so I go with that. *)

Lemma zagzigE g : otherP g -> zagP g <-> localP g.
Proof.
move=>Ot; split=>Z p s s' C.
- move=>O /(Z _ _ _ C) [s2][/=]; rewrite !zg_unlock.
  by case=>->-> _ H; rewrite -O [in X in g _ X](Ot _ _ H).
move=>H; set s'' := State (self s') (joint s') (other s).
have E : s' = s'' \> p by rewrite [s']stEX' -(Ot _ _ H) !zg_unlock.
by rewrite E in H; exists s''; split=>//; apply: Z H.
Qed.

Lemma det_func a v : deterP a -> {In coh, functional (a v)}.
Proof. by move=>H s C s1 s2 G /(H _ _ _ _ _ C G) []. Qed.

Lemma det_funcr a : deterP a -> funcrP a.
Proof. by move=>H s r1 s1 r2 s2 C G /(H _ _ _ _ _ C G) []. Qed.

(* Some basic constructions *)
Definition trr g s s' := g (tr s) (tr s').
Definition ex_rel (a : A -> srel U T) s s' := exists x : A, a x s s'.

(* We always seem to take transition products wrt. coherence predicates. *)
(* This requires combination of reinv and nprod; we define it directly *)

End SubjectiveRelsDefs.

(* Some basic lemmas about monoP *)

Lemma idM U T coh : monoP coh (fun x : state U T => True).
Proof. by []. Qed.

Lemma andM U T coh (P1 P2 : state U T -> Prop) :
        monoP coh P1 -> monoP coh P2 -> monoP coh (fun s => P1 s /\ P2 s).
Proof. by move=>H1 H2 p s C [/(H1 _ _ C) X1 /(H2 _ _ C) X2]. Qed.

Lemma prodM U1 U2 T1 T2 (coh1 : cohpred U1 T1) (coh2 : cohpred U2 T2) P1 P2 :
        monoP coh1 P1 -> monoP coh2 P2 -> monoP [* coh1, coh2] (fun s => P1 s\1 /\ P2 s\2).
Proof.
move=>H1 H2 [p1 p2] s; rewrite /= /prod_raw !prodE /=; case=>C1 C2 _.
by case=>/(H1 _ _ C1) X1 /(H2 _ _ C2) X2.
Qed.

Arguments idM [U T coh] p s _.
Prenex Implicits idM.

(*************************)
(* Transposing relations *)
(*************************)

Section TransposedRels.
Variables (U : pcm) (T : Type).
Implicit Types (g : srel U T) (coh : cohpred U T).

Definition selfP g := forall s s', g s s' -> self s = self s'.
Definition zigzagP (coh : _ -> Prop) g := forall p s s',
  coh (s \< p) -> self s = self s' ->
  g (s \< p) (s' \< p) -> g (s \> p) (s' \> p).
Definition zigP (coh : _ -> Prop) g := forall p s s',
  coh (s \< p) -> g (s \< p) s' ->
  exists s'', s' = s'' \< p /\ g (s \> p) (s'' \> p).

Lemma trrI g : trr (trr g) = g.
Proof. by apply/fext=>s; apply/fext=>s'; rewrite /trr !trI. Qed.

Lemma trr_func coh g : {In coh, functional (trr g)} <-> {In coh, functional g}.
Proof.
split=>H s /coh_tr C s1 s2; last by move=>H1 /(H _ C _ _ H1)/tr_inj.
by rewrite -[s]trI -[s1]trI -[s2]trI => H1 /(H _ C _ _ H1) ->.
Qed.

Lemma trr_self g : selfP (trr g) <-> otherP g.
Proof.
split=>H s s' X; rewrite -(trE0 s) -(trE0 s'); apply: H=>//.
by rewrite /trr !trI.
Qed.

Lemma trr_zigzag (coh : cohpred U T) g :
        zigzagP coh (trr g) <-> localP coh g.
Proof.
split=>H p s s' C; rewrite -{1}(trE0 s) -{1}(trE0 s');
move/(H p); rewrite -!trE1 !coh_tr -cohZ; move/(_ C);
by rewrite !tr_unlock /= !zg_unlock /= /trr /= !trEX /= !(joinC p).
Qed.

Lemma trr_zig (coh : cohpred U T) g : zigP coh (trr g) <-> zagP coh g.
Proof.
split=>H p s s' C.
- rewrite -[g]trrI {1 3}/trr !trE.
  case/H; first by rewrite -trE1 coh_tr -cohZ.
  by move=>s2 [E R]; exists (tr s2); rewrite -[s']trI E !trE.
rewrite /trr /= !trE.
case/H; first by rewrite -trE1 coh_tr -cohZ.
by move=>s2 [E R]; exists (tr s2); rewrite -[s']trI E !trE.
Qed.

Lemma zigzagE (coh : cohpred U T) g : selfP g -> zigP coh g <-> zigzagP coh g.
Proof. by rewrite -[g]trrI trr_self trr_zig trr_zigzag; apply: zagzigE. Qed.

End TransposedRels.


(*********************************************************)
(* Reflexive-transitive closure of a subjective relation *)
(*********************************************************)

Section SubjIteratedRels.
Variables (U : pcm) (T : Type).
Implicit Type (g : srel U T).
Variable (coh : cohpred U T).

Lemma iter_other' g n : otherP g -> otherP (iter' g n).
Proof. by move=>Ot s s1; elim: n s=>[s ->|n IH s] // [s2][/Ot ->] /IH. Qed.

Lemma iter_self' g n : selfP g -> selfP (iter' g n).
Proof. by move=>St s s1; elim: n s=>[s ->|n IH s] // [s2][/St ->] /IH. Qed.

Lemma iter_zag' g n : preserved coh g -> zagP coh g -> zagP coh (iter' g n).
Proof.
move=>pf Z p s s' C; elim: n s C=>[|n IH] s /= C; first by exists s.
case=>_ [/(Z _ _ _ C)][s1][-> G] /(IH _ (pf _ _ G C)) [s2][-> H].
by exists s2; split=>//; exists (s1 \< p).
Qed.

Lemma iter_zig' g n : preserved coh g -> zigP coh g -> zigP coh (iter' g n).
Proof.
move=>pf Z p s s'; elim: n s=>[|n IH] s /= C; first by exists s.
case=>_ [/(Z _ _ _ C)][s1][-> G]; move/cohZ/(pf _ _ G)/cohZ: C=>C.
by case/(IH _ C)=>s2 [-> H]; exists s2; split=>//; exists (s1 \> p).
Qed.

Lemma iter_zagzig' g n :
        otherP g -> preserved coh g ->
        zagP coh g -> localP coh (iter' g n).
Proof.
move=>Ot pf Z p s s'; elim: n s=>[|n IH] s /= C.
- by rewrite !zg_unlock=>-> [->->].
move=>O [_][/(Z _ _ _ C)][s2][->] G H; move/esym: (Ot _ _ G).
by rewrite !zgE /= {1}O=>/(IH _ (pf _ _ G C))/(_ H); exists (s2 \< p).
Qed.

Lemma iter_other g : otherP g -> otherP (iter g).
Proof. by move=>Ot s s' [n] /(iter_other' Ot). Qed.

Lemma iter_self g : selfP g -> selfP (iter g).
Proof. by move=>St s s' [n] /(iter_self' St). Qed.

Lemma iter_zag g : preserved coh g -> zagP coh g -> zagP coh (iter g).
Proof.
move=>pf Z p s _ C [n] /(iter_zag' pf Z C) [s'][->].
by exists s'; split=>//; exists n.
Qed.

Lemma iter_zig g : preserved coh g -> zigP coh g -> zigP coh (iter g).
Proof.
move=>pf Z p s _ C [n] /(iter_zig' pf Z C) [s'][->].
by exists s'; split=>//; exists n.
Qed.

Lemma iter_zagzig g :
        otherP g -> preserved coh g ->
        zagP coh g -> localP coh (iter g).
Proof.
by move=>Ot pf Z p s s' C O [n] /(iter_zagzig' Ot pf Z C O); exists n.
Qed.

End SubjIteratedRels.


(**************)
(**************)
(* Flattening *)
(**************)
(**************)

Section Flattening.
Variables (U : pcm) (T : Type).
Implicit Type (coh : cohpred U T) (r : srel U T).

Definition footeqP coh r :=
  forall s s', coh s -> r s s' -> dom_eq (flat coh s) (flat coh s').

Definition footacqP coh r h :=
  forall s s', coh s -> r s s' -> dom_eq (h \+ flat coh s) (flat coh s').

Definition footrelP coh r h :=
  forall s s', coh s -> r s s' -> dom_eq (flat coh s) (h \+ flat coh s').

End Flattening.

(***************************)
(* Invariance of relations *)
(***************************)

(* wrt. arbitrary resource invariant *)

Section CohPreservationLemmas.
Variables (U U' : pcm) (T T' : Type) (coh : cohpred U T) (coh' : cohpred U' T').

Lemma trr_coh (g : srel U T) : preserved coh g -> preserved coh (trr g).
Proof. by move=>S s s' /S H /coh_tr /H /coh_tr. Qed.

Lemma post_coh g : preserved coh (post_rel g coh).
Proof. by move=>s s' []. Qed.

Lemma ex_coh A (g : A -> srel U T) :
        (forall x, preserved coh (g x)) -> preserved coh (ex_rel g).
Proof. by move=>S s s' [x] /S. Qed.

End CohPreservationLemmas.


(***************)
(***************)
(* Transitions *)
(***************)
(***************)

(* these are relations with uniqP, otherP and localP properties *)

Module Trans.
Section ClassDef.
Variables (U : pcm) (T : Type) (coh : cohpred U T).

(* let's make transitions uniquely determine the end state *)
(* we thus don't allow the unioning of transitions *)
(* also, disentangle transitions from coh space *)

Structure mixin_of (trans : srel U T) := Mixin {
  _ : {In coh, functional trans};
  _ : otherP trans;
  _ : localP coh trans;
  _ : preserved coh trans}.

Definition class_of := mixin_of.

Structure trans := Pack {sort : srel U T; _ : class_of sort}.
Local Coercion sort : trans >-> Funclass.

Variables (rel : state U T -> state U T -> Prop) (cT : trans).
Definition class := let: Pack _ c as cT' := cT return class_of cT' in c.
Definition pack c := @Pack rel c.
Definition clone := fun c & rel = cT & phant_id (pack c) cT => pack c.

End ClassDef.

Module Exports.
Coercion sort : trans >-> Funclass.
Notation trans := trans.
Notation TransMixin f o z c := (@Mixin _ _ _ _ f o z c).
Notation Trans m := (@pack _ _ _ _ m).

Section CombinedMixin.
Variables (U : pcm) (T : Type) (g : srel U T) (coh : cohpred U T).

Definition transF := [/\ {In coh, functional g}, otherP g, localP coh g & preserved coh g].

Definition TransMixF (pf : transF) :=
  let: And4 a b c d := pf in @Mixin U T coh g a b c d.

Definition transP := forall s1 s2,
  [/\ forall s, g s s1 -> g s s2 -> coh s -> s1 = s2,
      g s1 s2 -> idfun.o s1 = idfun.o s2,
      forall p, idfun.o s1 = idfun.o s2 ->
        g (s1 \> p) (s2 \> p) -> coh (s1 \> p) -> coh (s1 \< p) ->
        g (s1 \< p) (s2 \< p) &
      g s1 s2 -> coh s1 -> coh s2].

Lemma tcP : transP -> transF.
Proof.
move=>H; split=>[s C||p|] s1 s2; rewrite ?lT;
case: (H s1 s2)=>H1 H2 H3 H4; try by eauto.
by move=>C O G; apply: H3 O G _ (C); apply/cohZ.
Qed.

Definition TransMix pf :=
  @Mixin U T coh g (let: And4 a b c d := tcP pf in a)
                   (let: And4 a b c d := tcP pf in b)
                   (let: And4 a b c d := tcP pf in c)
                   (let: And4 a b c d := tcP pf in d).

Definition TransPx pf := @pack _ _ _ _ (TransMix pf).

Definition transP2 := forall s1 s2,
  [/\ forall s, g s s1 -> g s s2 -> coh s -> s1 = s2,
      g s1 s2 -> idfun.o s1 = idfun.o s2,
      forall p, allvalid (s1 \< p) -> allvalid (s2 \< p) ->
        idfun.o s1 = idfun.o s2 -> g (s1 \> p) (s2 \> p) ->
        coh (s1 \> p) -> coh (s1 \< p) -> g (s1 \< p) (s2 \< p) &
      g s1 s2 -> coh s1 -> coh s2].

Lemma tcP2 : transP2 -> transF.
Proof.
move=>H; split=>[s C||p|] s1 s2; rewrite ?lT;
case: (H s1 s2)=>H1 H2 H3 H4; try by eauto.
move=>C O X; apply: H3=>//; first by apply: coh_valid C.
- move/cohZ: C=>C; rewrite zgVE.
  by case: (H (s1 \> p) (s2 \> p))=>_ _ _ /(_ X C)/coh_valid.
by rewrite -cohZ.
Qed.

Definition TransMix2 pf :=
  @Mixin U T coh g (let: And4 a b c d := tcP2 pf in a)
                   (let: And4 a b c d := tcP2 pf in b)
                   (let: And4 a b c d := tcP2 pf in c)
                   (let: And4 a b c d := tcP2 pf in d).

Definition TransPx2 pf := @pack _ _ _ _ (TransMix2 pf).

End CombinedMixin.

Arguments TransPx [U T] g coh _.
Arguments TransPx2 [U T] g coh _.

Notation "[ 'trans' 'of' rel 'for' C ]" := (@clone _ _ C rel _ (erefl _) idfun)
  (at level 0, format "[ 'trans'  'of'  rel  'for'  C ]") : form_scope.
Notation "[ 'trans' 'of' rel ]" := (@clone _ _ _ rel _ (erefl _) id)
  (at level 0, format "[ 'trans'  'of'  rel ]") : form_scope.

Section Repack.
Variables (U : pcm) (T : Type) (coh : cohpred U T) (g : trans coh).

Lemma trans_func : {In coh, functional g}.
Proof. by case: g=>G [Gc Ge Gz C]; apply: Gc. Qed.

Lemma trans_other : otherP g.
Proof. by case: g=>G [Gc Ge Gz C]; apply: Ge. Qed.

Lemma transZ : localP coh g.
Proof. by case: g=>G [Gc Ge Gz C]; apply: Gz. Qed.

Lemma trans_coh : preserved coh g.
Proof. by case: g=>G [Gc Ge Gz C]; apply: C. Qed.

End Repack.


Lemma trans_zag U T coh (g : @trans U T coh) : zagP coh g.
Proof. by apply/zagzigE; [apply: trans_other | apply: transZ]. Qed.

End Exports.
End Trans.

Export Trans.Exports.


(********************************)
(* Constructors for transitions *)
(********************************)

(* empty transition *)

Module EmpTrans.
Section EmpTrans.
Variables (U : pcm) (T : Type) (coh : cohpred U T).
Local Notation emp_rel := (@emp_rel (state U T)).

Lemma emp_func : {In coh, functional emp_rel}.
Proof. apply: In1W; apply: emp_rel_func. Qed.

Lemma emp_other : otherP emp_rel.
Proof. by []. Qed.

Lemma emp_zagzig : localP coh emp_rel.
Proof. by []. Qed.

Lemma emp_coh : preserved coh emp_rel.
Proof. apply: emp_rel_pres. Qed.

End EmpTrans.

Module Exports.

Definition empTransMixin U T coh :=
  TransMixin (@emp_func U T coh) (@emp_other U T)
             (@emp_zagzig U T coh) (@emp_coh U T coh).
Canonical emp_trans U T coh := Eval hnf in Trans (@empTransMixin U T coh).

Arguments emp_trans {U  T coh}.

End Exports.
End EmpTrans.

Export EmpTrans.Exports.

(* Identity transition *)

Module IdTrans.
Section IdTrans.
Variables (U : pcm) (T : Type) (coh : cohpred U T).
Local Notation id_tr := (@id_rel (state U T)).

Lemma id_func : {In coh, functional id_tr}.
Proof. by apply: In1W; apply: id_rel_func. Qed.

Lemma id_other : otherP id_tr.
Proof. by move=>s s' ->. Qed.

Lemma id_zagzig : localP coh id_tr.
Proof. by move=>???? O; rewrite !zg_unlock; case=>->->; rewrite O. Qed.

Lemma id_coh : preserved coh id_tr.
Proof. by apply: id_rel_pres. Qed.

End IdTrans.

Module Exports.

Definition idTransMixin U T coh :=
  TransMixin (@id_func U T coh) (@id_other U T)
             (@id_zagzig U T coh) (@id_coh U T coh).
Canonical id_tr U T coh := Eval hnf in Trans (@idTransMixin U T coh).

Arguments id_tr [U][T] coh.

Lemma idF U T coh : footeqP coh (@id_tr U T coh).
Proof. by move=> ??? ->. Qed.

Arguments idF [U][T] coh _ _ _.

End Exports.
End IdTrans.

Export IdTrans.Exports.

(* We often add a pre-condition to a transition *)

Module PreTrans.
Section PreTrans.
Variables (U : pcm) (T : Type) (coh : cohpred U T) (g : trans coh).
Variables (P : state U T -> Prop) (pf : monoP coh P).

Lemma pre_func : {In coh, functional (pre_rel P g)}.
Proof. by move=>s C s1 s2 [_ H][_ /(trans_func C H)]. Qed.

Lemma pre_other : otherP (pre_rel P g).
Proof. by move=>s s' [_ /trans_other]. Qed.

Lemma pre_zagzig : localP coh (pre_rel P g).
Proof.  by move=>p s s' C O [/(pf C) H1 /(transZ C O)]. Qed.

Lemma pre_coh : preserved coh (pre_rel P g).
Proof. by move=>s s' [_ /trans_coh]. Qed.

Definition preTransMixin :=
  TransMixin pre_func pre_other pre_zagzig pre_coh.
Canonical pre_tr := Eval hnf in Trans preTransMixin.

End PreTrans.

Module Exports.
Notation pre_tr := pre_tr.
Canonical pre_tr.

Lemma condF U T coh (g : trans coh) P pf :
        footeqP coh g -> footeqP coh (@pre_tr U T coh g P pf).
Proof. by move=>F s s' C [_]; apply: F. Qed.

Arguments pre_tr [U T coh] g [P] pf.

End Exports.
End PreTrans.

Export PreTrans.Exports.


Definition usub_rel (U : pcm) (P : U -> Prop) (pc : pcm_closed P)
                    (X : Type) (Q : X -> Prop) (g : srel U X)
                    (s s' : substate pc Q) := g (ustval pc Q s) (ustval pc Q s').

Arguments usub_rel [U P] pc [X] Q g s s'.

Section SubUTrans.
Variables (U : pcm) (P : U -> Prop) (pc : pcm_closed P).
Variables (T : Type) (Q : T -> Prop) (coh : cohpred U T) (g : trans coh).

Lemma subtrans_func : {In (usub_coh pc Q coh), functional (usub_rel pc Q g)}.
Proof. by move=>? C ?? H /trans_func-/(_ C _ H)-/ust_inj. Qed.

Lemma subtrans_other : otherP (usub_rel pc Q g).
Proof. by move=>?? /trans_other; rewrite -!ustEQ0=>/sval_inj. Qed.

Lemma subtrans_zagzig : localP (usub_coh pc Q coh) (usub_rel pc Q g).
Proof.
move=>p s s'; rewrite /= /usub_rel !ustE=>C O H.
apply/transZ=>//.
- by rewrite /SubUCoh.sub_coh_raw /= ustE in C.
by rewrite -!ustEQ0 O.
Qed.

Lemma subtrans_coh : preserved (usub_coh pc Q coh) (usub_rel pc Q g).
Proof. by move=>s s';  apply: trans_coh. Qed.

Definition usubTransMix :=
  TransMixin subtrans_func subtrans_other subtrans_zagzig subtrans_coh.
Canonical usub_tr := Eval hnf in Trans usubTransMix.
End SubUTrans.



(* Taking a sub-PCM of the given PCM of the transition *)

Section Preservation.
Variables (V : pcm) (T : Type) (g : srel V T) (coh : cohpred V T).

(* g preserves P *)
Definition preserveP (P : state V T -> Prop) :=
  forall (s1 s2 : state V T), coh s1 -> g s1 s2 -> P s1 -> P s2.

(* persistence on a sub_pcm is a frequent abbreviation *)
Definition sub_preserveP (D : rel V) (U : sub_pcm D) P :=
  preserveP (fun s => D (idfun.s s) (idfun.o s) /\ P s).

End Preservation.

Section SubTrans.
Variables (V : pcm) (D : sep_rel V) (U : sub_pcm D).
Variables (T : Type) (coh : cohpred V T) (P : state V T -> Prop).
Variables (gb_pf : cglobalP coh P).
Variables (g : trans coh) (sp_pf : sub_preserveP g coh U P).

Let sp_rel (s s' : state U T) := g (spval s) (spval s').

Lemma xsub_func : {In (sub_coh U gb_pf), functional sp_rel}.
Proof. by move=>? [_ H C] ?? G /(trans_func C G) /sp_inj. Qed.

Lemma xsub_other : otherP sp_rel.
Proof. by move=>?? /trans_other; rewrite -!spEQ0=>/pval_inj. Qed.

(* tricky proof *)
Lemma xsub_zagzig : localP (sub_coh U gb_pf) sp_rel.
Proof.
move=>p s s' [A Hps Cps] O; rewrite /sp_rel=>G.
have W : valid (self s \+ p \+ other s) by rewrite allvalidE !zgE in A.
rewrite spE ?(validL W) // in Cps Hps *; rewrite spE ?(validAR W) // in G.
have Csp : coh (spval s \> pval p) by apply/cohZ.
have Hsp : P (spval s \> pval p) by apply/gb_pf.
case: (sp_pf Csp G)=>[|{Hsp} Xsp Hsp].
- move: (W); rewrite -joinA valid_sep=>/andP [_].
  by rewrite !lSXst !zgE pfjoin ?spE // ?(validAR W).
have W1 : valid (p \+ other s').
- move/coh_valid: (trans_coh G Csp); rewrite !allvalidE spunlock !zgE.
  by move/validR/valid_pvalE.
suff W2 : valid (self s' \+ p).
- by rewrite !spE // in G *; apply: transZ G=>//; rewrite -!spE O.
have Xsp' : D (pval p) (pval (other s')).
- by move: W1; rewrite valid_sep=>/andP [].
have {} W : valid (pval (self s') \+ (pval p \+ pval (other s'))).
- by move/coh_valid: (trans_coh G Csp); rewrite allvalidE spunlock !zgE pfjoin.
rewrite valid_sep (validAL W) /=.
move: Xsp; rewrite !lSXst !spE // !zgE -!spE.
by rewrite sepC // => X; rewrite (sepAxx _ Xsp' X) // joinC.
Qed.

Lemma xsub_coh : preserved (sub_coh U gb_pf) sp_rel.
Proof.
move=>s s' G [A H C]; case: (sp_pf C G)=>[|X H'].
- by rewrite !spVE in A; rewrite (andX A).
split=>//; last by apply: trans_coh G C.
by rewrite spVE (coh_valid (trans_coh G C)) X.
Qed.

Definition subTransMix :=
  TransMixin xsub_func xsub_other xsub_zagzig xsub_coh.
Canonical sub_tr := Eval hnf in Trans subTransMix.
End SubTrans.


Module ProdTrans.
Section ProdTrans.
Variables (U1 U2 : pcm) (T1 T2 : Type).
Variables (coh1 : cohpred U1 T1) (g1 : trans coh1).
Variables (coh2 : cohpred U2 T2) (g2 : trans coh2).

Definition prod_rel s s' :=
  [/\ g1 s\1 s'\1, g2 s\2 s'\2 & valid (flat coh1 s'\1 \+ flat coh2 s'\2)].

Lemma xprod_func : {In (prod_coh coh1 coh2), functional prod_rel}.
Proof.
move=>s [Cf Cs _] s1 s2 [F1 S1 _][F2 S2 _]; rewrite [s1]seta [s2]seta.
by rewrite (trans_func Cf F1 F2) (trans_func Cs S1 S2).
Qed.

Lemma xprod_other : otherP prod_rel.
Proof. by move=>s s' [/trans_other X /trans_other Y]; rewrite !prodX X Y. Qed.

Lemma xprod_zagzig : localP (prod_coh coh1 coh2) prod_rel.
Proof.
move=>p s s' []; rewrite /= /prod_raw /prod_rel !prodE !prodX.
move=>C1 C2 _ [O1 O2][G1 G2 V]; try by apply/transZ.
move: (transZ C1 O1 G1)=>G1'; move: (transZ C2 O2 G2)=>G2'.
move/cohZ: (C1)=>C1'; move/cohZ: (trans_coh G1 C1')=>X1.
move/cohZ: (C2)=>C2'; move/cohZ: (trans_coh G2 C2')=>X2.
by rewrite !flatZ.
Qed.

Lemma xprod_coh : preserved (prod_coh coh1 coh2) prod_rel.
Proof.
move=>s s' [G1 G2 V][/= C1 C2 _]; split=>//;
by [apply: trans_coh G1 C1 | apply: trans_coh G2 C2].
Qed.

Definition prodTransMixin :=
  TransMixin xprod_func xprod_other xprod_zagzig xprod_coh.
Canonical prod_tr := Eval hnf in Trans prodTransMixin.

End ProdTrans.

Module Exports.
Notation prod_tr := prod_tr.
Notation prod_rel := prod_rel.
Canonical prod_tr.

Notation "[ '|' r1 , r2 ]" := (prod_tr r1 r2)
  (at level 0, format "[ '|'  r1 ,  r2 ]").

End Exports.
End ProdTrans.

Export ProdTrans.Exports.


(* existential transition *)

Module ExTrans.
Section ExTrans.
Variables (U : pcm) (T A : Type) (coh : cohpred U T) (a : A -> trans coh).
Variables (pf1 : funcrP coh a).

Lemma ex_func : {In coh, functional (ex_rel [eta a])}.
Proof.
move=>s C s1 s2 [x X][y Y]; move: (pf1 C X Y)=>E.
by rewrite -{y}E in Y *; apply: trans_func X Y.
Qed.

Lemma ex_other : otherP (ex_rel [eta a]).
Proof. by move=>s s' [_ /trans_other]. Qed.

Lemma ex_zagzig : localP coh (ex_rel [eta a]).
Proof. by move=>????? [x]; exists x; apply: transZ. Qed.

Lemma ex_coh : preserved coh (ex_rel [eta a]).
Proof. by move=>s s' [x]; apply: trans_coh. Qed.

Definition exTransMixin :=
  TransMixin ex_func ex_other ex_zagzig ex_coh.
Canonical ex_tr := Eval hnf in Trans exTransMixin.

End ExTrans.

Module Exports.
Notation ex_tr := ex_tr.
Canonical ex_tr.
Arguments ex_tr [U T A] coh a pf1.

End Exports.
End ExTrans.

Export ExTrans.Exports.


(* Union of transitions *)
(* it is not a transition itself, but we only ever union transitions *)
(* and not arbitrary relations *)
(* since it's not a transition, we don't wrap in a module *)

Section UnionRels.
Variables (U : pcm) (T : Type) (coh : cohpred U T).

Inductive union_rel (Ts : Pred (trans coh)) s s' : Prop :=
  Union g of g \In Ts & g s s'.

(* union of tranitions is not a transition, since it has no funcrP *)

Lemma union_other Ts : otherP (union_rel Ts).
Proof. by move=>s s' [g _] /trans_other. Qed.

Lemma union_sub Ts Ts' s1 s2 :
        Ts <=p Ts' -> union_rel Ts s1 s2 -> union_rel Ts' s1 s2.
Proof. by move=>H [g /H]; exists g. Qed.

End UnionRels.


(*************)
(*************)
(* Resources *)
(*************)
(*************)

Section ResourceDefs.
Variables (U : pcm) (T Tr : Type) (coh : cohpred U T).
Variables (interp: Tr -> trans coh) (internal : Tr -> Prop).

Lemma union_coh P : (forall tr, preserved coh (interp tr)) ->
                    preserved coh (union_rel (Image interp P)).
Proof. by move=>S s s' [g][x -> _ /S]. Qed.

Definition res_internalP := forall tr, internal tr -> footeqP coh (interp tr).

End ResourceDefs.


Module Resource.
Section Resource.

Structure mixin_of (U : pcm) (T Tr : Type) := Mixin {
  coh_op : cohpred U T;
  interp_op : Tr -> trans coh_op;
  internal_op : Tr -> Prop;
  (* each internal transition preserves footprints *)
  _ : res_internalP interp_op internal_op}.

Section ClassDef.

Notation class_of := mixin_of (only parsing).

Structure resource :=
  Pack {pcm_of : pcm; tp_of : Type; tr_of : Type; _ : class_of pcm_of tp_of tr_of}.

Variables (U : pcm) (T Tr : Type) (cT : resource).

Definition class :=
  let: Pack _ _ _ c as cT' :=
    cT return class_of (pcm_of cT') (tp_of cT') (tr_of cT') in c.

Definition pack c := @Pack U T Tr c.
Definition clone :=
  fun c & U = pcm_of cT & T = tp_of cT
        & Tr = tr_of cT & phant_id (pack c) cT => pack c.

End ClassDef.
End Resource.

Module Exports.
Notation resource := resource.
Notation ResourceMixin := (@Mixin _ _).
Notation Resource m := (@pack _ _ _ m).
Notation Ux := pcm_of.
Notation Tp := tp_of.
Notation Tr := tr_of.

Definition coh W := coh_op (class W).
Definition interp W := interp_op (class W).
Definition internal W := internal_op (class W).
Arguments interp : clear implicits.
Arguments internal : clear implicits.

Section CombinedMixin.
Variables (U : pcm) (T Tr : Type)  (coh : cohpred U T).
Variables (interp : Tr -> trans coh) (intern : Tr -> Prop).

Definition resP :=
  forall t, intern t -> footeqP coh (interp t).

Lemma res_mix : resP -> res_internalP interp intern.
Proof. by []. Qed.

Definition ResMix pf :=
  @Mixin U T Tr coh interp intern (res_mix pf).
Definition ResPx pf := @Resource.pack U T Tr (ResMix pf).

End CombinedMixin.

Arguments ResPx [U T Tr] coh interp intern _.

Lemma iint W : res_internalP (interp W) (internal W).
Proof. by case: W=>U T Tr []. Qed.

Lemma icoh W tr : preserved (coh W) (interp W tr).
Proof. by move=>s s' /trans_coh. Qed.

Lemma itransZ W tr : localP (coh W) (interp W tr).
Proof. by move=>??? C O /transZ; apply. Qed.

Lemma itrans_zag W tr : zagP (coh W) (interp W tr).
Proof. by move=>??? C /trans_zag; apply. Qed.

End Exports.

End Resource.

Export Resource.Exports.


(* resource with a TPCM instead of a PCM *)

Module TResource.
Section TResource.

Structure mixin_of (U : tpcm) (T Tr : Type) := Mixin {
  coh_op : cohpred U T;
  interp_op : Tr -> trans coh_op;
  internal_op : Tr -> Prop;
  (* each internal transition preserves footprints *)
  _ : res_internalP interp_op internal_op}.

Section ClassDef.

Notation class_of := mixin_of (only parsing).

Structure tresource :=
  Pack {pcm_of : tpcm; tp_of : Type; tr_of : Type; _ : class_of pcm_of tp_of tr_of}.

Variables (U : tpcm) (T Tr : Type) (cT : tresource).

Definition class :=
  let: Pack _ _ _ c as cT' :=
    cT return class_of (pcm_of cT') (tp_of cT') (tr_of cT') in c.

Definition pack c := @Pack U T Tr c.
Definition clone :=
  fun c & U = pcm_of cT & T = tp_of cT
        & Tr = tr_of cT & phant_id (pack c) cT => pack c.

End ClassDef.
End TResource.

Module Exports.
Notation tresource := tresource.
Notation TResourceMixin := (@Mixin _ _).
Notation TResource m := (@pack _ _ _ m).
Notation Uxt := pcm_of.

(* need a coercion from TResource to plain Resource, so that I can reuse *)
(* the projections and lemmas from Resource *)

Definition to_rmixin U T Tr (x : mixin_of U T Tr) : Resource.mixin_of U T Tr :=
  let: Mixin coh interp internal resinter := x in
  @Resource.Mixin U T Tr coh interp internal resinter.

Definition to_resource (x : tresource) : resource :=
  @Resource.Pack (pcm_of x) (tp_of x) (tr_of x) (to_rmixin (class x)).

Coercion to_rmixin : mixin_of >-> Resource.mixin_of.
Coercion to_resource : tresource >-> resource.

Lemma tpcm_eq (R : tresource) : Uxt R = Ux R :> pcm.
Proof. by []. Qed.

End Exports.
End TResource.

Export TResource.Exports.



(* Notation for state type and union of all transitions of a resource *)

Notation st W := (state (Ux W) (Tp W)).

(* all_coh is obsolete; I now care about preservation for all *)
(* not just internal transitions *)
Definition all_cohP W P :=
  forall t s s', internal W t -> coh W s -> interp W t s s' -> P s -> P s'.

(* P is persistent if preserved by all transitions of W *)
Definition persP W P :=
  forall t s s', coh W s -> interp W t s s' -> P s -> P s'.

Arguments all_cohP W P : clear implicits.
Arguments persP W P : clear implicits.

(***********************************)
(***********************************)
(* stepping relation of a resource *)
(***********************************)
(***********************************)

(* We need two notions of stepping.

- ostep: stepping by other transitions only
- mstep: stepping by mixed transitions, where the set of self transitions is
obtained by morphism application from another resource.

We define ostep now, as it is needed in the definition of morphisms.
Then we define mstep later.

*)

(* stepping just by others *)

Section Ostep.
Variable W : resource.

Inductive ostep s1 s2 : Prop := iostep a of interp W a (tr s1) (tr s2).

Definition iosteps s1 s2 a H := iter1 (@iostep s1 s2 a H).

Lemma oself : selfP (iter ostep).
Proof. by apply/iter_self=>?? [] // a /trans_other; rewrite !trE0. Qed.

Lemma ozig : zigP (coh W) (iter ostep).
Proof.
apply/iter_zig=>[|p] s s'; first by case=>a /icoh; rewrite !coh_tr.
move=>C [a]; rewrite trE=>/trans_zag [].
- by rewrite -trE1 coh_tr -cohZ.
move=>s'' [E H]; exists (tr s''); rewrite -{1}[s']trI E trE.
by split=>//; apply: (iostep (a:=a)); rewrite !trE.
Qed.

Lemma ozigzag : zigzagP (coh W) (iter ostep).
Proof. by apply/zigzagE; [apply/oself | apply/ozig]. Qed.

Lemma ocoh1 : preserved (coh W) ostep.
Proof. move=>s1 s2 [a] /trr_coh; apply; apply: icoh. Qed.

Lemma ocoh : preserved (coh W) (iter ostep).
Proof. by apply: iter_pres; apply: ocoh1. Qed.


End Ostep.

Arguments ostep W s1 s2 : clear implicits.
Arguments iostep [W s1 s2] a _.
Arguments iosteps [W s1 s2] a H.
Notation osteps W s1 s2 := (iter (ostep W) s1 s2).


(**********)
(**********)
(* Action *)
(**********)
(**********)

Definition actionP W A (a : st W -> Tr W -> A -> Prop) :=
  [/\ forall s t1 r1 t2 r2, coh W s ->
        a s t1 r1 -> a s t2 r2 -> t1 = t2 /\ r1 = r2,
      forall s t r, coh W s -> a s t r -> internal W t &
      forall s p t v, coh W (s \< p) ->
        a (s \> p) t v -> a (s \< p) t v].

Module Action.
Section ClassDef.
Variables (W : resource) (A : Type).

Record mixin_of (sort : st W -> Tr W -> A -> Prop) := Mixin {
  _ : actionP sort}.

Notation class_of := mixin_of (only parsing).

Structure action := Pack {sort : st W -> Tr W -> A -> Prop; _ : class_of sort}.
Local Coercion sort : action >-> Funclass.

Variables (sort : st W -> Tr W -> A -> Prop) (cT : action).
Definition class := let: Pack _ c as cT' := cT return class_of cT' in c.
Definition clone c of phant_id class c := @Pack sort c.
Definition pack c := @Pack sort c.

End ClassDef.

Module Exports.
Coercion sort : action >-> Funclass.
Notation action := action.
Notation ActionMixin := @Mixin.
Notation Action m := (@pack _ _ _ m).

Notation "[ 'action' 'of' rel 'for' C ]" :=
  (@clone _ _ rel C _ id)
  (at level 0, format "[ 'action'  'of'  rel  'for'  C ]") : form_scope.
Notation "[ 'action' 'of' rel ]" := (@clone _ _ rel _ id)
  (at level 0, format "[ 'action'  'of'  rel ]") : form_scope.

Section Repack.
Variables (W : resource) (A : Type) (a : action W A).

Lemma act_func s t1 r1 t2 r2 :
        coh W s -> a s t1 r1 -> a s t2 r2 -> (t1 = t2) * (r1 = r2).
Proof. by case: a=>tr [[f i z]] /= C H1 H2; case: (f _ _ _ _ _ C H1 H2). Qed.

Lemma act_int s t r : coh W s -> a s t r -> internal W t.
Proof. by case: a=>tr [[f i z]]; apply: i. Qed.

Lemma actZ s p t v :
        coh W (s \< p) -> a (s \> p) t v -> a (s \< p) t v.
Proof. by case: a=>tr [[f i z]]; apply: z. Qed.

(* the effect of an action is the following derived relation on states *)
Definition eff v (s s' : st W) :=
  exists t, [/\ a s t v & interp W t s s'].

Lemma eff_func (s s1 s2 : st W) (v1 v2 : A) :
        coh W s -> eff v1 s s1 -> eff v2 s s2 -> (s1 = s2) * (v1 = v2).
Proof.
move=>C [t1][Ha1 H1][t2][Ha2 H2]; case: (act_func C Ha1 Ha2)=>??; subst t2 v2.
by rewrite (trans_func C H1 H2).
Qed.

Lemma eff_coh s s' v : eff v s s' -> coh W s -> coh W s'.
Proof. by case=>t [Ha]; apply: icoh. Qed.

End Repack.

(* custom ways to build actions *)

Definition actF W A g (pf : actionP g) :=
  Action (ActionMixin W A g pf).

Notation actD g pf :=
  (Action (ActionMixin _ _ g pf)).

(* often times, actions are written in a "backwards" way *)
(* i.e., the transition is directly determined by the output value *)
(* we want some "smart constructors" for that *)

Definition smart_actP W A (g : A -> Tr W) :=
  funcrP (coh W) (fun r => interp W (g r)) /\ forall r, internal W (g r).

Definition smart_act_raw W A (g : A -> Tr W) (s : st W) t v :=
  exists s', t = g v /\ interp W t s s'.

Section SmartActions.
Variables (W : resource) (A : Type) (g : A -> Tr W) (pf : smart_actP g).

Lemma smartP : actionP (smart_act_raw g).
Proof.
case: (pf)=>H1 H2; split=>[s' t1 r1 t2 r2 C [s][-> H]|????|? p ?? C][x][->] //.
- by move/(H1 _ _ _ _ _ C H)=>->.
by case/(trans_zag C)=>s' [_]; exists (s' \< p).
Qed.

Definition smart_act := Action (ActionMixin W A (smart_act_raw g) smartP).

Lemma effE v : eff smart_act v = interp W (g v).
Proof.
apply/fext=>s; apply/fext=>s'; apply/pext.
by split=>[[t][[x][->]] //|H]; exists (g v); split=>//; exists s'.
Qed.

End SmartActions.

Definition actS W A (g : A -> Tr W) (pf : smart_actP g) :=
  Action (ActionMixin _ _ (smart_act pf) (smartP pf)).

End Exports.
End Action.

Export Action.Exports.

(* Some properties to relate to the legacy development. They can all
be proved from the three properties of transitions. They are used in
theory of process trees *)

Section SafetyFraming.
Variables (W : resource) (A : Type) (a : action W A).

Definition act_safe := [Pred s : st W | exists s' v, eff a v s s'].

Lemma act_safe_mono : monoP (coh W) act_safe.
Proof.
move=>p s C [s'][v][t][/(actZ C) H] /(itrans_zag C) [s2][].
by exists (s2 \< p), v, t.
Qed.

Lemma act_frame :
        frameP (coh W) act_safe (fun s s' => exists v, eff a v s s').
Proof.
move=>p s s' C [s2][v1][t1][H1 T1][v2][t2][H2 T2].
move: (actZ C H1) (itrans_zag C T1)=>H1' [s''][E1' T1']; subst s2.
case: (act_func C H1' H2)=>??; subst t2 v2.
move: (trans_func C T1' T2)=>?; subst s'.
by exists s''; split=>//; exists v1, t1.
Qed.

End SafetyFraming.

Section HelperLemmas.
Variables (U1 U2 : pcm) (T1 T2 : Type).
Variables (coh1 : cohpred U1 T1) (t1 : trans coh1).
Variables (coh2 : cohpred U2 T2) (t2 : trans coh2).

Lemma act2_safe s1 s2 x1 x2 :
         s2 = [@ x1, x2] ->
         t1 (s1\1) x1 -> t2 (s1\2) x2 ->
         valid (flat coh1 x1 \+ flat coh2 x2) ->
         [| t1, t2] s1 s2.
Proof. by move=>-> H1 H2 V; split; rewrite !prodE. Qed.

Lemma act2 s1 s2 x1 x2 :
         s2 = [@ x1, x2] ->
         (valid (flat coh1 x1) -> t1 (s1\1) x1) ->
         (valid (flat coh2 x2) -> t2 (s1\2) x2) ->
         valid (flat coh1 x1 \+ flat coh2 x2) ->
         [| t1, t2] s1 s2.
Proof.
by move=>-> H1 H2 V; split; rewrite ?prodE !(validE2 V) // in H1 H2 *;
   [apply: H1 | apply: H2].
Qed.

Variables (U3 : pcm) (T3 : Type) (coh3 : cohpred U3 T3) (t3 : trans coh3).

Lemma act3_safe s1 s2 x1 x2 x3 :
         s2 = [@ x1, [@ x2, x3]] ->
         t1 (s1\1) x1 -> t2 (s1\2\1) x2 -> t3 (s1\2\2) x3 ->
         valid (flat coh1 x1 \+ (flat coh2 x2 \+ flat coh3 x3)) ->
         [| t1, [| t2, t3]] s1 s2.
Proof.
by move=>-> H1 H2 H3 V; do !split; rewrite !prodE ?(validRE3 V) // flatPE.
Qed.

End HelperLemmas.

(**********************)
(**********************)
(* Resource morphisms *)
(**********************)
(**********************)

Section MorphDefs.
Variables (X : Type) (V W : resource).
Variables (phi : X -> st W -> st V -> Prop) (tau : X -> Tr V -> st W -> Tr W -> Prop).

(* properties that we want to prove *)

(* note : other steps don't change the seed *)
Definition morph_localP :=
  exists f, forall x p sw sv, coh W (sw \< p) ->
    coh V sv -> phi x (sw \> p) sv ->
    exists sv', sv = sv' \> f p /\ phi x (sw \< p) (sv' \< f p).

Definition morph_otherP :=
  forall x x' sw sw' sv sv',
    coh W sw -> coh W sw' -> coh V sv -> coh V sv' ->
    phi x sw sv -> phi x' sw' sv' ->
    other sw = other sw' -> other sv = other sv'.

Definition morph_funcP :=
  forall x sw sv1 sv2, coh W sw -> coh V sv1 -> coh V sv2 ->
    phi x sw sv1 -> phi x sw sv2 -> sv1 = sv2.

Definition morph_backwardP :=
  forall x x' sw sw' sv, coh W sw -> coh W sw' -> coh V sv ->
    phi x sw sv -> phi x' sw' sv -> x = x'.

Definition morph_internalP :=
  forall t t' x s, internal V t -> coh W s -> tau x t s t' -> internal W t'.

Definition morph_tfuncP :=
  forall x t sw t1 t2, coh W sw -> tau x t sw t1 -> tau x t sw t2 -> t1 = t2.

Definition morph_tlocalP :=
  forall x t sw p t', coh W (sw \< p) ->
     tau x t (sw \> p) t' -> tau x t (sw \< p) t'.

(* properties stored in the structure; packing some coherence conditions *)

Definition morph_localP' :=
  exists f, forall x p sw sv, coh W (sw \< p) ->
    phi x (sw \> p) sv ->
    exists sv', sv = sv' \> f p /\ phi x (sw \< p) (sv' \< f p).

Definition morph_otherP' :=
  forall x x' sw sw' sv sv',
    coh W sw -> coh W sw' ->
    phi x sw sv -> phi x' sw' sv' ->
    other sw = other sw' -> other sv = other sv'.

Definition morph_funcP' :=
  forall x sw sv1 sv2, coh W sw -> phi x sw sv1 -> phi x sw sv2 -> sv1 = sv2.

Definition morph_backwardP' :=
  forall x x' sw sw' sv, coh W sw -> coh W sw' -> phi x sw sv -> phi x' sw' sv -> x = x'.

Definition morph_cohP' := forall x sw sv, coh W sw -> phi x sw sv -> coh V sv.

End MorphDefs.


Module Morph.
Section Morph.
Variables (X : Type) (V W : resource).

Structure mixin_of (tau : X -> Tr V -> st W -> Tr W -> Prop) := Mixin {
  phi_op : X -> st W -> st V -> Prop;
  _ : morph_localP' phi_op;
  _ : morph_otherP' phi_op;
  _ : morph_funcP' phi_op;
  _ : morph_backwardP' phi_op;
  _ : morph_internalP tau;
  _ : morph_tfuncP tau;
  _ : morph_tlocalP tau;
  _ : morph_cohP' phi_op}.

Section ClassDef.

Notation class_of := mixin_of (only parsing).

Structure morph : Type := Pack {tau_of : _; _ : class_of tau_of}.
Local Coercion tau_of : morph >-> Funclass.

Variables (tau : X -> Tr V -> st W -> Tr W -> Prop) (cT : morph).
Definition class := let: Pack _ c as cT' := cT return class_of cT' in c.

Definition pack c := @Pack tau c.
Definition clone := fun c & tau = cT & phant_id (pack c) cT => pack c.

Definition phi := phi_op class.

End ClassDef.
End Morph.

Module Exports.
Coercion tau_of : morph >-> Funclass.
Notation morph := morph.
Notation Morph m := (@pack _ _ _ _ m).

Notation "[ 'morph' f ]" := (Morph.phi f)
  (at level 0, format "[ 'morph'  f ]").


Section Repack.
Variables (V W : resource) (X : Type) (tau : morph X V W).

Lemma morph_coh : morph_cohP' [morph tau].
Proof. by case: tau=>tau_of []. Qed.

Lemma morph_local : morph_localP' [morph tau].
Proof. by case: tau=>f []. Qed.

Lemma morph_other : morph_otherP' [morph tau].
Proof. by case: tau=>f []. Qed.

Lemma morph_func : morph_funcP' [morph tau].
Proof. by case: tau=>f []. Qed.

Lemma morph_backw : morph_backwardP' [morph tau].
Proof. by case: tau=>f []. Qed.

Lemma morph_int : morph_internalP tau.
Proof. by case: tau=>tau_of []. Qed.

Lemma morph_tfunc : morph_tfuncP tau.
Proof. by case: tau=>tau_of []. Qed.

Lemma morph_tlocal : morph_tlocalP tau.
Proof. by case: tau=>tau_of []. Qed.

End Repack.

Prenex Implicits morph_coh morph_func morph_int morph_tfunc morph_tlocal morph_backw.


(* repackaging, to give us some more coherence hypotheses *)
(* when proving concrete morphisms *)

Section Morph_from_Morph2.
Variables (X : Type) (V W : resource) (phi : X -> st W -> st V -> Prop).
Variables (tau : X -> Tr V -> st W -> Tr W -> Prop).

Lemma morph_local2 : morph_cohP' phi -> morph_localP phi -> morph_localP' phi.
Proof.
move=>C [f H]; exists f=>x p sw sv Cw T.
by apply: H (Cw) _ (T); apply: C T; apply/cohZ.
Qed.

Lemma morph_other2 : morph_cohP' phi -> morph_otherP phi -> morph_otherP' phi.
Proof.
move=>C O x x' sw sw' sv sv' Cw Cw' T T'.
by apply: O (C _ _ _ Cw T) (C _ _ _ Cw' T') T T'.
Qed.

Lemma morph_func2 : morph_cohP' phi -> morph_funcP phi -> morph_funcP' phi.
Proof.
move=>C F x sw sv1 sv2 Cw T1 T2.
by apply: F (C _ _ _ Cw T1) (C _ _ _ Cw T2) T1 T2.
Qed.

Lemma morph_backw2 : morph_cohP' phi -> morph_backwardP phi -> morph_backwardP' phi.
Proof.
move=>C B x x' sw sw' sv Cw Cw' T T'.
by apply: B Cw Cw' (C _ _ _ Cw' T') T T'.
Qed.

Definition MorphMixin lc ot fn bw (int : morph_internalP tau) tf tl ch :=
  Mixin (morph_local2 ch lc) (morph_other2 ch ot)
        (morph_func2 ch fn) (morph_backw2 ch bw) int tf tl ch.

End Morph_from_Morph2.
End Exports.
End Morph.

Export Morph.Exports.


(* An f-simulation is a predicate on W's state *)

Section MorphSimDefs.
Variables (X : Type) (V W : resource).
Variables (f : morph X V W) (sim : X -> st W -> Prop).

Definition morph_ostepP :=
  forall x sw sw' sv, coh W sw -> coh V sv ->
    [morph f] x sw sv -> sim x sw -> ostep W sw sw' ->
    exists sv', [/\ [morph f] x sw' sv', sim x sw' & osteps V sv sv'].

Definition morph_ostepP' :=
  forall x sw sw' sv, coh W sw ->
    [morph f] x sw sv -> sim x sw -> ostep W sw sw' ->
    exists sv', [/\ [morph f] x sw' sv', sim x sw' & osteps V sv sv'].

Definition morph_forwardP :=
  forall t x sw sv sv', internal V t ->
    coh W sw -> coh V sv -> [morph f] x sw sv -> sim x sw -> interp V t sv sv' ->
  exists t' x' sw', [/\ f x t sw t', [morph f] x' sw' sv',
                        sim x' sw' & interp W t' sw sw'].

Definition morph_forwardP' :=
  forall t x sw sv sv', internal V t ->
    coh W sw -> [morph f] x sw sv -> sim x sw -> interp V t sv sv' ->
  exists t' x' sw', [/\ f x t sw t', [morph f] x' sw' sv',
                        sim x' sw' & interp W t' sw sw'].

Definition morph_simP := morph_ostepP /\ morph_forwardP.

End MorphSimDefs.


Module Simulation.
Section Simulation.
Variables (X : Type) (V W : resource) (f : morph X V W).

Structure mixin_of (sim : X -> st W -> Prop) := Mixin {
  _ : morph_ostepP' f sim;
  _ : morph_forwardP' f sim}.

Section ClassDef.

Notation class_of := mixin_of (only parsing).

Structure simulation : Type := Pack {pred_of : _; _ : class_of pred_of}.
Local Coercion pred_of : simulation >-> Funclass.

Variables (sim : X -> st W -> Prop) (cT : simulation).
Definition class := let: Pack _ c as cT' := cT return class_of cT' in c.

Definition pack c := @Pack sim c.
Definition clone := fun c & sim = cT & phant_id (pack c) cT => pack c.

End ClassDef.
End Simulation.

Module Exports.
Coercion pred_of : simulation >-> Funclass.
Notation simulation := simulation.

Section Repack.
Variables (X : Type) (V W : resource) (f : morph X V W) (sim : simulation f).

Lemma morph_ostep : morph_ostepP' f sim.
Proof. by case: sim=>s []. Qed.

Lemma morph_osteps x sw sw' sv :
        coh W sw -> [morph f] x sw sv -> sim x sw -> osteps W sw sw' ->
        exists sv', [/\ [morph f] x sw' sv', sim x sw' & osteps V sv sv'].
Proof.
move=>Cw Z T [n]; elim: n x sw sv Cw Z T=>[|n IH] x sw sv Cw Z T /=.
- by move=><-; exists sv.
case=>sw1 [O]; case: (morph_ostep Cw Z T O)=>sv1 [Z1 T1 O1].
case/(IH _ _ _ (ocoh (iter1 O) Cw) Z1 T1)=>sv' [].
by move=>T' Z' /(iter_trans O1); exists sv'.
Qed.

(* often used consequence of morph_osteps *)
Lemma morph_osteps_sim x sw sw' sv :
        coh W sw -> [morph f] x sw sv -> sim x sw -> osteps W sw sw' -> sim x sw'.
Proof. by move=>Cw Z T /(morph_osteps Cw Z T) [sv'][]. Qed.

Lemma morph_forw : morph_forwardP' f sim.
Proof. by case: sim=>s []. Qed.

End Repack.

Section SimFromSim2.
Variables (X : Type) (V W : resource).
Variables (f : morph X V W) (sim : X -> st W -> Prop).

Lemma morph_ostep1 : morph_ostepP f sim -> morph_ostepP' f sim.
Proof. by move=>H x sw sw' sv Cw T; apply: H (morph_coh Cw T) T. Qed.

Lemma morph_ostep2 : morph_simP f sim -> morph_ostepP' f sim.
Proof. by case=>H _; apply: morph_ostep1. Qed.

Lemma morph_forw1 : morph_forwardP f sim -> morph_forwardP' f sim.
Proof. by move=>F t x sw sv sv' I Cw T; apply: F (morph_coh Cw T) T. Qed.

Lemma morph_forw2 : morph_simP f sim -> morph_forwardP' f sim.
Proof. by case=>_ F; apply: morph_forw1. Qed.

Definition SimMixin H := Mixin (morph_ostep2 H) (morph_forw2 H).
Definition SimMixin2 mo mf := Mixin (morph_ostep1 mo) (morph_forw1 mf).

End SimFromSim2.

Notation Simulation H := (@pack _ _ _ _ _ (SimMixin H)).
Notation Simulation2 H1 H2 := (@pack _ _ _ _ _ (SimMixin2 H1 H2)).

Arguments morph_osteps_sim [X V W f] sim x [sw sw' sv].

End Exports.
End Simulation.

Export Simulation.Exports.


(* Stepping by a mix of other and morphed transitions *)

Section MxStep.
Variables (X : Type) (V W : resource) (f : morph X V W) (sim : simulation f).

Inductive mxstep x1 sw1 x2 sw2 : Prop :=
| SelfMxStep t of internal V t &
    exists tw sv1 sv2, [/\ f x1 t sw1 tw, [morph f] x1 sw1 sv1, [morph f] x2 sw2 sv2,
                        interp V t sv1 sv2 & interp W tw sw1 sw2]
| OtherMxStep t of x1 = x2 & interp W t (tr sw1) (tr sw2).

Lemma omx x s1 s2 : ostep W s1 s2 -> mxstep x s1 x s2.
Proof. by case=>a; apply: OtherMxStep. Qed.

Lemma omxs x s1 s2 : osteps W s1 s2 -> iterc mxstep x s1 x s2.
Proof.
case=>n; elim: n s1=>[|n IH] /= s1; first by move=>->; apply: iter_refl.
by case=>s [/(omx x)] M /IH; apply: iter_trans; apply: iter1 M.
Qed.

Lemma omcoh x1 x2 s1 s2 : iterc mxstep x1 s1 x2 s2 -> coh W s1 -> coh W s2.
Proof.
case=>n; elim: n x1 s1=>[|n IH] /= x1 s1; first by case=>_ ->.
case; case=>x s /= [M] /IH C; case: M=>t _.
- by case=>t' [s1'][s2'][_ _ _ _ H] /(icoh H).
by move/icoh=>H /coh_tr /H /coh_tr.
Qed.

Lemma omsim x x' sw sw' sv :
        iterc mxstep x sw x' sw' ->
        coh W sw -> [morph f] x sw sv -> sim x sw ->
        exists sv', [/\ coh V sv', [morph f] x' sw' sv' & sim x' sw'].
Proof.
case=>n; elim: n x sw sv=>[|n IH] /= x sw sv.
- by case=>->-> C T S; exists sv; split=>//; apply: morph_coh C T.
case; case=>x1 sw1 /= [M H C T S].
suff [sv1 []]: exists sv1, [/\ coh W sw1, [morph f] x1 sw1 sv1 & sim x1 sw1].
- by apply: IH H.
case: {H} M=>t; last first.
- move=><-{x1} /iostep O; case/(morph_ostep C T S):(O)=>sv' [M' S' O'].
  by exists sv'; split=>//; apply: ocoh (iter1 O) C.
move=>I [t'][xx][sv1][E1 /(morph_func C T) E1'] T1 Hv Hw; subst xx.
move: (icoh Hw C)=>C1.
case/(morph_forw I C T S): Hv=>t2 [x2][xx][E2 T2 S2].
move: (morph_tfunc C E1 E2)=>?; subst t'; move/(trans_func C Hw)=>?; subst xx.
by move/(morph_backw C1 C1 T1): T2=>?; subst x2; exists sv1.
Qed.

End MxStep.

Arguments mxstep [X V W] f x1 sw1 x2 sw2.
Notation mxsteps f x1 s1 x2 s2 := (iterc (mxstep f) x1 s1 x2 s2).


(* stability along mxsteps *)
Lemma omxind_s V W (f : morph unit V W) (C :  _ -> Prop) (F : _ -> _ -> Prop) :
        (forall s, C s -> F s s) -> Transitive F ->
        (forall s1 s2, mxsteps f tt s1 tt s2 -> C s1 -> C s2) ->
        (forall s1 s2, mxstep f tt s1 tt s2 -> C s1 -> F s1 s2) ->
         forall s1 s2, mxsteps f tt s1 tt s2 -> C s1 -> F s1 s2.
Proof.
move=>H1 H2 H3 H4 s1 s2.
apply: (@itercs_ind _ _ (mxstep f) (fun a x => C x) (fun a x b y => F x y))=>//.
- by case; case=>s1' [_ s2'][_ s3']; apply: H2.
- by case=>x [y]; apply: H3.
by case=>x [y]; apply: H4.
Qed.


(* I've noticed that I often don't need the heavy definition of mxstep *)
(* and a simpler one suffices, when we don't have seeds *)
(* Not sure, however, that this is always true *)
(* In cases of non-trivial simulations, probably isn't *)

Section Mstep.
Variables (V W : resource) (f : Tr V -> st W -> Tr W -> Prop).

Inductive mstep (s1 s2 : st W) : Prop :=
| SelfMStep a ft of internal V a & f a s1 ft & interp W ft s1 s2
| OtherMStep a of interp W a (tr s1) (tr s2).

Lemma om s s' : ostep W s s' -> mstep s s'.
Proof. by case=>a; apply: OtherMStep. Qed.

Lemma oms s s' : osteps W s s' -> iter mstep s s'.
Proof. by apply /iter_sub/om. Qed.

Lemma mcoh : preserved (coh W) (iter mstep).
Proof.
by apply: iter_pres=>s s' [a ft I H /icoh|a /trans_coh H /coh_tr/H/coh_tr].
Qed.

End Mstep.

Arguments mstep [V W] f s1 s2.
Notation msteps f s1 s2 := (iter (mstep f) s1 s2).

Lemma mxm V W (f : morph unit V W) s1 s2 :
        mxstep f tt s1 tt s2 -> mstep (f tt) s1 s2.
Proof.
case=>[t I [tw][x1][x2][X _ _ _]|t _];
by [apply: SelfMStep I X | apply: OtherMStep].
Qed.

Lemma mxms V W (f : morph unit V W) s1 s2 :
       mxsteps f tt s1 tt s2 -> msteps (f tt) s1 s2.
Proof.
case=>n; elim: n s1 s2=>[|n IH] /= s1 s2 []; first by move=>->.
by case; case=>x [/=] M /IH; apply: iter_trans (iter1 (mxm M)).
Qed.

Arguments mxms [V W] f [s1 s2] _.

(* doesn't seem we have this lemma *)
(* thus, we must use mxstep, not mstep *)
Lemma msim V W (f : morph unit V W)  (sim : simulation f) sw sw' sv :
        msteps (f tt) sw sw' ->
        coh W sw -> [morph f] tt sw sv -> sim tt sw ->
        exists sv', [/\ coh V sv', [morph f] tt sw' sv' & sim tt sw'].
Proof.
case=>n; elim: n sw sw' sv=>[|n IH] sw sw' sv /=.
- by move=><- C T S; exists sv; split=>//; apply: morph_coh C T.
case=>sw1 /= [M H C T S].
suff [sv1 []]: exists sv1, [/\ coh W sw1, [morph f] tt sw1 sv1 & sim tt sw1].
- by apply: IH H.
case: {H} M=>t; last first.
- move/iostep=>O; case/(morph_ostep C T S):(O)=>sv' [M' S' O'].
  by exists sv'; split=>//; apply: ocoh (iter1 O) C.
move=>ft I Ev Hv.
move: morph_forw. rewrite /morph_forwardP'.
move/(_ _ _ _ _ _ _ _ _ _ _ _ _ _ _ Hv).
move: (@morph_int). rewrite /morph_internalP. move/(_ _ _ _ _ _ _ _ _ I).
move/(_ _ _ f). move/(_ _ tt). move/(_ _ _ _ Ev). move/(_ C).
move=>I'. move/(_ _ _ _ _ _ _ I').
(* and now we can't proceed *)
Abort.

(************************************************)
(* Categorical properties of seedless morphisms *)
(************************************************)

(* This section is not necessary in the development at all. However, I
formalize it anyway, because it's satisfying that morphisms do behave
categorically as they should, even if the behavior is categorically
propper only when we have no seeds.

In the presence of seeds, the morphisms are actually parametrized
families, and their composition has to accumulate the parameters by
taking products thereoff. Correspondinly, morphism composition is not
associative, because products are not associative.

A bit more strongly -- we can construct a natural transformation
between a composition and its reassociation, but not an equality. *)

Lemma morphE X V W (tau1 tau2 : morph X V W) :
       tau1 = tau2 <->
       [/\ [morph tau1] = [morph tau2] &
           tau1 = tau2 :> (X -> Tr V -> st W -> Tr W -> Prop)].
Proof.
case: tau1 tau2=>f1 m1 [f2 m2] /=; split.
- by case=>?; subst f2; move/inj_pair2=>->.
case: m1 m2=>p1 ???????? [p2 ????????][E ?].
have ? : p1 = p2 by apply: E.
by subst f2 p2; congr Morph.Pack; congr Morph.Mixin; apply: pf_irr.
Qed.

Lemma simE X V W (f : morph X V W) (S1 S2 : simulation f) :
        S1 = S2 <-> S1 = S2 :> (X -> st W -> Prop).
Proof.
case: S1 S2=>s1 m1 [s2 m2] /=; split; first by case.
case: m1 m2=>?? [??] ?; subst s2.
by congr Simulation.Pack; apply: pf_irr.
Qed.


Section MorphComp.
Variables (V1 V2 V3 : resource).
Variables (tau1 : morph unit V1 V2) (tau2 : morph unit V2 V3).

Definition phi_raw (_ : unit) (s3 : st V3) (s1 : st V1) :=
  exists s2, [morph tau2] tt s3 s2 /\ [morph tau1] tt s2 s1.

Definition comp_raw (_ : unit) (t1 : Tr V1) (s3 : st V3) (t3 : Tr V3) :=
  exists s2 t2, [/\ [morph tau2] tt s3 s2, tau1 tt t1 s2 t2 & tau2 tt t2 s3 t3].

Lemma comp_local : morph_localP phi_raw.
Proof.
case: (morph_local tau1) (morph_local tau2)=>f1 H1 [f2 H2].
exists (f1 \o f2); case=>p s3 x C3' _ [y][T2]; move/cohZ: (C3')=>C3.
case/(H2 _ _ _ _ C3'): T2 (T2)=>s2 [->{y}] T2' T2.
move: (morph_coh C3 T2) (morph_coh C3' T2')=>C2 C2'.
case/(H1 _ _ _ _ C2')=>s1 [->{x}].
by exists s1; do !split=>//; exists (s2 \< f2 p).
Qed.

Lemma comp_other : morph_otherP phi_raw.
Proof.
case; case=>s3 s3' s1 s1' C3 C3' _ _ [s2][T2 T1][s2'][T2' T1'] O.
move: (morph_coh C3 T2) (morph_coh C3' T2')=>C2 C2'.
by apply: morph_other C2 C2' T1 T1' (morph_other C3 C3' T2 T2' O).
Qed.

Lemma comp_func : morph_funcP phi_raw.
Proof.
case=>sw sv1 sv2 Cw _ _ [sv][T2 T1] [_][/(morph_func Cw T2) <-].
by apply: morph_func (morph_coh Cw T2) T1.
Qed.

Lemma comp_backw : morph_backwardP phi_raw.
Proof. by case; case. Qed.

Lemma comp_int : morph_internalP comp_raw.
Proof.
move=>t t' [s] I C [s2][t2][M T1 T2]; apply: morph_int T2=>//.
by apply: morph_int I (morph_coh C M) T1.
Qed.

Lemma comp_tfunc : morph_tfuncP comp_raw.
Proof.
case=>t s3 t3 t3' C3 [x2][x2'][M1 X1 X2] [y2][y2'][M2 Y1 Y2].
move: (morph_func C3 M1 M2)=>?; subst y2.
move: (morph_tfunc (morph_coh C3 M1) X1 Y1)=>?; subst y2'.
by apply: morph_tfunc X2 Y2.
Qed.

Lemma comp_tlocal : morph_tlocalP comp_raw.
Proof.
case=>t s3 p t' C3' [x2][t2][M T1 T2]; move/cohZ: (C3')=>C3.
move: (morph_coh C3 M)=>C2.
case: (morph_local tau2)=>f /(_ _ _ _ _ C3' M) [sv'][?].
subst x2=>{}M; move: (morph_coh C3' M)=>C2'.
move/(morph_tlocal C3'): T2=>T2.
move/(morph_tlocal C2'): T1=>T1.
by exists (sv' \< f p), t2.
Qed.

Lemma comp_coh : morph_cohP' phi_raw.
Proof. by case=>s3 s1 C3 [s2][/(morph_coh C3) C2] /(morph_coh C2). Qed.

Definition compMorphMixin :=
  MorphMixin comp_local comp_other comp_func
             comp_backw comp_int comp_tfunc comp_tlocal comp_coh.
Definition comp_morph := Morph compMorphMixin.

Variables (sim1 : simulation tau1) (sim2 : simulation tau2).

Definition sim_raw (_ : unit) (s3 : st V3) :=
  sim2 tt s3 /\ forall s2, [morph tau2] tt s3 s2 -> sim1 tt s2.

Lemma comp_sim_pf : morph_simP comp_morph sim_raw.
Proof.
split.
- case=>s3 s3' s1 C3 _ [s2][T3 T2][S3] /(_ _ T3) S2 O.
  move: (ocoh (iter1 O) C3)=>C3'.
  case/(morph_ostep C3 T3 S3): O=>s2' [T3' S3'].
  case/(morph_osteps (morph_coh C3 T3) T2 S2)=>s1' [T2' S2'].
  exists s1'; do !split=>//; first by exists s2'.
  by move=>x /(morph_func C3' T3') <-.
move=>t1 [] s3 s1 s1' I C3 _ [s2][T2 T1][S3 /(_ _ T2) S2].
case/(morph_forw I (morph_coh C3 T2) T1 S2)=>t2 [[[s2']]][H2' T2' S2'].
case/(morph_forw (morph_int I (morph_coh C3 T2) H2') C3 T2 S3).
move=>t3 [[[s3']]][H3' T3' S3'] H.
exists t3, tt, s3'; split=>//; first by exists s2, t2.
- by exists s2'.
by split=>// x /(morph_func _ T3') <- //; apply: icoh H C3.
Qed.

Canonical comp_sim := Simulation comp_sim_pf.

End MorphComp.


Section MorphId.
Variable W : resource.

Definition id_phi (_ : unit) (s s' : st W) := s' = s.
Definition id_raw (_ : unit) (t : Tr W) (s : st W)  (t' : Tr W) := t' = t.

Lemma id_local : morph_localP id_phi.
Proof. by exists id; case=>p sw sv C _ ->; exists sw. Qed.

Lemma id_other : morph_otherP id_phi.
Proof. by case; case=>?? ?? C C' _ _ ->->. Qed.

Lemma id_func : morph_funcP id_phi.
Proof. by case=>sw ?? C ?? ->->. Qed.

Lemma id_backw : morph_backwardP id_phi.
Proof. by case; case. Qed.

Lemma id_int : morph_internalP id_raw.
Proof. by move=>t t' [s I] C ->. Qed.

Lemma id_tfunc : morph_tfuncP id_raw.
Proof. by case=>t sw ?? C ->->. Qed.

Lemma id_tlocal : morph_tlocalP id_raw.
Proof. by case=>t sw p t' C ->. Qed.

Lemma id_coh : morph_cohP' id_phi.
Proof. by case=>s s' C ->. Qed.

Definition idMorphMixin :=
  MorphMixin id_local id_other id_func id_backw id_int id_tfunc id_tlocal id_coh.
Definition id_morph := Morph idMorphMixin.

Definition idsim_raw (_ : unit) (s : st W) := True.

Lemma id_ostep : morph_ostepP id_morph idsim_raw.
Proof.
case=>sw sw' sv C _ -> _ /iter1; exists sw'.
by split=>//; apply: ocoh C.
Qed.

Lemma id_forw : morph_forwardP id_morph idsim_raw.
Proof.
move=>t [sw] sv sv' I C _ -> _ H; exists t, tt, sv'.
by split=>//; apply: trans_coh H C.
Qed.

Canonical id_sim := Simulation2 id_ostep id_forw.

End MorphId.


(* Morphism equality and categorical properties *)
(* We are only interested in equality up to coherence *)
(* and up to domain *)
(* a domain of a morphism is the set of coherent states *)
(* on which it's state image is defined *)
(* we say that two morphisms are equal if they *)
(* have the same domains, and behave equally on them *)

Definition morph_eq (V W : resource) (f1 f2 : morph unit V W) :=
  (forall sw sw', coh W sw ->
      [morph f1] tt sw sw' <-> [morph f2] tt sw sw') /\
  (forall s t t', coh W s -> (exists s', [morph f1] tt s s') ->
                             (exists s', [morph f2] tt s s') ->
      f1 tt t s t' <-> f2 tt t s t').

Notation "f1 =m f2" := (@morph_eq _ _ f1 f2) (at level 40).

Lemma meq_refl V W (f : morph unit V W) : f =m f.
Proof. by []. Qed.

Lemma meq_sym V W (f1 f2 : morph unit V W) : f1 =m f2 -> f2 =m f1.
Proof.
case=>H1 H2; split; first by move=>sw sw' C; rewrite H1.
by move=>s t t' C D1 D2; rewrite H2.
Qed.

Lemma meq_trans V W (f1 f2 f3 : morph unit V W) :
        f1 =m f2 -> f2 =m f3 -> f1 =m f3.
Proof.
case=>H1 T1 [H2 T2]. split.
- by move=>sw sw' C; rewrite H1 // H2.
move=>s t t' C D1 D3.
have D2 : exists s', [morph f2] tt s s'.
- by case: D1=>s' /(H1 _ _ C); exists s'.
by rewrite T1 // T2.
Qed.

Lemma comp_morphA V1 V2 V3 W (f1 : morph unit V1 V2)
                 (f2 : morph unit V2 V3) (f3 : morph unit V3 W) :
        comp_morph f1 (comp_morph f2 f3) =m comp_morph (comp_morph f1 f2) f3.
Proof.
split.
- move=>sw s1 C3; split.
  - by case=>s2 [[s3]][T3 T2] T1; exists s3; split=>//; exists s2.
  by case=>s3 [T3][s2][T2 T1]; exists s2; split=>//; exists s3.
move=>s t t' C _ _; split.
- case=>s2 [t2][[s3]][T3 T2 E1][_][t3][/(morph_func C T3) <- E2 E3].
  by exists s3, t3; split=>//; exists s2, t2.
case=>s3 [t3][T3][s2][t2][T2 T1 E2 E3].
by exists s2, t2; split=>//; exists s3=>//; exists t3.
Qed.

Lemma comp_morph0f V W (f : morph unit V W) :
        comp_morph (id_morph V) f =m f.
Proof.
split; first by move=>sw sv C; split; [case=>sv' [T ->] | exists sv].
move=>s t t' C H1 H2; split; first by case=>sv [tv][M ->].
by move=>M; case: H2=>s' T; exists s', t.
Qed.

Lemma comp_morphf0 V W (f : morph unit V W) :
        comp_morph f (id_morph W) =m f.
Proof.
split; first by move=>sw sv sv'; split; [case=>sw' [->] | exists sw].
move=>sw t t' C H1 H2; split; first by case=>sw' [tw][-> M ->].
by exists sw, t'.
Qed.


Lemma mxstep_eq V W (f1 f2 : morph unit V W) sw sw' :
        coh W sw -> f1 =m f2 ->
        mxstep f1 tt sw tt sw' -> mxstep f2 tt sw tt sw'.
Proof.
move=>Cw [E1 E2][tv I [tw][sv1][sv2]|t _ H]; last by apply: OtherMxStep H.
case=>Ef1 M1 M2 Hv Hw; move: (icoh Hw Cw)=>Cw'.
move/(E1 _ _ Cw): (M1)=>M1'; move/(E1 _ _ Cw'): (M2)=>M2'.
move/(E2 _ _ _ Cw (ex_intro _ _ M1) (ex_intro _ _ M1')): Ef1=>Ef2.
by apply: SelfMxStep I _; exists tw, sv1, sv2.
Qed.

Lemma mxsteps_eq V W (f1 f2 : morph unit V W) sw sw' :
        coh W sw -> f1 =m f2 ->
        mxsteps f1 tt sw tt sw' -> mxsteps f2 tt sw tt sw'.
Proof.
move=>C E [n]. elim: n sw sw' C =>[|n IH] sw sw' C /=.
- by case=>->; apply: iter_refl.
case; case; case=>s' [M] /IH; move/iter2: (mxstep_eq C E M)=>M'.
by move/(_ (omcoh (iter1 M) C)); apply: iter_trans M'.
Qed.


(* In the case of id morphism I can switch between mstep and mxstep *)

Lemma msmxx W s1 s2 :
        coh W s1 -> mstep (id_morph W tt) s1 s2 ->
        mxstep (id_morph W) tt s1 tt s2.
Proof.
move=>C. case=>[a _ I -> H|a H]; last by apply: OtherMxStep H.
by apply: SelfMxStep I _; exists a, s1, s2; do !split=>//; apply: icoh H C.
Qed.

Lemma msmx W s1 s2 :
        coh W s1 -> msteps (id_morph W tt) s1 s2 ->
        mxsteps (id_morph W) tt s1 tt s2.
Proof.
move=>C [n]; elim: n s1 C=>[|n IH] s1 C /=.
- by move=>->; apply: iter_refl.
case=>s [M H]; have [C' M'] : coh W s /\ mxsteps (id_morph W) tt s1 tt s.
- by split; [apply: mcoh (iter1 M) C | apply: iter1; apply: msmxx].
by apply: iter_trans M' (IH _ C' H).
Qed.

Lemma omind V W (f : morph unit V W) (P : st W -> Prop) :
       (forall s1 s2, mstep (f tt) s1 s2 -> coh W s1 -> P s1 -> P s2) ->
        forall s1 s2, msteps (f tt) s1 s2 -> coh W s1 -> P s1 -> P s2.
Proof. by move=>H; apply: iters_ind=>// s1 s2 /mcoh. Qed.

Lemma omind_f V W (f : morph unit V W) (C : st W -> Prop) (F : st W -> st W -> Prop) :
        (forall s, coh W s -> C s -> F s s) ->
        Transitive F ->
        (forall s1 s2, mstep (f tt) s1 s2 -> coh W s1 -> C s1 -> C s2 /\ F s1 s2) ->
        forall s1 s2, msteps (f tt) s1 s2 -> coh W s1 -> C s1 -> F s1 s2.
Proof.
move=>H1 H2 H3 s1 s2 O X1 X2; move: s1 s2 {X1 X2} O (conj X1 X2).
apply: iterf_ind=>//; first by move=>s []; apply: H1.
move=>s1 s2 O [X Y]; case: (H3 _ _ O X Y)=>??; do !split=>//.
by apply: mcoh (iter1 O) X.
Qed.

(* relax transitivity *)
Lemma omind_f' V W (f : morph unit V W) (C : st W -> Prop) (F : st W -> st W -> Prop) :
        (forall s, coh W s -> C s -> F s s) ->
        (forall s1 s2 s3, coh W s1 -> coh W s2 -> coh W s3 ->
                  F s1 s2 -> F s2 s3 -> F s1 s3) ->
        (forall s1 s2, mstep (f tt) s1 s2 -> coh W s1 -> C s1 -> C s2 /\ F s1 s2) ->
        forall s1 s2, msteps (f tt) s1 s2 -> coh W s1 -> C s1 -> F s1 s2.
Proof.
move=>H1 H2 H3 s1 s2 O X1 X2; case: O=>n.
elim: n s1 s2 X1 X2=>[|n IH] s1 s2 X1 X2 /=; first by move=><-; apply: H1.
case=>s [M H]; case: (H3 _ _ M X1 X2)=>Cs Fs.
have M' : msteps (f tt) s s2 by exists n.
move: (mcoh (iter1 M) X1)=>Cohs; apply: H2 (IH _ _  Cohs Cs H)=>//.
by apply: mcoh M' Cohs.
Qed.

Lemma omind_s V W (f : morph unit V W) (C : st W -> Prop) (F : st W -> st W -> Prop) :
        (forall s, coh W s -> C s -> F s s) ->
        Transitive F ->
        (forall s1 s2, msteps (f tt) s1 s2 -> coh W s1 -> C s1 -> C s2) ->
        (forall s1 s2, mstep (f tt) s1 s2 -> coh W s1 -> C s1 -> F s1 s2) ->
        forall s1 s2, msteps (f tt) s1 s2 -> coh W s1 -> C s1 -> F s1 s2.
Proof.
move=>H1 H2 H3 H4 s1 s2 O X1 X2; move: s1 s2 {X1 X2} O (conj X1 X2).
apply: iterf_ind=>//; first by move=>s []; apply: H1.
move=>s1 s2 O [X Y]; move: (H3 _ _ (iter1 O) X Y) (H4 _ _ O X Y)=>??.
by do !split=>//; apply: mcoh (iter1 O) X.
Qed.

(* relax transitivity *)
Lemma omind_s' V W (f : morph unit V W) (C : st W -> Prop) (F : st W -> st W -> Prop) :
        (forall s, coh W s -> C s -> F s s) ->
        (forall s1 s2 s3, coh W s1 -> coh W s2 -> coh W s3 ->
                          F s1 s2 -> F s2 s3 -> F s1 s3) ->
        (forall s1 s2, msteps (f tt) s1 s2 -> coh W s1 -> C s1 -> C s2) ->
        (forall s1 s2, mstep (f tt) s1 s2 -> coh W s1 -> C s1 -> F s1 s2) ->
        forall s1 s2, msteps (f tt) s1 s2 -> coh W s1 -> C s1 -> F s1 s2.
Proof.
move=>H1 H2 H3 H4 s1 s2 O X1 X2; case: O=>n.
elim: n s1 s2 X1 X2=>[|n IH] s1 s2 X1 X2 /=; first by move=><-; apply: H1.
case=>s [M H]; move: (H3 _ _ (iter1 M) X1 X2)=>X3; move: (H4 _ _ M X1 X2)=>X4.
move: (mcoh (iter1 M) X1)=>Cohs.
have M' : msteps (f tt) s s2 by exists n.
by apply: H2 (IH _ _ Cohs X3 H)=>//; apply: mcoh M' Cohs.
Qed.

(* relax transitivity some more *)
Lemma omind_s'' V W (f : morph unit V W) (C : st W -> Prop) (F : st W -> st W -> Prop) :
        (forall s, coh W s -> C s -> F s s) ->
        (forall s1 s2 s3, coh W s1 -> coh W s2 -> coh W s3 ->
                          C s1 -> C s2 -> C s3 ->
                          F s1 s2 -> F s2 s3 -> F s1 s3) ->
        (forall s1 s2, msteps (f tt) s1 s2 -> coh W s1 -> C s1 -> C s2) ->
        (forall s1 s2, mstep (f tt) s1 s2 -> coh W s1 -> C s1 -> F s1 s2) ->
        forall s1 s2, msteps (f tt) s1 s2 -> coh W s1 -> C s1 -> F s1 s2.
Proof.
move=>H1 H2 H3 H4 s1 s2 O X1 X2; case: O=>n.
elim: n s1 s2 X1 X2=>[|n IH] s1 s2 X1 X2 /=; first by move=><-; apply: H1.
case=>s [M H]; move: (H3 _ _ (iter1 M) X1 X2)=>X3; move: (H4 _ _ M X1 X2)=>X4.
move: (mcoh (iter1 M) X1)=>Cohs.
have M' : msteps (f tt) s s2 by exists n.
apply: H2 (IH _ _ Cohs X3 H)=>//.
- by apply: mcoh M' Cohs.
by apply: H3 M' Cohs X3.
Qed.

(* we don't seem to have the following lemma in general *)
(* because there is no connection between states and transitions *)
(* we need that connection as a side-condition *)
(* special case is the lemma msub_spval_comp below *)
(* where f2 is taken to be the sub-resource morphism *)
Lemma mstep_comp V1 V2 W (sw sw' : st W) (f1 : morph unit V1 V2) (f2 : morph unit V2 W) :
        mstep ((comp_morph f1 f2) tt) sw sw' ->
        exists sv sv', [/\ [morph f2] tt sw sv, [morph f2] tt sw' sv' &
                            mstep (f1 tt) sv sv'].
Proof.
Abort.

(***************************************************)
(* Some resource constructions                     *)
(* Experimental for now, not sure I will need them *)
(***************************************************)

(* Empty resource *)

Section EmptyRes.
Definition ecoh := emp_coh.
Definition einterp (t : False) := @emp_trans _ _ ecoh.

Lemma eres_int : res_internalP einterp (fun x => false).
Proof. by []. Qed.

Definition emp_resResourceMix := ResourceMixin eres_int.
Canonical emp_res := Eval hnf in Resource emp_resResourceMix.

End EmptyRes.


Section SubURes.
Variables W : resource.
Variables (P : Ux W -> Prop) (Q : Tp W -> Prop) (pc : pcm_closed P).
Variable pf : all_cohP W (fun s : st W => exists s1, s = ustval pc Q s1).

Definition scoh := usub_coh pc Q (coh W).

Definition sinterp t := usub_tr pc Q (interp W t).

Lemma sres_int : res_internalP sinterp (internal W).
Proof. by move=>t /iint H s s' C /(H _ _ C). Qed.

Definition ust_resResourceMix := ResourceMixin sres_int.
Canonical ust_res := Eval hnf in Resource ust_resResourceMix.

Definition ust_sigma (_ : unit) (s : st ust_res) (x : st W) := x = ustval pc Q s.
Definition ust_delta (_ : unit) (t1 : Tr W) (s : st ust_res) (t : Tr ust_res) :=
  t = t1.

Lemma ust_local : morph_localP ust_sigma.
Proof.
exists (fun x : Ux ust_res => sval x); case=>p s s' _ _ ->.
by exists (ustval pc Q s); rewrite /ust_sigma !ustE.
Qed.

Lemma ust_other : morph_otherP ust_sigma.
Proof. by case; case=>???? _ _ _ _ ->-> O; rewrite -!ustEQ0 O. Qed.

Lemma ust_func : morph_funcP ust_sigma.
Proof. by case=>??? _ _ _ ->->. Qed.

Lemma ust_backw : morph_backwardP ust_sigma.
Proof. by case; case. Qed.

Lemma ust_int : morph_internalP ust_delta.
Proof. by move=>t t' [s I] C ->. Qed.

Lemma ust_tfunc : morph_tfuncP ust_delta.
Proof. by case=>t sw _ _ C ->->. Qed.

Lemma ust_tlocal : morph_tlocalP ust_delta.
Proof. by []. Qed.

Lemma ust_coh : morph_cohP' ust_sigma.
Proof. by case=>sw _ Cw ->. Qed.

Definition ustMorphMix := MorphMixin ust_local ust_other
             ust_func ust_backw ust_int ust_tfunc ust_tlocal ust_coh.
Canonical ust_morph := Eval hnf in Morph ustMorphMix.

Lemma ust_ostep : morph_ostepP ust_morph (fun => xPredT).
Proof.
case=>sw sw' sv _ _ -> _ [t] /iosteps O; exists (ustval pc Q sw'); split=>//.
case: O=>n; elim: n sw sw'=>[|n IH] sw sw' /=; first by move=>->.
case=>s' [O] /IH; apply/iter_trans/iter1.
by case: O=>a; rewrite /= /usub_rel !ustE; apply: iostep.
Qed.

Lemma ust_forw : morph_forwardP ust_morph (fun => xPredT).
Proof.
move=>t [] sw _ sv I Cw _ -> _ H; case: (pf I Cw H) (H); first by exists sw.
by move=>sv' ->; exists t, tt, sv'.
Qed.

Canonical ust_sim : simulation ust_morph := Simulation2 ust_ostep ust_forw.

End SubURes.


(* 3rd attemp. Universal construction of taking a subPCM of PCM, but where *)
(* we insist that all transitions, including externals, preserve the subPCM *)

(* We express the preservation requirement using the persistence of the sep relation *)
(* which is in practice better than above approach where we require *)
(* preservation of spval *)

(* while the condition is better, proving forward simulation *)
(* requires that D is a seprel, rather than ordinary relation *)
(* this is fine, as we always have this property *)

Definition sub_persP W (D : rel (Ux W)) (P : st W -> Prop) :=
  forall t (s1 s2 : st W),
    coh W s1 ->
    (* the validity follows from coh, but I give it to myself explicitly here *)
    (* so that I don't have to massage it from coh in client proofs *)
    allvalid s1 ->
    allvalid s2 ->
    interp W t s1 s2 ->
    D (idfun.s s1) (idfun.o s1) -> P s1 -> D (idfun.s s2) (idfun.o s2) /\ P s2.

Module SubResource.
Section SubResource.
Variables (W : resource) (D : sep_rel (Ux W)) (P : st W -> Prop).
Variables (U : sub_pcm D) (gb_pf : cglobalP (coh W) P).
Variables (sp_pf : sub_persP D P).

Lemma all_pers t : sub_preserveP (interp W t) (coh W) U P.
Proof.
move=>s x C H [X Y]; apply: sp_pf (C) (coh_valid C) _ (H) X Y.
by apply: coh_valid (trans_coh H C).
Qed.

Let xcoh := sub_coh U gb_pf.
Let xinterp t : trans xcoh := sub_tr gb_pf (@all_pers t).

Lemma subres_int : res_internalP xinterp (internal W).
Proof. by move=>t /iint H s s' [A X C] /(H _ _ C). Qed.

Definition sub_resMix := ResourceMixin subres_int.
Canonical sub_res := Eval hnf in Resource sub_resMix.

(* sep_persP implies the following *)
(* that we need for forward simulation in sub_forw lemma *)
(* but this requires that D is a seprel *)
Lemma sub_pers : persP W (fun s : st W => P s /\
        exists s1 : st sub_res, allvalid s1 /\ s = spval s1).
Proof.
move=>t s s' C H [X][s1][V E]; rewrite {s}E in C H X.
have V1: allvalid (spval s1) by rewrite (coh_valid C).
have V2: allvalid s' by move/(@icoh W t): H; move/(_ C)/coh_valid.
have S : D (idfun.s (spval s1)) (idfun.o (spval s1)).
- by move: V; rewrite !allvalidE valid_sep !spEQ0 !lT; case/andP.
move: (sp_pf C V1 V2 H S X); rewrite -!lT; case=>Sx Xx.
split=>//; exists (State (psub U (self s')) (joint s') (psub U (other s'))).
split; first by rewrite allvalidE valid_psubU Sx andbT -allvalidE.
by rewrite allvalidE in V2; apply/stE'; rewrite spE !pval_psub
     ?(sep0E V2 Sx) ?(validE2 V2).
Qed.

Let sub_sigma (_ : unit) (s : st sub_res) (x : st W) := x = spval s.
Let sub_delta (_ : unit) (t : Tr W) (x : st sub_res) (t' : Tr sub_res) := t' = t.

Lemma sub_local : morph_localP sub_sigma.
Proof.
exists (fun x : Ux sub_res=>pval x); case=>p s s' /coh_valid V _ ->.
rewrite allvalidE !zgE in V; exists (spval s).
by rewrite /sub_sigma /= !spE //= (validLE3 V).
Qed.

Lemma sub_other : morph_otherP sub_sigma.
Proof. by case; case=>???? _ _ _ _ ->-> O; rewrite -!spEQ0 O. Qed.

Lemma sub_func : morph_funcP sub_sigma.
Proof. by case=>??? _ _ _ ->->. Qed.

Lemma sub_backw : morph_backwardP sub_sigma.
Proof. by case; case. Qed.

Lemma sub_int : morph_internalP sub_delta.
Proof. by move=>t t' [s I] C ->. Qed.

Lemma sub_tfunc : morph_tfuncP sub_delta.
Proof. by case=>t _ _ _ _ ->->. Qed.

Lemma sub_tlocal : morph_tlocalP sub_delta.
Proof. by []. Qed.

Lemma sub_coh : morph_cohP' sub_sigma.
Proof. by case=>sw _ [_ _ Cw] ->. Qed.

Definition sub_morphMix :=
  MorphMixin sub_local sub_other
             sub_func sub_backw sub_int sub_tfunc sub_tlocal sub_coh.
Canonical sub_morph := Eval hnf in Morph sub_morphMix.

Lemma sub_ostep : morph_ostepP sub_morph (fun => xPredT).
Proof.
case=>sw sw' _ _ _ -> _ [t] /iosteps O; exists (spval sw'); split=>//.
case: O=>n; elim: n sw sw'=>[|n IH] sw sw' /=; first by move=>->.
case=>s' [O] /IH; apply/iter_trans/iter1.
by case: O=>a; rewrite /= !spE; apply: iostep.
Qed.

Lemma sub_forw : morph_forwardP sub_morph (fun => xPredT).
Proof.
move=>t [] sw x sv I [A X C1] C2 E _ H.
rewrite /Morph.phi /= /sub_sigma /sub_delta in E *; subst x.
case: (sub_pers C2 H) (H)=>[|Sy [y][Ay E] G].
- by split=>//; exists sw; move/coh_valid: C1.
by exists t, tt, y; rewrite /= -E.
Qed.

Canonical sub_sim := Simulation2 sub_ostep sub_forw.

End SubResource.

Module Exports.
Notation sub_resMix := sub_resMix.
Notation sub_res := sub_res.
Canonical sub_res.
Notation sub_morphMix := sub_morphMix.
Notation sub_morph := sub_morph.
Canonical sub_morph.
Notation sub_sim := sub_sim.

(* Are there a lemma to be proved about osteps and msteps of the two resources? *)
Section OstepsMstepsLemmas.
Variables (W : resource) (D : sep_rel (Ux W)) (P : st W -> Prop).
Variables (U : sub_pcm D) (pf : cglobalP (coh W) P).
Variables (qf : sub_persP D P).
Notation sub_res := (sub_res U pf qf).
Notation sub_morph := (sub_morph U pf qf).

Lemma osub_spval' s1 s2 : ostep sub_res s1 s2 -> ostep W (spval s1) (spval s2).
Proof. by case=>a; rewrite /= !spE; apply: iostep. Qed.

Lemma msub_spval' s1 s2 :
        mstep (sub_morph tt) s1 s2 ->
        mstep (id_morph W tt) (spval s1) (spval s2).
Proof.
case=>a; last by move/iostep/osub_spval'; apply: om.
by move=>_  /= I ->; apply: SelfMStep I _.
Qed.

Lemma osub_spval s1 s2 : osteps sub_res s1 s2 -> osteps W (spval s1) (spval s2).
Proof.
case=>n; elim: n s1 s2=>[|n IH] s1 s2 /=; first by move=>->.
by case=>s [/osub_spval' O] /IH; apply: iter_trans (iter1 O).
Qed.

Lemma msub_spval s1 s2 :
        msteps (sub_morph tt) s1 s2 ->
        msteps (id_morph W tt) (spval s1) (spval s2).
Proof.
case=>n; elim: n s1 s2=>[|n IH] s1 s2 /=; first by move=>->.
by case=>s [/msub_spval' O] /IH; apply: iter_trans (iter1 O).
Qed.

(* we actually have a more general lemma for mstep *)
(* whereby we can compose morphisms *)
Lemma msub_spval_comp R s1 s2 (f : morph unit R W):
        msteps (comp_morph f sub_morph tt) s1 s2 ->
        msteps (f tt) (spval s1) (spval s2).
Proof.
move=>M; move: s1 s2 M I; apply: iters_ind=>// s1 s2.
- by move=>?; apply: iter_trans.
case; last by move=>t /=; rewrite !spE; move/iostep/iter1/oms.
move=>t /= ft I [_][t2][-> F S] H _; rewrite /sub_morph /= in S; subst t2.
by apply: iter1; apply: SelfMStep I F H.
Qed.

(* otoh, for the other direction, we need *)
(* coherence of s1, but this is not severe *)
Lemma interp_sub t s1 s2' :
        interp W t (spval s1) s2' -> coh sub_res s1 ->
        exists s2, s2' = spval s2 /\ coh sub_res s2.
Proof.
move=>H [V X C]; case: (sub_pers (U:=U) (gb_pf:=pf) (sp_pf:=qf) C H).
- by split=>//; exists s1.
move=>Sx [x][Vx E]; subst s2'; exists x; split=>//.
by split=>//; apply: icoh H C.
Qed.

Lemma ospval_subX' s1 s2' :
        ostep W (spval s1) s2' -> coh sub_res s1 ->
        exists s2, s2' = spval s2 /\ ostep sub_res s1 s2.
Proof.
case=>a H /coh_tr C1; rewrite -spE in H.
case: (interp_sub H C1)=>s2 [E C2]; exists (tr s2).
rewrite spE -E trI; split=>//; apply: (@iostep sub_res _ _ a).
by rewrite /= trI -E.
Qed.

Lemma mspval_subX' s1 s2' :
        mstep (id_morph W tt) (spval s1) s2' -> coh sub_res s1 ->
        exists s2, s2' = spval s2 /\ mstep (sub_morph tt) s1 s2.
Proof.
case=>a; last by move/iostep/ospval_subX'=>H /H [s2][->] {H} /om; exists s2.
move=>_ I -> H /(interp_sub H) [s2][E] C; subst s2'.
by exists s2; split=>//; apply: SelfMStep I _ _.
Qed.

Lemma ospval_subX s1 s2' :
        osteps W (spval s1) s2' -> coh sub_res s1 ->
        exists s2, [/\ s2' = spval s2, osteps sub_res s1 s2 & coh sub_res s2].
Proof.
case=>n; elim: n s1 s2'=>[|n IH] s1 s2' /=; first by move=><-; exists s1.
case=>s [O H] C; case: (ospval_subX' O C) H=>x {O} [->] O1.
case/IH/(_ (ocoh (iter1 O1) C))=>s2 [->] O2.
by exists s2; split=>//; apply: iter_trans (iter1 O1) O2.
Qed.

Lemma mspval_subX s1 s2' :
        msteps (id_morph W tt) (spval s1) s2' -> coh sub_res s1 ->
        exists s2, [/\ s2' = spval s2, msteps (sub_morph tt) s1 s2 & coh sub_res s2].
Proof.
case=>n; elim: n s1 s2'=>[|n IH] s1 s2' /=; first by move=><-; exists s1.
case=>s [O H] C; case: (mspval_subX' O C) H=>x {O} [->] O1.
case/IH/(_ (mcoh (iter1 O1) C))=>s2 [->] O2.
by exists s2; split=>//; apply: iter_trans (iter1 O1) O2.
Qed.

Lemma ospval_sub s1 s2 :
        osteps W (spval s1) (spval s2) ->
        coh sub_res s1 -> osteps sub_res s1 s2.
Proof. by move=>O /(ospval_subX O) [_][/sp_inj <-]. Qed.

Lemma mspval_sub s1 s2 :
        msteps (id_morph W tt) (spval s1) (spval s2) ->
        coh sub_res s1 -> msteps (sub_morph tt) s1 s2.
Proof. by move=>O /(mspval_subX O) [_][/sp_inj <-]. Qed.

(* combining ospval_sub1 and ospval_sub *)
Lemma ospval_subX1 t s1 s2' s3 :
        osteps W s2' (spval s3) -> coh sub_res s1 ->
        interp W t (spval s1) s2' ->
        exists s2, [/\ s2' = spval s2, osteps sub_res s2 s3 & coh sub_res s2].
Proof.
move=>O C H; case: (interp_sub H C)=>x [E] C'.
by exists x; split=>//; apply: ospval_sub C'; rewrite -E.
Qed.

Lemma mspval_subX1 t s1 s2' s3 :
        msteps (id_morph W tt) s2' (spval s3) -> coh sub_res s1 ->
        interp W t (spval s1) s2' ->
        exists s2, [/\ s2' = spval s2, msteps (sub_morph tt) s2 s3 & coh sub_res s2].
Proof.
move=>O C H; case: (interp_sub H C)=>x [E] C'.
by exists x; split=>//; apply: mspval_sub C'; rewrite -E.
Qed.

(* a massaging of sep_persP *)
Lemma ost_sep s1 s2 :
        osteps W s1 s2 -> coh W s1 ->
        D (idfun.s s1) (idfun.o s1) -> P s1 ->
        D (idfun.s s2) (idfun.o s2) /\ P s2.
Proof.
move=>O C X Y; move: s1 s2 O C {X Y}(conj X Y).
apply: iters_ind=>// s1 s2; first by apply: ocoh.
move=>O C; move: (ocoh (iter1 O) C)=>C2.
case: O C C2=>a /qf H /coh_tr C /coh_tr C2 [S X].
move: (coh_validT C); rewrite !trE0=>V1.
move: (coh_validT C2); rewrite !trE0=>V2.
move: (H C (coh_valid C) (coh_valid C2)).
rewrite -!lT !trE0 sepC // !lT !(cglob_tr pf) -1?coh_tr //.
by move/(_ S X); rewrite sepC // -!lT.
Qed.

End OstepsMstepsLemmas.

Arguments interp_sub [W D P U pf qf] t [s1 s2'] H _.
Arguments ospval_subX1 [W D P U pf qf] t [s1 s2' s3] O C H.
Arguments mspval_subX1 [W D P U pf qf] t [s1 s2' s3] O C H.

End Exports.
End SubResource.

Export SubResource.Exports.


(* If we just have an a predicate I stable under f-msteps *)
(* it doesn't suffice to build a morphims *)

Section MorphInv.
Variables (V W : resource) (f : morph unit V W) (I : st W -> Prop).
Variable pf : forall s s', I s -> msteps (f tt) s s' -> I s'.

Definition sphi (_ : unit) (sw : st W) (sv : st V) := [morph f] tt sw sv /\ I sw.
Definition stau (_ : unit) (t : Tr V) := f tt t.

End MorphInv.


(* automation of oself *)

Module OselfAuto.
Section OselfAuto.

Structure tagged_state U T := Tag {untag : state U T}.
Definition spval_tag U T (s : state U T) := @Tag U T s.
Definition ssnd_tag U T (s : state U T) := spval_tag s.
Definition sfst_tag U T (s : state U T) := ssnd_tag s.
Canonical found_tag U T (s : state U T) := sfst_tag s.

Local Coercion untag : tagged_state >-> state.

(* stability when appearing under self *)

Definition axiom W U T (f : st W -> state U T) pivot s :=
 untag pivot = f s /\ forall i j, osteps W i j -> self (f i) = self (f j).

Structure self_stable W U T (f : st W -> state U T) (s : st W) :=
  SS {pivot : tagged_state U T; _ : axiom f pivot s}.

Local Coercion pivot : self_stable >-> tagged_state.

(* base case; identify function *)
Lemma found_ax W (s : st W) : axiom (fun x => x) (found_tag s) s.
Proof. by split=>// i j /oself. Qed.
Canonical found_form W (s : st W) := SS (found_ax s).

(* sfst case *)
Lemma sfst_ax W U1 U2 T1 T2 (f : st W -> state (prodPCM U1 U2) (T1 * T2))
                            (F : forall s, self_stable f s) s :
        axiom (fun s => (f s)\1) (sfst_tag (F s)\1) s.
Proof. by case: (F s)=>p [-> H]; split=>// i j /H E; rewrite -!prodEQ0 E. Qed.

Canonical sfst_form W U1 U2 T1 T2 f F s := SS (@sfst_ax W U1 U2 T1 T2 f F s).

(* ssnd case *)
Lemma ssnd_ax W U1 U2 T1 T2 (f : st W -> state (prodPCM U1 U2) (T1 * T2))
                            (F : forall s, self_stable f s) s :
        axiom (fun s => (f s)\2) (ssnd_tag (F s)\2) s.
Proof. by case: (F s)=>p [-> H]; split=>// i j /H E; rewrite -!prodEQ0 E. Qed.

Canonical ssnd_form W U1 U2 T1 T2 f F s := SS (@ssnd_ax W U1 U2 T1 T2 f F s).

(* spval case *)

Lemma spval_ax W V D (U : @sub_pcm V D) T (f : st W -> state U T)
                     (F : forall s, self_stable f s) s :
        axiom (fun s => spval (f s)) (spval_tag (spval (F s))) s.
Proof. by case: (F s)=>p [-> H]; split=>// i j /H E; rewrite -!spEQ0 E. Qed.

Canonical spval_form W V D U T f F s := SS (@spval_ax W V D U T f F s).

End OselfAuto.

Module Exports.
Coercion untag : tagged_state >-> state.
Coercion pivot : self_stable >-> tagged_state.
Canonical found_tag.
Canonical found_form.
Canonical sfst_form.
Canonical ssnd_form.
Canonical spval_form.

Lemma oselfR W U T (f : _ -> state U T) (F : forall s, self_stable f s) i j :
        osteps W i j -> self (F i) = self (F j).
Proof. by case: (F i) (F j)=>_ [-> _][_ [->]]; apply. Qed.

Arguments oselfR [W U T f F i j].
Notation oselfW W O := (@oselfR W _ _ _ _ _ _ O).

End Exports.

End OselfAuto.

Export OselfAuto.Exports.

Lemma osf W (V : pcm) (f : Ux W -> V) s1 s2 :
        osteps W s1 s2 -> f.s s1 = f.s s2.
Proof. by move=>O; rewrite !lbX (oself O). Qed.

Arguments osf [W V f s1 s2] _.

Definition osfw := osf.
Arguments osfw W [V f s1 s2] _.
