From Coq Require Import ssreflect ssrbool ssrfun.
From fcsl Require Import axioms prelude pcm morphism.
From fcsl Require Import options.

Definition allvalid (U : pcm) (s : U * U) := locked (valid (s.1 \+ s.2)).

Lemma allvalid_unlock (U : pcm) (s : U * U) : allvalid s = valid (s.1 \+ s.2).
Proof. by rewrite /allvalid -lock. Qed.

(* we hide the definition of the two applications *)
(* to be able to apply custom rewrite laws later on *)
Module Type Private_NotationSig.
Section Private_NotationSig.
Variables (U : pcm) (V : Type).
Parameter so_app : (U -> V) -> U * U -> V.
Parameter os_app : (U -> V) -> U * U -> V.
Axiom so_appE : forall f s, so_app f s = f s.1.
Axiom os_appE : forall f s, os_app f s = f s.2.
Axiom app_lock : forall A B, (A -> B) -> A -> B.
Axiom app_lockE : forall A B (f : A -> B) x, app_lock f x = f x.
End Private_NotationSig.
End Private_NotationSig.

Module Private_Notation : Private_NotationSig.
Section Private_Notation.
Variables (U : pcm) (V : Type).
Definition so_app (f : U -> V) (s : U * U) := f s.1.
Definition os_app (f : U -> V) (s : U * U) := f s.2.
Definition so_appE f s := erefl (so_app f s).
Definition os_appE f s := erefl (os_app f s).
Definition app_lock A B (f : A -> B) x := f x.
Definition app_lockE A B f x := erefl (@app_lock A B f x).
End Private_Notation.
End Private_Notation.

Notation "f '.s' x" := (Private_Notation.so_app f x) (at level 10,
  format "f '.s'  x") : fun_scope.
Notation "f '.o' x" := (Private_Notation.os_app f x) (at level 10,
  format "f '.o'  x") : fun_scope.
Notation "f '.t' x" := (f.s x \+ f.o x) (at level 10,
  format "f '.t'  x") : pcm_scope.

(* Coq doesn't respect the order in multi-rewrite rules *)
(* because it favors literal matches. In particular, no insertions *)
(* of coercions and canonical projections. *)
(* This gives me trouble, so I need a hack *)
(* to postpone removing the application lock *)

Lemma app_lockE (U V : pcm) (D : rel U) (f : {morphism D >-> V}) x :
        Private_Notation.app_lock f x = f x.
Proof. by rewrite Private_Notation.app_lockE. Qed.

Lemma app_lockE' U V (f : U -> V) x : Private_Notation.app_lock f x = f x.
Proof. by rewrite Private_Notation.app_lockE. Qed.

(* application laws *)

Section LabelSOLaws.
Variables (U : pcm) (V V1 V2 : Type) (f : U -> V) (f1 : U -> V1) (f2 : U -> V2).

(* we don't automatically rewrite with the first two *)
(* they expand the definitions of .s and .o *)
(* so should be used only in libraries, not in clients *)
Lemma lSX (s : U * U) : (f.s s = f s.1) * (f.o s = f s.2).
Proof. by rewrite Private_Notation.so_appE Private_Notation.os_appE. Qed.

(* we expand automatically when f is applied to a pair... *)
(* but guard the expansion with a lock in order to facilitate *)
(* a further check that the argument is of a special kind *)
(* the lock will be removed as a last step of rewriting *)
Lemma lSso (s o : U) :
        (f.s (s, o) = Private_Notation.app_lock f s) *
        (f.o (s, o) = Private_Notation.app_lock f o).
Proof. by rewrite !lSX !Private_Notation.app_lockE. Qed.

(* and one that replaces self/other with idfun *)
Lemma lSXid (s : U * U) :
        (f.s s = f (idfun.s s)) * (f.o s = f (idfun.o s)).
Proof. by rewrite !Private_Notation.so_appE !Private_Notation.os_appE. Qed.

(* label applying products *)
Lemma lSprod (s : U * U) :
        ((f1 \** f2).s s = (f1.s s, f2.s s)) *
        ((f1 \** f2).o s = (f1.o s, f2.o s)).
Proof. by rewrite !Private_Notation.so_appE !Private_Notation.os_appE. Qed.

End LabelSOLaws.


Section LabelTLaws.
Variables (U : pcm) (V V1 V2 : pcm) (f : U -> V) (f1 : U -> V1) (f2 : U -> V2).

Lemma lTso (s o : U) : f.t (s, o) =
        Private_Notation.app_lock f s \+ Private_Notation.app_lock f o.
Proof. by rewrite !lSso. Qed.

Lemma lTXid (s : U * U) : f.t s = f (idfun.s s) \+ f (idfun.o s).
Proof. by rewrite lSXid joinC lSXid joinC. Qed.

Lemma lTprod (s : U * U) :
        (f1 \** f2).t s = (f1.t s, f2.t s).
Proof. by rewrite !lSprod pcmPE. Qed.

End LabelTLaws.


(* label laws for interacting with all *)
(* NOTE: in the first lemma, it suffices that f : U -> V, not a morphism *)
(* but such rewrite rule is too eager and ends up disturbing the goals *)

Lemma lSA1_morphism (U V : pcm) (D : rel U) (f : {morphism D >-> V}) (s : U * U) :
        (f (idfun.s s) = f.s s) * (f (idfun.o s) = f.o s).
Proof. by rewrite !lSX. Qed.

Definition lSA := (lSA1_morphism).

Lemma lTA1_morphism (U V : pcm) (D : rel U) (f : {morphism D >-> V}) (s : U * U) :
        allvalid s -> D (idfun.s s) (idfun.o s) ->
        f (idfun.t s) = f.t s.
Proof.
move=>W S; rewrite pfjoin ?lSA // !lSX.
by rewrite /allvalid -lock in W.
Qed.

Definition lTA := (lTA1_morphism).

(* Though we mostly want to use these lemmas in Private_Notation versions *)
Lemma lSApply (U : pcm) V (f : U -> V) (s : U * U) :
        (Private_Notation.app_lock f (idfun.s s) = f.s s) *
        (Private_Notation.app_lock f (idfun.o s) = f.o s).
Proof. by rewrite !Private_Notation.app_lockE !lSX. Qed.

Lemma lTApply (U V : pcm) (D : rel U) (f : @morphism U V D) (s : U * U) :
        allvalid s -> D (idfun.s s) (idfun.o s) ->
        Private_Notation.app_lock f (idfun.t s) = f.t s.
Proof. by move=>W S; rewrite Private_Notation.app_lockE lTA. Qed.

(* this one not for automatic rewriting *)
Lemma allvE (U : pcm) (s : U * U) : allvalid s = valid (idfun.t s).
Proof. by rewrite /allvalid -lock !lSX. Qed.

(* and the implication *)
Lemma allvI (U : pcm) (s : U * U) : allvalid s -> valid (idfun.t s).
Proof. by rewrite allvE. Qed.

(* rewriting of \f.s and \f.o *)
Definition lSE' := (lSso, (* lSsw,*) lSprod).
(* rewriting of \f.t *)
Definition lTE' := (lTso, (* lTsw,*) lTprod).

(* rewrite without pfjoin; lTE before lSE because TE overrides s/o order *)
Definition lE := ((lTE', lSE'), (lTApply, lSApply), app_lockE).

(* rewrite with pfjoin *)
Definition lE1 := (lE, pfjoin).

(* interaction with all *)
(* lJA rewrites too much *)
Definition lA := (lTA, lSA, (* lJA, *) idfun0E).

(* lemmas for reindexing, i.e., automatically rewriting *)
(* projections out of composition *)
Section CompositionGenericsSO.
Variables (U1 U2 U V W : pcm).

Lemma lCSX (g : V -> W) (f : U -> V) x :
        (g (f.s x) = (g \o f).s x) *
        (g (f.o x) = (g \o f).o x).
Proof. by rewrite !lSX. Qed.

Lemma lCS1 (g : V -> W) x :
        (g (idfun.s x) = g.s x) *
        (g (idfun.o x) = g.o x).
Proof. by rewrite !lSX. Qed.

Lemma lprodScomp (f : _ -> U1 * U2) (s : U * U) :
        (((f.s s).1 = (fst \o f).s s) * ((f.o s).1 = (fst \o f).o s)) *
        (((f.s s).2 = (snd \o f).s s) * ((f.o s).2 = (snd \o f).o s)).
Proof. by rewrite !lSX. Qed.

End CompositionGenericsSO.

Section CompositionGenericsT.
Variables (U : pcm) (V W U1 U2 : pcm).

Lemma lCTX (D : rel V) (g : @morphism V W D) (f : U -> V) x :
        valid (f.t x) -> D (f.s x) (f.o x) -> g (f.t x) = (g \o f).t x.
Proof. by rewrite !lSX=>L S; rewrite pfjoin. Qed.

Lemma lCT1 (D : rel V) (g : @morphism V W D) x :
        valid (idfun.t x) -> D (idfun.s x) (idfun.o x) -> g (idfun.t x) = g.t x.
Proof. by rewrite !lSX=>L S; rewrite pfjoin. Qed.

(* need some phantoms on lCTX lCT1, so when I have to provide g's *)
(* explicitly, I can provide only the functions, not the morphisms *)

Lemma lCTX_phant (D : rel V) (g : @morphism V W D) of phantom (_ -> _) g :
        forall (f : U -> V) x,
          valid (f.t x) -> D (f.s x) (f.o x) -> g (f.t x) = (g \o f).t x.
Proof. by apply: lCTX. Qed.

Lemma lCT1_phant (D : rel V) (g : @morphism V W D) of phantom (_ -> _) g :
        forall x,
          valid (idfun.t x) -> D (idfun.s x) (idfun.o x) -> g (idfun.t x) = g.t x.
Proof. by apply: lCT1. Qed.

Lemma lprodTcomp (f : _ -> U1 * U2) (s : U * U) :
        ((f.t s).1 = (fst \o f).t s) * ((f.t s).2 = (snd \o f).t s).
Proof. by rewrite !lSX !pcmPE. Qed.

End CompositionGenericsT.

Notation lCTXf g := (@lCTX_phant _ _ _ _ _ (Phantom (_ -> _) g)).
Notation lCT1f g := (@lCT1_phant _ _ _ _ (Phantom (_ -> _) g)).

Definition lC1 := (lCS1, lCT1).



(* Afterthoughts *)

(* Turns out we also need the following notation, as insisting *)
(* on _.s, _.o, and _.t *always* requires a lot of glue proofs by *)
(* lCS1, lCT1, etc. *)

(* first some helper notation *)

Notation "f '.l' x" := (f (idfun.s x)) (at level 15,
  format "f '.l'  x") : fun_scope.
Notation "f '.r' x" := (f (idfun.o x)) (at level 15,
  format "f '.r'  x") : fun_scope.
Notation "f '.x' x" := (f.l x \+ f.r x) (at level 15,
  format "f '.x'  x").

Lemma lL (U V : pcm) (f : U -> V) x : f.l x = f.s x.
Proof. by rewrite lCS1. Qed.

Lemma lR (U V : pcm) (f : U -> V) x : f.r x = f.o x.
Proof. by rewrite lCS1. Qed.

Lemma lX (U V : pcm) (f : U -> V) x : f.x x = f.t x.
Proof. by rewrite !lCS1. Qed.

Definition lV := (lX, lL, lR).
